/**************************************************************************************************************
*
*       This file,
*
*                            /Air.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Air.h"
#include "fileio.h"
#include "Atmosphere.h"
#include "error.h"

Air::Air()
{
	cs_data=0;
	RRSEvent=0;

	IOName="air";
	setProfileName("air");
	
	b_isotropic_molecular_polarizability=false;
	b_RRS=false;
	
	b_consider_Herzberg_continuum=false;
	RCSFormula=RCS_vanDeelen;

	RRS_JMax=50;

	addParameter(fn_air_data, type_file, "fn_air_data", 0, "air data filename. 2 columns: pressure [hPa] TAB temperature [K]");
	addParameter(fn_air_map, type_file, "fn_air_map", 0, "air map filename");
	addParameter(&b_RRS, type_bool, "b_consider_rotational_raman_scattering|b_RRS|b_Ring_effect", 0, "consider rotational Raman scattering for the simulation of the Ring effect?");
	addParameter(&RRS_JMax, type_int, "RRS_JMax|J_max", 0, "maximum rotational quatum number considered");
	addParameter(&RCSFormula, type_enum, "RCSFormula", "Penndorf|Nicolet|BodhaineEq29|Bodhaine|vanDeelen|file", "Rayleigh cross section formula to use.");
	addParameter(&b_isotropic_molecular_polarizability, type_bool, "b_isotropic_molecular_polarizability", 0, "disable depolarization effects in the Rayleigh phase function?");
	addParameter(fn_rayleigh_cross_section, type_file, "fn_rayleigh_cross_section", 0, "external Rayleigh cross section file. 2 columns, wavelength [nm] TAB cs [cm^2]");
	addParameter(&b_consider_Herzberg_continuum, type_bool, "b_consider_Herzberg_continuum", 0, "consider an analytic description of the Herzberg continuum in the UV?");
	addParameter(&b_emission, type_bool, "b_emission", 0, "simulate spontaneous emission?");
	addParameter((Profile *)this, type_profile, "air", 0, "air profile data, usage: air[cluster index,i] where i=1: pressure [hPa], i=2: temperature [K]");

	RayleighPhF.setMoleculePtr(&Molecule);
	CabannesPhF.setMoleculePtr(&Molecule);
	RRPhaseFunction.setMoleculePtr(&Molecule);
	Molecule._vmr()=1.0;	// average air :)

	b_emission=false;
}

Air::~Air()
{
	if(cs_data)
		delete [] cs_data;
	if(RRSEvent)
		delete [] RRSEvent;
}

/*double Air::calculateRefractiveIndexEdlen(const double& wl)
{
	double l2=1e12*wl*wl;
	double nm1e8=6432.8+2949810.0/(146.0-l2)+25540.0/(41.0-l2);
	return 1.0+nm1e8*1e-8;
}

double Air::calculateRefractiveIndexStumpe(const double& wl, const double& p, const double& T,const double& RH)
{
	double rho,ret,b1;
	// wl:	wavelength
	// p:	pressure
	// T:	absolute temperature
	// RH:	relative humidity
	// rho: 	vapour pressure of water
	// formula: see tracydoc refraction index (Stumpe formula)
	double a[]={
	2371.34, 83939.7, 130, 4547.3, 38.99, 57.90e-8,
	9.3250e-4, 0.25844, 6487.31, 58.058, 0.71150,
	0.08851, 1.037, 2.3321e-3, 2.23366, 710.792, 7.75141e4,
	10474, 116.43, 0.43284, 7.75151e4};
	rho=RH/100.0*(-a[17]+a[18]*T-a[19]*T*T+a[20]*T*T*T);
	ret=(a[0]+a[1]/(a[2]-1/(wl*wl))+a[3]/(a[4]-1/(wl*wl)))*p/T;
	ret*=(1+p*(a[5]-a[6]/T+a[7]/(T*T)));
	b1=a[8]+a[9]/(wl*wl)+a[10]*pow(wl,-4)+a[11]*pow(wl,-6);
	ret+=b1*rho/T*(1+a[12]*rho*(-a[13]+a[14]/T-a[15]/(T*T)+a[16]/(T*T*T)))+1;
	return ret;
}
*/

double Air::calculateRefractiveIndexCiddor(const double& wl, const double& p_hp, const double& T,const double& RH, const double& x_CO2)
{
	double p=1e2*p_hp;
/*
	wl:	wavelength			[m]
	p_hp:	pressure			[hPa]
	T:	absolute temperature		[K]
	RH:	relative humidity		[%]
	x_CO2: 	number density mixing ratio in 	[ppm]
*/

/*
	these formulas had been taken from 
	http://emtoolbox.nist.gov/Wavelength/Documentation.asp
	the last one is called 

						Ciddor-Equation

	[thanks to Jack A. Stone and Jay H. Zimmerman for their excellent work]
*/
	
/* 
	#######################################################################
	#	Calculating Saturation Vapor Pressure, taken from                 #
	#	http://emtoolbox.nist.gov/Wavelength/Documentation.asp#AppendixAI #
        #######################################################################
*/
	double p_snu;
	if(T>273.15)
	{
		//over water
		const double K_1=1.16705214528E+03;
		const double K_2=-7.24213167032E+05;
		const double K_3=-1.70738469401E+01;
		const double K_4=1.20208247025E+04;
		const double K_5=-3.23255503223E+06;
		const double K_6=1.49151086135E+01;
		const double K_7=-4.82326573616E+03;
		const double K_8=4.05113405421E+05;
		const double K_9=-2.38555575678E-01;
		const double K_10=6.50175348448E+02;
		double Omega=T+K_9/(T-K_10);							//(A2)
		double A=Omega*Omega+K_1*Omega+K_2;						//(A3)
		double B=K_3*Omega*Omega+K_4*Omega+K_5;					//(A4)
		double C=K_6*Omega*Omega+K_7*Omega+K_8;					//(A5)
		double X=-B+sqrt(B*B-4*A*C);							//(A6)
		p_snu=1e6*pow(2*C/X,4.0);								//(A7)
	}
	else
	{
		//over ice
		const double A_1=-13.928169;							//(A8)
		const double A_2=34.7078238;							//(A9)
		double Phi=T/273.16;									//(A11)
		double Y=A_1*(1-pow(Phi,-1.5))+A_2*(1-pow(Phi,-1.25));	//(A12)
		p_snu=611.657*exp(Y);									//(A13)
	}
/*  #######################################################################
	#			Determining Humidity, taken from						  #
	# http://emtoolbox.nist.gov/Wavelength/Documentation.asp#AppendixAII  #
    #######################################################################		*/
	//calculate the enhancement factor f
	const double alpha=1.00062;									//(A17)
	const double beta=3.14e-8;									//(A17)
	const double gamma=5.6e-7;									//(A17)
	double t=T-273.15;
	double f=alpha+beta*p+gamma*t*t;									//(A16)
	//mole fraction calculation
	double x_nu=(RH/100.0)*f*p_snu/p;							//(A19)
/*  #######################################################################
	#		Ciddor Calculation of Index of Refraction, taken from		  #
	# http://emtoolbox.nist.gov/Wavelength/Documentation.asp#AppendixAIII #
    #######################################################################		*/	
	const double omega_0=295.235;								//(A21)
	const double omega_1=2.6422;								//(A21)
	const double omega_2=-0.03238;								//(A21)
	const double omega_3=0.004028;								//(A21)
	const double k_0=238.0185;									//(A22)
	const double k_1=5792105.0;									//(A22)
	const double k_2=57.362;									//(A22)
	const double k_3=167917.0;									//(A22)
	const double a_0=1.58123e-6;								//(A23)
	const double a_1=-2.9331e-8;									//(A23)
	const double a_2=1.1043e-10;								//(A23)
	const double b_0=5.707e-6;									//(A24)
	const double b_1=-2.051e-8;									//(A24)
	const double c_0=1.9898e-4;									//(A25)
	const double c_1=-2.376e-6;									//(A25)
	const double d=1.83e-11;									//(A26)
	const double e=-0.765e-8;									//(A26)
	const double p_R1=101325.0;									//(A27)
	const double T_R1=288.15;									//(A27)
	const double Z_a=0.9995922115;								//(A28)
	const double rho_nus=0.00985938;							//(A29)
	const double R=8.314472;									//(A30)
	const double M_nu=0.018015;									//(A30)
	double wl_micro=1e6*wl;
	double S=1.0/(wl_micro*wl_micro);							//(A31)
	double r_as=1e-8*(k_1/(k_0-S)+k_3/(k_2-S));					//(A32)
	double r_nus=1.022e-8*(omega_0+omega_1*S+omega_2*S*S+omega_3*S*S*S);				//(A33)
	//x_CO2 in micromol/mol
	double M_a=0.0289635+1.2011e-8*(x_CO2-400.0);				//(A34)
	double r_axs=r_as*(1+5.34e-7*(x_CO2-450.0));				//(A35)
	double Z_m=1.0-(p/T)*(a_0+a_1*t+a_2*t*t+(b_0+b_1*t)*x_nu+(c_0+c_1*t)*x_nu*x_nu)+(p/T)*(p/T)*(d+e*x_nu*x_nu);	//(A37)
	double rho_axs=p_R1*M_a/(Z_a*R*T_R1);						//(A38)
	double rho_nu=x_nu*p*M_nu/(Z_m*R*T);						//(A39)
	double rho_a=(1.0-x_nu)*p*M_a/(Z_m*R*T);					//(A40)
	double n=1.0+(rho_a/rho_axs)*r_axs+(rho_nu/rho_nus)*r_nus;	//(A41)
	return n;	

}

void Air::setClusterDimensions(int input_cluster_dimension)
{
	cluster_dimension=input_cluster_dimension;
	auxiliary_data_dimension=0;
	optical_cluster_dimension=no_L1_elements;
}

bool Air::loadFiles()
{
	char filename[BasicIO_string_size];

	// determine number of gases
	no_gases=pAtmosphere->getNOListedGases();

	// cluster and map data
	pAtmosphere->formatAirDataFilepath(filename);
	if(!loadClusterData(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load air cluster data from file \"%s\"!\n", filename);
		return false;
	}
	pAtmosphere->formatAirMapFilepath(filename);
	RayleighPhF.enableUseIsotropicMolecularPolarizability(b_isotropic_molecular_polarizability);
	if(!ClusterMap::load(filename) || !initializeClusterArray())
	{
		fprintf(stderr, "failed to load and initialize air map data from file \"%s\"!\n", filename);
		return false;
	}
	if(getClusterDimension()<2)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "Air profile in file \"%s\" has wrong dimension! ", pAtmosphere->getAirDataFilename());
		fprintf(stderr, "Found %d instead of at least 2 for pressure and temperature!\n",  getClusterDimension());
		return false;
	}
	/*if(RCSFormula==RCS_Bodhaine && getClusterDimension()!=3 && getClusterDimension()!=4)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "Air profile in file \"%s\" has wrong dimension! ", pAtmosphere->getAirDataFilename());
		fprintf(stderr, "Refractive index calculation after Bodhaine requires: p,T,RH and vmr_CO2[ppm]\n");
		return false;
	}*/
	// load Rayleigh cross section if necessary
	if(RCSFormula==RCS_file)
	{
		pAtmosphere->formatRayleighCrossSectionFilepath(filename);
		FILE *fptr=fopen(filename, "r");
		if(!fptr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not load Rayleigh cross section from file \"%s\"!\n", filename);
			return false;
		}
		no_Rayleigh_cross_sections=numberOfRows(fptr);
		if(cs_data)
			delete [] cs_data;
		cs_data=new double[2*no_Rayleigh_cross_sections];
		if(!cs_data)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not allocate %d Rayleigh cross sections according to file \"%s\"!\n",\
				no_Rayleigh_cross_sections, filename);
			fclose(fptr);
			return false;
		}
		lambda_Rayleigh=cs_data;
		cs_Rayleigh=cs_data+no_Rayleigh_cross_sections;
		int i;
		for(i=0;i<no_Rayleigh_cross_sections;i++)
		{
			if(fscanf(fptr, "%lg\t%lg\n", lambda_Rayleigh+i, cs_Rayleigh+i)!=2)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error reading line %d of Rayleigh cross section file \"%s\"!\n", i+1, filename);
				fclose(fptr);
				return false;
			}
			lambda_Rayleigh[i]*=1e-9;
		}
		fclose(fptr);
		cs_lambda_A=lambda_Rayleigh[0];
		cs_lambda_B=lambda_Rayleigh[no_Rayleigh_cross_sections-1];
		index_Rayleigh_cs=0;	// reset index into cross section table
	}
	return initializeData();
}

void Air::resetCluster(struct Profile::Cluster_s *pCluster)
{
	double T=pCluster->input_data[temperature];
	pCluster->optical_data[density]=1e-4*pCluster->input_data[pressure]/(const_kB*T);	// [1/cm³]
	pCluster->optical_data[wavelength]=0.0;
	pCluster->optical_data[epsilon_RRS]=0.0;
}

bool Air::initializeData()
{
	int i;
	// RRS setup
	if(b_RRS)
	{
		if(!(MolecularRRSCrossSection::setJMax(RRS_JMax) && GasMixtureDCS::initializeMolecules(RRS_JMax)))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error setting J_max to %d\n", RRS_JMax);
			return false;
		}
		if(RRSEvent)
			delete [] RRSEvent;
		RRSEvent=new RingEvent[no_clusters];
		if(!RRSEvent)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"error allocating %d RRS event objects!\n", no_clusters);
			return false;
		}
		for(i=0;i<no_clusters;i++)
		{
			resetCluster(cluster+i);
			if(!RRSEvent[i].allocate())
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			RRSEvent[i].associateWithAirCluster(cluster+i);
			RRSEvent[i].calculateKineticConstants();
		}
		if(RCSFormula==RCS_vanDeelen)
			pElasticMolecularPhaseFunction=AirElasticDCS.getCabannesPhaseFunctionPtr();
		else
			pElasticMolecularPhaseFunction=&CabannesPhF;
	}
	else
	{
		if(!GasMixtureDCS::initializeMolecules(RRS_JMax))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error setting J_max to %d\n", RRS_JMax);
			return false;
		}
		for(i=0;i<no_clusters;i++)
			resetCluster(cluster+i);
		if(RCSFormula==RCS_vanDeelen)
			pElasticMolecularPhaseFunction=AirElasticDCS.getRayleighPhaseFunctionPtr();
		else
			pElasticMolecularPhaseFunction=&RayleighPhF;
	}
	return true;
}

bool Air::calculateRayleighScatteringCrossSection(const double& wavelength, double *pcs)
{
	double cs;
	switch(RCSFormula)
	{
		case RCS_Penndorf:
		{
			double x=1e-6/wavelength; // x = 1 / (wavelength [micron])
			double x2=x*x;
			double x4=x2*x2;
			cs=1e-28*3.9993e-4*x4/(1-1.069e-2*x2-6.681e-5*x4);
			break;
		}
		case RCS_BodhaineEq29:
		{
			double l2=1e12*wavelength*wavelength;
			cs=1e-32*(1.0455996-341.29061/l2-0.90230850*l2)/(1.0+0.0027059889/l2-85.968563*l2);
			break;
		}
		case RCS_file:
		{
			if(wavelength<cs_lambda_A || wavelength>cs_lambda_B)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "wavelength %lg out of range [%lg,%lg]!\n", wavelength, cs_lambda_A, cs_lambda_B);
				return false;
			}
			if(lambda_Rayleigh[index_Rayleigh_cs]<=wavelength)
			{
				do
				{
					index_Rayleigh_cs++;
					if(index_Rayleigh_cs>=no_Rayleigh_cross_sections)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "no Rayleigh cross section data for wavelength %lg, index=%d!\n", wavelength, index_Rayleigh_cs);
						return false;
					}
				}while(lambda_Rayleigh[index_Rayleigh_cs]<wavelength);
				index_Rayleigh_cs--;
			}
			else
			{
				do
				{
					index_Rayleigh_cs--;
					if(index_Rayleigh_cs<0)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "no Rayleigh cross section data for wavelength %lg, index=%d!\n", wavelength, index_Rayleigh_cs);
						return false;
					}
				}while(lambda_Rayleigh[index_Rayleigh_cs]>wavelength);
			}
			// found intervall
			double f, l_a, l_b, cs_a, cs_b;
			l_a=lambda_Rayleigh[index_Rayleigh_cs];
			l_b=lambda_Rayleigh[index_Rayleigh_cs+1];
			f=(wavelength-l_a)/(l_b-l_a);

			cs_a=cs_Rayleigh[index_Rayleigh_cs];
			cs_b=cs_Rayleigh[index_Rayleigh_cs+1];
			cs=1e-4*(cs_a+f*(cs_b-cs_a));
			break;
		}
		default: 
			errorIn(__FILE__, __LINE__,__func__);
			fprintf(stderr, "failed to calculate Rayleigh cross section\n");
			return false;
	}
	*pcs=cs;	// in m^2
	return true;
}

void Air::setCenterWavelength(const double& wavelength)
{
	if(RCSFormula!=RCS_vanDeelen)
		Molecule.calculateKingCorrectionFactorOfAir(wavelength, 450.0);
}

bool Air::setWavelength(const double& lambda, const int& cluster_index)
{
	if(cluster_index<0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "cluster_index=%d\n", cluster_index);
		return false;
	}
	struct Profile::Cluster_s *pCluster=cluster+cluster_index;
	double *od=pCluster->optical_data;
	double *id=pCluster->input_data;
	if(od[wavelength]==lambda)
		return true;

	// calculate refractive index
	double n;
	switch(getClusterDimension())
	{
		default:
		case 2:
			n=calculateRefractiveIndexCiddor(lambda, id[pressure], id[temperature], 0.0);
			break;
		case 3:
			n=calculateRefractiveIndexCiddor(lambda, id[pressure], id[temperature], id[relative_humidity]);
			break;
		case 4:
			n=calculateRefractiveIndexCiddor(lambda, id[pressure], id[temperature], id[relative_humidity], id[vmr_CO2]);
			break;
	}
	od[refractive_index]=n;
	double cs, cs_RRS, number_density=od[density];
	// calculate rotational Raman scattering cross sections
	if(b_RRS)
	{
		class RingEvent *pRingEvent=RRSEvent+cluster_index;
		cs_RRS=pRingEvent->calculateCrossSections(lambda);
		// cs=AirElasticDCS.getRayleighCrossSection()-AirElasticDCS.getCabannesCrossSection();
		od[epsilon_RRS]=number_density*cs_RRS*1e5;
	}
	else
		od[epsilon_RRS]=cs_RRS=0.0;
	// Rayleigh / Cabannes cross section
	if(RCSFormula==RCS_vanDeelen)
	{
		if(b_RRS)
		{
			AirElasticDCS.calculateOpticalProperties(lambda);
			// cs=AirElasticDCS.getCabannesCrossSection();
			cs=AirElasticDCS.getRayleighCrossSection()-cs_RRS;
		}
		else
		{
			AirElasticDCS.calculateOpticalPropertiesElastic(lambda);
			cs=AirElasticDCS.getRayleighCrossSection();
		}
		cs*=1e-4;
	}
	else
	{
		Molecule.calculateKingCorrectionFactorOfAir(lambda, 450.0);
		// calculate Rayleigh (Cabannes) cross section
		if(RCSFormula==RCS_Bodhaine)
		{
			double lambda2=1e4*lambda*lambda;	// lambda needed in [cm]
			double nn=n*n, nnm1=nn-1.0, nnp2=nn+2.0;
			double x=nnm1/(lambda2*number_density*nnp2), xx=x*x;
			cs=744.150640327196*xx*Molecule._KingFactor()*1e-4;	// 1e-4 conversion cm^2 to m^2
		}
		else
		{
			if(!calculateRayleighScatteringCrossSection(lambda, &cs))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
		}
		if(b_RRS)
		{
			CabannesPhF.updateCoefficients();
			cs*=Molecule._f_Cabannes();
		}
		else
			RayleighPhF.updateCoefficients();
	}
	od[epsilon_elastic_molecular]=number_density*cs*1e9;
	od[wavelength]=lambda;
	return true;
}

bool Air::logRRSData()
{
	char path[BasicIO_string_size], file_name[BasicIO_string_size];
	formatOutputDirectory(path);
	sprintf(file_name, "%s%s_RRS.txt", path, getProfileName());
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not write RRS data to file \"%s\"!\n", file_name);
		return false;
	}
	double sigma_RRS=(AirElasticDCS.getRayleighCrossSection()-AirElasticDCS.getCabannesCrossSection());
	for(int i=0;i<getNOClusters();i++)
	{
		fprintf(fptr, "%lg", sigma_RRS);
		RRSEvent[i].print(fptr);
	}
	fclose(fptr);
	return true;
}

bool Air::logData()
{
	if(RCSFormula!=RCS_vanDeelen)
		return true;
	char path[BasicIO_string_size], file_name[BasicIO_string_size];
	formatOutputDirectory(path);
	sprintf(file_name, "%s%s_mp.txt", path, getProfileName());
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not write molecule parameter data to file \"%s\"!\n", file_name);
		return false;
	}
	AirElasticDCS.logMoleculeParameters(fptr);
	fclose(fptr);
	return true;
}




