/**************************************************************************************************************
*
*       This file,
*
*                            /Air.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Air_h
#define _Air_h

#include "Profile.h"

#include "RingEvent.h"

#include "PhaseFunctions/RayleighPhaseFunction.h"
#include "PhaseFunctions/CabannesPhaseFunction.h"
#include "PhaseFunctions/RotationalRamanPhaseFunction.h"

#include "Molecule.h"
#include "GasMixtureDCS.h"

#include "BasicIO.h"

class Atmosphere;

class Air: public BasicIO, public Profile
{
	public:
		enum L1Element_e
		{
			density,
			wavelength,
			refractive_index,
			epsilon_elastic_molecular,
			epsilon_RRS,
			no_L1_elements
		};

		enum L0Element_e
		{
			pressure,
			temperature,
			relative_humidity,
			vmr_CO2
		};
	
		enum RayleighCrossSectionFormula_e
		{
			RCS_Penndorf,
			RCS_Nicolet,
			RCS_BodhaineEq29,
			RCS_Bodhaine,
			RCS_vanDeelen,
			RCS_file
		};

	protected:
	// IO #########################################################################
		char fn_air_data[BasicIO_string_size];			// air profiles filename
		char fn_air_map[BasicIO_string_size];			// air map filename

		// rrs
		bool b_RRS;
		int RRS_JMax;						// 

		// spectroscopic parameters
		char fn_rayleigh_cross_section[BasicIO_string_size];	// external rayleigh cross section file
		enum RayleighCrossSectionFormula_e RCSFormula;
		bool b_consider_Herzberg_continuum;
		bool b_isotropic_molecular_polarizability;
		bool b_emission;
	// IO ####################################################################


		bool logRRSData();
		virtual void setClusterDimensions(int input_cluster_dimension);
		// virtual int defineAuxiliaryDataDimension(){return 0;}

		class Molecule Molecule;

		class PhaseFunction *pElasticMolecularPhaseFunction;

		class RayleighPhaseFunction RayleighPhF;
		class RayleighPhaseFunction CabannesPhF;
		class RotationalRamanPhaseFunction RRPhaseFunction;

		class GasMixtureDCS AirElasticDCS;

	private:

		class Atmosphere *pAtmosphere;

		// gases
		int no_gases;

		// simulate Ring?
		class RingEvent *RRSEvent;

		// Rayleigh scattering cross section ..........
		// class Matrix CSData;
		double cs_lambda_A, cs_lambda_B;
		double *cs_data, *lambda_Rayleigh, *cs_Rayleigh;
		int no_Rayleigh_cross_sections, index_Rayleigh_cs;

		bool calculateRayleighScatteringCrossSection(const double& wavelength, double *pcs);

		bool initializeData();
		void resetCluster(struct Profile::Cluster_s *pCluster);

	public:

		Air();
		~Air();

		bool loadFiles();
		bool logData();

		void setCenterWavelength(const double& lambda);
		bool setWavelength(const double& lambda, const int& cluster_index);

		// inline functions
		inline void setAtmospherePtr(class Atmosphere *AtmospherePtr){pAtmosphere=AtmospherePtr;}
		inline bool simulateRingEffect(){return b_RRS;}
		inline bool simulateEmission(){return b_emission;}

		inline class PhaseFunction *getRotationalRamanPhaseFunctionPtr(){return &RRPhaseFunction;}
		inline class PhaseFunction *getElasticMolecularPhaseFunctionPtr(){return pElasticMolecularPhaseFunction;}

		inline class RingEvent *getRingEventPtr(){return RRSEvent;}

		// static service --------------------------------------------------------
		static double calculateRefractiveIndexCiddor(const double& wl, const double& p_hp, const double& T,const double& RH, const double& x_CO2=450.0);
	
};

#endif //_Air_h


