/**************************************************************************************************************
*
*       This file,
*
*                            /Atmosphere.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Atmosphere.h"
#include "error.h"

Atmosphere::Atmosphere()
{
	setAtmospherePtr(this);
	gas=0;
	particle=0;
	OC=0;

	GasOpticalStructPtr=0;
	ParticleOpticalStructPtr=0;
	ActiveGasOpticalStructPtr=0;
	ActiveParticleOpticalStructPtr=0;

	addParameter(gases, type_string, "gases|absorbers|gaseous_absorbers", 0, "comma separated list of gaseous absorber configuration filenames, each without '.conf'");
	addParameter(particles, type_string, "particles|scatterers", 0, "comma separated list of particle configuration filenames, each without '.conf'");
}

Atmosphere::~Atmosphere()
{
	if(gas)
		delete [] gas;
	if(particle)
		delete [] particle;

	if(OC)
		delete [] OC;

	if(GasOpticalStructPtr)
		delete [] GasOpticalStructPtr;
	if(ParticleOpticalStructPtr)
		delete [] ParticleOpticalStructPtr;

	if(ActiveGasOpticalStructPtr)
		delete [] ActiveGasOpticalStructPtr;
	if(ActiveParticleOpticalStructPtr)
		delete [] ActiveParticleOpticalStructPtr;
}

bool Atmosphere::loadFiles()
{
	// air
	if(!Air::loadFiles())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to initialize air object!\n");
		return false;
	}
	int i;
	char filename[1000], buffer[50];

	// setup gases ------------------------------------------------
	if(gas)
		delete [] gas;
	gas=0;
	no_gases=0;
	if(strcmp(gases, "none"))
	{
		no_gases=numberOf(',', gases)+1;
		gas=new Gas[no_gases];
		if(!gas)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"failed to allocate %d gases!\n", no_gases);
			return false;
		}
		for(i=0;i<no_gases;i++)
		{
			gas[i].setAirPtr(this);
			extractWordFromList(gases, i, ',', buffer);
			sprintf(filename, "%s%s.conf", dir_base, buffer);
			strcpy(gas[i].getBaseDirectory(), dir_base);
			if(!gas[i].read(filename) || !gas[i].loadFiles())
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "failed to read gas definition from file \"%s\"!\n", filename);
				return false;
			}
		}
	}

	// setup particles --------------------------------------------
	if(particle)
		delete [] particle;
	particle=0;
	no_particles=0;
	if(strcmp(particles, "none"))
	{
		no_particles=numberOf(',', particles)+1;
		particle=new Particle[no_particles];
		if(!particle)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "failed to allocate %d gases!\n", no_particles);
			return false;
		}	
		for(i=0;i<no_particles;i++)
		{
			extractWordFromList(particles, i, ',', buffer);
			sprintf(filename, "%s%s.conf", dir_base, buffer);
			strcpy(particle[i].getBaseDirectory(), dir_base);
			if(!particle[i].read(filename) || !particle[i].loadFiles())
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "failed to read particle definition from file \"%s\"!\n", filename);
				return false;
			}
		}
	}
	return allocateOpticalCells();
}

bool Atmosphere::allocateOpticalCells()
{
	no_optical_cells=pGrid->getNOCells();

	// first determine number of unique cells
	int i, j;
	long int UniqueOCId, resolution;
	TVector<long int> UniqueOCIdList;
	if(!UniqueOCIdList.allocate(no_optical_cells)  || !UniqueOCMap.allocate(no_optical_cells))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	no_unique_optical_cells=0;
	for(i=0;i<no_optical_cells;i++)
	{
		resolution=1;
		// calculate Id for optical cell
		UniqueOCId=Air::getClusterIndex(i);
		resolution*=Air::getNOClusters();
		for(j=0;j<no_gases;j++)
		{
			UniqueOCId+=resolution*(1+gas[j].getClusterIndex(i));
			resolution*=gas[j].getNOClusters()+1;
		}
		for(j=0;j<no_particles;j++)
		{
			UniqueOCId+=resolution*(1+particle[j].getClusterIndex(i));
			resolution*=particle[j].getNOClusters()+1;
		}
		// add unique cell id if not already present
		for(j=0;j<no_unique_optical_cells;j++)
			if(UniqueOCIdList[j]==UniqueOCId)
			{
				UniqueOCMap[i]=j;
				break;
			}
		if(j==no_unique_optical_cells)
		{
			// id not found in list, new id
			UniqueOCIdList[no_unique_optical_cells]=UniqueOCId;
			UniqueOCMap[i]=no_unique_optical_cells;
			no_unique_optical_cells++;
		}
	}

	// allocate unique cells
	if(OC)
		delete [] OC;
	OC=new struct OpticalCell_s[no_unique_optical_cells];
	if(!OC)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating optical cells!\n");
		return false;
	}
	memset(OC, 0, sizeof(struct OpticalCell_s) * no_unique_optical_cells);

	// count gas and particle structs in unique optical cells
	struct OpticalCell_s *pOC;
	int no_GasOpticalStructs=0, no_ParticleOpticalStructs=0, cluster_index;
	for(i=0;i<no_optical_cells;i++)
	{
		pOC=OC+UniqueOCMap[i];
		if(pOC->valid)
			continue;
		pOC->air_cluster_index=Air::getClusterIndex(i);
		// pOC->cell_index=i;
		pOC->no_gases=0;
		for(j=0;j<no_gases;j++)
			if(gas[j].isCellOccupied(i))
				pOC->no_gases++;
		no_GasOpticalStructs+=pOC->no_gases;
		pOC->no_particles=0;
		for(j=0;j<no_particles;j++)
			if(particle[j].isCellOccupied(i))
				pOC->no_particles++;
		no_ParticleOpticalStructs+=pOC->no_particles;
		pOC->valid=true;
	}

	// allocate Gas::OpticalCell_s and Particle::OpticalCell_s
	if(GasOpticalStructPtr)
		delete [] GasOpticalStructPtr;
	GasOpticalStructPtr=0;
	if(no_GasOpticalStructs)
	{
		GasOpticalStructPtr=new struct Gas::OpticalStruct_s *[no_GasOpticalStructs];
		if(!GasOpticalStructPtr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating %d optical cells for gases !\n", no_GasOpticalStructs);
			return false;
		}
	}

	// ---------------------------------------
	if(ParticleOpticalStructPtr)
		delete [] ParticleOpticalStructPtr;
	ParticleOpticalStructPtr=0;
	if(no_ParticleOpticalStructs)
	{
		ParticleOpticalStructPtr=new struct Particle::OpticalStruct_s *[no_ParticleOpticalStructs];
		if(!ParticleOpticalStructPtr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating %d optical cells for particles!\n", no_ParticleOpticalStructs);
			return false;
		}
	}
	
	// adjust pointers to Gas::OpticalCell_s and Particle::OpticalCell_s:
	class Gas *pGas;
	struct Gas::OpticalStruct_s **pGOSPtr=GasOpticalStructPtr;

	class Particle *pParticle;
	struct Particle::OpticalStruct_s **pPOSPtr=ParticleOpticalStructPtr;

	// clear valid flag for next step
	for(i=0;i<no_unique_optical_cells;i++)
		OC[i].valid=false;

	for(i=0;i<no_optical_cells;i++)
	{
		pOC=OC+UniqueOCMap[i];
		if(pOC->valid)
			continue;

		pOC->GasOpticalStructPtr=pGOSPtr;
		for(j=0;j<no_gases;j++)
		{
			pGas=gas+j;
			if(pGas->isCellOccupied(i))
			{
				cluster_index=pGas->getClusterIndex(i);
				*pGOSPtr=pGas->getOpticalStructPtr(cluster_index);
				(*pGOSPtr)->cluster_index=cluster_index;
				pGOSPtr++;
			}
		}

		pOC->ParticleOpticalStructPtr=pPOSPtr;
		for(j=0;j<no_particles;j++)
		{
			pParticle=particle+j;
			if(pParticle->isCellOccupied(i))
			{
				cluster_index=pParticle->getClusterIndex(i);
				*pPOSPtr=pParticle->getOpticalStructPtr(cluster_index);
				(*pPOSPtr)->cluster_index=cluster_index;
				pPOSPtr++;
			}
		}
		// Ring effect
		if(simulateRingEffect())
			pOC->pRRSEvent=getRingEventPtr()+map[i];

		pOC->valid=true;
	}
	// log unique cell hash list data
	return true;
}

void Atmosphere::resetOpticalCells()
{
	int i;
	struct OpticalCell_s *pOC=OC;
	for(i=0;i<no_unique_optical_cells;i++)
	{
		pOC->valid=false;
		// pOC->cell_index=i;
		pOC->weight_splitting_limit=1.0;
		pOC->weight_russian_roulette_limit=0.0;
		pOC->weight_optimum=1.0;
		pOC->psi_sampling_probability=0.0;
		pOC++;
	}
	for(i=0;i<no_particles;i++)
		particle[i].invalidateAllClusters();
	for(i=0;i<no_gases;i++)
		gas[i].invalidateAllClusters();
	// Ring effect statistics reset
	/*if(b_consider_rotational_raman_scattering)
		for(i=0;i<Air::getNOClusters();i++)
			getRingEventPtr()[i].resetHits();*/
}

/*struct OpticalCell_s *Atmosphere::getOpticalCellPtr(const struct CellIndex_s& CI)
{
	int i, cell_index=pGrid->calculateCellIndex(CI);
	struct OpticalCell_s *pOC=OC+cell_index;
	if(pOC->valid)
		return pOC;
	return 0;
}*/

bool Atmosphere::calculateOpticalCellGasCrossSections(struct OpticalCell_s *pOC, const double& wavelength)
{
	int i;

	// total optical properties
	double absorption_coefficient=0.0;

	// calculate gases ###########################################################
	struct Gas::OpticalStruct_s *pGOS;
	for(i=0;i<pOC->no_gases;i++)
	{
		pGOS=pOC->GasOpticalStructPtr[i];
		pGOS->wavelength=wavelength;
		if(!pGOS->pGas->calculateOpticalStruct(pGOS))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error calculating Gas::OpticalStruct of gas \"%s\" cluster %d!\n", pGOS->pGas->getName(), 1+pGOS->cluster_index);
			return false;
		}
		absorption_coefficient+=pGOS->absorption_coefficient;
	}

	// recalculate total optical properties ##############################################
	pOC->absorption_coefficient=absorption_coefficient;
	pOC->extinction_coefficient=absorption_coefficient+pOC->scatter_coefficient;
	pOC->wavelength=wavelength;

	// done !
	return pOC->valid=true;
}

bool Atmosphere::setCenterWavelength(const double& center_wavelength)
{
	int i, j;
	class OpticalCell_s *pOC=OC;
	struct Gas::OpticalStruct_s *pGOS;
	struct Particle::OpticalStruct_s *pPOS;
	Air::setCenterWavelength(center_wavelength);
	for(j=0;j<no_unique_optical_cells;j++)
	{
		if(!calculateOpticalCell(pOC, center_wavelength))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}

		// carry results over -------------------------
		// air
		pOC->extinction_coefficient_at_center_wavelength=pOC->extinction_coefficient;
		pOC->absorption_coefficient_at_center_wavelength=pOC->absorption_coefficient;
		pOC->emission_at_center_wavelength=pOC->emission;
		pOC->scatter_coefficient_at_center_wavelength=pOC->scatter_coefficient;
		pOC->rayleigh_scatter_coefficient_at_center_wavelength=pOC->rayleigh_scatter_coefficient;
		pOC->rotational_raman_scatter_coefficient_at_center_wavelength=pOC->rotational_raman_scatter_coefficient;

		// gases 
		for(i=0;i<pOC->no_gases;i++)
		{
			pGOS=pOC->GasOpticalStructPtr[i];
			pGOS->absorption_coefficient_at_center_wavelength=pGOS->absorption_coefficient;
		}

		// particles 
		for(i=0;i<pOC->no_particles;i++)
		{
			pPOS=pOC->ParticleOpticalStructPtr[i];
			pPOS->extinction_coefficient_at_center_wavelength=pPOS->extinction_coefficient;
			pPOS->scatter_coefficient_at_center_wavelength=pPOS->scatter_coefficient;
			pPOS->absorption_coefficient_at_center_wavelength=pPOS->absorption_coefficient;
			pPOS->varpi0_at_center_wavelength=pPOS->varpi0;
		}

		// next optical cell
		pOC++;
	}
	if(b_RRS)
		logRRSData();
	Air::logData();
	return true;
}

bool Atmosphere::calculateOpticalCell(struct OpticalCell_s *pOC, const double& wavelength)
{
	int i;

	// total optical properties
	double absorption_coefficient=0.0, scatter_coefficient;

	// air #######################################################################
	if(!Air::setWavelength(wavelength, pOC->air_cluster_index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error setting air cluster wavelength to %le!\n", wavelength);
		return false;
	}
	struct Profile::Cluster_s *pCluster=cluster+pOC->air_cluster_index;
	pOC->air_number_density=pCluster->optical_data[density];
	pOC->rayleigh_scatter_coefficient=pCluster->optical_data[epsilon_elastic_molecular];
	pOC->rotational_raman_scatter_coefficient=pCluster->optical_data[epsilon_RRS];
	scatter_coefficient=pOC->rayleigh_scatter_coefficient+pOC->rotational_raman_scatter_coefficient;
	double g_DIS=0.0;

	// calculate gases ###########################################################
	struct Gas::OpticalStruct_s *pGOS;
	for(i=0;i<pOC->no_gases;i++)
	{
		pGOS=pOC->GasOpticalStructPtr[i];
		pGOS->wavelength=wavelength;
		if(!pGOS->pGas->calculateOpticalStruct(pGOS))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error calculating Gas::OpticalStruct of gas \"%s\" cluster %d!\n", pGOS->pGas->getName(), pGOS->cluster_index);
			return false;
		}
		absorption_coefficient+=pGOS->absorption_coefficient;
	}

	// calculate particles ########################################################
	struct Particle::OpticalStruct_s *pPOS;
	for(i=0;i<pOC->no_particles;i++)
	{
		pPOS=pOC->ParticleOpticalStructPtr[i];
		pPOS->wavelength=wavelength;
		if(!pPOS->pParticle->calculateOpticalStruct(pPOS))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error calculating Particle::OpticalStruct of particle \"%s\" cluster %d!\n", pPOS->pParticle->getName(), pPOS->cluster_index);
			return false;
		}
		absorption_coefficient+=pPOS->absorption_coefficient;
		g_DIS+=pPOS->scatter_coefficient*pPOS->asymmetry_parameter;
		scatter_coefficient+=pPOS->scatter_coefficient;
	}
	if(!scatter_coefficient)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "scatter coefficient of cell is zero. OpticalCell info:\n");
		pOC->printInfo(stderr);
		return false;
	}
	if(ISNAN(scatter_coefficient) || ISNAN(absorption_coefficient))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// recalculate total optical properties ##############################################
	g_DIS/=scatter_coefficient;
	// pOC->HGDIS.defineParameter(0, sqrt(g_DIS));
	pOC->HGDIS.defineParameter(0, g_DIS);
	pOC->absorption_coefficient=absorption_coefficient;
	pOC->scatter_coefficient=scatter_coefficient;
	pOC->extinction_coefficient=absorption_coefficient+scatter_coefficient;
	pOC->wavelength=wavelength;

	// calculate emission
	if(b_emission)
	{
		const double c1 = 2.0 * M_PI * PlanckConstant * speed_of_light * speed_of_light ;
		const double c2 = PlanckConstant * speed_of_light / k_B ;
		double T=getClusterPtr()[pOC->air_cluster_index].input_data[temperature];
		pOC->emission = 1e-14 * c1 * ( exp ( c2 / ( wavelength * T ) ) - 1.0 ) / pow ( wavelength, 5.0 );
	}

	// done !
	return pOC->valid=true;
}

struct Atmosphere::OpticalCell_s *Atmosphere::getOpticalCellPtr(const int& cell_index, const double& wavelength)
{
	struct OpticalCell_s *pOC=OC+UniqueOCMap[cell_index];
	if(pOC->valid && pOC->wavelength==wavelength)
		return pOC;
	if(calculateOpticalCell(pOC, wavelength))
		return pOC;
	return 0;
}

bool Atmosphere::createActiveConstituentLists()
{
	int i, j, k, no_active_gas_structs=0, no_active_particle_structs=0;
	struct Atmosphere::OpticalCell_s *pOC;
	bool is_active_cell;

	// release memory
	if(ActiveParticleOpticalStructPtr)
		delete [] ActiveParticleOpticalStructPtr;
	ActiveParticleOpticalStructPtr=0;
	if(ActiveGasOpticalStructPtr)
		delete [] ActiveGasOpticalStructPtr;
	ActiveGasOpticalStructPtr=0;

	// count structs
	for(i=0;i<no_unique_optical_cells;i++)
		OC[i].initialized=false;
	for(i=0;i<no_optical_cells;i++)
	{
		pOC=OC+UniqueOCMap[i];
		if(pOC->initialized)
			continue;	
		is_active_cell=false;
		pOC->no_active_gases=0;
		for(j=0;j<pOC->no_gases;j++)
			if(pOC->GasOpticalStructPtr[j]->pGas->isFunctionalCalculationEnabled())
			{
				pOC->no_active_gases++;
				no_active_gas_structs++;
				is_active_cell=true;
			}
		pOC->no_active_particles=0;
		for(j=0;j<pOC->no_particles;j++)
			if(pOC->ParticleOpticalStructPtr[j]->pParticle->isFunctionalCalculationEnabled())
			{
				no_active_particle_structs++;
				pOC->no_active_particles++;
				is_active_cell=true;
			}
		pOC->is_active_cell=is_active_cell;
		pOC->initialized=true;
	}

	// allocate
	if(no_active_gas_structs)
	{
		ActiveGasOpticalStructPtr=new Gas::OpticalStruct_s *[no_active_gas_structs];
		if(!ActiveGasOpticalStructPtr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating active gas optical structs!\n");
			return false;
		}
	}
	if(no_active_particle_structs)
	{
		ActiveParticleOpticalStructPtr=new Particle::OpticalStruct_s *[no_active_particle_structs];
		if(!ActiveParticleOpticalStructPtr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating active particle optical structs!\n");
			return false;
		}
	}

	// set pointers
	struct Gas::OpticalStruct_s **ppGOS=ActiveGasOpticalStructPtr;
	struct Particle::OpticalStruct_s **ppPOS=ActiveParticleOpticalStructPtr;
	for(i=0;i<no_unique_optical_cells;i++)
		OC[i].initialized=false;
	for(i=0;i<no_optical_cells;i++)
	{
		pOC=OC+UniqueOCMap[i];
		if(pOC->initialized)
			continue;	
		pOC->ActiveGasOpticalStructPtr=ppGOS;
		ppGOS+=pOC->no_active_gases;
		k=0;
		for(j=0;j<pOC->no_gases;j++)
			if(pOC->GasOpticalStructPtr[j]->pGas->isFunctionalCalculationEnabled())
				pOC->ActiveGasOpticalStructPtr[k++]=pOC->GasOpticalStructPtr[j];
		
		pOC->ActiveParticleOpticalStructPtr=ppPOS;
		ppPOS+=pOC->no_active_particles;
		k=0;
		for(j=0;j<pOC->no_particles;j++)
			if(pOC->ParticleOpticalStructPtr[j]->pParticle->isFunctionalCalculationEnabled())
				pOC->ActiveParticleOpticalStructPtr[k++]=pOC->ParticleOpticalStructPtr[j];
		pOC->initialized=true;
	}
	return true;
}

void Atmosphere::writeAuxiliaryFiles()
{
	int i;
	Air::writeAuxiliaryFiles();
	for(i=0;i<no_gases;i++)
		gas[i].writeAuxiliaryFiles();
	for(i=0;i<no_particles;i++)
		particle[i].writeAuxiliaryFiles();
}

void Atmosphere::disableFunctionalCalculation()
{
	int i;
	for(i=0;i<no_gases;i++)
		gas[i].enableFunctionalCalculation(false);
	for(i=0;i<no_particles;i++)
		particle[i].enableFunctionalCalculation(false);
}

void Atmosphere::OpticalCell_s::printInfo(FILE *fptr)
{
	fprintf(fptr, "optical cell valid                                         %s\n", valid ? "yes" : "no");
	fprintf(fptr, "optical cell initialized                                   %s\n", initialized ? "yes" : "no");
	fprintf(fptr, "air cluster index                                          %d\n", air_cluster_index );
	fprintf(fptr, "wavelength                                                 %lg\n", wavelength);

	fprintf(fptr, "extinction_coefficient                                     %lg\n", extinction_coefficient);
	fprintf(fptr, "extinction_coefficient_at_center_wavelength                %lg\n", extinction_coefficient_at_center_wavelength);

	fprintf(fptr, "absorption_coefficient                                     %lg\n", absorption_coefficient);
	fprintf(fptr, "absorption_coefficient_at_center_wavelength                %lg\n", absorption_coefficient_at_center_wavelength);

	fprintf(fptr, "scatter_coefficient                                        %lg\n", scatter_coefficient);
	fprintf(fptr, "scatter_coefficient_at_center_wavelength                   %lg\n", scatter_coefficient_at_center_wavelength);

	fprintf(fptr, "air_number_density                                         %lg\n", air_number_density);

	if(pRRSEvent)
	{
		fprintf(fptr, "Cabannes_scatter_coefficient                               %lg\n", rayleigh_scatter_coefficient);
		fprintf(fptr, "Cabannes_scatter_coefficient_at_center_wavelength          %lg\n", rayleigh_scatter_coefficient_at_center_wavelength);

		fprintf(fptr, "rotational_raman_scatter_coefficient                       %lg\n", rotational_raman_scatter_coefficient);
		fprintf(fptr, "rotational_raman_scatter_coefficient_at_center_wavelength  %lg\n", rotational_raman_scatter_coefficient_at_center_wavelength);
	}
	else
	{
		fprintf(fptr, "rayleigh_scatter_coefficient                               %lg\n", rayleigh_scatter_coefficient);
		fprintf(fptr, "rayleigh_scatter_coefficient_at_center_wavelength          %lg\n", rayleigh_scatter_coefficient_at_center_wavelength);
	}

	// class RingEvent *pRRSEvent;
	fprintf(fptr, "optical cell active                                        %s\n", is_active_cell ? "yes" : "no" );

	fprintf(fptr, "number of gases                                            %d\n", no_gases );
	fprintf(fptr, "number of active gases                                     %d\n", no_active_gases );

	fprintf(fptr, "number of particles                                        %d\n", no_particles );
	fprintf(fptr, "number of active particles                                 %d\n", no_active_particles );

	fprintf(fptr, "refractive_index                                           %lg\n", refractive_index );
}




