/**************************************************************************************************************
*
*       This file,
*
*                            /Atmosphere.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Atmosphere_h
#define _Atmosphere_h

#include "Air.h"

#include "Particle.h"
#include "Gas.h"

#include "Grid.h"

class Atmosphere: public Air
{
	public:
		struct OpticalCell_s
		{
			bool valid, initialized;

			// int cell_index;
			int air_cluster_index;
			double wavelength;

			// total optical properties
			double extinction_coefficient;
			double extinction_coefficient_at_center_wavelength;

			double absorption_coefficient;
			double absorption_coefficient_at_center_wavelength;

			double emission;
			double emission_at_center_wavelength;

			double scatter_coefficient;
			double scatter_coefficient_at_center_wavelength;

			// air
			double air_number_density;						// [1/cm³]

			double rayleigh_scatter_coefficient;					// [1/km]
			double rayleigh_scatter_coefficient_at_center_wavelength;		// [1/km]

			double rotational_raman_scatter_coefficient;				// [1/km]
			double rotational_raman_scatter_coefficient_at_center_wavelength;	// [1/km]

			class RingEvent *pRRSEvent;

			bool is_active_cell;

			// gaseous absorbers
			int no_gases, no_active_gases;
			struct Gas::OpticalStruct_s **GasOpticalStructPtr, **ActiveGasOpticalStructPtr;
			// list<struct Gas::OpticalStruct_s *> ActiveGasOpticalStructPtrList;

			// scatterers 
			int no_particles, no_active_particles;
			struct Particle::OpticalStruct_s **ParticleOpticalStructPtr, **ActiveParticleOpticalStructPtr;
			// list<struct Particle::OpticalStruct_s *> ActiveParticleOpticalStructPtrList;

			double refractive_index;

			// variance reduction ------------------------------------
			double weight_splitting_limit;
			double weight_optimum;
			double weight_russian_roulette_limit;
			double psi_sampling_probability;

			// directional importance sampling
			class HenyeyGreensteinPhaseFunction HGDIS;
			void printInfo(FILE *fptr=stdout);
		};
	// IO #########################################################################
		char gases[BasicIO_string_size];			// list of gaseous absorbers
		char particles[BasicIO_string_size];			// list of particles
	// IO ####################################################################

	protected:

		// optical cells
		int no_optical_cells, no_unique_optical_cells;
		struct OpticalCell_s *OC;
		class TVector<int> UniqueOCMap;

		struct Gas::OpticalStruct_s **GasOpticalStructPtr, **ActiveGasOpticalStructPtr;
		struct Particle::OpticalStruct_s  **ParticleOpticalStructPtr, **ActiveParticleOpticalStructPtr;

		bool allocateOpticalCells();
			
		// constituents

		int no_gases;
		class Gas *gas;

		int no_particles;
		class Particle *particle;
		
	public:

		Atmosphere();
		~Atmosphere();

		void disableFunctionalCalculation();

		bool setCenterWavelength(const double& center_wavelength);

		// void initializeProfileIO();
	
		virtual bool loadFiles();
		
		inline int getNOGases(){return no_gases;}	
		inline class Gas *getGasPtr(){return gas;}

		inline int getNOParticles(){return no_particles;}
		inline class Particle *getParticlePtr(){return particle;}

		// bool setWavelength(int cell_index, const double& wavelength);
		void writeAuxiliaryFiles();
		
		void resetOpticalCells();
		void invalidateCell(const int& cell_index){OC[UniqueOCMap[cell_index]].valid=false;}
		// inline struct OpticalCell_s *getOpticalCellPtr(){return OC;}

		inline struct OpticalCell_s *getOpticalCellPtr(const struct CellIndex_s& CI){return OC+UniqueOCMap[pGrid->calculateCellIndex(CI)];}
		inline struct OpticalCell_s *getOpticalCellPtr(const int& cell_index){return OC+UniqueOCMap[cell_index];}

		struct OpticalCell_s *getOpticalCellPtr(const int& cell_index, const double& wavelength);
		inline struct OpticalCell_s *getOpticalCellPtr(const struct CellIndex_s& CI, const double& wavelength)
			{return getOpticalCellPtr(pGrid->calculateCellIndex(CI), wavelength);}

		// struct OpticalCell_s *getOpticalCellPtr(const struct CellIndex_s& CI);
		bool calculateOpticalCell(struct OpticalCell_s *pOC, const double& wavelength);
		bool calculateOpticalCellGasCrossSections(struct OpticalCell_s *pOC, const double& wavelength);

		bool createActiveConstituentLists();

		void formatAirDataFilepath(char *filepath) {formatFilePath(filepath, fn_air_data);}
		void formatAirMapFilepath(char *filepath) {formatFilePath(filepath, fn_air_map);}
		
		inline char *getAirDataFilename(){return fn_air_data;}
		inline char *getAirMapFilename(){return fn_air_map;}

		void formatRayleighCrossSectionFilepath(char *filepath) {formatFilePath(filepath, fn_rayleigh_cross_section);}
		
		// inline bool useExternalRayleighCrossSection(){return RCSFormula==RCS_file;}

		inline char *getGasesList(){return gases;}
		inline int getNOListedGases(){return !strcmp(gases, "none")?0:numberOf(',', gases)+1;}

		inline char *getParticleList(){return particles;}
		inline int getNOListedParticles(){return !strcmp(particles, "none")?0:numberOf(',', particles)+1;}

		// bool writeStatistics();

};

#endif//_Atmosphere_h



