/**************************************************************************************************************
*
*       This file,
*
*                            /BinaryGridFile.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "BinaryGridFile.h"
#include "Depiction/EPSFile.h"
#include "error.h"
#include "fileio.h"

#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <limits.h>

BinaryGridFile::BinaryGridFile()
{
	data=(double *)0;
	additional_data=(void *)0;

	// plane stuff
	GridPlane=Z_Plane;
	x_plane_index=0;
	y_plane_index=0;
	z_plane_index=0;
	layer_index=0;

	memset(&FileHeader, 0, sizeof(struct FileHeader_s));
	FileHeader.additional_datasize=0;
	FileHeader.code[0]='b';
	FileHeader.code[1]='l';
	FileHeader.code[2]='f';
}

BinaryGridFile::~BinaryGridFile()
{
	freeBinaryGridFile();
}

void BinaryGridFile::freeBinaryGridFile()
{
	if(data)
	{
		delete [] data;
		data=(double *)0;
	}
	additional_data=(void *)0;
}

bool BinaryGridFile::allocate(int X_Res, int Y_Res, int Z_Res, int no_layers)
{
	int n=(X_Res+Y_Res+Z_Res+3)+no_layers*X_Res*Y_Res*Z_Res;
	if(data) delete [] data;
	data=new double[n];
	if(!data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	memset(data,0, sizeof(double)*n);
	FileHeader.X_Res=X_Res;
	FileHeader.Y_Res=Y_Res;
	FileHeader.Z_Res=Z_Res;
	FileHeader.no_layers=no_layers;
	FileHeader.datasize=sizeof(double)*n;
	FileHeader.additional_datasize=0;
	Grid.X=data;
	Grid.Y=Grid.X+X_Res+1;
	Grid.Z=Grid.Y+Y_Res+1;
	Grid.cell=Grid.Z+Z_Res+1;

	GridPlane=Z_Plane;
	x_plane_index=0;
	y_plane_index=0;
	z_plane_index=0;
	layer_index=0;

	return true;
}

bool BinaryGridFile::initializeFromGrid(class Grid *pGrid, int no_layers)
{
	int XRes,YRes,ZRes;
	class Grid1D *grid1D=pGrid->getExternalGrid1DPtr();
	XRes=grid1D[0].getResolution();
	YRes=grid1D[1].getResolution();
	ZRes=grid1D[2].getResolution();

	if(!allocate(XRes,YRes,ZRes,no_layers))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}

	memcpy(Grid.X,grid1D[0].getGridPtr(),sizeof(double)*(1+FileHeader.X_Res));
	memcpy(Grid.Y,grid1D[1].getGridPtr(),sizeof(double)*(1+FileHeader.Y_Res));
	memcpy(Grid.Z,grid1D[2].getGridPtr(),sizeof(double)*(1+FileHeader.Z_Res));
	return true;
}

void BinaryGridFile::defineAdditionalData(void *additional_dataptr, int sizeof_additional_data)
{
	additional_data=additional_dataptr;
	FileHeader.additional_datasize=sizeof_additional_data;
}

bool BinaryGridFile::writeToFile(char *filename)
{
	FILE *fptr=fopen(filename,"wb");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing to file \"%s\"\n", filename);
		return false;
	}
	bool success=writeToFile(fptr);
	fclose(fptr);
	return success;
}

bool BinaryGridFile::writeToFile(FILE *fptr)
{
	if(fwrite(&FileHeader, sizeof(struct FileHeader_s), 1,fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing header\n");
		return false;
	}
	if(FileHeader.additional_datasize>0)
		if(fwrite(additional_data, FileHeader.additional_datasize, 1, fptr)!=1)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error writing additional data\n");
			return false;
		}
	if(fwrite(data, FileHeader.datasize, 1, fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing data\n");
		return false;
	}
	return true;
}

bool BinaryGridFile::loadFromFile(char *filename)
{
	//fprintf(
	FILE *fptr=fopen(filename,"rb");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading binary grid data from file \"%s\"\n", filename);
		return false;
	}
	bool success=loadFromFile(fptr);
	fclose(fptr);
	return success;
}

bool BinaryGridFile::loadFromFile(FILE *fptr)
{
	//struct FileHeader_s FileHeader;
	if(fread(&FileHeader,sizeof(struct FileHeader_s),1,fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading header\n");
		return false;
	}
	if(FileHeader.code[0]!='b' || FileHeader.code[1]!='l' || FileHeader.code[2]!='f')
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file is not a binary grid file!\n");
		return false;
	}
	if(!allocate(&FileHeader))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating grid data according to header!\n");
		return false;
	}
	//this->FileHeader.additional_datasize=FileHeader.additional_datasize;
	if(FileHeader.additional_datasize)
		fseek(fptr, FileHeader.additional_datasize, SEEK_CUR); // skip additional data...
	if(fread(data, this->FileHeader.datasize, 1, fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading %d data elements!\n", this->FileHeader.datasize);
		return false;
	}
	return true;
}

bool BinaryGridFile::loadAdditionalData(FILE *fptr)
{
	if(!additional_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"additional data ptr is null.\n");
		return false;
	}
	long filepos=ftell(fptr);
	fseek(fptr, sizeof(struct FileHeader_s), SEEK_SET);
	bool success=true;
	success=fread(additional_data, FileHeader.additional_datasize,1,fptr)==1;
	fseek(fptr,filepos,SEEK_SET);
	return success;
}

bool BinaryGridFile::setPlaneAndIndex(enum GridPlane_e Plane, int index)
{
	GridPlane=Plane;
	switch(Plane)
	{
		case X_Plane:
			xres=FileHeader.Y_Res;
			yres=FileHeader.Z_Res;
			break;

		case Y_Plane:
			xres=FileHeader.X_Res;
			yres=FileHeader.Z_Res;
			break;

		case Z_Plane:
			xres=FileHeader.X_Res;
			yres=FileHeader.Y_Res;
			break;
	}
	return setPlaneIndex(index);
}

bool BinaryGridFile::setPlaneIndex(int index)
{
	bool allowed=false;
	switch(GridPlane)
	{
		case X_Plane:
			if(index>=0 && index<FileHeader.X_Res)
			{
				x_plane_index=index;
				allowed=true;
			}
			break;

		case Y_Plane:
			if(index>=0 && index<FileHeader.Y_Res)
			{
				y_plane_index=index;
				allowed=true;
			}
			break;

		case Z_Plane:
			if(index>=0 && index<FileHeader.Z_Res)
			{
				z_plane_index=index;
				allowed=true;
			}
			break;
	}
	return allowed;
}

bool BinaryGridFile::setLayerIndex(int index)
{
	if(index<0 || index>=FileHeader.no_layers)
	{	
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "index %d out of range!\n", index);
		return false;
	}
	layer_index=index;
	return true;
}

void BinaryGridFile::getPlaneData(double *buffer)
{	
	int x,y, index, ix,iy,iz, Rx,Ry, Rl;
	Rx=FileHeader.X_Res;
	Ry=FileHeader.Y_Res;
	// Rz=FileHeader.Z_Res;
	Rl=FileHeader.no_layers;
	switch(GridPlane)
	{
		case X_Plane:
			ix=x_plane_index;
			for(y=0;y<FileHeader.Z_Res;y++)
			{
				iz=y;
				for(x=0;x<FileHeader.Y_Res;x++)
				{
					iy=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					*buffer++=Grid.cell[index];
				}
			}
			break;

		case Y_Plane:
			iy=y_plane_index;
			for(y=0;y<FileHeader.Z_Res;y++)
			{
				iz=y;
				for(x=0;x<FileHeader.X_Res;x++)
				{
					ix=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					*buffer++=Grid.cell[index];
				}
			}
			break;

		case Z_Plane:
			iz=z_plane_index;
			for(y=0;y<FileHeader.Y_Res;y++)
			{
				iy=y;
				for(x=0;x<FileHeader.X_Res;x++)
				{
					ix=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					*buffer++=Grid.cell[index];
				}
			}
			break;
	}
}

void BinaryGridFile::setPlaneData(double *buffer)
{	
	int x,y, index, ix,iy,iz, Rx,Ry, Rl;
	Rx=FileHeader.X_Res;
	Ry=FileHeader.Y_Res;
	Rl=FileHeader.no_layers;
	switch(GridPlane)
	{
		case X_Plane:
			ix=x_plane_index;
			for(y=0;y<FileHeader.Z_Res;y++)
			{
				iz=y;
				for(x=0;x<FileHeader.Y_Res;x++)
				{
					iy=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					Grid.cell[index]=*buffer++;
				}
			}
			break;

		case Y_Plane:
			iy=y_plane_index;
			for(y=0;y<FileHeader.Z_Res;y++)
			{
				iz=y;
				for(x=0;x<FileHeader.X_Res;x++)
				{
					ix=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					Grid.cell[index]=*buffer++;
				}
			}
			break;

		case Z_Plane:
			iz=z_plane_index;
			for(y=0;y<FileHeader.Y_Res;y++)
			{
				iy=y;
				for(x=0;x<FileHeader.X_Res;x++)
				{
					ix=x;
					index=layer_index+Rl*(ix+Rx*(iy+Ry*iz));
					Grid.cell[index]=*buffer++;
				}
			}
			break;
	}
}

#ifdef WIN32

int __cdecl int_compare(const void *elem1, const void *elem2)
{
	return *((int *)elem1)-*((int *)elem2);
}

#else

int int_compare(const void *elem1, const void *elem2)
{
	return *((int *)elem1)-*((int *)elem2);
}

#endif

int BinaryGridFile::getMaxSliceBufferEntries()
{
	int R[3];
	R[0]=FileHeader.X_Res;
	R[1]=FileHeader.Y_Res;
	R[2]=FileHeader.Z_Res;
	qsort(R,3,sizeof(int),&int_compare);
	return R[2]*R[1];
}

void BinaryGridFile::getMaxSliceBufferResolution(int *pXRes, int *pYRes)
{
	int R[3];
	R[0]=FileHeader.X_Res;
	R[1]=FileHeader.Y_Res;
	R[2]=FileHeader.Z_Res;
	qsort(R,3,sizeof(int),&int_compare);
	*pXRes=R[2];
	*pYRes=R[2];
}


bool BinaryGridFile::isValid()
{
	return data?true:false;
}

void BinaryGridFile::writeFileHeaderInfo(FILE *fptr)
{
	fprintf(fptr,"no_layers=%d\n",FileHeader.no_layers);
	fprintf(fptr,"X_Res=%d\n",FileHeader.X_Res);
	fprintf(fptr,"Y_Res=%d\n",FileHeader.Y_Res);
	fprintf(fptr,"Z_Res=%d\n",FileHeader.Z_Res);
	fprintf(fptr,"additional_datasize=%d\n",FileHeader.additional_datasize);
	fprintf(fptr,"datasize=%d\n",FileHeader.datasize);
}

bool BinaryGridFile::exportToFilename(char *filename, const char *formatName, bool (BinaryGridFile::*exportToFPtr)(FILE *fptr))
{
	FILE *fptr=fopen(filename,"w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not export to file \"%s\" (format: %s)",filename, formatName);
		return false;
	}
	bool success=(this->*exportToFPtr)(fptr);
	fclose(fptr);
	return success;
}

bool BinaryGridFile::exportSliceToEPS(char *filename, bool logarithmic_scale)
{
	int nt=getMaxSliceBufferEntries();
	double *total_slice_data=new double[2*nt];
	if(!total_slice_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "slice data allocation error\n");
		return false;
	}
	getPlaneData(total_slice_data);
	char buffer[200];
	int txres,tyres,z,i,n,xres,yres;
	getGridPlaneInformation(GridPlane, &txres, &tyres, &z, buffer);
	// copy data to reduced slice....
	double *slice_data=total_slice_data+nt, *valueptr, *source_valueptr;
	xres=txres-2;
	yres=tyres-2;
	n=xres*yres;
	if(n<=0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"slice has negative dimensions!\n");
		return false;
	}
	// create reduced slice...
	source_valueptr=total_slice_data+txres+1;
	valueptr=slice_data;
	for(i=0;i<yres;i++)
	{
		memcpy(valueptr, source_valueptr, sizeof(double)*xres);
		source_valueptr+=txres;
		valueptr+=xres;
	}
	//fprintf(fptr,"# dataset %d, %s", layer_index+1, buffer);
	// #################################### prepare EPS file...
	class EPSFile EPSOutputFile;
	// define EPS file header:
	struct EPSFile::Header_s *pEPSFileHeader=EPSOutputFile.getHeaderPtr();
	pEPSFileHeader->bb_x1=0;
	pEPSFileHeader->bb_y1=0;
	pEPSFileHeader->bb_x2=xres;
	pEPSFileHeader->bb_y2=yres;
	strcpy(pEPSFileHeader->creator,"Tim Deutschmann, IUP Heidelberg");
	time_t long_time;
	time(&long_time);
	strcpy(pEPSFileHeader->date, asctime(localtime(&long_time)));
	if(!EPSOutputFile.open(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not create EPS file \"%s\"!\n", filename);
		return false;
	}
	// normalisation
	double min_value=DBL_MAX, max_value=DBL_MIN, value;
	valueptr=slice_data;
	for(i=0;i<n;i++)
	{
		value=*valueptr++;
		if(value<min_value)
			min_value=value;
		if(value>max_value)
			max_value=value;
	}
	double delta_min_max=max_value-min_value;
	// convert to [0,1]
	valueptr=slice_data;
	for(i=0;i<n;i++)
	{
		value=*valueptr;
		*valueptr++=(value-min_value)/delta_min_max;
	}
	//printf("min=%lf, max=%lf\n", min_value, max_value);
	if(logarithmic_scale)
	{
		double log_min_value=DBL_MAX, log_max_value=DBL_MIN, delta_log_min_max;
		valueptr=slice_data;
		for(i=0;i<n;i++)
		{
			value=log(min_value+*valueptr++);
			if(value<log_min_value)
				log_min_value=value;
			if(value>log_max_value)
				log_max_value=value;
		}
		delta_log_min_max=log_max_value-log_min_value;
		//printf("log_min=%lf, log_max=%lf\n", log_min_value, log_max_value);
		// calculate logarithmic depiction
		valueptr=slice_data;
		for(i=0;i<n;i++)
		{
			value=log(min_value+*valueptr);
			*valueptr++=(value-log_min_value)/delta_log_min_max;
		}
	}
	EPSOutputFile.greyImage(0,0,xres,yres,slice_data);
	delete [] total_slice_data;
	EPSOutputFile.close();
	return true;
}

bool BinaryGridFile::exportSliceToASCII(FILE *fptr)
{
	int n=getMaxSliceBufferEntries();
	double *slice_data=new double[n];
	if(!slice_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"allocation error\n");
		return false;
	}
	getPlaneData(slice_data);
	char buffer[200];
	int xres,yres,x,y,index=0,z;
	getGridPlaneInformation(GridPlane, &xres, &yres, &z, buffer);
	fprintf(fptr,"# dataset %d, %s", layer_index+1, buffer);
	for(y=0;y<yres;y++)
	{
		for(x=0;x<xres-1;x++)
			fprintf(fptr,"%lg\t",slice_data[index++]);
		fprintf(fptr,"%lg\n",slice_data[index++]);
	}
	delete [] slice_data;
	return true;
}

bool BinaryGridFile::exportLayerToASCII(FILE *fptr)
{
	int n=getMaxSliceBufferEntries();
	double *slice_data=new double[n];
	if(!slice_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"allocation error\n");
		return false;
	}
	char buffer[200];
	int z,zres=getNumberOfPlanes(GridPlane),xres,yres,x,y,index,old_plane_index=getPlaneIndex();
	getGridPlaneInformation(GridPlane, &xres, &yres, &z, buffer);
	fprintf(fptr,"# dataset %d, %s\n", layer_index+1, buffer);
	fprintf(fptr,"# %d slices total\n", zres);
	for(z=0;z<zres;z++)
	{
		index=0;
		setPlaneIndex(z);
		getPlaneData(slice_data);
		fprintf(fptr,"# slice %d\n", z+1);
		for(y=0;y<yres;y++)
		{
			for(x=0;x<xres-1;x++)
				fprintf(fptr,"%lg\t",slice_data[index++]);
			fprintf(fptr,"%lg\n",slice_data[index++]);
		}
	}
	delete [] slice_data;
	// restore old plane:
	setPlaneIndex(old_plane_index);
	return true;
}

bool BinaryGridFile::exportToASCII(FILE *fptr)
{
	int n=getMaxSliceBufferEntries();
	double *slice_data=new double[n];
	if(!slice_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"allocation error\n");
		return false;
	}
	char buffer[200];
	int z,xres,yres,x,y,index=0;
	getGridPlaneInformation(GridPlane, &xres, &yres, &z, buffer);
	fprintf(fptr,"# dataset %d, %s, plane %d\n", layer_index+1, buffer, getPlaneIndex()+1);
	getPlaneData(slice_data);
	for(y=0;y<yres;y++)
	{
		for(x=0;x<xres-1;x++)
			fprintf(fptr,"%le\t",slice_data[index++]);
		fprintf(fptr,"%le\n",slice_data[index++]);
	}
	delete [] slice_data;
	return true;
}

void BinaryGridFile::getGridPlaneInformation(enum GridPlane_e Plane, int *pxres, int *pyres, int *pzindex, char *buffer)
{
	switch(GridPlane)
	{
		case X_Plane:
			if(pxres && pyres)
			{
				*pxres=FileHeader.Y_Res;
				*pyres=FileHeader.Z_Res;
			}
			if(*pzindex)
				*pzindex=x_plane_index;
			if(buffer)
				sprintf(buffer, "X slice %d with resolution %dx%d (Y x Z)",x_plane_index+1,FileHeader.Y_Res,FileHeader.Z_Res);
			break;

		case Y_Plane:
			if(pxres && pyres)
			{
				*pxres=FileHeader.X_Res;
				*pyres=FileHeader.Z_Res;
			}
			if(*pzindex)
				*pzindex=y_plane_index;
			if(buffer)
				sprintf(buffer, "Y slice %d with resolution %dx%d (X x Z)",y_plane_index+1,FileHeader.X_Res,FileHeader.Z_Res);
			break;

		case Z_Plane:
			if(pxres && pyres)
			{
				*pxres=FileHeader.X_Res;
				*pyres=FileHeader.Y_Res;
			}
			if(*pzindex)
				*pzindex=z_plane_index;
			if(buffer)
				sprintf(buffer,"Z slice %d with resolution %dx%d (X x Y)",z_plane_index+1,FileHeader.X_Res,FileHeader.Y_Res);
			break;
	}
}

void BinaryGridFile::getGridPlaneInformation(enum GridPlane_e Plane, int *pxres, int *pyres)
{
	switch(Plane)
	{
		case X_Plane:
			*pxres=FileHeader.Y_Res;
			*pyres=FileHeader.Z_Res;
			break;

		case Y_Plane:
			*pxres=FileHeader.X_Res;
			*pyres=FileHeader.Z_Res;
			break;

		case Z_Plane:
			*pxres=FileHeader.X_Res;
			*pyres=FileHeader.Y_Res;
			break;
	}
}

bool BinaryGridFile::Character2GridPlane(char character, enum GridPlane_e *pGridPlane)
{
	switch(character)
	{
		case '1':
		case 'x':
		case 'X':
			*pGridPlane=X_Plane;
			return true;
		case '2':
		case 'y':
		case 'Y':
			*pGridPlane=Y_Plane;
			return true;
		case '3':
		case 'z':
		case 'Z':
			*pGridPlane=Z_Plane;
			return true;
		default:;
	}
	return false;
}

bool BinaryGridFile::readTextProfile(FILE *fptr)
{
	double *cell_value;
	int i=0, j, no_layers, ix, iy, iz;
	char line_buffer[2047], word_buffer[1023];
	while(fgets(line_buffer, 2047, fptr))
	{
		extractWordFromList(line_buffer, 0, '\t', word_buffer);
		ix=atoi(word_buffer)-1;

		extractWordFromList(line_buffer, 1, '\t', word_buffer);
		iy=atoi(word_buffer)-1;

		extractWordFromList(line_buffer, 2, '\t', word_buffer);
		iz=atoi(word_buffer)-1;

		if(ix<0 || ix>=FileHeader.X_Res || \
			iy<0 || iy>=FileHeader.Y_Res || \
			iz<0 || iz>=FileHeader.Z_Res)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "line %d: index (%d,%d,%d) out of range!\n", i+1, ix+1, iy+1, iz+1);
			return false;
		}
		
		cell_value=Grid.cell+calculateIndex(ix, iy, iz);
		no_layers=numberOf('\t', line_buffer)-2;
		if(no_layers!=FileHeader.no_layers)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"error reading line %d: inconsistent number of layers (%d and %d)!\n", i+1, no_layers, FileHeader.no_layers);
			return false;
		}

		for(j=0;j<no_layers;j++)
		{
			extractWordFromList(line_buffer, 3+j, '\t', word_buffer);
			cell_value[j]=atof(word_buffer);
		}
		i++;
	}
	return true;
}

bool BinaryGridFile::loadTextProfileAndInitialize(char *filename)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\" for reading!\n", filename);
		return false;
	}
	// jobs to do:
	// 1.: analyse file
	// 1a.: get x/y/z res
	// 1a.: get number of layers	
	int i, no_rows=numberOfRows(fptr);
	int no_layers=numberOfColumns(fptr)-3;
	int ix,xres, iy,yres, iz,zres;
	int minx=INT_MAX,maxx=INT_MIN, miny=INT_MAX,maxy=INT_MIN, minz=INT_MAX,maxz=INT_MIN;
	char line_buffer[2047];
	for(i=0;i<no_rows;i++)
	{
		fscanf(fptr, "%d\t%d\t%d\t%s", &ix, &iy, &iz, line_buffer);
		if(ix<minx)
			minx=ix;
		if(ix>maxx)
			maxx=ix;

		if(iy<miny)
			miny=iy;
		if(iy>maxy)
			maxy=iy;

		if(iz<minz)
			minx=iz;
		if(iz>maxz)
			maxz=iz;
	}
	xres=1+maxx-minx;
	yres=1+maxy-miny;
	zres=1+maxz-minz;
	if(!allocate(xres,yres,zres,no_layers))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not allocate %dx%dx%dx%d memory block!\n", xres, yres, zres, no_layers);
		fclose(fptr);
		return false;
	}
	rewind(fptr);
	if(!readTextProfile(fptr))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fclose(fptr);
		fprintf(stderr, "error reading file \"%s\"!\n", filename);
		return false;
	}
	fclose(fptr);
	return true;
}

bool BinaryGridFile::loadTextProfile(char *filename)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
		return false;
	if(!readTextProfile(fptr))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fclose(fptr);
		fprintf(stderr, "error reading file \"%s\"!\n", filename);
		return false;
	}
	fclose(fptr);
	return true;
}



