/**************************************************************************************************************
*
*       This file,
*
*                            /BinaryGridFile.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _BinaryGridFile_h
#define _BinaryGridFile_h

#include "Grid.h"

class BinaryGridFile
{

public:
	// public structures:
	struct FileHeader_s
	{
		char code[3];	// "bgf"
		int no_layers;

		int X_Res, XA, XB;
		int Y_Res, YA, YB;
		int Z_Res, ZA, ZB;

		int additional_datasize;
		int datasize;

		FileHeader_s(){code[0]='b';code[1]='g';code[2]='f';}
	};

	struct Grid_s
	{
		double *X, *Y, *Z;
		double *cell;
	};

	enum GridPlane_e
	{
		X_Plane,
		Y_Plane,
		Z_Plane
	};

	//const char *GridPlaneName[]={"R","Phi","Theta"};

private:

	void freeBinaryGridFile();
	
	bool exportToFilename(char *filename, const char *formatName, bool (BinaryGridFile::*exportToFPtr)(FILE *));
public:

	BinaryGridFile();
	~BinaryGridFile();

	void getGridPlaneInformation(enum GridPlane_e Plane, int *pxres, int *pyres);

	inline int calculateIndex(int ix, int iy, int iz, int il)
		{return il+FileHeader.no_layers*(ix+FileHeader.X_Res*(iy+FileHeader.Y_Res*iz));}

	inline int calculateIndex(int ix, int iy, int iz)
		{return FileHeader.no_layers*(ix+FileHeader.X_Res*(iy+FileHeader.Y_Res*iz));}

	// void determineValueBoundaries(

	bool allocate(int X_Res, int Y_Res, int Z_Res, int no_layers);
	inline bool allocate(struct FileHeader_s *pFileHeader){return allocate(pFileHeader->X_Res, pFileHeader->Y_Res, pFileHeader->Z_Res, pFileHeader->no_layers);}

	void getMaxSliceBufferResolution(int *pXRes, int *pYRes);

	// memory and allocation
	bool initializeFromGrid(class Grid *pGrid, int no_layers);
	void defineAdditionalData(void *additional_dataptr, int sizeof_additional_data);

	// IO
	bool writeToFile(char *filename), writeToFile(FILE *fptr);
	bool loadFromFile(char *filename), loadFromFile(FILE *fptr);
	bool loadTextProfile(char *filename);
	bool readTextProfile(FILE *fptr);
	bool loadTextProfileAndInitialize(char *filename);
	bool loadAdditionalData(FILE *fptr);

	// plane and layer functions
	bool setPlaneAndIndex(enum GridPlane_e Plane, int index);
	bool setPlaneIndex(int index);

	bool setLayerIndex(int index);

	void getPlaneData(double *buffer);
	void setPlaneData(double *buffer);

	int getMaxSliceBufferEntries();

	void writeFileHeaderInfo(FILE *fptr);

	bool Character2GridPlane(char character, enum GridPlane_e *pGridPlane);

	// information
	bool isValid();

	// ASCII
	bool exportSliceToASCII(FILE *fptr);
	bool exportLayerToASCII(FILE *fptr);

	bool exportToASCII(FILE *fptr);

	inline bool exportSliceToASCII(char *filename){return exportToFilename(filename, "ASCII", &BinaryGridFile::exportSliceToASCII);}
	inline bool exportLayerToASCII(char *filename){return exportToFilename(filename, "ASCII", &BinaryGridFile::exportLayerToASCII);}

	inline bool exportToASCII(char *filename){return exportToFilename(filename, "ASCII", &BinaryGridFile::exportToASCII);}

	// EPS
	bool exportSliceToEPS(char *filename, bool logarithmic_scale);

	// inline functions:

	// internal structures access
	inline struct Grid_s *getGridPtr(){return &Grid;}
	inline struct FileHeader_s *getFileHeaderPtr(){return &FileHeader;}
	inline double *getCellPtr(){return Grid.cell;}

	// plane and layer functions
	inline void setPlane(enum GridPlane_e Plane){GridPlane=Plane;}
	inline enum GridPlane_e getPlane(){return GridPlane;}
	inline int getNumberOfPlanes(enum GridPlane_e Plane){return Plane==X_Plane?FileHeader.X_Res:(Plane==Y_Plane?FileHeader.Y_Res:FileHeader.Z_Res);}
	inline int getNumberOfPlanes(){return getNumberOfPlanes(GridPlane);}
	inline int getPlaneIndex(){return GridPlane==X_Plane?x_plane_index:(GridPlane==Y_Plane?y_plane_index:z_plane_index);}
	inline int getLayerIndex(){return layer_index;}

	inline int getXRes(){return xres;}
	inline int getYRes(){return yres;}

	inline int getGridXRes(){return FileHeader.X_Res;}
	inline int getGridYRes(){return FileHeader.Y_Res;}
	inline int getGridZRes(){return FileHeader.Z_Res;}

	inline int getCellSize(){return FileHeader.no_layers;}
	inline int getNOLayers(){return FileHeader.no_layers;}
	inline double getPlaneHeight(int iZ){return Grid.Z[iZ+1]-Grid.Z[iZ];}

protected:
	struct FileHeader_s FileHeader;
	struct Grid_s Grid;

	// plane stuff
	enum GridPlane_e GridPlane;
	int x_plane_index, y_plane_index, z_plane_index;
	int layer_index;
	int xres, yres;

	// data...
	double *data;
	void *additional_data;
	void getGridPlaneInformation(enum GridPlane_e Plane, int *pxres, int *pyres, int *pindex, char *buffer);
};

#endif //BinaryGridFile



