/**************************************************************************************************************
*
*       This file,
*
*                            /BoxDetector.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "BoxDetector.h"

#include "Math/prand.h"
#include "trigonometry.h"

#include "SurfaceGrid.h"
#include "constants.h"
#include "error.h"

bool BoxDetector::assignDefinition(char *box_definition)
{
	if(numberOf('@', box_definition)!=1 || numberOf('!', box_definition)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"wrong box detector definition : \"%s\"\ncorrect format: ", box_definition);
		writeDefinitionTemplate(stderr);
		return false;
	}
	char buffer[100];

	// extract type --------------------------------------------------------------
	extractWordFromList(box_definition, 0, '@', buffer);
	if(!strcmp(buffer,"fullbox"))
		SamplingElement=fullbox;
	else if(!strcmp(buffer,"boxface"))
		SamplingElement=boxface;
	else if(!strcmp(buffer,"boxnorthface"))
		SamplingElement=boxnorthface;
	else if(!strcmp(buffer,"boxsouthface"))
		SamplingElement=boxsouthface;
	else if(!strcmp(buffer,"boxeastface"))
		SamplingElement=boxeastface;
	else if(!strcmp(buffer,"boxwestface"))
		SamplingElement=boxwestface;
	else if(!strcmp(buffer,"boxtopface"))
		SamplingElement=boxtopface;
	else if(!strcmp(buffer,"boxbottomface"))
		SamplingElement=boxbottomface;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"unknown box element :\"%s\"\n", buffer);
		writeDefinitionTemplate(stderr);
		return false;
	}

	// extract emissivity -----------------------------------------------------------
	extractWordFromList(box_definition, 1, '!', buffer);
	switch(*buffer)
	{
		case '-':
			b_invert_faces=true;
			break;
		case '+':
			b_invert_faces=false;
			break;
		default:
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"sign in \"%s\" is missing!\n", buffer);
			writeDefinitionTemplate(stderr);
			return false;
	}

/*	//-------------------------------------------------------------------------------
	if(!strcmp(buffer+1,"diffuse"))
		FaceEmissivity=diffuse;
	else if(!strcmp(buffer+1,"normal"))
		FaceEmissivity=normal;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"unknown face emissivity type :\"%s\"\n", buffer);
		writeDefinitionTemplate(stderr);
		return false;
	}*/

	// box definition ---------------------------------------------------------------
	char definition_buffer[100];
	extractWordFromList(box_definition, 1, '@', buffer);
	extractWordFromList(buffer, 0, '!', definition_buffer);

	if(!EmissionCell.assignDefinition(definition_buffer))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"invalid box definition: \"%s\"!\n", definition_buffer);
		return false;
	}
	// check if
	switch(SamplingElement)
	{
		default:
		case fullbox:
		case boxface:
		case boxtopface:
		case boxbottomface:
			break;

		case boxnorthface:
		case boxsouthface:
		case boxeastface:
		case boxwestface:
			if(pGrid->getGridDimension()!=3)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"box detector type \"%s\" is incompatible with grid!\n", getTypeString(SamplingElement));
				return false;
			}
			break;
	}

	// calculateRuntimeConstants();
	pGrid->initializeBoxSamplingStructure(&EmissionCell, &SamplingStructure);
	return true;
}

const char *BoxDetector::getTypeString(enum SamplingElement_e be)
{
	switch(be)
	{
		case fullbox: 		return "fullbox";
		case boxface: 		return "boxface";
		case boxnorthface:	return "boxnorthface";
		case boxsouthface:	return "boxsouthface";
		case boxeastface:	return "boxeastface";
		case boxwestface:	return "boxwestface";
		case boxtopface:	return "boxtopface";
		case boxbottomface:	return "boxbottomface";
		default: ;
	}
	return (const char *)0;
}

double BoxDetector::getArea()
{
	switch(SamplingElement)
	{
		case fullbox:
			return 1.0;
		default:
			return 1e6*SamplingStructure.size[(int)SamplingElement];
	}
}

/*const char *BoxDetector::getFaceEmissivityString(enum FaceEmissivity_e fe)
{
	switch(fe)
	{
		case diffuse:	return "diffuse";
		case normal:	return "normal";
		default: ;
	}
	return (const char *)0;
}*/

void BoxDetector::writeDefinitionTemplate(FILE *fptr)
{
	fprintf(fptr,"fullbox|boxface|boxtopface|boxbottomface|boxnorthface|boxsouthface|boxeastface|boxwestface @ a~b:*:c ! +|- \n");
}

inline void BoxDetector::sampleDirection(class Vector3D& direction)
{
	double phi=prand()*2.0*M_PI;
	double cos_phi=cos(phi);
	double sin_phi=sin(phi);

	double cos_theta=correctMu(2.0*prand()-1.0);
	double sin_theta=sqrt(1-cos_theta*cos_theta);

	direction[0]=cos_phi*sin_theta;
	direction[1]=sin_phi*sin_theta;
	direction[2]=cos_theta;
}

bool BoxDetector::takeEmissionSample(class Vector3D *pPosition, class Vector3D *pDirection)
{
	// class Vector3D position, direction, normal, vx, vy;
	// double cos_phi, sin_phi, phi, cos_theta, sin_theta, r;
	bool sample_from_face=false;
	enum SamplingElement_e face;
	class Vector3D face_normal;
	double x, y, z;
	switch(SamplingElement)
	{
		case fullbox:
			// get random point inside the box
			x=SamplingStructure.takeXSample(SamplingStructure.Nx, SamplingStructure.cx);
			y=SamplingStructure.takeYSample(SamplingStructure.Ny, SamplingStructure.cy);
			z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
			pGrid->convertToVector3D(x, y, z, *pPosition);
			// direction
			sampleDirection(*pDirection);
			return true;

		case boxface:
		{
			// select face
			face=boxwestface;
			double a=prand()*SamplingStructure.size[(int)boxface];
			for(int i=(int)boxtopface;i<=(int)boxwestface;i++)
				if((a-=SamplingStructure.size[i])<0.0)
				{
					face=(enum SamplingElement_e)i;
					break;
				}
			sample_from_face=true;
			break;
		}

		default:
			face=SamplingElement;
			sample_from_face=true;
			break;
	}
	double f=b_invert_faces?-1.0:1.0;
	if(sample_from_face)
		switch(face)
		{
			case boxtopface:
				x=SamplingStructure.takeXSample(SamplingStructure.Nx, SamplingStructure.cx);
				y=SamplingStructure.takeYSample(SamplingStructure.Ny, SamplingStructure.cy);
				z=SamplingStructure.z2+f*SamplingStructure.dz;
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateAltitudeAxis(x, y, z, face_normal);
				face_normal*=f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;

			case boxbottomface:
				x=SamplingStructure.takeXSample(SamplingStructure.Nx, SamplingStructure.cx);
				y=SamplingStructure.takeYSample(SamplingStructure.Ny, SamplingStructure.cy);
				z=SamplingStructure.z1-f*SamplingStructure.dz;
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateAltitudeAxis(x, y, z, face_normal);
				face_normal*=-f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;

			case boxnorthface:
				if(cartesic_grid)
				{
					x=SamplingStructure.x1-f*SamplingStructure.dx;
					y=SamplingStructure.takeXSample(SamplingStructure.Ny, SamplingStructure.cy);
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				else
				{
					x=SamplingStructure.takeXSample(SamplingStructure.Nx, SamplingStructure.cx);
					y=SamplingStructure.y1-f*SamplingStructure.dy;
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateSouthAxis(x, y, z, face_normal);
				face_normal*=-f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;

			case boxsouthface:
				if(cartesic_grid)
				{
					x=SamplingStructure.x2+f*SamplingStructure.dx;
					y=SamplingStructure.takeXSample(SamplingStructure.Ny, SamplingStructure.cy);
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				else
				{
					x=SamplingStructure.takeXSample(SamplingStructure.Nx, SamplingStructure.cx);
					y=SamplingStructure.y2+f*SamplingStructure.dy;
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateSouthAxis(x, y, z, face_normal);
				face_normal*=f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;

			case boxwestface:
				if(cartesic_grid)
				{
					x=SamplingStructure.takeYSample(SamplingStructure.Nx, SamplingStructure.cx);
					y=SamplingStructure.y1-f*SamplingStructure.dy;
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				else
				{
					x=SamplingStructure.x1-f*SamplingStructure.dx;
					y=SamplingStructure.takeYSample(SamplingStructure.Ny, SamplingStructure.cy);
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateEastAxis(x, y, z, face_normal);
				face_normal*=-f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;

			case boxeastface:
				if(cartesic_grid)
				{
					x=SamplingStructure.takeYSample(SamplingStructure.Nx, SamplingStructure.cx);
					y=SamplingStructure.y2+f*SamplingStructure.dy;
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				else
				{
					x=SamplingStructure.x1+f*SamplingStructure.dx;
					y=SamplingStructure.takeYSample(SamplingStructure.Ny, SamplingStructure.cy);
					z=SamplingStructure.takeZSample(SamplingStructure.Nz, SamplingStructure.cz);
				}
				pGrid->convertToVector3D(x, y, z, *pPosition);
				
				pGrid->calculateEastAxis(x, y, z, face_normal);
				face_normal*=f;
				SurfaceGrid::sampleLambertianScatterDirection(face_normal, *pDirection);
				break;
			default:;
		}
	return true;
}






