/**************************************************************************************************************
*
*       This file,
*
*                            /Camera.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Camera.h"
#include "Depiction/EPSFile.h"

#include <float.h>
#include <time.h>

#include "error.h"

void Camera::setupBasicIO()
{
	addParameter(&xres, type_int, "xres", 0, "vertical resolution of the image/camera?");
	addParameter(&yres, type_int, "yres", 0, "horizontal resolution of the image/camera?");
	addParameter(fn_camera_rgb_sensitivity, type_string, "fn_camera_rgb_sensitivity", 0, "nm<TAB>r<TAB>g<TAB>b");
	addParameter(fn_simulation, type_string, "fn_simulation", 0, "simulation.conf path");
	addParameter(fn_outfile, type_string, "fn_outfile", 0, "output file");
	addParameter(fn_outbase, type_string, "fn_outbase", 0, "directory to put auxiliary files");
	addParameter(&b_logarithmic_saturation, type_bool, "b_logarithmic_saturation", 0, "calculate saturation logarithmically?");
	addParameter(&b_output_auxiliary_files, type_bool, "b_output_auxiliary_files", 0, "output auxiliary files?");
}

Camera::Camera(class Simulation *SimulationPtr)
{
	DoubleRGB=(struct DoubleRGB_s *)0;
	RGBSensitivity=(struct RGBSensitivity_s *)0;
	RGB=(struct RGB_s *)0;
	pSimulation=SimulationPtr;
	pGeometry=pSimulation->getGeometryPtr();
	pRaytracer=pSimulation->getRaytracerPtr();
	setupBasicIO();
}

Camera::~Camera()
{
	if(DoubleRGB)
		delete [] DoubleRGB;
	if(RGBSensitivity)
		delete [] RGBSensitivity;
	if(RGB)
		delete [] RGB;
}

bool Camera::allocateScreen()
{
	// allocation
	int n=xres*yres;
	if(DoubleRGB)
		delete [] DoubleRGB;

	DoubleRGB=new struct DoubleRGB_s[n];
	if(RGB)
		delete [] RGB;
	RGB=new struct RGB_s[n];
	if(!RGB || !DoubleRGB)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	memset(DoubleRGB, 0, sizeof(DoubleRGB_s)*n);
	return true;
}

bool Camera::loadSensitivityFunctions()
{
	FILE *fptr=fopen(fn_camera_rgb_sensitivity, "r");
	if(!fptr)
		return false;
	spectral_resolution=numberOfRows(fptr);
	if(RGBSensitivity)
		delete [] RGBSensitivity;
	RGBSensitivity=new struct RGBSensitivity_s[spectral_resolution];
	if(!RGBSensitivity)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fclose(fptr);
		return false;
	}
	int i;
	for(i=0;i<spectral_resolution;i++)
		if(fscanf(fptr, "%lg\t%lg\t%lg\t%lg", &(RGBSensitivity[i].wavelength_nm), \
			&(RGBSensitivity[i].DoubleRGB.r), \
			&(RGBSensitivity[i].DoubleRGB.g), \
			&(RGBSensitivity[i].DoubleRGB.b))!=4)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d!\n", i+1);
			fclose(fptr);
			return false;
		}
	fclose(fptr);
	return true;
}

bool Camera::calculateScreenPixelDirections()
{
	// calculate directions
	int x, y;
	class Vector3D direction;
	double alpha_x, fx=atan(0.5*pGeometry->getOmegaHorizontal());
	double alpha_y, fy=atan(0.5*pGeometry->getOmegaVertical());
	bool is_spherical_grid=pSimulation->getGridPtr()->isSphericalGrid();
	for(y=0;y<yres;y++)
	{
		alpha_y=fy*(1.0-double(1+2*y)/double(yres-1));
		for(x=0;x<xres;x++)
		{
			alpha_x=fx*(-1.0+double(1+2*x)/double(xres-1));
			if(is_spherical_grid)
			{
				direction=norm(Vector3D(tan(alpha_x), 1.0, tan(alpha_y)));
				pGeometry->rotateIntoDetectorFOVSpherical(direction);
			}
			else
			{
				direction=norm(Vector3D(1.0, -tan(alpha_x), tan(alpha_y)));
				pGeometry->rotateIntoDetectorFOVCartesic(direction);
			}
			DoubleRGB[y*xres+x].direction=direction;
		}
	}
	return true;
}

bool Camera::setup()
{
	if(!pSimulation->readAndLoadFiles(fn_simulation) || !pSimulation->initialize())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"load simulation failed!\n");
		return false;
	}
	pRaytracer->enableSaveTrajectories(false);
	pGeometry->setFOVOmega(0.0);
	pGeometry->enableFixedInitialDirection();
	// pRaytracer->enableFixedInitialDirection();

	// allocations: screen and spectrum
	if(!allocateScreen())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not allocate memory for screen with resolution %dx%d!\n", xres, yres);
		return false;
	}
	if(!loadSensitivityFunctions())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not load camera rgb sensitivity functions from file \"%s\"!\n", fn_camera_rgb_sensitivity);
		return false;
	}
	return calculateScreenPixelDirections();
}

bool Camera::run(const enum Simulation::RTESolvingMethod_e& method)
{
	int x, y, i;
	struct DoubleRGB_s *pDoubleRGB;
	double wavelength_nm, intensity, r, g, b;
	char filename[2048];
	FILE *fptr=0;
	class Result result;
	
	for(i=0;i<spectral_resolution;i++)
	{
		wavelength_nm=RGBSensitivity[i].wavelength_nm;
		printf("%lgnm\n", wavelength_nm);
		pSimulation->setWavelengthNM(wavelength_nm);
		pSimulation->prepareOpticalMedium();
		r=RGBSensitivity[i].DoubleRGB.r;
		g=RGBSensitivity[i].DoubleRGB.g;
		b=RGBSensitivity[i].DoubleRGB.b;
		if(b_output_auxiliary_files)
		{
			sprintf(filename,"%s_wl%02d.txt", fn_outbase, i+1);
			fptr=fopen(filename,"w");
			if(!fptr)
			{	
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"could not open auxiliary data file \"%s\"!\n", filename);
				return false;
			}
			fprintf(fptr,"# lambda[nm]=%lg\n", wavelength_nm);
			fprintf(fptr,"# r=%lg\n", r);
			fprintf(fptr,"# g=%lg\n", g);
			fprintf(fptr,"# b=%lg\n", b);
		}
		pDoubleRGB=DoubleRGB;
		for(y=0;y<yres;y++)
			for(x=0;x<xres;x++)
			{
				pGeometry->setFixedInitialDirection(pDoubleRGB->direction);
				if(method==Simulation::MonteCarlo)
				{
					pRaytracer->run(&result);
					intensity=result.scalar_radiance.getMean();
				}
				else
					pRaytracer->runAnalyticalSSA(&intensity);
				pDoubleRGB->r+=intensity*r;
				pDoubleRGB->g+=intensity*g;
				pDoubleRGB->b+=intensity*b;
				if(b_output_auxiliary_files)
					fprintf(fptr, "%lg%c", intensity, x<xres-1?'\t':'\n');
				pDoubleRGB++;
			}
		if(fptr)
			fclose(fptr);
	}
	// search for maximum grey value
	pDoubleRGB=DoubleRGB;
	double grey_value, max_grey_value=DBL_MIN, min_grey_value=DBL_MAX, cf;
	for(y=0;y<yres;y++)
		for(x=0;x<xres;x++)
		{
			r=pDoubleRGB->r;
			g=pDoubleRGB->g;
			b=pDoubleRGB->b;
			grey_value=sqrt(r*r+g*g+b*b);
			if(grey_value>max_grey_value)
				max_grey_value=grey_value;
			else if(grey_value<min_grey_value)
				min_grey_value=grey_value;
			pDoubleRGB++;
		}
	double log_max_grey_value=log(max_grey_value);
	double log_min_grey_value=log(min_grey_value);
	// normalisation
	pDoubleRGB=DoubleRGB;
	struct RGB_s *pRGB=RGB;
	for(y=0;y<yres;y++)
		for(x=0;x<xres;x++)
		{
			r=pDoubleRGB->r;
			g=pDoubleRGB->g;
			b=pDoubleRGB->b;
			grey_value=sqrt(r*r+g*g+b*b);
			if(b_logarithmic_saturation)
				cf=(255.0*(log(grey_value)-log_min_grey_value))/((log_max_grey_value-log_min_grey_value)*grey_value);
			else
				cf=255.0/max_grey_value;
			pRGB->r=(unsigned char)(cf*r);
			pRGB->g=(unsigned char)(cf*g);
			pRGB->b=(unsigned char)(cf*b);
			pDoubleRGB++;
			pRGB++;
		}
	printf("%d\t%d\t%lg\t%lg\t%s\n", xres, yres, min_grey_value, max_grey_value, b_logarithmic_saturation?"log":"lin");
	return true;
}

bool Camera::writeResult()
{
	/*FILE *fptr=fopen(fn_outfile,"wb");
	if(!fptr)
	{
		fprintf(stderr,"could not write to file \"%s\"\n!", fn_outfile);
		return false;
	}
	fwrite(RGB, sizeof(struct RGB_s), xres*yres, fptr);
	fclose(fptr);
	return false;*/



	// map to data
	/*int i, n=xres*yres;
	unsigned int *data=new unsigned int[n];
	if(!data)
	{
		allocation_error(__FILE__, __LINE__);
		return false;
	}
	for(i=0;i<n;i++)
		data[i]=(RGB[i].r<<16)|(RGB[i].g<<8)|RGB[i].b;*/

	class EPSFile EPSOutputFile;
	// define EPS file header:
	struct EPSFile::Header_s *pEPSFileHeader=EPSOutputFile.getHeaderPtr();
	pEPSFileHeader->bb_x1=0;
	pEPSFileHeader->bb_y1=0;
	pEPSFileHeader->bb_x2=xres;
	pEPSFileHeader->bb_y2=yres;
	strcpy(pEPSFileHeader->creator,"IUP Heidelberg");
	time_t long_time;
	time(&long_time);
	strcpy(pEPSFileHeader->date, asctime(localtime(&long_time)));
	if(!EPSOutputFile.open(fn_outfile))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not create EPS file \"%s\"!\n", fn_outfile);
		// delete [] data;
		return false;
	}
	// EPSOutputFile.colorImage(0, 0, xres, yres, data);
	EPSOutputFile.colorImage(1, -1, xres, yres, (unsigned int *)RGB);
	EPSOutputFile.close();
	// delete [] data;
	return true;
}

