/**************************************************************************************************************
*
*       This file,
*
*                            /ClusterMap.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "fileio.h"
#include <stdlib.h>
#include <limits.h>

#include "ClusterMap.h"
#include "Grid.h"
#include "error.h"

ClusterMap::ClusterMap()
{
	map=0;
	no_clusters=0;
}

ClusterMap::~ClusterMap()
{
	if(map)
		delete [] map;
}

bool ClusterMap::load(char *file_name)
{
	// no_clusters MUST BE VALID HERE!!!!!!!!!!!!!!!!
	FILE *fptr=fopen(file_name,"r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\"!\n", file_name);
		return false;
	}
	no_cells=pGrid->getNOCells();
	if(map)
		delete [] map;
	map=new int[no_cells];
	if(!map)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		fclose(fptr);
		return false;
	}
	char line_buffer[1000], index_buffer[100];
	int grid_dimension=pGrid->getGridDimension(), i=0, cell_index_array[3], j, index;
	// set map to undefined:
	for(i=0;i<no_cells;i++)
		map[i]=-1;

	// read first line and check for ~ characters 
	fgets(line_buffer, 1000, fptr);
	if(numberOf('~', line_buffer)!=0 || numberOf(',', line_buffer)!=0 || numberOf(':', line_buffer)!=0)
	{
		int no_commas, k, l, no_cells_of_cluster, i1, i2, g[3]=
		{
			pGrid->mapCellRangeIndexToGrid1DIndex(0),
			pGrid->mapCellRangeIndexToGrid1DIndex(1),
			pGrid->mapCellRangeIndexToGrid1DIndex(2)
		}, I[3][2];
		class Grid1D *grid1D=pGrid->getExternalGrid1DPtr(), *pGrid1D;
		char range_buffer[20];
		double c1, c2;
		i=0;
		do
		{
			condenseLine(line_buffer);
			// format: x1~x2:y1~y2:z1~z2, x1~x2:y1~y2:z1~z2,...
			no_commas=numberOf(',', line_buffer);
			for(j=0;j<=no_commas;j++)
			{
				extractWordFromList(line_buffer, j, ',', index_buffer);
				// format: x1~x2:y1~y2:z1~z2
				if(numberOf(':', index_buffer)!=grid_dimension-1)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "error in line %d, v-slab %d (\"%s\"): wrong number of colons (grid dimension)!\n", i+1, j+1, index_buffer);
					fclose(fptr);
					return false;
				}
				no_cells_of_cluster=1;
				for(k=0;k<grid_dimension;k++)
				{
					extractWordFromList(index_buffer, k, ':', range_buffer);
					pGrid1D=grid1D+g[k];
					// format: c1~c2
					if(range_buffer[0]=='*' && range_buffer[1]=='\0')
					{
						c1=pGrid1D->getA();
						c2=pGrid1D->getB();
						i1=0;
						i2=pGrid1D->getResolution();
					}
					else if(sscanf(range_buffer, "%le~%le", &c1, &c2)!=2)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "error in line %d, v-slab %d (\"%s\"):\n", i+1, j+1, index_buffer);
						fprintf(stderr, "range no. %d (\"%s\") has wrong format (correct: c1~c2)\n", k+1, range_buffer); 
						fclose(fptr);
						return false;
					}
					if(!pGrid1D->findIndex(c1, &i1)\
						|| !pGrid1D->findIndex(c2, &i2)\
						|| i2<=i1)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "error in line %d, v-slab %d (\"%s\"):\n", i+1, j+1, index_buffer);
						fprintf(stderr, "range no. %d (\"%s\") is not representable by the %s grid\n", k+1, range_buffer, pGrid1D->getName()); 
						fclose(fptr);
						return false;
					}
					
					I[k][0]=i1;
					I[k][1]=i2;
					no_cells_of_cluster*=i2-i1;
				}
				for(l=0;l<no_cells_of_cluster;l++)
				{
					// calculate cell_index_array
					// i = iX + RX * ( iY + RY * iZ ) = iX + iY * RX + iZ * RX * RY
					// iX = i % Rx
					// iY = ( ( i - iX ) / RX ) % RY 		; rest = ( i - iX ) / RX
					// iZ = ( ( ( i - iX ) / RX ) - iY ) / RY = ( rest - iY ) / Ry	
					int R, rest=l;
					for(k=0;k<grid_dimension;k++)
					{
						R = I[k][1] - I[k][0];
						rest = ( rest - (cell_index_array[k] = rest % R) ) / R;
						cell_index_array[k]+=I[k][0];
					}
					map[pGrid->calculateCellIndex(cell_index_array)]=i;
				}
			}
			i++;
		} while(fgets(line_buffer, 1000, fptr));
	}
	else
	{
		//  grid indices and cluster indices
		i=0;
		do
		{
			condenseLine(line_buffer);
			if (numberOf('\t', line_buffer)!=grid_dimension)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in line %d of file \"%s\": grid cell index dimension mismatch!\n",i+1, file_name);
				fclose(fptr);
				return false;
			}
			for (j=0;j<grid_dimension;j++)
			{
				extractWordFromList(line_buffer, j, '\t', index_buffer);
				cell_index_array[j]=atoi(index_buffer)-1;
			}
			j=pGrid->calculateCellIndex(cell_index_array);
			extractWordFromList(line_buffer, grid_dimension, '\t', index_buffer);
			index=atoi(index_buffer)-1;
			if (index<0 || index>=no_clusters)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in line %d of file \"%s\": cluster index out of range [1, %d] mismatch!\n", i+1, file_name, no_clusters);
				fclose(fptr);
				return false;
			}
			map[j]=index;
			// cluster[index].no_cells++;
			i++;
		} while(fgets(line_buffer, 1000, fptr));
	}
	fclose(fptr);
	return true;
}

bool ClusterMap::determineNOClusters(char *file_name, int *pn)
{
	// no_clusters MUST BE VALID HERE!!!!!!!!!!!!!!!!
	FILE *fptr=fopen(file_name,"r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\"!\n", file_name);
		return false;
	}
	char line_buffer[1000], index_buffer[10];
	int i, index, max_index=INT_MIN, line=1;
	while(fgets(line_buffer, 1000, fptr))
	{
		condenseLine(line_buffer);
		i=numberOf('\t', line_buffer);
		extractWordFromList(line_buffer, i, '\t', index_buffer);
		index=atoi(index_buffer);
		if(index<=0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "index out of range in line %d of file \"%s\"!\n", line, file_name);
			fclose(fptr);
			return false;
		}
		if(index>max_index)
			max_index=index;
		line++;
	};
	fclose(fptr);
	*pn=max_index;
	return true;
}




