/**************************************************************************************************************
*
*       This file,
*
*                            /Depiction/PointCollection.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "PointCollection.h"
#include <memory.h>
#include <stdlib.h>
#include "../Math/Matrix.h"

PointCollection::PointCollection()
{
	/*point=(class Vector3D *)0;
	color=(unsigned int *)0;*/
	cp=(struct ColoredPoint_s *)0;
	no_points=0;
}

PointCollection::~PointCollection()
{
	if(cp)
		delete [] cp;
}

bool PointCollection::allocatePoints(const int& n)
{
	if(cp)
		delete [] cp;
	cp=new struct ColoredPoint_s[n];
	if(!cp)
		return false;
	no_points=n;
	return true;
}

bool PointCollection::addPointCollection(class PointCollection *pPC)
{
	if(isValid())
	{
		int old_no_points=no_points;
		struct ColoredPoint_s *old_cp=cp;
		cp=(struct ColoredPoint_s *)0;
		int n=no_points+pPC->getNOPoints();
		if(!allocatePoints(n))
			return false;
		// old content
		memcpy(cp, old_cp, old_no_points*sizeof(struct ColoredPoint_s));
		n=pPC->getNOPoints();
		memcpy(cp+old_no_points, pPC->getColoredPointBase(), n*sizeof(struct ColoredPoint_s));
	}
	else
	{
		int n=pPC->getNOPoints();
		if(!allocatePoints(n))
			return false;
		memcpy(cp, pPC->cp, n*sizeof(struct ColoredPoint_s));
	}
	return true;
}

void PointCollection::transformPoints(class Matrix3D &m)
{
	if(!no_points)
		return;
/*#if defined(WIN32) && not defined(MINGW)
	unsigned char FPU_state[108];
	double a1,a2,a3,b1,b2,b3,c1,c2,c3,x,y,z; //collumn vectors a,b,c; result vector (x,y,z)
	a1=m.a[0][0]; b1=m.a[0][1]; c1=m.a[0][2];
	a2=m.a[1][0]; b2=m.a[1][1]; c2=m.a[1][2];
	a3=m.a[2][0]; b3=m.a[2][1]; c3=m.a[2][2];
	void *data=(void *)cp;
	int n=no_points;
	__asm
	{	
		push eax
		push ecx
		push edi
		fnsave FPU_state
		fnclex
		mov ecx,n
		mov eax,data
		add eax,16	// start with z
		mov edi,eax
point_loop:		
		fld qword ptr [edi]
		sub edi,8
		fld qword ptr [edi]
		sub edi,8
		fld qword ptr [edi]
		//x
		fld c1
		fmul st(0),st(3)
		fld b1
		fmul st(0),st(3)
		fld a1 
		fmul st(0),st(3)
		fadd st(2),st(0)
		ffree st(0)
		fincstp
		fadd st(1),st(0)
		ffree st(0)
		fincstp
		fstp x
		//y		
		fld c2
		fmul st(0),st(3)
		fld b2
		fmul st(0),st(3)
		fld a2 
		fmul st(0),st(3)
		fadd st(2),st(0)
		ffree st(0)
		fincstp
		fadd st(1),st(0)
		ffree st(0)
		fincstp
		fstp y
		//x
		fld c3
		fmul st(0),st(3)
		fld b3
		fmul st(0),st(3)
		fld a3 
		fmul st(0),st(3)
		fadd st(2),st(0)
		ffree st(0)
		fincstp
		fadd st(1),st(0)
		ffree st(0)
		fincstp
		fstp z
		//aufrumen
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp
		//speichern
		fld z
		fld y
		fld x		
		fstp qword ptr [edi]
		add edi,8
		fstp qword ptr [edi]
		add edi,8
		fstp qword ptr [edi]
		add edi,32			// 24+4
		dec ecx
		cmp ecx,0
		je exit_point_loop
		jmp point_loop
exit_point_loop:
		frstor FPU_state
		pop edi
		pop ecx
		pop eax
	}
#else*/
	for(int i=0;i<no_points;i++)
		cp[i].point*=m;
//#endif
}

void PointCollection::translatePoints(class Vector3D& p)
{
	if(!no_points)
		return;
/*#if defined (WIN32) && not defined(MINGW)
	unsigned char FPU_state[108];
	double deltax=p.x[0],deltay=p.x[1],deltaz=p.x[2];
	void *data=(void *)cp;
	int n=no_points;
	__asm
	{	
		push eax
		push ecx
		push edi
		fnsave FPU_state
		fnclex
		fld deltaz
		fld deltay
		fld deltax

		mov ecx,n
		mov eax,data
		mov edi,eax
point_loop:
		// x
		fld qword ptr [edi]
		fadd st(0),st(1)
		fst qword ptr [edi]
		add edi,8
		ffree st(0)
		fincstp

		// y
		fld qword ptr [edi]
		fadd st(0),st(2)
		fst qword ptr [edi]
		add edi,8
		ffree st(0)
		fincstp

		// z
		fld qword ptr [edi]
		fadd st(0),st(3)
		fst qword ptr [edi]
		add edi,16		// 8+4+4
		ffree st(0)
		fincstp

		dec ecx
		cmp ecx,0
		je exit_point_loop
		jmp point_loop
exit_point_loop:
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp
		frstor FPU_state
		pop edi
		pop ecx
		pop eax
	}
#else*/
	for(int i=0;i<no_points;i++)
		cp[i].point+=p;
//#endif
}

int compare_x0_ColoredPoints(const void *cp1, const void *cp2)
{
	return ((struct ColoredPoint_s *)cp1)->point[0]>((struct ColoredPoint_s *)cp2)->point[0]?-1:1;
}

void PointCollection::sortByX0()
{
	qsort(cp, no_points, sizeof(struct ColoredPoint_s), compare_x0_ColoredPoints);
}

bool PointCollection::loadPointsFromTxtFile(char *file_path, const int& iX, const int& iY, const int& iZ, const int& iC)
{
	int i;
	class Matrix data;
	if(!data.load(file_path))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(iX<0 || iX>=data.getN() || iY<0 || iY>=data.getN() || iZ<0 || iZ>=data.getN() || iC<0 || iC>=data.getN())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "indices out of range!\n");
		return false;
	}
	if(!allocatePoints(data.getM()))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	for(i=0;i<no_points;i++)
	{
		cp[i].point[0]=data[i][iX];
		cp[i].point[1]=data[i][iY];
		cp[i].point[2]=data[i][iZ];
		cp[i].color=(unsigned int)(data[i][iC]);
	}
	return true;
}











