/**************************************************************************************************************
*
*       This file,
*
*                            /Depiction/PointCollectionRenderList.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "PointCollectionRenderList.h"
#include "../error.h"
#include <float.h>

void PointCollectionRenderList::renderPointCollections()
{
	pRenderScreen->clearScreen();
	list<PointCollection *>::iterator ppc;
	for(ppc=PCPtrList.begin();ppc!=PCPtrList.end();ppc++)
		if((*ppc)->isValid())
			pRenderScreen->renderPointCollection(*ppc);
}

void PointCollectionRenderList::translatePoints(class Vector3D& delta_p)
{
	if(PCPtrList.empty())
		return;
	list<PointCollection *>::iterator ppc;
	for(ppc=PCPtrList.begin();ppc!=PCPtrList.end();ppc++)
		(*ppc)->translatePoints(delta_p);
	perspective_vector[3]+=delta_p;
}

void PointCollectionRenderList::transformPoints(class Matrix3D& m)
{
	if(PCPtrList.empty())
		return;
	list<PointCollection *>::iterator ppc;
	for(ppc=PCPtrList.begin();ppc!=PCPtrList.end();ppc++)
		(*ppc)->transformPoints(m);
	for(int i=0;i<4;i++)
		perspective_vector[i]*=m;
}

void PointCollectionRenderList::resetPerspective()
{
	perspective_vector[0]=Vector3D(1.0,0.0,0.0);
	perspective_vector[1]=Vector3D(0.0,1.0,0.0);
	perspective_vector[2]=Vector3D(0.0,0.0,1.0);
	perspective_vector[3]=Vector3D(0.0,0.0,0.0);
}

void PointCollectionRenderList::restorePerspective()
{
	class Matrix3D m=Matrix3D(perspective_vector);
	class Vector3D oldT=perspective_vector[3];

	resetPerspective();

	transformPoints(m);
	translatePoints(oldT);
}

bool PointCollectionRenderList::savePerspective(char *file_name)
{
	return pRenderScreen->savePerspective(file_name);
}

bool PointCollectionRenderList::loadPerspective(char *file_name)
{
	if(!pRenderScreen->loadPerspective(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	restorePerspective();
	return true;
}

bool PointCollectionRenderList::exportToEPS(char *file_name, bool export_image)
{
	class PointCollection AllPoints;
	list<PointCollection *>::iterator ppc;
	for(ppc=PCPtrList.begin();ppc!=PCPtrList.end();ppc++)
		if(!AllPoints.addPointCollection(*ppc))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	pRenderScreen->exportToEPSFile(&AllPoints, file_name, export_image);
	return true;
}


class Vector3D PointCollectionRenderList::findClosestPointTo(const int& sx, const int& sy)
{
	if(PCPtrList.empty())
		return Vector3D(0.0, 0.0, 0.0);
	class Vector3D pmin;
	double rmin=DBL_MAX, r, dx, dy;
	struct ColoredPoint_s *cp;
	int i, n, x, y;
	list<PointCollection *>::iterator ppc;
	for(ppc=PCPtrList.begin();ppc!=PCPtrList.end();ppc++)
	{
		cp=(*ppc)->getColoredPointBase();
		n=(*ppc)->getNOPoints();
		for(i=0;i<n;i++)
		{
			pRenderScreen->toScreenPoint(cp[i].point, x, y);
			dx = x - sx;
			dy = y - sy;
			r = dx * dx + dy * dy;
			if(r < rmin)
			{
				rmin = r;
				pmin = cp[i].point;
			}
		}
	}
	return pmin;
}


