/**************************************************************************************************************
*
*       This file,
*
*                            /Depiction/RenderScreen.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "RenderScreen.h"
#include <memory.h>
#include <time.h>
#include <string.h>
#include "EPSFile.h"

RenderScreen::RenderScreen()
{
	data=(unsigned int *)0;
	z_buffer=(double *)0;
	pixel=(unsigned int **)0;
	stereo_mode=false;
	half_eye_distance=10.0;
	background_color=0;
}

RenderScreen::~RenderScreen()
{
	if(data)
	{
		delete [] data;
		data=(unsigned int *)0;
	}
	if(pixel)
	{
		delete [] pixel;
		pixel=(unsigned int **)0;
	}
	if(z_buffer)
	{
		delete [] z_buffer;
		z_buffer=(double *)0;
	}
}

bool RenderScreen::allocate(int xres, int yres, bool use_z_buffer)
{
	// pure data
	if(data)
		delete [] data;
	data=new unsigned int[xres*yres];
	if(!data)
			return false;
	// pixel array
	if(pixel)
		delete [] pixel;
	pixel=new unsigned int *[yres];
	if(!pixel)
		return false;
	for(int i=0;i<yres;i++)
		pixel[i]=data+i*xres;
	// z_buffer
	if(z_buffer)
	{
		delete [] z_buffer;
		z_buffer=(double *)0;
	}
	if(use_z_buffer)
	{
		int n=xres*yres;
		z_buffer=new double[2*n];
		if(!z_buffer)
			return false;
		weight_buffer=z_buffer+n;
	}
	this->xres=xres;
	this->yres=yres;
	x0=xres/2;
	y0=yres/2;
	return true;
}

/*#define BC			0xff
#define BACKGROUND		BC|(BC<<8)|(BC<<16)|(BC<<24)*/


void RenderScreen::clearScreen()
{
/*	unsigned int *data;		// 32 Bit Pixel
	double *z_buffer;		// depth information*/

	int i,n=xres*yres;
	if(z_buffer)
		memset(z_buffer, 0, 2*n*sizeof(double));
	if(!background_color)
		memset(data,0,n*sizeof(unsigned int));
	else
	{
		unsigned int *cptr=data;
		for(i=0;i<n;i++)
			*cptr++=background_color;
	}
}

#define PTR_INCREMENT		(sizeof(struct ColoredPoint_s)-3*sizeof(double))

bool RenderScreen::renderPointCollection(class PointCollection *pPC)
{
	int n,xres,yres,x0,y0;
	double d, fac;

	n=pPC->getNOPoints();
	if(!n)
		return false;
	d=Perspective.distance;
	fac=Perspective.factor;
	xres=this->xres;
	yres=this->yres;
	x0=this->x0;
	y0=this->y0;
/*#ifdef WIN32
	void *bitmapdata=data;
	void *zdata=z_buffer;
	void *pointdata=pPC->getColoredPointBase();
	int index=0,x,y;
	unsigned int bitmapindex;
	unsigned char FPU_state[108];
	//32bit-Version
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		push esi
		push edi
		//fill out with zeros
		mov edi,zdata
		mov esi,bitmapdata

		//save FPU state and clear exception flags
		fsave FPU_state		
		//reset FPU
		finit		
//######### point loop #############################
		//load permanently d and fac on stack
		fld fac
		fld d
		mov esi,pointdata		
		mov ecx,n

RenderPoints_point_loop:		
		push ecx		//save ecx
		//FPU:
		//st(0):d
		//st(1):fac
		//calculate screenpoint with the following formula:
		//sx=fac*y/(x+d)
		//sy=fac*z/(x+d)
		fld qword ptr [esi]
		fadd st(0),st(1)	//x+d
		//calculate sx
		add esi,8
		fld qword ptr [esi]			//y->st(0)
		fmul st(0),st(3)	//y*fac->st(0)
		fdiv st(0),st(1)	//fac*y/(x+d)->st(0)
		frndint
		fist x
		//calculate sy
		add esi,8
		fld qword ptr [esi]			//z->st(0)
		fmul st(0),st(4)	//z*fac->st(0)
		fdiv st(0),st(2)	//fac*z/(x+d)->st(0)
		frndint
		fist y
		//add esi,8			//esi points to next Vector3D now
		add esi,8			//esi points to color of the point now...
		//FPU aufrumen
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp

//		FPU state:
//		st(0):	zvalue
//		st(1):	d
//		st(2):	fac
		
		//adjust it to screen middle
		mov eax,x
		add eax,x0
		cmp eax,0
		jge RenderPoints_test_clip_x_xres
		jmp RenderPoints_next_point

RenderPoints_test_clip_x_xres:
		cmp eax,xres
		jl RenderPoints_test_clip_y
		jmp RenderPoints_next_point

RenderPoints_test_clip_y:
		mov x,eax
		mov eax,y0
		add eax,y
		cmp eax,0
		jge RenderPoints_test_clip_y_yres
		jmp RenderPoints_next_point

RenderPoints_test_clip_y_yres:
		cmp eax,yres
		jl RenderPoints_index_calculation
		jmp RenderPoints_next_point

RenderPoints_index_calculation:
		mov y,eax
		//point lies in screen bitmap, index calculation
		mov eax,y
		mul xres
		add eax,x		//index in eax
		mov bitmapindex,eax
		//check if there is an existing pixel ....				
		mov ebx,eax
		mov edi,bitmapdata
		shl ebx,2		//32bit bitmap		
		cmp dword ptr [edi+ebx],0
		jne RenderPoints_found_point
		jmp RenderPoints_new_point

RenderPoints_found_point:
		//found point check if it needs to get overdrawn
		mov edi,zdata
		shl ebx,1		//z-values are double values
		fcom qword ptr [edi+ebx]
		fstsw ax
		sahf
		jb RenderPoints_overdraw_point
		jmp RenderPoints_next_point

RenderPoints_overdraw_point:
		fst qword ptr [edi+ebx]
		shr ebx,1
		//mov edi,pcolordata
		//mov edx,dword ptr [edi]
		mov edx,dword ptr [esi]
		mov edi,bitmapdata
		mov [edi+ebx],edx
		jmp RenderPoints_next_point		

RenderPoints_new_point:
		//mov edi,pcolordata
		//mov edx,dword ptr [edi]
		mov edx,dword ptr [esi]
		mov edi,bitmapdata		
		mov [edi+ebx],edx
		shl ebx,1
		mov edi,zdata
		fst qword ptr [edi+ebx]

RenderPoints_next_point:		
		ffree st(0)
		fincstp
//RenderPoints_skip_this_point:
		pop ecx		//restore ecx

		//add dword ptr pcolordata,4
		add esi, 8	// next vector..., pragma pack /Zp ,8 byte Elementausrichtung ,skip 4 bytes behind sizeof(unsigned int)
		inc index
		dec ecx
		cmp ecx,0
		je RenderPoints_point_loop_exit
		jmp RenderPoints_point_loop

RenderPoints_point_loop_exit:
		//release d and fac in st(1) and st(0)
		ffree st(0)
		fincstp
		ffree st(0)
		fincstp
//########## end point loop ######################################
		//restore FPU state
		frstor FPU_state
		pop edi
		pop esi
		pop edx
		pop ecx
		pop ebx
		pop eax
	}
#else*/
/*
	void *bitmapdata, *zdata, *pointdata;
	int n,xres,yres,x0,y0,index,x,y;
	unsigned int  bitmapindex;
	double d,fac;
 * */
	// int sx,sy,i,ix,iy=0;
	int sx,sy,i,ix;
	double *doublePtr;
	struct ColoredPoint_s *cpp=pPC->getColoredPointBase();
	double _x, _y, _z, _xd, _x0=double(x0), _y0=double(y0);
	unsigned int pixel_color, index, *pixel_data=data, bc=background_color;
	double *z_data=z_buffer;
	double x_gridvector=0.0, y_gridvector=0.0;
	//printf("%d\n",n);
	if(stereo_mode)
	{
		for(i=0;i<n;i++)
		{
			doublePtr=(double *)cpp;
			_x=*doublePtr++;
			_y=*doublePtr++;
			_z=*doublePtr++;
			pixel_color=*((unsigned long *)doublePtr);
			_xd=_x+d;
			// 2D-Grid
			/*for(iy=-1;iy<=1;iy++)
				for(ix=-1;ix<=1;ix++)
				{
					x_gridvector=ix*half_eye_distance;
					y_gridvector=iy*half_eye_distance;

					sx=(int)(fac*(_y+x_gridvector)/_xd+_x0);
					sy=(int)(fac*(_z+y_gridvector)/_xd+_y0);
					//clipping:
					if((sx>=0 && sx<xres) && (sy>=0 && sy<yres))
					{
						index=(unsigned int)(sy*xres+sx);	//offset
						if(pixel_data[index]!=bc)
						{
							if(_xd<z_data[index])
							{
								pixel_data[index]=pixel_color;
								z_data[index]=_xd;
							}
						}
						else
						{
							pixel_data[index]=pixel_color;
							z_data[index]=_xd;				
						}
					}
				}*/
			for(ix=0;ix<2;ix++)
				{
					x_gridvector=(double(ix)-0.5)*half_eye_distance;
					sx=(int)(fac*(_y+x_gridvector)/_xd+_x0);
					sy=(int)(fac*(_z+y_gridvector)/_xd+_y0);
					//clipping:
					if((sx>=0 && sx<xres) && (sy>=0 && sy<yres))
					{
						index=(unsigned int)(sy*xres+sx);	//offset
						if(pixel_data[index]!=bc)
						{
							if(_xd<z_data[index])
							{
								pixel_data[index]=pixel_color;
								z_data[index]=_xd;
							}
						}
						else
						{
							pixel_data[index]=pixel_color;
							z_data[index]=_xd;				
						}
					}
				}
			cpp++;
		}
	}
	else
	{
		for(i=0;i<n;i++)
		{
			doublePtr=(double *)cpp;
			_x=*doublePtr++;
			_y=*doublePtr++;
			_z=*doublePtr++;
			pixel_color=*((unsigned long *)doublePtr);
			_xd=_x+d;
			sx=(int)((fac*_y)/_xd+_x0);
			sy=(int)((fac*_z)/_xd+_y0);
			//clipping:
			if((sx>=0 && sx<xres) && (sy>=0 && sy<yres))
			{
				index=(unsigned int)(sy*xres+sx);	//offset
				if(pixel_data[index]!=bc)
				{
					if(_xd<z_data[index])
					{
						pixel_data[index]=pixel_color;
						z_data[index]=_xd;
					}
				}
				else
				{
					pixel_data[index]=pixel_color;
					z_data[index]=_xd;				
				}
			}
			cpp++;
		}
	}
//#endif // WIN32
	return true;
}

/*inline void RenderScreen::renderGridPosition(double& _x, double& _y, double& _xd, double&_z, double& x_gridvector, double& y_gridvector, unsigned int pixel_color)
{
	unsigned int *_pixel_data=(unsigned int *)pixel_data;
	int sx=(int)(fac*(_y+x_gridvector)/_xd+double(x0));
	int sy=(int)(fac*(_z+y_gridvector)/_xd+double(y0));
	//clipping:
	if((sx>=0 && sx<xres) && (sy>=0 && sy<yres))
	{
		unsigned int index=(unsigned int)(sy*xres+sx);	//offset
		if(_pixel_data[index])
		{
			if(_xd<z_data[index])
			{
				_pixel_data[index]=pixel_color;
				z_data[index]=_xd;
			}
		}
		else
		{
			_pixel_data[index]=pixel_color;
			z_data[index]=_xd;				
		}
	}
}*/

void RenderScreen::toScreenPoint(class Vector3D &p, int &x, int &y)
{
	double rx,ry,d=p[0]+Perspective.distance;
	rx=Perspective.factor*p[1]/d;
	ry=Perspective.factor*p[2]/d;
	if(floor(rx)+0.5>rx)
		x=x0+(int)ceil(rx);
	else
		x=x0+(int)floor(rx);
	if(floor(ry)+0.5>ry)
		y=y0+(int)ceil(ry);
	else
		y=y0+(int)floor(ry);
}

void RenderScreen::absoluteToScreenPoint(class Vector3D p, int &x, int &y, double& depth)
{
	p*=Matrix3D(Perspective.vector);
	p+=Perspective.vector[3];
	double rx,ry,d=p[0]+Perspective.distance;
	rx=Perspective.factor*p[1]/d;
	ry=Perspective.factor*p[2]/d;
	if(floor(rx)+0.5>rx)
		x=x0+(int)ceil(rx);
	else
		x=x0+(int)floor(rx);
	if(floor(ry)+0.5>ry)
		y=y0+(int)ceil(ry);
	else
		y=y0+(int)floor(ry);
	depth=d;
}

void RenderScreen::toScreenPoint(class Vector3D& p, double& x, double& y)
{
	double d=p[0]+Perspective.distance;
	x=double(x0)+Perspective.factor*p[1]/d;
	y=double(y0)-Perspective.factor*p[2]/d;
}

bool RenderScreen::exportToEPSFile(class PointCollection *pOriginalPC, char *filename, bool export_image)
{
	class EPSFile EPSOutputFile;
	// define EPS file header:
	struct EPSFile::Header_s *pEPSFileHeader=EPSOutputFile.getHeaderPtr();
	double x1=double(pEPSFileHeader->bb_x1=x0-xres/2);
	double y1=double(pEPSFileHeader->bb_y1=y0-yres/2);
	double x2=double(pEPSFileHeader->bb_x2=x0+xres/2);
	double y2=double(pEPSFileHeader->bb_y2=y0+yres/2);
	double x,y;
	strcpy(pEPSFileHeader->creator,"Tim Deutschmann IUP Heidelberg");
	time_t long_time;
	time(&long_time);
	strcpy(pEPSFileHeader->date, asctime(localtime(&long_time)));
	if(!EPSOutputFile.open(filename))
	{
		fprintf(stderr,"could not create EPS file \"%s\"!\n", filename);
		return false;
	}
	if(export_image)
		EPSOutputFile.colorImage(x1,y1,xres,yres,data);
	else
	{
		class PointCollection PC;
		if(!PC.addPointCollection(pOriginalPC))
		{
			fprintf(stderr,"could not make PointCollection copy!\n");
			return false;
		}
		// create screen points...
		PC.sortByX0();
		struct ColoredPoint_s *cp=PC.getColoredPointBase();
		//EPSOutputFile.setRGBColor(0.0,0.0,0.0);
		EPSOutputFile.filledRectangle(x1,y1,x2,y2,0.0,0.0,0.0);
		int i, n=pOriginalPC->getNOPoints();
		for(i=0;i<n;i++)
		{
			toScreenPoint(cp->point, x, y);
			if(x>=x1 && x<=x2 && y>=y1 && y<=y2)
			{
				double b=double(cp->color&0xff)/255.0;
				double g=double((cp->color&0xff00)>>8)/255.0;
				double r=double((cp->color&0xff0000)>>16)/255.0;
				EPSOutputFile.putPixel(x,y,r,g,b);
			}
			cp++;
		}
	}
	EPSOutputFile.close();
	return true;
}

bool RenderScreen::savePerspective(char *file_name)
{
	int i,j;
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		fprintf(stderr,"could not write perspective to file \"%s\"!\n", file_name);
		return false;
	}
	fprintf(fptr,"%lg\n", Perspective.distance);
	fprintf(fptr,"%lg\n", Perspective.factor);
	for(i=0;i<3;i++)
	{
		for(j=0;j<3;j++)
			fprintf(fptr,"%lg\t", Perspective.vector[j][i]);
		fprintf(fptr,"%lg\n", Perspective.vector[j][i]);
	}
	fclose(fptr);
	return true;
}

bool RenderScreen::loadPerspective(char *file_name)
{
	int i,j;
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		fprintf(stderr,"could not write perspective to file \"%s\"!\n", file_name);
		return false;
	}
	fscanf(fptr,"%lg\n", &Perspective.distance);
	fscanf(fptr,"%lg\n", &Perspective.factor);
	for(i=0;i<3;i++)
	{
		for(j=0;j<3;j++)
			fscanf(fptr,"%lg\t", Perspective.vector[j].getCoefficientPtr()+i);
		fscanf(fptr,"%lg\n", Perspective.vector[j].getCoefficientPtr()+i);
	}
	fclose(fptr);
	return true;
}




