/**************************************************************************************************************
*
*       This file,
*
*                            /Depiction/SurfaceGridPointCollection.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "SurfaceGridPointCollection.h"
#include "../error.h"

SurfaceGridPointCollection::SurfaceGridPointCollection()
{
	pSurfaceGrid=(class SurfaceGrid *)0;
}

bool SurfaceGridPointCollection::initialize()
{
	if(!pSurfaceGrid)
		return false;
	else if(!pSurfaceGrid->useReliefData())
	{
		//PointCollection::freeCollection();
		return true;		// no relief data present.... done: nothing to see!
	}

	// declarations
	int x,y,xres,yres,index,base_index,xresm1,yresm1,xresp1,n=0,i;
	class Vector3D *gridPoint,*middlePoint,m; 
	pSurfaceGrid->getReliefData(&xres, &yres, &gridPoint, &middlePoint);
	xresm1=xres-1;
	yresm1=yres-1;
	xresp1=xres+1;
	// determine number of points to be allocated: leave the edge away!
	// x stripes
	for(y=1;y<yres;y++)
	{
		base_index=y*xresp1;
		for(x=1;x<xresm1;x++)
		{
			index=base_index+x;
			n+=int((gridPoint[index]-gridPoint[index+1]).length()/granularity);
		}
	}

	// y stripes
	for(x=1;x<xres;x++)
		for(y=1;y<yresm1;y++)
		{
			index=y*xresp1+x;
			n+=int((gridPoint[index]-gridPoint[index+xresp1]).length()/granularity);
		}

	// the crosses....
	for(y=1;y<yresm1;y++)		
		for(x=1;x<xresm1;x++)
		{
			m=middlePoint[y*xres+x];
			n+=int((gridPoint[y*xresp1+x]-m).length()/granularity);
			n+=int((gridPoint[y*xresp1+x+1]-m).length()/granularity);
			n+=int((gridPoint[(y+1)*xresp1+x]-m).length()/granularity);
			n+=int((gridPoint[(y+1)*xresp1+x+1]-m).length()/granularity);			
		}

	// allocate points
	if(!PointCollection::allocatePoints(n))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"allocation error!\n");
		return false;
	}

	// calculate the points:
	class Vector3D A,B,D;
	struct ColoredPoint_s *cpptr=cp;
	// x stripes
	for(y=1;y<yres;y++)
	{
		base_index=y*xresp1;
		for(x=1;x<xresm1;x++)
		{
			index=base_index+x;
			A=gridPoint[index];
			B=gridPoint[index+1];
			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);
			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}
		}
	}

	// y stripes
	for(x=1;x<xres;x++)
		for(y=1;y<yresm1;y++)
		{
			index=y*xresp1+x;
			A=gridPoint[index];
			B=gridPoint[index+xresp1];
			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);
			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}
		}

	// the crosses....
	for(y=1;y<yresm1;y++)
		for(x=1;x<xresm1;x++)
		{
			A=middlePoint[y*xres+x];
			B=gridPoint[y*xresp1+x];

			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);

			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}

			B=gridPoint[y*xresp1+x+1];

			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);

			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}

			B=gridPoint[(y+1)*xresp1+x];

			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);

			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}

			B=gridPoint[(y+1)*xresp1+x+1];

			D=B-A;
			n=int((B-A).length()/granularity);
			D=norm(D);

			for(i=0;i<n;i++)
			{
				/**pointPtr++=A+(double(i)*granularity)*D;
				*pcolor++=color;*/
				cpptr->point=A+(double(i)*granularity)*D;
				cpptr->color=color;
				cpptr++;
			}
		}

	return true;
}

/*bool SurfaceGridPointCollection::loadFromFile(char *filename, int *p_maximal_index, bool *enable)
{
	File file;
	if(!file.open(filename,"rb"))
	{
		fprintf(stderr,"could not open file \"%s\"!\n",filename);
		return false;
	}
	FILE *fptr=file.getFILEPtr();
	int n,i,j,index,no_events,total_no_events=0,max_no_events=0,event_type;
	fread(&n,1,sizeof(n),fptr);
	if(n<=0)
	{
		fprintf(stderr,"something is wrong with trace event file \"%s\"!\n",filename);
		return false;
	}
	// count total number of events
	if(enable) // apply filter
	{
		for(i=0;i<n;i++)
		{
			fread(&no_events,1,sizeof(n),fptr);
			if(no_events<=0)
			{
				fprintf(stderr,"something is wrong with trace event file \"%s\":\n",filename);
				fprintf(stderr,"trajectory %d is invalid!\n",i+1);
				return false;
			}
			int count=0;
			struct RadiativeTransferEvent_s Event;
			for(j=0;j<no_events;j++)
			{
				fread(&Event,sizeof(struct RadiativeTransferEvent_s),1,fptr);
				if(Event.type>=0 && enable[Event.type])
				{
					total_no_events++;
					count++;
				}
			}
			if(count>max_no_events)
				max_no_events=count;
		}
		// allocate
		if(!allocatePoints(total_no_events))
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"load trace events: allocation error: n=%d\n", no_events);
			return false;
		}
		// read-in events
		index=0;
		fseek(fptr,sizeof(n),SEEK_SET);
		for(i=0;i<n;i++)
		{
			fread(&no_events,1,sizeof(n),fptr);
			if(no_events<=0)
			{
				fprintf(stderr,"something is wrong with trace event file \"%s\":\n",filename);
				fprintf(stderr,"trajectory %d is invalid!\n",i+1);
				return false;
			}
			struct RadiativeTransferEvent_s Event;
			for(j=0;j<no_events;j++)
			{
				fread(&Event,sizeof(struct RadiativeTransferEvent_s),1,fptr);
				if(Event.type>=0 && enable[Event.type])
				{					
					this->m_point[index]=Event.position;
					event_type=Event.type;
					this->m_event_type[index]=event_type;
					this->m_color[index]=mapTypeToColor(event_type);
					index++;
				}
			}
		}
	}
	else // do not apply filter
	{
		for(i=0;i<n;i++)
		{
			fread(&no_events,1,sizeof(n),fptr);
			if(no_events<=0)
			{
				fprintf(stderr,"something is wrong with trace event file \"%s\":\n",filename);
				fprintf(stderr,"trajectory %d is invalid!\n",i+1);
				return false;
			}
			total_no_events+=no_events;
			if(no_events>max_no_events)
				max_no_events=no_events;
			fseek(fptr,no_events*sizeof(RadiativeTransferEvent_s),SEEK_CUR);
		}
		// allocate
		if(!allocatePoints(total_no_events))
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"load trace events: allocation error: n=%d\n", no_events);
			return false;
		}
		struct RadiativeTransferEvent_s *event=new struct RadiativeTransferEvent_s[max_no_events];
		if(!event)
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"load trace events: allocation error: n=%d\n", max_no_events);
			return false;
		}
		// read-in events
		index=0;
		fseek(fptr,sizeof(n),SEEK_SET);
		for(i=0;i<n;i++)
		{
			fread(&no_events,1,sizeof(n),fptr);
			if(no_events<=0\
			|| fread(event,sizeof(struct RadiativeTransferEvent_s),no_events,fptr)!=(unsigned int)no_events)
			{
				fprintf(stderr,"something is wrong with trace event file \"%s\":\n",filename);
				fprintf(stderr,"trajectory %d is invalid!\n",i+1);
				delete [] event;
				return false;
			}
			// extract_information
			for(j=0;j<no_events;j++)
			{
				this->m_point[index]=event[j].position;
				event_type=event[j].type;
				this->m_event_type[index]=event_type;
				this->m_color[index]=mapTypeToColor(event_type);
				index++;
			}
		}
		delete [] event;
	}
	if(p_maximal_index)
	{
		int maximal_index=0,type;
		for(i=0;i<total_no_events;i++)
		{
			type=m_event_type[i];
			if(type>maximal_index)
				maximal_index=type;
		}
		*p_maximal_index=maximal_index;
	}
	strcpy(this->filename,filename);
	return true;
}

unsigned int SurfaceGridPointCollection::mapTypeToColor(int event_type)
{
	int scatterer_index;
	switch(event_type)
	{
		case TLC_REFRACTIVE_INDEX:				return 0x00aaaaaa;	// grey
		case TLC_TOTAL_ABSORPTION_COEFFICIENT:	return 0x000000ff;	// blue
		case TLC_GROUND:						return 0x00ffff00;	// yellow
		case TLC_EPSILON_RAYLEIGH:				return 0x00ff0000;	// red
		case TLC_EPSILON_ROTATIONAL_RAMAN:		return 0x0000ffff;	// pink
		default:
			scatterer_index=event_type-TLC_EPSILON_AEROSOL;
			if(scatterer_index>=0)
				return 0x0000aa00+0x0011*scatterer_index;			// green tones
	}		
	return 0x00ffffff;
}
*/
