/**************************************************************************************************************
*
*       This file,
*
*                            /Depiction/TraceEventCollection.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "TraceEventCollection.h"
#include "../File.h"
#include "../Raytracer.h"
#include "../error.h"
#include <float.h>

// #include "../OpticalMedium.h"

TraceEventCollection::TraceEventCollection()
{
	event_type=(int *)0;
	weight=(double *)0;
	contribution=0;
	contribution_center=0.5;
	contribution_sigma=0.3;
}

TraceEventCollection::~TraceEventCollection()
{
	if(event_type)
		delete [] event_type;
	if(weight)
		delete [] weight;
	if(contribution)
		delete [] contribution;
}

bool TraceEventCollection::allocatePoints(int n)
{
	if(!PointCollection::allocatePoints(n))
		return false;
	if(event_type)
		delete [] event_type;
	event_type=new int[n];
	if(weight)
		delete [] weight;
	weight=new double[n];
	if(contribution)
		delete [] contribution;
	contribution=new double[n];
	return (event_type) && (weight) && (contribution);
}

bool TraceEventCollection::loadFromFile(char *file_name, bool *enable)
{
	File file;
	if(!file.open(file_name, "rb"))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not open file \"%s\"!\n", file_name);
		return false;
	}
	FILE *fptr=file.getFILEPtr();
	int n, i, index, total_no_events=0, et;
	
	// determine number of trace events in file
	fseek(fptr, 0, SEEK_END);
	n=ftell(fptr)/sizeof(struct RadiativeTransferEvent_s);
	fseek(fptr, 0, SEEK_SET);
	if(n<=0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"something is wrong with trace event file \"%s\"!\n", file_name);
		return false;
	}
	// printf("%d RT events in file \"%s\"!\n", n, file_name);

	// count total number of events
	/*double c;
	maximum_contribution=DBL_MIN;
	minimum_contribution=DBL_MAX;*/
	struct RadiativeTransferEvent_s Event;
	for(i=0;i<n;i++)// for each trajectory
	{
		fread(&Event, sizeof(struct RadiativeTransferEvent_s), 1, fptr);
		if(Event.type>=0 && enable[Event.type])
		{
			total_no_events++;
			/*c=Event.contribution;
			if(c>0.0)
			{
				if(c>maximum_contribution)
					maximum_contribution=c;
				if(c<minimum_contribution)
					minimum_contribution=c;
			}*/
		}
	}
	/*contribution_range=maximum_contribution-minimum_contribution;
	printf("event contributions in [%le,%le]\n", minimum_contribution, maximum_contribution);*/
	// allocate
	if(!allocatePoints(total_no_events))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"load trace events: allocation error: n=%d\n", total_no_events);
		return false;
	}
	// read-in events
	fseek(fptr, 0, SEEK_SET);
	index=0;
	for(i=0;i<n;i++)
	{
		fread(&Event, sizeof(struct RadiativeTransferEvent_s), 1, fptr);
		if(Event.type>=0 && enable[Event.type])
		{
			cp[index].point=Event.position;
			et=Event.type;
			event_type[index]=et;
			contribution[index]=Event.contribution;
			cp[index].color=mapTypeToColor(et, Event.contribution);
			index++;
		}
	}
	strcpy(this->file_name, file_name);
	return true;
}

void TraceEventCollection::recalculateColors()
{
	for(int i=0;i<no_points;i++) cp[i].color=mapTypeToColor(event_type[i], contribution[i]);	
}

unsigned int TraceEventCollection::mapTypeToColor(int event_type, const double& contribution)
{
	double d=(log(contribution)-contribution_center)/contribution_sigma, f=exp(-0.5*d*d);
	unsigned int color=56+(unsigned int)(f*200.0);
	switch(event_type)
	{
		case TE_SOURCE:				return 0x00ff00ff;	// purple  
		case TE_REFRACTION_X:			return 0x00ff6666;	// red grey
		case TE_REFRACTION_Y:			return 0x0066ff66;	// green grey
		case TE_REFRACTION_Z:			return 0x006666ff;	// blue grey
		case TE_ABSORPTION:			return 0x000000ff;	// blue
		case TE_GROUND_SCATTERING:
			return (color<<16) | (color<<8);	// 0x00ffff00;	// yellow
		case TE_GROUND_REFLECTION:
			return (color<<8) | color;	// 0x00ffff00;	// turquoise
		case TE_RAYLEIGH_SCATTERING:
			return color<<16;		// return 0x00ff0000;	// red
		case TE_ROTATIONAL_RAMAN_SCATTERING:
			return (color<<8) | color; //return 0x0000ffff;	// pink
		default:
			/*scatterer_index=event_type-TE_PARTICLE_SCATTERING;
			if(scatterer_index>=0)
				return 0x0000aa00+0x0011*scatterer_index;			// green tones
			*/
			return color<<8;
	}		
	return 0x00ffffff;
}

void TraceEventCollection::renderSensitivityTo(class RenderScreen *pRenderScreen)
{
	// structure of the point list
}







