/**************************************************************************************************************
*
*       This file,
*
*                            /Estimate.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Estimate.h"
#include "Result.h"
#include "error.h"

#include "Grid.h"

Estimate::Estimate()
{
	recursion_depth=0;
	GasProfileJacobian=0;
	ParticleProfileJacobian=0;
	allow_psi_sampling=false;
}

Estimate::~Estimate()
{
	if(GasProfileJacobian)
		delete [] GasProfileJacobian;
	if(ParticleProfileJacobian)
		delete [] ParticleProfileJacobian;
}

bool Estimate::associateWithResult(class Result *ResultPtr)
{
	pResult=ResultPtr;
	int i, no_clusters, no_phase_function_parameters, cluster_size, no_parameters;
	calculate_Jacobians=false;
	N_Parameters=0;

	// cell path lengths...
	calculate_cell_path_length=pResult->calculateCellPathLength();
	if(calculate_cell_path_length && !TrajectoryInformationObject::allocate())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}

	// cell sensitivity...
	calculate_cell_sensitivity=pResult->calculateCellSensitivity();
	if(calculate_cell_sensitivity)
	{
		if(!CellSensitivity.allocate(no_cells))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "allocation error!\n");
			return false;
		}
		CellSensitivity.setToZero();
	}

	// Jacobians and Hessians ###################################
	// gases
	if(GasProfileJacobian)
		delete [] GasProfileJacobian;
	GasProfileJacobian=0;
	no_active_gases=pResult->no_active_gases;
	if(no_active_gases)
	{
		GasProfileJacobian=new struct GasProfileJacobian_s[no_active_gases];
		if(!GasProfileJacobian)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		// initialize structs vectors therein
		class Gas *pGas;
		struct GasProfileJacobian_s *pGPJ;
		for(i=0;i<no_active_gases;i++)
		{
			pGPJ=GasProfileJacobian+i;
			pGas=pResult->GasResult[i].pGas;
			pGPJ->pGas=pGas;
			pGPJ->no_clusters=no_clusters=pGas->getNOClusters();
			if(!pGPJ->CumulativeContribution.allocate(no_clusters) ||
				!pGPJ->CurrentScatterOrderContribution.allocate(no_clusters) ||
				!pGPJ->Jacobian.allocate(no_clusters))// ||
				//!pGPJ->updateCluster.allocate(no_clusters))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			
		}
		calculate_Jacobians=true;
	}
	// particles
	if(ParticleProfileJacobian)
		delete [] ParticleProfileJacobian;
	ParticleProfileJacobian=0;
	no_active_particles=pResult->no_active_particles;
	if(no_active_particles)
	{
		ParticleProfileJacobian=new struct ParticleProfileJacobian_s[no_active_particles];
		if(!ParticleProfileJacobian)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		// initialize structs and matrices therein
		class Particle *pParticle;
		struct ParticleProfileJacobian_s *pPPJ;
		for(i=0;i<no_active_particles;i++)
		{
			pParticle=pResult->ParticleResult[i].pParticle;
			pPPJ=ParticleProfileJacobian+i;
			pPPJ->pParticle=pParticle;
			pPPJ->no_clusters=no_clusters=pParticle->getNOClusters();
			pPPJ->cluster_size=cluster_size=pResult->ParticleResult[i].cluster_size;
			pPPJ->no_parameters=no_parameters=pResult->ParticleResult[i].no_parameters;
			pPPJ->no_phase_function_parameters=no_phase_function_parameters=pParticle->getNOPhaseFunctionParameters();
			if(pParticle->hasSinglePhaseFunction())
			{
				if(!pPPJ->SPF_CumulativeContribution.allocate(no_parameters) ||
					!pPPJ->SPF_CurrentScatterOrderContribution.allocate(no_parameters) ||
					!pPPJ->SPF_Jacobian.allocate(no_parameters))
				{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
			}
			else
			{
				if(!pPPJ->CumulativeContribution.allocate(no_clusters, cluster_size) ||
					!pPPJ->CurrentScatterOrderContribution.allocate(no_clusters, cluster_size) ||
					!pPPJ->Jacobian.allocate(no_clusters, cluster_size))
				{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
			}
		}
		calculate_Jacobians=true;
	}
	if(calculate_Jacobians)
	{
		N_Parameters=pResult->N_Parameters;
		if(!LogarithmicDerivative.allocate(N_Parameters))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		calculate_Hessian=pResult->calculateHessian();
		if(calculate_Hessian)
		{
			if(!Hessian.allocate(N_Parameters, N_Parameters) || !CumulativeLogarithmic2ndDerivativeHessian.allocate(N_Parameters, N_Parameters))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
		}
	}
	return true;
}

void Estimate::addEstimate(class Estimate *pEstimate)
{
	scalar_radiance			+= pEstimate->scalar_radiance;
	elastic_scalar_radiance		+= pEstimate->elastic_scalar_radiance;

	direct_light_radiance		+= pEstimate->direct_light_radiance;

	vector_radiance			+= pEstimate->vector_radiance;
	elastic_vector_radiance		+= pEstimate->elastic_vector_radiance;

	scalar_RRS_count 		+= pEstimate->scalar_RRS_count;
	scalar_elastic_scatter_count	+= pEstimate->scalar_elastic_scatter_count;
	scalar_ground_scatter_count	+= pEstimate->scalar_ground_scatter_count;

	tau += pEstimate->tau;
	tau_RRS += pEstimate->tau_RRS;
	tau_absorption += pEstimate->tau_absorption;

	if(calculate_cell_sensitivity)
		addCellSensitivities(CellSensitivity.getCoefficientPtr(), pEstimate->CellSensitivity.getCoefficientPtr());
}

void Estimate::addCellSensitivities(double *cs_destination, double *cs_source)
{
	int i, index, n=getNOUpdateCells();
	int *pUpdateCellindex = getUpdateCellIndexPtr();
	for(i=0;i<n;i++)
	{
		index = *pUpdateCellindex++;
		cs_destination[index] += cs_source[index];
		cs_source[index] = 0.0;
	}
}

void Estimate::resetTransitionStokesMatrix()
{
	TransitionStokesMatrix.setToUnityMatrix();
	RRSTransitionStokesMatrix.setToUnityMatrix();
}

void Estimate::GasProfileJacobian_s::reset()
{
	// updateCluster.setToZero();
	CumulativeContribution.setToZero();
	CurrentScatterOrderContribution.setToZero();
	Jacobian.setToZero();
}

void Estimate::ParticleProfileJacobian_s::reset()
{
	// updateCluster.setToZero();
	if(pParticle->hasSinglePhaseFunction())
	{
		SPF_CumulativeContribution.setToZero();
		SPF_CurrentScatterOrderContribution.setToZero();
		SPF_Jacobian.setToZero();
	}
	else
	{
		CumulativeContribution.setToZero();
		CurrentScatterOrderContribution.setToZero();
		Jacobian.setToZero();
	}
}

void Estimate::reset()
{
	weight=1.0;
	tau=0.0;
	RRS_elastic_weight=1.0;
	tau_RRS=0.0;
	RRS_adjoint_weight=1.0;
	tau_absorption=0.0;

	twilight_photon=false;
	
	// has_RRSEvent=false;
	no_RRS_events=0;
	no_elastic_scatter_events=0;
	no_ground_scatter_events=0;

	// event counts
	scalar_RRS_count=0.0;
	scalar_multiple_RRS_count=0.0;
	scalar_elastic_scatter_count=0.0;
	scalar_ground_scatter_count=0.0;

	// radiances
	scalar_radiance=0.0;
	elastic_scalar_radiance=0.0;
	direct_light_radiance=0.0;
	
	// polarized radiances
	vector_radiance=0.0;
	elastic_vector_radiance=0.0;
	
	resetTransitionStokesMatrix();

	TrajectoryInformationObject::resetCumulativeUpdateStructs();
	TrajectoryInformationObject::resetUpdateCellIndices();
	/*if(calculate_cell_sensitivity)
	{
		CellPathLength.setToZero();
		CellSensitivity.setToZero();
	}*/
	if(calculate_cell_sensitivity)
	{
		double *cs = CellSensitivity.getCoefficientPtr();
		int *pUpdateCellindex = getUpdateCellIndexPtr();
		int i, n=getNOUpdateCells();
		for(i=0;i<n;i++)
			cs[*pUpdateCellindex++] = 0.0;
	}

	if(calculate_Jacobians)
	{
		int i;
		for(i=0;i<no_active_gases;i++)
			GasProfileJacobian[i].reset();
		for(i=0;i<no_active_particles;i++)
			ParticleProfileJacobian[i].reset();
		if(calculate_Hessian)
		{
			CumulativeLogarithmic2ndDerivativeHessian.setToZero();
			Hessian.setToZero();
		}
	}
}

void Estimate::accumulateSimpleHessianContributions(const double& scalar_contribution, const class StokesVector& vector_contribution)
{
	int i, j;
	double *ScalarHessian_i;
	double c, logJ_i, *CumulativeLogarithmic2ndDerivativeHessian_i;
	for(i=0;i<N_Parameters;i++)
	{
		logJ_i=LogarithmicDerivative[i];
		ScalarHessian_i=Hessian[i];
		CumulativeLogarithmic2ndDerivativeHessian_i=CumulativeLogarithmic2ndDerivativeHessian[i];
		for(j=0;j<N_Parameters;j++)
		{
			c = logJ_i * LogarithmicDerivative[j];
			c += CumulativeLogarithmic2ndDerivativeHessian_i[j];
			ScalarHessian_i[j] += c * scalar_contribution;
		}
	}
}






