/**************************************************************************************************************
*
*       This file,
*
*                            /Estimate.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Estimate_h
#define _Estimate_h

#include "Particle.h"
#include "Gas.h"

#include "Math/Vector.h"
#include "Math/Matrix.h"

#include "Math/StokesMatrix.h"
#include "Math/StokesVector.h"

#include "TrajectoryInformationObject.h"

class Result;

class Estimate: public TrajectoryInformationObject
{
	private:

/*

	concept for estimate structure: Its content is filled order by order
	there are intermediate variables for the current scatter order and for
	the cumulative values. These cumulative values needs to be carried over
	in a splitting event.

	After a splitting event the result must be carried over to the
	result while beeing weighted properly.
	
*/

	// Jacobians for the absorption coefficient and optical particle properties (epsilon, ssa and phf parameters)

	class Result *pResult;
	
	bool calculate_Jacobians, calculate_Hessian;
	
	int N_Parameters;
	int recursion_depth;

	bool allow_psi_sampling;

	bool calculate_cell_path_length;

	bool calculate_cell_sensitivity;
	// int no_air_clusters;

	// bool has_RRSEvent;
	int no_RRS_events, no_elastic_scatter_events, no_ground_scatter_events;

	public:

		Estimate();
		~Estimate();

		inline void setRecursionDepth(const int& d){recursion_depth=d;}
		inline int getRecursionDepth(){return recursion_depth;}

		inline void enableAllowPsiSampling(bool enable=true){allow_psi_sampling=enable;}
		inline bool allowPsiSampling(){return allow_psi_sampling;}

		inline bool calculateJacobians(){return calculate_Jacobians;}

		inline void enableCalculateCellPathLength(bool enable=true){calculate_cell_path_length=enable;}
		inline bool calculateCellPathLength(){return calculate_cell_path_length;}

		/*inline void setScatterOrder(int o){scatter_order=o;}
		inline void increaseScatterOrder(){scatter_order++;}
		inline int getScatterOrder(){return scatter_order;}*/
		
		void reset();

		inline int getNORRSEvents(){return no_RRS_events;}
		inline int getNOElasticScatterEvents(){return no_elastic_scatter_events;}
		inline int getNOGroundScatterEvents(){return no_ground_scatter_events;}

		inline void registerRRSEvent(){no_RRS_events++;}
		inline void registerElasticScatterEvent(){no_elastic_scatter_events++;}
		inline void registerGroundScatterEvent(){no_ground_scatter_events++;}

		/*inline void registerRRSEvent(){has_RRSEvent=true;}
		inline bool hasRRSEvent(){return has_RRSEvent;}*/

		bool associateWithResult(class Result *ResultPtr);
		inline bool initializeFromEstimate(class Estimate *pEstimate){return associateWithResult(pEstimate->pResult);}
		void addEstimate(class Estimate *pEstimate);
		// void registerInResult();
		
		void resetTransitionStokesMatrix();

		bool twilight_photon;
		double weight, tau;

		class StokesMatrix TransitionStokesMatrix;			// account for polarization
		class StokesMatrix RRSTransitionStokesMatrix;

		// elastic and inelastic radiances
		double direct_light_radiance;
		double scalar_radiance, elastic_scalar_radiance;
		class StokesVector vector_radiance, elastic_vector_radiance;	

		// RRS
		double RRS_elastic_weight;	// weight to correct for elastic scattering (and absorption)
		double RRS_adjoint_extinction;	// temporary variable
		double RRS_adjoint_weight;	// temporary variable
		double tau_RRS;
		double tau_absorption;		// tau_absorption

		// scatter event counts
		double scalar_RRS_count, scalar_multiple_RRS_count, scalar_elastic_scatter_count, scalar_ground_scatter_count;
		
		// light path
		class Vector CellSensitivity;

		void addCellSensitivities(double *cs_destination, double *cs_source);

		inline void addCellSensitivitiesTo(double *cs_destination)
			{addCellSensitivities(cs_destination, CellSensitivity.getCoefficientPtr());}

		inline void registerLocalEstimateInSensitivityGrid(double& c)
			{TrajectoryInformationObject::registerLocalEstimateInSensitivityGrid(c, CellSensitivity.getCoefficientPtr());}

		// Jacobians

		int no_active_gases;

		struct GasProfileJacobian_s
		{
			void reset();

			int no_clusters;

			class TVector<int> UpdateClusterIndex;
			int no_clusters_to_update;
			
			class Vector CumulativeContribution;		// cumulative logarithmic derivative
			class Vector CurrentScatterOrderContribution;	// current scatter order logarithmic derivative
			class Vector Jacobian;				// Jacobian sub estimate
			/*
				- the vectors have as many rows as there are clusters in the respective profile
			*/
			class Gas *pGas;
			// class TVector<bool> updateCluster;

		} *GasProfileJacobian;
		
		int no_active_particles;

		struct ParticleProfileJacobian_s
		{
			void reset();

			int no_clusters, cluster_size, no_phase_function_parameters, no_parameters;

			class TVector<int> UpdateClusterIndex;
			int no_clusters_to_update;

			// single phase function
			class Matrix CumulativeContribution;		// cumulative logarithmic derivative
			class Matrix CurrentScatterOrderContribution;	// current scatter order logarithmic derivative
			class Matrix Jacobian;				// Jacobian sub estimate
			/*
				- the matrices have as many rows as there are clusters in the respective profile
				- there are furthermore as many columns as there are optical parameters:
				  i.e.: 2 (ext. + ssa) + n.o. phase function parameters
			*/

			// single phase function stuff...
			class Vector SPF_CumulativeContribution;		// cumulative logarithmic derivative
			class Vector SPF_CurrentScatterOrderContribution;	// current scatter order logarithmic derivative
			class Vector SPF_Jacobian;			// Jacobian sub estimate
			/*
				- vectors have the length:
					 n.o. clusters (extinction coefficients) + ssa + n.n. phase function parameters
			*/
			class Particle *pParticle;
			// class TVector<bool> updateCluster;
		}*ParticleProfileJacobian;

		// Hessian

		class Vector LogarithmicDerivative;
		class Matrix CumulativeLogarithmic2ndDerivativeHessian;
		class Matrix Hessian;

		void accumulateSimpleHessianContributions(const double& scalar_contribution, const class StokesVector& vector_contribution);

};

#endif // _Estimate_h


