/**************************************************************************************************************
*
*       This file,
*
*                            /FunctionalLayer.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "FunctionalLayer.h"
#include "Raytracer.h"
#include "Geometry.h"

#include "coordinate_sampling.h"
#include "trigonometry.h"
#include <float.h>
#include "error.h"

FunctionalLayer::FunctionalLayer()
{
	/*F0_data=0;
	F0_spectrum=0;*/

	ActiveParticlePtr=0;
	ActiveGasPtr=0;

	b_calculate_cell_path_length=false;
}

FunctionalLayer::~FunctionalLayer()
{

	/*if(F0_data)
		delete [] F0_data;
	if(F0_spectrum)
		delete [] F0_spectrum;*/

	if(ActiveParticlePtr)
		delete [] ActiveParticlePtr;
	if(ActiveGasPtr)
		delete [] ActiveGasPtr;

	/*if(estimated_scatter_order)
		delete [] estimated_scatter_order;*/
}

bool FunctionalLayer::resetFunctionals(class Result *pResult)
{
	no_cells=pGrid->getNOCells();
	
	// release/allocate memory
	if(!forced_cell_path_length.allocate(no_cells))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}

	// reset flags and active constituent ptr lists
	b_calculate_cell_path_length=b_create_sensitivity_grid_file;

	// disable all functionals
	pAtmosphere->disableFunctionalCalculation();
	
	// evaluate constituents
	createListOfStrings(constituents, ',', constituents_list, "none");	
	createListOfStrings(path_integrals, ',', path_integrals_list, "none");	
	
	// create total list
	list<string> c_list=list<string>(constituents_list);
	list<string> pi_list=list<string>(path_integrals_list);
	total_constituents_list.clear();
	total_constituents_list.merge(c_list);
	total_constituents_list.merge(pi_list);
	total_constituents_list.sort();
	total_constituents_list.unique();

	if(!total_constituents_list.empty())
	{
		int i;
		int gas_offset=0;
		int particle_offset=0;
		list<string>::iterator it;
		for(it=total_constituents_list.begin();it!=total_constituents_list.end();it++)
		{
			bool found_constituent=false;
			pGas=gas;
			for(i=0;i<no_gases;i++)
			{
				if(!strcmp(it->c_str(), pGas->getName()))
				{
					pGas->enableFunctionalCalculation();
					found_constituent=true;
					pGas->setJacobianOffset(gas_offset);
					gas_offset+=pGas->getNOClusters();
					break;
				}
				pGas++;
			}
			if(!found_constituent)
			{
				// list entry is not a gas, maybe it is a particle?
				pParticle=particle;
				for(i=0;i<no_particles;i++)
				{
					if(!strcmp(it->c_str(), pParticle->getName()))
					{
						pParticle->enableFunctionalCalculation();
						found_constituent=true;
						pParticle->setJacobianOffset(particle_offset);
						if(pParticle->hasSinglePhaseFunction())
							particle_offset+=1+pParticle->getNOPhaseFunctionParameters()+pParticle->getNOClusters();
						else
							particle_offset+=(2+pParticle->getNOPhaseFunctionParameters())*pParticle->getNOClusters();
						break;
					}
					pParticle++;
				}
				if(!found_constituent)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "constituent %s not in atmosphere!\n", it->c_str());
					return false;
				}
			}
		}
		// displace aerosols Jacobian index
		for(i=0;i<no_particles;i++)
			if(particle[i].isFunctionalCalculationEnabled())
				particle[i].setJacobianOffset(gas_offset+particle[i].getJacobianOffset());
		b_calculate_cell_path_length=true;
		if(!constituents_list.empty() && !path_integrals_list.empty())
			b_calculate_Hessian=true;
	}
	if(!pAtmosphere->createActiveConstituentLists())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error creating active constituent lists\n");
		return false;
	}
	// define things about the result
	if(b_use_precision)
		pResult->defineAccuracy(b_precision_major_limit, precision);
	pResult->enableOutputOfPolarizedFunctionals(b_output_polarized_quantities);
	pResult->enableCalculateCellPathLength(b_calculate_cell_path_length);
	pResult->enableCalculateCellSensitivity(b_create_sensitivity_grid_file);
	RRS=pAtmosphere->simulateRingEffect();
	pResult->enableRRS(RRS);
	pResult->enableEventCounts(b_event_counts);

	if(!pResult->allocate(pAtmosphere, b_calculate_Hessian))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating result\n");
		return false;
	}
	pResult->reset();
	pResult->enableNormalizeWithF0(b_normalize_with_F0);
	pResult->setF0(getF0(wavelength));
	pResult->disableRadianceHessianOutput(b_suppress_Hessian_output);
	pResult->disableRadianceJacobianOutput(b_suppress_Jacobian_output);

	// reset (stochastic) variables
	SVIOne=0.0;
	
	// sensitivity things
	if(b_create_sensitivity_grid_file)
	{
		if(!pGrid->is3DGrid())
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "sensitivity grid setup: grid is not a 3D grid\n");
			return false;
		}
		if(!SensitivityBGF.initializeFromGrid(pGrid, 1))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating sensitivity grid file!\n");
			return false;
		}
		// b_calculate_cell_path_length=true;
	}
	// LEC histogram
	if(b_create_LEC_histogram)
	{
		int i;
		if(!LECBin.allocate( no_LEC_bins + 1, 9))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating local estimate contribution histogram!\n");
			return false;
		}
		LECBin.setToZero();
		LEC_range=maximum_LEC-minimum_LEC;
		for(i=0;i<=no_LEC_bins;i++)
			LECBin[i][0]=minimum_LEC+(LEC_range*(double(i)+0.5))/double(no_LEC_bins);
	}
	return true;
}

bool FunctionalLayer::calculateEscapeOpticalThickness(class Estimate *pEstimate, class Vector3D p, struct CellIndex_s CI)
{
	int i, n, cell_index, cluster_index;
	class Estimate::GasProfileJacobian_s *pGPJ;
	class Estimate::ParticleProfileJacobian_s *pPPJ;

	// old code
	bool continue_forcing, RRS=pAtmosphere->simulateRingEffect();
	double tau=0.0, distance=0.0, tau_at_center_wavelength=0.0, tau_absorption_at_center_wavelength=0.0;
	struct CellIndex_s NextCI;
	struct Atmosphere::OpticalCell_s *pOC;
	class Vector3D next_p, direction;
	
	if(b_sun_disc_sampling)
	{
		class Vector3D reference_direction=norm(SunPosition-p);
		sampleSphericalConeFace(reference_direction, SUN_SOLID_ANGLE_RADIUS, direction);
	}	
	else
		direction=norm(SunPosition-p);
	SunDirection=-direction;	// unit vector from the Sun to the point p
	if(b_calculate_cell_path_length)
	{
		pEstimate->resetForcedUpdateStructs();
		/*
		if(b_create_sensitivity_grid_file)
			forced_cell_path_length.setToZero();
		*/
		// gases
		n=pEstimate->no_active_gases;
		pGPJ=pEstimate->GasProfileJacobian;
		for(i=0;i<n;i++)
			pGPJ++->CurrentScatterOrderContribution.setToZero();

		// particles
		n=pEstimate->no_active_particles;
		pPPJ=pEstimate->ParticleProfileJacobian;
		for(i=0;i<n;i++)
		{
			if(pPPJ->pParticle->hasSinglePhaseFunction())
				pPPJ->SPF_CurrentScatterOrderContribution.setToZero();
			else
				pPPJ->CurrentScatterOrderContribution.setToZero();
			pPPJ++;
		}
	}
	/*
		this function could potentially get optimized for the respective grid type!
	*/
	do
	{
		continue_forcing=pGrid->calculateCellIntersectionPoint(p, &CI, direction, &distance, &next_p, &NextCI);
		if(!continue_forcing)
			return false;
		cell_index=pGrid->calculateCellIndex(CI);
		pOC=pAtmosphere->getOpticalCellPtr(cell_index, wavelength);
		if(!pOC)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not set wavelength of cell\n");
			return false;
		}
		tau += distance * pOC->extinction_coefficient;
		if(RRS)
		{
			tau_at_center_wavelength += distance * pOC->extinction_coefficient_at_center_wavelength;
			if(b_eliminate_absorption)
				tau_absorption_at_center_wavelength += distance * pOC->absorption_coefficient_at_center_wavelength;
		}
		if(b_calculate_cell_path_length)
		{
			pEstimate->addLengthToForcedUpdateStruct(cell_index, distance);
			// profile Jacobians

			// gases
			n=pEstimate->no_active_gases;
			pGPJ=pEstimate->GasProfileJacobian;
			for(i=0;i<n;i++)
			{
				cluster_index=pGPJ->pGas->getClusterIndex(cell_index);
				if(cluster_index>=0)
					pGPJ->CurrentScatterOrderContribution[cluster_index]-=distance;
				pGPJ++;
			}

			// particles
			n=pEstimate->no_active_particles;
			pPPJ=pEstimate->ParticleProfileJacobian;
			for(i=0;i<n;i++)
			{
				cluster_index=pPPJ->pParticle->getClusterIndex(cell_index);
				if(cluster_index>=0)
				{
					if(pPPJ->pParticle->hasSinglePhaseFunction())
						pPPJ->SPF_CurrentScatterOrderContribution[cluster_index]-=distance;
					else
						pPPJ->CurrentScatterOrderContribution[cluster_index][0]-=distance;
				}
				pPPJ++;
			}
		}
		// update position and voxel index
		p=next_p;
		CI=NextCI;
		if(NextCI.CellPosition>=CELL_POSITION_ABOVE_TOA)
			continue_forcing=false;
		else if(NextCI.CellPosition==CELL_POSITION_SURFACE)
			return false;
	}while(continue_forcing);
	tau_forced=tau;
	F0_exp_minus_tau_forced = getF0() * exp ( -tau_forced );
	if(RRS)
	{
		tau_forced_at_center_wavelength=tau_at_center_wavelength;
		tau_forced_absorption_at_center_wavelength=tau_absorption_at_center_wavelength;
		F0_exp_minus_tau_forced_at_center_wavelength = F0_at_center_wavelength * exp ( -tau_forced_at_center_wavelength );
	}
	//tau_forced_inelastic=tau_inelastic;
	if(b_calculate_cell_path_length)
		pEstimate->registerForcedUpdateStructsInUpdateCellList();
	return true;
}

/*void FunctionalLayer::registerContributionInSensitivityGrid(class Estimate *pEstimate, const double& c)
{
	int i;
	double *cpl, *fcpl, *cs, l;
	cpl=pEstimate->CellPathLength.getCoefficientPtr();
	fcpl=forced_cell_path_length.getCoefficientPtr();
	cs=pEstimate->CellSensitivity.getCoefficientPtr();
	for(i=0;i<no_cells;i++)
	{
		l=*fcpl++ + *cpl++;
		*cs++=c*l;
	}
}*/

void FunctionalLayer::registerDirectLightContribution(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI)
{
	if(calculateEscapeOpticalThickness(pEstimate, p, CI))
	{
		double scalar_contribution=getF0()*exp(-tau_forced);
		pEstimate->direct_light_radiance+=scalar_contribution;
		pCurrentRTEvent->contribution=scalar_contribution;
	}
	else
		pCurrentRTEvent->contribution=0.0;
}

void FunctionalLayer::processSunHit(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, double weight)
{
	double scalar_contribution;

	scalar_contribution = weight * getF0();
	pEstimate->scalar_radiance += scalar_contribution;

	SVIOne[0]=scalar_contribution;
	
	pEstimate->vector_radiance+=SVIOne;

	// calculate other functionals
	pCurrentRTEvent->contribution=scalar_contribution;

	/*if(b_create_sensitivity_grid_file)
		registerContributionInSensitivityGrid(pEstimate, scalar_contribution);	*/

	/*
		the raytracer updated cluster(cell) path lengths gradually in the CumulativeContribution
		
		CurrentScatterOrderContribution contains forced cluster path lengths
		
	*/
	if(b_calculate_cell_path_length)	// synonymous for b_calculate_Jacobians
	{
		int i, j, k, n, no_clusters, cluster_size, no_phase_function_parameters;

		double *logarithmic_derivative_ptr=pEstimate->LogarithmicDerivative.getCoefficientPtr();
		double *pCC, *pCSOC, *pJ, c;

		// gases
		n=pEstimate->no_active_gases;
		class Estimate::GasProfileJacobian_s *pGPJ=pEstimate->GasProfileJacobian;
		for(k=0;k<n;k++)
		{
			no_clusters=pGPJ->no_clusters;
			pCC=pGPJ->CumulativeContribution.getCoefficientPtr();
			pCSOC=pGPJ->CurrentScatterOrderContribution.getCoefficientPtr();
			pJ=pGPJ->Jacobian.getCoefficientPtr();
			for(i=0;i<no_clusters;i++)
			{
				c = *pCC++ + *pCSOC++;
				*logarithmic_derivative_ptr++ = c;
				
				*pJ++ += c * scalar_contribution;
			}
			pGPJ++;
		}

		// particles
		n=pEstimate->no_active_particles;
		class Estimate::ParticleProfileJacobian_s *pPPJ=pEstimate->ParticleProfileJacobian;
		for(k=0;k<n;k++)
		{
			no_clusters=pPPJ->no_clusters;
			no_phase_function_parameters=pPPJ->no_phase_function_parameters;

			if(pPPJ->pParticle->hasSinglePhaseFunction())
			{
				pCC=pPPJ->SPF_CumulativeContribution.getCoefficientPtr();
				pCSOC=pPPJ->SPF_CurrentScatterOrderContribution.getCoefficientPtr();
				pJ=pPPJ->SPF_Jacobian.getCoefficientPtr();
				for(i=0;i<no_clusters;i++)
				{
					// extinction coefficient
					c = *pCC++ + *pCSOC++;

					*logarithmic_derivative_ptr++ = c;
					*pJ++ += c * scalar_contribution;
				}
					
				// single scattering albedo
				c = *pCC++;
				*logarithmic_derivative_ptr++ = c;
				*pJ++ += c * scalar_contribution;
				
				// phase function parameters
				for(j=0;j<no_phase_function_parameters;j++)
				{
					c = *pCC++;
					*logarithmic_derivative_ptr++ = c;
					*pJ++ += c * scalar_contribution;
				}
			}
			else // not hasSinglePhaseFunction ------------------------------------
			{
				pCC=pPPJ->CumulativeContribution.getCoefficientPtr();
				pCSOC=pPPJ->CurrentScatterOrderContribution.getCoefficientPtr();
				pJ=pPPJ->Jacobian.getCoefficientPtr();
				cluster_size=pPPJ->cluster_size;
				for(i=0;i<no_clusters;i++)
				{
					// extinction coefficient
					c = *pCC++ + *pCSOC;

					*logarithmic_derivative_ptr++ = c;
					*pJ++ += c * scalar_contribution;
					
					// single scattering albedo
					c = *pCC++;

					*logarithmic_derivative_ptr++ = c;
					*pJ++ += c * scalar_contribution;
					
					// phase function parameters
					for(j=0;j<no_phase_function_parameters;j++)
					{
						c = *pCC++;

						*logarithmic_derivative_ptr++ = c;
						*pJ++ += c * scalar_contribution;
					}
					
					// done
					pCSOC+=cluster_size;
				}
			}
			pPPJ++;
		}

		// calculate all logarithmic derivatives, now calculate Hessian
		if(b_calculate_Hessian)
			pEstimate->accumulateSimpleHessianContributions(scalar_contribution, SVIOne);
	}
}

void FunctionalLayer::registerContributionInLECHistogram(const double& LEC)
{
	if(LEC<minimum_LEC || LEC>maximum_LEC)
	{
		LECBin[no_LEC_bins][1]+=LEC;
		LECBin[no_LEC_bins][2]+=LEC*LEC;
		LECBin[no_LEC_bins][3]++;
	}
	else
	{
		int index=int(double(no_LEC_bins)*(LEC-minimum_LEC)/LEC_range)%no_LEC_bins;
		LECBin[index][1]+=LEC;
		LECBin[index][2]+=LEC*LEC;
		LECBin[index][3]++;
	}
}

/*void FunctionalLayer::updateVarianceReductionParameters()
{
	VRGrid.updateDIVRSettings();
	VRGrid.exportToBinaryGridFile((char *) "directional_importances.bgf");
	VRGrid.writeCellData((char *) "cell_importances.txt");
}*/

void FunctionalLayer::processPropagation(class Estimate *pEstimate, const int& cell_index, const double &length)
{
	/*if(b_create_sensitivity_grid_file)
		pEstimate->CellPathLength[cell_index]+=length;*/
	if(b_calculate_cell_path_length)
		pEstimate->addLengthToCumulativeUpdateStruct(cell_index, length);
	if(pEstimate->calculateJacobians())
	{
		int i, n, cluster_index;

		// gases
		n=pEstimate->no_active_gases;
		struct Estimate::GasProfileJacobian_s *pGPJ=pEstimate->GasProfileJacobian;
		for(i=0;i<n;i++)
		{
			cluster_index=pGPJ->pGas->getClusterIndex(cell_index);
			if(cluster_index>=0)
				pGPJ->CumulativeContribution[cluster_index]-=length;
			pGPJ++;
		}

		// particles
		n=pEstimate->no_active_particles;
		struct Estimate::ParticleProfileJacobian_s *pPPJ=pEstimate->ParticleProfileJacobian;
		class Particle *pParticle;
		for(i=0;i<n;i++)
		{
			pParticle=pPPJ->pParticle;
			cluster_index=pParticle->getClusterIndex(cell_index);
			if(cluster_index>=0)
			{
				if(pParticle->hasSinglePhaseFunction())
					pPPJ->SPF_CumulativeContribution[cluster_index]-=length;
				else
					pPPJ->CumulativeContribution[cluster_index][0]-=length;
			}
			pPPJ++;
		}
	}

	struct Atmosphere::OpticalCell_s *pOC=pAtmosphere->getOpticalCellPtr(cell_index);
	pEstimate->tau += length * pOC->extinction_coefficient;
	
	// RRS weight
	if(wavelength!=central_wavelength)
	{
		double delta_extinction_coefficient;
		delta_extinction_coefficient = pOC->extinction_coefficient_at_center_wavelength;
		delta_extinction_coefficient -= pOC->extinction_coefficient;
		pEstimate->tau_RRS += length * delta_extinction_coefficient;
	}
	if(b_eliminate_absorption)
		pEstimate->tau_absorption += length * pOC->absorption_coefficient_at_center_wavelength;
}

void FunctionalLayer::processGroundScatterEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI, const class Vector3D& surface_normal)
{
	if(calculateEscapeOpticalThickness(pEstimate, p, CI))
	{
		double scalar_contribution, weight=pEstimate->weight, c;
		c = weight * pEstimate->RRS_adjoint_weight * fabs( SunDirection * surface_normal ) / M_PI;

		scalar_contribution = c * F0_exp_minus_tau_forced;
		pEstimate->scalar_radiance += scalar_contribution;

		SVIOne[0]=scalar_contribution;
		/*vector_F0[1]=0.0;
		vector_F0[2]=0.0;
		vector_F0[3]=0.0;*/
		
		pEstimate->vector_radiance+=SVIOne;
		/*estimated_scatter_order[pEstimate->getScatterOrder()]=true;
		pEstimate->increaseScatterOrder();*/

		if(pAtmosphere->simulateRingEffect())
		{
			// event counts
			if(b_event_counts)
			{
				int nc=pEstimate->getNORRSEvents();
				if(nc>1)
					pEstimate->scalar_multiple_RRS_count += scalar_contribution * double(nc);
				else
					pEstimate->scalar_RRS_count += scalar_contribution * double(nc);
			}

			// calculate RRS weight
			weight=pEstimate->RRS_elastic_weight*=exp( - pEstimate->tau_RRS );
			pEstimate->tau_RRS=0.0;

			// calculate contribution without RRS
			c *= weight * F0_exp_minus_tau_forced_at_center_wavelength;
			if(b_eliminate_absorption)
				c *= exp(pEstimate->tau_absorption+tau_forced_absorption_at_center_wavelength);

			pEstimate->elastic_scalar_radiance += c;

			SVIOne[0] = c;
			pEstimate->elastic_vector_radiance += SVIOne;
			
		}
		if(b_event_counts)
		{
			pEstimate->scalar_elastic_scatter_count += scalar_contribution * double(pEstimate->getNOElasticScatterEvents());
			pEstimate->scalar_ground_scatter_count += scalar_contribution * double(pEstimate->getNOGroundScatterEvents());
		}

		// calculate other functionals

		pCurrentRTEvent->contribution=scalar_contribution;

		/*if(b_enable_variance_reduction)
		{
			double weight=pEstimate->weight;
			ScatterOrderContributionMatrix[no_scatter_events][0]+=scalar_contribution;
			ScatterOrderContributionMatrix[no_scatter_events][3]+=weight;
			ScatterOrderContributionMatrix[no_scatter_events][4]++;
			ScatterOrderContributionMatrix[no_scatter_events][5]+=scalar_contribution/weight;
		}*/

		if(b_create_sensitivity_grid_file)
			pEstimate->registerLocalEstimateInSensitivityGrid(scalar_contribution);

		/*
			the raytracer updated cluster(cell) path lengths gradually in the CumulativeContribution
			
			CurrentScatterOrderContribution contains forced cluster path lengths
			
		*/
		if(b_calculate_cell_path_length)	// synonymous for b_calculate_Jacobians
		{
			
			int i, j, k, n, no_clusters, cluster_size, no_phase_function_parameters;

			double c, *logarithmic_derivative_ptr=pEstimate->LogarithmicDerivative.getCoefficientPtr();
			double *pCC, *pCSOC, *pJ;

			// gases
			n=pEstimate->no_active_gases;
			class Estimate::GasProfileJacobian_s *pGPJ=pEstimate->GasProfileJacobian;
			for(k=0;k<n;k++)
			{
				no_clusters=pGPJ->no_clusters;
				pCC=pGPJ->CumulativeContribution.getCoefficientPtr();
				pCSOC=pGPJ->CurrentScatterOrderContribution.getCoefficientPtr();
				pJ=pGPJ->Jacobian.getCoefficientPtr();
				for(i=0;i<no_clusters;i++)
				{
					c = *pCC++ + *pCSOC++;
					*logarithmic_derivative_ptr++ = c;
					
					*pJ++ += c * scalar_contribution;
				}
				pGPJ++;
			}

			// particles
			n=pEstimate->no_active_particles;
			class Estimate::ParticleProfileJacobian_s *pPPJ=pEstimate->ParticleProfileJacobian;
			for(k=0;k<n;k++)
			{
				no_clusters=pPPJ->no_clusters;
				no_phase_function_parameters=pPPJ->no_phase_function_parameters;
				// no_parameters=pPPJ->no_parameters;
				if(pPPJ->pParticle->hasSinglePhaseFunction())
				{
					pCC=pPPJ->SPF_CumulativeContribution.getCoefficientPtr();
					pCSOC=pPPJ->SPF_CurrentScatterOrderContribution.getCoefficientPtr();
					pJ=pPPJ->SPF_Jacobian.getCoefficientPtr();
					for(i=0;i<no_clusters;i++)
					{
						// extinction coefficient
						c = *pCC++ + *pCSOC++;

						*logarithmic_derivative_ptr++ = c;
						*pJ++ += c * scalar_contribution;
					}
						
					// single scattering albedo
					c = *pCC++;
					*logarithmic_derivative_ptr++ = c;
					*pJ++ += c * scalar_contribution;
					
					// phase function parameters
					for(j=0;j<no_phase_function_parameters;j++)
					{
						c = *pCC++;
						*logarithmic_derivative_ptr++ = c;
						*pJ++ += c * scalar_contribution;
					}
				}
				else // not hasSinglePhaseFunction ------------------------------------
				{
					pCC=pPPJ->CumulativeContribution.getCoefficientPtr();
					pCSOC=pPPJ->CurrentScatterOrderContribution.getCoefficientPtr();
					pJ=pPPJ->Jacobian.getCoefficientPtr();
					cluster_size=pPPJ->cluster_size;
					for(i=0;i<no_clusters;i++)
					{
						// extinction coefficient
						c = *pCC++ + *pCSOC;

						*logarithmic_derivative_ptr++ = c;
						*pJ++ += c * scalar_contribution;
						
						// single scattering albedo
						c = *pCC++;

						*logarithmic_derivative_ptr++ = c;
						*pJ++ += c * scalar_contribution;
						
						// phase function parameters
						for(j=0;j<no_phase_function_parameters;j++)
						{
							c = *pCC++;

							*logarithmic_derivative_ptr++ = c;
							*pJ++ += c * scalar_contribution;
						}
						
						// done
						pCSOC+=cluster_size;
					}
				}
				pPPJ++;
			}

			// calculate all logarithmic derivatives, now calculate Hessian
			if(b_calculate_Hessian)
				pEstimate->accumulateSimpleHessianContributions(scalar_contribution, SVIOne);
		}
	}
	else
		pCurrentRTEvent->contribution=0.0;

	pEstimate->resetTransitionStokesMatrix();
}

void FunctionalLayer::calculateScatterPhaseMatrices(class Atmosphere::OpticalCell_s *pOC, const double& mu, class StokesMatrix& ScatterPhaseMatrix, class StokesMatrix& ElasticScatterPhaseMatrix, double& esP, double& esP_at_center_wavelength)
{
	int i;
	double scatter_coefficient, PhF;

	// 1.) Rayleigh scattering 
	PhF = pMolecularPhF->evaluate(mu);
	pMolecularPhF->PhaseFunction::initializePhaseMatrix (mu);

	scatter_coefficient = pOC -> rayleigh_scatter_coefficient ;
	esP = scatter_coefficient * PhF;
	ScatterPhaseMatrix = pMolecularPhF->getPhaseMatrix () * scatter_coefficient;

	scatter_coefficient = pOC -> rayleigh_scatter_coefficient_at_center_wavelength ;
	esP_at_center_wavelength = scatter_coefficient * PhF;
	ElasticScatterPhaseMatrix = pMolecularPhF->getPhaseMatrix () * scatter_coefficient;

	// 2.) rotational Raman scattering
	if(pAtmosphere->simulateRingEffect())
	{
		// 3.) rotational Raman scattering 
		PhF=pRotationalRamanPhF->evaluate (mu);
		pRotationalRamanPhF->PhaseFunction::initializePhaseMatrix (mu);

		scatter_coefficient = pOC -> rotational_raman_scatter_coefficient;
		esP += scatter_coefficient * PhF;
		ScatterPhaseMatrix += pRotationalRamanPhF->getPhaseMatrix () * scatter_coefficient;

		scatter_coefficient = pOC -> rotational_raman_scatter_coefficient_at_center_wavelength;
		esP_at_center_wavelength += scatter_coefficient * PhF;
		ElasticScatterPhaseMatrix += pRotationalRamanPhF->getPhaseMatrix () * scatter_coefficient;
	}

	// 3.) particle scattering
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s **pPOSPtr = pOC -> ParticleOpticalStructPtr, *pPOS;
	for(i = 0 ; i<pOC -> no_particles ; i ++)
	{
		pPOS=*pPOSPtr;

		pPhaseFunction = pPOS->pPhaseFunction;
		PhF = pPOS->PhaseFunctionValue = pPhaseFunction -> evaluate (mu);
		pPhaseFunction -> initializePhaseMatrix(pPOS->ScatterPhaseMatrix, mu);

		scatter_coefficient = pPOS->scatter_coefficient;
		esP += scatter_coefficient * PhF;
		ScatterPhaseMatrix += pPOS->ScatterPhaseMatrix * scatter_coefficient;

		scatter_coefficient = pPOS->scatter_coefficient_at_center_wavelength;
		esP_at_center_wavelength += scatter_coefficient * PhF;
		ElasticScatterPhaseMatrix += pPOS->ScatterPhaseMatrix * scatter_coefficient;

		pPOSPtr ++;
	}
}

void FunctionalLayer::calculateScatterPhaseMatrices(class Atmosphere::OpticalCell_s *pOC, const double& mu, class StokesMatrix& ScatterPhaseMatrix, double& esP)
{
	int i;
	double scatter_coefficient, PhF;

	// 1.) Rayleigh scattering 
	PhF = pMolecularPhF->evaluate(mu);
	pMolecularPhF->PhaseFunction::initializePhaseMatrix (mu);

	scatter_coefficient = pOC -> rayleigh_scatter_coefficient ;
	esP = scatter_coefficient * PhF;
	ScatterPhaseMatrix = pMolecularPhF->getPhaseMatrix () * scatter_coefficient;

	// 2.) rotational Raman scattering
	if(pAtmosphere->simulateRingEffect())
	{
		// 3.) rotational Raman scattering 
		PhF=pRotationalRamanPhF->evaluate (mu);
		pRotationalRamanPhF->PhaseFunction::initializePhaseMatrix (mu);

		scatter_coefficient = pOC -> rotational_raman_scatter_coefficient;
		esP += scatter_coefficient * PhF;
		ScatterPhaseMatrix += pRotationalRamanPhF->getPhaseMatrix () * scatter_coefficient;
	}

	// 3.) particle scattering
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s **pPOSPtr = pOC -> ParticleOpticalStructPtr, *pPOS;
	for(i = 0 ; i<pOC -> no_particles ; i ++)
	{
		pPOS=*pPOSPtr;

		pPhaseFunction = pPOS->pPhaseFunction;
		PhF = pPOS->PhaseFunctionValue = pPhaseFunction -> evaluate (mu);
		pPhaseFunction -> initializePhaseMatrix(pPOS->ScatterPhaseMatrix, mu);

		scatter_coefficient = pPOS->scatter_coefficient;
		esP += scatter_coefficient * PhF;
		ScatterPhaseMatrix += pPOS->ScatterPhaseMatrix * scatter_coefficient;

		pPOSPtr ++;
	}
}

void FunctionalLayer::calculateDifferentialScatteringCoefficient(class Atmosphere::OpticalCell_s *pOC, const double& mu, double& esP)
{
	int i;
	double scatter_coefficient, PhF;

	// 1.) Rayleigh scattering 
	PhF = pMolecularPhF->evaluate(mu);

	scatter_coefficient = pOC -> rayleigh_scatter_coefficient ;
	esP = scatter_coefficient * PhF;

	// 2.) rotational Raman scattering
	if(pAtmosphere->simulateRingEffect())
	{
		// 3.) rotational Raman scattering 
		PhF=pRotationalRamanPhF->evaluate (mu);

		scatter_coefficient = pOC -> rotational_raman_scatter_coefficient;
		esP += scatter_coefficient * PhF;
	}

	// 3.) particle scattering
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s **pPOSPtr = pOC -> ParticleOpticalStructPtr, *pPOS;
	for(i = 0 ; i<pOC -> no_particles ; i ++)
	{
		pPOS=*pPOSPtr;

		pPhaseFunction = pPOS->pPhaseFunction;
		PhF = pPOS->PhaseFunctionValue = pPhaseFunction -> evaluate (mu);

		scatter_coefficient = pPOS->scatter_coefficient;
		esP += scatter_coefficient * PhF;

		pPOSPtr ++;
	}
}

void FunctionalLayer::calculateElasticDifferentialScatteringCoefficient(class Atmosphere::OpticalCell_s *pOC, const double& mu, double& esP)
{
	int i;
	double scatter_coefficient, PhF;

	// 1.) Rayleigh scattering 
	PhF = pMolecularPhF->evaluate(mu);

	scatter_coefficient = pOC -> rayleigh_scatter_coefficient ;
	esP = scatter_coefficient * PhF;

	// 3.) particle scattering
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s **pPOSPtr = pOC -> ParticleOpticalStructPtr, *pPOS;
	for(i = 0 ; i<pOC -> no_particles ; i ++)
	{
		pPOS=*pPOSPtr;

		pPhaseFunction = pPOS->pPhaseFunction;
		PhF = pPOS->PhaseFunctionValue = pPhaseFunction -> evaluate (mu);

		scatter_coefficient = pPOS->scatter_coefficient;
		esP += scatter_coefficient * PhF;

		pPOSPtr ++;
	}
}

void FunctionalLayer::processAtmosphericAbsorptionEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, class Atmosphere::OpticalCell_s *pOC)
{
	class StokesVector vector_contribution;
	double scalar_contribution, elastic_scalar_contribution, weight=pEstimate->weight, c;
	double RRS_adjoint_weight;

	if(pAtmosphere->simulateRingEffect())
	{
		RRS_adjoint_weight = pEstimate->RRS_adjoint_weight;

		// scalar contribution 
		scalar_contribution = weight * RRS_adjoint_weight * pOC->emission;
	
		// vector contribution
		SVIOne[0] = scalar_contribution;
		vector_contribution = pEstimate->TransitionStokesMatrix * SVIOne;

		// update Neumann series
		pEstimate->scalar_radiance += scalar_contribution;
		pEstimate->vector_radiance += vector_contribution;
		
		// ################ elastic biasing radiances importance sampling ###################################

		// calculate RRS weight
		pEstimate->RRS_elastic_weight *= exp( - pEstimate->tau_RRS );

		// pEstimate->RRS_elastic_weight *= pEstimate->RRS_adjoint_extinction / pOC->extinction_coefficient_at_center_wavelength; 
		pEstimate->tau_RRS=0.0;

		c = weight * pEstimate->RRS_elastic_weight * pOC->emission_at_center_wavelength;

		// consider absorption weight
		if(b_eliminate_absorption)
			c *= exp ( pEstimate->tau_absorption + tau_forced_absorption_at_center_wavelength );

		// scalar contribution 
		elastic_scalar_contribution = c ;
		pEstimate->elastic_scalar_radiance += elastic_scalar_contribution;
	
		// vector contribution
		SVIOne[0] = c;
		vector_contribution = pEstimate->RRSTransitionStokesMatrix * SVIOne;
		pEstimate->elastic_vector_radiance += vector_contribution;

	}
	else
	{	
		// elastic scattering ************************************************************************
		
		c = weight * pOC->emission;
		// scalar contribution 
		scalar_contribution = c ;
		pEstimate->scalar_radiance += scalar_contribution;
	
		// vector contribution
		SVIOne[0] = c ; 
		vector_contribution = pEstimate->TransitionStokesMatrix * SVIOne;
		pEstimate->vector_radiance += vector_contribution;
		
	}

	// register scalar contribution ---------------------------------------------------------------
	pCurrentRTEvent->contribution=scalar_contribution;
}

void FunctionalLayer::processAtmosphericScatterEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI, class Atmosphere::OpticalCell_s *pOC, const class Vector3D& k_out)
{
	/*
		n: backward MC scatter event ordering
		k_out=-k_n
		k_in=-k_{n+1}
	*/
	int i, j, k, no_clusters, no_phase_function_parameters, cluster_index, cluster_size;
	int no_active_gases, no_active_particles, no_parameters; // , cell_index;
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s *pPOS;
	bool is_active_cell=pOC->is_active_cell;
	double c, mu, sigma1, sigma2, esP;
	double *pCC, *pCSOC, *pJ, *logarithmic_derivative_ptr;

	class Estimate::GasProfileJacobian_s *pGPJ;
	class Estimate::ParticleProfileJacobian_s *pPPJ;

	if(calculateEscapeOpticalThickness(pEstimate, p, CI))
	{
		class StokesVector vector_contribution, F_vector;
		double scalar_contribution, elastic_scalar_contribution, weight=pEstimate->weight;
		double RRS_adjoint_weight;

		// cell_index=pGrid->calculateCellIndex(CI);
		pPhaseFunction = pCurrentRTEvent->pPhaseFunction ;

		// calculate effective phase function value ##################################
		PhaseFunction :: calculateScatterAngles (SunDirection, k_out, mu, sigma1, sigma2);
		if(pAtmosphere->simulateRingEffect())
		{
			double esP_at_center_wavelength;
			class StokesMatrix ElasticScatterPhaseMatrix;
			RRS_adjoint_weight = pEstimate->RRS_adjoint_weight;

			c = weight * RRS_adjoint_weight * F0_exp_minus_tau_forced / ( 4.0 * M_PI );

			// scalar contribution 
			scalar_contribution = c * pPhaseFunction->evaluate(mu) ;
		
			// vector contribution
			pPhaseFunction->initializePhaseMatrix(ScatterPhaseMatrix, mu);
			ScatterPhaseMatrix = StokesMatrix(-sigma2) * ( ScatterPhaseMatrix * StokesMatrix(M_PI-sigma1) );
			SVIOne[0] = c;
			vector_contribution = pEstimate->TransitionStokesMatrix * (ScatterPhaseMatrix * SVIOne);
	
			// update Neumann series
			pEstimate->scalar_radiance += scalar_contribution;
			pEstimate->vector_radiance += vector_contribution;
			
			// event counts
			if(b_event_counts)
			{
				int nc=pEstimate->getNORRSEvents();
				if(nc>1)
					pEstimate->scalar_multiple_RRS_count += scalar_contribution * double(nc);
				else
					pEstimate->scalar_RRS_count += scalar_contribution * double(nc);
				pEstimate->scalar_elastic_scatter_count += scalar_contribution * double(pEstimate->getNOElasticScatterEvents());
				pEstimate->scalar_ground_scatter_count += scalar_contribution * double(pEstimate->getNOGroundScatterEvents());
			}

			// ################ elastic biasing radiances importance sampling ###################################

			calculateScatterPhaseMatrices(pOC, mu, ScatterPhaseMatrix, ElasticScatterPhaseMatrix, esP, esP_at_center_wavelength);

			// calculate RRS weight
			pEstimate->RRS_elastic_weight *= exp( - pEstimate->tau_RRS );
			// pEstimate->RRS_elastic_weight *= pEstimate->RRS_adjoint_extinction / pOC->extinction_coefficient_at_center_wavelength; 
			pEstimate->tau_RRS=0.0;

			c = weight * pEstimate->RRS_elastic_weight * F0_exp_minus_tau_forced_at_center_wavelength / (4.0 * M_PI );
			// consider absorption weight
			if(b_eliminate_absorption)
				c *= exp ( pEstimate->tau_absorption + tau_forced_absorption_at_center_wavelength );

			// scalar contribution 
			elastic_scalar_contribution = c * esP_at_center_wavelength / pOC->scatter_coefficient_at_center_wavelength;
			pEstimate->elastic_scalar_radiance += elastic_scalar_contribution;
		
			// vector contribution
			c/=pOC->scatter_coefficient_at_center_wavelength;
			ElasticScatterPhaseMatrix=StokesMatrix(-sigma2)*(ElasticScatterPhaseMatrix*StokesMatrix(M_PI-sigma1));
			SVIOne[0] = c;
			vector_contribution = pEstimate->RRSTransitionStokesMatrix * ElasticScatterPhaseMatrix * SVIOne;
			pEstimate->elastic_vector_radiance += vector_contribution;

		}
		else
		{	
			// elastic scattering ************************************************************************
			
			c = weight * F0_exp_minus_tau_forced / ( 4.0 * M_PI );
			// scalar contribution 
			scalar_contribution = c * pPhaseFunction->evaluate(mu) ;
			pEstimate->scalar_radiance += scalar_contribution;
		
			// vector contribution
			pPhaseFunction->initializePhaseMatrix(ScatterPhaseMatrix, mu);
			ScatterPhaseMatrix=StokesMatrix(-sigma2)*(ScatterPhaseMatrix*StokesMatrix(M_PI-sigma1));
			SVIOne[0] = c ; 
			vector_contribution = pEstimate->TransitionStokesMatrix * (ScatterPhaseMatrix * SVIOne);
			pEstimate->vector_radiance += vector_contribution;
			
			// event counts
			if(b_event_counts)
			{
				pEstimate->scalar_elastic_scatter_count += scalar_contribution * double(pEstimate->getNOElasticScatterEvents());
				pEstimate->scalar_ground_scatter_count += scalar_contribution * double(pEstimate->getNOGroundScatterEvents());
			}
			if(is_active_cell)
				calculateDifferentialScatteringCoefficient(pOC, mu, esP);
		}

		// register scalar contribution ---------------------------------------------------------------
		pCurrentRTEvent->contribution=scalar_contribution;
		
		// ....
		if(b_create_sensitivity_grid_file)
			pEstimate->registerLocalEstimateInSensitivityGrid(scalar_contribution);
		
		if(b_calculate_cell_path_length)
		{
			/*
				gas derivatives:
				
					J = sum_n { w_n * ( - l_n^c - l_n^f ) }

				particle derivatives:

					J = sum_n { w_n * ( - l_n^c - l_n^f + y_n^c + y_n^f ) }

			*/
			
			// calculate cumulative part of the components of the logarithmic derivative vector

			logarithmic_derivative_ptr=pEstimate->LogarithmicDerivative.getCoefficientPtr();
			// gases
			no_active_gases=pEstimate->no_active_gases;
			pGPJ=pEstimate->GasProfileJacobian;
			for(k=0;k<no_active_gases;k++)
			{
				no_clusters=pGPJ->no_clusters;
				pCC=pGPJ->CumulativeContribution.getCoefficientPtr();
				pCSOC=pGPJ->CurrentScatterOrderContribution.getCoefficientPtr();
				for(i=0;i<no_clusters;i++)
					*logarithmic_derivative_ptr++ = *pCC++ + *pCSOC++;
				pGPJ++;
			}
			// gas derivatives are ready here

			// particles
			no_active_particles=pEstimate->no_active_particles;
			pPPJ=pEstimate->ParticleProfileJacobian;
			for(k=0;k<no_active_particles;k++)
			{
				no_clusters=pPPJ->no_clusters;
				no_phase_function_parameters=pPPJ->no_phase_function_parameters;
				if(pPPJ->pParticle->hasSinglePhaseFunction())
				{
					pCC=pPPJ->SPF_CumulativeContribution.getCoefficientPtr();
					pCSOC=pPPJ->SPF_CurrentScatterOrderContribution.getCoefficientPtr();

					// extinction coefficient
					for(i=0;i<no_clusters;i++)
						*logarithmic_derivative_ptr++ = *pCC++ + *pCSOC++;

					// single scattering albedo
					*logarithmic_derivative_ptr++ = *pCC++;
					
					// phase function parameters
					for(j=0;j<no_phase_function_parameters;j++)
						*logarithmic_derivative_ptr++ = *pCC++;
				}
				else
				{
					cluster_size=pPPJ->cluster_size;
					pCC=pPPJ->CumulativeContribution.getCoefficientPtr();
					pCSOC=pPPJ->CurrentScatterOrderContribution.getCoefficientPtr();
					for(i=0;i<no_clusters;i++)
					{
						// extinction coefficient
						*logarithmic_derivative_ptr++ = *pCC++ + *pCSOC;
						
						// single scattering albedo
						*logarithmic_derivative_ptr++ = *pCC++;
						
						// phase function parameters
						for(j=0;j<no_phase_function_parameters;j++)
							*logarithmic_derivative_ptr++ = *pCC++;
						
						// cluster done
						pCSOC+=cluster_size;
					}
				}
				pPPJ++;
			}
			// particle derivatives: cumulative scatter event part and path length part (both: cum. and curr.s.o.) have been calculated

			// proceed with additional scatter-event-related contributions to the particle logarithmic derivatives
			if(is_active_cell)
			{
				//	particle optical property Jacobian:

				//		the edges of the trajectory contribute through their path lengths
				//		to the Jacobian (-> call of accumulateClusterPathLengths)
				//		a single scatter event contributes only in the cluster associated with the
				//		location of the scatter event.
						
				//		For a single scatter event two quantities must be calculated, one for the
				//		scatter events that follow (mu_trajectory) and the other is the contribution 
				//		to the current scatter order (mu_forced)

				no_active_particles=pOC->no_active_particles;
				for(k=0;k<no_active_particles;k++)
				{
					pPOS=pOC->ActiveParticleOpticalStructPtr[k];
					pParticle=pPOS->pParticle;
					no_clusters=pParticle->getNOClusters();
					no_phase_function_parameters=pParticle->getNOPhaseFunctionParameters();
					
					// calculate index of logarithmic derivative
					cluster_index=pPOS->cluster_index;
					logarithmic_derivative_ptr=pEstimate->LogarithmicDerivative.getCoefficientPtr()+pParticle->getJacobianOffset();
					
					pPhaseFunction=pPOS->pPhaseFunction;
					pPhaseFunction->evaluateJacobian(mu);

					// extinction coefficient Jacobians
					if(pParticle->hasSinglePhaseFunction())
					{
						logarithmic_derivative_ptr[cluster_index] += (pPOS->varpi0 * pPOS->PhaseFunctionValue) / esP;
						logarithmic_derivative_ptr+=no_clusters;
					}
					else
					{
						// logarithmic_derivative_ptr += cluster_index*(2+no_phase_function_parameters);
						logarithmic_derivative_ptr += cluster_index*pParticle->getClusterDimension();
						*logarithmic_derivative_ptr++ += (pPOS->varpi0 * pPOS->PhaseFunctionValue) / esP;
					}

					// single scattering albedo
					*logarithmic_derivative_ptr++ += (pPOS->extinction_coefficient * pPOS->PhaseFunctionValue) / esP;
					
					// phase function parameter derivatives
					for(i=0;i<no_phase_function_parameters;i++)
						*logarithmic_derivative_ptr++ += (pPhaseFunction->Jacobian(i)*pPOS->scatter_coefficient) / esP ;

					if(b_calculate_Hessian)
						pPhaseFunction->evaluateHessian(mu);
					
				}
			}

			// calculate Jacobians finally ---------------------------------------------------------------------------------------
			// input is the logarithmic derivative vector

			logarithmic_derivative_ptr=pEstimate->LogarithmicDerivative.getCoefficientPtr();

			// gases
			no_active_gases=pEstimate->no_active_gases;
			pGPJ=pEstimate->GasProfileJacobian;
			for(k=0;k<no_active_gases;k++)
			{
				no_clusters=pGPJ->no_clusters;
				pJ=pGPJ->Jacobian.getCoefficientPtr();
				for(i=0;i<no_clusters;i++)
					*pJ++ += *logarithmic_derivative_ptr++ * scalar_contribution;
				pGPJ++;
			}

			// particles
			no_active_particles=pEstimate->no_active_particles;
			pPPJ=pEstimate->ParticleProfileJacobian;
			for(k=0;k<no_active_particles;k++)
			{
				if(pPPJ->pParticle->hasSinglePhaseFunction())
					pJ=pPPJ->SPF_Jacobian.getCoefficientPtr();
				else
					pJ=pPPJ->Jacobian.getCoefficientPtr();
				no_parameters=pPPJ->no_parameters;
				
				for(i=0;i<no_parameters;i++)
					*pJ++ += *logarithmic_derivative_ptr++ * scalar_contribution;
				pPPJ++;
			}
				
			// calculated all logarithmic derivatives, now calculate Hessian!
			if(b_calculate_Hessian)
			{
				pEstimate->accumulateSimpleHessianContributions(scalar_contribution, vector_contribution);
				if(is_active_cell)
				{
					int index_i,index_j, i_i,i_j;
					int no_phase_function_parameters_i, no_phase_function_parameters_j;

					int k_POS_i, k_POS_j;

					bool has_single_phase_function_i, has_single_phase_function_j;
										
					struct Particle::OpticalStruct_s *pPOS_i, *pPOS_j;
					class Particle *pParticle_i, *pParticle_j;
					class PhaseFunction *pPhaseFunction_i, *pPhaseFunction_j;
					// only particles have significant contributions here

					for(k_POS_i=0;k_POS_i<pOC->no_active_particles;k_POS_i++)
					{
						pPOS_i=pOC->ActiveParticleOpticalStructPtr[k_POS_i];
						pParticle_i=pPOS_i->pParticle;
						index_i=pParticle_i->getJacobianOffset();

						has_single_phase_function_i=pParticle_i->hasSinglePhaseFunction();

						if(has_single_phase_function_i)
							index_i+=pPOS_i->cluster_index;
						else
							index_i+=pPOS_i->cluster_index*pParticle_i->getClusterDimension();
					
						// extinction coefficient --------------------------------------------------------------------
						for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
						{
							pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
							pParticle_j=pPOS_j->pParticle;
							index_j=pParticle_j->getJacobianOffset();

							has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

							if(has_single_phase_function_j)
								index_j+=pPOS_j->cluster_index;
							else
								index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();

							// extinction coefficient
							c=-( pPOS_i->varpi0 * pPOS_i->PhaseFunctionValue * pPOS_j->varpi0 * pPOS_j->PhaseFunctionValue ) / (esP*esP);
							pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;
							
							if(has_single_phase_function_j)
								index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
							else // single scattering albedo is next to extinction coefficient inside a cluster
								index_j++;

							// single scattering albedo
							if(pPOS_j==pPOS_i)
								c=esP * pPOS_j->PhaseFunctionValue;
							else
								c=0.0;
							c -=  pPOS_j->varpi0 * pPOS_j->PhaseFunctionValue  *  pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue ;
							c /= esP * esP;
							 pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;

							index_j++;
							
							// phase function parameters
							pPhaseFunction_j = pPOS_j->pPhaseFunction;
							no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
							for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
							{
								if(pPOS_j==pPOS_i)
									c = esP * pPOS_i->varpi0;
								else
									c = 0.0;
								c -= pPOS_i->varpi0 * pPOS_i->PhaseFunctionValue * pPOS_j->extinction_coefficient * pPOS_j->varpi0;
								c *= pPhaseFunction_j->Jacobian(i_j) / (esP * esP);
								pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;
								index_j++;
							}
						}
						
						if(has_single_phase_function_i)
							index_i=pParticle_i->getJacobianOffset()+pParticle_i->getNOClusters();
						else // single scattering albedo is next to extinction coefficient inside a cluster
							index_i++;

						// single scattering albedo -----------------------------------------------------------------
						for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
						{
							pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
							pParticle_j=pPOS_j->pParticle;
							index_j=pParticle_j->getJacobianOffset();

							has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

							if(has_single_phase_function_j)
								index_j+=pPOS_j->cluster_index;
							else
								index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();

							// extinction coefficient
							if(pPOS_j==pPOS_i)
								c=esP * pPOS_j->PhaseFunctionValue;
							else
								c=0.0;
							
							c -= pPOS_j -> varpi0 * pPOS_j -> PhaseFunctionValue * pPOS_i -> extinction_coefficient * pPOS_i -> PhaseFunctionValue;
							c /= esP*esP;

							pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;

							if(has_single_phase_function_j)
								index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
							else // single scattering albedo is next to extinction coefficient inside a cluster
								index_j++;

							// single scattering albedo
							c = - pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue * pPOS_j->extinction_coefficient * pPOS_j->PhaseFunctionValue ;
							c /= esP * esP;

							pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;

							index_j++;
							
							// phase function parameters
							pPhaseFunction_j=pPOS_j->pPhaseFunction;
							no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
							for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
							{
								if(pPOS_j==pPOS_i)
									c=esP *pPOS_i->extinction_coefficient;
								else
									c=0.0;
								c-= pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue * pPOS_j->varpi0 * pPOS_j->extinction_coefficient;

								c*= pPhaseFunction_j->Jacobian(i_j) / (esP * esP);

								pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;
								index_j++;
							}
						}

						index_i++;

						// phase function parameters ----------------------------------------------------
						pPhaseFunction_i = pPOS_i->pPhaseFunction;
						no_phase_function_parameters_i=pParticle_i->getNOPhaseFunctionParameters();
						for(i_i=0;i_i<no_phase_function_parameters_i;i_i++)
						{
							for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
							{
								pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
								pParticle_j=pPOS_j->pParticle;

								index_j=pParticle_j->getJacobianOffset();
								has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

								if(has_single_phase_function_j)
									index_j+=pPOS_j->cluster_index;
								else
									index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();
								

								// extinction coefficient
								if(pPOS_j==pPOS_i)
									c = esP * pPOS_j->varpi0;
								else
									c = 0.0;
								
								c -= pPOS_j -> varpi0 * pPOS_j -> PhaseFunctionValue * pPOS_i -> extinction_coefficient * pPOS_i -> varpi0;
								c *= pPhaseFunction_i->Jacobian(i_i) / ( esP * esP);

								pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;

								
								if(has_single_phase_function_j)
									index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
								else // single scattering albedo is next to extinction coefficient inside a cluster
									index_j++;

								// single scattering albedo
								if(pPOS_j==pPOS_i)
									c = esP * pPOS_j->extinction_coefficient;
								else
									c=0.0;
								c -= pPOS_j->extinction_coefficient * pPOS_j->PhaseFunctionValue * pPOS_i->varpi0 * pPOS_i->extinction_coefficient;
								c *= pPhaseFunction_i->Jacobian(i_i) / ( esP * esP );
								pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;

								index_j++;
								
								// phase function parameters
								pPhaseFunction_j=pPOS_j->pPhaseFunction;
								no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
								for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
								{
									if(pPOS_j==pPOS_i)
										c=esP * pPOS_i->varpi0 *pPOS_i->extinction_coefficient * pPhaseFunction_i->Hessian(i_i, i_j);
									else
										c=0.0;
									c-= pPOS_i->varpi0 * pPOS_i->extinction_coefficient * pPhaseFunction_i->Jacobian(i_i) * \
									    pPOS_j->varpi0 * pPOS_j->extinction_coefficient * pPhaseFunction_j->Jacobian(i_j);
									c/=  esP * esP;

									pEstimate->Hessian[index_i][index_j] += c * scalar_contribution;
									index_j++;
								}
							}
							index_i++;
						}
					}
				}
			}
		}
	}
	else
		pCurrentRTEvent->contribution=0.0;
}

void FunctionalLayer::computeScatterEventTransition(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, class Atmosphere::OpticalCell_s *pOC, const class Vector3D& k_in, const class Vector3D& k_out)
{
	/*
		n: backward MC scatter event ordering
		k_out=-k_n
		k_in=-k_{n+1}
	*/
	int i, k, no_phase_function_parameters;
	int no_active_particles, cluster_index;
	class PhaseFunction *pPhaseFunction;
	struct Particle::OpticalStruct_s *pPOS;

	bool is_active_cell=pOC->is_active_cell;
	double c, mu, sigma1, sigma2, esP;

	double *logarithmic_derivative_ptr;

	// update transition Stokes matrix ######################################################################################################

	// calculate effective phase function value ##################################
	pPhaseFunction = pCurrentRTEvent -> pPhaseFunction;
	PhaseFunction :: calculateScatterAngles (k_in, k_out, mu, sigma1, sigma2);

	if(pAtmosphere->simulateRingEffect())
	{
		double esP_at_center_wavelength;

		class StokesMatrix ElasticScatterPhaseMatrix;

		pPhaseFunction->initializeWeightedPhaseMatrix(ScatterPhaseMatrix, mu);
		ScatterPhaseMatrix=StokesMatrix(-sigma2)*(ScatterPhaseMatrix*StokesMatrix(M_PI-sigma1));
		pEstimate->TransitionStokesMatrix*=ScatterPhaseMatrix;
		
		calculateScatterPhaseMatrices(pOC, mu, ScatterPhaseMatrix, ElasticScatterPhaseMatrix, esP, esP_at_center_wavelength);
		// RRS weights
		// pEstimate->RRS_elastic_weight *= (esP_at_center_wavelength*pOC->scatter_coefficient)/(esP*pOC->scatter_coefficient_at_center_wavelength);
		pEstimate->RRS_elastic_weight *= esP_at_center_wavelength / esP;

		c=ElasticScatterPhaseMatrix[0][0];
		ElasticScatterPhaseMatrix /= c;

		ElasticScatterPhaseMatrix=StokesMatrix(-sigma2)*(ElasticScatterPhaseMatrix*StokesMatrix(M_PI-sigma1));
		pEstimate->RRSTransitionStokesMatrix*=ElasticScatterPhaseMatrix;

	}
	else
	{
		pPhaseFunction->initializeWeightedPhaseMatrix(ScatterPhaseMatrix, mu);

		ScatterPhaseMatrix=StokesMatrix(-sigma2)*(ScatterPhaseMatrix*StokesMatrix(M_PI-sigma1));
		pEstimate->TransitionStokesMatrix*=ScatterPhaseMatrix;

		if(is_active_cell)
			calculateDifferentialScatteringCoefficient(pOC, mu, esP);
	}

	// -------------------------------
	if(is_active_cell)
	{
		// calculate cumulative contribution to the logarithmic derivative
		// and to the particle Jacobian in the estimate structure
		no_active_particles=pOC->no_active_particles;
		for(k=0;k<no_active_particles;k++)
		{
			pPOS=pOC->ActiveParticleOpticalStructPtr[k];
			pParticle=pPOS->pParticle;
			no_phase_function_parameters=pParticle->getNOPhaseFunctionParameters();
			cluster_index=pPOS->cluster_index;//pParticle->getClusterIndex(pOC->cell_index);
			
			if(pParticle->hasSinglePhaseFunction())
			{
				// acquire pointer to correct cumulative derivative sub vector
				logarithmic_derivative_ptr=pEstimate->ParticleProfileJacobian[pParticle->getResultStructIndex()].SPF_CumulativeContribution.getCoefficientPtr();

				// extinction coefficient Jacobians
				logarithmic_derivative_ptr[cluster_index] += (pPOS->varpi0 * pPOS->PhaseFunctionValue) / esP;
				logarithmic_derivative_ptr += pParticle->getNOClusters();

				// single scattering albedo
				*logarithmic_derivative_ptr++ +=(pPOS->extinction_coefficient * pPOS->PhaseFunctionValue) / esP;
				
				// phase function parameter derivatives
				// pPOS->pPhaseFunction->evaluateDerivatives(mu, PhaseFunctionParameterJacobian);
				pPhaseFunction=pPOS->pPhaseFunction;
				pPhaseFunction->evaluateJacobian(mu);
				for(i=0;i<no_phase_function_parameters;i++)
					*logarithmic_derivative_ptr++ += (pPhaseFunction->Jacobian(i)*pPOS->scatter_coefficient) / esP;
			}
			else
			{
				// acquire pointer to correct cumulative derivative sub vector
				logarithmic_derivative_ptr=pEstimate->ParticleProfileJacobian[pParticle->getResultStructIndex()].CumulativeContribution.getCoefficientPtr()+\
					pPOS->cluster_index*(2+no_phase_function_parameters);

				// extinction coefficient Jacobians
				*logarithmic_derivative_ptr++ += (pPOS->varpi0 * pPOS->PhaseFunctionValue) / esP;

				// single scattering albedo
				*logarithmic_derivative_ptr++ +=(pPOS->extinction_coefficient * pPOS->PhaseFunctionValue) / esP;
				
				// phase function parameter derivatives
				// pPOS->pPhaseFunction->evaluateDerivatives(mu, PhaseFunctionParameterJacobian);
				pPhaseFunction=pPOS->pPhaseFunction;
				pPhaseFunction->evaluateJacobian(mu);
				for(i=0;i<no_phase_function_parameters;i++)
					*logarithmic_derivative_ptr++ += (pPhaseFunction->Jacobian(i)*pPOS->scatter_coefficient) / esP;
			}
			if(b_calculate_Hessian)
				pPhaseFunction->evaluateHessian(mu);
		}
		// calculated all logarithmic derivatives, now calculate Hessian cumulative contributions!
		if(b_calculate_Hessian)
		{
			int index_i,index_j, i_i,i_j;
			int no_phase_function_parameters_i, no_phase_function_parameters_j;
			int k_POS_i, k_POS_j;
			struct Particle::OpticalStruct_s *pPOS_i, *pPOS_j;
			class Particle *pParticle_i, *pParticle_j;
			class PhaseFunction *pPhaseFunction_i, *pPhaseFunction_j;

			bool has_single_phase_function_i, has_single_phase_function_j;

			for(k_POS_i=0;k_POS_i<pOC->no_active_particles;k_POS_i++)
			{
				pPOS_i=pOC->ActiveParticleOpticalStructPtr[k_POS_i];
				pParticle_i=pPOS_i->pParticle;
				index_i=pParticle_i->getJacobianOffset();

				has_single_phase_function_i=pParticle_i->hasSinglePhaseFunction();

				if(has_single_phase_function_i)
					index_i+=pPOS_i->cluster_index;
				else
					index_i+=pPOS_i->cluster_index*pParticle_i->getClusterDimension();
			
				// extinction coefficient --------------------------------------------------------------------
				for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
				{
					pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
					pParticle_j=pPOS_j->pParticle;
					index_j=pParticle_j->getJacobianOffset();

					has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

					if(has_single_phase_function_j)
						index_j+=pPOS_j->cluster_index;
					else
						index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();

					// extinction coefficient
					c =- ( pPOS_i->varpi0 * pPOS_i->PhaseFunctionValue * pPOS_j->varpi0 * pPOS_j->PhaseFunctionValue ) / ( esP * esP );
					
					pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;
					
					if(has_single_phase_function_j)
						index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
					else // single scattering albedo is next to extinction coefficient inside a cluster
						index_j++;

					// single scattering albedo
					if(pPOS_j==pPOS_i)
						c=esP * pPOS_j->PhaseFunctionValue;
					else
						c=0.0;
					c -=  pPOS_j->varpi0 * pPOS_j->PhaseFunctionValue  *  pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue ;
					c /= esP * esP;
					pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;

					index_j++;
					
					// phase function parameters
					pPhaseFunction_j = pPOS_j->pPhaseFunction;
					no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
					for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
					{
						if(pPOS_j==pPOS_i)
							c=esP * pPOS_i->varpi0;
						else
							c=0.0;
						c -= pPOS_i->varpi0 * pPOS_i->PhaseFunctionValue * pPOS_j->varpi0 * pPOS_j->extinction_coefficient;
						c *= pPhaseFunction_j->Jacobian(i_j) / (esP * esP);
						pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;
						index_j++;
					}
				}
				if(has_single_phase_function_i)
					index_i=pParticle_i->getJacobianOffset()+pParticle_i->getNOClusters();
				else // single scattering albedo is next to extinction coefficient inside a cluster
					index_i++;

				// single scattering albedo -----------------------------------------------------------------
				for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
				{
					pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
					pParticle_j=pPOS_j->pParticle;
					index_j=pParticle_j->getJacobianOffset();

					has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

					if(has_single_phase_function_j)
						index_j+=pPOS_j->cluster_index;
					else
						index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();

					// extinction coefficient
					if(pPOS_j==pPOS_i)
						c=esP * pPOS_j->PhaseFunctionValue;
					else
						c=0.0;
					
					c -= pPOS_j -> varpi0 * pPOS_j -> PhaseFunctionValue * pPOS_i -> extinction_coefficient * pPOS_i -> PhaseFunctionValue;
					c /= esP*esP;

					pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;

					if(has_single_phase_function_j)
						index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
					else // single scattering albedo is next to extinction coefficient inside a cluster
						index_j++;

					// single scattering albedo
					c = - pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue * pPOS_j->extinction_coefficient * pPOS_j->PhaseFunctionValue ;
					c /= esP * esP;

					pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;

					index_j++;
					
					// phase function parameters
					pPhaseFunction_j=pPOS_j->pPhaseFunction;
					no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
					for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
					{
						if(pPOS_j==pPOS_i)
							c=esP * pPOS_i->extinction_coefficient;
						else
							c=0.0;
						c-= pPOS_i->extinction_coefficient * pPOS_i->PhaseFunctionValue * pPOS_j->varpi0 * pPOS_j->extinction_coefficient;
						c*= pPhaseFunction_j->Jacobian(i_j) / (esP * esP);

						pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;
						index_j++;
					}
				}
				
				index_i++;

				// phase function parameters ----------------------------------------------------
				pPhaseFunction_i = pPOS_i->pPhaseFunction;
				no_phase_function_parameters_i=pParticle_i->getNOPhaseFunctionParameters();
				for(i_i=0;i_i<no_phase_function_parameters_i;i_i++)
				{
					for(k_POS_j=0;k_POS_j<pOC->no_active_particles;k_POS_j++)
					{
						pPOS_j=pOC->ActiveParticleOpticalStructPtr[k_POS_j];
						pParticle_j=pPOS_j->pParticle;
						index_j=pParticle_j->getJacobianOffset();

						has_single_phase_function_j=pParticle_j->hasSinglePhaseFunction();

						if(has_single_phase_function_j)
							index_j+=pPOS_j->cluster_index;
						else
							index_j+=pPOS_j->cluster_index*pParticle_j->getClusterDimension();

						// extinction coefficient
						if(pPOS_j==pPOS_i)
							c=esP * pPOS_j->varpi0;
						else
							c=0.0;
						
						c -= pPOS_j -> varpi0 * pPOS_j -> PhaseFunctionValue * pPOS_i -> extinction_coefficient * pPOS_i -> varpi0;
						c *= pPhaseFunction_i->Jacobian(i_i) / ( esP * esP);

						pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;

						if(has_single_phase_function_j)
							index_j=pParticle_j->getJacobianOffset()+pParticle_j->getNOClusters();
						else // single scattering albedo is next to extinction coefficient inside a cluster
							index_j++;

						// single scattering albedo
						if(pPOS_j==pPOS_i)
							c = esP * pPOS_j->extinction_coefficient;
						else
							c = 0.0;
						c -= pPOS_j->extinction_coefficient * pPOS_j->PhaseFunctionValue * pPOS_i->varpi0 * pPOS_i->extinction_coefficient;
						c *= pPhaseFunction_i->Jacobian(i_i) / ( esP * esP );
						pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;

						index_j++;
						
						// phase function parameters
						pPhaseFunction_j=pPOS_j->pPhaseFunction;
						no_phase_function_parameters_j=pParticle_j->getNOPhaseFunctionParameters();
						for(i_j=0;i_j<no_phase_function_parameters_j;i_j++)
						{
							if(pPOS_j==pPOS_i)
								c=esP * pPOS_i->varpi0 *pPOS_i->extinction_coefficient * pPhaseFunction_i->Hessian(i_i, i_j);
							else
								c=0.0;
							c-= pPOS_i->varpi0 * pPOS_i->extinction_coefficient * pPhaseFunction_i->Jacobian(i_i) * \
							    pPOS_j->varpi0 * pPOS_j->extinction_coefficient * pPhaseFunction_j->Jacobian(i_j);
							c/=  esP * esP;

							pEstimate->CumulativeLogarithmic2ndDerivativeHessian[index_i][index_j]+=c;
							index_j++;
						}
					}
					index_i++;
				}
			}
		}
	}
}

void FunctionalLayer::printResult(class Result *pResult, FILE *fptr)
{
	pResult->finish();
	if(pResult->outputEnabled())
		pResult->print(fptr, output_format);
	if(!path_integrals_list.empty())
	{
		list<string>::iterator it;
		for(it=path_integrals_list.begin();it!=path_integrals_list.end();it++)
			pResult->printPathIntegralResults(it->c_str(), constituents_list, fptr);
			
	}
	fputc('\n', fptr);
	double L=pResult->scalar_radiance.getMean();
	if(b_create_sensitivity_grid_file)
	{
		int i;
		double *s=SensitivityBGF.getCellPtr(), N=L*pResult->scalar_radiance.getN();
		for(i=0;i<no_cells;i++)
			s[i]=pResult->CellSensitivity[i]/N;
		if(!SensitivityBGF.writeToFile(fn_sensitivity_grid_file))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not write RT cell sensitivity to file \"%s\"\n", fn_sensitivity_grid_file);
		}
	}
	if(b_create_LEC_histogram)
	{
		int i;
		double sum=0.0, var_sum=0.0;
		double LEC_sum=0.0, LECLEC_sum=0.0, N_sum=0.0;
		
		for(i=0;i<=no_LEC_bins;i++)
		{
			LEC_sum+=LECBin[i][1];
			LECLEC_sum+=LECBin[i][2];
			N_sum+=LECBin[i][3];
		}
		for(i=0;i<=no_LEC_bins;i++)
		{
			/*
				1: LEC
				2: LEC^2
				3: N
			*/
			LECBin[i][4]=LECBin[i][1]/LEC_sum;
			LECBin[i][5]=LECBin[i][2]/LECLEC_sum;
			LECBin[i][6]=LECBin[i][3]/N_sum;
			LECBin[i][7]=sum+=LECBin[i][4];
			LECBin[i][8]=var_sum+=LECBin[i][5];
		}
		const char *file_name="LECHistogram.txt";
		if(!LECBin.save(file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error saving \"%s\"\n", file_name);
		}
	}
}





