/**************************************************************************************************************
*
*       This file,
*
*                            /FunctionalLayer.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _FunctionalLayer_h
#define _FunctionalLayer_h

#include "Math/Vector3D.h"
#include "Math/Matrix3D.h"
#include "Atmosphere.h"

#include "Math/LinearelyInterpolatedFunctionTemplate.h"
#include "Math/StochasticVariable.h"
#include "Math/StokesVector.h"
#include "Math/StokesMatrix.h"
// #include "Math/SphericalHarmonicsExpansion.h"

#include "Math/Vector.h"
#include "Math/Matrix.h"

#include "Depiction/RenderScreen.h"
#include "TrajectoryInformationObject.h"

// #include "CellSensitivity.h"
#include "BinaryGridFile.h"

#include "SimulationObject.h"
#include "FunctionalLayerIO.h"

#include "Estimate.h"
#include "Result.h"

#include "RadiativeTransferEvent.h"

#include <list>
#include <string>

using namespace std;

class Geometry;
class Surface;
class Simulation;

class FunctionalLayer: public FunctionalLayerIO, public SimulationObject
{

protected:

	bool b_calculate_cell_path_length;
	bool RRS;

	/*bool *estimated_scatter_order;
	inline void resetESOFlags(){memset(estimated_scatter_order, 0, sizeof(bool)*maximum_no_trace_events);}*/

	// transition and scatter phase matrix
	class StokesMatrix ScatterPhaseMatrix;
	class StokesVector SVIOne;

	void calculateDifferentialScatteringCoefficient(class Atmosphere::OpticalCell_s *pOC, const double& mu, double& esP);
	void calculateElasticDifferentialScatteringCoefficient(class Atmosphere::OpticalCell_s *pOC, const double& mu, double& esP);
	void calculateScatterPhaseMatrices(class Atmosphere::OpticalCell_s *pOC, const double& mu, class StokesMatrix& ScatterPhaseMatrix, double& esP);
	void calculateScatterPhaseMatrices(class Atmosphere::OpticalCell_s *pOC, const double& mu, class StokesMatrix& ScatterPhaseMatrix, class StokesMatrix& ElasticScatterPhaseMatrix, double& esP, double& esP_at_center_wavelength);

	class Vector forced_cell_path_length;

	// ray tracing variables
	class Vector3D SunDirection, SunPosition;

	double tau_forced, tau_forced_at_center_wavelength, F0_at_center_wavelength;
	double tau_forced_absorption_at_center_wavelength;
	double F0_exp_minus_tau_forced, F0_exp_minus_tau_forced_at_center_wavelength;

	class Vector3D position, direction;
 	struct CellIndex_s CellIndex; 
	double wavelength, central_wavelength;

	class PhaseFunction *pMolecularPhF, *pRotationalRamanPhF;

	class Gas *pGas, **ActiveGasPtr;
	int no_active_gases;
	
	class Particle *pParticle, **ActiveParticlePtr;
	int no_active_particles;

	// sensitivity 
	class BinaryGridFile SensitivityBGF;
	// void registerContributionInSensitivityGrid(class Estimate *pEstimate, const double& c);
	
	// local estimate histogram
	class Matrix LECBin;
	void registerContributionInLECHistogram(const double& LEC);

	// solar irradiance spectrum
	class LinearelyInterpolatedFunctionTemplate<double> SolarIrradianceSpectrum;
	inline double getF0(){return getF0(wavelength);}
	
	
	// functional calculation
	bool resetFunctionals(class Result *pResult);

	// void determineActiveCells();
	bool calculateEscapeOpticalThickness(class Estimate *pEstimate, class Vector3D p, struct CellIndex_s CI);
	void calculateEffectiveScatterPhaseMatrix(struct OpticalCell_s *pOC, class StokesMatrix& ESM, const class Vector3D& k_in, const class Vector3D& k_out);
	void calculateEffectiveTransitionScatterPhaseMatrix(struct OpticalCell_s *pOC, class StokesMatrix& ESM, const class Vector3D& k_in, const class Vector3D& k_out);

	void registerDirectLightContribution(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI);

	void processPropagation(class Estimate *pEstimate, const int& cell_index, const double &length);
	void processGroundScatterEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI, const class Vector3D& surface_normal);
	void processAtmosphericScatterEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, const class Vector3D& p, const struct CellIndex_s& CI, class Atmosphere::OpticalCell_s *pOC, const class Vector3D& k_out);
	void processAtmosphericAbsorptionEvent(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, class Atmosphere::OpticalCell_s *pOC);
	void processSunHit(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, double weight);
	// void computeScatterEventTransition(class Estimate *pEstimate, );
	void computeScatterEventTransition(struct RadiativeTransferEvent_s *pCurrentRTEvent, class Estimate *pEstimate, class Atmosphere::OpticalCell_s *pOC, const class Vector3D& k_in, const class Vector3D& k_out);

	void registerEstimates(class Result *pResult);
	// bool reachedFunctionalAccuracy(class Result *pResult);
	void printResult(class Result *pResult, FILE *fptr=stdout);

public:

	FunctionalLayer();
	~FunctionalLayer();

	inline bool calculateCellPathLength(){return b_calculate_cell_path_length;}

	inline void enableNormalizeWithF0(bool enable=true){b_normalize_with_F0=enable;}
	inline double getF0(const double& lambda){return SolarIrradianceSpectrum.evaluate(1e9*lambda);}
	inline bool loadSolarIrradianceSpectrum(char *file_name){return SolarIrradianceSpectrum.readFromFile(file_name);}
	inline class LinearelyInterpolatedFunctionTemplate<double> *getF0Ptr(){return &SolarIrradianceSpectrum;}


};

#endif // _FunctionalLayer_h


