/**************************************************************************************************************
*
*       This file,
*
*                            /FunctionalLayerIO.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "FunctionalLayerIO.h"

FunctionalLayerIO::FunctionalLayerIO()
{
	IOName="functionals";

	strcpy(constituents, "none");
	strcpy(path_integrals, "none");

	strcpy(output_format, "%.5lg");

	b_event_counts=false;

	b_output_polarized_quantities=false;
	b_correct_for_optical_factors=false;

	b_suppress_Jacobian_output=false;
	b_suppress_Hessian_output=false;

	b_use_precision=false;
	precision=0.005;
	b_precision_major_limit=false;
	precision_check_count=1000;
	b_create_sensitivity_grid_file=false;
	b_enable_variance_reduction=false;

	b_create_LEC_histogram=false;
	no_LEC_bins=100;
	minimum_LEC=0.0;
	maximum_LEC=1e3;
	
	b_normalize_with_F0=false;
	b_sun_disc_sampling=true;
	b_calculate_Hessian=false;
	// she_order=40;

	addParameter(output_format, type_string, "output_format", 0, "output format in printf style for doubles e.g. \"%.3le\" or \"%.4lf\"");
	addParameter(&b_normalize_with_F0, type_bool, "b_normalize_with_F0|b_nF0|b_nwF0", 0, "divide functionals by F0?");
	addParameter(&b_sun_disc_sampling, type_bool, "b_sun_disc_sampling|b_sds", 0, "sample from sun disc in the local estimates instead taking the middle of the disc?");
	addParameter(&b_use_precision, type_bool, "b_use_precision", 0, "use precision of functionals for determining number of trajectories");
	addParameter(&b_output_polarized_quantities, type_bool, "b_output_polarized_quantities|b_polarization|b_pq_output", 0, "output of polarized quantities?");
	addParameter(&precision, type_double, "precision|functional_precision", 0, "relative precision in (0,1] of the functionals");
	addParameter(&b_precision_major_limit, type_bool, "b_precision_major_limit|b_pml", 0, "shall all relative functional precisions be smaller then precision?");
	addParameter(&precision_check_count, type_int, "precision_check_count|pcc", 0, "precision check intervall");
	addParameter(constituents, type_string, "constituents_list|cl", 0, "list of constituents to calculate Jacobians for");
	addParameter(path_integrals, type_string, "path_integrals|pi", 0, "list of considered path integrals");
	addParameter(&b_correct_for_optical_factors, type_bool, "b_correct_for_optical_factors|b_cfof", 0, "e.g. for gases: calculate dI/dn instead of dI/dbeta (beta=absorption coefficient)?");
	addParameter(&b_calculate_Hessian, type_bool, "b_calculate_Hessian|b_Hessian", 0, "calculate Hesse matrix?");
	addParameter(&b_event_counts, type_bool, "b_event_counts|b_ec", 0, "count effective number of events?");
	addParameter(&b_eliminate_absorption, type_bool, "b_eliminate_absorption|b_ea", 0, "remove absorption structures by importance sampling?");
	addParameter(&b_suppress_Hessian_output, type_bool, "b_suppress_Hessian_output|b_no_Hessian_output", 0, "suppress output of Hessian elements?");
	addParameter(&b_suppress_Jacobian_output, type_bool, "b_suppress_Jacobian_output|b_no_Jacobian_output", 0, "suppress output of Jacobian elements?");
	
	// enable variance reduction?
	addParameter(&b_enable_variance_reduction, type_bool, "b_enable_variance_reduction|b_evr", 0, "apply variance reduction techniques?");

	// sensitivity grid file
	addParameter(&b_create_sensitivity_grid_file, type_bool, "b_create_sensitivity_grid_file", 0, "create (binary) sensitivity grid file?");
	addParameter(&fn_sensitivity_grid_file, type_file, "fn_sensitivity_grid_file|fn_sgf", 0, "binary sensitivity grid file");
	
	// LEC histogram
	addParameter(&b_create_LEC_histogram, type_bool, "b_create_LEC_histogram|b_LEC_histogram", 0, "create local estimate contribution (LEC) histogram?");
	addParameter(&no_LEC_bins, type_int, "no_LEC_bins|LEC_bins", 0, "number of local estimate histogram bins");
	addParameter(&minimum_LEC, type_double, "minimum_LEC|LEC_min|LECA", 0, "minimum LEC considered in the LEC histogram");
	addParameter(&maximum_LEC, type_double, "maximum_LEC|LEC_max|LECB", 0, "maximum LEC considered in the LEC histogram");
}

FunctionalLayerIO::~FunctionalLayerIO()
{
}


