/**************************************************************************************************************
*
*       This file,
*
*                            /Gas.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Gas.h"
#include "fileio.h"

#include <stdlib.h>

#include "error.h"


class HITRANDataBase Gas::HITRANDB;

Gas::Gas()
{
	IOName="gas";

	cs_row=0;

	OpticalStruct=0;
	HITRANLineListOfCluster=0;

	addParameter(name, type_string, "name", 0, "name of the gas");
	addParameter(fn_cluster_data, type_file, "fn_cluster_data|fn_data|fn_clusters", 0, "cluster data filename. 1 column, number density [N/cm^3]");
	addParameter(fn_map, type_file, "fn_map", 0, "map filename");
	addParameter(wavelength_range_nm, type_string, "wavelength_range_nm", 0, "lambdaA[nm]~lambdaB[nm]: wavelength range for which lines are extracted");
	addParameter(fn_cross_section, type_file, "fn_cross_section|fn_cs", 0, "1st row: 0 TAB T1[K] TAB ... TAB TN [K], following rows: wl[nm] TAB cs1 [cm^2] TAB ... TAB csN [cm^2]");
	addParameter(&cross_section_format, type_enum, "cross_section_format|cs_format|csf", "wavelength_temperature|cluster_wavelength|HITRAN", "format of cross section file");
	addParameter((Profile *)this, type_profile, "profile|data", 0, "gas profile data, cluster row format: 1=n[1/cm^3], for table processing: profile[cluster,parameter index]");
	
	cross_section_format=wavelength_temperature;
}

Gas::~Gas()
{
	if(OpticalStruct)
		delete [] OpticalStruct;
	if(HITRANLineListOfCluster)
		delete [] HITRANLineListOfCluster;
}

bool Gas::loadCSData()
{
	char file_name[BasicIO_string_size];
	if(cross_section_format==HITRAN)
	{
		int i;
		char buffer[30];
		double lambdaA, lambdaB, p, p_s, T;
		if(InputData.getN()!=3)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "\"%s\": for each cluster temperature and pressure must be specified. File row format: N<TAB>p[hPa]<TAB>T[K], i.e. two columns\n", file_name);
			return false;
		}
		// check whether molecule is in data base
		if(!HITRANDB.isInitialized())
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"HITRAN database not initialized!\n");
			return false;
		}
		pHITRANMolecule=HITRANDB.getMoleculePtr(name);
		if(!pHITRANMolecule|| !pHITRANMolecule->load())
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "molecule \"%s\" not in HITRAN database!\n", name);
			return false;
		}
		// got it, extract absorption lines data
		if(numberOf('~', wavelength_range_nm)!=1)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"wavelength range has wrong format. Correct format: lambdaA[nm]~lambdaB[nm]!\n");
			return false;
		}
		// get wavelength range
		extractWordFromList(wavelength_range_nm, 0, '~', buffer);
		lambdaA=1e-9*atof(buffer);
		extractWordFromList(wavelength_range_nm, 1, '~', buffer);
		lambdaB=1e-9*atof(buffer);
		// extract lines
		if(!pHITRANMolecule->extractRelevantLinesIn(lambdaA, lambdaB))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		// allocate SpectralWindow objects
		if(HITRANLineListOfCluster)
			delete [] HITRANLineListOfCluster;
		HITRANLineListOfCluster=new class SpectralWindow[getNOClusters()];
		if(!HITRANLineListOfCluster)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		for(i=0;i<getNOClusters();i++)
		{
			if(!HITRANLineListOfCluster[i].allocateSpectralRange(pHITRANMolecule))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "%s: error allocating spectral lines for cluster %d\n", name, i+1);
				return false;
			}
			p=InputData[i][1];
			T=InputData[i][2];
			p_s=1e4*InputData[i][0]*const_kB*T;
			HITRANLineListOfCluster[i].calculateLineParameters(p, p_s, T);
		}
	}
	else
	{
		formatFilePath(file_name, fn_cross_section);
		wavelength_index=0;
		switch(cross_section_format)
		{
			case wavelength_temperature:
			{
				class Matrix CSRawData;
				int i, j, no_temperatures;
				double *T, T_min, T_max, temperature;
				cs_row=CSRawData.getRowPtr()+1;
				if(!CSRawData.load(file_name))
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "could not read cross section data from file \"%s\"!\n", file_name);
					return false;
				}
				/*
					expected file format:
					
					0	T1	T2	T3	...
					wl1	cs_11	cs_12	cs_13	...
					wl2	cs_21	cs_22	cs_23	...
					wl3	cs_31	cs_32	cs_33	...
					...

				*/
				no_wavelengths=CSRawData.getM()-1;
				no_temperatures=CSRawData.getN()-1;

				T=CSRawData.getCoefficientPtr()+1;			// temperatures in the first row
				T_min=T[0];
				T_max=T[no_temperatures-1];

				cs_row=CSRawData.getRowPtr()+1;
				lambda_min=cs_row[0][0];
				lambda_max=cs_row[no_wavelengths-1][0];

				if(!CrossSectionData.allocate(getNOClusters()+1, no_wavelengths))
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "error allocating cluster absorption cross sections\n");
					return false;
				}
				// fill-in wavelengths
				/*
					format for internal data:
					wl1	wl2	wl3	...
					cs_11	cs_12	cs_13	...
					cs_21	cs_22	cs_23	...
					cs_31	cs_32	cs_33	...
					...

				*/
				for(i=0;i<no_wavelengths;i++)
					CrossSectionData[0][i]=CSRawData[1+i][0];
				if(no_temperatures==1)
				{
					double cs;
					for(j=0;j<no_wavelengths;j++)
					{
						// set wavelength
						CrossSectionData[0][j]=CSRawData[1+j][0];
						cs=CSRawData[1+j][1];
						for(i=0;i<getNOClusters();i++)
							CrossSectionData[1+i][j]=cs;
					}
				}
				else
				{
					if(InputData.getN()!=2)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "\"%s\": for each cluster a temperature must be specified.\nFile row format: N<TAB>T[K], i.e. two columns\n", fn_cluster_data);
						return false;
					}
					// interpolate temperatures
					int temperature_index=0, t1, t2;
					double *pcs, cs1, cs2, f;
					for(i=0;i<getNOClusters();i++)
					{
						temperature=InputData[i][1];
						if(temperature<T_min || temperature>T_max)
						{
							fprintf(stderr, "gas \"%s\": no cs for temperature %.2lf in cluster %d\n", name, temperature, i+1);
							return false;
						}
						// determine temperature index
						if(T[temperature_index]<temperature)
						{
							while(T[++temperature_index]<temperature);
							temperature_index--;
						}
						else
							while(T[--temperature_index]>temperature);
						t1=1+temperature_index;
						t2=2+temperature_index;
						pcs=CrossSectionData[1+i];
						f=(temperature-T[temperature_index])/(T[temperature_index+1]-T[temperature_index]);
						for(j=0;j<no_wavelengths-1;j++)
						{
							cs1=cs_row[j][t1];
							cs2=cs_row[j][t2];
							*pcs++=cs1+f*(cs2-cs1);
						}
					}
				}
				break;
			}
			case cluster_wavelength:
				/*
					expected file format:
					wl1	wl2	wl3	...
					cs_11	cs_12	cs_13	...
					cs_21	cs_22	cs_23	...
					cs_31	cs_32	cs_33	...
					...

				*/
				if(!CrossSectionData.load(file_name))
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "could not read cross section data from file \"%s\"!\n", file_name);
					return false;
				}
				if(getNOClusters()!=CrossSectionData.getM()-1)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "cross section data does not fit to cluster map!\n");
					return false;
				}
				break;
			default:
				errorIn(__FILE__, __LINE__, __func__);
		}
		no_wavelengths=CrossSectionData.getN();
		lambda_table=CrossSectionData.getCoefficientPtr();
		lambda_min=lambda_table[0];
		lambda_max=lambda_table[no_wavelengths-1];
		cs_row=CrossSectionData.getRowPtr()+1;
	}
	return true;
}

bool Gas::loadFiles()
{
	Profile::setProfileName(name);
	char filename[BasicIO_string_size];
	formatFilePath(filename, fn_cluster_data);
	if(!loadClusterData(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load cluster data of gas \"%s\" from file \"%s\"!\n", name, filename);
		return false;
	}
	formatFilePath(filename, fn_map);
	if(!ClusterMap::load(filename) || !initializeClusterArray())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load map data of gas \"%s\" from file \"%s\"!\n", name, filename);
		return false;
	}
	if(!loadCSData())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "gas \"%s\" failed to load cross section data!\n", name);
		return false;
	}
	if(OpticalStruct)
		delete [] OpticalStruct;
	OpticalStruct=new struct OpticalStruct_s[getNOClusters()];
	if(!OpticalStruct)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating OpticalStructs\n");
		return false;
	}
	initializeOpticalStructs();
	return true;
}

void Gas::initializeOpticalStructs()
{
	for(int i=0;i<getNOClusters();i++)
	{
		OpticalStruct[i].pCluster=cluster+i;
		OpticalStruct[i].cluster_index=i;
		OpticalStruct[i].pGas=this;
		OpticalStruct[i].number_density=cluster[i].input_data[number_density];
	}
}

bool Gas::calculateOpticalStruct(struct OpticalStruct_s *pOS)
{
	struct Cluster_s *pCluster=pOS->pCluster;
	double lambda=pOS->wavelength, cross_section_cm2;
	if(pCluster->valid && lambda==pCluster->optical_data[0])
		return true;
	pOS->number_density=pCluster->input_data[number_density];
	// recalculate cross section
	int cluster_index=pOS->cluster_index;
	switch(cross_section_format)
	{
		case wavelength_temperature:
		case cluster_wavelength:
		{
			lambda*=1e9;	// nm
			if(lambda<lambda_min || lambda>lambda_max)
				cross_section_cm2=0.0;
			else
			{
				// find wavelength intervall
				if(lambda_table[wavelength_index]<=lambda)
				{
					do
					{
						wavelength_index++;
						if(wavelength_index>=no_wavelengths)
						{
							errorIn(__FILE__, __LINE__);
							fprintf(stderr, "no absorption cross section data of gas \"%s\" for wavelength %lg!\n", name, lambda);
							return false;
						}
					}while(lambda_table[wavelength_index]<lambda);
					wavelength_index--;
				}
				else
				{
					do
					{
						wavelength_index--;
						if(wavelength_index<0)
						{
							errorIn(__FILE__, __LINE__);
							fprintf(stderr, "no absorption cross section data of gas \"%s\" for wavelength %lg!\n", name, lambda);
							return false;
						}
					}while(lambda_table[wavelength_index]>lambda);
				}
				// get cluster index	
				// struct Profile::Cluster_s *pCluster=cluster+getClusterIndex(cell_index);
				double l_a, l_b, cs_a, cs_b, f;
				l_a=lambda_table[wavelength_index];
				l_b=lambda_table[wavelength_index+1];
				f=(lambda-l_a)/(l_b-l_a);
				cs_a=cs_row[cluster_index][wavelength_index];
				cs_b=cs_row[cluster_index][wavelength_index+1];
				cross_section_cm2=cs_a+f*(cs_b-cs_a);
			}
			break;
		}
		case HITRAN:	
			cross_section_cm2=HITRANLineListOfCluster[cluster_index].calculateCrossSection(lambda);
			break;
		default:
			errorIn(__FILE__,__LINE__,__func__);
			return false;
	}
	pOS->cross_section_cm2=cross_section_cm2;
	pCluster->optical_data[0]=lambda;
	pCluster->optical_data[1]=cross_section_cm2;
	pOS->absorption_coefficient=1e5*pOS->number_density*cross_section_cm2;
	pCluster->valid=true;
	return true;	
}






