/**************************************************************************************************************
*
*       This file,
*
*                            /Gas.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Gas_h
#define _Gas_h

#include "BasicIO.h"
#include "Profile.h"
#include "AirRelated.h"

#include "HITRAN/HITRANDataBase.h"

#include "HITRAN/SpectralWindow.h"

class Gas;

class Gas: public BasicIO, public Profile, public AirRelated
{
	public:
		enum CrossSectionFormat_e
		{
			wavelength_temperature,
			cluster_wavelength,
			HITRAN
		};

	private:
	// IO ##################################################
		char name[30];							// name of tracegas

		enum CrossSectionFormat_e cross_section_format;
		char fn_cluster_data[BasicIO_string_size];			// number density / vmr cluster data
		char fn_map[BasicIO_string_size];				// map
		
		char fn_cross_section[BasicIO_string_size];			// filename of single cross section file
		char wavelength_range_nm[BasicIO_string_size];			// e.g. 1300~1400 
	// ##########################################

	public:
		struct OpticalStruct_s
		{
			class Gas *pGas;					// associated gas
			struct Cluster_s *pCluster;				// associated cluster
			int cluster_index;

			double wavelength;

			double number_density;					// [N/cm^3]

			double cross_section_cm2;				// [cm^3]
			double cross_section_cm2_at_center_wavelength;		// [cm^3]

			double absorption_coefficient;				// [1/km]
			double absorption_coefficient_at_center_wavelength;	// [1/km]
		};

	protected:

		enum L1Element_e
		{
			wavelength,
			epsilon
		};

		enum L0Element_e
		{
			number_density
		};
		
		int no_wavelengths;

		/*int no_temperatures;
		double *T, T_min, T_max;	// temperature range*/
		// double *cs_data; 		// cross section data matrix

		class Matrix CrossSectionData;

		double **cs_row;		// pointers to rows of the cs matrix
		double lambda_min, lambda_max, *lambda_table;
		int wavelength_index;

		bool loadCSData();
		
		struct OpticalStruct_s *OpticalStruct;

		// HITRAN stuff
		static class HITRANDataBase HITRANDB;
		class HITRANMolecule *pHITRANMolecule;
		
		class SpectralWindow *HITRANLineListOfCluster;
		
		
	public:
		
		Gas();
		~Gas();

		inline struct OpticalStruct_s *getOpticalStructPtr(const int& i=0){return OpticalStruct+i;}
		inline double getConcentration(const int& cluster_index){return cluster[cluster_index].input_data[0];}

		inline char *getName(){return name;}

		void initializeOpticalStructs();

		// bool calculateCrossSection(double lambda, int cell_index, double *pcs);
		bool calculateOpticalStruct(struct OpticalStruct_s *pOS);
		bool loadFiles();

		inline static bool setupHITRAN(char *config_filename){return HITRANDB.open(config_filename);}
};

#endif //_Gas_h




