/**************************************************************************************************************
*
*       This file,
*
*                            /GasMixtureDCS.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "GasMixtureDCS.h"
#include "trigonometry.h"

#include "error.h"

// static element initialization
class Molecule GasMixtureDCS::N2, GasMixtureDCS::O2, GasMixtureDCS::Ar;
class Molecule *GasMixtureDCS::MPtr[3];

int GasMixtureDCS::no_molecule_types;

bool GasMixtureDCS::initializeMolecules(const int& J_max)
{
	MPtr[Molecule::N2]=&N2;
	MPtr[Molecule::O2]=&O2;
	MPtr[Molecule::Ar]=&Ar;
	no_molecule_types=sizeof(MPtr)/sizeof(MPtr[0]);
	return N2.initialize(Molecule::N2, 0.7809, J_max) \
		&& O2.initialize(Molecule::O2, 0.2095, J_max) \
		&& Ar.initialize(Molecule::Ar, 0.0093, J_max);
}

void GasMixtureDCS::calculateOpticalPropertiesElastic(const double& lambda)
{
	// add all cross sections
	// calculate phase matrix parameters
	int i;
	class Molecule *pM;
	double sigma, epsilon, A, B, C, N, zp3;

	Molecule::setWavelength(lambda);

	RayleighPhF._sigma()=0.0;
	RayleighPhF._ACN()=0.0;
	RayleighPhF._AN()=0.0;
	RayleighPhF._BN()=0.0;
	for(i=0;i<no_molecule_types;i++)
	{
		pM=MPtr[i];
		pM->calculateOpticalProperties();
		// get epsilon

		epsilon=pM->_epsilon();

		// Rayleigh scattering
		sigma=pM->_vmr() * pM->_sigma_Rayleigh();
		A=3.0*(45.0+epsilon);
		B=30.0*(9.0-epsilon);
		C=36.0*epsilon;
		N=180.0+40.0*epsilon;
		
		// add things
		RayleighPhF._sigma()+=sigma;
		RayleighPhF._ACN() += (sigma*(A+C))/N;
		RayleighPhF._AN() += (sigma*A)/N;
		RayleighPhF._BN() += (sigma*B)/N;
	}
	f_Cabannes=1.0;
	
	// sampling constant
	zp3 = RayleighPhF._ACN()/ RayleighPhF._AN();
	zp3 *= zp3 * zp3;
	RayleighPhF._zp3()=zp3;

	// normalization
	sigma=RayleighPhF._sigma();
	RayleighPhF._ACN() /= sigma;
	RayleighPhF._AN() /= sigma;
	RayleighPhF._BN() /= sigma;
}

void GasMixtureDCS::calculateOpticalProperties(const double& lambda)
{
	// add all cross sections
	// calculate phase matrix parameters
	int i;
	class Molecule *pM;
	double sigma, epsilon, A, B, C, N, vmr, zp3;

	Molecule::setWavelength(lambda);

	CabannesPhF._sigma()=RayleighPhF._sigma()=0.0;
	CabannesPhF._ACN()=RayleighPhF._ACN()=0.0;
	CabannesPhF._AN()=RayleighPhF._AN()=0.0;
	CabannesPhF._BN()=RayleighPhF._BN()=0.0;
	for(i=0;i<no_molecule_types;i++)
	{
		pM=MPtr[i];
		pM->calculateOpticalProperties();
		
		// get epsilon, volume mixing ratio and calculate C
		epsilon=pM->_epsilon();
		vmr=pM->_vmr();

		C=36.0*epsilon;

		// Rayleigh ----------------------------------------
		sigma = vmr * pM->_sigma_Rayleigh();
		A=3.0*(45.0+epsilon);
		B=30.0*(9.0-epsilon);
		N=180.0+40.0*epsilon;
		
		RayleighPhF._sigma() += sigma;
		RayleighPhF._ACN() += (sigma*(A+C))/N;
		RayleighPhF._AN() += (sigma*A)/N;
		RayleighPhF._BN() += (sigma*B)/N;

		// Cabannes -------------------------------
		sigma = vmr * pM->_sigma_Cabannes();
		A=3.0*(180.0+epsilon);
		B=30.0*(36.0-epsilon);
		N=40.0*(18.0+epsilon);

		CabannesPhF._sigma() += sigma;
		CabannesPhF._ACN() += (sigma*(A+C))/N;
		CabannesPhF._AN() += (sigma*A)/N;
		CabannesPhF._BN() += (sigma*B)/N;
	}
	f_Cabannes=CabannesPhF._sigma()/RayleighPhF._sigma();

	// sampling constant
	zp3 = RayleighPhF._ACN()/ RayleighPhF._AN();
	zp3 *= zp3 * zp3;
	RayleighPhF._zp3()=zp3;

	// normalization
	sigma=RayleighPhF._sigma();
	RayleighPhF._ACN() /= sigma;
	RayleighPhF._AN() /= sigma;
	RayleighPhF._BN() /= sigma;

	// sampling constant
	zp3 = CabannesPhF._ACN()/ CabannesPhF._AN();
	zp3 *= zp3 * zp3;
	CabannesPhF._zp3()=zp3;

	// normalization
	sigma=CabannesPhF._sigma();
	CabannesPhF._ACN() /= sigma;
	CabannesPhF._AN() /= sigma;
	CabannesPhF._BN() /= sigma;
}

void GasMixtureDCS::logMoleculeParameters(FILE *fptr)
{
	int i;
	class Molecule *pM;
	for(i=0;i<no_molecule_types;i++)
	{
		pM=MPtr[i];
		pM->logParameters(fptr);
		fputc('\n', fptr);
	}
}

