/**************************************************************************************************************
*
*       This file,
*
*                            /Geometry.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Geometry.h"
#include "Math/Vector3D.h"
#include "Math/Matrix3D.h"

#include "Math/prand.h"
#include "Math/minmax.h"

#include "trigonometry.h"
#include "constants.h"
#include "error.h"

#include "SurfaceGrid.h"

#include "PhaseFunctions/PhaseFunction.h"

double Geometry::SunDistance=1.49592e+008;	// Sun radius
double Geometry::SunDiameter=1.392e+006;	// Sun diameter	

Geometry::Geometry()
{
	IOName="geometry";
	FOVType=round;

	use_fixed_initial_direction=false;

	InstrumentSensitivityFunction=unweighted;
	DetectorViewMode=normal_view_mode;

	SunPositionDescription=detector_relative;	// position description of the sun
	SunCoLatitude=0.0;
	SunLongitude=0.0;

	SZA=0.0;
	SRAA=0.0;

	Altitude=0.001;
	Longitude=0.001;
	CoLatitude=0.001;

	ReferenceAltitude=0.001;
	ReferenceLongitude=0.001;
	ReferenceCoLatitude=0.001;

	X=0.0;
	Y=0.0;
	Z=0.0;

	ReferenceX=0.0;
	ReferenceY=0.0;
	ReferenceZ=0.0;		// reference point geo-coordinates

	NorthAzimuth=0.0;				// azymuth angle towards north pole meridian
	Elevation=0.0;					// elevation angle beta
	Roll=0.0;

	Omega=0.0;						// sight opening angle
	OmegaHorizontal=0.0;				// horizontal aperture size in degree
	OmegaVertical=0.0;				// vertical aperture size in degree

	ISGaussSigma=0.0;				// round FOV instrument sensitivity sigma
	ISGaussSigmaVertical=0.0;		// rectangular FOV instrument sensitivity vertical sigma
	ISGaussSigmaHorizontal=0.0;		// rectangular FOV instrument sensitivity horizontal sigma
	
	b_invert_faces=false;

	TangentHeight=-1.0;
	TangentHeightPrecision=1e-4;
	b_tangent_height_with_refraction=false;

	b_forward_geometry=false;

	addParameter(&FOVType, type_enum, "FOVType|FOV", "round|elliptic|square|rectangular|box|2pi|4pi|table", "geometrical shape of the field of view");
	addParameter(&DetectorViewMode, type_enum, "view_mode|DetectorViewMode", "normal_view_mode|limb|look_at", "detector view mode");
	addParameter(&InstrumentSensitivityFunction, type_enum, "ISF|InstrumentSensitivityFunction", "unweighted|cosine|gaussian", "FOV sensitivity function");
	addParameter(&DetectorPositionDescription, type_enum, "position_description|DetectorPositionDescription", "spherical|cartesic", "description of the detector position");
	addParameter(&Altitude, type_double, "Altitude|Height|altitude|height|a|h", 0, "[km] geo altitude detector coordinate");
	addParameter(&CoLatitude, type_angle, "CoLatitude|colatitude|colat", 0, "[degree] geo co latitudinal detector coordinate");
	addParameter(&Longitude, type_angle, "Longitude|longitude|long", 0, "[degree] geo longitudinal detector coordinate");
	addParameter(&ReferenceAltitude, type_double, "ReferenceAltitude|ReferenceHeight|Referencealtitude|a_ref|h_ref", 0, "[km] geo altitude detector reference coordinate");
	addParameter(&ReferenceCoLatitude, type_angle, "ReferenceCoLatitude|reference_colatitude|colat_ref", 0, "[degree] geo co latitudinal detector reference coordinate");
	addParameter(&ReferenceLongitude, type_angle, "ReferenceLongitude|reference_longitude|long_ref", 0, "[degree] geo longitudinal detector reference coordinate");
	addParameter(&X, type_double, "X|x", 0, "[km] cartesic detector x coordinate");
	addParameter(&Y, type_double, "Y|y", 0, "[km] cartesic detector y coordinate");
	addParameter(&Z, type_double, "Z|z", 0, "[km] cartesic detector z coordinate");
	addParameter(&ReferenceX, type_double, "ReferenceX|x_ref", 0, "[km] cartesic detector x reference coordinate");
	addParameter(&ReferenceY, type_double, "ReferenceY|y_ref", 0, "[km] cartesic detector y reference coordinate");
	addParameter(&ReferenceZ, type_double, "ReferenceZ|z_ref", 0, "[km] cartesic detector z reference coordinate");
	addParameter(&NorthAzimuth, type_angle, "NorthAzimuth", 0, "[degree] azimuth angle towards north pole meridian");
	addParameter(&Elevation, type_angle, "Elevation|elevation", 0, "[degree] elevation angle");
	addParameter(&Roll, type_angle, "Roll|roll", 0, "[degree] FOV Roll angle");
	addParameter(&Omega, type_angle, "Omega|omega|FOV", 0, "[degree] FOV opening angle");
	addParameter(&OmegaHorizontal, type_angle, "OmegaHorizontal|FOVH", 0, "[degree] horizontal FOV size");
	addParameter(&OmegaVertical, type_angle, "OmegaVertical|FOVV", 0, "[degree] vertical FOV size");
	addParameter(&ISGaussSigma, type_angle, "ISGaussSigma", 0, "[degree] round FOV instrument sensitivity sigma");
	addParameter(&ISGaussSigmaVertical, type_angle, "ISGaussSigmaVertical", 0, "[degree] rectangular FOV instrument sensitivity vertical sigma");
	addParameter(&ISGaussSigmaHorizontal, type_angle, "ISGaussSigmaHorizontal", 0, "[degree] rectangular FOV instrument sensitivity horizontal sigma");
	addParameter(box_detector_definition, type_string, "box_detector_definition|box_detector", 0, "definition of a box detector");
	addParameter(&b_invert_faces, type_bool, "b_invert_faces", 0, "indicates whether to invert box faces and emmit in the opposite direction");
	addParameter(&TangentHeight, type_double, "TangentHeight", 0, "[km] for limb calculations");
	addParameter(&TangentHeightPrecision, type_double, "TangentHeightPrecision", 0, "[km] precision of the tangent height point calculation");
	addParameter(&b_tangent_height_with_refraction, type_bool, "b_tangent_height_with_refraction", 0, "calculate tangent height point considering refraction");
	addParameter(&SunPositionDescription, type_enum, "SunPositionDescription|SunPosition", "absolute|detector_relative", "position description of the sun");
	addParameter(&SunDistance, type_double, "SunDistance|distance_to_sun|AU", 0, "[km] distance of planet to the sun");
	addParameter(&SunDiameter, type_double, "SunDiameter|diameter_sun", 0, "[km] sun's diameter");
	addParameter(&SunLongitude, type_angle, "SunLongitude|sun_longitude|long_sun", 0, "[degree] sun longitude");
	addParameter(&SunCoLatitude, type_angle, "SunCoLatitude|sun_colatitude|colat_sun", 0, "[degree] sun colatitude");
	addParameter(&SZA, type_angle, "SZA|sza|solar_zenith_angle", 0, "[degree] solar zenith angle (detector depending)");
	addParameter(&SRAA, type_angle, "SRAA|sraa|sun_relative_azimuth_angle", 0, "[degree] sun relative azimuth angle (detector depending), math. sun - detector");
	addParameter(&b_forward_geometry, type_bool, "b_forward_geometry|b_forward_RT|b_forward_MC", 0, "forward Monte Carlo radiation transport simulation");
	addParameter(fn_FOV_function, type_file, "fn_FOV_function|fn_sf", 0, "FOV function matrix file");
}

bool Geometry::initializeGeometry()
{
	switch(DetectorViewMode)
	{
		case limb:
		{
			if(!pGrid->isSphericalGrid())
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"limb mode does not make sense in cartesic coordinates!\n");
				return false;
			}
			if(TangentHeight>0.0)
			{
				if(TangentHeight>Altitude)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr,"Tangent height (%lfkm) is higher then detector height (%lfkm)!\n", TangentHeight, Altitude);
					return false;
				}
				Elevation=-acos((pGrid->getPlanetRadius()+TangentHeight)/(pGrid->getPlanetRadius()+Altitude));
			}
			break;
		}
		case look_at:
		{
			class Vector3D ReferencePoint, Position, Direction;
			if(pGrid->isSphericalGrid())
			{
				ReferencePoint=Polar(pGrid->getPlanetRadius()+ReferenceAltitude, ReferenceCoLatitude, ReferenceLongitude);
				Position=Polar(pGrid->getPlanetRadius()+Altitude, CoLatitude, Longitude);
				Direction=norm(ReferencePoint-Position);
				Direction*=Matrix3D(Matrix3D::ROTZ, -Longitude);
				Direction*=Matrix3D(Matrix3D::ROTY,-CoLatitude);
				NorthAzimuth=atan2(Direction[0], Direction[1]); // azimuth
				Elevation=atan2(Direction[2], sqrt(Direction[0]*Direction[0]+Direction[1]*Direction[1])); // elevation
			}
			else
			{
				ReferencePoint=Vector3D(ReferenceX, ReferenceY, ReferenceZ);
				Position=Vector3D(X, Y, Z);
				Direction=norm(ReferencePoint-Position);
				NorthAzimuth=M_PI-atan2(Direction[1],Direction[0]); // north azimuth
				Elevation=atan2(Direction[2], sqrt(Direction[0]*Direction[0]+Direction[1]*Direction[1])); // elevation
			}
			break;
		}
		default:
		case normal_view_mode:
			if(!pGrid->acceptGeometry(this))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			break;
	}
	// detector ########################################################
	if(pGrid->isSphericalGrid())
	{
		// position
		DetectorPosition.fromPolar(pGrid->getPlanetRadius()+Altitude, CoLatitude, Longitude);
		// PhaseFunction::initializeSphericalLocalCSMatrix(DetectorPosition);
		// central field of view direction
		DetectorCentralFOVDirection=Vector3D(0.0, 1.0, 0.0);
		rotateIntoDetectorFOVSpherical(DetectorCentralFOVDirection);
	}
	else
	{
		// position
		DetectorPosition=Vector3D(X,Y,Z);
		// PhaseFunction::defineReferenceZAxis(Vector3D(0.0, 0.0, 1.0));
		PhaseFunction::resetLocalCSMatrix();
		// central field of view direction
		DetectorCentralFOVDirection=Vector3D(1.0, 0.0, 0.0);	// south
		rotateIntoDetectorFOVCartesic(DetectorCentralFOVDirection);
	}
	if(!pGrid->contains(DetectorPosition))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"detector is outside the lattice:\n");
		write(stderr);
		return false;
	}
	// sun ##############################################################
	switch(SunPositionDescription)
	{
		case absolute:
			SunPosition=Polar(SunDistance, SunCoLatitude, SunLongitude);
			break;
		default:
		case detector_relative:
		{
			if(pGrid->isSphericalGrid())
				SunPositionFromSZAandRAZY(&SunPosition, DetectorPosition, NorthAzimuth, SZA, SRAA);
			else
				SunPosition.fromPolar(AU, SZA, M_PI+NorthAzimuth+SRAA);
			break;
		}
	}
	// detector cell index #######################################
	resetCellIndex(&DetectorCellIndex);
	if(!pGrid->calculateCellIndex(DetectorPosition, DetectorCentralFOVDirection, &DetectorCellIndex))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error calculating detector cell index!\n");
		return false;
	}
	switch(FOVType)
	{
		case box:
			if(b_forward_geometry)
			{
				fprintf(stderr, "not implemented yet!\n");
				return false;
			}
			if(!boxDetector.assignDefinition(box_detector_definition))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"invalid box detector definition: \"%s\"!\n", box_detector_definition);
				return false;
			}
			boxDetector.enableCartesicGrid(!pGrid->isSphericalGrid());
			break;

		case table:
			if(!FOV_function.load(fn_FOV_function))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error loading FOV function from file \"%s\"!\n", fn_FOV_function);
				return false;
			}
			break;

		case round:
			if(Omega!=0.0)
			{
				double mu=cos(Omega);
				switch(InstrumentSensitivityFunction)
				{
					case unweighted:
						FOV_mu_norm=1.0-mu;
						break;

					case cosine:
						FOV_mu_norm=0.5*(1.0-mu*mu);
						break;

					default:
						FOV_mu_norm=0.0;
						break;
				}
			}
			break;
		default:;
	}
	SunDirection=norm(SunPosition-DetectorPosition);
	return true;
}

void Geometry::rotateIntoDetectorFOVSpherical(class Vector3D& v)
{
	// a neutral v needs to point along the y-axis!
	// direction points against earth rotation now if you put it
	// to greenwhich line/aequator
	v*=Matrix3D(Matrix3D::ROTY, Roll);
	v*=Matrix3D(Matrix3D::ROTX, Elevation);
	v*=Matrix3D(Matrix3D::ROTZ, NorthAzimuth+0.5*M_PI);
	// a neutral (0,1,0) vector and NorthAzimuth=0 now produces a vector looking into negative
	// x direction	

	// if you put the Geometry to greenwhich line/aequator it 
	// would already have the correct direction
	v*=Matrix3D(Matrix3D::ROTY, CoLatitude);
	v*=Matrix3D(Matrix3D::ROTZ, Longitude);
}

void Geometry::rotateIntoDetectorFOVCartesic(class Vector3D& v)
{
	// a neutral v needs to point along the x-axis!
	v*=Matrix3D(Matrix3D::ROTX, Roll);
	v*=Matrix3D(Matrix3D::ROTY, -Elevation);
	v*=Matrix3D(Matrix3D::ROTZ, M_PI+NorthAzimuth);
}

bool Geometry::sampleInitialState(class Vector3D *pPosition, struct CellIndex_s *pCellIndex, class Vector3D *pDirection)
{
	if(b_forward_geometry)
		pGrid->sampleTOAPoint(pPosition, pDirection, SunPosition);
	else
	{
		if(FOVType!=box)
		{
			*pPosition=DetectorPosition;
			if(FOVType==four_pi)
			{
				double phi=2.0*M_PI*prand();
				double mu=correctMu(1.0-2.0*prand());
				double sin_theta=sqrt(1.0-mu*mu);
				*pDirection=Vector3D(cos(phi)*sin_theta, sin(phi)*sin_theta, mu);
			}
			else if(FOVType==two_pi)
				SurfaceGrid::sampleLambertianScatterDirection(DetectorCentralFOVDirection, *pDirection);	
			else
			{
				if(use_fixed_initial_direction)
					*pDirection=FixedInitialDirection;
				else
				{
					class Vector3D k;
					sampleFOVDirection(k);
					if(pGrid->isSphericalGrid())
					{
						// neutral direction is y direction
						// neutral direction to right is x axis
						// neutral direction upwards is z axis
						(*pDirection)[0]=k[1];
						(*pDirection)[1]=k[2];
						(*pDirection)[2]=k[0];
						rotateIntoDetectorFOVSpherical(*pDirection);
					}
					else
					{
						// neutral direction is x direction
						// FOVDirection=norm(Vector3D(1.0, x, y));
						(*pDirection)[0]=k[2];
						(*pDirection)[1]=k[0];
						(*pDirection)[2]=k[1];
						rotateIntoDetectorFOVCartesic(*pDirection);
					}
				}
			}
			*pCellIndex=DetectorCellIndex;
			return true;
		}
		else if(!boxDetector.takeEmissionSample(pPosition, pDirection))
			return false;
	}
	return pGrid->calculateCellIndex(*pPosition, *pDirection, pCellIndex);
}

/*class Vector3D Geometry::produceFOVVector(const double& kx, const double& ky, const double& kz)
{
	class Vector3D FOVDirection;
	if(pGrid->isSphericalGrid())
	{
		// neutral direction is y direction
		FOVDirection=norm(Vector3D(x, 1.0, y));
		rotateIntoDetectorFOVSpherical(FOVDirection);
	}
	else
	{
		FOVDirection=norm(Vector3D(1.0, -x, y));
		rotateIntoDetectorFOVCartesic(FOVDirection);
	}
	return FOVDirection;
}*/

void Geometry::sampleFOVDirection(class Vector3D& k_FOV)
{
	double x,y, r,phi, mu, sin_theta;
	switch(FOVType)
	{
		default:
		case round:
			if(Omega==0.0)
			{
				k_FOV=Vector3D(0.0, 0.0, 1.0);
				return;
			}
			phi=2.0*M_PI*prand();
			switch(InstrumentSensitivityFunction)
			{
				default:
				case unweighted:
					mu=correctMu(1.0-FOV_mu_norm*prand());
					break;
				case cosine:
					mu=correctMu(sqrt(1.0-2.0*FOV_mu_norm*prand()));
					break;
				case gaussian:
					/*do
					{
						x=gaussSampleBoxMullerMethod(ISGaussSigma,0.0);
						y=gaussSampleBoxMullerMethod(ISGaussSigma,0.0);
						r=pow(x/Omega,2.0)+pow(y/Omega,2.0);
					}while(r>0.25);
					mu=cos(r);*/
					mu=cos(gaussSampleBoxMullerMethod(ISGaussSigma,0.0));
					break;
			}
			sin_theta=sqrt(1.0-mu*mu);
			k_FOV[0]=cos(phi)*sin_theta;
			k_FOV[1]=sin(phi)*sin_theta;
			k_FOV[2]=mu;
			break;
		case elliptic:
			switch(InstrumentSensitivityFunction)
			{
				default:
				case unweighted:
					do
					{
						x=OmegaHorizontal*(-0.5+prand());
						y=OmegaVertical*(-0.5+prand());
						r=pow(x/OmegaHorizontal,2.0)+pow(y/OmegaVertical,2.0);
					}while(r>0.25);
					break;
				case cosine:
					do
					{
						x=OmegaHorizontal*(-0.5+sqrt(prand()));
						y=OmegaVertical*(-0.5+sqrt(prand()));
						r=pow(x/OmegaHorizontal,2.0)+pow(y/OmegaVertical,2.0);
					}while(r>0.25);
					break;
				case gaussian:
					do
					{
						x=gaussSampleBoxMullerMethod(ISGaussSigmaHorizontal,0.0);
						y=gaussSampleBoxMullerMethod(ISGaussSigmaVertical,0.0);
						r=pow(x/OmegaHorizontal,2.0)+pow(y/OmegaVertical,2.0);
					}while(r>0.25);
					break;
			}
			k_FOV=norm(Vector3D(tan(y), tan(x), 1.0));
			break;
		case square:
			if(Omega==0.0)
			{
				k_FOV=Vector3D(0.0, 0.0, 1.0);
				return;
			}
			switch(InstrumentSensitivityFunction)
			{
				default:
				case unweighted:
					x=Omega*(-0.5+prand());
					y=Omega*(-0.5+prand());
					break;
				case cosine:
					x=Omega*(-0.5+sqrt(prand()));
					y=Omega*(-0.5+sqrt(prand()));
					break;
				case gaussian:
					x=gaussSampleBoxMullerMethod(ISGaussSigma,0.0);
					y=gaussSampleBoxMullerMethod(ISGaussSigma,0.0);
					break;
			}
			k_FOV=norm(Vector3D(tan(y), tan(x), 1.0));
			break;
		case rectangular:
			switch(InstrumentSensitivityFunction)
			{
				default:
				case unweighted:
					x=OmegaHorizontal*(-0.5+prand());
					y=OmegaVertical*(-0.5+prand());
					break;
				case cosine:
					x=OmegaHorizontal*(-0.5+sqrt(prand()));
					y=OmegaVertical*(-0.5+sqrt(prand()));
					break;
				case gaussian:
					x=gaussSampleBoxMullerMethod(ISGaussSigmaHorizontal,0.0);
					y=gaussSampleBoxMullerMethod(ISGaussSigmaVertical,0.0);
					break;
			}
			k_FOV=norm(Vector3D(tan(y), tan(x), 1.0));
			break;
		case table:
			FOV_function.getSample(&x, &y);
			k_FOV=norm(Vector3D(tan(y), tan(x), 1.0));
			break;
	}
}

void Geometry::SunPositionFromSZAandRAZY(Vector3D *pSunPosition, Vector3D& DetectorPosition, double azymuth, double SZA, double RAZY)
{
	Vector3D d, p=DetectorPosition;
	double r, theta, phi;
	DetectorPosition.toPolar(&r, &theta, &phi);
	// if SZA==0.0 and azymuth+RAZY==0.0, d points into negative x direction
	// see function rotateIntoDetectorFOVSpherical
	d=Polar(1.0, SZA, M_PI+azymuth+RAZY);
	// now rotate Sun direction to the detector position on Earth
	d*=Matrix3D(Matrix3D::ROTY,theta);
	d*=Matrix3D(Matrix3D::ROTZ,phi);
	//now calculate Sun position by stretching with AU. (is not entirely exact....)
	double a,b,x1,x2,x;
	a=p*d;
	//b=p*p-2.237776646e16;
	b=p*p-SunDistance*SunDistance;
	x1=-a+sqrt(a*a-b);
	x2=-a-sqrt(a*a-b);
	x=x1>0.0?x1:x2;
	*pSunPosition=p+x*d;
}

bool Geometry::isTwilightGeometry(double *pTwilightDistance)
{
	// Kegelgleichung aufstellen :		(x-s)/(|x-s|)*d=cos(theta/2)
	// s :		Kegelspitze
	// d :		Richtung des Kegels (Achse Sonne-Erde)
	class Vector3D ConeDirection=norm(-SunPosition);
	// find theta...
	double EarthRadius=pGrid->getPlanetRadius();
	double square=1.0-pow(EarthRadius/SunDistance,2.0);
	double R_xy=EarthRadius*sqrt(square);
	double half_theta_cone=atan(R_xy/SunDistance);
	double DayNightBorderSunDistance=SunDistance*square;
	// simple test for twilight situation: is the detector position inside or outside the cone
	class Vector3D DetectorSunVector=DetectorPosition-SunPosition;
	double theta=acos(norm(DetectorSunVector)*ConeDirection);
	double D=fabs(DetectorSunVector*ConeDirection);
	if(theta<half_theta_cone && D>DayNightBorderSunDistance)
	{
	/*
		calculate distance to night&day shadow border....
		a=DetectorSunVector
		d=ConeDirection
		m=DetectorDirection
	*/
		class Vector3D DetectorDirection=getDetectorCentralFOVDirection();
		double cosine_half_theta_square=pow(cos(half_theta_cone),2.0);
		double ad=DetectorSunVector*ConeDirection;
		double aa=DetectorSunVector*DetectorSunVector;
		// gamma0=(a*d)^2-cos^2(theta/2)*a^2
		double gamma_0=ad*ad-aa*cosine_half_theta_square;
		// gamma1=(a*d)*(m*d)-cos^2(theta/2)*(a*m)
		double md=DetectorDirection*ConeDirection;
		double am=DetectorSunVector*DetectorDirection;
		double gamma_1=ad*md-cosine_half_theta_square*am;
		// gamma_2=(m*d)^2-cos^2(theta/2)
		double gamma_2=md*md-cosine_half_theta_square;
		double radikand=gamma_1*gamma_1-gamma_0*gamma_2;
		if(gamma_2==0.0 || radikand<0.0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "numerical problem with distance to night & day shadow border calculation!\n");
			return false;
		}
		double mu1=(-gamma_1+sqrt(radikand))/gamma_2;
		double mu2=(-gamma_1-sqrt(radikand))/gamma_2;
		*pTwilightDistance=max(mu1,mu2);
		return true;
	}
	*pTwilightDistance=0.0;
	return false;
}

bool Geometry::isPointBelowTerminator(class Vector3D& p)
{
	class Vector3D k=norm(SunPosition-p);
	double lambda=k*p;						// attention: minus sign went to here v
	class Vector3D TangentPoint=p;
	TangentPoint-=lambda*k;
	return TangentPoint.length()>pGrid->getPlanetRadius();
}

class Vector3D Geometry::sampleSunDiscDirection(class Vector3D& p)
{
	// construct direction sample around the z-axis
	class Vector3D d, SunVector;
	double x,y,r, phi, rs,ts,ps;

	r=SUN_SOLID_ANGLE_RADIUS*sqrt(prand());
	phi=2.0*M_PI*prand();
	
	x=r*cos(phi);
	y=r*sin(phi);

	d=Vector3D(x, y, 1.0);
	
	// rotate direction sample into the Sun's direction
	SunVector=SunPosition-p;		// Sun disc center direction
	SunVector.toPolar(&rs, &ts, &ps);	// decompose sun vector
	
	// rotate direction sample into the Sun's direction
	d*=Matrix3D(Matrix3D::ROTY, ts);
	d*=Matrix3D(Matrix3D::ROTZ, ps);
	
	return d;
}

class StokesMatrix Geometry::getFOVStokesRotationMatrix()
{
	/*i1=acos(correctMu(x_in*k_plane));*/
	class StokesMatrix FOVStokesRotationMatrix;
	FOVStokesRotationMatrix.setToUnityMatrix();
	return FOVStokesRotationMatrix;
}




