/**************************************************************************************************************
*
*       This file,
*
*                            /Geometry.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Geometry_h
#define _Geometry_h

#include "MapSampler.h"
#include "Grid.h"
#include "SimulationObject.h"

#include "Math/Vector3D.h"
#include "Math/StokesMatrix.h"

#include "BasicIO.h"

#include "BoxDetector.h"

class Geometry: public BasicIO, public SimulationObject 
{
	
	class BoxDetector boxDetector;
	// IO ##############################################################################################

public:
	enum FOVType_e
	{
		round,
		elliptic,
		square,
		rectangular,
		box,
		two_pi,
		four_pi,
		table
	};

	enum DetectorViewMode_e
	{
		normal_view_mode,	// normal view mode
		limb,			// limb mode, elevation angle is evaluated
		look_at			// position and reference point (ref_phi,ref_theta,ref_height) define geometry
	};

	enum InstrumentSensitivityFunction_e
	{
		unweighted,	// unweighted
		cosine,		// cosine signature
		gaussian	// gauss signature
	};

	enum SunPositionDescription_e
	{
		absolute, 
		detector_relative
	};

	// ############# detector properties.... ########################
	enum DetectorPositionDescription_e
	{
		spherical,
		cartesic
	};

private:

	class MapSampler FOV_function;
	char fn_FOV_function[BasicIO_string_size];
	
	enum DetectorPositionDescription_e DetectorPositionDescription;
	// geo coordinates
	double Altitude;						// geoposition altitude
	double Longitude;						// geoposition longitude
	double CoLatitude;						// geoposition latitude
	// reference point for "look_at" DetectorViewMode
	double ReferenceAltitude, ReferenceLongitude, ReferenceCoLatitude;	// reference point geo-coordinates

	// cartesic coordinates
	double X,Y,Z;

	// reference point for "look_at" DetectorViewMode
	double ReferenceX, ReferenceY, ReferenceZ;	// reference point geo-coordinates

	double NorthAzimuth;			// azymuth angle towards north pole meridian
	double Elevation;			// elevation angle beta
	double Roll;				// FOV Roll angle

	enum FOVType_e FOVType;
	double Omega;				// sight opening angle
	double OmegaHorizontal;			// horizontal aperture size in degree
	double OmegaVertical;			// vertical aperture size in degree

	enum InstrumentSensitivityFunction_e InstrumentSensitivityFunction;
	double ISGaussSigma;			// round FOV instrument sensitivity sigma
	double ISGaussSigmaVertical;		// rectangular FOV instrument sensitivity vertical sigma
	double ISGaussSigmaHorizontal;		// rectangular FOV instrument sensitivity horizontal sigma

	// box detectors -------------------------------------------------------
	char box_detector_definition[BasicIO_string_size];
	bool b_invert_faces;			// indicates whether to invert box faces and emmit in the opposite direction

	// limb simplifications -------------------------------------------------
	enum DetectorViewMode_e DetectorViewMode;
	double TangentHeight;			// for limb calculations
	double TangentHeightPrecision;		// precision of thetangent heigtt point calculation
	bool b_tangent_height_with_refraction;	// calculate tangent height point considering refraction
	
	// ################# sun properties ########################
	enum SunPositionDescription_e SunPositionDescription;	// position description of the sun
	static double SunDistance;			// geoposition r
	static double SunDiameter;			// sun's diameter	

	double SunLongitude;				// sun geoposition longitude
	double SunCoLatitude;				// sun geoposition latitude

	double SZA;					// solar zenith angle (detector depending)
	double SRAA;					// sun relative azimuth angle (detector depending)

	bool b_forward_geometry;

	// ##################################################################################################
	
	bool use_fixed_initial_direction;
	class Vector3D FixedInitialDirection;
	double FOV_mu_norm;	// for FOV sampling

protected:
	
	struct CellIndex_s DetectorCellIndex;

	class Vector3D DetectorPosition, DetectorCentralFOVDirection;
	class Vector3D SunPosition, SunDirection;

	class StokesMatrix getFOVStokesRotationMatrix();

	// void sampleFOVSolidAnglePoint(double *px, double *py);
	void sampleFOVDirection(class Vector3D& k_FOV);
	//class Vector3D sampleDetectorDirection();

	static void SunPositionFromSZAandRAZY(Vector3D *pSunPosition, class Vector3D& DetectorPosition, double azymuth, double SZA, double RAZY);

public:

	Geometry();
	
	void rotateIntoDetectorFOVSpherical(class Vector3D& v);
	void rotateIntoDetectorFOVCartesic(class Vector3D& v);
	inline void rotateIntoDetectorFOV(class Vector3D& v){pGrid->isSphericalGrid()?rotateIntoDetectorFOVSpherical(v):rotateIntoDetectorFOVCartesic(v);}
	
	inline bool hasPointSource(){return FOVType!=box;}

	inline void enableFixedInitialDirection(bool enable=true){use_fixed_initial_direction=enable;}
	inline void setFixedInitialDirection(const Vector3D& direction){FixedInitialDirection=direction;}

	bool sampleInitialState(class Vector3D *pPosition, struct CellIndex_s *pCellIndex, class Vector3D *pDirection);
	inline class BoxDetector& getBoxDetector(){return boxDetector;}

	bool isTwilightGeometry(double *pTwilightDistance);
	bool isPointBelowTerminator(class Vector3D& p);
	bool initializeGeometry();
	
	inline void setFOVType(const enum FOVType_e& FOVType){this->FOVType=FOVType;}
	inline void setFOVOmega(const double& SolidAngle){Omega=SolidAngle;}
	
	inline double getOmega(){return Omega;}
	inline enum FOVType_e getFOVType(){return FOVType;}
	inline double getOmegaVertical(){return OmegaVertical;}
	inline double getOmegaHorizontal(){return OmegaHorizontal;}
	
	// --------------------------------------------------
	inline double getX(){return X;}
	inline double getY(){return Y;}
	inline double getZ(){return Z;}
	
	inline double getAltitude(){return Altitude;}
	inline double getLongitude(){return Longitude;}
	inline double getCoLatitude(){return CoLatitude;}

	inline class Vector3D getDetectorPosition(){return DetectorPosition;}
	inline class Vector3D getDetectorCentralFOVDirection(){return DetectorCentralFOVDirection;}

	inline class Vector3D getSunPosition(){return SunPosition;}
	inline class Vector3D getSunDirection(){return SunDirection;}

	inline struct CellIndex_s getDetectorCellIndex(){return DetectorCellIndex;}

	// class Vector3D produceFOVVector(const double &x, const double &y);

	class Vector3D sampleSunDiscDirection(class Vector3D& p);
};

#endif //_Geometry_h


