/**************************************************************************************************************
*
*       This file,
*
*                            /GeometryList.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "GeometryList.h"
#include "trigonometry.h"

#include <time.h>
#include <limits.h>
#include <float.h>

void GeometryList::setSimulationPtr(Simulation *SimulationPtr)
{
	pSimulation=SimulationPtr;
	pGeometry=pSimulation->getGeometryPtr();
}	

bool GeometryList::loadGeometryDescription(char *file_name, bool time_in_seconds)
{
	if(!pSimulation->createIOParameterList(&GeometryParameterList, file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not load geometry description from file \"%s\"\n", file_name);
		return false;
	}
	list<struct Simulation::DataTableIOParameter_s>::iterator it=GeometryParameterList.begin();
	int line_number;
	if(time_in_seconds)
	{
		if(it++->column_type!=Simulation::DataTableIOParameter_s::time_stamp)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "first column of the geometry table \"%s\" must be 'time_stamp'!\n", file_name);
			return false;
		}
		// remove first entry in
		GeometryParameterList.pop_front();
		line_number=2;
	}
	else
	{
		if(it++->column_type==Simulation::DataTableIOParameter_s::parameter \
			|| it++->column_type==Simulation::DataTableIOParameter_s::parameter)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "first two columns of the geometry table (rows in \"%s\") must be 'carry' or 'ignore'!\n", file_name);
			return false;
		}
		// remove first two entries in
		GeometryParameterList.pop_front();
		GeometryParameterList.pop_front();
		line_number=3;
	}
	for(;it!=GeometryParameterList.end();it++)
	{
		if(it->column_type!=Simulation::DataTableIOParameter_s::parameter || it->BasicIOPtr!=pGeometry)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "\ngeometry description file \"%s\" line %d: only geometry parameters allowed!\n", file_name, line_number);
			return false;
		}	
		line_number++;
	}
	strcpy(geometry_description_file, file_name);
	return true;
}

bool GeometryList::allocateGeometryData(int no_geometries, int no_parameters)
{
	if(!SiE_of_geometry.allocate(no_geometries))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	if(!geometryParameter.allocate(no_geometries, no_parameters))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	this->no_geometries=no_geometries;
	this->no_parameters=no_parameters;
	return true;
}

bool GeometryList::loadGeometryData(char *file_name, bool time_in_seconds)
{
/*
	the first two columns of the geometry file must contain date and time!
*/
	MinimumTimeSiE=DBL_MAX;
	MaximumTimeSiE=DBL_MIN;
	if(GeometryParameterList.empty())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "geometry parameter description not loaded!\n");
		return false;
	}
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open geometries file \"%s\"!\n", file_name);
		return false;
	}
	int i, no_rows=numberOfRows(fptr);
	int j, no_columns=numberOfColumns(fptr);
	int tsc = time_in_seconds ? 1 : 2;
	if( no_columns - tsc != (int) GeometryParameterList.size())
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "number of columns (%d) in \"%s\" does not fit to \ngeometry table description in file \"%s\"!\n", \
			no_columns, file_name, geometry_description_file);
		return false;
	}
	if(!allocateGeometryData(no_rows, no_columns-tsc))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating geometry data table with %d rows and %d columns!", no_rows, no_columns-tsc);
		return false;
	}
	double SiE;
	char line_buffer[1000], value_buffer[100];
	char date_buffer[20], time_buffer[20];
	for(i=0;i<no_rows;i++)
	{
		if(!fgets(line_buffer, 1000, fptr))
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "unexpected end of file \"%s\" in line %d", file_name, i+1);
			return false;	
		}
		if(numberOf('\t', line_buffer)!=no_columns-1)
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "wrong TAB count in line %d of file \"%s\"!\n", i+1, file_name);
			return false;
		}
		// extract date and time
		if(time_in_seconds)
		{
			extractWordFromList(line_buffer, 0, '\t', time_buffer);
			SiE=atof(time_buffer);
		}
		else
		{
			int SiE_int;
			extractWordFromList(line_buffer, 0, '\t', date_buffer);
			extractWordFromList(line_buffer, 1, '\t', time_buffer);
			if(!calculateSecondsInEpoche(&SiE_int, date_buffer, time_buffer))
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "could not extract date and time from \"%s\" \"%s\" in line %d of file \"%s\"!\n", date_buffer, time_buffer, i+1, file_name);
				return false;
			}
			SiE=double(SiE_int);
		}
		if(SiE<MinimumTimeSiE)
			MinimumTimeSiE=SiE;
		if(SiE>MaximumTimeSiE)
			MaximumTimeSiE=SiE;
		SiE_of_geometry[i]=SiE;
		// read in geometry parameters
		for(j=0;j<no_parameters;j++)
		{
			extractWordFromList(line_buffer, tsc+j, '\t', value_buffer);
			geometryParameter[i][j]=atof(value_buffer);
		}
	}
	fclose(fptr);
	return true;
}

bool GeometryList::calculateGeometryWeightList(double StartTimeSiE, double StopTimeSiE, list<pair<int,double> > *pIFList)
{
	int i;
	double time_slice;
	double SiE_i, SiE_ip1;
	if(!(StartTimeSiE>=MinimumTimeSiE && StopTimeSiE<=MaximumTimeSiE))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "spectrum time frame outside geometries time frame!\n");
		return false;
	}
	for(i=0;i<no_geometries-1;i++)
	{
		SiE_i=SiE_of_geometry[i];
		SiE_ip1=SiE_of_geometry[i+1];
		if((StartTimeSiE>=SiE_i) && (StartTimeSiE<SiE_ip1))
		{
			pIFList->clear();	
			time_slice=SiE_ip1-StartTimeSiE;
			pIFList->push_back(pair<int, double>(int(i), time_slice));
		}
		if((StartTimeSiE<SiE_i) && (StopTimeSiE>=SiE_ip1))
		{
			time_slice=SiE_ip1-SiE_i;
			pIFList->push_back(pair<int, double>(int(i), time_slice));
			if(StopTimeSiE==SiE_ip1)
				return true;
		}
		if((SiE_i<StopTimeSiE) && (StopTimeSiE<=SiE_ip1))
		{
			time_slice=StopTimeSiE-SiE_i;
			pIFList->push_back(pair<int, double>(int(i), time_slice));
			return true;
		}
	}
	errorIn(__FILE__, __LINE__, __func__);
	fprintf(stderr, "could not find any geometry that fits to the time frame of the spectrum:!\n");
	return false;
}

bool GeometryList::applyGeometry(const int& i)
{
	list<struct Simulation::DataTableIOParameter_s>::iterator it=GeometryParameterList.begin();
	// skip time information
	int j=0;
	for(;it!=GeometryParameterList.end();it++)
		switch(it->pIOParameter->type)
		{
			case BasicIO::type_double:
				*((double *)it->pIOParameter->ptr)=geometryParameter[i][j++];
				break;
			case BasicIO::type_angle:
				*((double *)it->pIOParameter->ptr)=Rad(geometryParameter[i][j++]);
				break;
			default:;
		}
	return true;	
}

void GeometryList::formatTimeStampOfGeometry(int i, char *time_stamp)
{
	time_t SiE=(time_t) int(SiE_of_geometry[i]);
	formatTimeFromSecondsInEpoche(SiE, time_stamp);
}

int GeometryList::findGeometryIndex(double TimeSiE)
{
	int i;
	if(!(TimeSiE>=MinimumTimeSiE && TimeSiE<=MaximumTimeSiE))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "spectrum time frame outside geometries time frame!\n");
		return -1;
	}
	for(i=0;i<no_geometries;i++)
	{
		if(SiE_of_geometry[i]==TimeSiE)
			return i;
	}
	char time_string[20];
	formatTimeFromSecondsInEpoche(int(TimeSiE), time_string);
	errorIn(__FILE__, __LINE__, __func__);
	fprintf(stderr, "could not find any geometry for time %s!\n", time_string);
	return -1;
}

int GeometryList::findFirstGeometryIndexOfTimeRange(double StartTimeSiE, double StopTimeSiE)
{
	int i;
	double SiE_i, SiE_ip1;
	if(!(StartTimeSiE>=MinimumTimeSiE && StopTimeSiE<=MaximumTimeSiE))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "spectrum time frame outside geometries time frame!\n");
		return -1;
	}
	for(i=0;i<no_geometries-1;i++)
	{
		SiE_i=SiE_of_geometry[i];
		SiE_ip1=SiE_of_geometry[i+1];
		if((StartTimeSiE>=SiE_i) && (StartTimeSiE<SiE_ip1))
			return i;
	}
	errorIn(__FILE__, __LINE__, __func__);
	fprintf(stderr, "could not find any geometry that fits to the time frame of the spectrum:!\n");
	return -1;
}

bool GeometryList::loadTimeFrameData(char *file_name, bool time_in_seconds)
{
	if(!hasData())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "geometry data not loaded!\n");
		return false;
	}
	// allocate mean geometry vector
	if(!meanGeometryParameter.allocate(no_parameters))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating geometry data sample vector\n");
		return false;
	}
	resetMeanTimeAndGeometry();
	// open file and time frame data
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not integration frames file \"%s\"!\n", file_name);
		return false;
	}
	int tcs = time_in_seconds ? 2 : 4 ;
	if(numberOfColumns(fptr) != tcs)
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file \"%s\" has wrong number of columns (%d required)\n", file_name, tcs);
		return false;
	}
	// ---------------------------
	// allocate integration frames
	no_integration_frames=numberOfRows(fptr);
	if(!IntegrationFrame.allocate(no_integration_frames))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %d time frames according to file \"%s\"!\n", no_integration_frames, file_name);
		return false;
	}
	// ---------------------------
	double *snippet;
	int i, j, j1, j2;
	double start_in_SiE, stop_in_SiE;
	char line_buffer[2000], date_string[20], time_string[20];
	struct TimeInterval_s *pTimeInterval;
	for(i=0;i<no_integration_frames;i++)
	{
		if(!fgets(line_buffer, 2000, fptr))	
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d of file \"%s\"!\n", i+1, file_name);
			return false;
		}
		if(time_in_seconds)
		{
			if(!extractWordFromList(line_buffer, 0, '\t', time_string))
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error extracting intgeration start time in line %d of file \"%s\"!\n", i+1, file_name);
				return false;
			}
			start_in_SiE=atof(time_string);
			if(!extractWordFromList(line_buffer, 1, '\t', time_string))
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error extracting intgeration stop time in line %d of file \"%s\"!\n", i+1, file_name);
				return false;
			}
			stop_in_SiE=atof(time_string);
		}
		else
		{
			int SiE_int;
			// start time
			if(!extractWordFromList(line_buffer, 0, '\t', date_string)\
				 || !extractWordFromList(line_buffer, 1, '\t', time_string)\
				|| !calculateSecondsInEpoche(&SiE_int, date_string, time_string))
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error reading and/or calculating intgeration start time in line %d of file \"%s\"!\n", i+1, file_name);
				return false;
				
			}
			start_in_SiE=double(SiE_int);
			// stop time
			if(!extractWordFromList(line_buffer, 2, '\t', date_string)\
				 || !extractWordFromList(line_buffer, 3, '\t', time_string)\
				|| !calculateSecondsInEpoche(&SiE_int, date_string, time_string))
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error reading and/or calculating intgeration stop time in line %d of file \"%s\"!\n", i+1, file_name);
				return false;
			}
			stop_in_SiE=double(SiE_int);
		}
		// assign IntegrationFrame pointer
		pTimeInterval=IntegrationFrame.getCoefficientPtr()+i;
		// determine duration
		pTimeInterval->start_in_SiE = start_in_SiE;
		pTimeInterval->stop_in_SiE = stop_in_SiE;
		pTimeInterval->duration_in_seconds = stop_in_SiE - start_in_SiE ;
		// determine geometry list indices
		for(j=0;j<no_geometries;j++)
			if(SiE_of_geometry[j]>start_in_SiE)
				break;
		if(j==0 || j==no_geometries)
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "file \"%s\", line %d: integration frame \"%s\" outside geometry list time range!\n", file_name, i+1, line_buffer);
			return false;
		}
		pTimeInterval->start_geometry_index = j1 = j - 1;
		for(;j<no_geometries;j++)
			if(SiE_of_geometry[j]>=stop_in_SiE)
				break;
		if(j==no_geometries)
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "file \"%s\", line %d: integration frame \"%s\" outside geometry list time range!\n", file_name, i+1, line_buffer);
			return false;
		}
		pTimeInterval->stop_geometry_index = j2 = j ;
		// determine snippet size
		if(!pTimeInterval->snippet.allocate(j2-j1+1))
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "snippet allocation error\n");
			return false;
		}
		snippet = pTimeInterval -> snippet.getCoefficientPtr();
		j = j1 + 1;
		snippet [ 0 ] = SiE_of_geometry [j] - start_in_SiE;
		for(; j<j2; j++)
			snippet [ j - j1 ] = SiE_of_geometry [ j + 1 ] -SiE_of_geometry [ j ] ; 
		snippet [ j - j1 ] = stop_in_SiE -SiE_of_geometry [ j - 1 ] ; 
	}
	fclose(fptr);
	return true;
}

void GeometryList::sampleGeometry(const int& time_frame_index)
{
	int i, j=0;
	struct TimeInterval_s *pTimeInterval = IntegrationFrame.getCoefficientPtr() + time_frame_index;
	int i1 = pTimeInterval -> start_geometry_index;
	int i2 = pTimeInterval -> stop_geometry_index;
	double ti, time_sample = pTimeInterval->start_in_SiE + prand() * pTimeInterval -> duration_in_seconds;
	double weight, time1, time2, delta_time, p;
	// apply bisection search
	do
	{
		i = (i1+i2)/2;
		ti = SiE_of_geometry[i];
		if( time_sample == ti )
		{
			weight=1.0;
			goto GeometryList_sampleGeometry_set;
		}
		else if(time_sample>ti)
			i1=i;
		else
			i2=i;
	}while(i1+1!=i2);

	i=i1;

	time1 = SiE_of_geometry [ i1 ] ;
	time2 = SiE_of_geometry [ i2 ] ;
	delta_time = time2 - time1 ;
	weight = ( time2 - time_sample)  / delta_time;

GeometryList_sampleGeometry_set:
	// apply geometries -------------------------------------
	list<struct Simulation::DataTableIOParameter_s>::iterator it=GeometryParameterList.begin();
	for(;it!=GeometryParameterList.end();it++)
	{
		p = weight * geometryParameter[i][j] + ( 1.0 - weight ) * geometryParameter[i+1][j] ; 
		meanGeometryParameter[j] += p;
		switch(it->pIOParameter->type)
		{
			case BasicIO::type_double:
				*((double *)it->pIOParameter->ptr)=p;
				break;
			case BasicIO::type_angle:
				*((double *)it->pIOParameter->ptr)=Rad(p);
				break;
			default:;
		}
		j++;
	}
	mean_geometry_time += time_sample;
	no_geometry_samples++;
}

void GeometryList::resetMeanTimeAndGeometry()
{
	mean_geometry_time=0.0;
	meanGeometryParameter.setToZero();
	no_geometry_samples=0;
}

void GeometryList::printMeanTimeAndGeometry(FILE *fptr, bool time_in_seconds)
{
	int i=0;
	double mean_time = mean_geometry_time / double ( no_geometry_samples ) ;
	if(time_in_seconds)
		fprintf(fptr, "%le\t", mean_time);
	else
	{
		char line_buffer[2000];
		int mean_geometry_time_int = int ( mean_time ) ;
		formatTimeFromSecondsInEpoche(mean_geometry_time_int, line_buffer);
		fprintf(stderr, "%s\t", line_buffer);
	}
	double p;
	list<struct Simulation::DataTableIOParameter_s>::iterator it=GeometryParameterList.begin();
	for(;it!=GeometryParameterList.end();it++)
	{
		p = meanGeometryParameter[i] / double(no_geometry_samples);
		fprintf(fptr, "%lf\t", p);
		i++;
	}
}






