/**************************************************************************************************************
*
*       This file,
*
*                            /GeometryList.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _GeometryList_h
#define _GeometryList_h

#include "Simulation.h"
#include "Math/TMatrix.h"
#include "Math/TVector.h"

class GeometryList
{
	class Simulation *pSimulation;	
	class Geometry *pGeometry;

	list<struct Simulation::DataTableIOParameter_s> GeometryParameterList;

	int no_parameters, no_geometries;
	double MinimumTimeSiE, MaximumTimeSiE;
	class TVector<double> SiE_of_geometry;

	// ------
	class TMatrix<double> geometryParameter;
	char geometry_description_file[BasicIO_string_size];

	// infos on sampled geometries
	class TVector<double> meanGeometryParameter;
	double mean_geometry_time;
	int no_geometry_samples;

	// time intervals and integration frames
	struct TimeInterval_s
	{
		double start_in_SiE, stop_in_SiE;
		double duration_in_seconds;
		int start_geometry_index, stop_geometry_index;	// inidices into the geometry parameter list
		Vector snippet;
	};

	class TVector<struct TimeInterval_s> IntegrationFrame;
	int no_integration_frames;

protected:

	bool allocateGeometryData(int no_geometries, int no_parameters);

public:
	
	inline bool hasData(){return GeometryParameterList.size()!=0;}
	
	void resetMeanTimeAndGeometry();
	void printMeanTimeAndGeometry(FILE *fptr=stdout, bool time_in_seconds=true);

	inline int getNOGeometries(){return no_geometries;}
	inline int getNOIntegrationFrames(){return no_integration_frames;}

	void setSimulationPtr(Simulation *SimulationPtr);

	bool calculateGeometryWeightList(double StartTimeSiE, double StopTimeSiE, list<pair<int, double> > *pIFList);
	bool applyGeometry(const int& i);
	void sampleGeometry(const int& time_frame_index);

	void formatTimeStampOfGeometry(int i, char *time_stamp);
	int findFirstGeometryIndexOfTimeRange(double StartTimeSiE, double StopTimeSiE);
	int findGeometryIndex(double TimeSiE);

	bool loadGeometryDescription(char *file_name, bool time_in_seconds=false);
	bool loadGeometryData(char *file_name, bool time_in_seconds=false);
	bool loadTimeFrameData(char *file_name, bool time_in_seconds=false);

};

#endif // _GeometryList_h


