/**************************************************************************************************************
*
*       This file,
*
*                            /Grid.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Grid_h
#define _Grid_h

#include "GridIO.h"

#include "Grid1D.h"
#include "cell.h"

#include "Math/Vector3D.h"

#include "SurfaceGrid.h"
#include "BoxDetector.h"

class Geometry;

class Grid: public GridIO
{

protected:

	class Grid1D externalX[3];
	class Grid1D internalX[3];

	int ground_layer_index;		// index into height X[2] grid level containing the ground_level
	class SurfaceGrid *pSurfaceGrid;

public:

	inline bool isSphericalGrid(){return (GridType==sp1D || GridType==sp3D);}
	inline bool is3DGrid(){return (GridType==pp3D || GridType==sp3D);} 

	inline class Grid1D *getExternalGrid1DPtr(){return externalX;}
	inline class Grid1D *getInternalGrid1DPtr(){return internalX;}
	
	inline int getGrid1DResolution(const int& i){return internalX[i].getResolution();}
	virtual void printCell(FILE *fptr, struct CellIndex_s& CI)=0;
	
	virtual bool calculateGroundLayerIndex()=0;
	virtual int mapCellRangeIndexToGrid1DIndex(const int& i){return i;}

	inline void setSurfaceGridPtr(class SurfaceGrid *SurfaceGridPtr){pSurfaceGrid=SurfaceGridPtr;}

	// pure virtual / abstract interface
	virtual bool acceptGeometry(class Geometry *pGeometry)=0;
	virtual bool loadFiles()=0;
	virtual int getGridDimension()=0;
	virtual bool contains(class Vector3D& point)=0;

	virtual bool calculateCellIntersectionPoint(class Vector3D& Position, struct CellIndex_s *pCI, class Vector3D& Direction, double *pDistance, class Vector3D *pNextPosition, struct CellIndex_s *pNextCI)=0;

	virtual int getNOCells()=0;
	virtual int calculateCellIndex(int *index)=0;
	virtual int calculateCellIndex(const struct CellIndex_s& CI)=0;
	virtual bool calculateCellIndex(class Vector3D& point, struct CellIndex_s *pCI)=0;
	virtual bool calculateCellIndex(class Vector3D& point, class Vector3D& direction, struct CellIndex_s *pCI)=0;

	virtual void getGroundMapResolution(int *pxres, int *pyres)=0;
	virtual void convertHeightToGridHeight(double &h)=0;
	virtual double getVerticalExtension(const struct CellIndex_s& CI)=0;
	virtual void convertToVector3D(const double&, const double&, const double&, class Vector3D& p)=0;

	// virtual double getGroundLevel()=0;
	virtual void calculateSurfaceProjection(class Vector3D& position, double *px, double *py)=0;

	virtual void convertEquidistantAlbedoMapCoordinates(double *pXA, double *pYA, double *pXB, double *pYB)=0;
	virtual void convertAlbedoMapGrids(double *x, int xres, double *y, int yres)=0;

	virtual void sampleTOAPoint(class Vector3D *pPosition, class Vector3D *pDirection, class Vector3D& SunPosition)=0;
	
	// BoxDetector things....

	virtual void calculateEastAxis(double& x1, double& x2, double& x3, class Vector3D& n)=0;
	virtual void calculateSouthAxis(double& x1, double& x2, double& x3, class Vector3D& n)=0;
	virtual void calculateAltitudeAxis(double& x1, double& x2, double& x3, class Vector3D& n)=0;

	virtual void initializeBoxSamplingStructure(class MacroCell *pMacroCell, struct BoxDetector::SamplingStructure_s *pSamplingStructure)=0;
};

#endif //_Grid_h


