/**************************************************************************************************************
*
*       This file,
*
*                            /Grid1D.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Grid1D.h"
#include "fileio.h"
#include "error.h"

#include <string.h>
#include <stdlib.h>

Grid1D::Grid1D()
{
	g=(double *)0;
	resolution=1;
}

Grid1D::~Grid1D()
{
	if(g)
		delete [] g;
}

bool Grid1D::setResolution(int n)
{
	if(g)
		delete [] g;
	g=new double[2*n+1];
	if(!g)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"Grid1D::setResolution(): allocation error!\n");
		return false;
	}
	mg=g+n+1;
	resolution=n;
	return true;
}

bool Grid1D::loadFromFile(char *filename)
{
	FILE *fptr=fopen(filename, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not open file \"%s\"\n",filename);
		return false;
	}
	bool success=loadFromFile(fptr);
	fclose(fptr);
	return success;
}

bool Grid1D::loadFromFile(FILE *fptr)
{
	int i, n=numberOfRows(fptr);
	if(n<=0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"no lines in file!\n");
		return false;
	}
	if(!setResolution(n-1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"error setting resolution to %d!\n", n-1);
		return false;
	}
	for(i=0;i<n;i++)
		if(fscanf(fptr, "%lf\n", g+i)!=1)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d\n", i+1);
			return false;
		}
	return checkGrid();
}

void Grid1D::printToFile(FILE *fptr)
{
	int i;
	for(i=0;i<=resolution;i++)
		fprintf(fptr,"%lf\n",g[i]);
}

void Grid1D::printGridInfo(FILE *fptr, bool as_comment)
{
	fprintf(fptr,"%s%d %s intervals in [%lg%s,%lg%s] \n",as_comment?"# ":"",resolution,name,A,unit,B,unit);
}

bool Grid1D::checkGrid()
{
	int i;
	// check grid order
	for(i=1;i<=resolution;i++)
		if(g[i]<=g[i-1])
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "multiple defined value or NOT ascending (with row number) value order in grid entries %d and %d!\n", i, i+1);
			return false;
		}
	// calculate middle grid...
	for(i=0;i<resolution;i++)
		mg[i]=0.5*(g[i]+g[i+1]);
	A=g[0];
	B=g[resolution];
	return true;
}

void Grid1D::defineNameAndUnit(const char *name, const char *unit)
{
	strcpy(this->name, name);
	strcpy(this->unit, unit);
}

bool Grid1D::findIndex(double& x, int *pIndex)
{
	for(int i=0;i<=resolution;i++)
		if(g[i]==x)
		{
			*pIndex=i;
			return true;
		}
	// errorIn(__FILE__, __LINE__, __func__);
	return false;
}

bool Grid1D::findIntervallIndexFastButUnsafe(double& x, int *pIndex)
{
	int i=*pIndex;
	if(i<0 || i>resolution)
	{
		for(i=0;i<resolution;i++)
			if(x>g[i] && x<=g[i+1])
			{
				*pIndex=i;
				return true;
			}
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(x>g[i])// look into positive direction
	{
		for(;i<resolution;i++)
			if(x>g[i] && x<=g[i+1])
			{
				*pIndex=i;
				return true;
			}
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	for(;i>=0;i--)
		if(x>g[i] && x<=g[i+1])
		{
			*pIndex=i;
			return true;
		}
	errorIn(__FILE__, __LINE__, __func__);
	return false;
}





