/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/HITRANCalculator.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HITRANCalculator.h"
#include "SpectralWindow.h"

HITRANCalculator::HITRANCalculator()
{
        strcpy(fn_height_grid,"height_grid.txt");
        strcpy(fn_pressure,"pressure.txt");
        strcpy(fn_temperature,"temperature.txt");

	// HITRANAtmosphericGrid
	HITRANAtmosphericGrid.data=(double *)0;
	HITRANAtmosphericGrid.cs=(double **)0;

	// BasicIO
	addParameter(fn_pressure, type_string, "fn_pressure", 0, "pressure profile filename");
	addParameter(fn_temperature, type_string, "fn_temperature", 0, "temperature profile filename");
	addParameter(fn_height_grid, type_string, "fn_height_grid", 0, "vertical height grid");
}

HITRANCalculator::~HITRANCalculator()
{
	if(HITRANAtmosphericGrid.data)
	{
		delete [] HITRANAtmosphericGrid.data;
		HITRANAtmosphericGrid.data=(double *)0;
	}
	if(HITRANAtmosphericGrid.cs)
	{
		delete [] HITRANAtmosphericGrid.cs;
		HITRANAtmosphericGrid.cs=(double **)0;
	}
}


bool HITRANCalculator::mapAndAllocateWavenumberGrid(double **pnu, double *wavelength, int no_wavelengths, double *pnu_min, double *pnu_max)
{
	double nu_min=1e300,nu_max=1e-300,*nu=new double[no_wavelengths];
	if(!nu)
	{
		fprintf(stderr,"wavenumber allocation error\n");
		return false;
	}
	for(int i=0;i<no_wavelengths;i++)
	{
		double nu_temp=1e-2/wavelength[i];
		if(nu_temp>nu_max)
			nu_max=nu_temp;
		if(nu_temp<nu_min)
			nu_min=nu_temp;
		nu[i]=nu_temp;
	}
	/*nu_min-=2.0*wing_cut_off;
	nu_max+=2.0*wing_cut_off;*/
	// update...
	if(*pnu)
		delete [] *pnu;
	*pnu=nu;
	*pnu_max=nu_max;
	*pnu_min=nu_min;
	return true;
}

bool HITRANCalculator::getCrossSection(char *MoleculeName, double vmr, double T, double p, double *wavelength, int no_wavelengths, double *cross_section)
{
	if(!database_initialized)
	{
		fprintf(stderr,"HITRAN database not initialised!\n");
		return false;
	}
	int molecule_index;
	if(!getMoleculeIndex(MoleculeName, &molecule_index))
	{
		fprintf(stderr,"molecule \"%s\" not in database!\n", MoleculeName);
		return false;
	}
	if(!hasLocalMoleculeDataSet(molecule_index))
	{
		fprintf(stderr,"HITRAN database contains datasets for molecule \"%s\"...\n",MoleculeName);
		fprintf(stderr,"...but not the local copy\n");
		return false;
	}
	if(T<temperature_minimum || T>temperature_maximum)
	{
		fprintf(stderr,"temperature %lfK out of possible range [%lfK,%lfK]!\n",T,temperature_minimum,temperature_maximum);
		return false;
	}
	class HITRANMolecule *pHITRANMolecule=molecule+molecule_index;
	if(!pHITRANMolecule->load())
		return false;
	double lambdaA=wavelength[0];
	double lambdaB=wavelength[no_wavelengths-1];
	if(!pHITRANMolecule->extractRelevantLinesIn(lambdaA, lambdaB))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	class SpectralWindow LineList;
	if(!LineList.allocateSpectralRange(pHITRANMolecule))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	double p_s=vmr*p;
	LineList.calculateLineParameters(p, p_s, T);
	int i;
	for(i=0;i<no_wavelengths;i++)
		cross_section[i]=LineList.calculateCrossSection(wavelength[i]);
	return true;
}

/*bool HITRANDataBase::initAtmosphericGrid(int no_wavelengths)
{
	char filename[1000];
	formatHeightGridFilename(filename);
	int i,no_layers=numberOfRows(filename)-1;
	if(no_layers<=0)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"error in height grid file \"%s\". Empty or not present?\n",filename);
		return false;
	}
	// allocation
	int no_doubles=no_layers+1+2*no_wavelengths+no_wavelengths*no_layers;
	if(HITRANAtmosphericGrid.data)
		delete [] HITRANAtmosphericGrid.data;
	HITRANAtmosphericGrid.data=new double[no_doubles];
	if(HITRANAtmosphericGrid.cs)
		delete [] HITRANAtmosphericGrid.cs;
	HITRANAtmosphericGrid.cs=new double *[no_layers];
	if(!HITRANAtmosphericGrid.data || !HITRANAtmosphericGrid.cs)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "height grid allocation error\n");
		return false;
	}
	// assign pointers
	HITRANAtmosphericGrid.height=HITRANAtmosphericGrid.data;
	HITRANAtmosphericGrid.wavelength=HITRANAtmosphericGrid.height+no_layers+1;
	HITRANAtmosphericGrid.nu=HITRANAtmosphericGrid.wavelength+no_wavelengths;
	double *base=HITRANAtmosphericGrid.nu+no_wavelengths;
	// file read-in
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
		return false;
	for(i=0;i<=no_layers;i++)
	{
		if(fscanf(fptr,"%lf\n",HITRANAtmosphericGrid.height+i)!=1)
		{
			fprintf(stderr,"height grid file \"%s\", line %d:  error in line\n",filename,i+1);
			fclose(fptr);
			return false;
		}
		HITRANAtmosphericGrid.cs[i]=base+i*no_wavelengths;
	}
	fclose(fptr);
	HITRANAtmosphericGrid.no_layers=no_layers;
	// temperature and pressure files....
	formatPressureFilename(filename);
	if(!HITRANAtmosphericGrid.pressure.readFromFile(filename))
	{
		fprintf(stderr,"error in pressure file \"%s\". Empty or not present?\n",filename);
		return false;
	}
	formatTemperatureFilename(filename);
	if(!HITRANAtmosphericGrid.temperature.readFromFile(filename))
	{
		fprintf(stderr,"error in temperature file \"%s\". Empty or not present?\n",filename);
		return false;
	}
	HITRANAtmosphericGrid.no_wavelengths=no_wavelengths;
	return true;
}

bool HITRANDataBase::processAtmosphericGrid(char *MoleculeName, double *wavelength, int no_wavelengths, enum ProfileModel_e ProfileModel, void *vmr_profile, bool compute_absorption_coefficient)
{
	double vmr=0.0;
	class LinearelyInterpolatedFunctionTemplate<double> *profilePtr=(class LinearelyInterpolatedFunctionTemplate<double> *)0;
	if(ProfileModel==PFM_constant_vmr)
		vmr=*((double *)vmr_profile);
	else
		profilePtr=(class LinearelyInterpolatedFunctionTemplate<double> *)vmr_profile;
	if(!database_initialized)
	{
		fprintf(stderr,"HITRAN database not initialised!\n");
		return false;
	}
	int molecule_index,i,no_layers;
	if(!getMoleculeIndex(MoleculeName, &molecule_index))
		return false;
	if(!hasLocalMoleculeDataSet(molecule_index))
	{
		fprintf(stderr,"HITRAN database contains datasets for molecule \"%s\"...\n",MoleculeName);
		fprintf(stderr,"...but not the local copy. Get it from the HITRAN website!\n");
		return false;
	}
	class HITRANMolecule *pMolecule=molecule+molecule_index;
	if(!pMolecule->load() || !initAtmosphericGrid(no_wavelengths))
		return false;
	no_layers=HITRANAtmosphericGrid.no_layers;
	// calculate wavenumber array and look for nu_min and nu_max...
	double nu_min,nu_max,*nu=(double *)0;
	if(!mapAndAllocateWavenumberGrid(&nu, wavelength, no_wavelengths, &nu_min, &nu_max))
		return false;
	struct SpectralWindow_s *pSpectralWindow=pMolecule->getSpectralWindowPtr();
	pSpectralWindow->nuA_pcm=nu_min;
	pSpectralWindow->nuB_pcm=nu_max;
	if(!pMolecule->extractRelevantLines())
	{
		fprintf(stderr,"error extracting spectral lines [%lgcm^-1,%lgcm^-1]!\n",nu_min,nu_max);
		delete [] nu;
		return false;
	}
	for(i=0;i<no_layers;i++)
	{
		double T,p,height=0.5*(HITRANAtmosphericGrid.height[i]+HITRANAtmosphericGrid.height[i+1]);
		T=HITRANAtmosphericGrid.temperature.evaluate(height);
		p=HITRANAtmosphericGrid.pressure.evaluate(height);
		double tracegas_number_density,air_density;
		air_density=p/(1e4*k_B*T);
		switch(ProfileModel)
		{
			case PFM_vmr_profile:
				vmr=profilePtr->evaluate(height);
				tracegas_number_density=vmr*air_density;
				break;
			case PFM_number_density_profile:
				tracegas_number_density=profilePtr->evaluate(height);
				vmr=tracegas_number_density/air_density;
				break;
			default:
			case PFM_constant_vmr:
				tracegas_number_density=vmr*air_density;
				break;
		}
		pSpectralWindow->T=T;
		pSpectralWindow->p=p;
		pSpectralWindow->vmr=vmr;
		pMolecule->correctLineParameters();
		if(!molecule[molecule_index].renderCrossSection(no_wavelengths,nu,HITRANAtmosphericGrid.cs[i]))
		{
			fprintf(stderr,"error while rendering spectral lines to given wavelength grid...\n");
			delete [] nu;
			return false;
		}
		if(compute_absorption_coefficient)
		{
			double *base=HITRANAtmosphericGrid.cs[i];
			for(int j=0;j<no_wavelengths;j++)
				base[j]*=tracegas_number_density*1e5;
		}
	}
	memcpy(HITRANAtmosphericGrid.nu,nu,sizeof(double)*no_wavelengths);
	memcpy(HITRANAtmosphericGrid.wavelength,wavelength,sizeof(double)*no_wavelengths);
	// got all line parameters... now: render cross section
	delete [] nu;
	return true;
}

*/
