/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/HITRANDataBase.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HITRANDataBase.h"	

#include <math.h>
#include <stdlib.h>

#include "../File.h"
#include "../constants.h"
#include "../error.h"

HITRANDataBase::HITRANDataBase()
{
        IOName="HITRANDataBase";

        // initialize void particle
        strcpy(database_name,"hit04");
        strcpy(fn_molparam,"Global_Data/molparam.txt");
        strcpy(fn_parsum,"Global_Data/parsum.dat");
        strcpy(dir_molecule_directory,"HITRAN2004/By-Molecule/Uncompressed-files/");

        /*T_ref=296.0;
        p_ref=1013.25;
        wing_cut_off=20.0;                // see dissertation of Christian Frankenberg
	*/

	suppress_warning_messages=true;
	database_initialized=false;

	molecule=(class HITRANMolecule *)0;
	parsum_data=(double *)0;
	partition_sum=(double **)0;

	addParameter(dir_molecule_directory, type_string, "dir_molecule_directory", 0, "directory of the by-molecule folder");
	addParameter(database_name, type_string, "database_name", 0, "e.g. hit04");
	addParameter(fn_molparam, type_string, "fn_molparam", 0, "filename of the molparam.txt");
	addParameter(fn_parsum, type_string, "fn_parsum", 0, "file with partition function sums at T_ref");
	/*addParameter(&T_ref, type_double, "T_ref", 0, "reference temperature");
	addParameter(&p_ref, type_double, "p_ref", 0, "reference pressure");*/
}

HITRANDataBase::~HITRANDataBase()
{
	if(molecule)
	{
		// ...
		delete [] molecule;
		molecule=(class HITRANMolecule *)0;
	}
	if(parsum_data)
	{
		// ...
		delete [] parsum_data;
		parsum_data=(double *)0;
	}
	if(partition_sum)
	{
		// ...
		delete [] partition_sum;
		partition_sum=(double **)0;
	}
}

bool HITRANDataBase::readGlobalMoleculeParameters()
{
	File file;
	char filename[1000];
	formatGlobalMoleculeParameterFilename(filename);
	if(!file.open(filename,"rb"))
	{
		fprintf(stderr,"could not open molecule parameter file \"%s\"...!\n",filename);
		return false;
	}
	FILE *fptr=file.getFILEPtr();
	char line_buffer[100];
	if(!fgets(line_buffer,sizeof(line_buffer),fptr) || strncmp(line_buffer,"Molecule",8))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "molecule parameter file \"%s\" probably has wrong format!\n",filename);
		return false;
	}
	// int filepos_second_line=ftell(fptr);
	// determine number of molecules:
	no_molecules=0;
	while(fgets(line_buffer,sizeof(line_buffer),fptr)!=0) if(numberOf('(',line_buffer)==1) no_molecules++;
	if(molecule)
		delete [] molecule;
	molecule=new HITRANMolecule[no_molecules];
	if(!molecule)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"molecule allocation error\n");
		return false;
	}
	// allocated molecules... initialize them...
	int molecule_index=0;
	rewind(fptr);
	fgets(line_buffer,sizeof(line_buffer),fptr);
	// fseek(fptr,filepos_second_line,SEEK_SET);		// back to second line
	this->no_isotopes=0;
	int line_index=1;
	for(molecule_index=0;molecule_index<no_molecules;molecule_index++)
	{
		if(!fgets(line_buffer,sizeof(line_buffer),fptr))
		{
				fprintf(stderr,"error reading line %d in file \"%s\"!\n", line_index, filename);
				return false;
		}
		condenseLine(line_buffer);
		if(sscanf(line_buffer,"%s\t(%d)",molecule[molecule_index].name,&(molecule[molecule_index].HITRANIndex))!=2)
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"error evaluating molecule definition \"%s\" line from line %d of file \"%s\"!\n", line_buffer, line_index, filename);
			return false;
		}
		line_index++;
		molecule[molecule_index].molecule_index=molecule_index;
		int filepos_first_isotope=ftell(fptr);
		int no_isotopes=0,isotope_index;
		while(fgets(line_buffer,sizeof(line_buffer),fptr)!=0)
		{
			if(numberOf('(',line_buffer)==1)
				break;
			no_isotopes++;
		}
		if(!molecule[molecule_index].allocateIsotopes(no_isotopes))
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr, "isotope allocation error\n");
			return false;
		}
		this->no_isotopes+=no_isotopes;
		struct HITRANMoleculeIsotope_s *isotope=molecule[molecule_index].getIsotopePtr();
		fseek(fptr,filepos_first_isotope,SEEK_SET);		// back to first isotope
		for(isotope_index=0;isotope_index<no_isotopes;isotope_index++)
		{
			//Iso Abundance     Q(296K)      gj    Molar Mass(g)
			//161  .997317E+00    1.7464E+02    1     18.010565
			if(fscanf(fptr,"%d\t%lf\t%lf\t%d\t%lf\n",&(isotope->iso),&(isotope->abundance),&(isotope->Q_ref),&(isotope->gj),&(isotope->molar_mass))!=5)
			{
				fprintf(stderr,"file \"%s\", molecule \"%s\" isotope no. %d: error in line %d of file \"%s\"!\n",filename,molecule[molecule_index].name,isotope_index+1, line_index, filename);
				return false;
			}
			line_index++;
			isotope++;
		}
	}
	file.close();
	return true;
}

void HITRANDataBase::compressPartitionSumLine()
{
	char *destination_char_ptr=isotopes_line_buffer;
	char *source_char_ptr=isotopes_line, character;
	bool last_character_was_space=false;
	bool found_first_character=false;
	int n=0,nt=0;
	while(*source_char_ptr!='\0')
	{
		character=*source_char_ptr++;
		switch(character)
		{
			case 0x20:
				if(found_first_character)	// delete leading white spaces
				{
					if(!last_character_was_space)
					{
						*destination_char_ptr++='\t';
						nt++;
					}
					last_character_was_space=true;
				}
				break;

			case 0xd:
			case 0xa:break;

			default:
				*destination_char_ptr++=character;
				found_first_character=true;
				last_character_was_space=false;
		}
		n++;
	}
	*(--destination_char_ptr)='\0';
	strcpy(isotopes_line,isotopes_line_buffer);
}

bool HITRANDataBase::readPartitionSums()
{
	File file;
	char filename[1000];
	formatPartitionSumFilename(filename);
	if(!file.open(filename,"rb"))
	{
		fprintf(stderr,"could not open partitition sum file \"%s\"...!\n",filename);
		return false;
	}
	FILE *fptr=file.getFILEPtr();
	no_temperatures=numberOfRows(fptr)-1;
	if(parsum_data)
		delete [] parsum_data;
	parsum_data=new double[(no_isotopes+1)*no_temperatures];
	if(!parsum_data)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"partition sum raw data allocation error\n");
		return false;
	}
	temperature=parsum_data;
	if(partition_sum)
		delete [] partition_sum;
	partition_sum=new double *[no_temperatures];
	if(!partition_sum)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"partition sum pointer allocation error\n");
		return false;
	}
	// adjust pointers
	int i,j;
	for(i=0;i<no_isotopes;i++)
		partition_sum[i]=parsum_data+(1+i)*no_temperatures;
	// read the hole table	#######################################################
	if(!fgets(isotopes_line,sizeof(isotopes_line),fptr))
	{
		fprintf(stderr,"error reading header line of partition sum file \"%s\"...!\n",filename);
		return false;
	}
	compressPartitionSumLine();
	int no_matrix_columns=numberOf('\t',isotopes_line)+1;
	// allocate partition sum matrix
	double **matrix, *matrix_data;
	matrix_data=new double[no_temperatures*no_matrix_columns];
	if(!matrix_data)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	matrix=new double *[no_temperatures];
	if(!matrix)
	{
		delete [] matrix_data;
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	// read in all partition sums
	char buffer[100];
	for(i=0;i<no_temperatures;i++)
	{
		matrix[i]=matrix_data+i*no_matrix_columns;
		if(!fgets(isotopes_line, sizeof(isotopes_line),fptr))
		{
			fprintf(stderr,"error reading partition sum file \"%s\" line %d!\n",filename,i+2);
			delete [] matrix_data;
			delete [] matrix;
			return false;
		}
		compressPartitionSumLine();
		char *charptr=isotopes_line, *destination_char_ptr, character;
		double *matrixRowEntry=matrix[i];
		for(j=0;j<no_matrix_columns;j++)
		{
			destination_char_ptr=buffer;
			do
			{
				character=*charptr++;
				if(character=='\t'|| character=='\0')
					break;
				*destination_char_ptr++=character;
			}while(1);
			*(++destination_char_ptr)='\0';
			matrixRowEntry[j]=atof(buffer);
		}
	}
	// go back to first row and get isotopes line
	rewind(fptr);
	fgets(isotopes_line,sizeof(isotopes_line),fptr);
	file.close();
	compressPartitionSumLine();
	// iterate over the all isotopes...
	int index=0;
	for(i=1;i<no_matrix_columns;i++)
	{
		char text_buffer[50];
		extractWordFromList(isotopes_line,i,'\t',buffer);
		extractWordFromList(buffer, 0, '_', text_buffer);
		int molecule_index;
		if(!getMoleculeIndex(text_buffer,&molecule_index) && !suppress_warning_messages)
			fprintf(stderr,"warning: unknown molecule \"%s\" in partition sum file!\n",text_buffer);
		else
		{
			int iso,n;
			extractWordFromList(buffer, 1, '_', text_buffer);
			iso=atoi(text_buffer);
			n=molecule[molecule_index].no_isotopes;
			struct HITRANMoleculeIsotope_s *isotope=molecule[molecule_index].getIsotopePtr();
			for(j=0;j<n;j++)
				if(isotope[j].iso==iso)
				{
					// i is the matrix column
					double *ps=partition_sum[index];
					isotope[j].partition_sum_index=index;
					for(int k=0;k<no_temperatures;k++)
						ps[k]=matrix[k][i];
					index++;
					break;
				}
			if(j==n)
			{
				extractWordFromList(buffer, 0, '_', text_buffer);
				if(!suppress_warning_messages)
					fprintf(stderr,"warning: iso=%d of molecule \"%s\" of partition sum table not present in \"%s\"!\n",iso, text_buffer, fn_molparam);
			}
		}
	}
	for(i=0;i<no_temperatures;i++)
		temperature[i]=matrix[i][0];
	temperature_minimum=temperature[0];
	temperature_maximum=temperature[no_temperatures-1];
	delete [] matrix_data;
	delete [] matrix;
	return true;
}

bool HITRANDataBase::open(char *filename)
{
	if(database_initialized)
		return true;
	if(!read(filename))
	{
		fprintf(stderr,"could not read database description IO file \"%s\"...\n", filename);
		return false;
	}
	if(!(readGlobalMoleculeParameters() && readPartitionSums()))
	{
		fprintf(stderr,"database initialisation failed... see above!\n");
		return false;
	}
	HITRANMolecule::setHITRANDataBasePtr(this);
	return database_initialized=true;
}

void HITRANDataBase::formatMoleculeParameterFile(int molecule_index, char *filename)
{
	sprintf(filename,"%s%s%02d_%s.par", dir_base, dir_molecule_directory, molecule[molecule_index].HITRANIndex, database_name);
}

bool HITRANDataBase::hasLocalMoleculeDataSet(int molecule_index)
{
	char filename[2000];
	formatMoleculeParameterFile(molecule_index,filename);
	return doesFileExist(filename);
}

bool HITRANDataBase::determinePartitionSumIndex(int *pPartitionIndex, int molecule_index, int isotope_index)
{
	*pPartitionIndex=-1;
	// determine partition sum index
	int j;
	char text_buffer[30];
	HITRANMoleculeIsotope_s *isotope=molecule[molecule_index].getIsotopePtr();
	sprintf(text_buffer,"%s_%d",molecule[molecule_index].name,isotope[isotope_index].iso);
	for(j=0;j<no_isotopes;j++)
	{
		char extracted_isotope_name[20];
		extractWordFromList(isotopes_line,j+1,'\t',extracted_isotope_name);
		if(!strcmp(extracted_isotope_name,text_buffer))
			break;
	}
	if(j==no_isotopes)
	{
		fprintf(stderr, "Isotope \"%s\" not found!\n",text_buffer);
		return false;
	}
	*pPartitionIndex=j;
	return true;
}


bool HITRANDataBase::getMoleculeIndex(char *MoleculeName, int *pMoleculeIndex)
{
	int molecule_index;
	for(molecule_index=0;molecule_index<no_molecules;molecule_index++)
		if(!strcmp(molecule[molecule_index].name,MoleculeName))
			break;
	if(molecule_index==no_molecules)
	{
		fprintf(stderr,"HITRAN database does not contain datasets for molecule \"%s\"\n",MoleculeName);
		return false;
	}
	*pMoleculeIndex=molecule_index;
	return true;
}


/*

	typical volume mixing ratios (vmr)

	N2		0,7809
	O2		0,2095
	Ar		0,0093
	CO2		0,00033
	Ne		18 ppm
	He		5 ppm
	Kr		1 ppm
	Xe		0,09 ppm
	CH4		1,5 ppm
	CO		0,1 ppm
	H2		0,5 ppm
	N2O		0,25 ppm
	O3 	bis zu 10 ppm in der Stratosphre reine Luft 	5 - 50 ppb in Erdbodennhe belastete Luft 	bis 500 ppb in Erdbodennhe
	SO2 	0,2 ppb
	NH3 	6 ppb
	NO2 	1 ppb, 100 ppb in belasteter Luft
	CH2O 	0 - 10 ppb



	p_s ist der Partialdruck. Dazu wikipedia:

	Das Dalton-Gesetz (Daltonsches Gesetz, Gesetz der Partialdrcke), 1805 von John Dalton formuliert, besagt, dass die Summe aller Partialdrcke pi 
	bei idealen Gasen gleich dem Gesamtdruck des Gemisches pGesamt ist. Fr k Komponenten ergibt sich:

    	p_\mathrm{Gesamt} = \sum_{i=1}^k p_i

	Hieraus leitet sich ab, dass der Partialdruck des i-ten Gases gleich dem Produkt aus Molenbruch yi des Gases mal Gesamtdruck des Gemisches 
	(beispielsweise Luftdruck) ist. Dies ist zwar eine idealisierte Darstellung fr den Fall, dass die Teilchen der Gasphase auer der 
	mechanischen keine gegenseitige Wechselwirkung haben (ideale Gase), jedoch kann man diese im Normalfall vernachlssigen.

		p_i = y_i \cdot p

	Das Verhltnis der Teilchenanzahl (Stoffmenge) n_i einer Komponente i zur Gesamtteilchenzahl nGesamt des Gemisches entspricht hierbei 
	dem Molenbruch yi und somit auch dem Partialdruck pi der Komponente i zum Gesamtdruck pGesamt des Gemisches. Es gilt folglich:

		\frac{n_i}{n_\mathrm{Gesamt}} = \frac{p_i}{p_\mathrm{Gesamt}}

	Beispiel: trockene Luft in Meereshhe

	Die folgende Tabelle zeigt die Zusammensetzung von vollstndig trockener Luft auf Meereshhe (Normalbedingung) also bei einem Luftdruck von 1013,25 hPa.

	Man kann den Volumenanteil mit dem Teilchenanteil (Molenbruch) gleichsetzen, da es sich um annhernd ideale Gase handelt. Sonstige Bestandteile
	der Luft kann man aufgrund ihres geringen Anteils vernachlssigen.
	Tabelle I: bersicht Partialdrcke trockene Luft in Meereshhe (unter Verwendung aller blicher Einheiten) Komponente 	Volumenanteil
	% 	Partialdruck
					hPa (mbar) 	kPa 		mmHg		(Torr) 		bar
	Luft 			100,00 		1.013,25 	101,325 	759,96 		1,01325
	Stickstoff 		78,090 		791,25 		79,12 		593,45 		0,79125
	Sauerstoff 		20,950 		212,28 		21,23 		159,21 		0,21228
	Argon 			0,927 		9,39 		0,939 		7,04 		0,00939
	Kohlendioxid 	0,033 		0,33 		0,033 		0,25 		0,00033

*/

double HITRANDataBase::calculatePartitionSum(int partition_sum_index, double T)
{
	int i,k,index,n;
	double c,d,Q=0.0;
	if(T<temperature_minimum)
	{
		// 3 point Lagrange interpolation
		n=3;
		index=2;
	}
	else if(T>temperature_maximum)
	{
		// 3 point Lagrange interpolation
		n=3;
		index=no_temperatures-1;
	}
	else
	{
		// 4 point Lagrange interpolation
		// find first temperature which is above T
		for(index=0;index<no_temperatures;index++)
			if(temperature[index]>T)
				break;
		if(index==no_temperatures)
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"temperature error\n");
			return 0.0;
		}
		n=4;
	}
	double *QBase=partition_sum[partition_sum_index];
	for(i=0;i<n;i++)
	{
		c=1.0;
		d=1.0;
		for(k=0;k<n;k++)
			if(k!=i)
			{
				c*=T-temperature[index-2+k];
				d*=temperature[index-2+i]-temperature[index-2+k];
			}
		Q+=(c*QBase[index-2+i])/d;
	}
	return Q;
}


class HITRANMolecule *HITRANDataBase::getMoleculePtr(char *MoleculeName)
{
	int molecule_index;
	return getMoleculeIndex(MoleculeName, &molecule_index)?molecule+molecule_index:0;
}


