/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/HITRANMolecule.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HITRANMolecule.h"
#include "HITRANDataBase.h"
#include "SpectralWindow.h"
#include "../error.h"

// static variables
class HITRANDataBase *HITRANMolecule::pHITRANDataBase=(class HITRANDataBase *)0;

HITRANMolecule::HITRANMolecule()
{
        T_ref=296.0;
        p_ref=1013.25;
        wing_cut_off=20.0;                // see dissertation of Christian Frankenberg
	suppress_warning_messages=true;
	HITRANMoleculeIsotope=(struct HITRANMoleculeIsotope_s *)0;
}

HITRANMolecule::~HITRANMolecule()
{
	if(HITRANMoleculeIsotope)
	{
		delete [] HITRANMoleculeIsotope;
		HITRANMoleculeIsotope=(struct HITRANMoleculeIsotope_s *)0;
	}
}

bool HITRANMolecule::allocateIsotopes(int n)
{
	if(HITRANMoleculeIsotope)
		delete [] HITRANMoleculeIsotope;
	HITRANMoleculeIsotope=new struct HITRANMoleculeIsotope_s[n];
	if(!HITRANMoleculeIsotope)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;	
	}
	no_isotopes=n;
	return true;
}

bool HITRANMolecule::load()
{
	char filename[2047];
	pHITRANDataBase->formatMoleculeParameterFile(molecule_index, filename);
	if(!loadBlockData(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	resetEnabledParameters();
	enableParameter(HLP_nu);
	createEnabledParametersList();
	extractEnabledLineParameters();
	return true;
}

bool HITRANMolecule::extractRelevantLinesIn(const double& lambdaA, const double& lambdaB)
{
	// lambdaA < lambdaB !!!
	if(lambdaA>lambdaB)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int i;
	double nu;
	SpectralRange.nuA_pcm=1e-2/lambdaB;
	SpectralRange.nuB_pcm=1e-2/lambdaA;

	// determine index range
	nu=SpectralRange.nuA_pcm-wing_cut_off;
	for(i=0;i<no_spectral_lines;i++)
		if(LineParameterStruct[i].nu>nu)
			break;
	if(i==no_spectral_lines)
	{
		// fprintf(stderr,"spectral window too large!\n");
		SpectralRange.no_lines=0;
		return true;
	}
	SpectralRange.iA=i-1;  // <- iA
	nu=SpectralRange.nuB_pcm+wing_cut_off;
	for(i=i;i<no_spectral_lines;i++)
		if(LineParameterStruct[i].nu>nu)
			break;
	if(i==no_spectral_lines)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"spectral window too large!\n");
		return false;
	}
	SpectralRange.iB=i;	// <- iB
	// extract defulat parameters for index range
	enableDefaultParameters();
	createEnabledParametersList();
	extractEnabledLineParameters(SpectralRange.iA, SpectralRange.iB);
	
	SpectralRange.no_lines=SpectralRange.iB-SpectralRange.iA+1;
	/*// allocate corrected spectral line parameter structs...
	if(SpectralWindow.CorrectedLPS)
		delete [] SpectralWindow.CorrectedLPS;
	SpectralWindow.CorrectedLPS=new struct CorrectedLineParameterStruct_s[n];
	if(!SpectralWindow.CorrectedLPS)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not allocate %d corrected line parameter records....!\n", n);
		return false;
	}*/
	SpectralRange.lambdaA=1e-2/(SpectralRange.nuB_pcm+wing_cut_off);
	SpectralRange.lambdaB=1e-2/(SpectralRange.nuA_pcm-wing_cut_off);
	return true;
}

double HITRANMolecule::calculatePartitionSum(struct HITRANMoleculeIsotope_s *pMI, const double& T)
{
	return pHITRANDataBase->calculatePartitionSum(pMI->partition_sum_index, T);
}




