/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/LineParameterFile.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "LineParameterFile.h"

struct LineParameterStruct_s LineParameterFile::DummyLPS;

struct LineParameterFile::LineParameter_s LineParameterFile::LineParameter[HITRANLineParameters]=
{
	{2	,0	, LPType_integer	, int((char *)&(LineParameterFile::DummyLPS.M)-(char *)&(LineParameterFile::DummyLPS))},
	{1	,2	, LPType_integer	, int((char *)&(LineParameterFile::DummyLPS.I)-(char *)&(LineParameterFile::DummyLPS))},
	{12	,3	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.nu)-(char *)&(LineParameterFile::DummyLPS))},
	{10	,15	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.S)-(char *)&(LineParameterFile::DummyLPS))},
	{10	,25	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.A)-(char *)&(LineParameterFile::DummyLPS))},
	{5	,35	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.gamma_air)-(char *)&(LineParameterFile::DummyLPS))},
	{5	,40	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.gamma_self)-(char *)&(LineParameterFile::DummyLPS))},
	{10	,45	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.E_dash_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{4	,55	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.n_air)-(char *)&(LineParameterFile::DummyLPS))},
	{8	,59	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.delta_air)-(char *)&(LineParameterFile::DummyLPS))},
	{15	,67	, LPType_Hollerith	, int((char *)&(LineParameterFile::DummyLPS.V_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{15	,82	, LPType_Hollerith	, int((char *)&(LineParameterFile::DummyLPS.V_dash_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{15	,97	, LPType_Hollerith	, int((char *)&(LineParameterFile::DummyLPS.Q_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{15	,112	, LPType_Hollerith	, int((char *)&(LineParameterFile::DummyLPS.Q_dash_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{6	,127	, LPType_string		, int((char *)&(LineParameterFile::DummyLPS.I_err)-(char *)&(LineParameterFile::DummyLPS))},
	{12	,133	, LPType_string		, int((char *)&(LineParameterFile::DummyLPS.I_ref)-(char *)&(LineParameterFile::DummyLPS))},
	{1	,145	, LPType_flag		, int((char *)&(LineParameterFile::DummyLPS.flag)-(char *)&(LineParameterFile::DummyLPS))},
	{7	,146	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.g_dash)-(char *)&(LineParameterFile::DummyLPS))},
	{7	,153	, LPType_real		, int((char *)&(LineParameterFile::DummyLPS.g_dash_dash)-(char *)&(LineParameterFile::DummyLPS))}
};


LineParameterFile::LineParameterFile()
{
	block_data=(char *)0;
	LineParameterStruct=(struct LineParameterStruct_s *)0;
	resetEnabledParameters();
	no_enabled_parameters=0;
}

LineParameterFile::~LineParameterFile()
{
	if(block_data)
	{
		delete [] block_data;
		block_data=(char *)0;
	}
	if(LineParameterStruct)
	{
		delete [] LineParameterStruct;
		LineParameterStruct=(struct LineParameterStruct_s *)0;
	}
}

void LineParameterFile::resetEnabledParameters()
{
	for(int i=0;i<HITRANLineParameters;i++)
		parameterMask[i]=false;
}

bool LineParameterFile::loadBlockData(char *filename)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
	{
		fprintf(stderr,"could not open line parameter file \"%s\"!\n",filename);
		return false;
	}
	fseek(fptr,0,SEEK_END);
	long filesize=ftell(fptr);
	fptr=freopen(filename,"rb",fptr);
	if(!fptr)
	{
		fprintf(stderr,"could not open line parameter file \"%s\" in binary mode!\n", filename);
		return false;
	}
	no_spectral_lines=int(filesize/HITRANLineParameterRecordSize);
	if(block_data)
		delete [] block_data;
	block_data=new char[filesize];
	if(!block_data || fread(block_data,HITRANLineParameterRecordSize, no_spectral_lines, fptr)!=(unsigned int) no_spectral_lines)
	{
		fprintf(stderr,"could not %s block data for line parameter file \"%s\"!\n", block_data?"read":"allocate", filename);
		fclose(fptr);
		return false;
	}
	fclose(fptr);
	setFilename(filename);
	if(LineParameterStruct)
		delete [] LineParameterStruct;
	LineParameterStruct=new struct LineParameterStruct_s[no_spectral_lines];
	if(!LineParameterStruct)
	{
		fprintf(stderr,"file \"%s\": could not allocate %d line parameter records\n", filename, no_spectral_lines);
		return false;	
	}
	return true;
}

void LineParameterFile::enableDefaultParameters()
{
	resetEnabledParameters();
	//parameterMask[HLP_nu]=true;
	parameterMask[HLP_I]=true;
	parameterMask[HLP_n_air]=true;
	parameterMask[HLP_gamma_air]=true;
	parameterMask[HLP_gamma_self]=true;
	parameterMask[HLP_nu]=true;
	parameterMask[HLP_delta_air]=true;
	parameterMask[HLP_S]=true;
	parameterMask[HLP_E_dash_dash]=true;
}

void LineParameterFile::createEnabledParametersList()
{	
	no_enabled_parameters=0;
	for(int i=0;i<HITRANLineParameters;i++)
		if(parameterMask[i])
			pLineParameter[no_enabled_parameters++]=LineParameter+i;
}

void LineParameterFile::extractLineParameters(int index, bool parameterMask[HITRANLineParameters])
{
	int j,k, n=no_enabled_parameters, l;
	// count and link enabled parameters
	char buffer[16], *blockdataptr=block_data+index*HITRANLineParameterRecordSize;
	char *source_ptr, *destination_ptr;
	void *varptr;
	for(j=0;j<n;j++)
	{
		l=pLineParameter[j]->length;
		source_ptr=blockdataptr+pLineParameter[j]->position;
		destination_ptr=buffer;
		for(k=0;k<l;k++)
			*destination_ptr++=*source_ptr++;
		buffer[k]='\0';
		varptr=(void *)((char *)(LineParameterStruct+index)+pLineParameter[j]->offset);
		switch(pLineParameter[j]->type)
		{
			case LPType_real:
				*((double *)varptr)=atof(buffer);
				break;
			case LPType_integer:
				*((int *)varptr)=atoi(buffer);
				break;
			case LPType_flag:
				*((char *)varptr)=buffer[0];
				break;
			case LPType_string:
				strcpy((char *)varptr,buffer);
				break;
			case LPType_Hollerith:
				strcpy((char *)varptr,buffer);
				break;
			default:;
		}
	}
}

bool LineParameterFile::extractLineParameters(int iA, int iB, bool parameterMask[HITRANLineParameters])
{
	if(!no_enabled_parameters)
		return true;
	for(int i=iA;i<=iB;i++)
		extractLineParameters(i, parameterMask);
	return true;
}

