/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/SpectralWindow.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "SpectralWindow.h"
#include "HITRANMolecule.h"

#include <stdio.h>
#include <memory.h>
#include <math.h>
#include <limits.h>

#include "../constants.h"
#include "Voigt.h"
#include "../error.h"

#define c2						1.4399		// hc/k_B in [cm/K]

// class HITRANDataBase *SpectralWindow::pHITRANDataBase=0;

SpectralWindow::SpectralWindow()
{
	pHITRANMolecule=0;
	line_index=0;
}

bool SpectralWindow::allocateSpectralRange(class HITRANMolecule *HITRANMoleculePtr)
{
	pHITRANMolecule=HITRANMoleculePtr;
	pSpectralRange=pHITRANMolecule->getSpectralRangePtr();
	if(!AtmosphericLPSData.allocate(pSpectralRange->no_lines, 8))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	AtmosphericLPS=(AtmosphericLineParameterStruct_s *)AtmosphericLPSData.getDataPtr();
	// allocate Q_T vector
	if(!Q_T.allocate(pHITRANMolecule->getNOIsotopes()))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	return true;
}

void SpectralWindow::calculateLineParameters(const double& p, const double& p_s, const double& T)
{
	// struct AtmosphericLineParameterStruct_s *pALPS, struct LineParameterStruct_s *pLPS)
	struct HITRANMoleculeIsotope_s *pMI, *MI=pHITRANMolecule->getIsotopePtr();
	struct LineParameterStruct_s *pLPS=pHITRANMolecule->getLineParameterStructPtr();
	struct AtmosphericLineParameterStruct_s *pALPS=AtmosphericLPS;
	double S, nu, gamma_L, gamma_D;

	double T_ref=pHITRANMolecule->getReferenceTemperature();
	double p_ref=pHITRANMolecule->getReferencePressure();
	double wing_cut_off=pHITRANMolecule->getWingCutOff();
	int i, isotope_index, n=pHITRANMolecule->getNOIsotopes();
	int iA=pSpectralRange->iA, iB=pSpectralRange->iB;

	// calculate Q_T
	for(isotope_index=0;isotope_index<n;isotope_index++)
		Q_T[isotope_index]=pHITRANMolecule->calculatePartitionSum(MI+isotope_index, T);
	
	
	pLPS+=iA;
	for(i=iA;i<=iB;i++)
	{
		isotope_index=pLPS->I-1;
		pMI=MI+isotope_index;

		pALPS->nu=nu=pLPS->nu+(p*pLPS->delta_air)/p_ref;
		pALPS->nu_A=nu-wing_cut_off;
		pALPS->nu_B=nu+wing_cut_off;

		gamma_L=pALPS->gamma_L=(pow(T_ref/T,pLPS->n_air)*pLPS->gamma_air*(p-p_s)+pLPS->gamma_self*p_s)/p_ref;
		gamma_D=pALPS->gamma_D=(nu*sqrt((2e3*k_B*Avogadro*T)/pMI->molar_mass))/speed_of_light;
		pALPS->y=gamma_L/gamma_D;

		S = pLPS->S;
		S *= pMI->Q_ref / Q_T[isotope_index];
		S *= exp(-c2*pLPS->E_dash_dash*(1.0/T-1.0/T_ref));
		S *= (1.0-exp(-c2*nu/T))/(1.0-exp(-c2*nu/T_ref));		// <- unclear: vacuum wavenumber or corrected wavenumber?
		pALPS->S = S;
		
		pALPS->Area=S/(gamma_D*sqrt(M_PI));

		// increment
		pALPS++;
		pLPS++;
	}

	// set internal state variables
	this->p=p;
	this->p_s=p_s;
	this->T=T;
}

int SpectralWindow::findLineIndex(const double& nu)
{
	int i=line_index;
	// determine smallest possible i such that 
	// AtmosphericLPS[i].nu_min < nu  && nu < AtmosphericLPS[i].nu_max
	if(AtmosphericLPS[i].nu_B<nu) // go right?
	{
		while(AtmosphericLPS[i].nu_B<nu)
			if(++i>=pSpectralRange->no_lines)
			{
				--i;
				break;
			}
		return i;
	}
	else
	{
		while(AtmosphericLPS[i].nu_B>nu)
			if(--i==0)
				break;
		++i;
		return i;
	}
	return 0;
}

double SpectralWindow::calculateCrossSection(const double& lambda)
//bool HITRANMolecule::renderCrossSection(int no_wavenumbers, double* nu, double *cross_section)
{
	if(lambda<pSpectralRange->lambdaA || lambda>pSpectralRange->lambdaB)
		return 0.0;
	int i, no_lines;
	double nu, nu_line, x, y, gamma_D, Area, cross_section=0.0;
	struct AtmosphericLineParameterStruct_s *pALPS=AtmosphericLPS;
	nu=1e-2/lambda;
	line_index=i=findLineIndex(nu);
	pALPS+=i;
	no_lines=pSpectralRange->no_lines;
	for(;i<no_lines;i++)
	{
		if(pALPS->nu_A>nu)
			break;
		// nu
		nu_line	= pALPS->nu;

		// line shape parameters
		gamma_D	= pALPS->gamma_D;
		Area	= pALPS->Area;
		y	= pALPS->y;

		x=(nu-nu_line)/gamma_D;
		cross_section+=Area*KuntzHumlicekVoigtProfile(x,y);
		pALPS++;
	}
	return cross_section;
}

/*
bool HITRANMolecule::renderCrossSection(int no_wavenumbers, double* nu, double *cross_section)
{
	int i,j,n=SpectralWindow.no_lines;
	memset(cross_section, 0, sizeof(double)*no_wavenumbers); // reset line_shape
	double nu_line, nu_min, nu_max, nu_j, gamma_D, gamma_L, S, x, y;
	struct CorrectedLineParameterStruct_s *pALPS=SpectralWindow.CorrectedLPS;
	if(!pALPS)
		return true;
	for(i=0;i<n;i++)
	{
		// nu
		nu_line=pALPS->nu;
		nu_min=nu_line-wing_cut_off;
		nu_max=nu_line+wing_cut_off;
		// line shape parameters
		gamma_D=pALPS->gamma_D;
		gamma_L=pALPS->gamma_L;
		y=gamma_L/gamma_D;
		// intensity
		S=pALPS->S;
		for(j=0;j<no_wavenumbers;j++)
		{
			// i: spectral line index
			// j: wavelength/wavenumber index
			nu_j=nu[j];
			if(nu_j>nu_min && nu_j<nu_max)
			{
				x=(nu_j-nu_line)/gamma_D;
				cross_section[j]+=(S*KuntzHumlicekVoigtProfile(x,y))/(gamma_D*sqrt(M_PI));
			}
		}
		pALPS++;
	}
	return true;
}
*/


