/**************************************************************************************************************
*
*       This file,
*
*                            /HITRAN/main_HITRANExtract.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HITRANCalculator.h"
#include <stdio.h>
#include <stdlib.h>

void print_usage(FILE *fptr)
{
	// fprintf(fptr,"usage: HITRANAccess HITRAN.conf [commands] [parameters]\n");
	fprintf(fptr,"usage: HITRANExtract HITRAN.conf molecule_name(e.g. O2) vmr(0-1)[1] T[K] p[hPa] wl1[m] wl2[m] n\n");
	// fprintf(fptr,"commands:\n");
	// fprintf(fptr,"\t scs (single cross section calculation)\n");
	// fprintf(fptr,"\t\tparameters  : molecule_name(e.g. O2) vmr(0-1)[1] T[K] p[hPa] wl1[m] wl2[m] n\n");
	/*fprintf(fptr,"\t gcs (grid cross section calculation)\n");
	fprintf(fptr,"\t\tparameters  : molecule_name[e.g. O2] vmrmode[cvmr|ndprf|vmrprf] vmr|profile wl1[m] wl2[m] n\n");
	fprintf(fptr,"\t gac (grid absorption coefficient)\n");
	fprintf(fptr,"\t\tparameters  : molecule_name[e.g. O2] vmrmode[cvmr|ndprf|vmrprf] vmr|profile wl1[m] wl2[m] n\n");*/
/*	fprintf(fptr,"\n\ntemplate for HITRAN.conf:\n\n");
	HITRANDataBaseIO HDBIO;
	HDBIO.writeTemplate(fptr);
	fprintf(fptr,"\n\ntemplate for HITRAN.conf:\n\n");*/
}

int main(int argc, char *argv[])
{
	HITRANCalculator Calculator;

	Calculator.enableWarnings();
	char *database_IO_file=argv[1];
	if(argc!=9)
	{
		print_usage(stdout);
		return 0;
	}
	if(!Calculator.open(database_IO_file))
	{
		fprintf(stderr,"could not open HITRAN database with file \"%s\"...\n",database_IO_file);
		return -1;
	}
	/*if(!strcmp(argv[2],"scs"))
	{
		if(argc!=10)
		{
			print_usage(stderr);
			return -1;
		}*/
		// cross section...
		char *molecule_name=argv[2];
		double vmr=atof(argv[3]);
		double T=atof(argv[4]);
		double p=atof(argv[5]);
		double wavelength_A=atof(argv[6]);
		double wavelength_B=atof(argv[7]);
		int no_wavelengths=atoi(argv[8]);
		// wavelength array		
		int i;
		double *wavelength=new double[2*no_wavelengths],*cross_section;
		if(!wavelength)
		{
			fprintf(stderr,"%d wavelengths allocation error!\n",no_wavelengths);
			return -1;
		}
		cross_section=wavelength+no_wavelengths;
		for(i=0;i<no_wavelengths;i++)
			wavelength[i]=wavelength_A+(double(i)*(wavelength_B-wavelength_A))/double(no_wavelengths-1);
		
		
		if(!Calculator.getCrossSection(molecule_name,vmr,T,p,wavelength,no_wavelengths,cross_section))
		{
			fprintf(stderr,"cross section extraction failed...\n");
			delete [] wavelength;
			return -1;
		}
		printf("# gas                 :	%s\n", molecule_name);
		printf("# volume mixing ratio :	%lf\n", vmr);
		printf("# temperature         :	%lfK\n", T);
		printf("# pressure            :	%lfhPa\n", p);
		for(i=0;i<no_wavelengths;i++)
			printf("%.14lg\t%.14lg\n", wavelength[i],cross_section[i]);
		delete [] wavelength;
	// }
	/*else if(!strcmp(argv[2],"gcs") || !strcmp(argv[2],"gac"))
	{
		bool compute_absorption_coefficient=false;
		if(!strcmp(argv[2],"gac"))
			compute_absorption_coefficient=true;
		if(argc!=9)
		{
			print_usage(stderr);
			return -1;
		}
		char *molecule_name=argv[3];
		// vmr stuff
		char *vmrmode=argv[4];
		double vmr;
		void *vmr_profile;
		ProfileModel_e VMRProfileModel;
		class LinearelyInterpolatedFunctionTemplate<double> VMRProfile;
		if(!strcmp(vmrmode,"cvmr"))
		{
			VMRProfileModel=PFM_constant_vmr;
			vmr=atof(argv[5]);
			vmr_profile=(void *)&vmr;
		}
		else if(!strcmp(vmrmode,"ndprf"))
		{
			if(!VMRProfile.readFromFile(argv[5]))
			{
				fprintf(stderr,"error reading number density profile \"%s\"!\n",argv[5]);
				return -1;
			}
			VMRProfileModel=PFM_number_density_profile;
			vmr_profile=(void *)&VMRProfile;
		}
		else if(!strcmp(vmrmode,"vmrprf"))
		{
			if(!VMRProfile.readFromFile(argv[5]))
			{
				fprintf(stderr,"error reading vmr profile \"%s\"!\n",argv[5]);
				return -1;
			}
			VMRProfileModel=PFM_vmr_profile;
			vmr_profile=(void *)&VMRProfile;
		}
		else
		{
			fprintf(stderr,"unknown vmr profile mode \"%s\"!\n",vmrmode);
			print_usage(stderr);
			return -1;
		}
		// wavelength array		
		double wavelength_A=atof(argv[6]);
		double wavelength_B=atof(argv[7]);
		int no_wavelengths=atoi(argv[8]);
		double *wavelength=new double[no_wavelengths];
		if(!wavelength)
		{
			fprintf(stderr,"%d wavelengths allocation error!\n",no_wavelengths);
			return -1;
		}
		int i,j;
		for(i=0;i<no_wavelengths;i++)
			wavelength[i]=wavelength_A+(double(i)*(wavelength_B-wavelength_A))/double(no_wavelengths-1);
		// here we go...
		if(!Calculator.processAtmosphericGrid(molecule_name, wavelength, no_wavelengths, VMRProfileModel, vmr_profile, compute_absorption_coefficient))
		{
			fprintf(stderr,"cross section extraction failed...\n");
			delete [] wavelength;
			return -1;
		}
		// output
		struct HITRANAtmosphericGrid_s *pHITRANAtmosphericGrid=Calculator.getHITRANAtmosphericGridPtr();
		double **cs=pHITRANAtmosphericGrid->cs;
		printf("# v wavelength height ->");
		for(i=0;i<pHITRANAtmosphericGrid->no_layers;i++)
		{
			double height=0.5*(pHITRANAtmosphericGrid->height[i]+pHITRANAtmosphericGrid->height[i+1]);
			printf("\t%.3lf",height);
		}
		fputc('\n',stdout);
		for(j=0;j<pHITRANAtmosphericGrid->no_wavelengths;j++)
		{
			printf("%.14lg",wavelength[j]);
			for(i=0;i<pHITRANAtmosphericGrid->no_layers;i++)
				printf("\t%.14lg",cs[i][j]);
			fputc('\n',stdout);
		}
		delete [] wavelength;
	}*/
	return 0;
}




