/**************************************************************************************************************
*
*       This file,
*
*                            /MapSampler.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>
#include <float.h>

#include "MapSampler.h"
#include "fileio.h"
#include "Math/prand.h"
#include "error.h"
#include "trigonometry.h"

bool MapSampler::load(const char *file_name)
{
	Matrix A;
	if(!A.load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not load file \"%s\"!\n", file_name);
		return false;
	}
	int Nx, Ny, Nxp1, Nyp1, i, j;
	Ny=A.getM()-2;
	Nx=A.getN()-2;
	Nxp1=Nx+1;
	Nyp1=Ny+1;
	class Vector cx, cy;
	class Matrix data;
	if(!cx.allocate(Nxp1)\
		|| !cy.allocate(Nyp1)\
		|| !data.allocate(Nyp1 ,Nxp1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}

	for(j=0;j<=Nx;j++)
		cx[j]=Rad(A[0][1+j]);
	double value, max_value=DBL_MIN, factor;
	for(i=0;i<=Ny;i++)
	{
		cy[i]=Rad(A[1+i][0]);
		for(j=0;j<=Nx;j++)
		{
			value=A[1+i][1+j];	
			data[i][j]=value;
			if(value>max_value)
				max_value=value;
		}
	}
	factor=pow(10.0, 1.0-ceil(log10(max_value)));
	data*=factor;
	return initializeFromMatrix(data, cx, cy);
}

bool MapSampler::initializeFromMatrix(class Matrix& data, class Vector& cx, class Vector& cy)
{
	int Nxp1=data.getN();
	int Nyp1=data.getM();
	Nx=Nxp1-1;
	Ny=Nyp1-1;
	if(!x.allocate(Nxp1) \
		|| !y.allocate(Nyp1) \
		|| !s.allocate(Nyp1 ,Nxp1) \
		|| !s0.allocate(Nyp1 ,Nxp1) \
		|| !p.allocate(Ny ,Nx) \
		|| !a.allocate(Ny ,Nx) \
		|| !b.allocate(Ny ,Nx) \
		|| !c.allocate(Ny ,Nx) \
		|| !d.allocate(Ny ,Nx))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}

	// copy data
	memcpy(x.getCoefficientPtr(), cx.getCoefficientPtr(), sizeof(double) * Nxp1); 
	memcpy(y.getCoefficientPtr(), cy.getCoefficientPtr(), sizeof(double) * Nyp1); 
	memcpy(s.getDataPtr(), data.getDataPtr(), sizeof(double) * Nxp1 * Nyp1 );

	// normalization
	if(!normalize())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	memcpy(s0.getDataPtr(), s.getDataPtr(), sizeof(double) * Nxp1 * Nyp1 );
	return true;
}

bool MapSampler::normalize()
{
	int i, j;
	// normalisation
	double area=0.0, x0, x1, y0, y1;
	double e, f, dxdy, _a, _b, _c, _d, _p;
	double A00, A01, A10, A11;
	for(i=0;i<Ny;i++)
	{
		y0=y[i];
		y1=y[i+1];
		for(j=0;j<Nx;j++)
		{
			x0=x[j];
			x1=x[j+1];
			dxdy=(x1-x0)*(y1-y0);
			
			A00=s[i][j]/dxdy;
			A01=s[i][j+1]/dxdy;
			A10=s[i+1][j]/dxdy;
			A11=s[i+1][j+1]/dxdy;

			_a=a[i][j]=A00-A10-A01+A11;
			_b=b[i][j]=y0*(A10-A11)+y1*(A01-A00);
			_c=c[i][j]=x0*(A01-A11)+x1*(A10-A00);
			_d=d[i][j]=A00*x1*y1-A10*x1*y0-A01*x0*y1+A11*x0*y0;
			
			e=0.5*_a*(x1*x1-x0*x0)+_c*(x1-x0);
			f=0.5*_b*(x1*x1-x0*x0)+_d*(x1-x0);

			_p=p[i][j]=0.5*e*(y1*y1-y0*y0)+f*(y1-y0);
			if(_p<0.0)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "negative probability\n");
				return false;
			}

			area+=_p;
		}
	}
	if(area<=0.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "negative area (norm)\n");
		return false;
	}
	for(i=0;i<Ny;i++)
		for(j=0;j<Nx;j++)
			p[i][j]/=area;
	return true;
}

bool MapSampler::getSample(double *px, double *py)
{
	int i, j;
	// double alpha=prand(), sx, sy, e, f, radicand;
	double alpha=prand();
	for(i=0;i<Ny;i++)
		for(j=0;j<Nx;j++)
			if((alpha-=p[i][j])<0.0)
			{
				double x0, x1, y0, y1, radicand;
				double N, Ne, Nf, foe, sx, sy, e, f;

				x0=x[j];
				x1=x[j+1];

				y0=y[i];
				y1=y[i+1];
				
				// sample y
				e=0.5*a[i][j]*(x1*x1-x0*x0)+c[i][j]*(x1-x0);
				f=0.5*b[i][j]*(x1*x1-x0*x0)+d[i][j]*(x1-x0);
				
				Ne=0.5*e*(y1*y1-y0*y0);
				Nf=f*(y1-y0);
				if(fabs(Ne/Nf)<1e-8)
					sy=y0+prand()*(y1-y0);
				else
				{
					N=Ne+Nf;
					foe=f/e;
					radicand=2.0*(prand()*N+f*y0)/e+y0*y0+foe*foe;
					sy=-foe+sqrt(radicand);
					if(sy<y0 || sy>y1)
					{
						sy=-foe-sqrt(radicand);
						if(sy<y0 || sy>y1)
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "error sampling y!\n");
							return false;
						}
					}
				}
				
				// sample x
				e=a[i][j]*sy+b[i][j];
				f=c[i][j]*sy+d[i][j];
				
				Ne=0.5*e*(x1*x1-x0*x0);
				Nf=f*(x1-x0);
				N=Ne+Nf;
				if(fabs(Ne/Nf)<1e-8)
					sx=x0+prand()*(x1-x0);
				else
				{
					foe=f/e;
					radicand=2.0*(prand()*N+f*x0)/e+x0*x0+foe*foe;
					sx=-foe+sqrt(radicand);
					if(sx<x0 || sx>x1)
					{
						sx=-foe-sqrt(radicand);
						if(sx<x0 || sx>x1)
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "error sampling x!\n");
							return false;
						}
					}
				}

				*px=sx;
				*py=sy;

				return true;
			}
	return false;
}




