/**************************************************************************************************************
*
*       This file,
*
*                            /Math/AssociatedLegendrePolynomial.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "AssociatedLegendrePolynomial.h"
#include "../error.h"
#include <math.h>

AssociatedLegendrePolynomial::AssociatedLegendrePolynomial()
{
	data=0;
	basis=0;
}

AssociatedLegendrePolynomial::~AssociatedLegendrePolynomial()
{
	if(data)
		delete [] data;
	if(basis)
		delete [] basis;
}

bool AssociatedLegendrePolynomial::setOrder(int n)
{
	int i, np1=n+1;
	if(data)
		delete [] data;
	data=new double[np1*np1];
	if(basis)
		delete [] basis;
	basis=new double *[np1];
	if(!data || !basis)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	for(i=0;i<=n;i++)
	{
		np1=i;
		basis[i]=data+np1*np1;
	}
	order=n;
	no_coefficients=(order+1)*(order+1);
	return true;
}

void AssociatedLegendrePolynomial::calculateBasis(const double& x)
{
	double somxx=sqrt((1.+x)*(1.-x));
	if(order==0)
		basis[0][0]=1.;
	else if(order==1)
	{
		basis[0][0]=1.;
		basis[1][0]=.5*somxx;
		basis[1][1]=x;
		basis[1][2]=-somxx;
	}
	else
	{
		int l, m, index;

		// calculate everything until l=1
		basis[0][0]=1.;
		basis[1][0]=.5*somxx;
		basis[1][1]=x;
		basis[1][2]=-somxx;

		// calculate l,m=0
		double plm20=1., plm10=x, pl0;
		int two_lm1=3, lm1=1;
//	# = processed
//	* = processing
//	o = to be done

//                             #                           
//                            ###                          
//                           oo*oo                         
//                          ooo*ooo                        
//                         oooo*oooo                       
//                        ooooo*ooooo                      

		for(l=2;l<=order;l++)
		{
			pl0=(x*(double(two_lm1))*plm10-(double(lm1))*plm20)/double(l);

			basis[l][l]=pl0;
			// increment
			plm20=plm10;
			plm10=pl0;
			two_lm1+=2;
			lm1++;
		}
		// calculate l,l=m
		double pmm=1., pmmp1, fact=1.;
		double plm2m, plm1m, plm;
		int two_mp1, lpmm1, lmm;
		two_mp1=3;
		for(m=1;m<=order;m++)
		{
//	# = processed
//	* = processing
//	o = to be done

//                             #                           
//                            ###                          
//                           oo#o*                         
//                          ooo#oo*                        
//                         oooo#ooo*                       
//                        ooooo#oooo*                      

			pmm*=-1.*fact*somxx;
			fact+=2.;
			index=two_mp1-1;
			basis[m][index]=pmm;

			if(m<order)
			{
				plm2m=pmm;

				// calculate l=m+1
//	# = processed
//	* = processing
//	o = to be done

//                             #                           
//                            ###                          
//                           oo#*#                         
//                          ooo#o*#                        
//                         oooo#oo*#                       
//                        ooooo#ooo*#                      

				pmmp1=x*(double(two_mp1))*pmm;
				basis[m+1][index+1]=pmmp1;
				plm1m=pmmp1;


//	# = processed
//	* = processing
//	o = to be done

//                             #                           
//                            ###                          
//                           oo###                         
//                          ooo#*##                        
//                         oooo#**##                       
//                        ooooo#***##                      
//                       oooooo#****##                     
				if(m<order-1)
				{
					two_lm1=2*m+3;
					lpmm1=2*m+1;
					lmm=2;
					index=2*m+2;
					for(l=m+2;l<=order;l++)
					{
						plm=(x*(double(two_lm1))*plm1m-(double(lpmm1))*plm2m)/(double(lmm));

						basis[l][index]=plm;
						// increment
						plm2m=plm1m;
						plm1m=plm;

						two_lm1+=2;
						lpmm1++;
						lmm++;
						index++;
					}
				}
			}
			two_mp1+=2;
		}
		// calculate l,-m from l,m
		double f;
		for(l=2;l<=order;l++)
		{
			f=double(-l*(l+1));
			for(m=1;m<=l;m++)
			{
				basis[l][l-m]=basis[l][l+m]/f;
				// basis[l][l-m]=f;
				f*=double(-(l+m)*(l-m+1));
			}
		}
	}
}

void AssociatedLegendrePolynomial::writeBasis(FILE *fptr)
{
	int l, m, i=0, n=1+order;
	n*=n;
	for(l=0;l<=order;l++)
	{
		for(m=-l;m<=l;m++)
		{
			i=l+m;
			fprintf(fptr, "\t%le", basis[l][i]);
		}
		fputc('\n', fptr);
	}
}




