/**************************************************************************************************************
*
*       This file,
*
*                            /Math/InterpolatedFunctionTemplate.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "InterpolatedFunctionTemplate.h"

template<class T> InterpolatedFunctionTemplate<T>::InterpolatedFunctionTemplate()
{
	no_points=0;
	yp=(T *)0;
	xp=(double *)0;
}

template<class T> InterpolatedFunctionTemplate<T>::~InterpolatedFunctionTemplate()
{
	if(yp)
	{
		delete [] yp;
		yp=(T *)0;
	}
	if(xp)
	{
		delete [] xp;
		xp=(double *)0;
	}
}

template<class T> bool InterpolatedFunctionTemplate<T>::readFromFile(char *filename, int no_rows)
{	
	if(!filename) return false;
	File file;
	if(!file.open(filename,(char *)"r"))
	{
		fprintf(stderr,"file \"%s\" either not present or not readable\n",filename);
		return false;
	}
	if(!readFromFile(file.getFILEPtr(),no_rows))
	{
		fprintf(stderr,"error reading file \"%s\". Check its content!\n",filename);
		return false;
	}
	strcpy(data_filename,filename);
	return true;
}

template<class T> bool InterpolatedFunctionTemplate<T>::writeToFile(char *filename)
{
	FILE *fptr=fopen(filename,"w");
	if(!fptr)
	{
		fprintf(stderr,"could not write to file \"%s\"!\n",filename);
		return false;
	}
	bool success=writeToFile(fptr);
	fclose(fptr);
	strcpy(data_filename,filename);
	return success;
}

template<class T> bool InterpolatedFunctionTemplate<T>::readFromFile(FILE *fptr, int no_rows)
{	
	if(no_rows<0)
		no_rows=numberOfRows(fptr);
	if(!allocate(no_rows))
	{
		fprintf(stderr,"allocation error...\n");
		return false;
	}
	char line_buffer[InterpolatedFunctionTemplate_MAX_CHARACTERS_PER_LINE], buffer[100], *charptr;
	double *pxp=xp;
	T *pyp=yp;
	for(int i=0;i<no_points;i++)
	{
		fgets(line_buffer, InterpolatedFunctionTemplate_MAX_CHARACTERS_PER_LINE, fptr);
		condenseLine(line_buffer);
		extractWordFromList(line_buffer, 0, '\t', buffer);
		*pxp++=atof(buffer);
		charptr=line_buffer;
		while(*charptr++!='\t') ;
		if(!convert(charptr, pyp++))
		{
			fprintf(stderr,"error reading line %d\n",i+1);
			return false;
		}
	}
	return true;
}

template<class T> bool InterpolatedFunctionTemplate<T>::writeToFile(FILE *fptr)
{
	if(no_points<=0)
	{
		fprintf(stderr,"linearely interpolated function undefined...\n");
		return false;
	}
	T *pyp=yp;
	double *pxp=xp;
	char buffer[InterpolatedFunctionTemplate_MAX_CHARACTERS_PER_LINE];
	for(int i=0;i<no_points;i++)
	{
		if(!convert(pyp, buffer))
		{
			fprintf(stderr,"could not convert data point %i...\n",i+1);
			return false;
		}
		fprintf(fptr,"%lg\t%s\n", *pxp++, buffer);
	}
	return true;
}

template<class T> bool InterpolatedFunctionTemplate<T>::allocate(int n)
{
	if(yp)
		delete [] yp;
	yp=new T[n];
	if(xp)
		delete [] xp;
	xp=new double[n];
	if(!yp || !xp)
		return false;
	no_points=n;
	return true;
}

