/**************************************************************************************************************
*
*       This file,
*
*                            /Math/LeastSquaresMinimizer.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "LeastSquaresMinimizer.h"
#include "../error.h"
#include "nan.h"

LeastSquaresMinimizer::LeastSquaresMinimizer()
{
	gamma=1e-300;
	minimal_convergence_rate=1e-4;
	addParameter(&gamma, type_double, "gamma", 0, "LM regularisation parameter");
	addParameter(&minimal_convergence_rate, type_double, "minimal_convergence_rate|mRc", 0, "minimal convergence rate (chi^2 percent update) at which to stop LM fit");
}

bool LeastSquaresMinimizer::setup(int n)
{
	N=n;
	int i;

	class Vector *pVector [] = {&x, &xt, &J, &d, &dnm1, &g, &gnm1};

	for(i=0;i<int(sizeof(pVector)/sizeof(Vector *));i++)	
		if(!pVector[i]->allocate(n))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}

	class Matrix *pMatrix[] = {&H, &B};

	for(i=0;i<int(sizeof(pMatrix)/sizeof(Matrix *));i++)	
		if(!pMatrix[i]->allocate(n, n))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}

	if(!defineX(x))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	return true;
}

bool LeastSquaresMinimizer::fit(enum Method_e Method, double mRc)
{
	switch(Method)
	{
		case LevenbergMarquardt:
			return LevenbergMarquardtFit(mRc);
		case ConjugateGradient:
			return ConjugateGradientFit(mRc);
		case HessianApproachLevenbergMarquardt:
			return HessianApproachLevenbergMarquardtFit(mRc);
		default:;
	}
	errorIn(__FILE__, __LINE__, __func__);
	return false;
}

bool LeastSquaresMinimizer::LevenbergMarquardtFit(double mRc)
{
	int i, j, iteration=0;
	double chi_square, chi_square_trial, gamma=this->gamma, Rc;
	if(!calculateChiSquare(x, chi_square))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	// writeState(0);
	printf("%d: chi^2=%le\n", iteration, chi_square);
LM_fit_loop:

	if(!calculateJAndH(x, chi_square, J, H))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

LM_recalculate_H:

	for(i=0;i<N;i++)
		for(j=0;j<N;j++)
		{
			if(i==j)
				B[i][j]=gamma+H[i][j];
			else
				B[i][j]=H[i][j];
		}
	if(!B.calculateInverse())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	xt = x - B * J;
	if(!calculateChiSquare(xt, chi_square_trial))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(chi_square_trial>chi_square)
	{
		gamma*=7.0;
		goto LM_recalculate_H;
	}
	gamma/=11.0;
	x.copy(xt);
	foundNewState(x, chi_square_trial, chi_square);
	Rc=(chi_square-chi_square_trial)/chi_square;

	chi_square=chi_square_trial;
	iteration++;

	printf("%d: chi^2=%le\n", iteration, chi_square);
	
	//writeState(iteration);
	if(Rc>mRc)
		goto LM_fit_loop;
	// writeState(-1);
	printf("# %d iterations\n", iteration);
	return true;
}

bool LeastSquaresMinimizer::singleLevenbergMarquardtStep(double *pgamma, const int& count)
{
	int i, j, iteration, c;
	double chi_square, chi_square_trial, gamma=*pgamma;
	for(c=0;c<count;c++)
	{
		// start
		xt=x;
		if(!calculateJAndH(x, chi_square, J, H))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		iteration=0;
		printf("%d.%d: %le\n", c+1, iteration, chi_square);
		if(ISNAN(chi_square))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "chi^2 is nan! Cancelling fit...\n");
			x=xt;
			return false;
		}
SingleLM_recalculate_H:
		iteration++;
		/*if(iteration>max_iterations)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "iterations count exceeds maximum of %d!\n", max_iterations);
			return false;
		}*/

		for(i=0;i<N;i++)
			for(j=0;j<N;j++)
			{
				if(i==j)
					B[i][j]=gamma+H[i][j];
				else
					B[i][j]=H[i][j];
			}

		if(!B.calculateInverse())
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}

		xt = x - B * J;

		if(!calculateChiSquare(xt, chi_square_trial))
		{
			errorIn(__FILE__, __LINE__, __func__);
			gamma*=2.0;
			goto SingleLM_recalculate_H;
		}

		// printf("%d (%le):\t", iteration, chi_square_trial);
		// xt.print(stdout);
		// fputc('\n', stdout);
		printf("%d.%d: %le\n", c+1, iteration, chi_square_trial);
		if(ISNAN(chi_square_trial))
		{
			calculateChiSquare(x, chi_square);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "chi^2 is nan, staying at chi^2=%le! Cancelling fit...\n", chi_square);
			return false;
		}

		if(chi_square_trial>chi_square)
		{
			gamma*=2.0;
			goto SingleLM_recalculate_H;
		}
		gamma*=0.3;
		if(gamma<1e-200)
			gamma=1e-200;

		x.copy(xt);
	}

	foundNewState(x, chi_square_trial, chi_square);

	/*fputc('\n', stdout);
	x.print(stdout);
	fputc('\n', stdout);*/
	*pgamma=gamma;

	return true;

}

bool LeastSquaresMinimizer::HessianApproachLevenbergMarquardtFit(double mRc)
{
	int i, j, iteration=0;
	bool continue_LM_loop=true, recalculate_H;
	double chi_square, chi_square_trial, gamma=this->gamma, Rc;

	// start
	if(!calculateJ(x, chi_square, J))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	H.setToUnity();
	do
	{
		recalculate_H=true;
		do
		{
			for(i=0;i<N;i++)
				for(j=0;j<N;j++)
				{
					if(i==j)
						B[i][j]=gamma+H[i][j];
					else
						B[i][j]=H[i][j];
				}
			if(!B.calculateInverse())
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			xt = x - B * J;
			if(!calculateChiSquare(xt, chi_square_trial))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			if(chi_square_trial>chi_square)
				gamma*=2.0;
			else
				recalculate_H=false;
		} while(recalculate_H);

		iteration++;

		gamma*=0.5;	// relax

		xnm1=x;
		x=xt;

		foundNewState(x, chi_square_trial, chi_square);

		Rc=(chi_square-chi_square_trial)/chi_square;
		chi_square=chi_square_trial;
		if(Rc<mRc)
			continue_LM_loop=false;
		else
		{
			// calculate Hessian approach
			Jnm1=J;
			if(!calculateJ(x, chi_square, J))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			class Vector &y=g, &s=d, &ymHs=xt;
			y = J - Jnm1;
			s = x - xnm1;
			ymHs = y - H * s;
			
			double yTs=y.scalarProductWith(s), h_ij;
			double ymHsTs=ymHs.scalarProductWith(s);
			
			for(i=0;i<N;i++)
				for(j=0;j<N;j++)
				{
					h_ij=ymHs[i]*y[j]+y[i]*ymHs[j];
					H[i][j]+=(h_ij-(ymHsTs*y[i]*y[j])/yTs)/yTs;
				}
		}
	} while(continue_LM_loop);
	return true;
}

bool LeastSquaresMinimizer::singleBFGSLMStep(double *pgamma)
{
	int i, j, iteration;
	bool recalculate_H;
	double chi_square, chi_square_trial, gamma=*pgamma;

	// start
	if(!calculateJ(x, chi_square, J))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	recalculate_H=true;
	do
	{
		iteration=1;
		for(i=0;i<N;i++)
			for(j=0;j<N;j++)
			{
				if(i==j)
					B[i][j]=gamma+H[i][j];
				else
					B[i][j]=H[i][j];
			}
		if(!B.calculateInverse())
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		xt = x;
		xt -= B * J;
		if(!calculateChiSquare(xt, chi_square_trial))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		printf("%d:\t%lg\t%lg\t%lg\n", iteration, chi_square, chi_square_trial, gamma);
		if(chi_square_trial>chi_square)
			gamma*=7.0;
		else
			recalculate_H=false;
		iteration++;
	} while(recalculate_H);


	gamma*=0.2;	// relax

	xnm1=x;
	x=xt;


	fputc('\n', stdout);
	x.print(stdout);
	fputc('\n', stdout);

	chi_square=chi_square_trial;

	// calculate Hessian approach
	Jnm1=J;
	if(!calculateJ(x, chi_square, J))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	class Vector &y=g, &s=d, &ymHs=xt;
	y = J - Jnm1;
	s = x - xnm1;
	ymHs = y ;
	ymHs -= H * s;
	
	double yTs=y.scalarProductWith(s), h_ij;
	double ymHsTs=ymHs.scalarProductWith(s);
	
	for(i=0;i<N;i++)
		for(j=0;j<N;j++)
		{
			h_ij=ymHs[i]*y[j]+y[i]*ymHs[j];
			H[i][j]+=(h_ij-(ymHsTs*y[i]*y[j])/yTs)/yTs;
		}
	*pgamma=gamma;
	return true;

}

void LeastSquaresMinimizer::resetFit()
{
	H.setToUnity();
}


bool LeastSquaresMinimizer::lineSearch(class Vector& x, const class Vector& d, double& chi_square, double& alpha)
{
	/*double alpha1=0.0, alpha2=alpha;
	double chi_square1=chi_square, chi_square2;

	xt=x+alpha2*d;
	if(!calculateChiSquare(xt, chi_square2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	bool continue_line_search=false;
	do
	{
		xt=x+alpha*d;
		if(!calculateChiSquare(xt, chi_square2))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		
		alpha2=0.5*(alpha1+alpha2);
		if(chi_square_trial>chi_square)
		else
		{
			alpha*=0.5;
			
		}
		
	}while(continue_line_search);
	return false;*/
	return false;
}

bool LeastSquaresMinimizer::ConjugateGradientFit(double mRc)
{
	double beta, beta_PR, chi_square, old_chi_square, Rc, alpha=1.0;
	if(!calculateJ(x, chi_square, J))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	d=-J;
	if(!lineSearch(x, d, chi_square, alpha))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	
	dnm1=d;
	gnm1=d;

	bool continue_CG_loop=true;
	do
	{
		if(!calculateJ(x, chi_square, J))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		old_chi_square=chi_square;
		d=-J;

		// calculate beta
		beta_PR=d.scalarProductWith(d-dnm1)/dnm1.scalarProductWith(dnm1);
		beta=max(0.0, beta_PR);

		// g=d;
		g=d+(gnm1*beta);
		if(!lineSearch(x, g, chi_square, alpha))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		Rc=(old_chi_square-chi_square)/old_chi_square;
		if(Rc<mRc)
			continue_CG_loop=false;
		else
		{
			dnm1=d;
			gnm1=g;
		}
	}while(continue_CG_loop);

	return true;
}

bool LeastSquaresMinimizer::steepestDescentStep(double *pstep_size_factor)
{
	int iteration=0;
	double chi_square, chi_square_trial;
	double step_size_factor=*pstep_size_factor;
	bool not_found_state=true;
	
	/*double chi_square;*/
	// x.print(stdout);
	if(!calculateJAndH(x, chi_square, J, H))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	do
	{
		xt.copy(x);
		xt-=J*step_size_factor;
		if(!calculateChiSquare(xt, chi_square_trial))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		iteration++;
		printf("%d:\t%lg\t%lg\t%lg\n", iteration, chi_square, chi_square_trial, step_size_factor);

		if(chi_square_trial>chi_square)
			step_size_factor*=0.123;	
		else
			not_found_state=false;
	} while(not_found_state);

	// xt.print(stdout);
	x=xt;
	fputc('\n', stdout);
	x.print(stdout);
	fputc('\n', stdout);
	
	*pstep_size_factor=step_size_factor;
	return true;
}

/*bool LeastSquaresMinimizer::writeState(int iteration)
{
	int i;
	char file_name[2000];
	sprintf(file_name, "iteration%d_yf.txt", iteration);
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\" for writing!\n", file_name);
		return false;
	}
	for(i=0;i<M;i++)
		fprintf(fptr, "%d\t%lg\t%lg\n", i+1, y[i], f[i]);
	fclose(fptr);

	sprintf(file_name, "iteration%d_xJ.txt", iteration);
	fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\" for writing!\n", file_name);
		return false;
	}
	for(i=0;i<N;i++)
		fprintf(fptr, "%d\t%lg\t%lg\n", i+1, x[i], J[i]);
	fclose(fptr);

	sprintf(file_name, "iteration%d_H.txt", iteration);
	H.save(file_name);
	return true;
}*/







