/**************************************************************************************************************
*
*       This file,
*
*                            /Math/LeastSquaresMinimizer.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _LeastSquaresMinimizer_h
#define _LeastSquaresMinimizer_h

#include "../BasicIO.h"

#include "Vector.h"
#include "Matrix.h"


class LeastSquaresMinimizer: public BasicIO
{

public:
/*

	generic least square minimizing algorithm

	chi^2(x_1, x_2, ..., x_n)
	
*/
	enum Method_e
	{
		LevenbergMarquardt,
		HessianApproachLevenbergMarquardt,
		ConjugateGradient
	};

private:

	int N;

	double gamma, minimal_convergence_rate;

protected:

	class Vector x, xnm1, xt, J, Jnm1;
	class Vector g, gnm1, d, dnm1;		// conjugate gradients
	class Matrix H, B;


	inline int getN(){return N;}

	virtual bool defineX(class Vector& x)=0;
	virtual void foundNewState(const class Vector& x, double new_chi_square, double old_chi_square){}

	virtual bool calculateChiSquare(const class Vector& x, double& chiSquare)=0;
	virtual bool calculateJAndH(const class Vector& x, double& chiSquare, class Vector& J, class Matrix& H)=0;
	virtual bool calculateJ(const class Vector& x, double& chiSquare, class Vector& J)=0;
	
	bool lineSearch(class Vector& x, const class Vector& J, double& chi_square, double& alpha);

	bool LevenbergMarquardtFit(double minimal_convergence_rate);
	bool ConjugateGradientFit(double minimal_convergence_rate);
	bool HessianApproachLevenbergMarquardtFit(double mRc);

	void resetFit();

public:
	
	inline void setGamma(double gamma){this->gamma=gamma;}
	bool setup(int n);
	
	inline void setMinimalConvergenceRate(double mRc){minimal_convergence_rate=mRc;}	

	inline bool fit(enum Method_e Method){return fit(Method, minimal_convergence_rate);}
	bool fit(enum Method_e Method, double mRc);

	bool steepestDescentStep(double *pstep_size_factor);
	bool singleBFGSLMStep(double *pgamma);
	bool singleLevenbergMarquardtStep(double *pgamma, const int& count);

	// bool writeState(int iteration);

	LeastSquaresMinimizer();
	
};

#endif // _LeastSquaresMinimizer_h




