/**************************************************************************************************************
*
*       This file,
*
*                            /Math/LegendrePolynomial.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <stdlib.h>

#include "LegendrePolynomial.h"
#include "../fileio.h"
#include "../error.h"

bool LegendrePolynomial::allocate(int n)
{
	if(!LP.allocate(n, 3) || !factor.allocate(n, 4))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	no_coefficients=n;
	calculateConstantFactors();
	return true;
}

bool LegendrePolynomial::load(char *file_name)
{
	if(!Coefficient.load(file_name) || !allocate(Coefficient.getM()))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	return true;
}

double LegendrePolynomial::evaluate(const double& x, const int& index)
{
	calculateBasisFunctions(x);
	register int D_LP=LP.getN(), D_C=Coefficient.getN();
	register double *LP_i=LP[0]+Basis, *c_i=Coefficient[0]+index;
	register double result=0.0;
	for(register int i=0;i<no_coefficients;i++)
	{
		result+=*c_i**LP_i;
		LP_i+=D_LP;
		c_i+=D_C;
	}
	return result;
}

double LegendrePolynomial::primitive(const double& x)
{
	calculateBasisPrimitives(x);
	register int D_LP=LP.getN(), D_C=Coefficient.getN();
	register double *LP_i=LP[0]+Primitive, *c_i=Coefficient[0];
	register double result=0.0;
	for(register int i=0;i<no_coefficients;i++)
	{
		result+=*c_i**LP_i;
		LP_i+=D_LP;
		c_i+=D_C;
	}
	return result;
}

void LegendrePolynomial::calculateConstantFactors()
{
	register int l;
	register int *f;
	for(l=0;l<no_coefficients;l++)
	{
		f=factor[int(l)];
		f[F_l]=l;
		f[F_lp1]=l+1;
		f[F_lp2]=l+2;
		f[F_twolp1]=2*l+1;
	}
}

void LegendrePolynomial::calculateBasisFunctions(const double& x)
{
	if(old_x==x)
		return;
	register double *LP_i=LP[0]+Basis;

	*LP_i=1.0;
	if(no_coefficients==1)
		return;

	register int N=LP.getN();
	LP_i+=N;

	*LP_i=x;
	if(no_coefficients==2)
		return;

	LP_i+=N;

	register int *f=factor[1];
	register double P_lm1=1.0, P_l=x, P_lp1;
	for(int i=2;i<no_coefficients;i++)
	{
		*LP_i=P_lp1=(P_l*x*double(f[F_twolp1])-P_lm1*double(f[F_l]))/double(f[F_lp1]);

		P_lm1=P_l;
		P_l=P_lp1;

		f+=4;
		LP_i+=N;
	}
	old_x=x;
}

void LegendrePolynomial::calculateBasisPrimitives(const double& x)
{
	if(old_x==x)
		return;
	register double *LP_i=LP[0]+Primitive;
	*LP_i=x;
	if(no_coefficients==1)
		return;

	register int N=LP.getN();
	LP_i+=N;

	*LP_i=0.5*x*x;
	if(no_coefficients==2)
		return;

	LP_i+=N;

	register int *f=factor[1];
	register double P_lm1=1.0, P_l=x, P_lp1;
	for(register int i=2;i<no_coefficients;i++)
	{
		P_lp1=(P_l*x*double(f[F_twolp1])-P_lm1*double(f[F_l]))/double(f[F_lp1]);

		*LP_i=(x*P_lp1-P_l)/double(f[F_lp2]);

		P_lm1=P_l;
		P_l=P_lp1;

		f+=4;
		LP_i+=N;
	}
	old_x=x;
}

void LegendrePolynomial::calculateBasisFunctionsAndPrimitives(const double& x)
{
	/*double *cf=basis_function,*cp=basis_primitive;
	*cf++=1.0;
	*cp++=x;
	if(no_coefficients==1)
		return;
	*cf++=x;
	*cp++=0.5*x*x;
	if(no_coefficients==2)
		return;
	double P_lm1=1.0,P_l=x,P_lp1,l;
	for(int i=2;i<no_coefficients;i++)
	{
		l=(double)(i-1);
		P_lp1=(P_l*x*(2.0*l+1.0)-P_lm1*l)/(l+1);
		*cf++=P_lp1;
		*cp++=(x*P_lp1-P_l)/(l+2.0);
		P_lm1=P_l;
		P_l=P_lp1;
	}*/

	if(old_x==x)
		return;

	register double *LP_i=LP[0];
	LP_i[Basis]=1.0;
	LP_i[Primitive]=x;

	if(no_coefficients==1)
		return;

	register int N=LP.getN();
	LP_i+=N;

	LP_i[Basis]=x;
	LP_i[Primitive]=0.5*x*x;
	if(no_coefficients==2)
		return;

	LP_i+=N;

	register int *f=factor[1];
	register double P_lm1=1.0, P_l=x, P_lp1;
	for(register int i=2;i<no_coefficients;i++)
	{
		LP_i[Basis] = P_lp1 = (P_l*x*double(f[F_twolp1])-P_lm1*double(f[F_l]))/double(f[F_lp1]);
		LP_i[Primitive] = (x*P_lp1-P_l)/double(f[F_lp2]);

		P_lm1=P_l;
		P_l=P_lp1;

		f+=4;
		LP_i+=N;
	}
	old_x=x;
}

double LegendrePolynomial::calculateNorm()
{
	return 2.0*LP[0][Basis];
}

bool LegendrePolynomial::normalize()
{
	double N=LP[0][Basis];
	if(N<=0.0)
		return false;
	LP[0][Basis]=1.0;
	return true;
}


