/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Matrix.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>
#include <float.h>

#include "Matrix.h"
#include "Vector.h"

#include "../error.h"
#include "../fileio.h"
#include "minmax.h"


Matrix::Matrix()
{
	M=0;
	N=0;
	data=0;
	c=0;	
}

Matrix::Matrix(const Matrix& a)
{
	if(allocate(a.M, a.N))
		memcpy(data, a.data, sizeof(double)*M*N);
	
	else
		errorIn(__FILE__, __LINE__, __func__);
}

Matrix::~Matrix()
{
	if(data)
		delete [] data;
	if(c)
		delete [] c;
}

bool Matrix::allocate(int m, int n)
{
	if(data)
		delete [] data;
	data=new double[m*n];
	if(c)
		delete [] c;
	c=new double *[m];
	if(!data || !c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %dx%d elements!\n", m, n);
		return false;
	}
	for(int i=0;i<m;i++)
		c[i]=data+i*n;
	M=m;
	N=n;
	return true;
}

bool Matrix::findRowIndex(const double& v, const int& column, int& i)
{
/*
	find row index i for which subsequent row elements (of column) are lower and upper boundary
*/
	/*if(v>c[i][column])
	{
		if(i>=N-1)
			return false;
		while(v>c[i][column])
		{
			i++;
			if(i==N)
				return false;
		}
		--i;
		return true;
	}
	else if(v<c[i][column])
	{
		if(i<=0)
			return false;
		while(v<c[i][column])
		{
			i--;
			if(i==0)
				return false;
		}
	}*/
	// check boundaries
	if(v<c[0][column] || v>c[N-1][column])
		return false;
	int l, m=i, h;
	double cmcolumn=c[m][column];
	if(v==cmcolumn)
		return true;
	else if(v>cmcolumn)
	{
		l=m;
		h=N-1;
	}
	else
	{
		l=0;
		h=m;
	}
	do
	{
		m = ( l + h ) / 2 ;
		cmcolumn=c[m][column];
		if(v==cmcolumn)
			return true;
		else if(v>cmcolumn)
			l=m;
		else
			h=m;
	}while(l+1!=h);
	i=l;
	return true;
}

class Vector Matrix::operator*(const class Vector& a) const
{
	int i, j;
	class Vector result;
	if(result.allocate(M) && N==a.N)
		for(i=0;i<M;i++)
		{
			double value=0.0;
			for(j=0;j<N;j++)
				value+=c[i][j]*a.c[j];
			result.c[i]=value;
		}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Matrix& Matrix::operator=(const class Matrix& a)
{
	int i, j;
	if(allocate(a.M, a.N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]=a.c[i][j];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

class Matrix& Matrix::operator+=(const class Matrix& a)
{
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
	}
	else
	{
		int i, j;
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]+=a.c[i][j];
	}
	return *this;
}

class Matrix& Matrix::operator-=(const class Matrix& a)
{
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
	}
	else
	{
		int i, j;
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]+=a.c[i][j];
	}
	return *this;
}

class Matrix Matrix::operator*(const class Matrix& a) const
{
	int i, j, k;
	class Matrix result;
	if(result.allocate(M, a.N) && a.M==N)
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
			{
				double value=0.0;
				for(k=0;k<N;k++)
					value+=c[i][k]*a.c[k][j];
				result.c[i][j]=value;
			}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Matrix Matrix::operator/(const double& q) const
{
	int i, j;
	class Matrix result;
	if(result.allocate(M, N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				result.c[i][j]=c[i][j]/q;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Matrix Matrix::operator*(const double& f) const
{
	int i, j;
	class Matrix result;
	if(result.allocate(M,N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				result.c[i][j]=c[i][j]*f;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Matrix& Matrix::operator*=(const class Matrix& a)
{
	int i, j, k;
	class Matrix result;
	if(result.allocate(M, a.N) && a.M==N)
	{
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
			{
				double value=0.0;
				for(k=0;k<N;k++)
					value+=c[i][k]*a.c[k][j];
				result.c[i][j]=value;
			}
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]=result.c[i][j];
	}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

class Matrix& Matrix::operator*=(const double& x)
{
	int i, j;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
			c[i][j]*=x;
	return *this;
}

class Matrix& Matrix::operator/=(const double& x)
{
	int i, j;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
			c[i][j]/=x;
	return *this;
}

bool Matrix::determineColumnRange(const int& column, double *pMinimum, double *pMaximum) const
{
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	int i;	
	double value;
	double minimum=DBL_MAX;
	double maximum=-DBL_MAX;
	for(i=0;i<M;i++)
	{
		value=c[i][column];
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	*pMinimum=minimum;
	*pMaximum=maximum;
	return true;
}

bool Matrix::determineRowRange(const int& row, double *pMinimum, double *pMaximum) const
{
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	int i;	
	double value;
	double minimum=DBL_MAX;
	double maximum=-DBL_MAX;
	for(i=0;i<N;i++)
	{
		value=c[row][i];
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	*pMinimum=minimum;
	*pMaximum=maximum;
	return true;
}

bool Matrix::load(const char *file_name)
{
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "lala could not load matrix from file \"%s\"!\n", file_name);
		return false;
	}
	if(!allocate(numberOfRows(fptr), numberOfColumns(fptr)))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int i, j;	//,n;
	double *data_ptr=data;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
		{
			fscanf(fptr, j<N-1?"%lg\t":"%lg\n", data_ptr++);
			/*if(fscanf(fptr, j<N-1?"%lg\t":"%lg\n", data_ptr++)!=1)
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in matrix file \"%s\", line %d, column %d!\n", file_name, i+1, j+1);
				return false;
			}*/
		}
	fclose(fptr);
	return true;
}

void Matrix::normalizeToUnityValueRange()
{
	int i, elements=M*N;
	double minimum=DBL_MAX, maximum=-DBL_MAX;
	double *pValue=data, value, C;
	for(i=0;i<elements;i++)
	{
		value=*pValue++;
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	pValue=data;
	double ami=fabs(minimum);
	double ama=fabs(maximum);
	C=max(ami, ama);
	for(i=0;i<elements;i++)
		*pValue++/=C;	
}

bool Matrix::calculateRange(double *pMinimum, double *pMaximum) const
{
	if(!data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	int i, n=M*N;	
	double value;
	double *pValue=data;
	double minimum=DBL_MAX;
	double maximum=-DBL_MAX;
	for(i=0;i<n;i++)
	{
		value=*pValue++;
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	*pMinimum=minimum;
	*pMaximum=maximum;
	return true;
}

bool Matrix::calculateInverse()
{
	if(M!=N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "not a quadratic matrix!\n");
		return false;
	}
	int i, j, s, t, el=0;
	double b, maximum, det, norm;
	int *v=new int[M];
	if(!v)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
/*
	Calculation of Hadamard Condition number: 
	determinant is divided by all geometrical row averages
*/
	norm=1.0;
	for(i=0;i<M;i++)
	{
		double sum=0.0;
		for(j=0;j<N;j++)
			sum+=c[i][j]*c[i][j];
		norm*=sqrt(sum);
	}
	det=1.0;
	for (s=0;s<M;s++)		// { s: column }
	{
		v[s]=s;
	        maximum=0.0;
	        for (t=s;t<M;t++)	//   { pivoting }
		{
			b=fabs(c[t][s]);
			if (b>maximum)
			{
				maximum=b;
				el=t;
			}
		}
	        if(maximum==0.0)
		{
			delete [] v;
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"found singular matrix!\n");
			return false;
		}
		if(el!=s)			//     { largest element at top }
		{
			for (j=0;j<M;j++)		//     { swap rows }
			{
				b=c[s][j];
				c[s][j]=c[el][j];
				c[el][j]=b;
			}
			v[s]=el;	//        { store nb. of swapped row }
			det=-det;   //    {each swap changes det-sign}
		}  // {if el<> s then ...}
	        b=1.0/c[s][s];               //    { normalize with largest element }
	        det*=c[s][s];
	        c[s][s]=1.0;
	        for(j=0;j<M;j++)
			c[s][j]*=b; //   { in row s all divide with max }
	        for(t=0;t<M;t++)
			if (t!=s)
			{ // in all other rows
				b=c[t][s];
				c[t][s]=0.0; // in column s make a 0
				for (j=0;j<M;j++)
					c[t][j]-=b*c[s][j];
	        	}
	}
	// swap back
	for (s=M-1;s>=0;s--)
	{
		if (v[s]!=s)
			for (j=0;j<M;j++)
			{
				b=c[j][s];
				c[j][s]=c[j][v[s]];
				c[j][v[s]]=b;
			}
	}
	delete [] v;
	return true;
}

void Matrix::setToUnity()
{
	int i, n=min(M, N);
	memset(data, 0, sizeof(double)*M*N);
	for(i=0;i<n;i++)
		c[i][i]=1.0;
}

class Matrix Matrix::getTransposed()
{
	class Matrix T;
	if(!T.allocate(N,M))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return T;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			T.c[j][i]=c[i][j];
	return T;
}

class Matrix Matrix::operator+(const class Matrix& a) const
{
	class Matrix result;
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
		return result;
	}
	if(!result.allocate(M, N))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return result;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			result.c[i][j]=c[i][j]+a.c[i][j];
	return result;
}

class Matrix Matrix::operator-(const class Matrix& a) const
{
	class Matrix result;
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
		return result;
	}
	if(!result.allocate(M, N))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return result;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			result.c[i][j]=c[i][j]-a.c[i][j];
	return result;
}

bool Matrix::save(const char *file_name)
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing to file \"%s\"!\n", file_name);
		return false;
	}
	dumpTo(fptr);
	fclose(fptr);
	return true;
}

void Matrix::dumpTo(FILE *fptr, bool transpose)
{
	int i, j;
	if(transpose)
		for(j=0;j<N;j++)
			for(i=0;i<M;i++)
				fprintf(fptr, i<M-1?"%.12lg\t":"%.12lg\n", c[i][j]);
	else
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				fprintf(fptr, j<N-1?"%.12lg\t":"%.12lg\n", c[i][j]);
}

bool Matrix::calculateMinimumAndMaximum(int i1, int i2, int j1, int j2, double& minimum, double& maximum) const
{
	if(!data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	if(i1<0 || i1>=M || i2<0 || i2>=M || i1>i2 || j1<0 || j1>=N || j2<0 || j2>=N || j1>j2)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "range error!\n");
		return false;
	}
	int i, j;	
	double value, *c_i;
	minimum=DBL_MAX;
	maximum=-DBL_MAX;
	for(i=i1;i<=i2;i++)
	{
		c_i=c[i];
		for(j=j1;j<=j2;j++)
		{
			value=c_i[j];
			if(value<minimum)
				minimum=value;
			if(value>maximum)
				maximum=value;
		}
	}
	return true;
}

