/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Matrix.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Matrix_h
#define _Matrix_h

#include <memory.h>
#include <stdio.h>
#include "../error.h"

class Vector;

class Matrix
{
		friend class Vector;

		int M, N;
		double *data, **c;

	public:
		
		Matrix();
		Matrix(const class Matrix& a);
		~Matrix();

		bool allocate(int m, int n);
		inline bool isValid() const {return data && c;}

		bool load(const char *file_name);
		bool save(const char *file_name);

		void dumpTo(FILE *fptr, bool transpose=false);
		
		inline double *getDataPtr(){return data;}
		inline double *getCoefficientPtr(){return data;}
		inline double **getRowPtr(){return c;}
		inline double *getRowPtr(int i){return c[i];}

		bool calculateInverse();
		class Matrix getTransposed();
		bool calculateRange(double *pMinimum, double *pMaximum) const;
		bool calculateMinimumAndMaximum(int i1, int i2, int j1, int j2, double& minimum, double& maximum) const;

		bool findRowIndex(const double& v, const int& column, int& i);

		inline int getM() const {return M;}
		inline int getN() const {return N;}
		inline int getNumberOfElements() const {return M*N;}

		void normalizeToUnityValueRange();

		bool determineColumnRange(const int& column, double *pMinimum, double *pMaximum) const;
		bool determineRowRange(const int& row, double *pMinimum, double *pMaximum) const;

		inline void setToZero() {memset(data, 0, sizeof(double)*M*N);}
		inline void setTo(const double& value) {for(register int i=0;i<M*N;i++) data[i]=value;}
		void setToUnity();
		
		// inline double& operator[](const int& i) const {return c[i];}
		inline double *operator[](const int& i) const
		{
			if(i<0 || i>=M)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "matrix row index %d!\n", i);
			}
			return c[i];
		}

		class Matrix operator+(const class Matrix& a) const;
		class Matrix operator-(const class Matrix& a) const;

		class Matrix& operator=(const class Matrix& a);
		class Matrix& operator+=(const class Matrix& a);
		class Matrix& operator-=(const class Matrix& a);

		class Matrix operator*(const class Matrix& a) const;
		class Vector operator*(const class Vector& a) const;
		class Matrix operator*(const double& f) const;
		class Matrix operator/(const double& q) const;

		class Matrix& operator*=(const class Matrix& a);
		class Matrix& operator*=(const double& x);

		class Matrix& operator/=(const double& x);
};

#endif //_Matrix_h




