/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Matrix3D.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "Matrix3D.h"
#include "Vector3D.h"

#include "minmax.h"

Matrix3D::Matrix3D()
{
/*	a[0][0]=0.0;		a[0][1]=0.0;		a[0][2]=0.0; 
	a[1][0]=0.0;		a[1][1]=0.0;		a[1][2]=0.0; 
	a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=0.0; */
}

Matrix3D::Matrix3D(const class Vector3D& a0, const class Vector3D& a1, const class Vector3D& a2)
{
	int i;
	for(i=0;i<3;i++)
	{
		a[i][0]=a0.x[i];
		a[i][1]=a1.x[i];
		a[i][2]=a2.x[i];
	}
}

Matrix3D::Matrix3D(const class Vector3D& EulerAngles)
{
	// Euler matrix: R_z,gamma * R_y,beta * R_x,alpha
	double c_alpha=cos(EulerAngles.x[0]);
	double s_alpha=sin(EulerAngles.x[0]);

	double c_beta=cos(EulerAngles.x[1]);
	double s_beta=sin(EulerAngles.x[1]);

	double c_gamma=cos(EulerAngles.x[2]);
	double s_gamma=sin(EulerAngles.x[2]);

	a[0][0]=c_gamma*c_beta;
	a[0][1]=c_gamma*s_beta*s_alpha-s_gamma*c_alpha;
	a[0][2]=c_gamma*s_beta*c_alpha+s_gamma*s_alpha;

	a[1][0]=s_gamma*c_beta;
	a[1][1]=s_gamma*s_beta*s_alpha+c_gamma*c_alpha;
	a[1][2]=s_gamma*s_beta*c_alpha-c_gamma*s_alpha;

	a[2][0]=-s_beta;
	a[2][1]=c_beta*s_alpha;
	a[2][2]=c_beta*c_alpha;
}

Matrix3D::Matrix3D(class Vector3D *base_vector)
{
	int i;
	for(i=0;i<3;i++)
	{
		a[0][i]=base_vector[i].x[0];
		a[1][i]=base_vector[i].x[1];
		a[2][i]=base_vector[i].x[2];
	}
}

Matrix3D::Matrix3D(enum Matrix3D::InitType_e type, const double &val)
{
	int i;
	switch(type)
	{
		case UNITY:			
		{
			a[0][0]=1.0;		a[0][1]=0.0;		a[0][2]=0.0; 
			a[1][0]=0.0;		a[1][1]=1.0;		a[1][2]=0.0; 
			a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=1.0; 
			break;
		}
		case VAL:			
		{
			a[0][0]=val;		a[0][1]=0.0;		a[0][2]=0.0; 
			a[1][0]=0.0;		a[1][1]=val;		a[1][2]=0.0; 
			a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=val; 
			break;
		}
		case ROTX:			
		{
			a[0][0]=1.0;		a[0][1]=0.0;		a[0][2]=0.0; 
			a[1][0]=0.0;		a[1][1]=cos(val);	a[1][2]=-sin(val); 
			a[2][0]=0.0;		a[2][1]=sin(val);	a[2][2]=cos(val); 
			break;
		}
		case ROTY:			
		{
			a[0][0]=cos(val);	a[0][1]=0.0;		a[0][2]=sin(val); 
			a[1][0]=0.0;		a[1][1]=1.0;		a[1][2]=0.0; 
			a[2][0]=-sin(val);	a[2][1]=0.0;		a[2][2]=cos(val); 		
			break;
		}
		case ROTZ:			
		{
			a[0][0]=cos(val);	a[0][1]=-sin(val);	a[0][2]=0.0; 
			a[1][0]=sin(val);	a[1][1]=cos(val);	a[1][2]=0.0; 
			a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=1.0; 					
			break;
		}
		default:  //and ZERO
		for(i=0;i<3;i++)
		{
			a[0][0]=0.0;		a[0][1]=0.0;		a[0][2]=0.0; 
			a[1][0]=0.0;		a[1][1]=0.0;		a[1][2]=0.0; 
			a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=0.0; 
		}			
	}
}

class Matrix3D axisRotation(const class Vector3D& v, const double& alpha)
{	
	class Matrix3D m;
	double cos_alpha=cos(alpha), sin_alpha=sin(alpha), one_minus_cos_alpha=1.0-cos_alpha;

	m.a[0][0]=cos_alpha+v.x[0]*v.x[0]*one_minus_cos_alpha;
	m.a[0][1]=v.x[0]*v.x[1]*one_minus_cos_alpha-v.x[2]*sin_alpha;
	m.a[0][2]=v.x[0]*v.x[2]*one_minus_cos_alpha+v.x[1]*sin_alpha;

	m.a[1][0]=v.x[0]*v.x[1]*one_minus_cos_alpha+v.x[2]*sin_alpha;
	m.a[1][1]=cos_alpha+v.x[1]*v.x[1]*one_minus_cos_alpha;
	m.a[1][2]=v.x[1]*v.x[2]*one_minus_cos_alpha-v.x[0]*sin_alpha;

	m.a[2][0]=v.x[2]*v.x[0]*one_minus_cos_alpha-v.x[1]*sin_alpha;
	m.a[2][1]=v.x[2]*v.x[1]*one_minus_cos_alpha+v.x[0]*sin_alpha;
	m.a[2][2]=cos_alpha+v.x[2]*v.x[2]*one_minus_cos_alpha;

	return m;
}

class Vector3D Matrix3D::operator*(const class Vector3D& v)
{
	register int i, j;
	register double sum;
	double result[3];
	for(i=0;i<3;i++)
	{
		sum=0.0;
		for(j=0;j<3;j++)
			sum+=a[i][j]*v.x[j];
		result[i]=sum;
	}
	return Vector3D(result[0], result[1], result[2]);
}

double Matrix3D::determinante() const
{
	//Cramer's rule: develop with first row
	register double sum=0.0, da, dp;
	register int i, p1, p2;
	for(i=0;i<3;i++)
	{
		p1=(i+1)%3;
		p2=(i+2)%3;
		dp=a[1][p1]*a[2][p2];
		da=a[1][p2]*a[2][p1];
		sum+=a[0][i]*(dp-da);		
	}
	return sum;
}

void Matrix3D::internalCalculateInverse(class Matrix3D& im, const double& d) const
{
	register int i, j;
	int ki,gi,kj,gj,ip1,ip2,jp1,jp2;
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
		{
			ip1=(i+1)%3;
			ip2=(i+2)%3;
			jp1=(j+1)%3;
			jp2=(j+2)%3;
			ki=min(ip1,ip2);
			gi=max(ip1,ip2);
			kj=min(jp1,jp2);
			gj=max(jp1,jp2);
			im.a[j][i]=a[gi][kj]*a[ki][gj]-a[ki][kj]*a[gi][gj];
			if((i+j)%2)
				im.a[j][i]/=d;
			else	
				im.a[j][i]/=-d;
		}
}

class Matrix3D Matrix3D::inverse() const
{	
	class Matrix3D im;
	double d=determinante();
	if(d==0.0)
		return Matrix3D(Matrix3D::ZERO,0.0);
	internalCalculateInverse(im, d);
	return im;
}

class Matrix3D Matrix3D::transpose() const
{
	class Matrix3D t;
	int i,j;
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
			t.a[i][j]=a[j][i];
	return t;
}


bool Matrix3D::calculateInverse(class Matrix3D& im) const
{
	double d=determinante();
	if(d==0.0)
		return false;
	internalCalculateInverse(im, d);
	return true;
}

void Matrix3D::initializeWithSphericalBasisInColumns(const class Vector3D& p)
{
/*
	construct right hand coordinate system:
	
		dx -> dtheta
		dy -> dphi
		dz -> dr
*/
	double r, theta, phi;
	p.toPolar(&r, &theta, &phi);

	double cos_phi=cos(phi);
	double sin_phi=sin(phi);
	double cos_theta=cos(theta);
	double sin_theta=sin(theta);

	// dtheta		dphi (normalised)	dr
	a[0][0]=cos_phi*cos_theta;	a[0][1]=-sin_phi;	a[0][2]=cos_phi*sin_theta;
	a[1][0]=sin_phi*cos_theta;	a[1][1]=cos_phi;	a[1][2]=sin_phi*sin_theta;
	a[2][0]=-sin_theta;		a[2][1]=0.0;		a[2][2]=cos_theta;
	
/*
	CS transformation:
	row coordinates vector v * CS matrix = column coordinates vector
*/
}

class Matrix3D Matrix3D::operator*(const class Matrix3D& m) const
{
	double sum;
	int i, j , k;
	class Matrix3D result;
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
		{
			sum=0.0;
			for(k=0;k<3;k++)
				sum+=a[i][k]*m.a[k][j];
			result.a[i][j]=sum;
		}
	return result;
}

class Matrix3D& Matrix3D::operator*=(const class Matrix3D& m)
{
	double sum;
	int i, j , k;
	class Matrix3D result;
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
		{
			sum=0.0;
			for(k=0;k<3;k++)
				sum+=a[i][k]*m.a[k][j];
			result.a[i][j]=sum;
		}
	// copy back
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
			a[i][j]=result.a[i][j];
	return *this;
}


