/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Polynomial.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Polynomial.h"
#include "../error.h"
#include "../fileio.h"

#include "Matrix.h"
#include "Vector.h"

Polynomial::Polynomial()
{
	coefficient=0;
}

Polynomial::~Polynomial()
{
	if(coefficient)
		delete [] coefficient;
}

bool Polynomial::allocate(int n)
{
	if(coefficient)
		delete [] coefficient;
	coefficient=new double[n];
	if(!coefficient)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating polynomial with %d coefficients!\n", n);
		return false;
	}
	no_coefficients=n;
	return true;
}

double Polynomial::evaluate(const double& x) const
{
	register int i=no_coefficients-1;
	register double y=coefficient[i--];
	for(;i>=0;i--)
		y+=coefficient[i]+x*y;
	return y;
}

bool Polynomial::load(char *file_name)
{
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not load polynomial from file \"%s\"!\n", file_name);
		return false;
	}
	if(!allocate(numberOfRows(fptr)))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating polynomial according to \"%s\"!\n", file_name);
		return false;
	}
	for(int i=0;i<no_coefficients;i++)
		if(fscanf(fptr, "%lf\n", coefficient+i)!=1)
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error reading polynomial coefficient %d from line %d of file \"%s\"!\n", i, i+1, file_name);
			return false;
		}
	fclose(fptr);	
	return true;
}

bool Polynomial::save(char *file_name)
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not save polynomial to file \"%s\"!\n", file_name);
		return false;
	}
	save(fptr);
	fclose(fptr);
	return true;
}

void Polynomial::save(FILE *fptr)
{
	for(int i=0;i<no_coefficients;i++)
		fprintf(fptr, "%lg\n", coefficient[i]);
}

bool Polynomial::fit(double *x, double *y, const int& M)
{
	class Matrix A;
	class Vector b;
	if(!A.allocate(no_coefficients, no_coefficients) || !b.allocate(no_coefficients))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	A.setToZero();
	b.setToZero();
	register int m, i, j;
	register double xm, xi, xj, ym, sum=0.0;
	for(m=0;m<M;m++)
	{
		xm=x[m];
		ym=y[m];
		xi=1.0;
		for(i=0;i<no_coefficients;i++)
		{
			xj=xi;
			for(j=0;j<no_coefficients;j++)
			{
				A[int(i)][j]+=xj;
				xj*=xm;
			}
			b[int(i)]+=xi*ym;
			xi*=xm;
		}
	}
	if(!A.calculateInverse())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "fit failed due to singular matrix\n");
		return false;
	}
	for(i=0;i<no_coefficients;i++)
	{
		sum=0.0;
		for(j=0;j<no_coefficients;j++)
			sum+=A[int(i)][j]*b[int(j)];
		coefficient[i]=sum;
	}
	return true;
}

