/**************************************************************************************************************
*
*       This file,
*
*                            /Math/SphericalHarmonicsExpansion.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "SphericalHarmonicsExpansion.h"
#include "../error.h"

SphericalHarmonicsExpansion::SphericalHarmonicsExpansion()
{
	alp_norm=0;
	alp_norm_data=0;
	coefficient_data=0;
	c=0;
}

SphericalHarmonicsExpansion::~SphericalHarmonicsExpansion()
{
	if(alp_norm)
		delete [] alp_norm;
	if(alp_norm_data)
		delete [] alp_norm_data;
	if(coefficient_data)
		delete [] coefficient_data;
	if(c)
		delete [] c;
}

bool SphericalHarmonicsExpansion::setOrder(int order)
{
	if(!AssociatedLegendrePolynomial::setOrder(order))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int two_lp1, l,m;
	if(alp_norm_data)
		delete [] alp_norm_data;
	alp_norm_data=new double[getNOCoefficients()];
	if(alp_norm)
		delete [] alp_norm;
	alp_norm=new double *[getOrder()+1];
	if(c)
		delete [] c;
	c=new complex<double> *[order+1];
	if(coefficient_data)
		delete [] coefficient_data;
	coefficient_data=new complex<double>[getNOCoefficients()+2*order+1];
	if(!alp_norm || !alp_norm_data || !coefficient_data || !c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// assign pointers and calculate norms
	azimuth_component=coefficient_data+getNOCoefficients();

	complex<double> *ptr=coefficient_data;
	double *double_ptr=alp_norm_data, f, double_two_lp1_over_4pi;
	two_lp1=1;
	for(l=0;l<=order;l++)
	{
		// first assign pointers
		c[l]=ptr;
		alp_norm[l]=double_ptr;

		// calculate alp norms
		double_two_lp1_over_4pi=(double(two_lp1))/(4.0*M_PI);
		for(m=0;m<=l;m++)
		{
			if(m)
			{
				// m not zero
				double_ptr[l+m]=sqrt(double_two_lp1_over_4pi*f);
				double_ptr[l-m]=sqrt(double_two_lp1_over_4pi/f);
			}
			else
			{
				// m  is zero
				double_ptr[l]=sqrt(double_two_lp1_over_4pi); // alp_norm[l];
				f=1.0;
			}
			f*=double((l+m)*(l-m+1));
		}

		// increment
		ptr+=two_lp1;
		double_ptr+=two_lp1;
		two_lp1+=2;
		
	}
	return true;
}

void SphericalHarmonicsExpansion::resetCoefficients()
{
	memset(coefficient_data, 0, sizeof(complex<double>)*getNOCoefficients());
}

void SphericalHarmonicsExpansion::calculateAzimuthComponents(const double& phi)
{
	int m;
	complex<double> exp_i_m_phi;
	azimuth_component[getOrder()]=1.0;
	for(m=1;m<=getOrder();m++)
	{
		exp_i_m_phi=exp(complex<double>(0.0, 1.0)*(double(m))*phi);
		azimuth_component[getOrder()+m]=exp_i_m_phi;	
		azimuth_component[getOrder()-m]=1.0/exp_i_m_phi;	
	}
}

complex<double> SphericalHarmonicsExpansion::evaluate(const double& theta, const double& phi)
{
	int l, m;
	complex<double> result=0.0;
	complex<double> *c_ptr=coefficient_data;
	complex<double> *azimuth_component_l;
	double *alp_ptr=AssociatedLegendrePolynomial::getDataPtr();
	double *norm_ptr=alp_norm_data, f=1.0;
	// -------------------
	AssociatedLegendrePolynomial::calculateBasis(cos(theta));
	calculateAzimuthComponents(phi);
	for(l=0;l<=getOrder();l++)
	{
		azimuth_component_l=azimuth_component+l;
		for(m=-l;m<=l;m++)
		{
			result+=(*c_ptr)*(*alp_ptr)*(*norm_ptr)*azimuth_component_l[m];
			// increment
			c_ptr++;
			alp_ptr++;
			norm_ptr++;
		}
	}
	return f*result;
}

