/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Spline.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <stdio.h> 

#include "Spline.h"
#include "../error.h"

bool Spline::allocate(int n)
{
	if(!xi.allocate(n) || !s.allocate(n, 4))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %d bins\n", n);
		return false;
	}
	no_bins=n;
	last_bin=no_bins-1;
	return true;
}

bool Spline::load(char *file_name)
{
	class Matrix M;
	if(!M.load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(M.getN()!=5)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file \"%s\" needs to have 5 columns\n", file_name);
		return false;
	}
	no_bins=M.getM();
	if(!allocate(no_bins))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int i, j;
	for(i=0;i<no_bins;i++)
	{
		xi[i]=M[i][0];
		for(j=0;j<4;j++)
			s[i][j]=M[i][1+j];
	}
	return true;
}

double Spline::evaluate(const double& x) const
{
	int index;
	if(x<xi[0])
		index=0;
	else if(x>xi[last_bin])
		index=last_bin;
	else
		for(index=0;index<last_bin;index++)
			if(x>=xi[index] && x<xi[index+1])
				break;
	// got index evaluate polynomial
	register double *c=s[index]+3;
	register double y=*c, d=x-xi[index];
	y+=*(--c)+d*y;
	y+=*(--c)+d*y;
	y+=*(--c)+d*y;
	return y;
}

int Spline::getBinOf(const double& x) const
{
	int index;
	if(x<xi[0])
		index=0;
	else if(x>=xi[last_bin])
		index=last_bin;
	else
		for(index=0;index<last_bin;index++)
			if(x>=xi[index] && x<xi[index+1])
				break;
	return index;
}

void Spline::save(FILE *fptr)
{
	int i, n=s.getM();
	for(i=0;i<n;i++)
		fprintf(fptr,"%lg\t%lg\t%lg\t%lg\t%lg\n", xi[i], s[i][0], s[i][1], s[i][2], s[i][3]);
}

bool Spline::save(char *file_name)
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing spline data to file \"%s\"!\n", file_name);
		return false;
	}
	save(fptr);
	fclose(fptr);
	return true;
}


