/**************************************************************************************************************
*
*       This file,
*
*                            /Math/SplineFit.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "SplineFit.h"
#include "../error.h"


SplineFit::SplineFit()
{
	lambda_a=1.0;
	lambda_b=1.0;
	constraint_strength=0.0;
}

bool SplineFit::loadDataPoints(char *file_name, int g)
{
	int i, j, n, index;
	Matrix DataMatrix;
	if(!DataMatrix.load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading points from \"%s\"\n", file_name);
		return false;
	}
	if(DataMatrix.getN()!=2)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file \"%s\" has wrong format\n", file_name);
		return false;
	}
	if(!data.allocate(DataMatrix.getM(), 3))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	for(i=0;i<DataMatrix.getM();i++)
	{
		for(j=0;j<2;j++)
			data[i][j]=DataMatrix[i][j];
		data[i][2]=0.0;
	}
	n=DataMatrix.getM()/g-1;

	if(!Spline::allocate(n) || !lambda.allocate(n-1, 3))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	// s matrix
	for(i=0;i<n;i++)
	{
		//index=i*delta;
		index=i*g;
		getXi(i)=data[index][0];
		// getS(i)[0]=data[index][1];
		getS(i)[0]=1.0;
	}
	getXi(i)=data[data.getM()-1][0]+1;
	// getS(i)[0]=data[data.getM()-1][1];
	getS(i)[0]=1.0;
	/*for(i=0;i<n;i++)
		getS(i)[1]=(getS(i+1)[0]-getS(i)[0])/(getXi(i+1)-getXi(i));*/

	// lambdas: constraint strength
	lambda_a=constraint_strength;
	for(i=0;i<n-1;i++)
	{
		lambda[i][0]=constraint_strength;
		lambda[i][1]=constraint_strength;
		lambda[i][2]=constraint_strength;
	}
	lambda_b=constraint_strength;
	return setup(4*Spline::getNOBins());
}

bool SplineFit::defineX(class Vector& x)
{
	int i, j, index=0;
	for(i=0;i<Spline::getNOBins();i++)
	{
		for(j=0;j<4;j++)
			x[index++]=Spline::getS(i)[j];
	}
	return true;
}

/*bool SplineFit::defineY(class Vector& y, class Vector& ye)
{
	for(int i=0;i<data.getM();i++)
	{
		y[i]=data[i][1];
		// ye[i]=0.01*y[i];
		ye[i]=1.0;
	}
	return true;
}*/

void SplineFit::setX(const class Vector& x)
{
	int i, j, index=0;
	for(i=0;i<Spline::getNOBins();i++)
	{
		for(j=0;j<4;j++)
			Spline::getS(i)[j]=x[index++];
	}
	lambda_b=x[index++];
}

/*bool SplineFit::evaluateF(const class Vector& x, class Vector& f)
{
	setX(x);
	for(int i=0;i<data.getM();i++)
		f[i]=Spline::evaluate(data[i][0]);
	return true;
}*/


bool SplineFit::calculateChiSquare(const class Vector& x, double& chiSquare)
{
	int i;
	double xi_i, xi_ip1, d, f;
	double dxi_1, dxi_2, dxi_3;
	register double sum=0.0;

	setX(x);

	for(i=0;i<data.getM();i++)
	{
		data[i][2]=f=Spline::evaluate(data[i][0]);
		d=f-data[i][1];	// no error considered here!!!
		sum+=d*d;
	}

	// constraints

	for(i=0;i<getNOBins()-1;i++)
	{
		xi_i=getXi(i);
		xi_ip1=getXi(i+1);

		// dxi
		dxi_1=xi_ip1-xi_i;
		dxi_2=dxi_1*dxi_1;
		dxi_3=dxi_2*dxi_1;

		sum+=lambda[i][0]*pow(s[i][0]+s[i][1]*dxi_1+s[i][2]*dxi_2+s[i][3]*dxi_3-s[i+1][0], 2.0);
		sum+=lambda[i][1]*pow(s[i][1]+2.0*s[i][2]*dxi_1+3.0*s[i][3]*dxi_2-s[i+1][1], 2.0);
		sum+=lambda[i][2]*pow(2.0*s[i][2]+6.0*s[i][3]*dxi_1-2.0*s[i+1][2], 2.0);
	}
	
	// edge constraints: natural spline
	sum+=lambda_a*pow(2.0*s[0][2], 2.0);
	sum+=lambda_b*pow(2.0*s[getNOBins()-1][2]+6.0*s[getNOBins()-1][3], 2.0);
	chiSquare=sum;
	return true;
}


bool SplineFit::calculateJAndH(const class Vector& x, double& chiSquare, class Vector& J, class Matrix& H)
{
	/*if(!evaluateF(x, f))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}*/
	int i, j, index, row_index, column_index;
	double xi_i, xi_ip1, _x, dy, c;
	double dx_1, dx_2, dx_3, dx_4, dx_5, dx_6;
	double dxi_1, dxi_2, dxi_3, dxi_4, dxi_5, dxi_6;
	double dxi_1_km1, dxi_2_km1, dxi_3_km1, dxi_4_km1, dxi_5_km1, dxi_6_km1;
	double C_k0, C_k1, C_k2, C_b;
	double C_km10, C_km11, C_km12;
	double l_k0, l_k1, l_k2;
	double l_km10, l_km11, l_km12;
	double s_k0, s_k1, s_k2, s_k3;

	if(!calculateChiSquare(x, chiSquare))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// calculate Jacobian J and Hessian H
	J.setToZero();
	H.setToZero();
	for(i=0;i<Spline::getNOBins()-1;i++)
	{
		index=4*i;
		xi_i=getXi(i);
		xi_ip1=getXi(i+1);
		for(j=0;j<data.getM();j++)
		{
			_x=data[j][0];
			if(_x>=xi_i && _x<xi_ip1)
			{
				dx_1=_x-xi_i;
				dx_2=dx_1*dx_1;
				dx_3=dx_2*dx_1;
				dx_4=dx_2*dx_2;
				dx_5=dx_3*dx_2;
				dx_6=dx_3*dx_3;

				dy=data[j][2]-data[j][1];

				c=2.0*dy;

				J[index]+=c;
				H[index][index]+=2.0;
				H[index][index+1]+=2.0*dx_1;
				H[index][index+2]+=2.0*dx_2;
				H[index][index+3]+=2.0*dx_3;

				J[index+1]+=c*dx_1;
				H[index+1][index]+=2.0*dx_1;
				H[index+1][index+1]+=2.0*dx_2;
				H[index+1][index+2]+=2.0*dx_3;
				H[index+1][index+3]+=2.0*dx_4;

				J[index+2]+=c*dx_2;
				H[index+2][index]+=2.0*dx_2;
				H[index+2][index+1]+=2.0*dx_3;
				H[index+2][index+2]+=2.0*dx_4;
				H[index+2][index+3]+=2.0*dx_5;

				J[index+3]+=c*dx_3;
				H[index+3][index]+=2.0*dx_3;
				H[index+3][index+1]+=2.0*dx_4;
				H[index+3][index+2]+=2.0*dx_5;
				H[index+3][index+3]+=2.0*dx_6;
			}
		}

		// dxi
		dxi_1=xi_ip1-xi_i;
		dxi_2=dxi_1*dxi_1;
		dxi_3=dxi_2*dxi_1;
		dxi_4=dxi_2*dxi_2;
		dxi_5=dxi_3*dxi_2;
		dxi_6=dxi_3*dxi_3;
		s_k0=s[i][0];
		s_k1=s[i][1];
		s_k2=s[i][2];
		s_k3=s[i][3];
		
		if(i<Spline::getNOBins()-2)
		{
			l_k0=lambda[i][0];
			l_k1=lambda[i][1];
			l_k2=lambda[i][2];
			
			C_k0=s_k0+s_k1*dxi_1+s_k2*dxi_2+s_k3*dxi_3-s[i+1][0];
			C_k1=s_k1+2.0*s_k2*dxi_1+3.0*s_k3*dxi_2-s[i+1][1];
			C_k2=2.0*s_k2+6.0*s_k3*dxi_1-2.0*s[i+1][2];

			J[index]+=2.0*l_k0*C_k0;
			H[index][index]+=2.0*l_k0;
			H[index][index+1]+=2.0*l_k0*dxi_1;
			H[index][index+2]+=2.0*l_k0*dxi_2;
			H[index][index+3]+=2.0*l_k0*dxi_3;
			H[index][index+4]+=-2.0*l_k0;

			J[index+1]+=2.0*l_k0*C_k0*dxi_1+2.0*l_k1*C_k1;
			H[index+1][index]+=2.0*l_k0*dxi_1;
			H[index+1][index+1]+=2.0*l_k0*dxi_2+2.0*l_k1;
			H[index+1][index+2]+=2.0*l_k0*dxi_3+4.0*l_k1*dxi_1;
			H[index+1][index+3]+=2.0*l_k0*dxi_4+6.0*l_k1*dxi_2;
			H[index+1][index+4]+=-2.0*l_k0*dxi_1;
			H[index+1][index+5]+=-2.0*l_k1;

			J[index+2]+=2.0*l_k0*C_k0*dxi_2+4.0*l_k1*C_k1*dxi_1+4.0*l_k2*C_k2;
			H[index+2][index]+=2.0*l_k0*dxi_2;
			H[index+2][index+1]+=2.0*l_k0*dxi_3+4.0*l_k1*dxi_1;
			H[index+2][index+2]+=2.0*l_k0*dxi_4+8.0*l_k1*dxi_2+8.0*l_k2;
			H[index+2][index+3]+=2.0*l_k0*dxi_5+12.0*l_k1*dxi_3+24.0*l_k2*dxi_1;
			H[index+2][index+4]+=-2.0*l_k0*dxi_2;
			H[index+2][index+5]+=-4.0*l_k1*dxi_1;
			H[index+2][index+6]+=-8.0*l_k2;

			J[index+3]+=2.0*l_k0*C_k0*dxi_3+6.0*l_k1*C_k1*dxi_2+12.0*l_k2*C_k2*dxi_1;
			H[index+3][index]+=2.0*l_k0*dxi_3;
			H[index+3][index+1]+=2.0*l_k0*dxi_4+6.0*l_k1*dxi_2;
			H[index+3][index+2]+=2.0*l_k0*dxi_5+12.0*l_k1*dxi_3+24.0*l_k2*dxi_1;
			H[index+3][index+3]+=2.0*l_k0*dxi_6+18.0*l_k1*dxi_4+72.0*l_k2*dxi_2;
			H[index+3][index+4]+=-2.0*l_k0*dxi_3;
			H[index+3][index+5]+=-6.0*l_k1*dxi_2;
			H[index+3][index+6]+=-24.0*l_k2*dxi_1;

			if(i>0)
			{
				l_km10=lambda[i-1][0];
				l_km11=lambda[i-1][1];
				l_km12=lambda[i-1][2];

				J[index]-=2.0*l_km10*C_km10;
				H[index][index-4]-=2.0*l_km10;
				H[index][index-3]-=2.0*l_km10*dxi_1_km1;
				H[index][index-2]-=2.0*l_km10*dxi_2_km1;
				H[index][index-1]-=2.0*l_km10*dxi_3_km1;
				H[index][index]+=2.0*l_km10;

				J[index+1]-=2.0*l_km11*C_km11;
				H[index+1][index-3]-=2.0*l_km11;
				H[index+1][index-2]-=4.0*l_km11*dxi_1_km1;
				H[index+1][index-1]-=6.0*l_km11*dxi_2_km1;
				H[index+1][index+1]-=2.0*l_km11;

				J[index+2]-=4.0*l_km12*C_km12;
				H[index+2][index-2]-=8.0*l_km12;
				H[index+2][index-1]-=24.0*l_km12*dxi_1_km1;
			}
			else
			{

				J[2]+=8.0*lambda_a*s_k2;
				H[2][2]+=8.0*lambda_a;
			}
		}
		else
		{
			C_b=2.0*s_k2+6.0*s_k3*dxi_1;

			J[index+2]+=4.0*lambda_b*C_b;
			H[index+2][index+2]+=8.0*lambda_b;
			H[index+2][index+3]+=24.0*lambda_b*dxi_1;

			J[index+3]+=12.0*lambda_b*C_b*dxi_1;
			H[index+3][index+2]+=24.0*lambda_b*dxi_1;
			H[index+3][index+2]+=72.0*lambda_b*dxi_2;
		}
		C_km10=C_k0;
		C_km11=C_k1;
		C_km12=C_k2;
		dxi_1_km1=dxi_1;
		dxi_2_km1=dxi_2;
		dxi_3_km1=dxi_3;
		dxi_4_km1=dxi_4;
		dxi_5_km1=dxi_5;
		dxi_6_km1=dxi_6;
	}
	
	// J*=-1.0;
	return true;
}

/*double SplineFit::calculateChiSquareSmoothnessConstraints()
{

	int i;
	double sum=0.0;
	double xi_i, xi_ip1;
	double dxi_1, dxi_2, dxi_3;
	for(i=0;i<getNOBins()-1;i++)
	{
		xi_i=getXi(i);
		xi_ip1=getXi(i+1);

		// dxi
		dxi_1=xi_ip1-xi_i;
		dxi_2=dxi_1*dxi_1;
		dxi_3=dxi_2*dxi_1;

		sum+=lambda[i][0]*pow(s[i][0]+s[i][1]*dxi_1+s[i][2]*dxi_2+s[i][3]*dxi_3-s[i+1][0], 2.0);
		sum+=lambda[i][1]*pow(s[i][1]+2.0*s[i][2]*dxi_1+3.0*s[i][3]*dxi_2-s[i+1][1], 2.0);
		sum+=lambda[i][2]*pow(2.0*s[i][2]+6.0*s[i][3]*dxi_1-2.0*s[i+1][2], 2.0);
	}
	
	// edge constraints: natural spline
	sum+=lambda_a*pow(2.0*s[0][2], 2.0);
	sum+=lambda_b*pow(2.0*s[getNOBins()-1][2]+6.0*s[getNOBins()-1][3], 2.0);
	return sum;
}

double SplineFit::chiSquare()
{
	int i;
	double d, sum=0.0;
	for(i=0;i<getM();i++)
	{
		d=(f[i]-y[i])/ye[i];
		sum+=d*d;
	}

	sum+=calculateChiSquareSmoothnessConstraints();
	return sum;
}*/

void SplineFit::foundNewState(const class Vector& x, double new_chi_square, double old_chi_square)
{
/*	double R=(old_chi_square-new_chi_square)/new_chi_square;	
	int i, n=lambda.getM();
	double Sc=constraint_strength;
	lambda_a=Sc;
	for(i=0;i<n-1;i++)
	{
		lambda[i][0]=Sc;
		lambda[i][1]=Sc;
		lambda[i][2]=Sc;
	}
	lambda_b=Sc;
*/
}















