/**************************************************************************************************************
*
*       This file,
*
*                            /Math/StochasticVariable.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _StochasticVariable_h
#define _StochasticVariable_h

#include "nan.h"

template<class T> class StochasticVariable
{
	int n;

	T x, xq, mean, deviation;

	bool finished;

	T variance()
	{
		T xbar=getMean();
		T n_xbar_xbar=xbar*(xbar*T(n));
		return (xq-n_xbar_xbar)/T(n-1);
	}
public:

	StochasticVariable(){reset();}
	~StochasticVariable(){}
	StochasticVariable(const T& mean, const T& deviation){this->mean=mean;this->deviation=deviation;finished=true;}

	// inline T& estimate(){return est;}

	void reset()
	{
		n=0;
		x=xq=0.0;
		mean=0.0;
		deviation=0.0;
		finished=false;
	}

	/*void registerEstimate()
	{
		n++;
		x+=est;
		xq+=est*est;
	}*/

	void registerEstimate(const T& p)
	{
		n++;
		x+=p;
		xq+=p*p;
	}

	void finish()
	{
		if(!finished)
		{
			if(n==0)
			{
				finished=true;
				return;
			}
			mean=getMean();
			deviation=getStddev();
			finished=true;
		}
	}

	T getMean()
	{
		if(finished)
			return mean;
		return x/T(n);
	}

	T getStddev()
	{
		if(finished)
			return deviation;
		//return sqrt(variance()/T(n));
		return sqrt(variance()/T(n));
	}

	inline T getPrecision()
	{
		return fabs(sqrt(variance()*T(n))/x);
	}


	// inline void print(FILE *fptr, const char *format){fprintf(fptr, format, average(), stddev());}
	class StochasticVariable& operator*=(const double& f)
	{
		if(!finished)
			finish();
		mean *= f;
		deviation *= f;
		return *this;
	}

	class StochasticVariable& operator/=(const double& f)
	{
		if(!finished)
			finish();
		mean /= f;
		deviation /= f;
		return *this;
	}

	inline class StochasticVariable& operator=(const T& f){finished=true; mean=f; deviation=0.0; return *this;}

	inline int getN(){return n;}

	inline bool isNaN() {return ISNAN(x) || ISNAN(xq) || ISNAN(mean) || ISNAN(deviation);}
	inline T getX(){return x;}
	inline T getMeanEstimate(){return x/T(n);}

};

#endif //_StochasticVariable_h


