/**************************************************************************************************************
*
*       This file,
*
*                            /Math/StokesMatrix.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "StokesMatrix.h"
#include "StokesVector.h"


StokesMatrix::StokesMatrix(const double& alpha)
{
	double c2a=cos(2.*alpha), s2a=sin(2.*alpha);

	c[0][0]=1.0;	c[0][1]=0.0;	c[0][2]=0.0;	c[0][3]=0.0;
	c[1][0]=0.0;	c[1][1]=c2a;	c[1][2]=-s2a;	c[1][3]=0.0;
	c[2][0]=0.0;	c[2][1]=s2a;	c[2][2]=c2a;	c[2][3]=0.0;
	c[3][0]=0.0;	c[3][1]=0.0;	c[3][2]=0.0;	c[3][3]=1.0;
}

StokesMatrix::StokesMatrix(const StokesMatrix& a)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]=a.c[i][j];
}

/*void StokesMatrix::initializeRotationMatrix(const double& alpha)
{
	double c2a=cos(2.*alpha), s2a=sin(2.*alpha);

	c[0][0]=1.0;	c[0][1]=0.0;	c[0][2]=0.0;	c[0][3]=0.0;
	c[1][0]=0.0;	c[1][1]=c2a;	c[1][2]=s2a;	c[1][3]=0.0;
	c[2][0]=0.0;	c[2][1]=-s2a;	c[2][2]=c2a;	c[2][3]=0.0;
	c[3][0]=0.0;	c[3][1]=0.0;	c[3][2]=0.0;	c[3][3]=1.0;

}*/

void StokesMatrix::setToZero()
{
	c[0][0]=0.0;	c[0][1]=0.0;	c[0][2]=0.0;	c[0][3]=0.0;
	c[1][0]=0.0;	c[1][1]=0.0;	c[1][2]=0.0;	c[1][3]=0.0;
	c[2][0]=0.0;	c[2][1]=0.0;	c[2][2]=0.0;	c[2][3]=0.0;
	c[3][0]=0.0;	c[3][1]=0.0;	c[3][2]=0.0;	c[3][3]=0.0;
}

void StokesMatrix::setToUnityMatrix()
{
	c[0][0]=1.0;	c[0][1]=0.0;	c[0][2]=0.0;	c[0][3]=0.0;
	c[1][0]=0.0;	c[1][1]=1.0;	c[1][2]=0.0;	c[1][3]=0.0;
	c[2][0]=0.0;	c[2][1]=0.0;	c[2][2]=1.0;	c[2][3]=0.0;
	c[3][0]=0.0;	c[3][1]=0.0;	c[3][2]=0.0;	c[3][3]=1.0;
}

class StokesVector StokesMatrix::operator*(const class StokesVector& a)
{
	double value;
	register int i, j;
	class StokesVector result;
	for(i=0;i<4;i++)
	{
		value=0.0;
		for(j=0;j<4;j++)
			value+=c[i][j]*a.c[j];
		result.c[i]=value;
	}
	return result;
}

class StokesMatrix StokesMatrix::operator*(const double& f)
{
	register int i, j;
	class StokesMatrix result;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			result.c[i][j]=c[i][j]*f;
	return result;
}

class StokesMatrix& StokesMatrix::operator=(const class StokesMatrix& a)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]=a.c[i][j];
	return *this;
}

class StokesMatrix StokesMatrix::operator*(const class StokesMatrix& a)
{
	register int i, j, k;
	double value;
	class StokesMatrix result;
	for(i=0;i<4;i++)	
		for(j=0;j<4;j++)
		{
			value=0.0;
			for(k=0;k<4;k++)
				value+=c[i][k]*a.c[k][j];
			result.c[i][j]=value;
		}
	return result;
}

class StokesMatrix StokesMatrix::operator/(const double& q)
{
	register int i, j;
	class StokesMatrix result;
	for(i=0;i<4;i++)	
		for(j=0;j<4;j++)
			result.c[i][j]=c[i][j]/q;
	return result;
}

class StokesMatrix& StokesMatrix::operator+=(const class StokesMatrix& a)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]+=a.c[i][j];
	return *this;
}

class StokesMatrix& StokesMatrix::operator-=(const class StokesMatrix& a)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]-=a.c[i][j];
	return *this;
}

class StokesMatrix& StokesMatrix::operator*=(const class StokesMatrix& a)
{
	register int i, j, k;
	double value;
	class StokesMatrix result;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
		{
			value=0.0;
			for(k=0;k<4;k++)
				value+=c[i][k]*a.c[k][j];
			result.c[i][j]=value;
		}
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]=result.c[i][j];
	return *this;
}

class StokesMatrix& StokesMatrix::operator*=(const double& x)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]*=x;
	return *this;
}

class StokesMatrix& StokesMatrix::operator/=(const double& x)
{
	register int i, j;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			c[i][j]/=x;
	return *this;
}


/*int main()
{
	StokesVector a;
	a.setTo(1.0);
	StokesVector aa=a*a;
	return 0;
}*/

