/**************************************************************************************************************
*
*       This file,
*
*                            /Math/StokesVector.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "StokesVector.h"

StokesVector::StokesVector(const StokesVector& a)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]=a.c[i];
}

StokesVector::StokesVector(const int& x)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]=x;
}

StokesVector::StokesVector(const double& I, const double& Q, const double& U, const double& V)
{
	c[0]=I;
	c[1]=Q;
	c[2]=U;
	c[3]=V;
}

class StokesVector StokesVector::operator+(const class StokesVector& a) const
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=c[i]+a.c[i];
	return result;
}

class StokesVector StokesVector::operator-(const class StokesVector& a) const
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=c[i]+a.c[i];
	return result;
}

class StokesVector StokesVector::operator*(const class StokesVector& a) const
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=c[i]*a.c[i];
	return result;
}

class StokesVector StokesVector::operator/(const class StokesVector& a) const
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=c[i]/a.c[i];
	return result;
}

class StokesVector& StokesVector::operator=(const class StokesVector& a)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]=a.c[i];
	return *this;
}

class StokesVector sqrt(const class StokesVector& r)
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=sqrt(r.c[i]);
	return result;
}

class StokesVector fabs(const class StokesVector& r)
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=fabs(r.c[i]);
	return result;
}

class StokesVector StokesVector::operator*(const double& f) const
{
	register int i;
	class StokesVector result;
	for(i=0;i<4;i++)
		result.c[i]=c[i]*f;
	return result;
}

class StokesVector& StokesVector::operator+=(const class StokesVector& a)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]+=a.c[i];
	return *this;
}

class StokesVector& StokesVector::operator-=(const class StokesVector& a)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]-=a.c[i];
	return *this;
}

class StokesVector& StokesVector::operator*=(const double& x)
{
	register int i;
	for(i=0;i<4;i++)
		c[i]*=x;
	return *this;
}


