/**************************************************************************************************************
*
*       This file,
*
*                            /Math/TMatrix.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "TMatrix.h"
#include "TVector.h"

#include "../error.h"
#include "../fileio.h"
#include "minmax.h"

#include <math.h>
#include <float.h>

template<class T> TMatrix<T>::TMatrix()
{
	M=0;
	N=0;
	data=0;
	c=0;	
}

template<class T> TMatrix<T>::TMatrix(const TMatrix<T>& a)
{
	if(allocate(a.M, a.N))
		memcpy(data, a.data, sizeof(T)*M*N);
	
	else
		errorIn(__FILE__, __LINE__, __func__);
}

template<class T> TMatrix<T>::~TMatrix()
{
	if(data)
		delete [] data;
	if(c)
		delete [] c;
}

template<class T> bool TMatrix<T>::allocate(int m, int n)
{
	if(data)
		delete [] data;
	data=new T[m*n];
	if(c)
		delete [] c;
	c=new T *[m];
	if(!data || !c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %dx%d elements!\n", m, n);
		return false;
	}
	for(int i=0;i<m;i++)
		c[i]=data+i*n;
	M=m;
	N=n;
	return true;
}

template<class T> bool TMatrix<T>::load(const char *file_name)
{
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not load matrix from file \"%s\"!\n", file_name);
		return false;
	}
	if(!allocate(numberOfRows(fptr), numberOfColumns(fptr)))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int i, j, n;
	char fs_tab[5], fs_ret[5];
	sprintf(fs_tab, "%%%s\t", getFormatString());
	sprintf(fs_ret, "%%%s\n", getFormatString());
	T *data_ptr=data;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
		{
			n=fscanf(fptr, j<N-1?fs_tab:fs_ret, data_ptr++);
			if(n!=1)
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in matrix file \"%s\", line %d, column %d!\n", file_name, i+1, j+1);
				return false;
			}
		}
	fclose(fptr);
	return true;
}

template<class T> bool TMatrix<T>::save(const char *file_name)
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing to file \"%s\"!\n", file_name);
		return false;
	}
	dumpTo(fptr);
	fclose(fptr);
	return true;
}

template<class T> void TMatrix<T>::dumpTo(FILE *fptr, const char *formatString, bool transpose)
{
	int i, j;
	char fs_tab[5], fs_ret[5];
	sprintf(fs_tab, "%%%s\t", formatString);
	sprintf(fs_ret, "%%%s\n", formatString);
	if(transpose)
		for(j=0;j<N;j++)
			for(i=0;i<M;i++)
				fprintf(fptr, i<M-1?fs_tab:fs_ret, c[i][j]);
	else
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				fprintf(fptr, j<N-1?fs_tab:fs_ret, c[i][j]);
}

template<class T> class TVector<T> TMatrix<T>::operator*(const class TVector<T>& a) const
{
	int i, j;
	class TVector<T> result;
	if(result.allocate(M) && N==a.N)
		for(i=0;i<M;i++)
		{
			T value=0.0;
			for(j=0;j<N;j++)
				value+=c[i][j]*a.c[j];
			result.c[i]=value;
		}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator=(const class TMatrix<T>& a)
{
	int i, j;
	if(allocate(a.M, a.N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]=a.c[i][j];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator+=(const class TMatrix<T>& a)
{
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
	}
	else
	{
		int i, j;
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]+=a.c[i][j];
	}
	return *this;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator-=(const class TMatrix<T>& a)
{
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
	}
	else
	{
		int i, j;
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]+=a.c[i][j];
	}
	return *this;
}

template<class T> class TMatrix<T> TMatrix<T>::operator*(const class TMatrix<T>& a) const
{
	int i, j, k;
	class TMatrix<T> result;
	if(result.allocate(M, a.N) && a.M==N)
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
			{
				T value=0.0;
				for(k=0;k<N;k++)
					value+=c[i][k]*a.c[k][j];
				result.c[i][j]=value;
			}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TMatrix<T> TMatrix<T>::operator/(const T& q) const
{
	int i, j;
	class TMatrix<T> result;
	if(result.allocate(M, N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				result.c[i][j]=c[i][j]/q;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TMatrix<T> TMatrix<T>::operator*(const T& f) const
{
	int i, j;
	class TMatrix<T> result;
	if(result.allocate(M,N))
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				result.c[i][j]=c[i][j]*f;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator*=(const class TMatrix<T>& a)
{
	int i, j, k;
	class TMatrix<T> result;
	if(result.allocate(M, a.N) && a.M==N)
	{
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
			{
				T value=0.0;
				for(k=0;k<N;k++)
					value+=c[i][k]*a.c[k][j];
				result.c[i][j]=value;
			}
		for(i=0;i<M;i++)
			for(j=0;j<N;j++)
				c[i][j]=result.c[i][j];
	}
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator*=(const T& x)
{
	int i, j;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
			c[i][j]*=x;
	return *this;
}

template<class T> class TMatrix<T>& TMatrix<T>::operator/=(const T& x)
{
	int i, j;
	for(i=0;i<M;i++)
		for(j=0;j<N;j++)
			c[i][j]/=x;
	return *this;
}

template<class T> bool TMatrix<T>::calculateInverse()
{
	if(M!=N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "not a quadratic matrix!\n");
		return false;
	}
	int i, j, s, t, el=0;
	T b, maximum, det, norm;
	int *v=new int[M];
	if(!v)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
/*
	Calculation of Hadamard Condition number: 
	determinant is divided by all geometrical row averages
*/
	norm=1.0;
	for(i=0;i<M;i++)
	{
		T sum=0.0;
		for(j=0;j<N;j++)
			sum+=c[i][j]*c[i][j];
		norm*=sqrt(sum);
	}
	det=1.0;
	for (s=0;s<M;s++)		// { s: column }
	{
		v[s]=s;
	        maximum=0.0;
	        for (t=s;t<M;t++)	//   { pivoting }
		{
			b=fabs(c[t][s]);
			if (b>maximum)
			{
				maximum=b;
				el=t;
			}
		}
	        if(maximum==0.0)
		{
			delete [] v;
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"found singular matrix!\n");
			return false;
		}
		if(el!=s)			//     { largest element at top }
		{
			for (j=0;j<M;j++)		//     { swap rows }
			{
				b=c[s][j];
				c[s][j]=c[el][j];
				c[el][j]=b;
			}
			v[s]=el;	//        { store nb. of swapped row }
			det=-det;   //    {each swap changes det-sign}
		}  // {if el<> s then ...}
	        b=1.0/c[s][s];               //    { normalize with largest element }
	        det*=c[s][s];
	        c[s][s]=1.0;
	        for(j=0;j<M;j++)
			c[s][j]*=b; //   { in row s all divide with max }
	        for(t=0;t<M;t++)
			if (t!=s)
			{ // in all other rows
				b=c[t][s];
				c[t][s]=0.0; // in column s make a 0
				for (j=0;j<M;j++)
					c[t][j]-=b*c[s][j];
	        	}
	}
	// swap back
	for (s=M-1;s>=0;s--)
	{
		if (v[s]!=s)
			for (j=0;j<M;j++)
			{
				b=c[j][s];
				c[j][s]=c[j][v[s]];
				c[j][v[s]]=b;
			}
	}
	delete [] v;
	return true;
}

template<class T> void TMatrix<T>::setToUnity()
{
	int i, n=min(M, N);
	memset(data, 0, sizeof(T)*M*N);
	for(i=0;i<n;i++)
		c[i][i]=1.0;
}

template<class T> class TMatrix<T> TMatrix<T>::getTransposed()
{
	class TMatrix<T> TransposedMatrix;
	if(!TransposedMatrix.allocate(N,M))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return TransposedMatrix;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			TransposedMatrix.c[j][i]=c[i][j];
	return TransposedMatrix;
}

template<class T> class TMatrix<T> TMatrix<T>::operator+(const class TMatrix<T>& a) const
{
	class TMatrix<T> result;
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
		return result;
	}
	if(!result.allocate(M, N))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return result;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			result.c[i][j]=c[i][j]+a.c[i][j];
	return result;
}

template<class T> class TMatrix<T> TMatrix<T>::operator-(const class TMatrix<T>& a) const
{
	class TMatrix<T> result;
	if(M!=a.M || N!=a.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch\n");
		return result;
	}
	if(!result.allocate(M, N))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return result;
	}
	int i, j, m=M, n=N;
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			result.c[i][j]=c[i][j]-a.c[i][j];
	return result;
}

