/**************************************************************************************************************
*
*       This file,
*
*                            /Math/TMatrix.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _TMatrix_h
#define _TMatrix_h

#include <memory.h>
#include <stdio.h>

template<class T> class TVector;

template<class T> class TMatrix
{

		int M, N;
		T *data, **c;
		
		// to be implemented
		const char *getFormatString();

	public:
		
		TMatrix();
		TMatrix(const class TMatrix& a);
		~TMatrix();

		bool allocate(int m, int n);
		inline bool isValid() const {return data && c;}

		bool load(const char *file_name);
		bool save(const char *file_name);

		inline void dumpTo(FILE *fptr){dumpTo(fptr, getFormatString());}
		void dumpTo(FILE *fptr, const char *formatString, bool transpose=false);

		inline T *getData(){return data;}
		inline T *getCoefficientPtr(){return data;}
		inline T **getRowPtr(){return c;}
		inline T *getRowPtr(int i){return c[i];}

		bool calculateInverse();
		class TMatrix<T> getTransposed();

		inline int getM() const {return M;}
		inline int getN() const {return N;}
		inline int getNumberOfElements() const {return M*N;}

		inline void setToZero() {memset(data, 0, sizeof(T)*M*N);}
		inline void setTo(const T& value) {for(register int i=0;i<M*N;i++) data[i]=value;}
		void setToUnity();
		
		inline T *operator[](const int& i) const {return c[i];}

		class TMatrix<T> operator+(const class TMatrix<T>& a) const;
		class TMatrix<T> operator-(const class TMatrix<T>& a) const;

		class TMatrix<T>& operator=(const class TMatrix<T>& a);
		class TMatrix<T>& operator+=(const class TMatrix<T>& a);
		class TMatrix<T>& operator-=(const class TMatrix<T>& a);

		class TMatrix<T> operator*(const class TMatrix<T>& a) const;
		class TVector<T> operator*(const class TVector<T>& a) const;
		class TMatrix<T> operator*(const T& f) const;
		class TMatrix<T> operator/(const T& q) const;

		class TMatrix<T>& operator*=(const class TMatrix<T>& a);
		class TMatrix<T>& operator*=(const T& x);

		class TMatrix<T>& operator/=(const T& x);
};

#endif //_TMatrix_h




