/**************************************************************************************************************
*
*       This file,
*
*                            /Math/TVector.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "TVector.h"
#include "../error.h"
#include "../fileio.h"
#include "nan.h"

#include <math.h>
#include <float.h>

template<class T> TVector<T>::TVector()
{
	N=0;
	c=0;
}

template<class T> TVector<T>::TVector(const TVector<T>& a)
{
	if((c=new T[a.N]))
		memcpy(c, a.c , sizeof(T)*a.N);
	else
		errorIn(__FILE__, __LINE__, __func__);
	N=a.N;
}

template<class T> bool TVector<T>::allocate(int n)
{
	if(c)
		delete [] c;
	c=new T[n];
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %d elements!\n", n);
		return false;
	}
	N=n;
	return true;
}

template<class T> TVector<T>::~TVector<T>()
{
	if(c) delete [] c;
}

template<class T> class TVector<T> TVector<T>::operator+(const class TVector<T>& a) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]+a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TVector<T> TVector<T>::operator-(const class TVector<T>& a) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]-a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);

	return result;
}

template<class T> class TVector<T> TVector<T>::operator*(const class TVector<T>& a) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]*a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TVector<T> TVector<T>::operator/(const class TVector<T>& a) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]/a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TVector<T>& TVector<T>::operator=(const class TVector<T>& a)
{
	int i;
	if(allocate(a.N))
		for(i=0;i<N;i++)
			c[i]=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);

	return *this;
}

template<class T> class TVector<T> TVector<T>::operator*(const T& f) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N))
		for(i=0;i<N;i++)
			result.c[i]=c[i]*f;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TVector<T> TVector<T>::operator/(const T& x) const
{
	int i;
	class TVector<T> result;
	if(result.allocate(N))
		for(i=0;i<N;i++)
			result.c[i]=c[i]/x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

template<class T> class TVector<T>& TVector<T>::operator+=(const class TVector<T>& a)
{
	int i;
	if(N==a.N)
		for(i=0;i<N;i++)
			c[i]+=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

template<class T> class TVector<T>& TVector<T>::operator-=(const class TVector<T>& a)
{
	int i;
	if(N==a.N)
		for(i=0;i<N;i++)
			c[i]-=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}


template<class T> class TVector<T>& TVector<T>::operator*=(const T& x)
{
	int i;
	if(c)
		for(i=0;i<N;i++)
			c[i]*=x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

template<class T> class TVector<T>& TVector<T>::operator/=(const T& x)
{
	int i;
	if(c)
		for(i=0;i<N;i++)
			c[i]/=x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

template<class T> class TVector<T> operator-(const class TVector<T>& v)
{
	class TVector<T> result;
	if(result.allocate(v.N))
	{
		int i;	
		for(i=0;i<v.N;i++)
			result.c[i]=-v.c[i];
	}
	else
		errorIn(__FILE__, __LINE__, __func__);

	return result;
}

template<class T> bool TVector<T>::copy(const class TVector<T>& v)
{
	if(!N==v.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	memcpy(c, v.c, sizeof(T)*N);
	return true;
}

template<class T> class TVector<T> operator*(const T& f, const class TVector<T>& v)
{
	int i, n=v.getN();
	class TVector<T> result;
	if(!result.allocate(n))
		errorIn(__FILE__, __LINE__, __func__);
	else
		for(i=0;i<n;i++)
			result.c[i]=f*v.c[i];
	return result;	
}

