/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Vector.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>
#include <float.h>

#include "Vector.h"
#include "../error.h"
#include "nan.h"
#include "../fileio.h"


Vector::Vector()
{
	N=0;
	c=0;
}

Vector::Vector(const Vector& a)
{
	if((c=new double[a.N]))
		memcpy(c, a.c , sizeof(double)*a.N);
	else
		errorIn(__FILE__, __LINE__, __func__);
	N=a.N;
}

bool Vector::allocate(int n)
{
	if(c)
		delete [] c;
	c=new double[n];
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating %d elements!\n", n);
		return false;
	}
	N=n;
	return true;
}

Vector::~Vector()
{
	if(c)
		delete [] c;
}

bool Vector::findIndex(const double& v, int& i)
{
	if(v>c[i])
	{
		if(i>=N-1)
			return false;
		while(v>c[i])
		{
			i++;
			if(i==N)
				return false;
		}
		--i;
		return true;
	}
	else if(v<c[i])
	{
		if(i<=0)
			return false;
		while(v<c[i])
		{
			i--;
			if(i==0)
				return false;
		}
	}
	return true;
}

class Vector Vector::operator+(const class Vector& a) const
{
	int i;
	class Vector result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]+a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector Vector::operator-(const class Vector& a) const
{
	int i;
	class Vector result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]-a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);

	return result;
}

class Vector Vector::operator*(const class Vector& a) const
{
	int i;
	class Vector result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]*a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector Vector::operator/(const class Vector& a) const
{
	int i;
	class Vector result;
	if(result.allocate(N) && N==a.N)
		for(i=0;i<N;i++)
			result.c[i]=c[i]/a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector& Vector::operator=(const class Vector& a)
{
	int i;
	if(allocate(a.N))
		for(i=0;i<N;i++)
			c[i]=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);

	return *this;
}

class Vector sqrt(const class Vector& r)
{
	int i;
	class Vector result;
	if(result.allocate(r.N))
		for(i=0;i<r.N;i++)
			result.c[i]=::sqrt(r.c[i]);
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector fabs(const class Vector& r)
{
	int i;
	class Vector result;
	if(result.allocate(r.N))
		for(i=0;i<r.N;i++)
			result.c[i]=::fabs(r.c[i]);
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector Vector::operator*(const double& f) const
{
	int i;
	class Vector result;
	if(result.allocate(N))
		for(i=0;i<N;i++)
			result.c[i]=c[i]*f;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector Vector::operator/(const double& x) const
{
	int i;
	class Vector result;
	if(result.allocate(N))
		for(i=0;i<N;i++)
			result.c[i]=c[i]/x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return result;
}

class Vector& Vector::operator+=(const class Vector& a)
{
	int i;
	if(N==a.N)
		for(i=0;i<N;i++)
			c[i]+=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

class Vector& Vector::operator-=(const class Vector& a)
{
	int i;
	if(N==a.N)
		for(i=0;i<N;i++)
			c[i]-=a.c[i];
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}


class Vector& Vector::operator*=(const double& x)
{
	int i;
	if(c)
		for(i=0;i<N;i++)
			c[i]*=x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

class Vector& Vector::operator/=(const double& x)
{
	int i;
	if(c)
		for(i=0;i<N;i++)
			c[i]/=x;
	else
		errorIn(__FILE__, __LINE__, __func__);
	return *this;
}

bool Vector::calculateRange(double *pMinimum, double *pMaximum) const
{
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	int i;	
	double value;
	double minimum=DBL_MAX;
	double maximum=DBL_MIN;
	for(i=0;i<N;i++)
	{
		value=c[i];
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	*pMinimum=minimum;
	*pMaximum=maximum;
	return true;
}

class Vector operator-(const class Vector& v)
{
	class Vector result;
	if(result.allocate(v.N))
	{
		int i;	
		for(i=0;i<v.N;i++)
			result.c[i]=-v.c[i];
	}
	else
		errorIn(__FILE__, __LINE__, __func__);

	return result;
}

bool Vector::load(const char *file_name)
{
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading file \"%s\"!\n", file_name);
		return false;
	}
	if(!allocate(numberOfRows(fptr)))
	{
		fclose(fptr);
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "Allocation error. Check file \"%s\"!\n", file_name);
		return false;
	}
	for(int i=0;i<N;i++)
		if(!fscanf(fptr, "%le", c+i))
		{
			fclose(fptr);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "Error in line %d of file \"%s\"\n", i+1, file_name);
			return false;
		}
	fclose(fptr);
	return true;
}

double Vector::scalarProductWith(const class Vector& a) const
{
	if(a.N!=N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "dimension mismatch in scalar product\n");
		return 0.0;
	}	
	int i, n=N;
	double sum=0.0, *c1p=c, *c2p=a.c;
	for(i=0;i<n;i++)
		sum += *c1p++ * *c2p++;
	return sum;
}

void Vector::print(FILE *fptr, bool transpose) const
{
	int i;
	// fprintf(fptr, "%d", N);
	if(transpose)
		for(i=0;i<N;i++)
			fprintf(fptr, "%le\n", c[i]);
	else
	{
		for(i=0;i<N;i++)
			fprintf(fptr, "\t%le", c[i]);
		fputc('\n', fptr);
	}
}

bool Vector::copy(const class Vector& v)
{
	if(N!=v.N)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	memcpy(c, v.c, sizeof(double)*N);
	return true;
}

class Vector operator*(const double& f, const class Vector& v)
{
	int i, n=v.getN();
	class Vector result;
	if(!result.allocate(n))
		errorIn(__FILE__, __LINE__, __func__);
	else
		for(i=0;i<n;i++)
			result.c[i]=f*v.c[i];
	return result;	
}

bool Vector::save(const char *file_name, bool transpose) const
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing to file \"%s\"!\n", file_name);
		return false;
	}
	print(fptr, transpose);
	fclose(fptr);
	return true;
}

bool Vector::getRangeMinimumAndMaximum(int i1, int i2, double& minimum, double& maximum) const
{
	if(!c)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "vector not initialized!\n");
		return false;
	}
	if(i1<0 || i1>=N || i2<0 || i2>=N || i2<i1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "invalid range!\n");
		return false;
	}
	
	int i;	
	double value;
	minimum=DBL_MAX;
	maximum=DBL_MIN;
	for(i=i1;i<=i2;i++)
	{
		value=c[i];
		if(value<minimum)
			minimum=value;
		if(value>maximum)
			maximum=value;
	}
	return true;
}





