/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Vector.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Vector_h
#define _Vector_h

#include <memory.h>
#include <stdio.h>

#include "../error.h"

class Vector
{
		friend class Matrix;

		int N;
		double *c;

	public:
		
		Vector();
		Vector(const class Vector& a);
		~Vector();

		bool allocate(int n);
		inline bool isValid() const {return c;}

		bool calculateRange(double *pMinimum, double *pMaximum) const;
		bool getRangeMinimumAndMaximum(int i1, int i2, double& minimum, double& maximum) const;
		bool findIndex(const double& v, int& i);

		bool load(const char *file_name);
		bool save(const char *file_name, bool transpose=false) const;
		void print(FILE *fptr, bool transpose=false) const;
		
		inline int getN() const {return N;}
		bool copy(const class Vector& v);

		inline void setTo(const double& x){for(register int i=0;i<N;i++) c[i]=x;}
		inline void setToZero(){setTo(0.0);}

		inline double& operator[](const int& i) const
		{
			if(i<0 || i>=N)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "index %d\n", i);
			}
			return c[i];
		}
		inline double *getCoefficientPtr() const {return c;}

		double scalarProductWith(const class Vector& a) const;

		class Vector operator+(const class Vector& a) const;
		class Vector operator-(const class Vector& a) const;

		class Vector operator*(const double& x) const;
		class Vector operator*(const class Vector& a) const;
		class Vector operator/(const double& x) const;
		class Vector operator/(const class Vector& a) const;
	
		class Vector& operator=(const class Vector& a);
		class Vector& operator+=(const class Vector& a);
		class Vector& operator-=(const class Vector& a);
		class Vector& operator*=(const double& x);
		class Vector& operator/=(const double& x);
		
		inline class Vector& operator=(const double& x){setTo(x); return *this;}

		friend class Vector sqrt(const class Vector& r);
		friend class Vector fabs(const class Vector& r);
		friend class Vector operator-(const class Vector& v);
		friend class Vector operator*(const double& f, const class Vector& v);
		
};

#endif //_Vector_h



