/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Vector3D.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>

#include "Vector3D.h"
#include "Matrix3D.h"


void Vector3D::toPolar(double *pR, double *pTheta, double *pPhi) const
{
	double sqxy=x[0]*x[0]+x[1]*x[1];
	
	*pTheta=atan2(sqrt(sqxy), x[2]);
	*pPhi=atan2(x[1], x[0]);
	*pR=sqrt(sqxy+x[2]*x[2]);
	/*if(ISNAN(*pTheta) || ISNAN(*pPhi) || ISNAN(*pR))
		fprintf(stderr, "file \"%s\" line %d: toPolar() -> NAN\n", __FILE__, __LINE__);*/
}

void Vector3D::toPolar(double *pTheta, double *pPhi) const
{
	double sqxy=x[0]*x[0]+x[1]*x[1];
	
	*pTheta=atan2(sqrt(sqxy), x[2]);
	*pPhi=atan2(x[1], x[0]);
	/*if(ISNAN(*pTheta) || ISNAN(*pPhi) || ISNAN(*pR))
		fprintf(stderr, "file \"%s\" line %d: toPolar() -> NAN\n", __FILE__, __LINE__);*/
}

void Vector3D::fromPolar(const double& r, const double& theta, const double& phi)
{
	double sin_theta=sin(theta);
	double cos_theta=cos(theta);
	x[0]=r*cos(phi)*sin_theta;
	x[1]=r*sin(phi)*sin_theta;
	x[2]=r*cos_theta;
}

void Vector3D::fromPolar(const double& theta, const double& phi)
{
	double sin_theta=sin(theta);
	double cos_theta=cos(theta);
	x[0]=cos(phi)*sin_theta;
	x[1]=sin(phi)*sin_theta;
	x[2]=cos_theta;
}
void Vector3D::toCartesic(double *pX, double *pY, double *pZ) const
{
	*pX=x[0]; *pY=x[1]; *pZ=x[2];
}

class Vector3D Vector3D::getUnityVector() const
{
	double l=sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
	return Vector3D(x[0]/l, x[1]/l, x[2]/l);
}

class Vector3D Polar(const double& r, const double& theta, const double& phi)
{
	double sin_theta=sin(theta);
	return Vector3D(r*cos(phi)*sin_theta, r*sin(phi)*sin_theta, r*cos(theta));
}

class Vector3D Vector3D::createDRVector(const double& theta, const double& phi)
{
	double sin_theta=sin(theta);
	return Vector3D(cos(phi)*sin_theta, sin(phi)*sin_theta, cos(theta));
}

class Vector3D Vector3D::createDThetaVector(const double& theta, const double& phi)
{
	double cos_theta=cos(theta);
	return Vector3D(cos(phi)*cos_theta, sin(phi)*cos_theta, -sin(theta));
}

class Vector3D Vector3D::createDPhiVector(const double& theta, const double& phi)
{
	double sin_theta=sin(theta);
	return Vector3D(-sin(phi)*sin_theta, cos(phi)*sin_theta, 0.0);
}

class Vector3D Vector3D::operator-(const class Vector3D& a) const
{
	return Vector3D(x[0]-a.x[0], x[1]-a.x[1], x[2]-a.x[2]);
}

class Vector3D Vector3D::operator+(const class Vector3D& a) const
{
	return Vector3D(x[0]+a.x[0], x[1]+a.x[1], x[2]+a.x[2]);
}

class Vector3D& Vector3D::operator*=(const Matrix3D& m)
{
/*
	calculates:

		*this=Matrix3D * *this
*/
	register int i, j;
	register double sum;
	double result[3];
	for(i=0;i<3;i++)
	{
		sum=0.0;
		for(j=0;j<3;j++)
			sum+=m.a[i][j]*x[j];
		result[i]=sum;
	}
	x[0]=result[0];
	x[1]=result[1];
	x[2]=result[2];
	return *this;
}

double Vector3D::operator*(const class Vector3D& a) const
{
	register int i;
	register double sum=0.0;
	for(i=0;i<3;i++)	
		sum+=x[i]*a.x[i];
	return sum;
}

class Vector3D Vector3D::cross(const class Vector3D& a) const
{
	return Vector3D(x[1]*a.x[2]-x[2]*a.x[1], x[2]*a.x[0]-x[0]*a.x[2], x[0]*a.x[1]-x[1]*a.x[2]);
}

class Vector3D& Vector3D::operator+=(const class Vector3D& a)
{
	x[0]+=a.x[0];
	x[1]+=a.x[1];
	x[2]+=a.x[2];
	return *this;
}

class Vector3D& Vector3D::operator-=(const class Vector3D& a)
{
	x[0]-=a.x[0];
	x[1]-=a.x[1];
	x[2]-=a.x[2];
	return *this;
}

class Vector3D Vector3D::operator*(const class Matrix3D& m) const
{
	register int i, j;
	register double sum;
	class Vector3D r;
	for(j=0;j<3;j++)
	{
		sum=0.0;
		for(i=0;i<3;i++)
			sum+=x[j]*m.a[i][j];
		r.x[j]=sum;
	}
	return r;
}

class Vector3D& Vector3D::operator*=(const double& f)
{
	x[0]*=f;
	x[1]*=f;
	x[2]*=f;
	return *this;
}

class Vector3D Vector3D::operator/(const double& d) const
{
	return Vector3D (x[0]/d, x[1]/d, x[2]/d);
}

