/**************************************************************************************************************
*
*       This file,
*
*                            /Math/Vector3D.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Vector3D_h
#define _Vector3D_h

#include <stdio.h>
#include <math.h>

#include "nan.h"

class Matrix3D;

class Vector3D
{

	double x[3];

	friend class Matrix3D;

public:

	Vector3D(){}
	Vector3D(const double& x0, const double& x1, const double& x2) {x[0]=x0; x[1]=x1; x[2]=x2;}
	Vector3D(const class Vector3D& v) {x[0]=v.x[0]; x[1]=v.x[1]; x[2]=v.x[2];}

	~Vector3D(){}

	inline double *getCoefficientPtr() {return x;}

	void toPolar(double *pR, double *pTheta, double *pPhi) const;
	void toPolar(double *pTheta, double *pPhi) const;
	void fromPolar(const double& r, const double& theta, const double& phi);
	void fromPolar(const double& theta, const double& phi);
	void toCartesic(double *pX, double *pY, double *pZ) const;
	

	class Vector3D cross(const class Vector3D&) const;
	class Vector3D getUnityVector() const;
	inline double length() const {return sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);}
	inline bool isValid() const {return (ISNAN(x[0]) || ISNAN(x[1]) || ISNAN(x[2]))? false: true;}

	// operators
	inline double& operator[](const int& i){return x[i];}
	class Vector3D operator-(const class Vector3D&) const;
	class Vector3D operator+(const class Vector3D&) const;
	double operator*(const class Vector3D&) const;
	class Vector3D& operator+=(const class Vector3D& a);
	class Vector3D& operator-=(const class Vector3D& a);
	inline class Vector3D operator*(const double& f) const {return Vector3D(x[0]*f, x[1]*f, x[2]*f);}
	class Vector3D operator*(const class Matrix3D& m) const;
	class Vector3D operator/(const double& d) const;

	// static member functions
	static class Vector3D createDRVector(const double& theta, const double& phi);
	static class Vector3D createDThetaVector(const double& theta, const double& phi);
	static class Vector3D createDPhiVector(const double& theta, const double& phi);

	// friends
	friend class Vector3D norm(const class Vector3D&);
	friend class Vector3D Polar(const double&, const double&, const double&);

	friend class Vector3D operator*(const double&, const class Vector3D&);
	friend class Vector3D operator-(const class Vector3D&);
	friend class Matrix3D axisRotation(const class Vector3D&, const double&);
	
	class Vector3D& operator*=(const double& f);	
	// mixtures with Matrix3D
	class Vector3D& operator*=(const class Matrix3D&);	// *this=Matrix3D * *this


};

/*// friends
inline class Vector3D norm(const class Vector3D& p){return p.getUnityVector();}
class Vector3D Polar(const double& r, const double& theta, const double& phi);
class Vector3D operator*(const double& x, const class Vector3D& v){return Vector3D(x*v.x[0], x*v.x[1], x*v.x[2]);}
class Vector3D operator-(const class Vector3D& v){return Vector3D(-v.x[0], -v.x[1], -v.x[2]);}
class Matrix3D axisRotation(const class Vector3D&, const double&);*/

inline class Vector3D norm(const class Vector3D& p){return p.getUnityVector();}
class Vector3D Polar(const double& r, const double& theta, const double& phi);
inline class Vector3D operator*(const double& x, const class Vector3D& v){return Vector3D(x*v.x[0], x*v.x[1], x*v.x[2]);}
inline class Vector3D operator-(const class Vector3D& v){return Vector3D(-v.x[0], -v.x[1], -v.x[2]);}
class Matrix3D axisRotation(const class Vector3D&, const double&);


#endif // _Vector3D_h


