/**************************************************************************************************************
*
*       This file,
*
*                            /Math/statistics.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <math.h>
#include <stdio.h>

#include "statistics.h"
#include "nan.h"
#include "prand.h"

#include "../error.h"

double covariance(double *px0, double *py0, const int& n0)
{
//	calculates the covariance of x and y
	double *px=px0;
	double *py=py0;
	double xm=0.0, ym=0.0;
	int i, n=n0;
	// calculate mean values
	for(i=0;i<n;i++)
	{
		xm+=*px++;
		ym+=*py++;
	}
	xm/=double(n);
	ym/=double(n);
	px=px0;
	py=py0;
	// calulate covariance
	double cov=0.0;
	for(i=0;i<n;i++)
		cov+=(*px++-xm)*(*py++-ym);
	cov/=double(n);
	return cov;
}

double correlation(double *px0, double *py0, const int& n0)
{
//	calculates the correlation coefficient of x and y
	double *px=px0;
	double *py=py0;
	double xm=0.0, ym=0.0;
	int i, n=n0;
	// calculate mean values
	for(i=0;i<n;i++)
	{
		xm+=*px++;
		ym+=*py++;
	}
	xm/=double(n);
	ym/=double(n);
	// calulate covariance
	px=px0;
	py=py0;
	double cov=0.0, dx, dy, vx=0.0, vy=0.0, cor;
	for(i=0;i<n;i++)
	{
		dx=*px++-xm;
		dy=*py++-ym;
		cov+=dx*dy;
		vx+=dx*dx;
		vy+=dy*dy;
	}
	if(vx==0.0 || vy==0.0)
		return 0.0;
	cor=cov/sqrt(vx*vy);
	if(ISNAN(cor))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "NAN\n");
		return 0.0;
	}
	return double(cor);
}

