/**************************************************************************************************************
*
*       This file,
*
*                            /MieModule.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "MieModule.h"
#include "error.h"

#define An_MAX		60000

MieModule::MieModule()
{
	x=0.0;
}

MieModule::~MieModule()
{
}

int MieModule::calculateNOCoefficients(const double& x)
{
	double N;
	/*if(x>=0.02 && x<8.0)
		N=x+4.0*pow(x, 1.0/3.0)+1.0;
	else if(x>=8.0 && x<4200.0)
		N=x+4.05*pow(x, 1.0/3.0)+2.0;
	else if(x>=4200 && x<20000.0)
		N=x+4.0*pow(x, 1.0/3.0)+2.0;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "x out of range [2e-2, 2e5]!\n");
		return false;
	}*/
	N=x+4.0*pow(x, 1.0/3.0)+12.0;
	return int(N);
}

bool MieModule::setX(const double& x)
{
	N_max=calculateNOCoefficients(x);
	if(!allocate(N_max))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	this->x=x;
	sinx=sin(x);
	cosx=cos(x);
	xx=x*x;
	return true;
}

bool MieModule::setRefractiveIndex(const complex<double>& m)
{
	if(imag(m)>0.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "sign convention states: imaginary part of the refractive index required to be negative!\n");
		return false;
	}
	z=m*x;
	this->m=m;
	mm=m*m;
	mm_one_o_mmx=(mm-1.0)/(x*mm);
	return true;
}

bool MieModule::allocate(int n)
{
	if(!A.allocate(An_MAX+1) || 
		!a.allocate(n) || 
		!b.allocate(n) || 
		!psi.allocate(n+1) || 
		!chi.allocate(n+1) ||
		!pi.allocate(n+1) ||
		!tau.allocate(n+1) ||
		!TwoNp1oNNp1.allocate(n+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	return true;
}

bool MieModule::setParameters(const double& lambda, const double& m_re, const double& m_im, const double& radius)
{
	k=(2.0*M_PI)/lambda;
	x=k*radius;
	if(!setX(x))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	m=complex<double>(m_re, m_im);
	if(!setRefractiveIndex(m))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	return true;
}

void MieModule::calculateAnPsiChi()
{
	//  calculate helper functions
	int i;

	chi[0]=cosx;
	psi[0]=sinx;

	chi[1]=cosx/x+sinx;	
	psi[1]=sinx/x-cosx;
	
	double f;
	for(i=1;i<N_max;i++)
	{
		f=double(2*i+1);
		chi[i+1]=(f*chi[i])/x-chi[i-1];
		psi[i+1]=(f*psi[i])/x-psi[i-1];
	}
	
	// calculate An
/*	complex<double> Ai=A[N_max-1]=(double)(2*N_max+1)/z;
	for(i=N_max-1;i>=1;i--)
		Ai=A[i-1]=double(2*i+1)/z-1.0/Ai;*/
	A[An_MAX-1]=(double)(2*An_MAX+1)/z;
	for(i=An_MAX-1;i>=1;i--)
		A[i-1]=double(2*i+1)/z-1.0/A[i];
}

void MieModule::calculateTwoNp1oNNp1(double *TwoNp1oNNp1, const int& n)
{
	int i;
	double *ptr=TwoNp1oNNp1;
	*ptr++=0.0;
	for(i=1;i<n;i++)
		*ptr++=(double)(2*i+1)/((double)(i)*(double)(i+1));
}

void MieModule::calculatePiAndTau(const double& mu, double* pi, double *tau, const int& n)
{
	int i;
	double s, t;

	// calculate Legendre coefficients for mu
	pi[0]=0.0;
	tau[0]=0.0;

	pi[1]=1.0;
	tau[1]=mu;
	
	for(i=2;i<=n;i++)
	{
		s=mu*pi[i-1];
		t=s-pi[i-2];
		pi[i]=s+(double(i)*t)/double(i-1);
		tau[i]=double(i)*mu*pi[i]-double(1+i)*pi[i-1];
	}
}

/*complex<double> MieModule::An(int n)
{
	// Lenzs method
	double nd=double(n);
	complex<double> alpha_j1, alpha_j2, zinv, aj;
	complex<double> alpha, result, ratio, runratio;
	zinv=2.0/z;
	alpha=(0.5+nd)*zinv;
	aj=(-1.5-nd)*zinv;
	alpha_j1=aj+1.0/alpha;
	alpha_j2=aj;
	ratio=alpha_j1/alpha_j2;
	runratio=alpha*ratio;
	do
	{
		aj=zinv-aj;
		alpha_j1=aj+1.0/alpha_j1;
		alpha_j2=aj+1.0/alpha_j2;
		ratio=alpha_j1/alpha_j2;
		if(fabs(abs(alpha_j1/aj))<=1e-2 || fabs(abs(alpha_j2/aj))<=1e-2)
		{
			zinv*=-1.0;
			aj=zinv-aj;
			ratio=(1.0+aj*alpha_j1)/(1.0+aj*alpha_j2);
			alpha_j1=1.0/alpha_j1+aj;
			alpha_j2=1.0/alpha_j2+aj;
		}
		zinv*=-1.0;
		runratio*=ratio;
	}while(fabs(abs(ratio)-1.0)>1e-16);
	return runratio-nd/z;
}*/

void MieModule::calculateMieCoefficients()
{
/*
	calculate Ricatti-Bessel functions, logarithmic derivatives and finally the Mie coefficients
*/

	calculateAnPsiChi();
	
	int i;
	complex<double> F, zeta_im1, zeta_i, a_im1, b_im1;
	double psi_i, chi_i, psi_im1=psi[0], chi_im1=chi[0], f;
	for(i=1;i<=N_max;i++)
	{
		psi_i=psi[i];
		chi_i=chi[i];

		zeta_i=complex<double>(psi_i, chi_i);
		zeta_im1=complex<double>(psi_im1, chi_im1);

		F=A[i-1]/m+double(i)*mm_one_o_mmx;
		a[i-1]=(F*psi_i-psi_im1)/(F*zeta_i-zeta_im1);

		F=m*A[i-1];
		b[i-1]=(F*psi_i-psi_im1)/(F*zeta_i-zeta_im1);

		psi_im1=psi_i;
		chi_im1=chi_i;
	}
	double Qext_sum=0.0, Qsca_sum=0.0, g_sum=0.0, aa, bb, f1, f2;
	for(i=1;i<=N_max;i++)
	{
		f=(double)(2*i+1);

		a_im1=a[i-1];
		b_im1=b[i-1];

		aa = abs(a_im1);
		aa *= aa;

		bb = abs(b_im1);
		bb *= bb;

		Qext_sum += f * ( a_im1.real() + b_im1.real() );
		Qsca_sum += f * ( aa + bb );

		f1=(double(i)*double(i+2))/(double(i+1));
		f2=double(2*i+1)/(double(i)*double(i+1));
		g_sum += f1 * ( a_im1 * conj ( a[i] ) + b_im1 * conj ( b[i] ) ) . real() + f2 * ( a_im1 * conj ( b_im1 ) ) . real();

		/*f1=(double(i-1)*double(i+1))/double(i);
		f2=double(2*i+1)/(double(i)*double(i+1));
		if(i>1)
			g_sum+=f1*(a[i-2]*conj(a_im1)+b[i-2]*conj(b[i])).real()+f2*(a_im1*conj(b_im1)).real();
		else
			g_sum+=f2*(a_im1*conj(b_im1)).real();*/
	}

	Qext = ( 2.0 * Qext_sum ) / xx;
	Qsca = ( 2.0 * Qsca_sum ) / xx;
	ssa = Qsca / Qext;
	g = (4.0 * g_sum ) / ( xx * Qsca );

}

void MieModule::calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P33, double& P34, complex<double> *a, complex<double> *b, double *pi, double *tau, int n)
{
	calculatePiAndTau(mu, pi, tau, n);
	// calculate scattering amplitudes
	int i;
	complex<double> S_p=0.0, S_m=0.0, a_im1, b_im1; //, S1=0.0, S2=0.0;
	double f, pi_i, tau_i;
	for(i=1;i<=n;i++)
	{
		f=(double)(2*i+1)/((double)(i)*((double)(i+1)));
		a_im1=a[i-1];
		b_im1=b[i-1];
		pi_i=pi[i];
		tau_i=tau[i];
		S_p+=f*(a_im1+b_im1)*(pi_i+tau_i);
		S_m+=f*(a_im1-b_im1)*(pi_i-tau_i);
		//S1+=f*(a_im1*pi_i+b_im1*tau_i);
		//S2+=f*(a_im1*tau_i+b_im1*pi_i);
	}
	
	// calculate phase matrix elements
	/*P11=S1.real();
	P12=S1.imag();
	P33=S2.real();
	P34=S2.imag();*/

	double Qsca_xx = Qsca * xx;
	P11=(conj(S_p)*S_p+conj(S_m)*S_m).real()/Qsca_xx;
	P12=-(conj(S_p)*S_m+conj(S_m)*S_p).real()/Qsca_xx;
	P33=(conj(S_p)*S_p-conj(S_m)*S_m).real()/Qsca_xx;
	P34=(conj(S_p)*S_m-conj(S_m)*S_p).imag()/Qsca_xx;

	/*f=4.0/(Qsca*xx);
	P11 = f * (conj(S2)*S2+conj(S1)*S1).real()/2.0;
	P12 = f * (conj(S2)*S2-conj(S1)*S1).real()/2.0;
	P33 = f * (conj(S2)*S1+conj(S1)*S2).real()/2.0;
	P34 = f * (conj(S2)*S1-conj(S1)*S2).imag()/2.0;*/
}

void MieModule::calculatePhaseMatrix(const double& mu, class StokesMatrix& PhaseMatrix, bool zero_rest)
{
	double P11, P12, P33, P34;
	
	calculatePhaseMatrixElements(mu, P11, P12, P33, P34);

	if(zero_rest)
	{
		// PhaseMatrix.setToZero();
		// fill rest with zeros
		PhaseMatrix[0][2]=0.0;
		PhaseMatrix[0][3]=0.0;
		PhaseMatrix[1][2]=0.0;
		PhaseMatrix[1][3]=0.0;
		PhaseMatrix[2][0]=0.0;
		PhaseMatrix[2][1]=0.0;
		PhaseMatrix[3][0]=0.0;
		PhaseMatrix[3][1]=0.0;
	}
	// fill elements
	PhaseMatrix[0][0]=PhaseMatrix[1][1]=P11;
	PhaseMatrix[1][0]=PhaseMatrix[0][1]=P12;
	PhaseMatrix[2][2]=PhaseMatrix[3][3]=P33;
	PhaseMatrix[3][2]=P34;
	PhaseMatrix[2][3]=-P34;

}

void MieModule::printMieCoefficients(FILE *fptr)
{
	int i, n=N_max;
	for(i=0;i<n;i++)
		fprintf(fptr, "%d\t%le\t%le\t%le\t%le\t%le\t%le\t%le\t%le\n", \
		i, a[i].real(), a[i].imag(), \
		b[i].real(), b[i].imag(), \
		A[i].real(), A[i].imag(), \
		psi[i], chi[i]);
}

bool MieModule::createLegendrePolynomialMatrix(int resolution, int no_coefficients, bool recreate)
{
/*
	in columns we have the coefficients belonging to each x
*/
	if(recreate || LegendrePolynomialMatrix.getM()!=resolution+1 || LegendrePolynomialMatrix.getN()!=no_coefficients)
	{
		if(!LegendrePolynomialMatrix.allocate(resolution+1, no_coefficients))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating Legendre coefficient matrix\n");
			return false;
		}
		recreate=true;
	}
	if(recreate)
	{
		int i, j;
		
		double x, R=double(resolution);//, dx=2.0/R;
		double *c, *cm1, *cm2;
		for(i=0;i<=resolution;i++)
		{
			c=LegendrePolynomialMatrix[i];
			x=-1.0+double(2*i)/R;
			cm2=c;
			*c++=1.0;
			cm1=c;
			*c++=x;
			for(j=2;j<no_coefficients;j++)
			{
				*c=(double(2*j-1)*(*cm1)*x-double(j-1)*(*cm2))/double(j);
				cm2++;
				cm1++;
				c++;
			}
		}
	}
	LegendrePolynomialMatrix.save("LPMatrix.txt");
	// done
	return true;
}

bool MieModule::developLegendrePhaseFunctions(LegendrePhaseFunction& F11, LegendrePhaseFunction& F12, LegendrePhaseFunction& F33, LegendrePhaseFunction& F34, const int& no_coefficients, const int& resolution)
{
	if(no_coefficients>resolution)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "for a Legendre polynomial expansion with %d coefficients one needs a higher resolution then %d\n",no_coefficients, resolution);
		return false;
	}
	if(!createLegendrePolynomialMatrix(resolution, no_coefficients))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(!F11.allocate(no_coefficients) ||
		!F12.allocate(no_coefficients) ||
		!F33.allocate(no_coefficients) ||
		!F34.allocate(no_coefficients))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	int i, j;
	double x, *c, *c11, *c12, *c33, *c34;
	double P11, P12, P33, P34, f;
	
	for(i=0;i<=resolution;i++)
	{
		c=LegendrePolynomialMatrix[i];
		x=c[1];
		calculatePhaseMatrixElements(x, P11, P12, P33, P34);
		
		// reset coefficient pointers
		c11=F11.getCoefficientPtr();
		c12=F12.getCoefficientPtr();
		c33=F33.getCoefficientPtr();
		c34=F34.getCoefficientPtr();
		for(j=0;j<no_coefficients;j++)
		{
			*c11++ += (*c) * P11;
			*c12++ += (*c) * P12;
			*c33++ += (*c) * P33;
			*c34++ += (*c) * P34;
			c++;
		}
	}

	// don't forget f=double(2*j+1);
	c11=F11.getCoefficientPtr();
	c12=F12.getCoefficientPtr();
	c33=F33.getCoefficientPtr();
	c34=F34.getCoefficientPtr();
	const char *file_name="PhaseMatrixElements.txt";
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not write to file \"%s\"\n",file_name); 
	}
	else
		fprintf(fptr, "# F11\tF12\tF33\tF34\n");
	double _c11=c11[0];
	double _c12=c12[0];
	double _c33=c33[0];
	double _c34=c34[0];
	for(j=0;j<no_coefficients;j++)
	{
		f=double(2*j+1);
		c11[j] *= f/_c11;
		c12[j] *= f/_c12;
		c33[j] *= f/_c33;
		c34[j] *= f/_c34;
		if(fptr)
			fprintf(fptr,"%le\t%le\t%le\t%le\n", c11[j], c12[j], c33[j], c34[j]);
	}
	if(fptr)
		fclose(fptr);
	return true;

}

// bool MieModule::processLogNormalDSD(const double& lambda, const double& mean_radius, const double& sigma, const double& re, const double& im, const int& res_DSD, const int& res_phf, bool doLPE, int no_LPE_coefficients)
bool MieModule::processLogNormalDSD(const double& lambda, const double& mean_radius, const double& sigma, const double& re, const double& im, const int& res_DSD, const int& res_phf)
{
	double log_mean_radius=log(mean_radius), log_radius, weight, weight_sum=0.0;
	double range=MIE_MODULE_SIGMA_RANGE*sigma, two_sigma_square=2.0*sigma*sigma, radius;
	int i, j;

/*
	effective quantities to be calculated:
		
		extinction cross section
		single scattering albedo
		asymmetry parameter
		phase matrix elements
*/
	class Vector P11Vector, P12Vector, P33Vector, P34Vector;
	if(!P11Vector.allocate(res_phf+1) ||
		!P12Vector.allocate(res_phf+1) ||
		!P33Vector.allocate(res_phf+1) ||
		!P34Vector.allocate(res_phf+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating phase matrix element vectors\n");
		return false;
	}
	P11Vector.setToZero();
	P12Vector.setToZero();
	P33Vector.setToZero();
	P34Vector.setToZero();
	double scattering_coefficient_sum=0.0, scattering_coefficient, epsilon, epsilon_sum=0.0, ssa_sum=0.0, g_sum=0.0, area;
	double P11, P12, P33, P34, mu, delta_log_radius;
	for(i=0;i<=res_DSD;i++)
	{
		delta_log_radius=-range+(double(2*i)*range)/double(res_DSD);
		log_radius=log_mean_radius+delta_log_radius;
		weight=exp(-(delta_log_radius*delta_log_radius)/two_sigma_square);
		radius=exp(log_radius);
		area=2.0*radius*radius;
		
		if(!setX(radius, lambda) || !setRefractiveIndex(re, im))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error setting size parameter or refractive index\n");
			return false;
			
		}
		calculateMieCoefficients();

		epsilon = area * Qext;
		scattering_coefficient = area * Qsca;

		ssa_sum += weight *  scattering_coefficient; 
		g_sum += weight * scattering_coefficient * g;
		// phase functions
		
		for(j=0;j<=res_phf;j++)
		{
			mu=-1.0+double(2*j)/double(res_phf);
			calculatePhaseMatrixElements(mu, P11, P12, P33, P34);
			P11Vector[j]+=weight * scattering_coefficient * P11;
			P12Vector[j]+=weight * scattering_coefficient * P12;
			P33Vector[j]+=weight * scattering_coefficient * P33;
			P34Vector[j]+=weight * scattering_coefficient * P34;
			
		}

		epsilon_sum += weight * epsilon;
		scattering_coefficient_sum += weight * scattering_coefficient;
		
		weight_sum+=weight;
	}
	double effective_extinction_coefficient = epsilon_sum / weight_sum;
	double effective_ssa = ssa_sum / epsilon_sum;
	double effective_g = g_sum / scattering_coefficient_sum;
	P11Vector /= scattering_coefficient_sum;
	P12Vector /= scattering_coefficient_sum;
	P33Vector /= scattering_coefficient_sum;
	P34Vector /= scattering_coefficient_sum;
	// out put scalar quantities
	printf("# ext. %le, ssa %.4lf, g %.3lf\n", effective_extinction_coefficient, effective_ssa, effective_g);
	/*if(doLPE)
	{
		if(!createLegendrePolynomialMatrix(res_phf, no_LPE_coefficients))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		class Vector P11_LC, P12_LC, P33_LC, P34_LC;  
		if(!P11_LC.allocate(no_LPE_coefficients) ||
			!P12_LC.allocate(no_LPE_coefficients) ||
			!P33_LC.allocate(no_LPE_coefficients) ||
			!P34_LC.allocate(no_LPE_coefficients))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error allocating Legendre polynomial expansion coefficients\n"); 
			return false;
		}
		double P11, P12, P33, P34, f;
		double *c, *c11, *c12, *c33, *c34;
		for(i=0;i<=res_phf;i++)
		{
			c=LegendrePolynomialMatrix[i];
			
			// reset coefficient pointers
			c11=P11_LC.getCoefficientPtr();
			c12=P12_LC.getCoefficientPtr();
			c33=P33_LC.getCoefficientPtr();
			c34=P34_LC.getCoefficientPtr();
			
			// reset P11, P12, P33, P34
			P11=P11Vector[i];
			P12=P12Vector[i];
			P33=P33Vector[i];
			P34=P34Vector[i];
			
			for(j=0;j<no_LPE_coefficients;j++)
			{
				*c11++ += (*c) * P11;
				*c12++ += (*c) * P12;
				*c33++ += (*c) * P33;
				*c34++ += (*c) * P34;
				c++;
			}
		}
		// finalize coefficients
		double _c11=P11_LC[0];
		double _c12=P12_LC[0];
		double _c33=P33_LC[0];
		double _c34=P34_LC[0];

		// reset pointers
		c11=P11_LC.getCoefficientPtr();
		c12=P12_LC.getCoefficientPtr();
		c33=P33_LC.getCoefficientPtr();
		c34=P34_LC.getCoefficientPtr();
		for(j=0;j<no_LPE_coefficients;j++)
		{
			f=double(2*j+1);
			c11[j] *= f/_c11;
			c12[j] *= f/_c12; c33[j] *= f/_c33; c34[j] *= f/_c34; printf("%le\t%le\t%le\t%le\n", c11[j], c12[j], c33[j], c34[j]);
		}
	}
	else*/
		for(j=0;j<=res_phf;j++)
		{
			mu=-1.0+double(2*j)/double(res_phf);
			printf("%lf\t%le\t%le\t%le\t%le\n", mu, P11Vector[j], P12Vector[j], P33Vector[j], P34Vector[j]); 
		}
	return true;
}

void MieModule::calculateCrossSectionSSAAndG(char *bins_file, char *data_file)
{
	// create list of radii
	class Vector BinSize, BinRadius;
	if(!BinSize.load(bins_file))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return;
	}
	if(!BinRadius.allocate(BinSize.getN()-1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return;
	}
	int i;
	for(i=0;i<BinRadius.getN();i++)
		BinRadius[i]=0.5*(BinSize[i]+BinSize[i+1]);
	// radii list done
	
}

void MieModule::calculateBulkDetailedPhaseFunction(char *bins_file, char *data_file, int Nphf)
{

}



