/**************************************************************************************************************
*
*       This file,
*
*                            /MieModule.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _MieModule_h
#define _MieModule_h

#include <complex>

#include "Math/Vector.h" 
#include "Math/Matrix.h" 
#include "Math/TVector.h"
#include "Math/StokesMatrix.h"
#include "PhaseFunctions/LegendrePhaseFunction.h"

using namespace std;

#define MIE_MODULE_SIGMA_RANGE	4.0

class MieModule
{

public:

	struct Mode_s
	{
		double N;	// number of particles in the mode
		double radius;	// mean radius
		double sigma;	// sigma of the log normal distribution in logarithmic radius space
	};

	enum DSDModel_e
	{
		NLogNormal,
		detailed
	};

private:

	double x, xx, k, sinx, cosx;
	complex<double> m, z, mm, mm_one_o_mmx;
	int N_max;	
	
	TVector < complex<double> > A, a, b;

	class Vector psi, chi;
	class Vector pi, tau;
	class Vector TwoNp1oNNp1;

	bool allocate(int n);
	// complex<double> An(int n);

	// parameters
	double Qext;
	double Qsca, Qsca_sum;
	double ssa;
	double g;
	
	// Legendre polynomial expansion
	class Matrix LegendrePolynomialMatrix;

	bool createLegendrePolynomialMatrix(int resolution, int no_coefficients, bool recreate=false);

	// FIRST set x
	bool setX(const double& x);
	inline bool setX(const double& r, const double& lambda){return setX((2.0*M_PI*r)/lambda);}

	// THEN refractive index
	bool setRefractiveIndex(const complex<double>& m);
	inline bool setRefractiveIndex(const double& m_re, const double& m_im){return setRefractiveIndex(complex<double>(m_re, -m_im));}

	void calculateAnPsiChi();
public:

	MieModule();
	~MieModule();

	bool setParameters(const double& lambda, const double& m_re, const double& m_im, const double &radius);
	
	int calculateNOCoefficients(const double& x);
	inline int getNOCoefficients(){return N_max;}

	void calculateMieCoefficients();	// calculate Ricatti-Bessel functions, logarithmic derivatives and finally the Mie coefficients
	void printMieCoefficients(FILE *fptr);

	void calculatePiAndTau(const double& mu, double* pi, double *tau, const int& n);
	inline void calculatePiAndTau(const double& mu){calculatePiAndTau(mu, pi.getCoefficientPtr(), tau.getCoefficientPtr(), N_max);}

	void calculateTwoNp1oNNp1(double *TwoNp1oNNp1, const int& n);
	inline void calculateTwoNp1oNNp1(){calculateTwoNp1oNNp1(TwoNp1oNNp1.getCoefficientPtr(), N_max);}

	// access to internal variables
	inline complex<double> *getAPtr(){return a.getCoefficientPtr();}
	inline complex<double> *getBPtr(){return b.getCoefficientPtr();}
	inline double *getPiPtr(){return pi.getCoefficientPtr();}
	inline double *getTauPtr(){return pi.getCoefficientPtr();}
	inline double *getTwoNp1oNNp1Ptr(){return TwoNp1oNNp1.getCoefficientPtr();}

	inline double getExtinctionEfficiency(){return Qext;}
	inline double getScatteringEfficiency(){return Qsca;}
	inline double getScatteringEfficiencySum(){return Qsca_sum;}
	inline double getSingleScatteringAlbedo(){return ssa;}
	inline double getAsymmetryParameter(){return g;}
	
		
	// void calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P33, double& P34, complex<double> *a, complex<double> *b, double *pi, double *tau, int n);
	void calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P33, double& P34, complex<double> *a, complex<double> *b, double *pi, double *tau, int n);

	inline void calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P33, double &P34)
		{calculatePhaseMatrixElements(mu, P11, P12, P33, P34, a.getCoefficientPtr(), b.getCoefficientPtr(), pi.getCoefficientPtr(), tau.getCoefficientPtr(), N_max);}

	void calculatePhaseMatrix(const double& mu, class StokesMatrix& PhaseMatrix, bool zero_rest=true);
	bool developLegendrePhaseFunctions(LegendrePhaseFunction& F11, LegendrePhaseFunction& F12, LegendrePhaseFunction& F33, LegendrePhaseFunction& F34, const int& no_coefficients, const int& resolution);
	
	// bool processLogNormalDSD(const double& lambda, const double& mean_radius, const double& sigma, const double& re, const double& im, const int& res_DSD, const int& res_phf, bool doLPE=false, int no_LPE_coefficients=100);
	bool processLogNormalDSD(const double& lambda, const double& mean_radius, const double& sigma, const double& re, const double& im, const int& res_DSD, const int& res_phf);
	bool processLogNormalDSDHGPhF(const double& lambda, struct Mode_s* mode, int no_modes, const double& re, const double& im, const int& res_DSD);

	void calculateCrossSectionSSAAndG(char *bins_file, char *data_file);
	void calculateBulkDetailedPhaseFunction(char *bins_file, char *data_file, int Nphf);
};

#endif // _MieModule_h


