/**************************************************************************************************************
*
*       This file,
*
*                            /MolecularRRSCrossSection.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "MolecularRRSCrossSection.h"
#include "error.h"

int MolecularRRSCrossSection::J_max=30;
class Matrix MolecularRRSCrossSection::PlaczekTellerCoefficients;
/*
struct MolecularRRSCrossSection::Parameters_s MolecularRRSCrossSection::DAMP[2]=
{
	MolecularRRSCrossSection::Parameters_s(1e-31, -6.01466, 2385.57, 186.099, 1.98973e2, 6, 3, 1),	// N2
	MolecularRRSCrossSection::Parameters_s(1e-30, 0.07149, 45.9364, 48.2716, 1.4378e2, 0, 1, 0)	// O2
};

MolecularRRSCrossSection::Parameters_s::Parameters_s(const double& A, const double& B, const double& C, const double& D, \
	const double& B0, const int& gJe, const int& gJo, const int& I)
{
	gamma.A=A;
	gamma.B=B;
	gamma.C=C;
	gamma.D=D;
	this->B0=B0;
	this->gJe=gJe;
	this->gJo=gJo;
	this->I=I;
}

bool MolecularRRSCrossSection::Parameters_s::allocateDeltaB0JJp1(const int& J_max)
{
	if(!DeltaB0JJp1.allocate(J_max, 2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// calculate coefficients
	int j, four_j;
	double *pDeltaB0JJp1=DeltaB0JJp1[0];
	for(j=0;j<J_max;j++)
	{
		four_j=4*j;
		*pDeltaB0JJp1++=B0*double(four_j+6);
		*pDeltaB0JJp1++=B0*double(2-four_j);
	}
	return true;
}
*/

bool MolecularRRSCrossSection::setJMax(const int& n)
{
	if(!PlaczekTellerCoefficients.allocate(n, 2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	J_max=n;

	int j;
	double *pPTC=PlaczekTellerCoefficients[0];
	for(j=0;j<J_max;j++)
	{
		*pPTC++=double(3*(j+1)*(j+2))/double(2*(2*j+1)*(2*j+3));
		*pPTC++=double(3*j*(j-1))/double(2*(2*j+1)*(2*j-1));
	}
	return true;
}

bool MolecularRRSCrossSection::allocate(class Molecule *pM)
{
	if(!wT.allocate(J_max) || \
		!lambda.allocate(J_max, 4) || \
		!sigma.allocate(J_max, 4) || \
		!scattering_probability.allocate(J_max, 2) || \
		!cumulative_scattering_probability.allocate(J_max, 2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	pMolecule=pM;
	return true;
}

void MolecularRRSCrossSection::print(const int& j, FILE *fptr)
{
	fprintf(fptr, "\t%le", wT[j]);
	fprintf(fptr, "\t%le\t%le", sigma_out, sigma_in);
	for(int i=0;i<4;i++)
		fprintf(fptr, "\t%le\t%le\t%le\t%le", lambda[j][i], sigma[j][i], scattering_probability[j][i], cumulative_scattering_probability[j][i]);
}

void MolecularRRSCrossSection::setTemperature(const double& temperature)
{
/*
	calculates thermodynamic weights (wT vector)
*/
	int j, twoIp1=2*pMolecule->_I()+1;
	double hcB_0ok_BT = ( const_h * const_c * pMolecule->_B0()) / ( const_kB * temperature ) ;
	double w_norm=double ( twoIp1 * twoIp1 ) / ( 2.0 * hcB_0ok_BT );
	double *pwT=wT.getCoefficientPtr();
	for(j=0;j<J_max;j++)
		*pwT++ = (( j%2 ? pMolecule->_gJo() : pMolecule->_gJe() ) * double(2*j+1) * exp(-hcB_0ok_BT*double(j*(j+1)))) / w_norm;
}

void MolecularRRSCrossSection::calculateCrossSections(const double& wavelength)
{
	if(this->wavelength==wavelength)
		return;
	int j;

	// lambda in/out  plus/minus
	double lambda_out_plus, lambda_out_minus, lambda_in_plus, lambda_in_minus;
	double *plambda=lambda[0];

	// DeltaB0JJp1
	double DeltaB0JJp1_plus, DeltaB0JJp1_minus, *pDeltaB0JJp1=pMolecule->_DeltaB0JJp1()[0];

	// gamma^2
	double gamma_squared, gamma_squared_center=pMolecule->calculateGamma(wavelength);
	gamma_squared_center*=gamma_squared_center;

	// k and k^4
	double k_center=1.0/wavelength;
	double l4, k4_center=k_center*k_center;
	k4_center*=k4_center;

	// b
	double wT;
	double *pb=PlaczekTellerCoefficients[0];
	double wb_plus, wb_minus;

	// w
	double *pwT=this->wT.getCoefficientPtr();
	
	// sigma in/out  plus/minus
	double sigma_out_plus, sigma_out_minus, sigma_in_plus, sigma_in_minus;
	double *psigma=sigma[0];

	// scattering_probability
	double *psp=scattering_probability[0];

	sigma_in=sigma_out=0.0;
	for(j=0;j<J_max;j++)
	{
		DeltaB0JJp1_plus=*pDeltaB0JJp1++;
		DeltaB0JJp1_minus=*pDeltaB0JJp1++;
		/* 
			precalculates:

				lambda_x,out^+(lambda,J)=1/(1/lambda-B_{0X}*DeltaJ^+)
				lambda_x,out^-(lambda,J)=1/(1/lambda-B_{0X}*DeltaJ^-)

				lambda_x,in^+(lambda,J)=1/(1/lambda+B_{0X}*DeltaJ^+)
				lambda_x,in^-(lambda,J)=1/(1/lambda+B_{0X}*DeltaJ^-)

		*/
		*plambda++=lambda_out_plus	=	1.0 / ( k_center - DeltaB0JJp1_plus );
		*plambda++=lambda_out_minus	=	1.0 / ( k_center - DeltaB0JJp1_minus );
		*plambda++=lambda_in_plus	=	1.0 / ( k_center + DeltaB0JJp1_plus );
		*plambda++=lambda_in_minus	=	1.0 / ( k_center + DeltaB0JJp1_minus );

		// get wT
		wT=*pwT++;

		//  get b^+ and b^-, multiply with wT
		wb_plus=wT**pb++;
		wb_minus=wT**pb++;

		/*
			1. cross section for out-scattering: sigma_out (free path length sampling)

				sigma_out=
					256*pi^5/27 
					* gamma_x^2(lambda) 
					---------------------------------
					* sum_J
						w_x(J,T) 
						*
						{
							b^+_J
							/
							lambda_x,out^{+4}(lambda,J)
							+
							b^-_J
							/
							lambda_x,out^{-4}(lambda,J)
						}

				sigma_out_plus=
					256*pi^5/27 
					* gamma_x^2(lambda)
					--------------------
					* w_x(J,T) 
					* b^+_J
					/ lambda_x,out^{+4}(lambda,J)

				sigma_out_minus=
					256*pi^5/27 
					* gamma_x^2(lambda) 
					--------------------
					* w_x(J,T) 
					* b^-_J
					/ lambda_x,out^{+4}(lambda,J)

		*/
		l4=lambda_out_plus;
		l4*=l4;
		l4*=l4;
		*psigma++=sigma_out_plus=wb_plus/l4;
		sigma_out+=sigma_out_plus;

		l4=lambda_out_minus;
		l4*=l4;
		l4*=l4;
		*psigma++=sigma_out_minus=wb_minus/l4;
		sigma_out+=sigma_out_minus;

		/*
			2. cross section for in-scattering: sigma_in

				sigma_in=
					256*pi^5/27 
					* 1/lambda^4
					-----------------------------
					* sum_J

						w_x(J,T) 
						*
						{
							b^+
							* gamma_x^2(lambda_x,in^+(lambda,J))
							+
							b^-_J
							* gamma_x^2(lambda_x,in^-(lambda,J))
						}

				sigma_in_plus=
					* 256*pi^5/27 
					* 1/lambda^4
					---------------------
					* w_x(J,T) 
					* b^+
					* gamma_x^2(lambda_x,in^+(lambda,J))

				sigma_in=
					256*pi^5/27 
					* 1/lambda^4
					------------------
					* w_x(J,T) 
					* b^-_J
					* gamma_x^2(lambda_x,in^-(lambda,J))

		*/
		gamma_squared=pMolecule->calculateGamma(lambda_in_plus);
		gamma_squared*=gamma_squared;
		*psigma++=sigma_in_plus=wb_plus*gamma_squared;
		*psp++=sigma_in_plus;
		sigma_in+=sigma_in_plus;

		gamma_squared=pMolecule->calculateGamma(lambda_in_minus);
		gamma_squared*=gamma_squared;
		*psigma++=sigma_in_minus=wb_minus*gamma_squared;
		*psp++=sigma_in_minus;
		sigma_in+=sigma_in_minus;
		
		// plambda+=4;
	}

	// finish calculations of single line sigma out/in and normalize scattering_probability, calculate scattering CDF
	psp=scattering_probability[0];
	psigma=sigma[0];
	double factor_out=const_RRSC*gamma_squared_center;
	double factor_in=const_RRSC*k4_center;
	double csp=0.0, sp, *pcsp=cumulative_scattering_probability[0];
	for(j=0;j<J_max;j++)
	{
		// sigma_out
		*psigma++*=factor_out;
		*psigma++*=factor_out;

		// sigma_in
		*psigma++*=factor_in;
		*psigma++*=factor_in;

		// normalize scattering probabilitie and calculate CDF
		// Stokes
		sp=*psp/sigma_in;
		csp+=sp;

		*psp++=sp;
		*pcsp++=csp;

		// Anti-Stokes
		sp=*psp/sigma_in;
		csp+=sp;

		*psp++=sp;
		*pcsp++=csp;
	}

	// finish total RRS cross sections
	sigma_out*=factor_out;
	sigma_in*=factor_in;

	this->wavelength=wavelength;
}

double MolecularRRSCrossSection::sampleWavelength(const double& p)
{
	int j;
	double *pcsp=cumulative_scattering_probability[0];
	for(j=0;j<J_max;j++)
	{
		if(p<*pcsp++)
			return lambda[j][in_Stokes];
		else if(p<*pcsp++)
			return lambda[j][in_AntiStokes];
	}
	errorIn(__FILE__, __LINE__, __func__);
	fprintf(stderr, "sampled center wavelength\n");
	return wavelength;
}



