/**************************************************************************************************************
*
*       This file,
*
*                            /Molecule.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Molecule.h"
#include "error.h"

#include <math.h>

#define Loschmidt		2.6867805e19	// 1/cm³ @ 760Torr (1013.25hPa), 0°C
#define epsilon0		8.854187817e-12
#define twenty_four_pipipi	744.150640327

double Molecule::lambda_mum=0.0;
double Molecule::lambda_cm=0.0;
double Molecule::ss=0.0;

Molecule::Molecule()
{
	B0=0.0;
	gJe=0;
	gJo=0;
	I=0;
}

bool Molecule::setWavelength(const double& lambda)
{
	lambda_mum=1e6*lambda;
	if(lambda_mum>2.06 || lambda_mum<0.181)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "wavelength %leµm for N2 refractive index \
			calculation out of possible range [0.181,2.06]!\n", lambda_mum);
		return false;
	}
	else if(lambda_mum<0.198 || lambda_mum>1.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "wavelength %leµm for O2 refractive index \
			calculation out of possible range [0.198,1.0]!\n", lambda_mum);
		return false;
	}
	lambda_cm=1e2*lambda;
	double s=1.0/lambda_mum;
	ss=s*s;
	return true;
}

// parameter functions ###################################################################

// RAYLEIGH SCATTERING BY AIR
	// D. R. Bates 
	// Planet. Space Sa., Vol.32, No.6, pp. 785-790, 1984 

// !!!!!!!!!!!!! refractive indices at 760 Torr and O°C	!!!!!!!!!!!!!!!!!!

double Molecule::calculateN2RefractiveIndex(const double& lambda_mu, const double& ss)
{
	if(lambda_mum>0.468)
		return 1.0+1e-8*(6855.2+3243157.0/(144.0-ss));
	else if(lambda_mum>0.254)
		return 1.0+1e-8*(5989.242+3363266.3/(144.0-ss));
	// else if(lambda_mum>0.181)
	return 1.0+1e-8*(6998.749+3233582.0/(144.0-ss));
}

double Molecule::calculateO2RefractiveIndex(const double& lambda_mu, const double& ss)
{
	if(lambda_mum>0.546)
		return 1.0+1e-8*(21351.1+218567.0/(40.9-ss));
	else if(lambda_mum>0.288)
		return 1.0+1e-8*(20564.8+248089.9/(40.9-ss));
	else if(lambda_mum>0.221)
		return 1.0+1e-8*(22120.4+203187.6/(40.9-ss));
	// else if(lambda_mum>0.198)
	return 1.0+1e-8*(23796.7+168988.4/(40.9-ss));
}

double Molecule::calculateArRefractiveIndex(const double& lambda_mu, const double& ss)
{
	return 1.0+1e-7*(678.6711+301829.43/(144.0-ss));
}

double Molecule::calculateN2KingFactor(const double& ss)
{
	return 1.034+3.17e-4*ss;
}

double Molecule::calculateO2KingFactor(const double& ss)
{
	return 1.096 + ss*(1.385e-3 + 1.448e-4*ss);
}

double Molecule::calculateArKingFactor(const double& ss)
{
	return 1.0;
}

double Molecule::calculateN2Gamma(const double& ss)
{
	return 1e-25*(-6.01466+2385.57/(186.099-ss));
}

double Molecule::calculateO2Gamma(const double& ss)
{
	return 1e-24*(0.07149+45.9364/(48.2716-ss));
}

double Molecule::calculateArGamma(const double& ss)
{
	return 0.0;
}

// #########################################################################################################

bool Molecule::allocateDeltaB0JJp1(const int& J_max)
{
	if(!DeltaB0JJp1.allocate(J_max, 2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	// calculate coefficients
	int j, four_j;
	double *pDeltaB0JJp1=DeltaB0JJp1[0];
	for(j=0;j<J_max;j++)
	{
		four_j=4*j;
		*pDeltaB0JJp1++=B0*double(four_j+6);
		*pDeltaB0JJp1++=B0*double(2-four_j);
	}
	return true;
}

bool Molecule::initialize(const enum Molecule::Molecule_e& M, const double& vmr, const int& J_max)
{
	switch(M)
	{
		case N2: // Nitrogen 
			B0=1.98973e2;
			gJe=6;
			gJo=3;
			I=1;
			calculateRefractiveIndex=&calculateN2RefractiveIndex;
			calculateKingFactor=&calculateN2KingFactor;
			calculateGammaInternal=&calculateN2Gamma;
			name="N2";
			break;

		case O2: // Oxygen
			B0=1.4378e2;
			gJe=0;
			gJo=1;
			I=0;
			calculateRefractiveIndex=&calculateO2RefractiveIndex;
			calculateKingFactor=&calculateO2KingFactor;
			calculateGammaInternal=&calculateO2Gamma;
			name="O2";
			break;

		case Ar: // Argon from Peck and Fisher
			calculateRefractiveIndex=&calculateArRefractiveIndex;
			calculateKingFactor=&calculateArKingFactor;
			calculateGammaInternal=&calculateArGamma;
			name="Ar";
			break;

		default:
		case CO2:
			name="CO2";
			errorIn(__FILE__, __LINE__, __func__);
			return false;
				
	}
	this->vmr=vmr;
	return allocateDeltaB0JJp1(J_max);
}

void Molecule::calculateOpticalProperties()
{
	n=(*calculateRefractiveIndex)(lambda_mum, ss);
	KingFactor=(*calculateKingFactor)(ss);
	epsilon=4.5*(KingFactor-1.0);
	f_Cabannes=(18.0+epsilon)/(18.0+4.0*epsilon);

	// calculate (n**2-1)/(n**2+2)
	double c=(n*n-1.0)/(Loschmidt*(n*n+2.0));
	
	/*alpha=3e6*c*epsilon0;
	gamma=alpha*sqrt(epsilon);*/

	c/=lambda_cm*lambda_cm;

	sigma_Rayleigh=twenty_four_pipipi*(c*c)*KingFactor;
	sigma_Cabannes=f_Cabannes*sigma_Rayleigh;
}

/*double Molecule::calculateGamma(const double& lambda)
{
	double lambda_mum=1e6*lambda;
	double s=1.0/lambda_mum;
	double ss=s*s;

	double n=(*calculateRefractiveIndex)(lambda_mum, ss);
	double KingFactor=(*calculateKingFactor)(ss);
	
	return (3e6*(n*n-1.0)*epsilon0*sqrt(4.5*(KingFactor-1.0)))/(Loschmidt*(n*n+2.0));
}*/

double Molecule::calculateGamma(const double& lambda)
{
	double s=1e-6/(lambda);
	double ss=s*s;
	return 1e-4*(*calculateGammaInternal)(ss);
}

double Molecule::KingCorrectionFactorOfAir(const double& lambda, const double& xCO2)
{
	// F(lambda)
	// D.R. Bates, Rayleigh scattering by Air, Planet. Space Science 1984
	/*double nu_vac=1/(lambda*1e2);
	return 1.0367+5.381e-12*pow(nu_vac,2.0)+0.304e-20*pow(nu_vac,4.0);*/

	double pCO2=1e-3*xCO2;
	double lambda2=1e12*lambda*lambda;
	double lambda4=lambda2*lambda2;

	double F_N2=1.034+3.17e-4/lambda2;
	double F_O2=1.096+1.385e-3/lambda2+1.448e-4/lambda4;
	return (78.084*F_N2+20.946*F_O2+0.934+pCO2*1.15)/(99.964+pCO2);
}

void Molecule::logParameters(FILE *fptr)
{
/*	
		// kinetic constants
		double B0;
		int gJe, gJo;
		int I;

		//  higher level kinetic parameters
		class Matrix DeltaB0JJp1;

		// polarizabilities, macroscopic refractive index
		double vmr;
		double n;		// refractive index
		double KingFactor;	// King Factor
		double alpha, gamma;

		double epsilon;		// ( gamma over alpha ) squared
		double f_Cabannes;
		
		// optical properties derived from the polarizabilities 
		double sigma_Rayleigh;
		double sigma_Cabannes;
*/
	fprintf(fptr, "######################## [ %s ] ##########################\n", name);
	fprintf(fptr, "B0=%lf\n", B0);
	fprintf(fptr, "gJe=%d\n", gJe);
	fprintf(fptr, "gJo=%d\n", gJo);
	fprintf(fptr, "I=%d\n", I);
	fprintf(fptr, "Delta B0JJp1:\n");
	
	DeltaB0JJp1.dumpTo(fptr);

	fprintf(fptr, "\nvmr=%lf\n",vmr);
	fprintf(fptr, "n=%lf\n", n);
	fprintf(fptr, "KingFactor=%lf\n", KingFactor);
	/*fprintf(fptr, "alpha=%le\n", alpha);
	fprintf(fptr, "gamma=%le\n", gamma);*/
	fprintf(fptr, "epsilon=%le\n", epsilon);
	fprintf(fptr, "f_Cabannes=%lf\n", f_Cabannes);
	fprintf(fptr, "sigma_Rayleigh=%le\n", sigma_Rayleigh);
	fprintf(fptr, "sigma_Cabannes=%le\n", sigma_Cabannes);
}






