/**************************************************************************************************************
*
*       This file,
*
*                            /Molecule.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Molecule_h
#define _Molecule_h

#include "Math/Matrix.h"

class Molecule
{
	public:
		enum Molecule_e
		{
			N2,
			O2,
			Ar,
			CO2
		};

	private:

		// kinetic constants
		double B0;
		int gJe, gJo;
		int I;

		//  higher level kinetic parameters
		class Matrix DeltaB0JJp1;
		bool allocateDeltaB0JJp1(const int& J_max);

		// polarizabilities, macroscopic refractive index
		double vmr;
		double n;		// refractive index
		double KingFactor;	// King Factor
		// double alpha, 
		// double gamma;

		double epsilon;		// ( gamma over alpha ) squared
		double f_Cabannes;
		
		// optical properties derived from the polarizabilities 
		double sigma_Rayleigh;
		double sigma_Cabannes;
		
		const char *name;

		double (*calculateGammaInternal)(const double& ss);
	public:

		Molecule();

		bool initialize(const enum Molecule_e& M, const double& vmr, const int& J_max);
		static bool setWavelength(const double& lambda);

		void logParameters(FILE *fptr);
		// pointer to member functions
		double (*calculateRefractiveIndex)(const double& lambda_mu, const double& ss);
		double (*calculateKingFactor)(const double& ss);

		double calculateGamma(const double& lambda);

		void calculateOpticalProperties();
	
		inline void setKingFactor(const double& FK) {epsilon=4.5*((KingFactor=FK)-1.0); f_Cabannes=(18.0+epsilon)/(18.0+4.0*epsilon);}
		static double KingCorrectionFactorOfAir(const double& lambda, const double& xCO2=450.0);
		inline void calculateKingCorrectionFactorOfAir(const double& lambda, const double& xCO2=450.0)
			{setKingFactor(KingCorrectionFactorOfAir(lambda, xCO2));}
		
		//inline access
		inline double& _vmr(){return vmr;}
		inline double& _KingFactor(){return KingFactor;}
		// inline double& _alpha(){return alpha;}
		// inline double& _gamma(){return gamma;}
		inline double& _f_Cabannes(){return f_Cabannes;}
		inline double& _epsilon(){return epsilon;}
		inline double& _sigma_Rayleigh(){return sigma_Rayleigh;}
		inline double& _sigma_Cabannes(){return sigma_Cabannes;}

		inline class Matrix& _DeltaB0JJp1(){return DeltaB0JJp1;}
		inline double& _B0(){return B0;}
		inline int& _gJe(){return gJe;}
		inline int& _gJo(){return gJo;}
		inline int& _I(){return I;}
		
	private:

		static double lambda_mum, lambda_cm, ss;

		// refractive index parametrisations
		static double calculateN2RefractiveIndex(const double& lambda_mu, const double& ss);
		static double calculateO2RefractiveIndex(const double& lambda_mu, const double& ss);
		static double calculateArRefractiveIndex(const double& lambda_mu, const double& ss);
		
		static double calculateN2KingFactor(const double& ss);
		static double calculateO2KingFactor(const double& ss);
		static double calculateArKingFactor(const double& ss);

		static double calculateN2Gamma(const double& ss);
		static double calculateO2Gamma(const double& ss);
		static double calculateArGamma(const double& ss);
};

#endif //_Molecule_h




