/**************************************************************************************************************
*
*       This file,
*
*                            /Particle.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Particle.h"
#include <math.h>

#include "error.h"

Particle::Particle()
{
	IOName="particle";
	b_has_single_phase_function=true;

	model=optical;
	PhaseFunctionType=HenyeyGreenstein;	// model of the phase function
	// g=0.7;					// asymmetry parameter
	no_Legendre_expansion_moments=10;		// number of Legendre expansion moments for both, optical and micro physical particle definition;
	phase_function_resolution=1000;			// cos(scatter_angle) resolution of the phase function
	b_use_MieLUT=false;				// use a look up table for the Mie calculations?
	confidence_intervall=4.0;			// number of sigmas to be considered around a maximum for DSD Mie calculations
	no_DSD_bins=100;				// number of bins for droplet size distribution Mie calculations

	// setOpticalClusterDimension(3);

	HGPhaseFunction=0;
	LEPhaseFunction=0;
	MPhaseFunction=0;
	HGQTPhaseFunction=0;
	PhaseFunctionPtr=0;
	OpticalStruct=0;

	DSDModel=MieModule::NLogNormal;

	addParameter(name, type_string, "name", 0, "name of the particle");
	addParameter(&model, type_enum, "model", "optical|microphysical", "input data model");
	addParameter(&b_has_single_phase_function, type_bool, "b_has_single_phase_function|b_single_PhF|b_one_PhF", 0, "use one phase function for the hole atmosphere?");
	addParameter(&PhaseFunctionType, type_enum, "PhaseFunctionType|PhFType|PhF", "HG|Mie|Rayleigh|Cabannes|RR|LE|HGQTable|Table", "model of the phase function");

	addParameter(&g, type_double, "asymmetry_parameter|g", 0, "asymmetry parameter");
	addParameter(&h, type_double, "h", 0, "in [0.5,1], relative contribution of the copolarized component to the total backscattered light if the incident light is completely linearly polarized");
	addParameter(&p, type_double, "p", 0, "in [0,1], maximum degree of linear polarization for incident unpolarized light");

	addParameter(&single_scattering_albedo, type_double, "single_scattering_albedo|ssa", 0, "single scattering albedo in [0,1]");
	addParameter(&phase_function_resolution, type_int, "phase_function_resolution|PhF_resolution|PhFRes", 0, "cos(scatter_angle) resolution of the phase function");
	addParameter(&DSDModel, type_enum, "DSDModel", "NLogNormal|detailed", "DSD model row format of fn_droplet_size_distribution: either NLogNormal (N_i<TAB>r_i<TAB>sigma_i<RET>) or detailed (N_i<RET>)");
	addParameter(fn_droplet_size_distribution, type_file, "fn_droplet_size_distribution|fn_DSD", 0, "DSD file, format depends on DSDModel");
	addParameter(fn_DSD_radius_list, type_file, "fn_DSD_radius_list|fn_DSD_radii", 0, "DSD radius list file");
	
	addParameter(fn_phase_function, type_file, "fn_phase_function|fn_PhF|fn_differential_cross_section|fn_dcs|fn_phase_matrix|fn_PhM", 0, "phase function / matrix / differential cross section data");
	addParameter(&b_use_phase_function_caching, type_bool, "b_use_phase_function_caching|b_phf_caching|b_use_phase_function_cache", 0, "use phase function caching?");
	addParameter(&b_interpolate_phase_function, type_bool, "b_interpolate_phase_function", 0, "interpolate phase function?");
	// addParameter(&b_use_MieLUT, type_bool, "b_use_MieLUT|b_MieLUT", 0, "use a look up table for the Mie calculations?");
	// addParameter(fn_MieLUT, type_file, "fn_MieLUT", 0, "Mie look up table (LUT) filename");
	addParameter(&confidence_intervall, type_double, "confidence_intervall|DSD_sigma", 0, "number of sigmas to be considered around a maximum for DSD Mie calculations");
	addParameter(fn_refractive_index, type_file, "fn_refractive_index|fn_refin", 0, "filename for the wavelength dependent complex refractive index");
	addParameter(&no_DSD_bins, type_int, "no_dsd_bins|no_DSD_bins", 0, "number of bins for droplet size distribution Mie calculations");
	addParameter(&no_Legendre_expansion_moments, type_int, "no_Legendre_expansion_moments|no_LE_moments", 0, "number of Legendre expansion moments for both, optical and micro physical particle definition");
	addParameter(fn_map, type_file, "fn_map", 0, "filename for map file");
	addParameter(fn_cluster_data, type_file, "fn_cluster_data|fn_data", 0, "cluster data. format: ext.@ref TAB ssa TAB [Phf parameters sep. by TABs], e.g. for HG: g TAB h TAB p");
	addParameter((Profile *)this, type_profile, "profile|data", 0, "particle profile data, cluster row format: 1=ext coeff[1/km] 2=ssa[0-1] 3,4,5...=Phf parameters");

	h=0.5;
	p=0.5;
	g=0.5;

	b_use_phase_function_caching=false;
	b_interpolate_phase_function=false;
	single_scattering_albedo=1.0;
}

Particle::~Particle()
{
	if(HGPhaseFunction)	delete [] HGPhaseFunction;
	if(MPhaseFunction)	delete [] MPhaseFunction;
	if(LEPhaseFunction)	delete [] LEPhaseFunction;
	if(HGQTPhaseFunction)	delete [] HGQTPhaseFunction;
	if(PhaseFunctionPtr)	delete [] PhaseFunctionPtr;
	if(OpticalStruct)	delete [] OpticalStruct;
}

bool Particle::loadFiles()
{
	Profile::setProfileName(name);
	int i, j;
	char filename[BasicIO_string_size];
	// fn_cluster_data
	formatFilePath(filename, fn_cluster_data);
	if(!loadClusterData(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load cluster data of particle \"%s\" from file \"%s\"!\n", name, filename);
		return false;
	}
	if(b_has_single_phase_function && getClusterDimension()!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "contradiction between cluster dimension and b_has_single_phase_function flag for particle \"%s\" defined in file \"%s\"!\n", name, filename);
		return false;
	}

	// fn_map
	formatFilePath(filename, fn_map);
	if(!ClusterMap::load(filename) || !initializeClusterArray())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load and initialize map data of particle \"%s\" from file \"%s\"!\n", name, filename);
		return false;
	}

	if(OpticalStruct)
		delete [] OpticalStruct;
	OpticalStruct=new struct OpticalStruct_s[getNOClusters()];
	if(!OpticalStruct)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating OpticalStructs\n");
		return false;
	}
	if(!allocatePhaseFunctions())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	char phase_function_name[100];
	strcpy(phase_function_name, name);
	if(model==optical)
	{
		// set phase function parameters
		if(b_has_single_phase_function)
		{
			if(getClusterDimension()!=1)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "data file \"%s\" has wrong number of columns, but should have 1\n", fn_cluster_data);
				return false;
			}
			// set single scattering albedo
			class PhaseFunction *pSinglePhaseFunction=0;
			switch(PhaseFunctionType)
			{
				case HenyeyGreenstein:
					if(!SingleHGPhaseFunction.defineParameter(0, g) ||
						!SingleHGPhaseFunction.defineParameter(1, h) ||
						!SingleHGPhaseFunction.defineParameter(2, p))
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "error defining phase function parameters according to \"%s\"!\n", filename);
						return false;
					}
					pSinglePhaseFunction=&SingleHGPhaseFunction;
					break;

				case LegendreExpansion:
					formatFilePath(filename, fn_phase_function);
					if(!SingleLEPhaseFunction.loadData(filename))
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "could not load Legendre phase function from file \"%s\"!\n", filename);
						return false;
					}
					SingleLEPhaseFunction.enableCache(b_use_phase_function_caching, phase_function_resolution);
					SingleLEPhaseFunction.enableInterpolation(b_interpolate_phase_function);
					pSinglePhaseFunction=&SingleLEPhaseFunction;
					break;

				case HGQTable:
					formatFilePath(filename, fn_phase_function);
					if(!SingleHGQTablePhaseFunction.loadData(filename))
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "could not load HG quadrature phase function from file \"%s\"!\n", filename);
						return false;
					}
					single_scattering_albedo=SingleHGQTablePhaseFunction.getSingleScatteringAlbedo();
					g=SingleHGQTablePhaseFunction.getAsymmetryParameter();
					pSinglePhaseFunction=&SingleHGQTablePhaseFunction;
					break;


				case Table:
					formatFilePath(filename, fn_phase_function);
					if(!SingleTablePhaseFunction.loadData(filename))
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "could not load table phase function from file \"%s\"!\n", filename);
						return false;
					}
					pSinglePhaseFunction=&SingleTablePhaseFunction;
					break;
				
				default:
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "unsupported phase function type!\n");
					return false;
			}
			setNOPhaseFunctionParameters(pSinglePhaseFunction->getNOParameters());
			pSinglePhaseFunction->setName(phase_function_name);
			if(!pSinglePhaseFunction->allocateJacobianAndHessian())
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
		}// if b_has_single_phase_function
		else
		{
			if(PhaseFunctionType==HGQTable)
			{
				formatFilePath(filename, fn_phase_function);
				FILE *fptr=fopen(filename, "rb");
				if(!fptr)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "could not load differential cross section file \"%s\"!\n", filename);
					return false;
				}
				int nc;
				if(!HGQTablePhaseFunction::readHGQTableFileHeader(fptr, &nc))
				{
					fclose(fptr);
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
				if(nc!=no_clusters)
				{
					fclose(fptr);
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "incompatible number of clusters: %d in file \"%s\", %d in configuration!\n", nc, filename, no_clusters);
					return false;
				}
				for(i=0;i<no_clusters;i++)
				{
					if(!HGQTPhaseFunction[i].loadData(fptr))
					{
						fclose(fptr);
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "error reading differential cross section of cluster %d from file \"%s\"\n", i+1, filename);
						return false;
					}
					cluster[i].optical_data[ssa]=HGQTPhaseFunction[i].getSingleScatteringAlbedo();
					sprintf(phase_function_name, "%s_c%d", name, i+1);
					HGQTPhaseFunction[i].setName(phase_function_name);
				}
				no_PhF_parameters=0;
				fclose(fptr);
			}
			else
			{
				switch(PhaseFunctionType)
				{
					case HenyeyGreenstein:
						if(getClusterDimension()!=5)
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "data file \"%s\" has wrong number of columns, but should have 5\n", fn_cluster_data);
							return false;
						}
						break;

					case LegendreExpansion:
						break;

					default:
			
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "unsupported phase function type!\n");
						return false;
				}
				for(i=0;i<no_clusters;i++)
				{
					// set single scattering albedo
					cluster[i].optical_data[ssa]=cluster[i].input_data[1];
					no_PhF_parameters=PhaseFunctionPtr[i]->getNOParameters();
					sprintf(phase_function_name, "%s_c%d", name, i+1);
					PhaseFunctionPtr[i]->setName(phase_function_name);
					for(j=0;j<no_PhF_parameters;j++)
						if(!PhaseFunctionPtr[i]->defineParameter(j, cluster[i].input_data[2+j]))
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "error defining phase function parameter %d in line %d of file \"%s\"!\n", j+1, i+1, fn_cluster_data);
							return false;
						}
					if(PhaseFunctionType==LegendreExpansion)
					{
						class LegendrePhaseFunction *pLEPhaseFunction=(class LegendrePhaseFunction *)PhaseFunctionPtr[i];
						pLEPhaseFunction->enableCache(b_use_phase_function_caching, phase_function_resolution);
						pLEPhaseFunction->enableInterpolation(b_interpolate_phase_function);
					}
					if(!PhaseFunctionPtr[i]->allocateJacobianAndHessian())
					{
						errorIn(__FILE__, __LINE__, __func__);
						return false;
					}
				}
			}
		}
	}
	else// model!=optical i.e. = microphysical ###############################################################################################
	{
		if(DSDModel==MieModule::detailed)
		{
			formatFilePath(filename, fn_DSD_radius_list);
			if(!DSDRadius.load(filename))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "could not load DSD radius list\n");
				return false;
			}
		}

		if(PhaseFunctionType!=Mie)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "microphysical model requires Mie phase function as phase function type\n");
			return false;
		}
		// load refractive index
		formatFilePath(filename, fn_refractive_index);
		if(!ComplexRefractiveIndexMatrix.load(filename))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error loading complex refractive index from file \"%s\"!\n", filename);
			return false;
		}
		// 
		
		if(b_has_single_phase_function)
		{
			if(ComplexRefractiveIndexMatrix.getN()!=3)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "complex refractive index file \"%s\" must have three columns for b_has_single_phase_function=yes\n", fn_refractive_index);
				return false;
			}
			// load DSD from file
			formatFilePath(filename, fn_droplet_size_distribution);
			if(!MieDSDData.load(filename))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "could not load droplet size distribution from file \"%s\"\n", filename);
				return false;
			}
			SingleMPhaseFunction.initializeComplexRefractiveIndex(ComplexRefractiveIndexMatrix, 0);
			SingleMPhaseFunction.enableCache(b_use_phase_function_caching, phase_function_resolution);
			SingleMPhaseFunction.enableInterpolation(b_interpolate_phase_function);
			strcpy(phase_function_name, name);
			SingleMPhaseFunction.setName(phase_function_name);
			SingleMPhaseFunction.defineDSDModel(DSDModel);

			switch(DSDModel)
			{
				default:
				case MieModule::NLogNormal:
					if(MieDSDData.getN()!=3)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "droplet size distribution file \"%s\" has %d columns.\nCorrect row format: N<TAB>r<TAB>log. sigma<RET>\n", filename, MieDSDData.getN());
						return false;
					}
					no_DSD_modes=MieDSDData.getM();	// number of rows = number of modes
					SingleMPhaseFunction.setQuadratureResolution(no_DSD_bins);
					SingleMPhaseFunction.setModePtr((struct MieModule::Mode_s *)MieDSDData.getDataPtr(), no_DSD_modes);
					break;

				case MieModule::detailed:
					if(MieDSDData.getN()!=1 || MieDSDData.getM()!=DSDRadius.getN())
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "droplet size distribution file \"%s\" has %d columns, but the \"detailed\" DSDModel \
							requires one column and as mayn rows as in %s\n", filename, MieDSDData.getN(), fn_DSD_radius_list);
						return false;
					}
					SingleMPhaseFunction.defineRadiusList(DSDRadius.getCoefficientPtr(), DSDRadius.getN());
					SingleMPhaseFunction.defineBinCounts(MieDSDData.getCoefficientPtr());
					break;
				
			}
		}
		else	// does not have single phase function
		{
			if(ComplexRefractiveIndexMatrix.getN()!=1+2*no_clusters)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "complex refractive index file \"%s\" must have 1 + 2 * #clusters (i.e. %d) columns\n", fn_refractive_index, 1+2*no_clusters);
				return false;
			}
			for(i=0;i<no_clusters;i++)
			{
				MPhaseFunction[i].initializeComplexRefractiveIndex(ComplexRefractiveIndexMatrix, i);
				MPhaseFunction[i].enableCache(b_use_phase_function_caching, phase_function_resolution);
				MPhaseFunction[i].enableInterpolation(b_interpolate_phase_function);
				sprintf(phase_function_name, "%s_c%d", this->name, i+1);
				MPhaseFunction[i].setName(phase_function_name);
				MPhaseFunction[i].defineDSDModel(DSDModel);
			}
			switch(DSDModel)
			{
				default:
				case MieModule::NLogNormal:
					// model=microphysical
					if(InputData.getN()%3!=0)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "micro physical model requires following input format: N1<TAB>r1<TAB>sigma1<TAB>...<TAB>Nn<TAB>rn<TAB>sigman<TAB>\n");
						return false;
					}
					no_DSD_modes=InputData.getN()/3;
					// assign pointers
					for(i=0;i<no_clusters;i++)
					{
						MPhaseFunction[i].setQuadratureResolution(no_DSD_bins);
						MPhaseFunction[i].setModePtr((struct MieModule::Mode_s *)cluster[i].input_data, no_DSD_modes);
					}
					break;
				case MieModule::detailed:
					if(InputData.getN()!=DSDRadius.getN())
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "number of input data columns in %s must equal number of rows in %s\n", fn_cluster_data, fn_DSD_radius_list);
						return false;
					}
					for(i=0;i<no_clusters;i++)
					{
						MPhaseFunction[i].defineRadiusList(DSDRadius.getCoefficientPtr(), DSDRadius.getN());
						MPhaseFunction[i].defineBinCounts(cluster[i].input_data);
					}
					break;
			}
		}
	}
	return true;
}

bool Particle::allocatePhaseFunctions()
{
	if(PhaseFunctionPtr)
		delete [] PhaseFunctionPtr;
	PhaseFunctionPtr=new PhaseFunction *[no_clusters];
	if(!PhaseFunctionPtr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "particle \"%s\": could not allocate %d phase function ptrs!\n", name, no_clusters);
		return false;
	}
	int i;
	if(b_has_single_phase_function)
	{
		class PhaseFunction *SinglePhaseFunctionPtr;
		switch(PhaseFunctionType)
		{
			case HenyeyGreenstein:
				SinglePhaseFunctionPtr=&SingleHGPhaseFunction;
				break;

			case LegendreExpansion:
				SinglePhaseFunctionPtr=&SingleLEPhaseFunction;
				break;

			case HGQTable:
				SinglePhaseFunctionPtr=&SingleHGQTablePhaseFunction;
				break;

			case Table:
				SinglePhaseFunctionPtr=&SingleTablePhaseFunction;
				break;

			case Mie:
				SinglePhaseFunctionPtr=&SingleMPhaseFunction;
				break;

			default:
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "phase function model not yet implemented!\n");
				return false;
		}
		for(i=0;i<no_clusters;i++)
			PhaseFunctionPtr[i]=SinglePhaseFunctionPtr;
	}
	else
		switch(PhaseFunctionType)
		{
			case HenyeyGreenstein:
				if(HGPhaseFunction)
					delete [] HGPhaseFunction;
				HGPhaseFunction=new HenyeyGreensteinPhaseFunction[no_clusters];
				if(!HGPhaseFunction)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "particle \"%s\": could not allocate %d HG phase functions\n", name, no_clusters);
					return false;
				}
				for(i=0;i<no_clusters;i++)
					PhaseFunctionPtr[i]=HGPhaseFunction+i;
				break;

			case LegendreExpansion:
				if(LEPhaseFunction)
					delete [] LEPhaseFunction;
				LEPhaseFunction=new LegendrePhaseFunction[no_clusters];
				if(!LEPhaseFunction)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "particle \"%s\": could not allocate %d Legendre phase functions\n", name, no_clusters);
					return false;
				}
				for(i=0;i<no_clusters;i++)
					PhaseFunctionPtr[i]=LEPhaseFunction+i;
				break;

			case HGQTable:
				if(HGQTPhaseFunction)
					delete [] HGQTPhaseFunction;
				HGQTPhaseFunction=new HGQTablePhaseFunction[no_clusters];
				if(!HGQTPhaseFunction)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "particle \"%s\": could not allocate %d Legendre phase functions\n", name, no_clusters);
					return false;
				}
				for(i=0;i<no_clusters;i++)
					PhaseFunctionPtr[i]=HGQTPhaseFunction+i;
				break;

			case Mie:
				if(MPhaseFunction)
					delete [] MPhaseFunction;
				MPhaseFunction=new MiePhaseFunction[no_clusters];
				if(!MPhaseFunction)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "particle \"%s\": could not allocate %d Mie phase functions\n", name, no_clusters);
					return false;
				}
				for(i=0;i<no_clusters;i++)
					PhaseFunctionPtr[i]=MPhaseFunction+i;
				break;

			default:
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "phase function model not yet implemented!\n");
				return false;

		}
	initializeOpticalStructs();
	return true;
}

void Particle::initializeOpticalStructs()
{
	for(int i=0;i<getNOClusters();i++)
	{
		OpticalStruct[i].pParticle=this;
		OpticalStruct[i].pPhaseFunction=PhaseFunctionPtr[i];
		OpticalStruct[i].pCluster=cluster+i;
		OpticalStruct[i].cluster_index=i;
	}
}

bool Particle::calculateOpticalStruct(struct OpticalStruct_s *pOS)
{
	struct Profile::Cluster_s *pCluster=pOS->pCluster;
	double lambda=pOS->wavelength;
	if(pCluster->valid && lambda==pCluster->optical_data[wavelength])
		return true;
	// recalculate cluster
	if(model==optical)
	{
		if(PhaseFunctionType==HGQTable)
		{
			if(b_has_single_phase_function)
			{
				pCluster->optical_data[epsilon]=pCluster->input_data[0]*1e5*SingleHGQTablePhaseFunction.getSigmaExtinction();
				pOS->varpi0=single_scattering_albedo;
				pOS->asymmetry_parameter=g;
			}
			else
			{
				class HGQTablePhaseFunction *pHGQTablePhaseFunction=HGQTPhaseFunction+pOS->cluster_index;
				pCluster->optical_data[epsilon]=pCluster->input_data[0]*1e5*pHGQTablePhaseFunction->getSigmaExtinction();
				pCluster->optical_data[ssa]=pOS->varpi0=pHGQTablePhaseFunction->getSingleScatteringAlbedo();
				pOS->asymmetry_parameter=pHGQTablePhaseFunction->getAsymmetryParameter();
			}
		}
		else
		{
			pCluster->optical_data[epsilon]=pCluster->input_data[0];
			if(b_has_single_phase_function)
			{
				// pCluster->optical_data[ssa]=single_scattering_albedo;
				if(PhaseFunctionType==HenyeyGreenstein)
				{
					pOS->pPhaseFunction->defineParameter(0, g);
					pOS->pPhaseFunction->defineParameter(1, h);
					pOS->pPhaseFunction->defineParameter(2, p);
				}
				pOS->varpi0=single_scattering_albedo;
				pOS->asymmetry_parameter=g;
			}
			else
			{
				int i, n=getNOPhaseFunctionParameters();
				for(i=0;i<n;i++)
					pOS->pPhaseFunction->defineParameter(i, pCluster->input_data[2+i]);
				pOS->varpi0=pCluster->input_data[1];
				pOS->asymmetry_parameter=pCluster->input_data[2];
				pCluster->optical_data[ssa]=pCluster->input_data[1];
				pCluster->optical_data[asymmetry_parameter]=pCluster->input_data[2];
			}
		}
		pOS->pPhaseFunction->setWavelength(lambda);
	}
	else
	{
		int cluster_index=pOS->cluster_index;
		if(b_has_single_phase_function)
		{
			if(SingleMPhaseFunction.getWavelength()!=lambda && !SingleMPhaseFunction.setWavelength(lambda))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			pCluster->optical_data[epsilon]=pCluster->input_data[0]*SingleMPhaseFunction.getExtinctionCoefficient();
			pOS->varpi0=SingleMPhaseFunction.getSingleScatteringAlbedo();
			pOS->asymmetry_parameter=SingleMPhaseFunction.getAsymmetryParameter();
		}
		else
		{
			class MiePhaseFunction *pMiePhaseFunction=MPhaseFunction+cluster_index;
			if(!pMiePhaseFunction->setWavelength(lambda))
			{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
			}
			
			pCluster->optical_data[epsilon]=pMiePhaseFunction->getExtinctionCoefficient();
			pCluster->optical_data[ssa]=pMiePhaseFunction->getSingleScatteringAlbedo();
			pCluster->optical_data[asymmetry_parameter]=pMiePhaseFunction->getAsymmetryParameter();
			pOS->varpi0=pMiePhaseFunction->getSingleScatteringAlbedo();
			pOS->asymmetry_parameter=pMiePhaseFunction->getAsymmetryParameter();
		}
	}
	// cluster optical data valid
	// printf("extinction=%le\n", pCluster->optical_data[epsilon]);
	double extinction_coefficient=pCluster->optical_data[epsilon];
	pOS->extinction_coefficient=extinction_coefficient;
	pOS->scatter_coefficient=extinction_coefficient*pOS->varpi0;
	// printf("ssa=%.5lf\n", pOS->varpi0);
	pOS->absorption_coefficient=extinction_coefficient*(1.0-pOS->varpi0);

	pCluster->optical_data[wavelength]=lambda;
	pCluster->valid=true;
	return true;
}


