/**************************************************************************************************************
*
*       This file,
*
*                            /Particle.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Particle_h
#define _Particle_h

#include "BasicIO.h"
#include "Profile.h"
#include "Math/Matrix.h"
#include "MieModule.h"
// #include "Math/LinearelyInterpolatedFunctionTemplate.h"

#include "PhaseFunctions/MiePhaseFunction.h"
#include "PhaseFunctions/HenyeyGreensteinPhaseFunction.h"
#include "PhaseFunctions/LegendrePhaseFunction.h"
#include "PhaseFunctions/TablePhaseFunction.h"
#include "PhaseFunctions/HGQTablePhaseFunction.h"

class Particle;

class Particle: public BasicIO, public Profile
{
	public:
		enum Model_e
		{
			optical,
			microphysical
		};

		struct OpticalStruct_s
		{
			class Particle *pParticle;				// associated particle
			struct Cluster_s *pCluster;				// associated cluster
			int cluster_index;

			double wavelength;					// wavelength in m

			double extinction_coefficient;				// [1/km]
			double extinction_coefficient_at_center_wavelength;	// [1/km]

			double scatter_coefficient;				// [1/km]
			double scatter_coefficient_at_center_wavelength;	// [1/km]

			double absorption_coefficient;				// [1/km]
			double absorption_coefficient_at_center_wavelength;	// [1/km]

			double varpi0;						// single scattering albedo
			double varpi0_at_center_wavelength;			// single scattering albedo at center_wavelength

			double asymmetry_parameter;

			double PhaseFunctionValue;			// Phase Function value evaluated at some mu
			class StokesMatrix ScatterPhaseMatrix;

			class PhaseFunction *pPhaseFunction;		// pointer to phase function
		};

	private:
	// IO #######################################
		char name[BasicIO_string_size];				// name of the particle
		enum Model_e model;					// input data model

		enum PhaseFunctionType_e PhaseFunctionType;		// model of the phase function
		double g;						// asymmetry parameter
		double h;						// see description
		double p;						// see description
		double single_scattering_albedo;
		
		bool b_has_single_phase_function;			// use a single phase function
		
		bool b_use_phase_function_caching;			// use phase function caching (together with phase function resulution)
		bool b_interpolate_phase_function;			// interpolate phase function
		
		enum MieModule::DSDModel_e DSDModel;
		int no_Legendre_expansion_moments;			// number of Legendre expansion moments for both, optical and micro physical particle definition;

		int phase_function_resolution;				// cos(scatter_angle) resolution of the phase function
		char fn_phase_function[BasicIO_string_size];		// phase function data

		bool b_use_MieLUT;					// use a look up table for the Mie calculations?
		char fn_MieLUT[BasicIO_string_size];			// Mie look up table (LUT) filename

		double confidence_intervall;				// number of sigmas to be considered around a maximum for DSD Mie calculations

		char fn_refractive_index[BasicIO_string_size];		// filename for the wavelength dependent complex refractive index

		int no_DSD_bins;					// number of bins for droplet size distribution Mie calculations
		char fn_droplet_size_distribution[BasicIO_string_size];	// file name for data for Mie module
		char fn_DSD_radius_list[BasicIO_string_size];		// file name for DSD radius list

		char fn_map[BasicIO_string_size];			// filename for map file
		char fn_cluster_data[BasicIO_string_size];		// filename for cluster data
	// ##################################################################################

	protected:

		enum L1Element_e
		{
			wavelength,
			epsilon,
			ssa,
			asymmetry_parameter
		};

		class PhaseFunction **PhaseFunctionPtr;	// pointers of phase functions
		
		int no_PhF_parameters;

		class MiePhaseFunction *MPhaseFunction, SingleMPhaseFunction;
		class HenyeyGreensteinPhaseFunction *HGPhaseFunction, SingleHGPhaseFunction;
		class LegendrePhaseFunction *LEPhaseFunction, SingleLEPhaseFunction;
		class TablePhaseFunction *TPhaseFunction, SingleTablePhaseFunction;
		class HGQTablePhaseFunction *HGQTPhaseFunction, SingleHGQTablePhaseFunction;

		bool allocatePhaseFunctions();

		struct OpticalStruct_s *OpticalStruct;			// OpticalStruct data

		int no_DSD_modes;

		class Matrix ComplexRefractiveIndexMatrix;
		class Matrix MieDSDData;

		// detailed DSDModel
		class Vector DSDRadius;
		
	public:

		Particle();
		~Particle();
		
		inline struct OpticalStruct_s *getOpticalStructPtr(const int& i=0){return OpticalStruct+i;}

		inline char *getName(){return name;}
		inline int getNOPhaseFunctionParameters(){return no_PhF_parameters;}
		inline void setNOPhaseFunctionParameters(const int& n){no_PhF_parameters=n;}

		bool loadFiles();
	
		inline bool hasSinglePhaseFunction(){return b_has_single_phase_function;}

		void initializeOpticalStructs();
		bool calculateOpticalStruct(struct OpticalStruct_s *pOS);
	
};

#endif //_Particle_h




