/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/GasMixturePhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "GasMixturePhaseFunction.h"

#include "../trigonometry.h"
#include "../error.h"

GasMixturePhaseFunction::GasMixturePhaseFunction()
{
	sigma=0.0;
	ACN=0.75;
	AN=0.75;
	BN=1.5;
	zp3=1.0;
}

void GasMixturePhaseFunction::setParameters(const double& sigma, const double& ACN, const double& AN, const double& BN, const double& zp3)
{
	// cross section
	this->sigma=sigma;

	// phase function
	this->ACN=ACN;
	this->AN=AN;
	this->BN=BN;
	this->zp3=zp3;
}

double GasMixturePhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
/*
	solve in principle
		u^3+p*u+q=0
	p=3*(A+C)/A
	q=(3*N-6*N*alpha)/A
	zq=q/2
	zp=p/3
	u=(-zq-(zq^2+zp^3)^2)^(1/3)-(zq+(zq^2+zp^3)^2)^(1/3)

*/
	double zq=(3.0-6.0*alpha)/(2.0*AN);
	double a=sqrt(zq*zq+zp3);
	if(pISWeight)
		*pISWeight=1.0;
	return correctMu(pow(a-zq, 1.0/3.0)-pow(a+zq, 1.0/3.0));
}

double GasMixturePhaseFunction::evaluate(const double& mu)
{
	return ACN+AN*mu*mu;
}

class StokesMatrix& GasMixturePhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
/*
	model taken from:
		Landgraf, Hasekamp, van Deelen, Aben: Rotational Raman scattering of polarized light in the Earth atmosphere:
			a vector radiative transfer model using radiative transfer pertubation theory approach

		Elsevier JQSRT 87, 399-433, (2004)
*/
	/*double h=A*(1.0+mu*mu);
	double P12=(A*(mu*mu-1.0))/N;
	
	PM[0][0]=(h+C)/N;	PM[0][1]=P12;		PM[0][2]=0.0;			PM[0][3]=0.0;
	PM[1][0]=P12;		PM[1][1]=h/N;		PM[1][2]=0.0;			PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=(2.0*A*mu)/N;		PM[2][3]=0.0;
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=0.0;			PM[3][3]=(B*mu)/N;
	*/
	double mumu=mu*mu;
	double P12=AN*(mumu-1.0);

	PM[0][0]=ACN+AN*mumu;	PM[0][1]=P12;		PM[0][2]=0.0;			PM[0][3]=0.0;
	PM[1][0]=P12;		PM[1][1]=AN*(1.0+mumu);	PM[1][2]=0.0;			PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=2.0*AN*mu;		PM[2][3]=0.0;
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=0.0;			PM[3][3]=BN*mu;

	return PM;
}

class StokesMatrix& GasMixturePhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
/*
	double A1pmumu=A*(1+mu*mu);
	double A1pmumupC=A1pmumu+C;

	double P12=A*(mu*mu-1)/A1pmumupC;
	double P22=A1pmumu/A1pmumupC;
	double P33=(2.0*A*mu)/A1pmumupC;
	double P44=(B*mu)/A1pmumupC;

	PM[0][0]=1.0;	PM[0][1]=P12;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=P12;	PM[1][1]=P22;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=P33;	PM[2][3]=0.0;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=0.0;	PM[3][3]=P44;
*/
	double mumu=mu*mu;
	double D=ACN+AN*mumu;
	double P12=AN*(mumu-1.0)/D;

	PM[0][0]=1.0;		PM[0][1]=P12;			PM[0][2]=0.0;		PM[0][3]=0.0;
	PM[1][0]=P12;		PM[1][1]=(AN*(1.0+mumu))/D;	PM[1][2]=0.0;		PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;			PM[2][2]=(2.0*AN*mu)/D;	PM[2][3]=0.0;
	PM[3][0]=0.0;		PM[3][1]=0.0;			PM[3][2]=0.0;		PM[3][3]=(BN*mu)/D;

	return PM;
}














