/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/HGQPhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HGQPhaseFunction.h"
#include "../error.h"
#include <float.h>

#include "../Math/TVector.h"
#include "../BasicIO.h"

HGQPhaseFunction::HGQPhaseFunction()
{
	wavelength=0.0;
	enable_phase_matrix_cache=false;
	enable_interpolate_phase_matrix=false;
}

void HGQPhaseFunction::writeCache()
{
	char path[BasicIO_string_size], file_name[BasicIO_string_size];
	formatOutputDirectory(path);
	sprintf(file_name, "%s%s_PhaseMatrixCache.txt", path, name);

	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error writing %s phase function cache to file %s\n", name, file_name); 
		return;
	}
	PhaseMatrixCache.dumpTo(fptr);
	fclose(fptr);
}

/*
#define 	PMC_mu	0
#define 	PMC_P11	1
#define 	PMC_P12	2
#define 	PMC_P22	3
#define 	PMC_P33	4
#define 	PMC_P34	5
#define 	PMC_P44	6
#define 	PMC_P12_P11	7
#define 	PMC_P22_P11	7
#define 	PMC_P33_P11	8
#define 	PMC_P34_P11	9
#define 	PMC_P44_P11	9
#define 	PMC_delta_mu	10
#define 	PMC_CDF	11
#define 	PMC_mu_q	12
#define 	PMC_hits	13
#define 	PMC_dimension	14
*/

int HGQPhaseFunction::calculateCacheIndex(const double& mu)
{
	int i=floor(double(phase_matrix_cache_resolution)*HenyeyGreensteinPhaseFunction::evaluateCDF(mu));
	if(i<0 || i>phase_matrix_cache_resolution)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "mu=%lf\n", mu);
		return 0;
	}
	return i;
}

bool HGQPhaseFunction::interpolateMuFromAlpha(double *pmu, const int& i, const double& alpha)
{
	// definitions I

	double mu_i=PhaseMatrixCache[i][PMC_mu];
	double mu_ip1=PhaseMatrixCache[i+1][PMC_mu];
	double delta_mu=PhaseMatrixCache[i][PMC_delta_mu];

	double P_i=PhaseMatrixCache[i][PMC_P11];
	double P_ip1=PhaseMatrixCache[i+1][PMC_P11];
	double delta_P=P_ip1-P_i;

	double muP=P_i*mu_ip1-P_ip1*mu_i;

	// definitions II
	
	double m=delta_P/delta_mu;
	double a=muP/delta_mu;
	double b=muP/delta_P;
	
	// definitions III

	double c=2.0*(2.0*alpha-PhaseMatrixCache[i][PMC_CDF]+mu_i*a)/m+mu_i*mu_i;
	
	// solution
	double radicand=c+b*b;
	if(radicand<0.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	double sr=sqrt(radicand);
	double s1=-b+sr;
	double s2=-b-sr;

	// validation
	if(s1>=mu_i && s1<=mu_ip1)
		*pmu=s1;
	else if(s2>=mu_i && s2<=mu_ip1)
		*pmu=s2;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "i=%d, F=%lf, alpha=%lf, s1=%lf, s2=%lf, mu_i=%lf, mu_ip1=%lf\n", i, 2.0*alpha, alpha, s1, s2, mu_i, mu_ip1);
		fprintf(stderr, "i=%d, mu=%lf, P11=%le, CDF=%lf, mu_q=%lf\n", i, \
			PhaseMatrixCache[i][PMC_mu], PhaseMatrixCache[i][PMC_P11], PhaseMatrixCache[i][PMC_CDF], PhaseMatrixCache[i][PMC_mu_q]);
		fprintf(stderr,"i=%d, mu=%lf, P11=%le, CDF=%lf, mu_q=%lf\n", i+1, \
			PhaseMatrixCache[i+1][PMC_mu], PhaseMatrixCache[i+1][PMC_P11], PhaseMatrixCache[i+1][PMC_CDF], PhaseMatrixCache[i+1][PMC_mu_q]);
		return false;
	}
	return true;
}

void HGQPhaseFunction::calculateCacheLine(const int& i, double& P11, double& P12, double& P22, double& P33, double& P34, double& P44)
{
	double mu=PhaseMatrixCache[i][0];
	calculatePhaseMatrixElements(mu, P11, P12, P22, P33, P34, P44);
}

bool HGQPhaseFunction::createPhaseMatrixCache()
{
	int i, j;
	double mu, P11, P12, P22, P33, P34, P44, *ptr;
	double P11_sum=0.0, N, alpha, F;
	// setup phase matrix cache ######################################################################################
	if(!PhaseMatrixCache.allocate(phase_matrix_cache_resolution+1, PMC_dimension) \
		|| !PMC_index.allocate(phase_matrix_cache_resolution+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating phase matrix cache with resolution %d\n", phase_matrix_cache_resolution);
		return false;
	}
	// initialize phase matrix cache
	PhaseMatrixCache.setToZero();

	// calculate all entries
	for(i=0;i<=phase_matrix_cache_resolution;i++)
	{
		ptr=PhaseMatrixCache[i];
		
		*ptr++=HenyeyGreensteinPhaseFunction::sampleMu(double(i)/double(phase_matrix_cache_resolution));

		calculateCacheLine(i, P11, P12, P22, P33, P34, P44);
		
		*ptr++=P11;
		*ptr++=P12;
		*ptr++=P22;
		*ptr++=P33;
		*ptr++=P34;
		*ptr++=P44;
		*ptr++=P12/P11;
		*ptr++=P22/P11;
		*ptr++=P33/P11;
		*ptr++=P34/P11;
		*ptr++=P44/P11;
	}

	for(i=0;i<phase_matrix_cache_resolution;i++)
		PhaseMatrixCache[i][PMC_delta_mu]=PhaseMatrixCache[i+1][PMC_mu]-PhaseMatrixCache[i][PMC_mu];

	// create inverse CDF ###############################
	//       1. normalize PDF and CDF
	PhaseMatrixCache[0][PMC_CDF]=0.0;
	for(i=1;i<=phase_matrix_cache_resolution;i++)
	{
		P11_sum+=0.5*(PhaseMatrixCache[i][PMC_P11]+PhaseMatrixCache[i-1][PMC_P11])*PhaseMatrixCache[i-1][PMC_delta_mu];
		PhaseMatrixCache[i][PMC_CDF]=P11_sum;
	}
	N=0.5*P11_sum;
	/*if(N!=1.0)
		fprintf(stderr, "HGQ P11 norm = %le\n", P11_sum);*/
	// printf("N=%lf\n", P11_sum);
	for(i=0;i<=phase_matrix_cache_resolution;i++)
	{
		PhaseMatrixCache[i][PMC_P11]/=N;
		PhaseMatrixCache[i][PMC_P12]/=N;
		PhaseMatrixCache[i][PMC_P22]/=N;
		PhaseMatrixCache[i][PMC_P33]/=N;
		PhaseMatrixCache[i][PMC_P34]/=N;
		PhaseMatrixCache[i][PMC_P44]/=N;
		PhaseMatrixCache[i][PMC_CDF]/=N;
	}

	// create quadrature of the inverse CDF
	N=double(phase_matrix_cache_resolution);

	// interpolate rest
	j=0;
	PhaseMatrixCache[0][PMC_mu_q]=-1.0;
	PMC_index[0]=0;
	for(i=1;i<phase_matrix_cache_resolution;i++)
	{
		// i loop over mu_qs
		alpha=(double(i))/N;
		F=2.0*alpha;
		for(;j<=phase_matrix_cache_resolution;j++)
			if(PhaseMatrixCache[j][PMC_CDF]>=F)
				break;
		if(j>phase_matrix_cache_resolution)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error calculating quadrature at alpha=%lf of %s Mie phase function!\n", alpha, name);
			return false;
		}
		j--;

		if(!interpolateMuFromAlpha(&mu, j, alpha))
		{
			writeCache();
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		PhaseMatrixCache[i][PMC_mu_q]=mu;
		PhaseMatrixCache[i][PMC_hits]=PMC_index[i]=j;
	}
	PhaseMatrixCache[phase_matrix_cache_resolution][PMC_mu_q]=1.0;
	PMC_index[phase_matrix_cache_resolution]=phase_matrix_cache_resolution;
	writeCache();
	return true;
}

double HGQPhaseFunction::evaluate(const double& mu)
{
	if(enable_phase_matrix_cache && enable_interpolate_phase_matrix)
	{
		// first step: determine index
		// by evaluating the cdf at mu and divide by the norm 2
		double *ptr, muA,muB, P11A,P11B;
		int i=calculateCacheIndex(mu);
		// A
		ptr=PhaseMatrixCache[i];
		muA=*ptr++;
		P11A=*ptr;
		// B
		ptr=PhaseMatrixCache[i+1];
		muB=*ptr++;
		P11B=*ptr;
		// interpolate
		return P11A+((mu-muA)*(P11B-P11A))/(muB-muA);
	}
	return evaluateExactly(mu);
}

double HGQPhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
	double mu;
	if(enable_phase_matrix_cache)
	{
		double F=2.0*alpha;
		int i=PMC_index[floor(alpha*double(phase_matrix_cache_resolution))];
		// fine adjust intervall
		if(F>PhaseMatrixCache[i+1][PMC_CDF])
			while(F>PhaseMatrixCache[i+1][PMC_CDF])
				i++;
		else if(F<PhaseMatrixCache[i][PMC_CDF])
			while(F<PhaseMatrixCache[i][PMC_CDF])
				i--;
		// interpolate in interval
		if(!interpolateMuFromAlpha(&mu, i, alpha))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return 1.0;
		}
		if(enable_interpolate_phase_matrix)
			*pISWeight=1.0;
		else
			*pISWeight=evaluateExactly(mu)/PhaseMatrixCache[i][PMC_P11];
	}
	else
	{
		mu=HenyeyGreensteinPhaseFunction::sampleMu(alpha, 0);
		*pISWeight=evaluateExactly(mu)/HenyeyGreensteinPhaseFunction::evaluate(mu);
	}
	return mu;
}

class StokesMatrix& HGQPhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	double P11, P12, P22, P33, P34, P44;

	if(enable_phase_matrix_cache && enable_interpolate_phase_matrix)
	{
		// first step: determine index
		// by evaluating the cdf at mu and divide by the norm 2
		double *ptrA,*ptrB, muA,muB, pA,pB;
		int i=calculateCacheIndex(mu);
		// A
		ptrA=PhaseMatrixCache[i];
		muA=*ptrA++;
		// B
		ptrB=PhaseMatrixCache[i+1];
		muB=*ptrB++;
		
		// calculate pA and pB
		pB=(mu-muA)/(muB-muA);
		pA=1.0-pB;

		// interpolate
		P11 = pA * (*ptrA++) + pB * (*ptrB++);
		P12 = pA * (*ptrA++) + pB * (*ptrB++);
		P22 = pA * (*ptrA++) + pB * (*ptrB++);
		P33 = pA * (*ptrA++) + pB * (*ptrB++);
		P34 = pA * (*ptrA++) + pB * (*ptrB++);
		P44 = pA * (*ptrA++) + pB * (*ptrB++);
	}
	else
		calculatePhaseMatrixElements(mu, P11, P12, P22, P33, P34, P44);
		
	PM[0][0]=P11;	PM[0][1]=P12;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=P12;	PM[1][1]=P22;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=P33;	PM[2][3]=P34;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=-P34;	PM[3][3]=P44;

	return PM;
}

class StokesMatrix& HGQPhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	double P11, P12, P22, P33, P34, P44;
	if(enable_phase_matrix_cache && enable_interpolate_phase_matrix)
	{
		// first step: determine index
		// by evaluating the cdf at mu and divide by the norm 2
		double *ptrA,*ptrB, muA,muB, pA,pB;
		int i=calculateCacheIndex(mu);
		// A
		ptrA=PhaseMatrixCache[i];
		muA=*ptrA;
		ptrA+=5;
		// B
		ptrB=PhaseMatrixCache[i+1];
		muB=*ptrB;
		ptrB+=5;
		
		// calculate pA and pB
		pB=(mu-muA)/(muB-muA);
		pA=1.0-pB;

		// interpolate
		P12 = pA * (*ptrA++) + pB * (*ptrB++);
		P22 = pA * (*ptrA++) + pB * (*ptrB++);
		P33 = pA * (*ptrA++) + pB * (*ptrB++);
		P34 = pA * (*ptrA++) + pB * (*ptrB++);
		P44 = pA * (*ptrA++) + pB * (*ptrB++);
	}
	else
	{
		calculatePhaseMatrixElements(mu, P11, P12, P22, P33, P34, P44);
		P12/=P11;
		P22/=P11;
		P33/=P11;
		P34/=P11;
		P44/=P11;
	}

	PM[0][0]=1.0;	PM[0][1]=P12;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=P12;	PM[1][1]=P22;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=P33;	PM[2][3]=P34;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=-P34;	PM[3][3]=P44;
	
	return PM;
}



