/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/HGQTablePhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HGQTablePhaseFunction.h"
#include "../error.h"

// #define LPF_LINEARISED_PHASE_FUNCTION			1

HGQTablePhaseFunction::HGQTablePhaseFunction()
{
	PhaseFunctionType=HGQTable;
	strcpy(name, "HGQTable");
	enable_phase_matrix_cache=true;
	enable_interpolate_phase_matrix=true;
}

bool HGQTablePhaseFunction::loadData(FILE *fptr)
{
	if(fread(&sigma_extinction, sizeof(double), 1, fptr)!=1 || \
		fread(&single_scattering_albedo, sizeof(double), 1, fptr)!=1 || \
		fread(&asymmetry_parameter, sizeof(double), 1, fptr)!=1 || \
		fread(&no_mus, sizeof(int), 1, fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file format!\n");
		return false;
	}
	if(!HGQData.allocate(no_mus+1, 6))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	HGQPhaseFunction::enableCache(true, no_mus);
	if(fread(HGQData.getCoefficientPtr(), sizeof(double)*6, no_mus+1, fptr) != (unsigned int)(no_mus+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file format!\n");
		return false;
	}
	return true;
}

bool HGQTablePhaseFunction::setWavelength(const double& lambda)
{
	if(wavelength==lambda)
		return true;
	// set asymmetry parameter
	if(!HenyeyGreensteinPhaseFunction::defineParameter(0, asymmetry_parameter))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(enable_phase_matrix_cache && !createPhaseMatrixCache())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	wavelength=lambda;
	return true;
}

bool HGQTablePhaseFunction::readHGQTableFileHeader(FILE *fptr, int *pN)
{
	char dcs_string[3];
	if(fread(dcs_string, sizeof(char), 3, fptr)!=3 || \
		dcs_string[0]!='d' || dcs_string[1]!='c' || dcs_string[2]!='s' ||
		fread(pN, sizeof(int), 1, fptr)!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file format!\n");
		return false;
	}
	return true;
}

bool HGQTablePhaseFunction::loadData(char *file_name)
{
	FILE *fptr=fopen(file_name, "rb");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open file \"%s\"!\n", file_name);
		return false;
	}
	int no_clusters;
	if(!readHGQTableFileHeader(fptr, &no_clusters) || no_clusters!=1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file format!\n");
		return false;
	}
	if(!loadData(fptr))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	fclose(fptr);
	return true;
}

void HGQTablePhaseFunction::calculateCacheLine(const int& i, double& P11, double& P12, double& P22, double& P33, double& P34, double& P44)
{
	double *ptr=HGQData[i];
	P11=*ptr++;
	P12=*ptr++;
	P22=*ptr++;
	P33=*ptr++;
	P34=*ptr++;
	P44=*ptr++;
}




