/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/HenyeyGreensteinPhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "HenyeyGreensteinPhaseFunction.h"
#include <math.h>

#include "../Math/prand.h"
#include "../error.h"

HenyeyGreensteinPhaseFunction::HenyeyGreensteinPhaseFunction()
{
	PhaseFunctionType=HenyeyGreenstein;
	h=0.5;
	p=0.5;
	g=0.5;
	gg=0.25;
	ggg=0.125;
	gggg=0.0625;
	one_plus_gg=1.25;
	one_minus_gg=0.75;
	two_g=1.0;
	strcpy(name, "HenyeyGreenstein");
}

double HenyeyGreensteinPhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
	if(pISWeight)
		*pISWeight=1.0;
	/*
		alpha = 0 -> backward scattering
		alpha = 1 -> forward scattering
	*/
	if(g)
		return (one_plus_gg-pow(one_minus_gg/(1.0+g*(2.0*alpha-1.0)),2.0))/two_g;
	return -1.0+2.0*alpha;
}

double HenyeyGreensteinPhaseFunction::evaluate(const double& mu)
{
	return one_minus_gg/pow(1.0-(2.0*mu-g)*g, 1.5);
}

double HenyeyGreensteinPhaseFunction::evaluateCDF(const double& mu)
{
	/*
		mu = -1 -> cdf = 0
		mu = 1 -> cdf = 1
	*/
	if(g)
		return ((1.0-g)*((1.0+g)/sqrt(one_plus_gg-two_g*mu)-1.0))/two_g;
	return 0.5 * ( 1.0 + mu ) ;
}

/*void HenyeyGreensteinPhaseFunction::evaluateDerivatives(double mu, double *parameter_derivative)
{
	*parameter_derivative=(mu*(gg+3.0)+g*(gg-5.0))/pow(1.0-g*(2.0*mu-g), 2.5);
}

void HenyeyGreensteinPhaseFunction::evaluateDerivatives(double mu, double *pPhF, double *parameter_derivative)
{
	double denominator_part=1.0-(2.0*mu-g)*g;
	*pPhF=one_minus_gg/pow(denominator_part, 1.5);
	*parameter_derivative=(mu*(gg+3.0)+g*(gg-5.0))/pow(denominator_part, 2.5);
}*/

bool HenyeyGreensteinPhaseFunction::defineParameter(int index, double value)
{
	switch(index)
	{
		case 0: // g
			g=value;
			gg=g*g;
			ggg=gg*g;
			gggg=gg*gg;
			one_plus_gg=1.0+gg;
			one_minus_gg=1.0-gg;
			two_g=2.0*g;
			break;
		case 1:	// h
			if(value<0.5 || value>1)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"Henyey-Greenstein parameter h(=%lf) is out of range [0.5,1]!\n", value);
				return false;
			}
			h=value;
			break;
		case 2: // p
			if(p>1 || p<0)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"Henyey-Greenstein parameter p(=%lf) is out of range [0,1]!\n", value);
				return false;
			}
			p=value;
			break;
		default:
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"Henyey-Greenstein phase function model has no parameter with index %d!\n", index+1);
			return false;
			
	}
	return true;
}


class StokesMatrix& HenyeyGreensteinPhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
/*
	after:

		C. J. Braak, J. F. de Haan, C. V. M. van der Mee, J. W. Hovenier and L. D. Travis:

		Parameterized Scattering Matrices for Nonspherical Particles in Planetary Atmospheres
*/

	double P11=one_minus_gg/pow(1.0-(2.0*mu-g)*g,1.5);
	double P22=P11*(h+(1.0-h)*mu);
	double one_p_mumu=1.0+mu+mu;
	double P33=(P22*2.0*mu)/one_p_mumu;
	double P21=-(P11*p*(1.0-mu*mu))/one_p_mumu;
	
	PM[0][0]=P11;	PM[0][1]=P21;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=P21;	PM[1][1]=P22;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=P33;	PM[2][3]=0.0;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=0.0;	PM[3][3]=P33;

	return PM;
}


class StokesMatrix& HenyeyGreensteinPhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
/*
	after:

		C. J. Braak, J. F. de Haan, C. V. M. van der Mee, J. W. Hovenier and L. D. Travis:

		Parameterized Scattering Matrices for Nonspherical Particles in Planetary Atmospheres
*/

	double P11=one_minus_gg/pow(1.0-(2.0*mu-g)*g,1.5);
	double P22=P11*(h+(1.0-h)*mu);
	double one_p_mumu=1.0+mu+mu;
	double P33=(P22*2.0*mu)/one_p_mumu;
	double P21=-(P11*p*(1.0-mu*mu))/one_p_mumu;
	
	PM[0][0]=1.0;		PM[0][1]=P21/P11;	PM[0][2]=0.0;		PM[0][3]=0.0;
	PM[1][0]=P21/P11;	PM[1][1]=P22/P11;	PM[1][2]=0.0;		PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=P33/P11;	PM[2][3]=0.0;
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=0.0;		PM[3][3]=P33/P11;

	return PM;
}


void HenyeyGreensteinPhaseFunction::evaluateJacobian(double mu)
{
	J[0]=(mu*(gg+3.0)+g*(gg-5.0))/pow(1.0-g*(2.0*mu-g), 2.5);
	J[1]=0.0;
	J[2]=0.0;
}

void HenyeyGreensteinPhaseFunction::evaluateHessian(double mu)
{
	//register double a=mu*mu*(gg+15.0)-4.0*g*mu*(gg-7.0)-2.0*gggg+23.0*gg-5.0;
	register double a=mu*mu*(gg+15.0)+mu*g*(6.0*gg-32.0)-2.0*gggg+23*gg-5.0;
	register double b=pow(1.0-2.0*mu*g+gg, 3.5);
	H[0][0]=a/b;
}





