/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/LegendrePhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "LegendrePhaseFunction.h"
#include "../trigonometry.h"
#include "../Math/prand.h"
#include "../Math/nan.h"
#include "../error.h"

// #define LPF_LINEARISED_PHASE_FUNCTION			1

LegendrePhaseFunction::LegendrePhaseFunction()
{
	PhaseFunctionType=LegendreExpansion;
	strcpy(name, "LegendreExpansion");
}

bool LegendrePhaseFunction::setWavelength(const double& lambda)
{
	// set asymmetry parameter
	double asymmetry_parameter=Coefficient[1][0]/3.0;
	if(!HenyeyGreensteinPhaseFunction::defineParameter(0, asymmetry_parameter))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(enable_phase_matrix_cache && !createPhaseMatrixCache())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	wavelength=lambda;
	return true;
}

bool LegendrePhaseFunction::loadData(char *file_name)
{
	if(!load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(Coefficient.getN()!=4 && Coefficient.getN()!=6)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "Legendre phase function coefficient matrix must have 4 or 5 columns!\n");
		return false;
	}
	return true;
}


/*#define MAX_ITERATIONS		10

bool LegendrePhaseFunction::setResolution(int resolution)
{
	if(lpf_data)
		delete [] lpf_data;
#ifdef LPF_LINEARISED_PHASE_FUNCTION		
	lpf_data=new double [3*resolution];
	mu_s=lpf_data;
	F_mu=lpf_data+resolution;
	P_mu=lpf_data+2*resolution;
#else
	lpf_data=new double [resolution];
	mu_s=lpf_data;
#endif
	if(!lpf_data)
		return false;
	n=resolution;
	if(!normalize())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"phase function normalisation failed!\n");
		return false;
	}
	F0=LegendrePolynomial::primitive(-1.0);
	double p,mu_n=-1.0,mu_np1,eps=2.0/double(200*(resolution<10000?10000:resolution)),P;
	int i,iteration_counter;
	for(i=0;i<resolution;i++)
	{
		p=double(2*i+1)/double(resolution);
		// find mu which belongs to this point with the Newton method
		iteration_counter=0;
		do
		{
			P=primitive(mu_n)-p;
			mu_np1=mu_n-P/evaluate(mu_n);
			mu_n=mu_np1;
			//printf("%lg %lg, ", mu_n, P);
			iteration_counter++;
			if(iteration_counter>MAX_ITERATIONS)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "Newton method iteration counter (%d) exceeded!\n",MAX_ITERATIONS);
				fprintf(stderr, "decrease phase function resolution to avoid this!\n");
				fprintf(stderr, "phase function:\n");
				for(i=0;i<resolution;i++)
				{
					mu_n=-1.0+double(2*i)/double(resolution);
					p=evaluate(mu_n);
					fprintf(stderr,"%lg\t%lg\n",mu_n,p);
				}
				return false;
			}
		}while(fabs(P)>eps);
		//fputc('\n',stdout);
		mu_s[i]=mu_n;
#ifdef LPF_LINEARISED_PHASE_FUNCTION
		F_mu[i]=primitive(mu_n);
		P_mu[i]=evaluate(mu_n);
#endif
	}
	return true;
}
*/

/*
double LegendrePhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
	
	double mu;
	int index=int(alpha*double(n))%n;
#ifdef LPF_LINEARISED_PHASE_FUNCTION
	mu=correctMu(mu_s[index]+(2.0*alpha-F_mu[index])/P_mu[index])
#else
	mu=mu_s[index];
#endif
	if(ISNAN(mu))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "sampled bad mu, returning 0!\n");
		return 0.0;
	}
	if(pISWeight)
		*pISWeight=1.0;
	return mu;
}*/

void LegendrePhaseFunction::calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P22, double& P33, double& P34, double& P44)
{
	LegendrePolynomial::calculateBasisFunctions(mu);
	register int D_LP=LP.getN(), D_C=Coefficient.getN();
	register double *LP_i=LP[0]+Basis;
	register double *c_i=Coefficient[0];
	register double P11_result=0.0;
	register double P12_result=0.0;
	register double P33_result=0.0;
	register double P34_result=0.0;
	if(Coefficient.getN()==4)
	{
		for(register int i=0;i<no_coefficients;i++)
		{
			P11_result += c_i[0]**LP_i;
			P12_result += c_i[1]**LP_i;
			P33_result += c_i[2]**LP_i;
			P34_result += c_i[3]**LP_i;
			LP_i += D_LP;
			c_i += D_C;
		}
		P11=P11_result;
		P12=P12_result;
		P22=P11_result;
		P33=P33_result;
		P34=P34_result;
		P44=P33_result;
	}
	else
	{
		register double P22_result=0.0;
		register double P44_result=0.0;
		for(register int i=0;i<no_coefficients;i++)
		{
			P11_result += c_i[0]**LP_i;
			P12_result += c_i[1]**LP_i;
			P22_result += c_i[2]**LP_i;
			P33_result += c_i[3]**LP_i;
			P34_result += c_i[4]**LP_i;
			P44_result += c_i[5]**LP_i;
			LP_i += D_LP;
			c_i += D_C;
		}
		P11 = P11_result;
		P12 = P12_result;
		P22 = P22_result;
		P33 = P33_result;
		P34 = P34_result;
		P44 = P44_result;
	}
}

void LegendrePhaseFunction::evaluateJacobian(double mu)
{
	/*double *parameter_derivative=J.getCoefficientPtr();
	LegendrePolynomial::calculateBasisFunctions(mu);
	int M=Coefficient.getM();
	for(int j=1;j<no_coefficients;j++)
		*parameter_derivative ++ = *bf ++ / N;

	double f=evaluate(mu), N=*coefficient, *bf=basis_function;
	* parameter_derivative ++ = ( N * *bf ++ -f) / ( N * N );
	for(int j=1;j<no_coefficients;j++)
		*parameter_derivative ++ = *bf ++ / N;
	*/

	LegendrePolynomial::calculateBasisFunctions(mu);

	register double *parameter_derivative=J.getCoefficientPtr();
	register double *bf=LP[0]+Basis;

	int M=Coefficient.getM(), N=LP.getN();
	register int i;
	for(i=0;i<M;i++)
	{
		*parameter_derivative ++ = *bf;
		bf += N;
	}
}

void LegendrePhaseFunction::evaluateHessian(double mu){;}

/*void LegendrePhaseFunction::evaluateDerivatives(double mu, double *parameter_derivative)
{
	double f=evaluate(mu),N=*coefficient,*bf=basis_function;
	*parameter_derivative++=(N**bf++-f)/(N*N);
	for(int j=1;j<no_coefficients;j++)
		*parameter_derivative++=*bf++/N;
}*/

bool LegendrePhaseFunction::defineParameter(int index, double value)
{
	int M,m, N,n;
	M=Coefficient.getM();
	N=Coefficient.getN();
	if(index>=M*N || index<0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"Legendre phase function coefficient index (=%d) out of range [1,%d]!\n", index+1, N*M);
		return false;
	}
	m=index/N;
	n=index%N;
	Coefficient[m][n]=value;
	return true;
}





