/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/MiePhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "MiePhaseFunction.h"
#include "../error.h"
#include <float.h>

#include "../Math/TVector.h"
#include "../BasicIO.h"

MiePhaseFunction::MiePhaseFunction()
{
	wavelength=0.0;
	bin_radius=0;
	NumberInBin=0;
	DSDModel=MieModule::NLogNormal;
}

void MiePhaseFunction::setModePtr(struct MieModule::Mode_s *mode, int no_modes)
{
	this->mode=mode;
	this->no_modes=no_modes;	
}

void MiePhaseFunction::calculatePhaseMatrixElements(const double& mu, double& P11, double& P12, double& P22, double& P33, double& P34, double& P44)
{
	int i, j;
	MM.calculatePiAndTau(mu, pi.getCoefficientPtr(), tau.getCoefficientPtr(), no_Mie_coefficients);
	for(i=0;i<=no_Mie_coefficients;i++)
	{
		double pi_i=pi[i];
		double tau_i=tau[i];
		gamma_p[i]=pi_i+tau_i;
		gamma_m[i]=pi_i-tau_i;
	}
	complex<double> SpSp=0.0, SpSm=0.0, SmSp=0.0, SmSm=0.0;
	for(i=0;i<no_Mie_coefficients;i++)
	{
		double gpi=gamma_p[i+1];
		double gmi=gamma_m[i+1];
		complex<double> *S_pp_i=S_pp[i];
		complex<double> *S_pm_i=S_pm[i];
		complex<double> *S_mp_i=S_mp[i];
		complex<double> *S_mm_i=S_mm[i];
		for(j=0;j<no_Mie_coefficients;j++)
		{
			double gpj=gamma_p[j+1];
			double gmj=gamma_m[j+1];
			SpSp += S_pp_i[j] * gpi * gpj;
			SpSm += S_pm_i[j] * gpi * gmj;
			SmSp += S_mp_i[j] * gmi * gpj;
			SmSm += S_mm_i[j] * gmi * gmj;
		}
	}
	P11 = ( complex<double>(0.25, 0.0) * (SpSp+SmSm)).real();
	P12 = (-complex<double>(0.25, 0.0) * (SpSm+SmSp)).real();
	P22 = P11;
	P33 = ( complex<double>(0.25, 0.0) * (SpSp-SmSm)).real();
	P34 = ( complex<double>(0.0, 0.25) * (SpSm-SmSp)).real();
	P44 = P33;
}

double MiePhaseFunction::evaluateExactly(const double& mu)
{
	int i, j;
	MM.calculatePiAndTau(mu, pi.getCoefficientPtr(), tau.getCoefficientPtr(), no_Mie_coefficients);
	for(i=0;i<=no_Mie_coefficients;i++)
	{
		double pi_i=pi[i];
		double tau_i=tau[i];
		gamma_p[i]=pi_i+tau_i;
		gamma_m[i]=pi_i-tau_i;
	}
	complex<double> SpSp=0.0, SmSm=0.0;
	for(i=0;i<no_Mie_coefficients;i++)
	{
		double gpi=gamma_p[i+1];
		double gmi=gamma_m[i+1];
		complex<double> *S_pp_i=S_pp[i];
		complex<double> *S_mm_i=S_mm[i];
		for(j=0;j<no_Mie_coefficients;j++)
		{
			double gpj=gamma_p[j+1];
			double gmj=gamma_m[j+1];
			SpSp += S_pp_i[j] * gpi * gpj;
			SmSm += S_mm_i[j] * gmi * gmj;
		}
	}
	return ( complex<double>(0.25, 0.0) * (SpSp+SmSm)).real();
}

bool MiePhaseFunction::initializeComplexRefractiveIndex(class Matrix& RIMatrix, int cluster_index)
{
	int i, n=RIMatrix.getM();
	if(!ComplexRefractiveIndex.allocate(n))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating complex refractive index function\n");
		return false;
	}
	double *wl=ComplexRefractiveIndex.getXPtr(), *column;
	complex< double > *ri=ComplexRefractiveIndex.getYPtr();
	int cr=1+2*cluster_index, ci=cr+1;
	for(i=0;i<n;i++)
	{
		column=RIMatrix.getRowPtr(i);
		*wl++=*column;
		*ri++=complex<double>(column[cr], column[ci]);
	}
	return true;
}

bool MiePhaseFunction::setWavelength(const double& lambda)
{
	// printf("calculating %s for %.2lfnm, ", name, 1e9*lambda);
	if(wavelength==lambda)	return true;

	// allocate number of required a's and b's
	int i, j, k, l, n;
	double radius, r_max=DBL_MIN, log_mu, sigma, log_sigma, *SS;
	double y, area, m_re, m_im, N, dr, wave_number;

	// result variables
	double extinction_coefficient_sum=0.0, extinction_coefficient;
	double scattering_coefficient_sum=0.0, scattering_coefficient, epsilon_s;
	double g_sum=0.0, g;

	// get refractive index
	refractive_index=ComplexRefractiveIndex.evaluate(lambda);
	m_re=refractive_index.real();
	m_im=refractive_index.imag();
	
	// printf("m_re=%lf, m_im=%lf, ", m_re, m_im);
	
	// size spectra integration:
	// create list of radii, weights and norms
	class Matrix SizeSpectrum;
	int no_radii=0, running_index;
	switch(DSDModel)
	{
		default:
		case MieModule::NLogNormal:
			for(i=0;i<no_modes;i++)
			{
				if(mode[i].sigma!=0.0)
					no_radii+=1+quadrature_resolution;
				else
					no_radii++;
			}
			break;
		case MieModule::detailed:
			no_radii=no_bins;
			break;
	}
	/*
		columns:

		0:	r, radius
		1:	N(r)
		2:	sigma_e(r)
		3:	sigma_s(r)
		4:	g(r)

	*/
	if(!SizeSpectrum.allocate(no_radii, 5))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	// 1.: determine all radii and calculate n(r)
	running_index=0;	
	switch(DSDModel)
	{
		default:
		case MieModule::NLogNormal:
			{
				class Matrix Y;
				if(!Y.allocate(1+quadrature_resolution, 3))
				{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
				// Y[i][0]=y(i-0.5)
				// Y[i][1]=y(i)
				// Y[i][2]=y(i+0.5)
				for(i=0;i<=quadrature_resolution;i++)
					for(k=0;k<3;k++)
						Y[i][k]=-MIE_MODULE_SIGMA_RANGE+(MIE_MODULE_SIGMA_RANGE*(double(2*i+k-1)))/double(quadrature_resolution);
				// process all modes ---------------------------
				for(i=0;i<no_modes;i++)
				{
					sigma=mode[i].sigma;
					N=1e9*mode[i].N;
					if(sigma==0.0)
					{
						SS=SizeSpectrum[running_index++];
						SS[0]=radius=mode[i].radius;
						SS[1]=N;
						if(radius>r_max)
							r_max=radius;
					}
					else
					{
						log_sigma=log(sigma);
						log_mu=log(mode[i].radius);
						for(j=0;j<=quadrature_resolution;j++)
						{
							SS=SizeSpectrum[running_index++];
							y=Y[j][1];
							SS[0]=radius=exp(log_mu+y*log_sigma);
							dr=exp(log_mu+Y[j][2]*log_sigma)-exp(log_mu+Y[j][0]*log_sigma);
							SS[1]=(dr*N)/(sqrt(2.0*M_PI)*log_sigma*radius)*exp(-0.5*y*y);
							if(radius>r_max)
								r_max=radius;
						}
					}
				}
			}
			break;

		case MieModule::detailed:
			for(i=0;i<no_bins;i++)
			{
				SS=SizeSpectrum[running_index++];
				SS[0]=radius=bin_radius[i];
				SS[1]=1e9*NumberInBin[i];
				if(radius>r_max)
					r_max=radius;
			}
	}
	wave_number=(2.0*M_PI)/lambda;
	no_Mie_coefficients=MM.calculateNOCoefficients(wave_number*r_max);

	// allocate Mie coefficients for S_pp, S_pm, S_mp and S_mm
	if(!S_pp.allocate(no_Mie_coefficients, no_Mie_coefficients) || \
		!S_pm.allocate(no_Mie_coefficients, no_Mie_coefficients) || \
		!S_mp.allocate(no_Mie_coefficients, no_Mie_coefficients) || \
		!S_mm.allocate(no_Mie_coefficients, no_Mie_coefficients))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	S_pp.setToZero();
	S_pm.setToZero();
	S_mp.setToZero();
	S_mm.setToZero();

	// allocate alpha_p, and alpha_m
	double f, *cptr;
	complex<double> *a, *b;
	TVector< complex<double> > alpha_p, alpha_m;
	if(!alpha_p.allocate(no_Mie_coefficients) || \
		!alpha_m.allocate(no_Mie_coefficients) || \
		!gamma_p.allocate(no_Mie_coefficients+1) || \
		!gamma_m.allocate(no_Mie_coefficients+1) || \
		!TwoNp1oNNp1.allocate(no_Mie_coefficients+1) || \
		!pi.allocate(no_Mie_coefficients+1) || \
		!tau.allocate(no_Mie_coefficients+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	MM.calculateTwoNp1oNNp1(TwoNp1oNNp1.getCoefficientPtr(), no_Mie_coefficients+1);
	// printf("start....");
	// fflush(stdout);
	// do actual n(r) integration
	for(i=0;i<no_radii;i++)
	{
		SS=SizeSpectrum[i];

		radius=SS[0];

		if(!MM.setParameters(lambda, m_re, -m_im, radius))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
			
		}
		n=MM.getNOCoefficients();
		MM.calculateMieCoefficients();

		// obtain optical properties
		area=M_PI*radius*radius;

		SS[2] = extinction_coefficient = area * MM.getExtinctionEfficiency();
		SS[3] = scattering_coefficient = area * MM.getScatteringEfficiency();
		SS[4] = g = MM.getAsymmetryParameter();
		
		// add optical properties
		N=SS[1];

		extinction_coefficient_sum += N * extinction_coefficient;
		scattering_coefficient_sum += epsilon_s = N * scattering_coefficient;
		g_sum += epsilon_s * g;
		
		// calculate alpha_p and alpha_m
		a=MM.getAPtr();
		b=MM.getBPtr();
		cptr=TwoNp1oNNp1.getCoefficientPtr()+1;
		for(k=0;k<n;k++)
		{
			f=*cptr++;
			alpha_p[k]=f*(*a+*b);
			alpha_m[k]=f*(*a-*b);
			a++;
			b++;
		}

		// calculate S_pp, S_pm, S_mp and S_mm components
		for(k=0;k<n;k++)
		{
			complex<double> *S_pp_k=S_pp[k];
			complex<double> *S_pm_k=S_pm[k];
			complex<double> *S_mp_k=S_mp[k];
			complex<double> *S_mm_k=S_mm[k];
			complex<double> N_conj_alpha_pk = N * conj ( alpha_p[k] );
			complex<double> N_conj_alpha_mk = N * conj ( alpha_m[k] );
			for(l=0;l<n;l++)
			{
				complex<double> alpha_pl = alpha_p[l];
				complex<double> alpha_ml = alpha_m[l];
				S_pp_k[l] += N_conj_alpha_pk * alpha_pl;
				S_pm_k[l] += N_conj_alpha_pk * alpha_ml;
				S_mp_k[l] += N_conj_alpha_mk * alpha_pl;
				S_mm_k[l] += N_conj_alpha_mk * alpha_ml;
			}
		}
	}
	// save SizeSpectrum matrix
	char path[BasicIO_string_size], file_name[BasicIO_string_size];
	formatOutputDirectory(path);
	sprintf(file_name, "%s%s_OpticalProperties.txt", path, name);
	SizeSpectrum.save(file_name);

	// finish bulk optical property calculation
	this->extinction_coefficient = extinction_coefficient_sum;
	this->ssa = scattering_coefficient_sum / extinction_coefficient_sum;
	g = g_sum / scattering_coefficient_sum;
	if(!HenyeyGreensteinPhaseFunction::defineParameter(0, g))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// phase matrix 
	f = ( 4.0 * M_PI ) / ( scattering_coefficient_sum * wave_number * wave_number );
	for(k=0;k<no_Mie_coefficients;k++)
	{
		complex<double> *S_pp_k=S_pp[k];
		complex<double> *S_pm_k=S_pm[k];
		complex<double> *S_mp_k=S_mp[k];
		complex<double> *S_mm_k=S_mm[k];
		for(l=0;l<no_Mie_coefficients;l++)
		{
			S_pp_k[l] *= f;
			S_pm_k[l] *= f;
			S_mp_k[l] *= f;
			S_mm_k[l] *= f;
		}
	}
	// printf("stop!\n");
	
	wavelength=lambda;
	// fprintf(phf_fptr,"# %le\t%le\t%le\t%le\n", this->extinction_coefficient, this->ssa, g, MM.getExtinctionEfficiency());
	if(enable_phase_matrix_cache && !createPhaseMatrixCache())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	// printf("done\n");
	return true;
}






