/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/MolecularPhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "MolecularPhaseFunction.h"
#include "../trigonometry.h"

double MolecularPhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
	double z1=(N*(3.0-6.0*alpha))/(2.0*A);
	double z2=(A+C)/A;
	double a=sqrt(z1*z1+z2*z2*z2);
	if(pISWeight)
		*pISWeight=1.0;
	return correctMu(pow(a-z1,1.0/3.0)-pow(a+z1,1.0/3.0));
}

double MolecularPhaseFunction::evaluate(const double& mu)
{
	return (A*(1.0+mu*mu)+C)/N;
}

class StokesMatrix& MolecularPhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
/*
	taken from:
		Landgraf, Hasekamp, van Deelen, Aben: Rotational Raman scattering of polarized light in the Earth atmosphere:
			a vector radiative transfer model using radiative transfer pertubation theory approach

		Elsevier JQSRT 87, 399-433, (2004)
*/
	double h=A*(1.0+mu*mu);
	double P12=(A*(mu*mu-1.0))/N;
	
	PM[0][0]=(h+C)/N;	PM[0][1]=P12;		PM[0][2]=0.0;			PM[0][3]=0.0;
	PM[1][0]=P12;		PM[1][1]=h/N;		PM[1][2]=0.0;			PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=(2.0*A*mu)/N;		PM[2][3]=0.0;
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=0.0;			PM[3][3]=(B*mu)/N;

	return PM;
}

class StokesMatrix& MolecularPhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	double A1pmumu=A*(1+mu*mu);
	double A1pmumupC=A1pmumu+C;

	double P12=A*(mu*mu-1)/A1pmumupC;
	double P22=A1pmumu/A1pmumupC;
	double P33=(2.0*A*mu)/A1pmumupC;
	double P44=(B*mu)/A1pmumupC;

	PM[0][0]=1.0;	PM[0][1]=P12;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=P12;	PM[1][1]=P22;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=P33;	PM[2][3]=0.0;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=0.0;	PM[3][3]=P44;

	return PM;
}





