/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/PhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "RayleighPhaseFunction.h"
#include "PhaseFunction.h"
#include "../Math/prand.h"
#include "../constants.h"
#include "../trigonometry.h"
#include "../error.h"

#include "RotationalRamanPhaseFunction.h"

#include "TablePhaseFunction.h"
#include "HenyeyGreensteinPhaseFunction.h"
#include "MiePhaseFunction.h"
#include "LegendrePhaseFunction.h"
#include "HGQTablePhaseFunction.h"

class Matrix3D PhaseFunction::localCSMatrix=Matrix3D(Matrix3D::UNITY, 0.0);

class PhaseFunction *PhaseFunction::newPhaseFunction(enum PhaseFunctionType_e type)
{
	switch(type)
	{
		case HenyeyGreenstein:		return new HenyeyGreensteinPhaseFunction;
		case Rayleigh: 			return new RayleighPhaseFunction;
		case RotationalRaman: 		return new RotationalRamanPhaseFunction;
		case LegendreExpansion: 	return new LegendrePhaseFunction;
		case HGQTable:			return new HGQTablePhaseFunction;
		case Table: 			return new TablePhaseFunction;
		case Mie:			return new MiePhaseFunction;
		default:;
	}
	return (class PhaseFunction *)0;
}

bool PhaseFunction::extractType(enum PhaseFunctionType_e *pPhaseFunctionType, char *text)
{
	if(!strcmp(text,"HenyeyGreenstein") || !strcmp(text,"HG"))
		*pPhaseFunctionType=HenyeyGreenstein;
	else if(!strcmp(text,"Mie") || !strcmp(text,"M"))
		*pPhaseFunctionType=Mie;
	else if(!strcmp(text,"Rayleigh") || !strcmp(text,"rayleigh"))
		*pPhaseFunctionType=Rayleigh;
	else if(!strcmp(text,"Cabannes"))
		*pPhaseFunctionType=Cabannes;
	else if(!strcmp(text,"RotationalRaman") || !strcmp(text,"raman"))		
		*pPhaseFunctionType=RotationalRaman;
	else if(!strcmp(text,"LegendreExpansion") || !strcmp(text,"LE"))
		*pPhaseFunctionType=LegendreExpansion;
	else if(!strcmp(text,"HGQTable") || !strcmp(text,"HGQ"))
		*pPhaseFunctionType=HGQTable;
	else if(!strcmp(text,"Table") || !strcmp(text,"tab"))
		*pPhaseFunctionType=Table;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"unknown phase function type \"%s\"!\n",text);
		return false;
	}
	return true;
}

const char *PhaseFunction::getTypeString(enum PhaseFunctionType_e PhaseFunctionType)
{
	switch(PhaseFunctionType)
	{	
		case HenyeyGreenstein:		return "HenyeyGreenstein";
		case Rayleigh:			return "Rayleigh";
		case Cabannes:			return "Cabannes";
		case RotationalRaman:		return "RotationalRaman";
		case LegendreExpansion:		return "LegendreExpansion";
		case HGQTable:			return "HGQTable";
		case Table:			return "Table";
		case Mie:			return "Mie";
		default:;
	}
	errorIn(__FILE__, __LINE__, __func__);
	return (const char *)0;
}

void PhaseFunction::printType(enum PhaseFunctionType_e PhaseFunctionType, char *text)
{
	strcpy(text,getTypeString(PhaseFunctionType));
}

void PhaseFunction::writeTypeTemplate(FILE *fptr)
{
	fprintf(fptr,"HG (Henyey-Greenstein), Mie, rayleigh, Cabannes, raman, LE (Legendre polynomial expansion), HGQTable, tab (Table)");
}

void PhaseFunction::scatter(class Vector3D *pDirection, double *pMu, double *pPhi, double *pISWeight, bool resample)
{
	double *px=pDirection->getCoefficientPtr();

	// ################ analyse incoming direction #####################
	double x_in=*px;	px++;
	double y_in=*px;	px++;		
	double z_in=*px;		

	// check relation to z-axis...
	double rq_xy_in=sqrt(x_in*x_in+y_in*y_in);
	double phi_in=atan2(y_in,x_in);
	double theta_in=atan2(rq_xy_in,z_in);

	// ################ sample phi and theta ###########################
	if(resample)
	{
		alpha_phi=prand_full();
		alpha_mu=prand_full();
	}
	double phi_sample=alpha_phi*M_2PI;
	double mu_sample=sampleMu(alpha_mu, pISWeight);

	/*if(ISNAN(mu_sample))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "invalid cosine of the scatter angle sampled, alpha=%le!\n", alpha_mu);
	}*/
	
	
	// get sin_theta according to mu
	double sin_theta_sample=sqrt(1.0-mu_sample*mu_sample);

	// ################## construct scattered vector ###################
	double x=sin_theta_sample*cos(phi_sample);
	double y=sin_theta_sample*sin(phi_sample);
	double z=mu_sample;

	// ######### rotate vector back into the incoming direction ########
	// rotate about theta
	double sina=sin(theta_in);
	double cosa=cos(theta_in);
	double rx=x*cosa+z*sina;
	double rz=-x*sina+z*cosa;
	// update x and z
	x=rx;
	z=rz;
	// rotate about phi
	sina=sin(phi_in);
	cosa=cos(phi_in);
	       rx=x*cosa-y*sina;
	double ry=x*sina+y*cosa;

	//  ###################### update direction ########################
	// check this!
	px=pDirection->getCoefficientPtr();
	*px++=rx;
	*px++=ry;
	*px++=rz;
	*pMu=mu_sample;
	*pPhi=phi_sample;

	/*if(!pDirection->isValid())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "invalid direction sampled!\n");
	}*/
}

void PhaseFunction::scatterWithAzimuthHGSunBias(Vector3D *pDirection, double *pMu, double *pPhi, class Vector3D SunDirection, const double& g, double *pweight)
{
	double *px=pDirection->getCoefficientPtr();

	// ################ analyse incoming direction #####################
	double x_in=*px;	px++;
	double y_in=*px;	px++;
	double z_in=*px;

	// check relation to z-axis...
	double rq_xy_in=sqrt(x_in*x_in+y_in*y_in);
	double phi_in=atan2(y_in,x_in);
	double theta_in=atan2(rq_xy_in,z_in);
	
	// rotate SunDirection into incomging axis

	SunDirection*=Matrix3D(Matrix3D::ROTZ, -phi_in);
	SunDirection*=Matrix3D(Matrix3D::ROTY, -theta_in);
	
	double phi_sun=atan2(SunDirection[1], SunDirection[0]);

	// ################ sample phi and theta ###########################

	// sample phi according to 2D HG phase function
	double gg=g*g;
	double phi_HG=2.0*atan((1.0-g)*tan(0.5*M_PI*(1.0-2.0*prand_full()))/(1.0+g));
	double phi_sample=phi_sun+phi_HG;
	double mu_phi=cos(phi_HG);

	double mu_sample=sampleMu(prand_full());
	*pweight=(1.0+g*(g-2.0*mu_phi))/(1.0-gg);

	/*if(ISNAN(mu_sample))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "invalid cosine of the scatter angle sampled, alpha=%le!\n", alpha_mu);
	}*/
	
	
	// get sin_theta according to mu
	double sin_theta_sample=sqrt(1.0-mu_sample*mu_sample);

	// ################## construct scattered vector ###################
	double x=sin_theta_sample*cos(phi_sample);
	double y=sin_theta_sample*sin(phi_sample);
	double z=mu_sample;

	// ######### rotate vector back into the incoming direction ########
	// rotate about theta
	double sina=sin(theta_in);
	double cosa=cos(theta_in);
	double rx=x*cosa+z*sina;
	double rz=-x*sina+z*cosa;
	// update x and z
	x=rx;
	z=rz;
	// rotate about phi
	sina=sin(phi_in);
	cosa=cos(phi_in);
	       rx=x*cosa-y*sina;
	double ry=x*sina+y*cosa;

	//  ###################### update direction ########################
	// check this!
	px=pDirection->getCoefficientPtr();
	*px++=rx;
	*px++=ry;
	*px++=rz;
	*pMu=mu_sample;
	*pPhi=phi_sample;

	/*if(!pDirection->isValid())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "invalid direction sampled!\n");
	}*/

}

class StokesMatrix& PhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	PM[0][0]=1.0;	PM[0][1]=0.0;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=0.0;	PM[1][1]=1.0;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=1.0;	PM[2][3]=0.0;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=0.0;	PM[3][3]=1.0;
	
	return PM;
}

class StokesMatrix& PhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	PM[0][0]=1.0;	PM[0][1]=0.0;	PM[0][2]=0.0;	PM[0][3]=0.0;
	PM[1][0]=0.0;	PM[1][1]=1.0;	PM[1][2]=0.0;	PM[1][3]=0.0;
	PM[2][0]=0.0;	PM[2][1]=0.0;	PM[2][2]=1.0;	PM[2][3]=0.0;
	PM[3][0]=0.0;	PM[3][1]=0.0;	PM[3][2]=0.0;	PM[3][3]=1.0;

	return PM;
}

// #define TRANSFORM_TO_LOCAL_CS

void PhaseFunction::calculateScatterAngles(const class Vector3D& k_in, const class Vector3D& k_out, double& mu, double& sigma1, double& sigma2)
{
/*
	excerpt from fomulas of the spherical trigonometry:

	three points A, B and C on the sphere. angles in the corners are A,B, C
	angles of the opposite sides are a,b and c

	sinus rule:

		sin(a)/sin(A) = sin(b)/sin(B) = sin(c)/sin(C)

	cosinus rules:

		cos(a) = cos(b) cos(c) + sin(b) sin(c) cos(A)
		cos(b) = cos(c) cos(a) + sin(c) sin(a) cos(B)
		cos(c) = cos(a) cos(b) + sin(a) sin(b) cos(C)

*/
#ifdef TRANSFORM_TO_LOCAL_CS
	// develop vectors into local coordinate system
	class Vector3D k1, k2;	// vectors in the local CS
	k1=k_in*localCSMatrix;
	k2=k_out*localCSMatrix;
	
	// apply usual algebra and calculate angles:
	double theta1,phi1, theta2,phi2, r;
	k1.toPolar(&r, &theta1, &phi1);
	k2.toPolar(&r, &theta2, &phi2);
	mu=correctMu(k1*k2);
#else
	double theta1,phi1, theta2,phi2, r;
	k_in.toPolar(&r, &theta1, &phi1);
	k_out.toPolar(&r, &theta2, &phi2);
	mu=correctMu(k_in*k_out);
#endif
	
	// apply spherical trigonometry formulas
	double c1,s1, c2,s2;
	double ct1=cos(theta1), st1=sin(theta1), ct2=cos(theta2), st2=sin(theta2);
	double sdp=sin(phi2-phi1);
	
	/*if(st1==0.0)
	{
		st1=theta1>0.5*M_PI?1.0:-1.0;
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "st1=0\n");
	}
	if(st2==0.0)
	{
		st2=theta2>0.5*M_PI?1.0:-1.0;
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "st2=0\n");
	}*/
	
	// NOTE: I ommit division through sin(theta_s) due 
	// to the later use of atan2
	
	c1=(ct2-ct1*mu)/st1;
	s1=sdp*st2;

	c2=(ct1-ct2*mu)/st2;
	s2=sdp*st1;
	
	// calculate sigmas
	sigma1=atan2(s1, c1);
	sigma2=atan2(s2, c2);
}

void PhaseFunction::fillStokesMatrix(class StokesMatrix& SM, const class Vector3D& k_in, const class Vector3D& k_out, bool weighted)
{
	double mu, sigma1, sigma2;

	calculateScatterAngles(k_in, k_out, mu, sigma1, sigma2);
	
	if(weighted)
		initializeWeightedPhaseMatrix(mu);
	else
		initializePhaseMatrix(mu);
	
	SM=StokesMatrix(M_PI-sigma2)*(PhaseMatrix*StokesMatrix(-sigma1));
	// fputc('.',stdout);
}

void PhaseFunction::fillStokesMatrix(class StokesMatrix& SM, class StokesMatrix& PM, double& mu, double& sigma1, double& sigma2, bool weighted)
{
	if(weighted)
		initializeWeightedPhaseMatrix(PM, mu);
	else
		initializePhaseMatrix(PM, mu);
	
	SM=StokesMatrix(M_PI-sigma2)*(PhaseMatrix*StokesMatrix(-sigma1));
	// fputc('.',stdout);
}

bool PhaseFunction::allocateJacobianAndHessian()
{
	int n=getNOParameters();
	if(!J.allocate(n) || !H.allocate(n, n))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	J.setToZero();
	H.setToZero();
	return true;
}


