/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/PhaseFunction.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _PhaseFunction_h
#define _PhaseFunction_h

#include "../Math/Vector.h"
#include "../Math/Matrix.h"

#include "../Math/Vector3D.h"
#include "../Math/Matrix3D.h"

#include "../Math/StokesVector.h"
#include "../Math/StokesMatrix.h"

#include "../Math/prand.h"
#include "../SimulationObject.h"

#include <string.h>

enum PhaseFunctionType_e
{
	HenyeyGreenstein,	// Henyey-Greenstein model phase function
	Mie,			// Mie phase function model
	Rayleigh,		// Rayleigh scattering phase function 
	Cabannes,		// Cabannes scattering phase function 
	RotationalRaman,	// rotational raman scattering phase function		
	LegendreExpansion,	// Legendre polyonomial expansion
	HGQTable,		// Henyey-Greenstein quadrature table
	Table			// linearely interpolated phase function
};

class PhaseFunction: public SimulationObject
{
protected:

	class Vector J;
	class Matrix H;
	
	static class Matrix3D localCSMatrix;
	
	enum PhaseFunctionType_e PhaseFunctionType;
	double wavelength;
	class StokesMatrix PhaseMatrix;

	// virtual double sampleMu()=0;
	virtual double sampleMu(const double& p, double *pISWeight=0)=0;

	double alpha_mu, alpha_phi;
	
	char name[100];

public:

	virtual ~PhaseFunction(){}
	
	inline void setName(const char *text){strcpy(name, text);}

	virtual bool setWavelength(const double& lambda){wavelength=lambda; return true;}
	inline double getWavelength(){return wavelength;}

	virtual double evaluate(const double& mu)=0;
	virtual inline class StokesMatrix& evaluatePhaseMatrix(const double& mu) {return initializePhaseMatrix(PhaseMatrix, mu);}

	inline class StokesMatrix& getPhaseMatrix(){return PhaseMatrix;}

	void scatter(Vector3D *pDirection, double *pMu, double *pPhi, double *pISWeight=0, bool resample=true);
	void scatterWithAzimuthHGSunBias(Vector3D *pDirection, double *pMu, double *pPhi, class Vector3D SunDirection, const double& g_IS, double *pweight);

	inline void getAlphas(double *alpha_mu_ptr, double *alpha_phi_ptr){*alpha_mu_ptr=alpha_mu;*alpha_phi_ptr=alpha_phi;}
	inline void setAlphas(const double& alpha_mu, const double& alpha_phi){this->alpha_mu=alpha_mu; this->alpha_phi=alpha_phi;}

	static void calculateScatterAngles(const class Vector3D& k_in, const class Vector3D& k_out, double& mu, double& sigma1, double& sigma2);

	inline class StokesMatrix& initializePhaseMatrix(const double& mu){return initializePhaseMatrix(PhaseMatrix, mu);}
	inline class StokesMatrix& initializeWeightedPhaseMatrix(const double& mu){return initializeWeightedPhaseMatrix(PhaseMatrix, mu);}

	virtual class StokesMatrix& initializePhaseMatrix(class StokesMatrix& PM, const double& mu);
	virtual class StokesMatrix& initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu);

	void fillStokesMatrix(class StokesMatrix& SM, const class Vector3D& k_in, const class Vector3D& k_out, bool weighted);
	void fillStokesMatrix(class StokesMatrix& SM, double& mu, double& sigma1, double& sigma2, bool weighted);
	void fillStokesMatrix(class StokesMatrix& SM, class StokesMatrix& PM, double& mu, double& sigma1, double& sigma2, bool weighted);

	// ################# J and H ######################################
	bool allocateJacobianAndHessian();

	inline double& Jacobian(const int& i){return J[i];}
	inline class Vector& Jacobian(){return J;}

	inline double& Hessian(const int& i, const int& j){return H[i][j];}
	inline class Matrix& Hessian(){return H;}

	// virtual parameter interface
	virtual int getNOParameters(){return 0;}
	virtual inline bool defineParameter(int index, double value){return false;}
	virtual double getParameter(int index){return 0.0;}

	// ######## J and H ################## 
	virtual void evaluateJacobian(double mu){}
	virtual void evaluateHessian(double mu){}

	/*virtual void evaluateDerivatives(double mu, double *parameter_derivative){}
	virtual void evaluateDerivatives(double mu, double *pPhF, double *parameter_derivative){}*/

	virtual bool requiresFileInput(){return false;}
	virtual bool loadData(char *filename){return true;}

	// static functions
	static void initializeSphericalLocalCSMatrix(const class Vector3D& p){localCSMatrix.initializeWithSphericalBasisInColumns(p);}
	static void resetLocalCSMatrix(){localCSMatrix=Matrix3D(Matrix3D::UNITY, 0.0);}

	static class PhaseFunction *newPhaseFunction(enum PhaseFunctionType_e type);
	static bool extractType(enum PhaseFunctionType_e *pPhaseFunctionType, char *text);
	static void printType(enum PhaseFunctionType_e PhaseFunctionType, char *text);
	static const char *getTypeString(enum PhaseFunctionType_e PhaseFunctionType);
	static void writeTypeTemplate(FILE *fptr);

};

#endif //_PhaseFunction_h

