/**************************************************************************************************************
*
*       This file,
*
*                            /PhaseFunctions/TablePhaseFunction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "TablePhaseFunction.h"
#include "../error.h"
#include <float.h>

TablePhaseFunction::TablePhaseFunction()
{
	//mu_bin=0;	
}

TablePhaseFunction::~TablePhaseFunction()
{
	/*if(mu_bin)
		delete [] mu_bin;*/
}

bool TablePhaseFunction::loadData(char *file_name)
{
	if(!data.load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load matrix from file \"%s\"!\n", file_name);
		return false;
	}
	if(data.getN()!=5)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "matrix from file \"%s\" has wrong format (%d columns instead of 5)!\n", file_name, data.getN());
		return false;
	}
	return initialize();
}

bool TablePhaseFunction::initialize()
{
	no_angles = data.getM();
	// check for equidistant mu values
	double newdx, dx=data[1][0]-data[0][0];
	int i, n = no_angles;
	for(i=2;i<n;i++)
	{
		newdx=data[i][0]-data[i-1][0];
		if(newdx!=dx)
		{
			has_equidistant_mu_grid=false;
			break;
		}
	}
	if(i==n)
		has_equidistant_mu_grid=true;
	no_intervals = no_angles - 1 ;
	// allocate sampling matrix
	if(!P11BinArea.allocate(no_intervals) || !AccumulatedP11BinArea.allocate(no_intervals) || !NormalizedP11.allocate(n))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating bin area vector!\n");
		return false;
	}
	
	// calculate P11 norm
	double x1,x2, f1,f2, area, total_area=0.0;
	for(i=0;i<no_intervals;i++)
	{
		x1=data[i][0];
		x2=data[i+1][0];
		f1=data[i][1];
		f2=data[i+1][1];

		area=0.5*(f1+f2)*(x2-x1);
		if(area<=0.0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "bin(row) %d of P11 Element has negative area!\n", i+1);
			return false;
		}

		P11BinArea[i]=area;
		total_area+=area;
	}
	/*const double max_relative_difference=0.001;
	double relative_difference=fabs(total_area/2.0-1.0);
	if(relative_difference>max_relative_difference)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "warning: phase function norm (%.4lf) differs %.4lf%% (max: %.4lf%%) from 2!\n", \
			total_area, 100.0*relative_difference, 100.0*max_relative_difference);
	}*/

	// normalize
	P11BinArea/=total_area;
	for(i=0;i<n;i++)
		NormalizedP11[i]=data[i][1]/total_area;
	
	// calculate accumulated P11 bin are
	total_area=0.0;
	for(i=0;i<no_intervals;i++)
	{
		area=P11BinArea[i];
		AccumulatedP11BinArea[i]=total_area;
		total_area+=area;
	}
	
	/*// make quadrature
	if(mu_bin)
		delete [] mu_bin;
	mu_bin=new int[no_angles];
	if(!mu_bin)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating mu_bin\n");
		return false;
	}
	double p;
	for(i=0;i<no_angles;i++)
	{
		p=(double(i)+0.5)/double(no_angles);
		j=0; 
		while(AccumulatedP11BinArea[j]<p)
			j++;	
		mu_bin[i]=j;
	}*/
	return true;
}

double TablePhaseFunction::sampleMu(const double& alpha, double *pISWeight)
{
	if(pISWeight)
		*pISWeight=1.0;

	double CDF;
	// bisection search
	int i, i1 = 0, i2 = no_intervals ;
	while ( i1 + 1 < i2 )
	{
		i = ( i1 + i2 ) / 2 ;
		CDF = AccumulatedP11BinArea[i];
		if ( alpha == CDF )
			return data [i][0];
		else if ( alpha > CDF )
			i1 = i;
		else
			i2 = i;
	}
	// quadratic interpolation
	double dp = alpha;
	dp -= AccumulatedP11BinArea[i1];
	// AccumulatedP11BinArea[mu_index-1];

	// interpolate
	double a = data[i1][0];
	double b = data[i2][0];

	double fa = NormalizedP11[i1];
	double fb = NormalizedP11[i2];

/*
	                 fb *
                            |
                            |
	     fa *           |
	        |           |
	        |           |
	        |           |
	        |           |
	        |           |
	        |           |
	--------+-----------+----
                a           b
*/
	double df = fb - fa;
	double bma = b - a;

	if ( df != 0.0 )
	{
		double m = df / bma;
		double c = fa;
		// f(x) = m * x + c , x in [ 0 , b - a ]

		// int f(x) = 0.5 * m * x^2 + c * x = dp
		
		double p = c / m ;
		double q = 2.0 * dp / m ;

		double radicand = p * p + q ;

		double sqrt_radicand = sqrt ( radicand ) ;	
		double solution = - p + sqrt_radicand ;

		if ( solution < 0.0 || solution > bma )
		{
			solution = - p - sqrt_radicand ;
			if ( solution < 0.0 || solution > bma )
			{
				errorIn ( __FILE__, __LINE__, __func__ ) ;
				fprintf ( stderr, "quadratic pg formula not solvable\n" ) ;
				return 0.5 * ( a + b ) ;
			}
		}
		return a + solution ;
	}
	return a + dp / fa  ;
}

int TablePhaseFunction::findMuIndex(const double& mu)
{
	// determine bin
	if(has_equidistant_mu_grid)
	{
		int index = int ( double ( no_intervals ) * ( 1.0 + mu ) / 2.0 ) ;
		return index % no_intervals;
	}	
	else
	{
		/*
			index=0;
			while(data[index][0]<mu)
				index++;
			--index;
		*/
		double mu_m;
		int l=0, m, h = no_intervals ;
		while ( l + 1 != h )
		{
			m = ( l + h ) / 2 ;
			mu_m = data[m][0];
			if ( mu == mu_m )
				return m;
			else if ( mu > mu_m )
				l = m;
			else
				h = m;
		}
		return l;
	}
}

double TablePhaseFunction::evaluate(const double& mu)
{
	if(mu<=-1.0)
		return data[0][1];
	else if(mu>1.0)
		return data[0][no_intervals];

	int index=findMuIndex(mu);
	
	// got bin, now interpolate
	double a=data[index][0];
	double b=data[index+1][0];
	double fa=data[index][1];
	double fb=data[index+1][1];
	return fa+((fb-fa)*(mu-a))/(b-a);
}

class StokesMatrix& TablePhaseFunction::initializePhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	int i, index=findMuIndex(mu);
	double P[4], a,b,fa, bma, muma;
	a=data[index][0];
	b=data[index+1][0];
	bma=b-a;
	muma=mu-a;
	for(i=0;i<4;i++)
	{
		fa=data[index][1+i];
		P[i]=fa+((data[index+1][1+i]-fa)*muma)/bma;
	}
	PM[0][0]=P[0];		PM[0][1]=P[1];		PM[0][2]=0.0;		PM[0][3]=0.0;
	PM[1][0]=P[1];		PM[1][1]=P[0];		PM[1][2]=0.0;		PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=P[2];		PM[2][3]=-P[3];
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=P[3];		PM[3][3]=P[2];

	return PM;
}

class StokesMatrix& TablePhaseFunction::initializeWeightedPhaseMatrix(class StokesMatrix& PM, const double& mu)
{
	int i, index=findMuIndex(mu);
	double P11, Pij, P[4], a,b, fa,  bma, muma;
	a=data[index][0];
	b=data[index+1][0];
	bma=b-a;
	muma=mu-a;

	fa=data[index][1];
	P11=fa+((data[index+1][1]-fa)*muma)/bma;
	P[0]=1.0;
	for(i=1;i<4;i++)
	{
		fa=data[index][1+i];
		Pij=fa+((data[index+1][1+i]-fa)*muma)/bma;
		P[i]=Pij/P11;
	}
	PM[0][0]=P[0];		PM[0][1]=P[1];		PM[0][2]=0.0;		PM[0][3]=0.0;
	PM[1][0]=P[1];		PM[1][1]=P[0];		PM[1][2]=0.0;		PM[1][3]=0.0;
	PM[2][0]=0.0;		PM[2][1]=0.0;		PM[2][2]=P[2];		PM[2][3]=-P[3];
	PM[3][0]=0.0;		PM[3][1]=0.0;		PM[3][2]=P[3];		PM[3][3]=P[2];

	return PM;
}

// #define TEST_TPF	1
#ifdef TEST_TPF

#include <stdlib.h>

static void print_usage(FILE *fptr)
{
	fprintf(fptr,"test_table_phf phf.txt bins N\n");
}

int main(int argc, char *argv[])
{
	if(argc!=4)
	{
		print_usage(stdout);
		return 0;
	}
	char *filename=argv[1];
	int bins=atoi(argv[2]),N=atoi(argv[3]),i,index;
	double *frequency=new double[bins],mu;
	if(!frequency)
	{
		fprintf(stderr,"%d bins allocation error!\n", bins);
		return -1;
	}
	class TablePhaseFunction tpf;
	if(!tpf.loadData(filename))
	{
		fprintf(stderr,"could not read file \"%s\"!\n", filename);
		delete [] frequency;
		return -1;
	}
	for(i=0;i<bins;i++)
		frequency[i]=0.0;
	for(i=0;i<N;i++)
	{
		mu=tpf.sampleMu();
		index=int((double(bins)*(mu+1.0))/2.0)%bins;
		frequency[index]++;
	}
	double factor=N/bins;
	for(i=0;i<bins;i++)
	{
		frequency[i]/=factor;
		mu=-1.0+double(2*i+1)/double(bins);
		printf("%lg\t%lg\n",mu,frequency[i]);
	}
	delete [] frequency;
	return 0;
}

#endif
