/**************************************************************************************************************
*
*       This file,
*
*                            /PlaneParallel1DGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "PlaneParallel1DGrid.h"
#include "Geometry.h"
#include "error.h"

bool PlaneParallel1DGrid::loadFiles()
{
	char filename[1000];
	formatZGridFilename(filename);
	
	externalX[2].defineNameAndUnit("altitude", "km");
	if(!externalX[2].loadFromFile(filename))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "failed to load altitude grid!\n");
		return false;
	}
	Z_Res=externalX[2].getResolution();

	internalX[2].defineNameAndUnit("altitude", "km");
	if(!internalX[2].setResolution(Z_Res))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"could not allocate internal altitude grid with resolution %d!\n", Z_Res);
		return false;
	}
	// construct tracing grid
	int i;
	double *ih=internalX[2].getGridPtr();
	double *eh=externalX[2].getGridPtr();
	for(i=0;i<=Z_Res;i++)
		*ih++=*eh++;
		
	return internalX[2].checkGrid() && calculateGroundLayerIndex();
}

bool PlaneParallel1DGrid::contains(class Vector3D& point)
{
	return point[2]>=ground_level && point[2]<=internalX[2].getB();
}

bool PlaneParallel1DGrid::calculateCellIntersectionPoint(class Vector3D& Position, struct CellIndex_s *pCI,\
	 class Vector3D& Direction, double *pDistance, class Vector3D *pNextPosition,\
		struct CellIndex_s *pNextCI)
{
	int iZ=pCI->coordinates.cartesic.iZ;

	double pz=Position[2], kz=Direction[2], lambda_z, *z=internalX[2].getGridPtr();

	bool positive_kz=(kz>=0.0);

	// z --------------------------------------------------------------------

	if(positive_kz)
		lambda_z=(z[iZ+1]-pz)/kz;
	else
		lambda_z=(z[iZ]-pz)/kz;

	// find direction -------------------------------------------------------


	pNextCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	*pNextCI=*pCI;

	if(positive_kz)
	{
		pNextCI->coordinates.cartesic.iZ++;
		if(pNextCI->coordinates.cartesic.iZ>=Z_Res)
			pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
	}
	else
	{
		pNextCI->coordinates.cartesic.iZ--;
		if(pNextCI->coordinates.cartesic.iZ<ground_layer_index)
			pNextCI->CellPosition=CELL_POSITION_SURFACE;
	}

	if(pSurfaceGrid->useReliefData() && pSurfaceGrid->isReliefCell(pCI))
	{

		// there is maybe a surface relief intersection....
		double distance;
		class Vector3D NextPosition;
		pSurfaceGrid->calculateReliefIntersectionPoint(pCI, Position, Direction, &distance, &NextPosition);
		if(distance<lambda_z)
		{
			*pNextPosition=NextPosition;
			*pDistance=distance;
			*pNextCI=*pCI;
			pNextCI->CellPosition=CELL_POSITION_SURFACE;
			return true;
		}
	}

	*pNextPosition=Position+lambda_z*Direction;
	*pDistance=lambda_z;
	if(pNextCI->CellPosition==CELL_POSITION_SURFACE)
	{
		class Vector3D ZDirection=Vector3D(0.0,0.0,1.0);
		pSurfaceGrid->defineSurfacePointAndNormal(*pNextPosition, ZDirection);
	}
	pNextCI->event=TE_REFRACTION_Z;
	return true;
}

bool PlaneParallel1DGrid::calculateCellIndex(class Vector3D& point, struct CellIndex_s *pCI)
{
	double z=point[2];
	if(z<ground_level)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(z>internalX[2].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	if(!internalX[2].findIntervallIndexFastButUnsafe(z, &(pCI->coordinates.cartesic.iZ)))
		return false;
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

bool PlaneParallel1DGrid::calculateCellIndex(class Vector3D& point, class Vector3D& direction, struct CellIndex_s *pCI)
{
	double z=point[2];
	if(z<ground_level)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(z>internalX[2].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	int iZ=pCI->coordinates.cartesic.iZ;
	if(!internalX[2].findIntervallIndexFastButUnsafe(z, &iZ))
		return false;
	if(internalX[2][iZ+1]==z && direction[2]>0.0)
		iZ++;
	pCI->coordinates.cartesic.iZ=iZ;
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

void PlaneParallel1DGrid::initializeBoxSamplingStructure(class MacroCell *pMacroCell, struct BoxDetector::SamplingStructure_s *pSamplingStructure)
{
	pSamplingStructure->x1=0.0;
	pSamplingStructure->x2=0.0;
	pSamplingStructure->Nx=0.0;
	pSamplingStructure->cx=0.0;
	pSamplingStructure->dx=1e-6;

	pSamplingStructure->y1=0.0;
	pSamplingStructure->y2=0.0;
	pSamplingStructure->Ny=0.0;
	pSamplingStructure->cy=0.0;
	pSamplingStructure->dy=1e-6;

	pSamplingStructure->z1=pMacroCell->internal_xA(2);
	pSamplingStructure->z2=pMacroCell->internal_xB(2);
	pSamplingStructure->Nz=pSamplingStructure->z2-pSamplingStructure->z1;
	pSamplingStructure->cz=pSamplingStructure->z1;
	pSamplingStructure->dz=1e-6;

	// sampling functions
	pSamplingStructure->takeXSample=&SampleCartesicCoordinate;
	pSamplingStructure->takeYSample=&SampleCartesicCoordinate;
	pSamplingStructure->takeZSample=&SampleCartesicCoordinate;

	// volumes / areas
	pSamplingStructure->size[(int)BoxDetector::fullbox]=pSamplingStructure->Nz;

	pSamplingStructure->size[(int)BoxDetector::boxtopface]=0.0;
	pSamplingStructure->size[(int)BoxDetector::boxbottomface]=0.0;

	pSamplingStructure->size[(int)BoxDetector::boxnorthface]=0.0;
	pSamplingStructure->size[(int)BoxDetector::boxsouthface]=0.0;

	pSamplingStructure->size[(int)BoxDetector::boxeastface]=0.0;
	pSamplingStructure->size[(int)BoxDetector::boxwestface]=0.0;

	pSamplingStructure->size[(int)BoxDetector::boxface]=0.0;
	for(int i=(int)BoxDetector::boxtopface;i<=(int)BoxDetector::boxwestface;i++)
		pSamplingStructure->size[(int)BoxDetector::boxface]+=pSamplingStructure->size[i];
}

void PlaneParallel1DGrid::printCell(FILE *fptr, struct CellIndex_s& CI)
{
	int iZ=CI.coordinates.cartesic.iZ;
	class Grid1D& Z=internalX[2];
	fprintf(fptr, "z[%d]=[%lf,%lf]\n", iZ, Z[iZ], Z[iZ+1]);
}

bool PlaneParallel1DGrid::acceptGeometry(class Geometry *pGeometry)
{
	int index;
	double c=pGeometry->getZ();
	if(externalX[2].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector Z coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	return true;
}

