/**************************************************************************************************************
*
*       This file,
*
*                            /PlaneParallel3DGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "PlaneParallel3DGrid.h"
#include "error.h"
#include "Geometry.h"

bool PlaneParallel3DGrid::loadFiles()
{
	char filename_X[1000], filename_Y[1000], filename_Z[1000];

	formatXGridFilename(filename_X);
	formatYGridFilename(filename_Y);
	formatZGridFilename(filename_Z);

	struct
	{
		char *filename;
		const char *name;
		int *pRes;
	}GridInitStruct[3]=
	{
		{filename_X, "x", &X_Res},
		{filename_Y, "y", &Y_Res},
		{filename_Z, "z", &Z_Res}
	};
	
	int i, j;
	double *ig, *eg;
	for(i=0;i<3;i++)
	{
		externalX[i].defineNameAndUnit(GridInitStruct[i].name, "km");
		if(!externalX[i].loadFromFile(GridInitStruct[i].filename))
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr, "failed to load %s grid, from file \"%s\"!\n", \
				GridInitStruct[i].name, GridInitStruct[i].filename);
			return false;
		}
		*GridInitStruct[i].pRes=externalX[i].getResolution();

		internalX[i].defineNameAndUnit(GridInitStruct[i].name, "km");
		if(!internalX[i].setResolution(*GridInitStruct[i].pRes))
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr,"could not allocate internal %s grid with resolution %d!\n", \
				GridInitStruct[i].name, *GridInitStruct[i].pRes);
			return false;
		}

		ig=internalX[i].getGridPtr();
		eg=externalX[i].getGridPtr();
		for(j=0;j<=*GridInitStruct[i].pRes;j++)
			*ig++=*eg++;
			
		if(!internalX[i].checkGrid())
		{
			errorIn(__FILE__, __LINE__);
			fprintf(stderr, "%s grid check failed!\n", GridInitStruct[i].name);
			return false;
		}
	}

	return calculateGroundLayerIndex();
}

bool PlaneParallel3DGrid::contains(class Vector3D& point)
{
	double x=point[0], y=point[1], z=point[2];
	return x>=internalX[0].getA() && x<=internalX[0].getB() &&\
		y>=internalX[1].getA() && y<=internalX[1].getB() &&\
		z>=internalX[2].getA() && z<=internalX[2].getB();
}

bool PlaneParallel3DGrid::calculateCellIntersectionPoint(class Vector3D& Position, struct CellIndex_s *pCI,\
	 class Vector3D& Direction, double *pDistance, class Vector3D *pNextPosition,\
		struct CellIndex_s *pNextCI)
{
	int iX=pCI->coordinates.cartesic.iX;
	int iY=pCI->coordinates.cartesic.iY;
	int iZ=pCI->coordinates.cartesic.iZ;

	double px=Position[0], kx=Direction[0], lambda_x, *x=internalX[0].getGridPtr();
	double py=Position[1], ky=Direction[1], lambda_y, *y=internalX[1].getGridPtr();
	double pz=Position[2], kz=Direction[2], lambda_z, *z=internalX[2].getGridPtr();

	bool positive_kx=(kx>=0.0);
	bool positive_ky=(ky>=0.0);
	bool positive_kz=(kz>=0.0);

	// x --------------------------------------------------------------------

	if(positive_kx)
		lambda_x=(x[iX+1]-px)/kx;
	else
		lambda_x=(x[iX]-px)/kx;

	// y --------------------------------------------------------------------

	if(positive_ky)
		lambda_y=(y[iY+1]-py)/ky;
	else
		lambda_y=(y[iY]-py)/ky;

	// z --------------------------------------------------------------------

	if(positive_kz)
		lambda_z=(z[iZ+1]-pz)/kz;
	else
		lambda_z=(z[iZ]-pz)/kz;

	// find direction -------------------------------------------------------

	// pNextCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	*pNextCI=*pCI;
	double lambda_min;
	if(lambda_x<lambda_y)
		if(lambda_x<lambda_z)
		{
			lambda_min=lambda_x;
			*pNextPosition=Position+lambda_min*Direction;
			pNextCI->event=TE_REFRACTION_X;
			if(positive_kx)
			{
				pNextCI->coordinates.cartesic.iX++;
				if(pNextCI->coordinates.cartesic.iX>=X_Res)
				{
					if(b_cyclic_x)
					{
						pNextCI->coordinates.cartesic.iX=0;
						(*pNextPosition)[0]=internalX[0].getA();
						//pNextPosition->x[1]=Position.x[1];
						//pNextPosition->x[2]=Position.x[2];
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_X_B;
				}
			}
			else
			{
				pNextCI->coordinates.cartesic.iX--;
				if(pNextCI->coordinates.cartesic.iX<0)
				{
					if(b_cyclic_x)
					{
						pNextCI->coordinates.cartesic.iX=X_Res-1;
						(*pNextPosition)[0]=internalX[0].getB();
						//pNextPosition->x[1]=Position.x[1];
						//pNextPosition->x[2]=Position.x[2];
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_X_A;
				}
			}
		}
		else
		{
			lambda_min=lambda_z;
			*pNextPosition=Position+lambda_min*Direction;
			pNextCI->event=TE_REFRACTION_Z;
			if(positive_kz)
			{
				pNextCI->coordinates.cartesic.iZ++;
				if(pNextCI->coordinates.cartesic.iZ>=Z_Res)
				{
					if(b_cyclic_z)
					{
						pNextCI->coordinates.cartesic.iZ=0;
						//pNextPosition->x[0]=Position.x[0];
						//pNextPosition->x[1]=Position.x[1];
						(*pNextPosition)[2]=ground_level;
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
				}
			}
			else
			{
				pNextCI->coordinates.cartesic.iZ--;
				if(pNextCI->coordinates.cartesic.iZ<ground_layer_index)
				{
					if(b_cyclic_z)
					{
						pNextCI->coordinates.cartesic.iZ=Z_Res-1;
						//pNextPosition->x[0]=Position.x[0];
						//pNextPosition->x[1]=Position.x[1];
						(*pNextPosition)[2]=internalX[2].getB();
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_SURFACE;
				}
			}
		}
	else
	{
		if(lambda_y<lambda_z)
		{
			lambda_min=lambda_y;
			*pNextPosition=Position+lambda_min*Direction;
			pNextCI->event=TE_REFRACTION_Y;
			if(positive_ky)
			{
				pNextCI->coordinates.cartesic.iY++;
				if(pNextCI->coordinates.cartesic.iY>=Y_Res)
				{
					if(b_cyclic_y)
					{
						pNextCI->coordinates.cartesic.iY=0;
						//pNextPosition->x[0]=Position.x[0];
						(*pNextPosition)[1]=internalX[1].getA();
						//pNextPosition->x[2]=Position.x[2];
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_Y_B;
				}
			}
			else
			{
				pNextCI->coordinates.cartesic.iY--;
				if(pNextCI->coordinates.cartesic.iY<0)
				{
					if(b_cyclic_y)
					{
						pNextCI->coordinates.cartesic.iY=Y_Res-1;
						//pNextPosition->x[0]=Position.x[0];
						(*pNextPosition)[1]=internalX[1].getB();
						//pNextPosition->x[2]=Position.x[2];
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_Y_A;
				}
			}
		}
		else
		{
			lambda_min=lambda_z;
			*pNextPosition=Position+lambda_min*Direction;
			pNextCI->event=TE_REFRACTION_Z;
			if(positive_kz)
			{
				pNextCI->coordinates.cartesic.iZ++;
				if(pNextCI->coordinates.cartesic.iZ>=Z_Res)
				{
					if(b_cyclic_z)
					{
						pNextCI->coordinates.cartesic.iZ=0;
						//pNextPosition->x[0]=Position.x[0];
						//pNextPosition->x[1]=Position.x[1];
						(*pNextPosition)[2]=ground_level;
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
				}
			}
			else
			{
				pNextCI->coordinates.cartesic.iZ--;
				if(pNextCI->coordinates.cartesic.iZ<ground_layer_index)
				{
					if(b_cyclic_z)
					{
						pNextCI->coordinates.cartesic.iZ=Z_Res-1;
						//pNextPosition->x[0]=Position.x[0];
						//pNextPosition->x[1]=Position.x[1];
						(*pNextPosition)[2]=internalX[2].getB();
						goto skip_set_next_position;
					}
					else
						pNextCI->CellPosition=CELL_POSITION_SURFACE;
				}
			}
		}
	}

	if(pSurfaceGrid->useReliefData() && pSurfaceGrid->isReliefCell(pCI))
	{
		// there is maybe a surface relief intersection....
		double distance;
		class Vector3D NextPosition;
		pSurfaceGrid->calculateReliefIntersectionPoint(pCI, Position, Direction, &distance, &NextPosition);
		if(distance<lambda_min)
		{
			*pNextPosition=NextPosition;
			*pDistance=distance;
			*pNextCI=*pCI;
			pNextCI->CellPosition=CELL_POSITION_SURFACE;
			return true;
		}
	}

skip_set_next_position:

	*pDistance=lambda_min;
	if(pNextCI->CellPosition==CELL_POSITION_SURFACE)
	{
		class Vector3D ZDirection=Vector3D(0.0, 0.0, 1.0);
		// pNextPosition->x[2]=1e-8;
		pSurfaceGrid->defineSurfacePointAndNormal(*pNextPosition, ZDirection);
	}

	return true;

}

bool PlaneParallel3DGrid::calculateCellIndex(class Vector3D& point, struct CellIndex_s *pCI)
{
	double x=point[0], y=point[1], z=point[2];
	if(z<ground_level)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(z>internalX[2].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}

	if(x<internalX[0].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_A;
		return false;
	}
	else if(x>internalX[0].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_B;
		return false;
	}

	if(y<internalX[1].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	else if(y>internalX[1].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_B;
		return false;
	}

	if(!(internalX[0].findIntervallIndexFastButUnsafe(x, &(pCI->coordinates.cartesic.iX)) && \
		internalX[1].findIntervallIndexFastButUnsafe(y, &(pCI->coordinates.cartesic.iY)) && \
		internalX[2].findIntervallIndexFastButUnsafe(z, &(pCI->coordinates.cartesic.iZ))))
		return false;

	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return  true;
}

bool PlaneParallel3DGrid::calculateCellIndex(class Vector3D& point, class Vector3D& direction, struct CellIndex_s *pCI)
{
	double x=point[0], y=point[1], z=point[2];
	if(z<ground_level)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(z>internalX[2].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}

	if(x<internalX[0].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_A;
		return false;
	}
	else if(x>internalX[0].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_B;
		return false;
	}

	if(y<internalX[1].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	else if(y>internalX[1].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_B;
		return false;
	}
	int iX=pCI->coordinates.cartesic.iX;
	int iY=pCI->coordinates.cartesic.iY;
	int iZ=pCI->coordinates.cartesic.iZ;
	if(!(internalX[0].findIntervallIndexFastButUnsafe(x, &iX) && \
		internalX[1].findIntervallIndexFastButUnsafe(y, &iY) && \
		internalX[2].findIntervallIndexFastButUnsafe(z, &iZ)))
		return false;

	if(internalX[0][iX+1]==x && direction[0]>0.0)
		iX++;
	pCI->coordinates.cartesic.iX=iX;

	if(internalX[1][iY+1]==y && direction[1]>0.0)
		iY++;
	pCI->coordinates.cartesic.iY=iY;

	if(internalX[2][iZ+1]==z && direction[2]>0.0)
		iZ++;

	pCI->coordinates.cartesic.iX=iX;
	pCI->coordinates.cartesic.iY=iY;
	pCI->coordinates.cartesic.iZ=iZ;

	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return  true;
}

void PlaneParallel3DGrid::initializeBoxSamplingStructure(class MacroCell *pMacroCell, struct BoxDetector::SamplingStructure_s *pSamplingStructure)
{
	pSamplingStructure->x1=pMacroCell->internal_xA(0);
	pSamplingStructure->x2=pMacroCell->internal_xB(0);
	pSamplingStructure->Nx=pSamplingStructure->x2-pSamplingStructure->x1;
	pSamplingStructure->cx=pSamplingStructure->x1;
	pSamplingStructure->dx=1e-6;

	pSamplingStructure->y1=pMacroCell->internal_xA(1);
	pSamplingStructure->y2=pMacroCell->internal_xB(1);
	pSamplingStructure->Ny=pSamplingStructure->y2-pSamplingStructure->y1;
	pSamplingStructure->cy=pSamplingStructure->y1;
	pSamplingStructure->dy=1e-6;

	pSamplingStructure->z1=pMacroCell->internal_xA(2);
	pSamplingStructure->z2=pMacroCell->internal_xB(2);
	pSamplingStructure->Nz=pSamplingStructure->z2-pSamplingStructure->z1;
	pSamplingStructure->cz=pSamplingStructure->z1;
	pSamplingStructure->dz=1e-6;

	// sampling functions
	pSamplingStructure->takeXSample=&SampleCartesicCoordinate;
	pSamplingStructure->takeYSample=&SampleCartesicCoordinate;
	pSamplingStructure->takeZSample=&SampleCartesicCoordinate;

	// volumes / areas
	pSamplingStructure->size[(int)BoxDetector::fullbox]=pSamplingStructure->Nx*pSamplingStructure->Ny*pSamplingStructure->Nz;

	pSamplingStructure->size[(int)BoxDetector::boxtopface]=pSamplingStructure->Nx*pSamplingStructure->Ny;
	pSamplingStructure->size[(int)BoxDetector::boxbottomface]=pSamplingStructure->Nx*pSamplingStructure->Ny;

	pSamplingStructure->size[(int)BoxDetector::boxnorthface]=pSamplingStructure->Nx*pSamplingStructure->Nz;
	pSamplingStructure->size[(int)BoxDetector::boxsouthface]=pSamplingStructure->Nx*pSamplingStructure->Nz;

	pSamplingStructure->size[(int)BoxDetector::boxeastface]=pSamplingStructure->Ny*pSamplingStructure->Nz;
	pSamplingStructure->size[(int)BoxDetector::boxwestface]=pSamplingStructure->Ny*pSamplingStructure->Nz;

	pSamplingStructure->size[(int)BoxDetector::boxface]=0.0;
	for(int i=(int)BoxDetector::boxtopface;i<=(int)BoxDetector::boxwestface;i++)
		pSamplingStructure->size[(int)BoxDetector::boxface]+=pSamplingStructure->size[i];
}

void PlaneParallel3DGrid::printCell(FILE *fptr, struct CellIndex_s& CI)
{
	int iX=CI.coordinates.cartesic.iX;
	int iY=CI.coordinates.cartesic.iY;
	int iZ=CI.coordinates.cartesic.iZ;
	class Grid1D& X=internalX[0];
	class Grid1D& Y=internalX[1];
	class Grid1D& Z=internalX[2];
	fprintf(fptr, "x[%d]=[%lf,%lf], y[%d]=[%lf,%lf], z[%d]=[%lf,%lf]\n", iX, X[iX], X[iX+1], iY, Y[iY], Y[iY+1], iZ, Z[iZ], Z[iZ+1]);
}

bool PlaneParallel3DGrid::acceptGeometry(class Geometry *pGeometry)
{
	int index;
	double c=pGeometry->getX();
	if(externalX[0].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector X coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	c=pGeometry->getY();
	if(externalX[1].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector Y coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	c=pGeometry->getZ();
	if(externalX[2].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector Z coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	return true;
}



