/**************************************************************************************************************
*
*       This file,
*
*                            /ProbabilityDensity1D.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "ProbabilityDensity1D.h"
#include "error.h"

#include "Math/prand.h"

#include <math.h>

bool ProbabilityDensity1D::normalize()
{
	int i;
	double dx, df, integral;
	normalisation_constant=0.0;

	// calculate intervall sizes
	for(i=0;i<no_intervals;i++)
	{
		PDFData[i][column_interval_size] = dx = PDFData[i+1][column_ordinate] - PDFData[i][column_ordinate] ;
		PDFData[i][column_bin_area] = df = ( dx * ( PDFData[i][column_function] + PDFData[i+1][column_function] ) ) / 2.0 ;
		normalisation_constant += df;
	}
	
	// normalisation
	integral=0.0;
	for(i=0;i<no_intervals;i++)
	{
		df = PDFData[i][column_bin_area] /= normalisation_constant ; 
		PDFData[i][column_density] = PDFData[i][column_function] / normalisation_constant;
		PDFData[i][column_cumulative_density] = integral ;
		integral += df;
	}
	PDFData[i][column_density] = PDFData[i][column_function] / normalisation_constant;
	PDFData[i][column_cumulative_density] = integral;

	// check
	if(fabs(integral-1.0)>1e-12)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "normalisation failed: N=%le ( i.e. NOT 1 ) \n", integral);
		return false;
	}
	return true;
}

bool ProbabilityDensity1D::load(char *file_name)
{
	int i;
	class Matrix Data;
	if(!Data.load(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(!PDFData.allocate(Data.getM(), no_columns))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	PDFData.setToZero();
	no_intervals=Data.getM()-1;
	for(i=0;i<=no_intervals;i++)
	{
		PDFData[i][0]=Data[i][0];
		PDFData[i][1]=Data[i][1];
	}
	return normalize();
}

double ProbabilityDensity1D::takeSample()
{
	double alpha = prand(), CDF;
	// bisection search
	int i, i1=0, i2=no_intervals;
	while ( i1 + 1 < i2 )
	{
		i = ( i1 + i2 ) / 2 ;
		CDF = PDFData[i][column_cumulative_density];
		if ( alpha == CDF )
			return PDFData[i][column_ordinate];
		else if ( alpha > CDF )
			i1 = i;
		else
			i2 = i;
	}
	// quadratic interpolation
	double dp = alpha;
	dp -= PDFData[i1][column_cumulative_density];
	// AccumulatedP11BinArea[mu_index-1];

	// interpolate
	double a = PDFData[i1][column_ordinate];
	double b = PDFData[i2][column_ordinate];

	double fa = PDFData[i1][column_density];
	double fb = PDFData[i2][column_density];
/*
	                 fb *
                            |
                            |
	     fa *           |
	        |           |
	        |           |
	        |           |
	        |           |
	        |           |
	        |           |
	--------+-----------+----
                a           b
*/
	double df = fb - fa;
	double bma = b - a;

	if ( df != 0.0 )
	{
		double m = df / bma;
		double c = fa;
		// f(x) = m * x + c , x in [ 0 , b - a ]

		// int f(x) = 0.5 * m * x^2 + c * x = dp
		
		double p = c / m ;
		double q = 2.0 * dp / m ;

		double radicand = p * p + q ;

		double sqrt_radicand = sqrt ( radicand ) ;	
		double solution = - p + sqrt_radicand ;

		if ( solution < 0.0 || solution > bma )
		{
			solution = - p - sqrt_radicand ;
			if ( solution < 0.0 || solution > bma )
			{
				errorIn ( __FILE__, __LINE__, __func__ ) ;
				fprintf ( stderr, "quadratic pg formula not solvable\n" ) ;
				return 0.5 * ( a + b ) ;
			}
		}
		return a + solution ;
	}
	return a + dp / fa  ;
}





