/**************************************************************************************************************
*
*       This file,
*
*                            /Profile.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Profile.h"
#include "error.h"

#include "fileio.h"
#include "stringio.h"

#include <stdlib.h>
#include <float.h>

#include "Atmosphere.h"
#include "Simulation.h"

bool Profile::b_output_polarized_quantities=false;

Profile::Profile()
{
	cluster=0;
	cell_index_data=0;

	optical_cluster_dimension=0;
	auxiliary_data_dimension=0;
	cluster_dimension=0;

	retrieve_profile=true;
}

Profile::~Profile()
{
	if(cluster)
		delete [] cluster;
	if(cell_index_data)
		delete [] cell_index_data;

}

bool Profile::initializeClusterArray()
{
	int i, j, *intptr;
	for(j=0;j<getNOClusters();j++)
		cluster[j].no_cells=0;
	for(i=0;i<no_cells;i++)
	{
		j=map[i];
		if(j>=0)
			cluster[j].no_cells++;
	}
	if(cell_index_data)
		delete [] cell_index_data;
	cell_index_data=new int[no_cells];
	if(!cell_index_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error\n");
		return false;
	}
	intptr=cell_index_data;
	// allocate cluster list
	for(j=0;j<getNOClusters();j++)
	{
		cluster[j].cell_index=intptr;
		intptr+=cluster[j].no_cells;
		cluster[j].no_cells=0;	// use no_cells as a counter
	}
	// create cluster list
	for(i=0;i<no_cells;i++)
	{
		j=map[i];
		if(j>=0)
		{
			cluster[j].cell_index[cluster[j].no_cells]=i;
			cluster[j].no_cells++;
		}
	}
	return true;
}

bool Profile::allocateClusterData()
{
	if(!OpticalData.allocate(no_clusters, optical_cluster_dimension))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating optical cluster data\n");
		return false;
	}
	if(auxiliary_data_dimension && !AuxiliaryData.allocate(no_clusters, auxiliary_data_dimension))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating auxiliary data\n");
		return false;
	}
	return true;
}

void Profile::setClusterDimensions(int input_cluster_dimension)
{
/*
	default function
*/
	cluster_dimension=input_cluster_dimension;
	auxiliary_data_dimension=1+cluster_dimension;
	optical_cluster_dimension=1+input_cluster_dimension;
}

bool Profile::loadClusterData(char *file_name)
{
	if(!InputData.load(file_name))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not open file \"%s\"!\n", file_name);
		return false;
	}

	// reset number of clusters
	setNOClusters(InputData.getM());
	setClusterDimensions(InputData.getN());
	if(!allocateClusterData())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(cluster)
		delete [] cluster;
	cluster=new Cluster_s[no_clusters];
	if(!cluster)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	int i;
	for(i=0;i<getNOClusters();i++)
	{
		cluster[i].valid=false;
		cluster[i].cell_index=0;
	}
	// allocate Jacobian data
	for(i=0;i<no_clusters;i++)
	{
		cluster[i].input_data=InputData[i];
		cluster[i].optical_data=OpticalData[i];
		cluster[i].auxiliary_data=auxiliary_data_dimension ? AuxiliaryData[i] : 0;
	}
	return true;
}

void Profile::invalidateAllClusters()
{
	int i, n=getNOClusters();
	for(i=0;i<n;i++)
		invalidateCluster(i);
}

void Profile::invalidateCluster(const int& i)
{
	struct Cluster_s *pCluster=cluster+i;
	int j, n=pCluster->no_cells;
	int *cell_index=pCluster->cell_index;
	for(j=0;j<n;j++)
		pAtmosphere->invalidateCell(cell_index[j]);
	pCluster->valid=false;
}

void Profile::writeAuxiliaryFiles()
{
	char path[BasicIO_string_size], file_name[BasicIO_string_size];
	formatOutputDirectory(path);
	sprintf(file_name, "%s%s_ocd.txt", path, profile_name);
	OpticalData.save(file_name);
}



