/**************************************************************************************************************
*
*       This file,
*
*                            /Raytracer.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Raytracer.h"
#include "Math/prand.h"

#include "refraction.h"

#include "trigonometry.h"
#include "constants.h"
#include "error.h"

#include <time.h>

#include "Atmosphere.h"
#include "Surface.h"
#include "Geometry.h"
#include "GeometryList.h"
#include "Simulation.h"

#ifdef USE_GUI
	#include "./gui/PathViewer/PathViewerWidget.h"
#endif

Raytracer::Raytracer()
{
	RTEvent=(struct RadiativeTransferEvent_s *)0;
	running=false;
	// use_fixed_initial_direction=false;

	wavelength=0.0;
}

Raytracer::~Raytracer()
{
	if(RTEvent)
		delete [] RTEvent;
}

bool Raytracer::allocateRTEvents(int n)
{
	if(RTEvent)
		delete [] RTEvent;
	RTEvent=new struct RadiativeTransferEvent_s[n];
	if(!RTEvent)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	memset(RTEvent, 0, sizeof(struct RadiativeTransferEvent_s)*n);
	/*
	if(estimated_scatter_order)
		delete [] estimated_scatter_order;
	estimated_scatter_order=new bool[n];
	if(!estimated_scatter_order)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	*/
	return true;
}

bool Raytracer::initialize()
{
	// radiative transfer events
	if(!allocateRTEvents(maximum_no_trace_events))
		return false;
	if(b_enable_variance_reduction)
		if(!VarianceReductionLayer::setup(maximum_no_trace_events))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	/*
	if(b_calculate_cell_path_length && !TrajectoryInformationObject::allocate(maximum_no_trace_events))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}*/
	//TwilightRTEvent=RTEvent+maximum_no_trace_events;
	// calculate cT
	Re=pGrid->getPlanetRadius();
	double sin_theta=Re/AU;
	double theta=asin(sin_theta);
	cT_square=cos(theta);
	// printf("sizeof(RadiativeTransferEvent_s)=%d\n", int(sizeof(RadiativeTransferEvent_s)));

	return true;
}

bool Raytracer::consistenceCheck(struct RadiativeTransferEvent_s *Event, int n)
{
	// tests the event chain for consistency
	int i;
	struct RadiativeTransferEvent_s *pEvent=Event;
	// cell index check
	struct CellIndex_s CI;
	bool indices_equal;
	for(i=0;i<n;i++)
	{
		if(!pGrid->calculateCellIndex(pEvent->position, &CI))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"event chain point outside the lattice. Event index: %d\n",i+1);
			return false;
		}
		indices_equal=(CI.coordinates.general.iC[0]==pEvent->CellIndex.coordinates.general.iC[0])\
	 		&& (CI.coordinates.general.iC[1]==pEvent->CellIndex.coordinates.general.iC[1])\
			&& (CI.coordinates.general.iC[2]==pEvent->CellIndex.coordinates.general.iC[2]);
		if(!indices_equal)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "Event index %d: point - cell index inconsistence\n", i+1);
			return false;
		}
	}
	return true;
}

/*void Raytracer::saveTrajectorySnippet(int i1, int i2)
{
	int nt = 1 + i2 - i1, n = nt + no_forced_path_RT_events;
	// write number trace events for the snippet
	if(b_ASCII_trajectory_output)
	{
		writeTrajectoryASCII ( scattered_stream, RTEvent + i1, nt);
		writeTrajectoryASCII ( scattered_stream, ForcedPathRTEvent, no_forced_path_RT_events);
	}
	else
	{
		// printf("%d (%d+%d) events!\n", n, nt, no_forced_path_RT_events);
		fwrite(&n, sizeof(int), 1, scattered_stream);
		if(nt)
			fwrite ( RTEvent + i1, sizeof(RadiativeTransferEvent_s), nt, scattered_stream );
		if(no_forced_path_RT_events)
			fwrite ( ForcedPathRTEvent, sizeof(RadiativeTransferEvent_s), no_forced_path_RT_events, scattered_stream );
	}
	total_no_events+=n;
}*/

class Vector3D Raytracer::retraceTrajectoryUntilExtinction(struct RadiativeTransferEvent_s *pRTEvent0, const double& tau_interaction)
{
	struct CellIndex_s CellIndex;
	double distance_to_scatter_event, wavelength;
	class Vector3D position, direction;
	register struct RadiativeTransferEvent_s *pRTEvent=pRTEvent0;
	while(pRTEvent->tau_extinction<tau_interaction)
	{
		pRTEvent++;
		if(pRTEvent>=RTEvent+maximum_no_trace_events)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return position;
		}
	}
	pCurrentRTEvent=pRTEvent;
	--pRTEvent;

	// get relevant information
	CellIndex=pRTEvent->CellIndex;
	position=pRTEvent->position;
	direction=pRTEvent->direction;
	wavelength=pRTEvent->wavelength;
	
	// determine extinction coefficient
	distance_to_scatter_event=(tau_interaction-pRTEvent->tau_extinction)/ \
		pAtmosphere->getOpticalCellPtr(CellIndex, wavelength)->extinction_coefficient;
	pRTEvent++;

	// calculate snippet
	pRTEvent->CellIndex=CellIndex;
	pRTEvent->direction=direction;
	pRTEvent->distance=distance_to_scatter_event;
	pRTEvent->wavelength=wavelength;
	return position+distance_to_scatter_event*direction;
}

void Raytracer::propagateTrajectory(struct RadiativeTransferEvent_s *pRTEventStart, struct RadiativeTransferEvent_s *pRTEventEnd, class Estimate *pEstimate)
{
	int cell_index;
	register struct RadiativeTransferEvent_s *pRTEvent=pRTEventStart;
	for ( pRTEvent = pRTEventStart ; pRTEvent <= pRTEventEnd ; pRTEvent++ )
	{
		cell_index=pGrid->calculateCellIndex((pRTEvent-1)->CellIndex);
		processPropagation(pEstimate, cell_index, pRTEvent->distance);
	}
}

bool Raytracer::generateSISTrajectory(class Estimate *pEstimate, class Vector3D p, class Vector3D k, CellIndex_s CI, double lambda)
{
/*
	This function uses the following global variables:
		position
		direction
		CellIndex
		wavelength

*/

	// static variables
	double distance, albedo, reflection_coefficient=0.0;
	class Vector3D next_p, last_k, k_Sun;
	struct CellIndex_s NextCI;
	double tau_interaction, tau_snippet;

	// weight window
	double ww_rr, ww_optimum, ww_split, weight;
	bool ww_enabled;
	
	// optical properties
	double tau_extinction=0.0, tau_extinction_detector_to_ground=0.0;
	double absorption_coefficient, scatter_coefficient, extinction_coefficient;
	struct Atmosphere::OpticalCell_s *pOC;

	double alpha_scatter_object, interaction_distance;

	struct Particle::OpticalStruct_s **pPOStructPtr, *pPOStruct;
	class PhaseFunction *pPhaseFunction;
	double PhaseFunctionISWeight;

	int cell_index;
	bool continue_tracing, no_interaction, found_scatter_event, fpl_tracing, propagate_light_path;
	struct RadiativeTransferEvent_s *pRTEvent0=pCurrentRTEvent, *pGroundRTEvent=0;

	// Markov chain creation loop
	memset(&NextCI, 0, sizeof(struct CellIndex_s));

	continue_tracing=true;
	no_interaction=true;
	while(1)
	{
		pCurrentRTEvent->contribution=0.0;
		tau_interaction=-log(nzprand());
find_extinction_point:
		if(!pGrid->calculateCellIntersectionPoint(p, &CI, k, &distance, &next_p, &NextCI))
		{
			double r, phi, theta;
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not calculate cell intersection point\n\n");
			if(pGrid->isSphericalGrid())
			{
				p.toPolar(&r, &theta, &phi);
				fprintf(stderr, "r=%lg, theta=%lg, phi=%lg, ", r, Degree(theta), Degree(phi));
				k.toPolar(&theta, &phi);
				fprintf(stderr, "theta_k = %lf deg, phi_k = %lf deg\n", Degree(theta), Degree(phi));
			}
			else
			{
				fprintf(stderr, "p = (%lg, %lg, %lg), ", p[0], p[1], p[2]);
				fprintf(stderr, "k = (%lg, %lg, %lg)\n", k[0], k[1], k[2]);
			}
			// fprintf(stderr, "CI[1]=%d, CI[2]=%d, CI[3]=%d\n", CI.iC[0], CI.iC[1], CI.iC[2]);
			pGrid->printCell(stderr, CI);
			writeTrajectoryASCII(stderr, RTEvent, no_events);
			return false;
		}
		cell_index=pGrid->calculateCellIndex(CI);
		pOC=pAtmosphere->getOpticalCellPtr(cell_index, lambda);
		// pCurrentRTEvent->cell_index=cell_index;

		// copy properties:
		absorption_coefficient	= pOC->absorption_coefficient;
		scatter_coefficient	= pOC->scatter_coefficient;
		extinction_coefficient	= pOC->extinction_coefficient;

		// calculate tau attenuation (tau_snippet)
		tau_snippet = distance * extinction_coefficient;

		// calculate flag for free path length tracing (following if block)
		if(b_automatic_switch_selection)
		{
			if(no_scatter_events)
			{
				if(tau_interaction>tau_snippet)
					fpl_tracing=true;
				else
					fpl_tracing=false;
				propagate_light_path=true;
			}
			else // no scatter event so far
			{
				tau_extinction += tau_snippet;
				pCurrentRTEvent->tau_extinction = tau_extinction;
				fpl_tracing=true;
				propagate_light_path=false;
			}
		}
		else
		{
			propagate_light_path=true;
			if(tau_interaction>tau_snippet)
				fpl_tracing=true;
			else
				fpl_tracing=false;
		}
		if(fpl_tracing)
		{
			pCurrentRTEvent->distance=distance;

			//interaction point not met yet
			tau_interaction-=tau_snippet;

			// update cell path length estimates
			if(propagate_light_path)
				processPropagation(pEstimate, cell_index, distance);

			switch(NextCI.CellPosition)
			{
				case CELL_POSITION_INSIDE_ATMOSPHERE:// refraction
				// case CELL_POSITION_GROUND_EDGE:
					CI=NextCI;
					pCurrentRTEvent->type=NextCI.event;
					/*if(b_simulate_refraction)
					{
						class Vector3D RadialVector=norm(position);
						double N_IN=pOC->refractive_index;
						pOC=pAtmosphere->getOpticalCellPtr(NextCellIndex);
						double N_OUT=pOC->refractive_index;
						direction=doRefraction(norm(next_position), direction, N_IN, N_OUT);
					}*/
					pCurrentRTEvent->alpha=0.0;
					found_scatter_event=false;
					break;

				case CELL_POSITION_SURFACE:
					// under the ground/relief
					if(b_automatic_switch_selection)
					{
						if(no_scatter_events)
							goto normal_ground_processing;
						else
						{
							reflection_coefficient=pSurface->calculateReflectionCoefficient(next_p, k, &(pEstimate->weight));
							if(reflection_coefficient)
							{
								pGroundRTEvent = pCurrentRTEvent;
								tau_extinction_detector_to_ground = tau_extinction;
								albedo = pSurface->getAlbedo(CI, next_p, lambda);
								goto simulate_specular_reflection_at_the_ground;
							}
							else
							{
								albedo=pSurface->getAlbedo(CI, next_p, lambda);
								// enforce scattering on the interval to the ground or ground scattering!	
								double p_transmission = exp( -tau_extinction ) ;
								// absolute probabilities for both processes
								double p_extinction = 1.0 - p_transmission;		// probability that extinction (not transmission) happens
								double p_ground_scattering = p_transmission * albedo;	// probability for transmission and scattering at the ground 

								// normalize
								double p_forced_scattering = p_extinction + p_ground_scattering; // extinction or ground scattering
								double alpha_extinction = nzprand() * p_forced_scattering;
								// modification of the photon weight
								pEstimate->weight *= p_forced_scattering;
								if(alpha_extinction<p_extinction)
								{
									// sample extinction point in atmosphere
									tau_interaction = - log ( 1.0 - alpha_extinction ) ;
									next_p=retraceTrajectoryUntilExtinction(pRTEvent0, tau_interaction);
									k=pCurrentRTEvent->direction;
									CI=pCurrentRTEvent->CellIndex;
									pOC=pAtmosphere->getOpticalCellPtr(CI, pCurrentRTEvent->wavelength);
									if(!pOC)
									{
										errorIn(__FILE__, __LINE__, __func__);
										return false;
									}
									scatter_coefficient=pOC->scatter_coefficient;
									pEstimate->weight *= scatter_coefficient / pOC->extinction_coefficient;
									propagateTrajectory(pRTEvent0, pCurrentRTEvent, pEstimate);
									goto estimate_scattering_angle;
								}
								else
								{
									propagateTrajectory(pRTEvent0, pCurrentRTEvent, pEstimate);
									goto estimate_ground_incoming_radiation; // force ground scattering
								}
							}
						}// if(no_scatter_events)
					} // if(b_automatic_switch_selection)
					else
					{
normal_ground_processing:
						reflection_coefficient=pSurface->calculateReflectionCoefficient(next_p, k, &(pEstimate->weight));
						if(nzprand()>reflection_coefficient)
						{
							albedo=pSurface->getAlbedo(CI, next_p, lambda);
							if(prand()>albedo)
							{
								continue_tracing=false;
								pCurrentRTEvent->type=TE_ABSORPTION;
								found_scatter_event=false;
							}
							else
							{
estimate_ground_incoming_radiation:
								p=next_p;
								this->wavelength=lambda;
								// pSurface->sampleLambertianScatterDirection(&pCurrentRTEvent->theta, &pCurrentRTEvent->phi, k);
								pSurface->sampleLambertianScatterDirection(k);
								pEstimate->registerGroundScatterEvent();
								pEstimate->registerElasticScatterEvent();
								FunctionalLayer::processGroundScatterEvent(pCurrentRTEvent, pEstimate, next_p, CI, pSurface->getSurfaceNormal());
								pCurrentRTEvent->type=TE_GROUND_SCATTERING;
								no_scatter_events++;
								found_scatter_event=true;
								if(b_Neumann_series_truncation && no_scatter_events>=no_scatter_orders)
									continue_tracing=false;
							}
							no_interaction=false;
						}
						else
						{
simulate_specular_reflection_at_the_ground:
							pSurface->calculateSpecularlyReflectedDirection(pSurface->getSurfaceNormal(), k);
							pCurrentRTEvent->type=TE_GROUND_REFLECTION;	 // HERE!!
						}
					}
					break;

				default:
					if(b_automatic_switch_selection)
					{
						// reached TOA with automatic switch selection
						if(no_scatter_events) // normal procedure
							goto normal_reached_TOA_procedure;
						else
						{
							if(reflection_coefficient)
							{
								// there was reflection at the ground
								double p_transmission_to_ground = exp( - tau_extinction_detector_to_ground ) ;

								// 3 possibilities: 
								// 	1. scattering before ground
								// 	2. ground scattering
								// 	3. scattering after the ground

								// 1. probability that extinction between detector and ground happens
								double p_extinction_before_ground = 1.0 - p_transmission_to_ground;

								// 2. probability for transmission to ground and scattering (i.e. no reflection and no surface absorption)
								double p_ground_scattering = p_transmission_to_ground * ( 1.0 - reflection_coefficient ) * albedo ;

								// 3. probability for transmission to ground, ground reflection and scattering between the ground and TOA
								double p_transmission_from_ground_to_TOA = exp ( tau_extinction_detector_to_ground - tau_extinction );
								double p_extinction_from_ground_to_TOA = 1.0 - p_transmission_from_ground_to_TOA ;
								double p_extinction_after_ground= p_transmission_to_ground * reflection_coefficient * p_extinction_from_ground_to_TOA ;
								
								// total probability
								double p_forced_scattering = p_extinction_before_ground + p_ground_scattering + p_extinction_after_ground;

								// modification of the photon weight
								pEstimate->weight *= p_forced_scattering;
	
								// sample random number
								double alpha_extinction = nzprand() * p_forced_scattering;
								if( alpha_extinction < p_extinction_before_ground )
								{
									// 1. extinction between detector and ground happens
									tau_interaction = - log ( 1.0 - alpha_extinction );
									next_p=retraceTrajectoryUntilExtinction(pRTEvent0, tau_interaction);
									goto enforce_scatter_event_after_TOA_exit_with_surface_reflection;
								}
								else if( (alpha_extinction-= p_extinction_before_ground) < p_ground_scattering)
								{
									// 2. transmission to ground and scattering (i.e. no reflection and no surface absorption)
									pCurrentRTEvent=pGroundRTEvent;
									next_p=pGroundRTEvent->position;
									k=(pGroundRTEvent-1)->direction;
									CI=pGroundRTEvent->CellIndex;
									propagateTrajectory(pRTEvent0, pGroundRTEvent, pEstimate);
									goto estimate_ground_incoming_radiation;
								}
								else
								{
									// 3. transmission to ground, ground reflection and scattering between the ground and TOA
									alpha_extinction -= p_ground_scattering;

									tau_interaction = tau_extinction_detector_to_ground;
									tau_interaction -= log ( 1.0 - (p_extinction_from_ground_to_TOA * alpha_extinction ) / p_extinction_after_ground);
									if(tau_interaction<0.0)
									{
										errorIn(__FILE__, __LINE__, __func__);
										return false;
									}
									next_p=retraceTrajectoryUntilExtinction(pGroundRTEvent, tau_interaction);
									goto enforce_scatter_event_after_TOA_exit_with_surface_reflection;
								}
enforce_scatter_event_after_TOA_exit_with_surface_reflection:
								k=pCurrentRTEvent->direction;
								CI=pCurrentRTEvent->CellIndex;
								pOC=pAtmosphere->getOpticalCellPtr(CI, pCurrentRTEvent->wavelength);
								if(!pOC)
								{
									errorIn(__FILE__, __LINE__, __func__);
									return false;
								}
								scatter_coefficient=pOC->scatter_coefficient;
								pEstimate->weight *= scatter_coefficient / pOC->extinction_coefficient;
								propagateTrajectory(pRTEvent0, pCurrentRTEvent, pEstimate);
								goto estimate_scattering_angle;
							}
							else // if(reflection_coefficient)
							{
								//  enforce scattering on interval from detector to TOA
								double p_transmission = exp( -tau_extinction ) ; 	// transmission probability
								double p_extinction = 1.0 - p_transmission ;		// probability that extinction happens

								if(b_consider_direct_light)
									checkDirectLightContribution(pEstimate, p_transmission);

								// sample random number from correct interval
								double alpha_extinction = p_transmission + nzprand() * p_extinction;

								// modification of the photon weight
								pEstimate->weight *= p_extinction;

								// sample extinction point in atmosphere
								tau_interaction = - log ( alpha_extinction );
								next_p=retraceTrajectoryUntilExtinction(pRTEvent0, tau_interaction);
								k=pCurrentRTEvent->direction;
								CI=pCurrentRTEvent->CellIndex;
								pOC=pAtmosphere->getOpticalCellPtr(CI, pCurrentRTEvent->wavelength);
								if(!pOC)
								{
									errorIn(__FILE__, __LINE__, __func__);
									return false;
								}
								scatter_coefficient=pOC->scatter_coefficient;
								pEstimate->weight *= scatter_coefficient / pOC->extinction_coefficient;
								propagateTrajectory(pRTEvent0, pCurrentRTEvent, pEstimate);
								goto estimate_scattering_angle;
							}// if(reflection_coefficient)
						}// if(no_scatter_events)
					}
					else // if(b_automatic_switch_selection)
					{
normal_reached_TOA_procedure:
						found_scatter_event=false;
						if(NextCI.CellPosition>=CELL_POSITION_ABOVE_TOA)
						{
							switch(NextCI.CellPosition)
							{
								case CELL_POSITION_OFF_X_A:
								case CELL_POSITION_OFF_X_B:
									pCurrentRTEvent->type=TE_REFRACTION_X;
									break;

								case CELL_POSITION_OFF_Y_A:
								case CELL_POSITION_OFF_Y_B:
									pCurrentRTEvent->type=TE_REFRACTION_Y;
									break;

								default:
									pCurrentRTEvent->type=TE_REFRACTION_Z;
							}
							if(b_consider_direct_light)
								checkDirectLightContribution(pEstimate, pEstimate->weight);
							continue_tracing=false;
						}
					}// if(b_automatic_switch_selection)
			} //switch(NextCellIndex.CellPosition)
		}
		else // if(fpl_tracing)
		{
			// ######### interaction ############################################
			// where?
			interaction_distance=tau_interaction/extinction_coefficient;
			pCurrentRTEvent->distance=interaction_distance;
			next_p=p+interaction_distance*k;

			// update cell path length
			// cell_path_length[cell_index]+=interaction_distance;
			//if(!b_automatic_switch_selection || (b_automatic_switch_selection && no_scatter_events))
			if(propagate_light_path)
				processPropagation(pEstimate, cell_index, interaction_distance);

			// interaction type
			if(extinction_coefficient*prand()>absorption_coefficient)
			{
estimate_scattering_angle:
				/*if(!scatter_coefficient)
				{
					errorIn(__FILE__, __LINE__ , __func__);
					fprintf(stderr, "scatter coefficient is zero. Check atmosphere configuration files!\n");
					return false;
				}*/
				// scattering occured
				no_interaction=false;
				alpha_scatter_object=prand()*scatter_coefficient;
				pPhaseFunction=0;
				pEstimate->RRS_adjoint_extinction=pOC->extinction_coefficient;
				if((alpha_scatter_object-=pOC->rayleigh_scatter_coefficient)<0.0)
				{
					// Rayleigh scattering
					pPhaseFunction=pMolecularPhF;
					pCurrentRTEvent->type=TE_RAYLEIGH_SCATTERING;
					pEstimate->registerElasticScatterEvent();
				}
				else if((alpha_scatter_object-=pOC->rotational_raman_scatter_coefficient)<0.0)
				{
					// rotational Raman scattering
					pPhaseFunction=pRotationalRamanPhF;
					pCurrentRTEvent->type=TE_ROTATIONAL_RAMAN_SCATTERING;
					lambda=pOC->pRRSEvent->sampleWavelength();
					this->wavelength=lambda;
					pEstimate->registerRRSEvent();
					pEstimate->RRS_adjoint_weight *= pOC->pRRSEvent->getRatioSigmaInSigmaOut();
					// pEstimate->RRS_elastic_weight*=pOC->
				}
				else
				{
					pPOStructPtr=pOC->ParticleOpticalStructPtr;
					int i, n=pOC->no_particles;
					for(i=0;i<n;i++)
					{
						pPOStruct=*pPOStructPtr;
						alpha_scatter_object-=pPOStruct->scatter_coefficient;
						if(alpha_scatter_object<0.0)
						{
							pPhaseFunction=pPOStruct->pPhaseFunction;
							pCurrentRTEvent->type=TE_PARTICLE_SCATTERING+i;
							pEstimate->registerElasticScatterEvent();
							break;
						}
						pPOStructPtr++;
					}
				}
				if(!pPhaseFunction)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "scattering without phase function!\n");
					return false;
				}
				pCurrentRTEvent->pPhaseFunction=pPhaseFunction;
				// scatter object found, sample scatter direction
				last_k=k;
				FunctionalLayer::processAtmosphericScatterEvent(pCurrentRTEvent, pEstimate, next_p, CI, pOC, -last_k);
				/*
					if(update_variance_reduction_parameters)
						VRGrid.registerQuadraturePoint(cell_index, p, -k_out, pCurrentRTEvent->contribution, pEstimate->weight);
				*/
				// update indices and counters
				found_scatter_event=true;
				no_scatter_events++;
				if(b_Neumann_series_truncation && no_scatter_events>=no_scatter_orders)
					continue_tracing=false;
				else
				{
					pPhaseFunction->scatter(&k, &(pCurrentRTEvent->mu), &(pCurrentRTEvent->phi), &PhaseFunctionISWeight);
					pEstimate->weight*=PhaseFunctionISWeight;
					FunctionalLayer::computeScatterEventTransition(pCurrentRTEvent, pEstimate, pOC, -k, -last_k);
					VRGrid.setCurrentPredecessorIndex();
				} // if(b_Neumann_series_truncation && no_scatter_events>=no_scatter_orders)
			}
			else
			{
				// absorption
				pCurrentRTEvent->type=TE_ABSORPTION;
				continue_tracing=false;
				found_scatter_event=false;
				if(pAtmosphere->simulateEmission())
					FunctionalLayer::processAtmosphericAbsorptionEvent(pCurrentRTEvent, pEstimate, pOC);
			}
			no_interaction=false;
			// CellIndex stays as it is...
		} // if (fpl_tracing)

		// update position
		p=next_p;
		pCurrentRTEvent->position=p;
		pCurrentRTEvent->direction=k;
		pCurrentRTEvent->wavelength=lambda;
		pCurrentRTEvent->CellIndex=CI;

		if(b_save_trajectories)
			logEvent(pCurrentRTEvent);
		// ....
		if(found_scatter_event)
		{
		 	//if(b_save_trajectories)
			//	saveTrajectorySnippet(lse_index+1, no_events);
			lse_index=no_events;
			/*if(b_create_sensitivity_plot)
				plotSubTrajectoryOnFunctionalScreen();*/
		}

		pCurrentRTEvent++;
		no_events++;
		if(no_events>=maximum_no_trace_events)
		{
			writeTrajectoryASCII(stderr, RTEvent, no_events);
			fprintf(stderr,"reached maximum number of trace events (%d)\n", maximum_no_trace_events);
			return false;
		}
		if(abort_tracing)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"abort tracing\n");
			return false;
		}
		if(!continue_tracing)
			break;
		if(no_interaction)
			goto find_extinction_point;
		if(found_scatter_event && b_enable_variance_reduction)
		{
			VRGrid.calculateWeightWindow(cell_index, k, ww_enabled, ww_rr, ww_optimum, ww_split);
			weight=pEstimate->weight;

			if(ww_enabled)
			{
				if(!b_generate_variance_reduction_parameter_file && weight<ww_rr)
				{
					double p_rr=weight/ww_optimum;
					if(prand()>p_rr)
						break;	// leads to return true at the end of the function
					// pEstimate->weight=weight=ww_optimum;
					pEstimate->weight=ww_optimum;
				}	
				else if(pEstimate->getRecursionDepth()>0 && weight>ww_split)
				{
					int current_qp_index=VRGrid.getCurrentQuadraturePointIndex();

					// we have a split event
					// create and pursue IS estimate structure ---------------------------------------------------------
					class Estimate SplitEstimate;
					if(!SplitEstimate.initializeFromEstimate(pEstimate))
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "unable to allocate split estimate object\n");
						return false;
					}

					// determine number of branches
					int i, n_IS_branches;
					double weight_ratio=weight/ww_optimum;
					double floor_weight_ratio=floor(weight_ratio);
					double p_inc=weight_ratio-floor_weight_ratio;
					if(prand()<p_inc)
						n_IS_branches=1+int(floor_weight_ratio);
					else
						n_IS_branches=int(floor_weight_ratio);

					// determine split weight
					double split_weight=1.0/double(n_IS_branches);
					SplitEstimate.setRecursionDepth(pEstimate->getRecursionDepth()-1);
					for(i=0;i<n_IS_branches;i++)
					{
						SplitEstimate.reset();
						SplitEstimate.weight = split_weight * weight;
						VRGrid.setPredecessorIndex(current_qp_index);
						if(!generateSISTrajectory(&SplitEstimate, p, k, CI, lambda))
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "error tracing IS trajectory\n");
							return false;
						}
						pEstimate->addEstimate(&SplitEstimate);
					}
					break;
				}
			}
		}
	}// while(1)
	pEstimate->registerCumulativeUpdateStructsInUpdateCellList();
	return true;
}

void Raytracer::checkDirectLightContribution(class Estimate *pEstimate, double weight)
{
	// check sun hit:
	double dx=SunPosition[0]-position[0];
	double dy=SunPosition[1]-position[1];
	double dz=SunPosition[2]-position[2];
	double r=sqrt(dx*dx+dy*dy+dz*dz);
	double mu=(dx*direction[0]+dy*direction[1]+dz*direction[2])/r;
	processSunHit ( pCurrentRTEvent, pEstimate, (mu>0.9999 && sqrt(2.0*(1.0-mu)) < SUN_SOLID_ANGLE_RADIUS ) ? weight : 0 ) ;
}

bool Raytracer::calculateDirectLightRadiance(class Result *pResult)
{
	if(!pGeometry->initializeGeometry())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"calculateDirectLightRadiance(): could not prepare geometry\n");
		return false;
	}
	wavelength=pSimulation->getCenterWavelength();
	if(!pGeometry->hasPointSource())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"calculateDirectLightRadiance(): detector is not a point source\n");
		return false;
	}
	position=pGeometry->getDetectorPosition();
	CellIndex=pGeometry->getDetectorCellIndex();
	SunPosition=pGeometry->getSunPosition();
	class Estimate estimate;
	estimate.enableCalculateCellPathLength(false);
	pResult->scalar_radiance=calculateEscapeOpticalThickness(&estimate, position, CellIndex) ? getF0()*exp(-tau_forced) : 0.0;
	return true;
}

bool Raytracer::run(class GeometryList *pGeometryList, const int& time_frame_index)
{
	class Result result;
	if(pGeometryList)
	{
		pGeometryList->resetMeanTimeAndGeometry();
		if(!run(&result, pGeometryList, time_frame_index))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		pGeometryList->printMeanTimeAndGeometry(stdout, pSimulation->timeInSeconds());
	}
	else
	{
		if(!run(&result, pGeometryList, time_frame_index))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	}
	FunctionalLayer::printResult(&result, stdout);
	if(update_variance_reduction_parameters)
	{
		if(!VRGrid.save(fn_variance_reduction_parameter_file))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error saving variance reduction parameter file \"%s\"\n", fn_variance_reduction_parameter_file);
		}
	}
	return true;
}

bool Raytracer::run(class Result *pResult, class GeometryList *pGeometryList, const int& time_frame_index)
{
	if(running)
		return false;
	if(!pGeometry->initializeGeometry())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not prepare geometry\n");
		return false;
	}
	delta_phf.defineParameter(0, delta_mu);

	if(b_use_external_seed)
		srand(external_seed);
	else
		srand((unsigned int)time(0));

	// control flags
	abort_tracing=false;

	// open streams
	if(!openStreams())
		return false;

	running=true;
	
	// center wavelength things.....
	central_wavelength=pSimulation->getCenterWavelength();
	F0_at_center_wavelength=getF0(central_wavelength);
	wavelength=central_wavelength;

	pSimulation->prepareOpticalMedium();

	// get pointers to air phase functions
	pMolecularPhF=pAtmosphere->getElasticMolecularPhaseFunctionPtr();
	pRotationalRamanPhF=pAtmosphere->getRotationalRamanPhaseFunctionPtr();

	// write all configuration files to log file
	if(!pGeometryList)
	{
		fprintf(log_stream,"# geometry #####################\n");
		pGeometry->write(log_stream);
	}
	fprintf(log_stream,"# raytracer ####################\n");
	RaytracerIO::write(log_stream);
	fprintf(log_stream,"# simulation ###################\n");
	// pSimulation->write(log_stream);
	
	// setup result structure
	if(!resetFunctionals(pResult))	// allocates result
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error resetting functionals!\n");
		return false;
	}
	
	// setup estimate structure
	class Estimate estimate;
	if(!estimate.associateWithResult(pResult))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating function estimate!\n");
		return false;
	}
	estimate.setRecursionDepth(maximum_recursion_depth);	// recursion depth
	estimate.enableAllowPsiSampling(b_importance_sampling_on_root_trajectory);

	// fpos_t fpos_scattered_stream;
	return run(pResult, &estimate, pGeometryList, time_frame_index);
}

bool Raytracer::run(class Result *pResult, class Estimate *pEstimate, class GeometryList *pGeometryList, const int& time_frame_index)
{
	// bool is_point_source=false; 		// always the same start point
	// class Vector3D position0;		// if is point source
	// struct CellIndex_s CellIndex0;	// if is point source
	int scattered_trajectories=0, unscattered_trajectories=0;
	bool continue_trajectory_sampling=true;

	// printf("F0=%lf\n", F0_at_center_wavelength);
	/*if(pGeometry->hasPointSource())// || use_fixed_initial_direction)
	{	
		is_point_source=true;
		position0=pGeometry->getDetectorPosition();
		CellIndex0=pGeometry->getDetectorCellIndex();
		if(b_automatic_switch_selection)
		{
			class Vector3D CentralFOVDirection;
			if(use_fixed_initial_direction)
				CentralFOVDirection=fixed_initial_direction;
			else
				CentralFOVDirection=pGeometry->getDetectorCentralFOVDirection();
			if(!determineForcedScatteringMode(&b_forced_scattering_mode, position0, CentralFOVDirection))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "could not determine forced scattering mode usage\n");
				running=false;
				return false;
			}
			fprintf(log_stream,"b_forced_scattering_mode=%s\n", b_forced_scattering_mode?"yes":"no");
		}
		b_use_twilight_forcing=false;
	}// pGeometry->hasPointSource() ? yes
	else
		resetCellIndex(&CellIndex0);*/
	SunPosition=pGeometry->getSunPosition();
#ifdef USE_GUI
	pPathViewerWidget->setRange(0, no_quadrature_trajectories/PROGRESS_BAR_UPDATE_GRANULARITY);
	pPathViewerWidget->setValue(0);
#endif
	do	// loop over adjoint trajectory samples
	{
		if(pGeometryList)
		{
			pGeometryList -> sampleGeometry ( time_frame_index ) ;
			if(!pGeometry->initializeGeometry())
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			SunPosition=pGeometry->getSunPosition();
		}
		memset(&CellIndex, 0, sizeof(struct CellIndex_s));
#ifdef USE_GUI
		if(scattered_trajectories%PROGRESS_BAR_UPDATE_GRANULARITY)
			pPathViewerWidget->setValue(scattered_trajectories/PROGRESS_BAR_UPDATE_GRANULARITY);
#endif
		/*if(use_fixed_initial_direction)
		{
			direction=fixed_initial_direction;
			position=position0;
			CellIndex=CellIndex0;
		}
		else*/
		while(!pGeometry->sampleInitialState(&position, &CellIndex, &direction))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error producing valid initial position and direction!\n");
		}
		if(update_variance_reduction_parameters)
			VRGrid.setSourceCellIndex(pGrid->calculateCellIndex(CellIndex));

		wavelength=pSimulation->getWavelengthSample();
		pEstimate->reset();
		if(b_automatic_switch_selection || (!b_automatic_switch_selection && b_use_twilight_forcing))
			pEstimate->twilight_photon = pGeometry->isPointBelowTerminator(position);
		else
			pEstimate->twilight_photon = false;
		// cell_path_length.setToZero();
		// resetESOFlags();	

		no_scatter_events=0;

		/*if(b_save_trajectories && !b_ASCII_trajectory_output)
		{
			// printf("\nplace holder for scatter events of size %d\n", int(sizeof(struct RadiativeTransferEvent_s)));
			fgetpos(scattered_stream, &fpos_scattered_stream);
			fwrite(&no_scatter_events, sizeof(int), 1, scattered_stream);
		}*/

		// initialize first RTEvent
		RTEvent[0].type=TE_SOURCE;
		RTEvent[0].position=position;
		RTEvent[0].direction=direction;
		RTEvent[0].wavelength=wavelength;
		RTEvent[0].tau_extinction=0.0;
		RTEvent[0].distance=0.0;
		//RTEvent[0].cell_index=pGrid->calculateCellIndex(CellIndex);
		RTEvent[0].CellIndex=CellIndex;
		/*if(b_forced_scattering_mode && (is_point_source || !b_automatic_switch_selection))
		{
			if(!findFirstScatterEvent(pEstimate, RTEvent+1))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"could not find (enforced) first scatter event!\n");
				running=false;
				return false;
			}
			no_scatter_events=1;
			lse_index=no_events-1;
			if(b_Neumann_series_truncation && no_scatter_orders<=1)
				goto skip_tracing;
		}
		else // normal tracing procedure
		{*/
			if(b_save_trajectories)
				logEvent(RTEvent);
			no_events=1;
			lse_index=0;
			no_scatter_events=0;
		//}
		pCurrentRTEvent=RTEvent+no_events;

		/*if(!is_point_source)
			registerDirectLightContribution(pEstimate, position, CellIndex);*/
		if(!generateSISTrajectory(pEstimate, position, direction, CellIndex, wavelength))
		{
			closeStreams();
			running=false;
			return false;
		}
// skip_tracing:
		// FunctionalLayer::registerEstimates();
		pResult->registerEstimate(pEstimate);
		if(b_create_LEC_histogram)
			registerContributionInLECHistogram(pEstimate->scalar_radiance);
		
		//printf("%d\n",++counter);
		if(no_scatter_events>0)
		{
			/*if(b_save_trajectories && !b_ASCII_trajectory_output)
			{
				// printf("%d scatter event%s recorded sizeof(int)=%d\n", no_scatter_events, no_scatter_events>1?"s":"", sizeof(int));
				fpos_t fpos_scattered_stream_current;
				fgetpos(scattered_stream, &fpos_scattered_stream_current);
				fsetpos(scattered_stream, &fpos_scattered_stream);
				fwrite(&no_scatter_events, sizeof(int), 1, scattered_stream);
				fsetpos(scattered_stream, &fpos_scattered_stream_current);
			}*/
			scattered_trajectories++;
		}
		else // absorption without scattering...
		{
			// int n=no_events;
			/*if(b_save_trajectories)
			{
				if(b_ASCII_trajectory_output)
					writeTrajectoryASCII(unscattered_stream, RTEvent, n);
				else
				{
					fwrite(&n, sizeof(n), 1, unscattered_stream);
					fwrite(RTEvent, sizeof(RadiativeTransferEvent_s), n, unscattered_stream);
				}
			}*/
			unscattered_trajectories++;
		}
		if(update_variance_reduction_parameters)
		{
			if(!(scattered_trajectories%getPrecisionCheckCount()))
			{
				// if(b_enable_directional_importance_sampling)
				// VRGrid.exportDirectionalImportancesToBinaryGridFile((char *) "directional_importances.bgf");
				VRGrid.exportImportancesToBinaryGridFile((char *) "importances.bgf");
				; // VRGrid.writeCellData((char *) "cell_importances.txt");
			}
			VRGrid.registerQuadraturePoints();
		}
		if(usePrecision() && scattered_trajectories)
		{
		 	if(!(scattered_trajectories%getPrecisionCheckCount()))
				continue_trajectory_sampling=!pResult->metAccuracyCriterium();
		}
		else if(scattered_trajectories>=no_quadrature_trajectories)
			continue_trajectory_sampling=false;

	}while(continue_trajectory_sampling);

	/*if(b_save_trajectories)
	{
		if(b_ASCII_trajectory_output)
		{
			fseek(scattered_stream, 0, SEEK_SET);
			fprintf(scattered_stream, "scattered trajectories: %d\n", scattered_trajectories);
			fseek(unscattered_stream, 0, SEEK_SET);
			fprintf(unscattered_stream, "unscattered trajectories: %d\n", unscattered_trajectories);
		}
		else
		{
			fseek(scattered_stream, 0, SEEK_SET);
			fwrite(&scattered_trajectories, sizeof(int), 1, scattered_stream);
			fseek(unscattered_stream, 0, SEEK_SET);
			fwrite(&unscattered_trajectories, sizeof(int), 1, unscattered_stream);
		}
	}*/
	closeStreams();
	// pAtmosphere->writeStatistics();
	running=false;
	pAtmosphere->writeAuxiliaryFiles();
	return true;
}

bool Raytracer::calculateEscapeOpticalThicknessSimple(class Vector3D p, struct CellIndex_s CI, const double& wavelength, double& tau_forced)
{
	int cell_index;
	bool continue_forcing; // , RRS=pAtmosphere->simulateRingEffect();
	double tau=0.0, distance=0.0;
	struct CellIndex_s NextCI;
	struct Atmosphere::OpticalCell_s *pOC;
	class Vector3D next_p, direction=SunDirection;
	do
	{
		continue_forcing=pGrid->calculateCellIntersectionPoint(p, &CI, direction, &distance, &next_p, &NextCI);
		if(!continue_forcing)
			return false;
			
		cell_index=pGrid->calculateCellIndex(CI);
		pOC=pAtmosphere->getOpticalCellPtr(cell_index, wavelength);
		if(!pOC)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not set wavelength of cell\n");
			return false;
		}
		tau += distance * pOC->extinction_coefficient;
		p=next_p;
		CI=NextCI;
		if(NextCI.CellPosition>=CELL_POSITION_ABOVE_TOA)
			continue_forcing=false;
		else if(NextCI.CellPosition==CELL_POSITION_SURFACE)
			return false;
	}while(continue_forcing);
	tau_forced=tau;
	return true;
}

bool Raytracer::runAnalyticalSSA(double *pIntensity)
{
	// basic initialization
	if(!pGeometry->initializeGeometry())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not prepare geometry\n");
		return false;
	}

	if(b_use_external_seed)
		srand(external_seed);
	else
		srand((unsigned int)time(0));

	// center wavelength things.....
	central_wavelength=pSimulation->getCenterWavelength();
	F0_at_center_wavelength=getF0(central_wavelength);
	pAtmosphere->setCenterWavelength(central_wavelength);
	wavelength=central_wavelength;

	// get pointers to air phase functions
	pMolecularPhF=pAtmosphere->getElasticMolecularPhaseFunctionPtr();
	pRotationalRamanPhF=pAtmosphere->getRotationalRamanPhaseFunctionPtr();

	pAtmosphere->resetOpticalCells();

	// test geometry
	if(pGeometry->getFOVType()!=Geometry::round || pGeometry->getOmega()!=0.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "single scattering approximation implemented only for round and delta FOVs!\n");
		return false;
	}

	// method: raytrace until infinity (or ground intersection)
	class Vector3D p, next_p, direction;
	struct CellIndex_s CI, NextCI;
	direction	= pGeometry->getDetectorCentralFOVDirection();
	SunPosition	= pGeometry->getSunPosition();

	int i, j;
	bool continue_forcing; //, ground_hit=false;
	double distance=0.0;

	double step_size_km;
	class Vector3D step_vector, p_source;

	struct Atmosphere::OpticalCell_s *pOC;
	double intensity = 0.0, F, mu;
	double extinction_coefficient;
	double tau = 0.0, delta_tau, half_delta_tau; 
	double EpsilonSP;

	// RRS source terms -------------
	class Matrix RRSSourceTerm;
	bool RRS = pAtmosphere -> simulateRingEffect ( ) ;
	bool calculated_F0=false;
	int no_RRS_lines = 0 ;
	if(RRS)
	{
		no_RRS_lines = 4 * MolecularRRSCrossSection :: getJMax ( ) ;

		// RRSSourceTerm row format:

		// 	1 : RRS line wavelength
		// 	2 : differential scattering coefficient times irradiance to center wavelength
		// 	3 : solar irradiance F0 at RRS line wavelength

		if ( !RRSSourceTerm . allocate ( 3, no_RRS_lines ) )
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	}

	// go !
	CI = pGeometry->getDetectorCellIndex();
	p = pGeometry->getDetectorPosition();
	do
	{
		continue_forcing=pGrid->calculateCellIntersectionPoint(p, &CI, direction, &distance, &next_p, &NextCI);
		if(!continue_forcing)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}

		// get pointer to optical cell
		pOC = pAtmosphere -> getOpticalCellPtr ( pGrid -> calculateCellIndex ( CI ), central_wavelength ) ;
		if ( ! pOC )
		{
			errorIn ( __FILE__, __LINE__, __func__ ) ;
			fprintf ( stderr, "could not set wavelength of cell\n" ) ;
			return false;
		}
		
		// calculate source function
		SunDirection = norm ( SunPosition - p ) ;
		mu = correctMu ( direction * SunDirection ) ;
		if(RRS)
		{
			// elastic contribution first
			calculateElasticDifferentialScatteringCoefficient( pOC, mu, EpsilonSP ) ;
			
			// calculate once solar irradiance at central wavelength
			if( ! calculated_F0 )
			{
				pOC -> pRRSEvent -> getInScatteringWavelengths ( RRSSourceTerm [ 0 ] ) ;
				for ( i = 0 ; i < no_RRS_lines ; i ++ )
					RRSSourceTerm [ 2 ] [ i ] = getF0 ( RRSSourceTerm [ 0 ] [ i ] ); 
				calculated_F0 = true;
			}

			//  RRS differential scattering coefficient
			pOC -> pRRSEvent -> getInScatteringCrossSections ( RRSSourceTerm [ 1 ] ) ;
			double factor = 1e5 * pOC -> air_number_density * pRotationalRamanPhF -> evaluate ( mu ) ; 
			for ( i = 0 ; i < no_RRS_lines ; i ++ )
				RRSSourceTerm [ 1 ] [ i ] *= factor * RRSSourceTerm [ 2 ] [ i ]; 
		}
		else
			calculateDifferentialScatteringCoefficient( pOC, mu, EpsilonSP ) ;

		// ...
		extinction_coefficient = pOC -> extinction_coefficient ;
		
		// determine number of steps in cell
		step_size_km = distance / double(no_quadrature_points_per_cell);
		step_vector = step_size_km * direction;

		delta_tau = step_size_km * extinction_coefficient ;
		half_delta_tau = 0.5 * delta_tau ;
	
		tau += half_delta_tau;
		for ( i = 0 ; i < no_quadrature_points_per_cell ; i ++ )
		{
			p_source = p + ( 0.5 + double ( i ) ) * step_vector;
			// scatter event in the middle of the step size
			if(RRS)
			{
				double incident_scattered_fraction;

				// elastic fraction
				if(!calculateEscapeOpticalThicknessSimple(p_source, CI, central_wavelength, tau_forced))
				{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
				incident_scattered_fraction = EpsilonSP * F0_at_center_wavelength * exp ( - tau_forced ) ; 

				// inelastic / RRS fraction
				for ( j = 0 ; j < no_RRS_lines ; j ++ )
				{
					if(!calculateEscapeOpticalThicknessSimple(p_source, CI, RRSSourceTerm [ 0 ] [ j ], tau_forced))
					{
						errorIn(__FILE__, __LINE__, __func__);
						return false;
					}
					incident_scattered_fraction += RRSSourceTerm [ 1 ] [ j ] * exp ( - tau_forced ) ; 
				}
				// accumulate into intensity integral 
				intensity += step_size_km * incident_scattered_fraction * exp ( - tau );
			}
			else
			{
				if(!calculateEscapeOpticalThicknessSimple(p_source, CI, central_wavelength, tau_forced))
				{
					errorIn(__FILE__, __LINE__, __func__);
					return false;
				}
				F = F0_at_center_wavelength * exp ( - tau_forced ) ; 
				intensity += step_size_km * EpsilonSP * F * exp ( - tau );
			}
			tau += delta_tau ;
		} 
		tau -= half_delta_tau;

		p = next_p;
		CI = NextCI;
		if ( NextCI.CellPosition >= CELL_POSITION_ABOVE_TOA )
			continue_forcing = false;
		else if ( NextCI.CellPosition == CELL_POSITION_SURFACE )
		/*{
			ground_hit = true;*/
			continue_forcing = false;
		//}
	} while ( continue_forcing ) ;
	intensity/=4.0*M_PI;
	// printf ( "distance to %s: %.2lfkm\n", total_distance, ground_hit ? "ground" : "TOA" ) ;
	// printf("%le\n", intensity);
	*pIntensity=intensity;
	return true;
}

/*
bool Raytracer::calculateTerminatorIntersection(const class Vector3D& P, const class Vector3D& k, double *pDistance)
{
	class Vector3D v=P;
	v-=SunPosition;
	double sk=SunPosition*k;
	double sv=SunPosition*v;
	double ss=SunPosition*SunPosition;
	double vk=v*k;
	double vv=v*v;

	// a,b,c
	double a=ss*cT_square-sk*sk;
	double b=ss*cT_square*vk-sv*sk;
	double c=ss*cT_square*vv-sv*sv;

	// Test if there is an intersection with the T within the atmosphere.
	if(a==0.0)
		return false;
	double p=b/a;
	double q=c/a;
	double radicand=p*p-q;
	if(radicand<=0.0)
		return false;
	double lambda=-p+sqrt(radicand);
	if(lambda<0.0)
		return false;
	*pDistance=lambda;
	return true;
}

bool Raytracer::isDirectLightPoint(const class Vector3D& p)
{
	class Vector3D k=norm(SunPosition-p);
	double lambda=k*p;						// attention minus sign went to here v
	class Vector3D TangentPoint=p;
	TangentPoint-=lambda*k;
	return TangentPoint.length()>Re;
}

bool Raytracer::generateTwilightEventEstimate(struct RadiativeTransferEvent_s *pRTEvent, int *pno_events, double *pWeight)
{
	int n=1;
	double tau=0.0, distance, weight, forced_weight;
	struct CellIndex_s CI=pRTEvent->CellIndex, NextCI;
	class Vector3D next_position;
	struct Atmosphere::OpticalCell_s *pOC;
	direction=(pRTEvent-1)->direction;
	position=pRTEvent->position;
	if(isDirectLightPoint(position))
		return false;
	if(!calculateTerminatorIntersection(position, direction, &distance))
		return false;
	position=pRTEvent->position+distance*direction;
	if(!pGrid->contains(position))
		return false;
	// begin raytracing --------------------------------------------------
	struct RadiativeTransferEvent_s *pTwilightRTEvent=TwilightRTEvent;
	*pTwilightRTEvent=*pRTEvent;
	pTwilightRTEvent->direction=direction;

	// trace until the terminator is reached...
	pTwilightRTEvent++;
	position=pRTEvent->position;
	do
	{
		if(!pGrid->calculateCellIntersectionPoint(position,&CI,direction,&distance,&next_position,&NextCI))
			return false;
		pOC=pAtmosphere->getOpticalCellPtr(CI);
		if(isDirectLightPoint(next_position))
			break;
		tau+=distance*pOC->extinction_coefficient;
		pTwilightRTEvent->position=position=next_position;
		pTwilightRTEvent->CellIndex=CI=NextCI;
		pTwilightRTEvent->direction=direction;
		pTwilightRTEvent->distance=distance;
		pTwilightRTEvent->type=TE_REFRACTION;
		pTwilightRTEvent++;
		n++;
		if(NextCI.CellPosition==CELL_POSITION_SURFACE || NextCI.CellPosition>=CELL_POSITION_ABOVE_TOA)
			return false;
		if(b_simulate_refraction)
		{
			class Vector3D RadialVector=norm(position);
			double N_IN=pOC->refractive_index;
			pOC=pAtmosphere->getOpticalCellPtr(NextCI);
			double N_OUT=pOC->refractive_index;
			direction=doRefraction(norm(next_position), direction, N_IN, N_OUT);
		}
	}while(1);
	// next_point is a direct light point
	// but point not!
	if(!calculateTerminatorIntersection(position, direction, &distance))
		return false;
	tau+=distance*pOC->extinction_coefficient;
	weight=exp(-tau);
	position+=distance*direction;
	if(!findFirstScatterEvent(&forced_weight, pTwilightRTEvent))
		return false;
	*pWeight=forced_weight*weight;
	*pno_events=n+no_events;
	return true;
}
*/


/*void Raytracer::writeTrajectory(FILE *fptr, struct RadiativeTransferEvent_s *RTEvent, int n, bool scattered)
{
	double r,theta,phi;
	struct RadiativeTransferEvent_s *pEvent=RTEvent;
	fprintf(fptr, "# %s trajectory with %d events. ", scattered ? "scattered" : "unscattered", n);
	fprintf(fptr, "How to read the following lines:\n");
	fprintf(fptr, "# n\td[km]\ttype\th[km]\tlat[]\tlong[]\tH/Z\tiLat/Y\tiLong/X\n");
	for(int i=0;i<n;i++)
	{
		fprintf(fptr, "%d\t", i+1);
		pEvent->position.toPolar(&r,&theta,&phi);
		fprintf(fptr, "%.4lf\t", pEvent->distance);
		fprintf(fptr, "%d\t%.3lf\t%.4lf\t%.4lf\t", pEvent->type, r-pGrid->getPlanetRadius(), 90.0-Degree(theta), Degree(phi));
		fprintf(fptr, "%d\t%d\t%d\t", pEvent->CellIndex.coordinates.general.iC[2],
			pEvent->CellIndex.coordinates.general.iC[1],
			pEvent->CellIndex.coordinates.general.iC[0]);
		fputc('\n',fptr);
		pEvent++;
	}
}*/

void Raytracer::writeTrajectoryASCII(FILE *fptr, struct RadiativeTransferEvent_s *pRTEvent, int n)
{
	int i;
	for(i=0;i<n;i++)
	{
		writeSingleEventASCII(fptr, pRTEvent);
		pRTEvent++;
	}
	fputc('\n', fptr);
}

/*bool Raytracer::determineForcedScatteringMode(bool *pFSMode, class Vector3D& Position, class Vector3D& Direction)
{
	bool ground_hit=false, continue_tracing;
	double TauAbsorption, TauScattering, distance;
	struct RadiativeTransferEvent_s *RTEventPtr=RTEvent;
	struct CellIndex_s CI, NextCI;	
	class Vector3D position=Position, direction=Direction, next_position;
	struct Atmosphere::OpticalCell_s *pOC;
	resetCellIndex(&CI);
	if(!pGrid->calculateCellIndex(position, &CI))
		return false;
	TauAbsorption=0.0;
	TauScattering=0.0;
	int i=0;
	RTEventPtr->position=Position;
	RTEventPtr->direction=Direction;
	RTEventPtr->wavelength=wavelength;
	RTEventPtr->CellIndex=CellIndex;
	RTEventPtr->distance=0.0;
	RTEventPtr->type=TE_SOURCE;
	RTEventPtr++;
	do
	{
		continue_tracing=pGrid->calculateCellIntersectionPoint(position, &CI, direction, &distance, &next_position, &NextCI);
		if(!continue_tracing)
		{
			writeTrajectory(stderr, RTEvent, i+1);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not determine LOS OT\n");
			return false;
		}
		RTEventPtr->position=next_position;
		RTEventPtr->direction=direction;
		RTEventPtr->wavelength=wavelength;
		RTEventPtr->CellIndex=NextCI;
		RTEventPtr->distance=distance;
		RTEventPtr->type=NextCI.event;
		RTEventPtr++;
		pOC=pAtmosphere->getOpticalCellPtr(CI, wavelength);
		if(!pOC)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not set wavelength of cell\n");
			return false;
		}

		TauAbsorption+=distance*pOC->absorption_coefficient;
		TauScattering+=distance*pOC->scatter_coefficient;

		CI=NextCI;
		position=next_position;
		if(NextCI.CellPosition==CELL_POSITION_SURFACE)
		{
			ground_hit=true;
			continue_tracing=false;
		}
		else if(NextCI.CellPosition==CELL_POSITION_ABOVE_TOA)
		{
			ground_hit=false;
			continue_tracing=false;
		}
		else if(NextCI.CellPosition>CELL_POSITION_ABOVE_TOA)
		{
			fprintf(stderr,"lateral exit!\n");
			return false;
		}
		if(++i>maximum_no_trace_events)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "exceeded maximum number of trace events!\n");
			return false;
		}
	}while(continue_tracing);

	double p_a=1.0-exp(-TauAbsorption);			// probability for absorption
	double p_s=1.0-exp(-TauScattering);			// probability for scattering
	double p_e=1.0-exp(-(TauAbsorption+TauScattering));	// probability for extinction
	*pFSMode=ground_hit?(p_e>0.367879441 && p_a>p_s):(p_e<0.367879441 || p_a>p_s);
	return true;
}


bool Raytracer::findFirstScatterEvent(class Estimate *pEstimate, struct RadiativeTransferEvent_s *pRTEventBase)
{
	//	purpose:	"enforce" first scatter event in case of the forced scattering mode....
	//			a.k.a. find a scatter event along the viewing direction
	//
	//			the function also calculates the weight for this trajectory
	//
	//	return value: true if successfull

	//				Will store number of events of the trajectory in this->no_events.
	//				The last event RTEvent[this->no_events-1] is a correct scatter
	//				event with position and new direction stored in this->direction.
	//
	//	annotations:	RTEventPtr->tau_absorption is used for TauAbsorption
	//			RTEventPtr->tau_scattering is used for TauScattering

	double TauAbsorption=0.0, TauScattering=0.0, TauExtinction=0.0, distance;
	double DeltaTauExtinction, DeltaTauAbsorption, DeltaTauScattering, PhaseFunctionISWeight;
	struct RadiativeTransferEvent_s *RTEventPtr=pRTEventBase;
	struct CellIndex_s NextCellIndex;
	bool continue_tracing, ground_hit=false;
	class Vector3D next_position, last_direction;
	int i, j, k, no_snippets=0, albedo_map_index=0, cell_index;
	struct Atmosphere::OpticalCell_s *pOC;
	no_events=1;
// ###############################################################################
// ######### determine optical thickness and calculate forced light path  ########
// ###############################################################################
	do
	{
		no_snippets++;
		continue_tracing=pGrid->calculateCellIntersectionPoint(position,&CellIndex,direction,&distance,&next_position,&NextCellIndex);
		if(!continue_tracing)
		{
			writeTrajectory(stderr, RTEvent, no_snippets+1);
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "assertion failure!\n");
			return false;
		}
		cell_index=pGrid->calculateCellIndex(CellIndex);
		pOC=pAtmosphere->getOpticalCellPtr(cell_index, wavelength);
		if(!pOC)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not set wavelength of cell\n");
			return false;
		}
		// update cell path length
		TauAbsorption+=DeltaTauAbsorption=distance*(pOC->absorption_coefficient);
		TauScattering+=DeltaTauScattering=distance*(pOC->scatter_coefficient);
		TauExtinction+=DeltaTauExtinction=DeltaTauAbsorption+DeltaTauScattering;
		
		RTEventPtr->distance=distance;
		RTEventPtr->distance_cell_index=cell_index;
		// update this->position, this->direction, this->CellIndex and *RTEventPtr
		position=next_position;
		CellIndex=NextCellIndex;
		// update RTEventPtr
		RTEventPtr->position=position;
		RTEventPtr->direction=direction;
		RTEventPtr->wavelength=wavelength;
		RTEventPtr->CellIndex=CellIndex;
		RTEventPtr->type=NextCellIndex.event;
		RTEventPtr->tau_absorption=DeltaTauAbsorption;
		RTEventPtr->tau_scattering=DeltaTauScattering;
		if(NextCellIndex.CellPosition==CELL_POSITION_SURFACE)
		{
			if(prand()>pSurface->calculateReflectionCoefficient(k))
			{
						// transmission
			ground_hit=true;
			continue_tracing=false;
		}
		else if(NextCellIndex.CellPosition>=CELL_POSITION_ABOVE_TOA)
		{
			ground_hit=false;
			continue_tracing=false;
		}
		//if(continue_tracing && b_simulate_refraction)
		//{
		//	class Vector3D RadialVector=norm(position);
		//	double N_IN=pOC->refractive_index;
		//	pOC=pAtmosphere->getOpticalCellPtr(NextCellIndex);
		//	double N_OUT=pOC->refractive_index;
		//	direction=doRefraction(norm(next_position),direction,N_IN,N_OUT);
		//}
		RTEventPtr++;
		if(no_snippets>=maximum_no_trace_events)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"error searching first scatter event: maximum number of trace events (%d) exceeded!\n", maximum_no_trace_events);
			return false;
		}
	}while(continue_tracing);
// ####################################################################################
// ######### find first scatter event, distinguish between ground hit/TOA hit #########
// ####################################################################################
	// reset RTEventPtr to second event in the chain
	double p_survive_extinction, tau_absorption;
	double tau_scatter_event_location, tau_next, alpha_extinction;
	bool sample_alpha_extinction;
find_sample:
	p_survive_extinction=exp(-TauExtinction);
	RTEventPtr=pRTEventBase;
	sample_alpha_extinction=true;
	if(ground_hit) // hit the ground on the path from the detector through the atmosphere...
	{
		double albedo;
		CellIndex=RTEvent[no_snippets-1].CellIndex;
		albedo=pSurface->getAlbedo(CellIndex, next_position, wavelength);
		if(albedo>0.0)
		{
			// surface is not black... we may consider the ground scatter probability by a weight
			alpha_extinction=prand();
			if(alpha_extinction<p_survive_extinction)
			{
				// update cell_path_length array
				register struct RadiativeTransferEvent_s *pRTEvent=RTEvent+1;
				for(k=1;k<=no_snippets;k++)
				{
					// cell_path_length[pRTEvent->distance_cell_index]=pRTEvent->distance;
					processPropagation(pEstimate, pRTEvent->distance_cell_index, pRTEvent->distance);
					pRTEvent++;
				}
			
				// photon reached the surface: ground scatter event
				pEstimate->registerGroundScatterEvent();
				pEstimate->registerElasticScatterEvent();
				RTEventPtr=RTEvent+no_snippets;
				pCurrentRTEvent=RTEventPtr;
				position=RTEventPtr->position;
				pSurface->sampleLambertianScatterDirection(&RTEventPtr->theta, &RTEventPtr->phi, direction);
			
				pEstimate->weight=albedo;
				FunctionalLayer::processGroundScatterEvent(pEstimate, position, CellIndex, pSurface->getSurfaceNormal());
				
				RTEventPtr->position=position;
				RTEventPtr->direction=direction;
				RTEventPtr->wavelength=wavelength;
				RTEventPtr->type=TE_GROUND_SCATTERING;
				RTEventPtr->CellIndex=CellIndex;
				RTEventPtr->albedo_map_index=albedo_map_index;
				no_events=1+no_snippets;
				// writeTrajectory(stderr, RTEvent, no_events, true);
				return true;
			}
			else // scattered above the surface
			{
				// this should be quite unlikely!!
				sample_alpha_extinction=false;
				pEstimate->weight=1.0;
			}
		}
		else
			pEstimate->weight=1.0-p_survive_extinction;
	} // if ground hit
	else
		pEstimate->weight=1.0-p_survive_extinction;		// basic weight: there is scattering
	if(sample_alpha_extinction)
		alpha_extinction=p_survive_extinction+prand()*(1.0-p_survive_extinction);
	// get tau_scatter_event_location from random number alpha_extinction
	tau_scatter_event_location=-log(alpha_extinction);
	// tau_scatter_event_location now is the (scatter) optical thickness until the event location
	for(i=0;i<no_snippets;i++)
	{
		tau_next=tau_scatter_event_location-RTEventPtr->tau_scattering-RTEventPtr->tau_absorption;
		if(b_save_trajectories)
			logEvent(RTEventPtr);
		if(tau_next<0.0) // found scatter event cell
		{
			CellIndex=(RTEventPtr-1)->CellIndex; // the same cell, of course
			cell_index=pGrid->calculateCellIndex(CellIndex);
			pOC=pAtmosphere->getOpticalCellPtr(cell_index, wavelength);
			distance=tau_scatter_event_location/pOC->extinction_coefficient;
			// calculate position of the first scatter event
			direction=(RTEventPtr-1)->direction;
			position=RTEventPtr->position=(RTEventPtr-1)->position+distance*direction;
			RTEventPtr->CellIndex=CellIndex; // the same cell, of course
			RTEventPtr->distance=distance;

			// update cell_path_length array
			// register struct RadiativeTransferEvent_s *pRTEvent=RTEvent+1;
			//for(k=1;k<2+i;k++)
			//{
			//	cell_path_length[pRTEvent->distance_cell_index]=pRTEvent->distance;
			//	pRTEvent++;
			//}
			processPropagation(pEstimate, cell_index, distance);
			// determine scatter object
			class PhaseFunction *pPhaseFunction=0;
			double alpha_scatter_object=prand()*pOC->scatter_coefficient;
			if((alpha_scatter_object-=pOC->rayleigh_scatter_coefficient)<0.0)
			{
				// Rayleigh scattering
				pPhaseFunction=pMolecularPhF;
				RTEventPtr->type=TE_RAYLEIGH_SCATTERING;
				pEstimate->registerElasticScatterEvent();
			}
			else if((alpha_scatter_object-=pOC->rotational_raman_scatter_coefficient)<0.0)
			{
				// rotational Raman scattering
				pPhaseFunction=pRotationalRamanPhF;
				RTEventPtr->type=TE_ROTATIONAL_RAMAN_SCATTERING;
				wavelength=pOC->pRRSEvent->sampleWavelength();
				pEstimate->registerRRSEvent();
				pEstimate->weight*=pOC->pRRSEvent->getRatioSigmaInSigmaOut();
				pEstimate->RRS_adjoint_extinction=pOC->extinction_coefficient;
			}
			else
			{
				struct Particle::OpticalStruct_s **pPOStructPtr=pOC->ParticleOpticalStructPtr, *pPOStruct;
				int n=pOC->no_particles;
				for(j=0;j<n;j++)
				{
					pPOStruct=*pPOStructPtr;
					alpha_scatter_object-=pPOStruct->scatter_coefficient;
					if(alpha_scatter_object<0.0)
					{
						pPhaseFunction=pPOStruct->pPhaseFunction;
						RTEventPtr->type=TE_PARTICLE_SCATTERING+j;
						pEstimate->registerElasticScatterEvent();
						break;
					}
					pPOStructPtr++;
				}
			}
			if(!pPhaseFunction)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "scattering without phase function!\n");
				return false;
			}
			RTEventPtr->pPhaseFunction=pPhaseFunction;
			// end of selecting phase function

			pCurrentRTEvent=RTEventPtr;
			// scatter object found, sample scatter direction
			last_direction=direction;
			pPhaseFunction->scatter(&direction, &(RTEventPtr->mu), &(RTEventPtr->phi), &PhaseFunctionISWeight);
			//calculate trajectory weight
			pEstimate->weight*=(PhaseFunctionISWeight*pOC->scatter_coefficient)/pOC->extinction_coefficient;
			FunctionalLayer::processAtmosphericScatterEvent(pEstimate, position, CellIndex, pOC, -last_direction);
			FunctionalLayer::computeScatterEventTransition(pEstimate, pOC, -direction, -last_direction);
			
			if(b_save_trajectories)
				logEvent(RTEventPtr);
			if(!direction.isValid())
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			// update no_events counter
			no_events=2+i;
			return true;
		} // if tau_next<0.0
		tau_scatter_event_location=tau_next;
		tau_absorption+=RTEventPtr->tau_absorption;
		processPropagation(pEstimate, RTEventPtr->distance_cell_index, RTEventPtr->distance);
		//direction=RTEventPtr->direction;
		RTEventPtr++;
	}// for(i=0;i<no_snippets
	if(i==no_snippets)
		goto find_sample;
	return true;
}
*/

/*bool Raytracer::traceRayWithRefraction(class Vector3D& Position, struct CellIndex_s& CI, class Vector3D& Direction, bool print_trajectory)
{
	struct CellIndex_s NextCI;
	class Vector3D NextPosition, NGradient, NextNDirection, NDirection;
	double refractive_index, distance;
	while(1)
	{
		// get refractive index and gradient at the old position
		pGrid->getRefractiveIndexAndGradient(Position, CI, &refractive_index, &NGradient);

		// calculate refractive index * direction
		NDirection=refractive_index*Direction;

		// make step...
		NextPosition=Position+refraction_step_size*Direction;
		NextNDirection=NDirection+refraction_step_size*NGradient;
		if(print_trajectory)
		{
			Position.print(stdout, true);
			printf("\t%.12lf\t", refractive_index);
			NGradient.print(stdout, true);
			fputc('\n', stdout);
		}
		// check if the next position is inside the model domain
		if(pGrid->calculateCellIndex(NextPosition, &NextCI))
		{
			CI=NextCI;
			Position=NextPosition;
			Direction=norm(NextNDirection);
		}
		else
		{
			// is it below the ground?
			if(NextCI.CellPosition==CELL_POSITION_SURFACE)
			{
				// if yes: calculate projected ground tunneling exit point
				while(pGrid->calculateCellIntersectionPoint(Position, &CI, Direction, &distance, &NextPosition, &NextCI))
				{
					Position=NextPosition;
					CI=NextCI;
				};
				if(pGrid->isSphericalGrid())
				{
					// calculate point at which surface is intersected after "tunneling" through earth...
					double R=pGrid->getMinimalRadius();
					double rk=Position*Direction;
					double radicand=R*R+rk*rk;
					if(radicand<0.0)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "spherical ground intersection point calculation failed!\n");
						Position.printSpherical(stderr);
						fputc('\n', stderr);
						return false;
					}
					Position+=Direction*(sqrt(radicand)-rk);
				}
				else
					return false;
			}
			// is it above the TOA? 
			else if(NextCI.CellPosition==CELL_POSITION_ABOVE_TOA)
			{
				// if yes: calculate exact exit point
				while(pGrid->calculateCellIntersectionPoint(Position, &CI, Direction, &distance, &NextPosition, &NextCI))
				{
					Position=NextPosition;
					CI=NextCI;
				};
				return true;
			}
		}
	}
	return true;
}

#define	STEP_SIZE		1e-8

bool Raytracer::findRefractiveSunDirection(const class Vector3D& p, class Vector3D *pSunDirection)
{
	// double distance, chi_square=2f, new_chi_square;
	double rd, td, pd;
	struct CellIndex_s CI, CI0;
	class Vector3D Position, Direction, Direction0, SunDirection;
	bool not_found_direction=true;

	// fitting variables
	class Vector3D k_theta, k_phi, delta_direction, center_direction;

	class MatrixMND C, J, Jt, E, gammaE, JtJ;
	class VectorND xi, xip1, Fxmy, JtFxmy, CJtFxmy;
	try
	{
		C=MatrixMND(2, 2);
		E=MatrixMND(2, 2);
		gammaE=MatrixMND(2, 2);
		JtJ=MatrixMND(2, 2);
		J=MatrixMND(3, 2);
		xi=VectorND(2);
		xip1=VectorND(2);
		Fxmy=VectorND(3);
		JtFxmy=VectorND(3);
		CJtFxmy=VectorND(3);
	}
	catch(class MatrixMNDMemoryException& mme)
	{
		mme.print(stderr);
		return false;
	}
	catch(class VectorNDMemoryException& vme)
	{
		vme.print(stderr);
		return false;
	}
	
	// initialize unity matrix
	E.setToUnity();

	// initialize primary direction
	SunPosition=pGeometry->getSunPosition();
	Position=p;
	Direction0=norm(SunPosition-Position);
	if(!pGrid->calculateCellIndex(Position, &CI0))
		return false;
	
	Direction0.toPolar(&rd, &td, &pd);
	xi[0]=td;
	xi[1]=pd;
	int n=0;
	do
	{
		// determine center direction ---------------------------------------------------
		CI=CI0;
		Position=p;
		Direction=Direction0;
		if(!traceRayWithRefraction(Position, CI, Direction))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"iteration %d, ray intersects the surface\n", n);
			return false;
		}
		SunDirection=norm(SunPosition-Position);
		center_direction=Direction;
		delta_direction=Direction-SunDirection;

		if(delta_direction.length()<refraction_fitness)
			not_found_direction=false;
		else
		{
			// determine Jacobians ##########################################################
			Direction0.toPolar(&rd, &td, &pd);

			// determine k theta -------------------------
			CI=CI0;
			Position=p;
			Direction.fromPolar(rd, td + STEP_SIZE, pd);
			if(!traceRayWithRefraction(Position, CI, Direction))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "iteration %d, ray intersects the surface\n", n);
				return false;
			}
			k_theta = ( Direction - center_direction ) / STEP_SIZE;

			// determine k phi ---------------------------
			CI=CI0;
			Position=p;
			Direction.fromPolar(rd, td, pd + STEP_SIZE);
			if(!traceRayWithRefraction(Position, CI, Direction))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "iteration %d, ray intersects the surface\n", n);
				return false;
			}
			k_phi = ( Direction - center_direction ) / STEP_SIZE;

			// Jacobians done .....
			J[0][0]=k_theta.x[0];
			J[1][0]=k_theta.x[1];
			J[2][0]=k_theta.x[2];

			J[0][1]=k_phi.x[0];
			J[1][1]=k_phi.x[1];
			J[2][1]=k_phi.x[2];

			// C matrix ..............
			Jt=J.getTransposed();
			JtJ=Jt*J;
			C=JtJ;
			//gammaE=E*gamma;
			//C=gammaE+JtJ;
			C.invert();

			// F(x) - y
			Fxmy[0]=delta_direction.x[0];
			Fxmy[1]=delta_direction.x[1];
			Fxmy[2]=delta_direction.x[2];

			// ...
			JtFxmy=Jt*Fxmy;
			CJtFxmy=C*JtFxmy;
			xip1=xi-CJtFxmy;
		
			// ...
			Direction0.fromPolar(1.0, xip1[0], xip1[1]);
			xi=xip1;
			n++;
			if(n>refraction_iterations_maximum)
				return false;
		}
	}while(not_found_direction);
	// printf("%d\t", n);
	Position=p;
	//traceRayWithRefraction(Position, CI0, Direction0, true);
	*pSunDirection=Direction0;
	return true;
}

*/








