/**************************************************************************************************************
*
*       This file,
*
*                            /Raytracer.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Raytracer_h
#define _Raytracer_h

#include "RaytracerIO.h"

#include "Grid.h"
#include "FunctionalLayer.h"
#include "VarianceReductionLayer.h"
#include "PhaseFunctions/DeltaPhaseFunction.h"

class GeometryList;

#ifdef USE_GUI
	class PathViewerWidget;
	#define PROGRESS_BAR_UPDATE_GRANULARITY		50
#endif

#include "cell.h"

#define MAXIMUM_NO_FORCED_PATH_RT_EVENTS	1000

class Raytracer: public RaytracerIO, public FunctionalLayer, public VarianceReductionLayer
{

#ifdef USE_GUI
		PathViewerWidget *pPathViewerWidget;
#endif
	// internal tracing structures and variables:
	
	// event chain
	struct RadiativeTransferEvent_s ForcedPathRTEvent[MAXIMUM_NO_FORCED_PATH_RT_EVENTS];
	int no_forced_path_RT_events;

	struct RadiativeTransferEvent_s *RTEvent, *pCurrentRTEvent;	// points to the current RTEvent to be determined
	bool allocateRTEvents(int n);

	int total_no_events, no_events;
	int no_scatter_events, lse_index;

	/*
	// twilight stuff
		struct RadiativeTransferEvent_s *TwilightRTEvent;
	*/

	/*bool calculateTerminatorIntersection(const class Vector3D& P, const class Vector3D& k, double *pDistance);
	bool isDirectLightPoint(const class Vector3D& P);*/
	
	double cT_square, Re;

	bool abort_tracing, running;
	bool generateSISTrajectory(class Estimate *pEstimate, class Vector3D position, class Vector3D direction, CellIndex_s CellIndex, double lambda);

	// forced scattering mode stuff
	// bool determineForcedScatteringMode(bool *pFSMode, class Vector3D& Position, class Vector3D& Direction);
	// bool findFirstScatterEvent(class Estimate *pEstimate, struct RadiativeTransferEvent_s *pRTEventBase);
	void propagateTrajectory(struct RadiativeTransferEvent_s *pRTEventStart, struct RadiativeTransferEvent_s *pRTEventEnd, class Estimate *pEstimate);
	class Vector3D retraceTrajectoryUntilExtinction(struct RadiativeTransferEvent_s *pRTEvent, const double& tau_interaction);
	bool consistenceCheck(struct RadiativeTransferEvent_s *Event, int n);

	void saveTrajectorySnippet(int i1, int i2);

	class DeltaPhaseFunction delta_phf;
	bool calculateEscapeOpticalThicknessSimple(class Vector3D p, struct CellIndex_s CI, const double& wavelength, double& tau_forced);

public:

	Raytracer();
	~Raytracer();

	inline class RaytracerIO *getRaytracerIOPtr(){return (RaytracerIO *)this;}
	inline class FunctionalLayer *getFunctionalLayerIOPtr(){return (FunctionalLayer *)this;}

	//  bool generateTwilightEventEstimate(struct RadiativeTransferEvent_s *pRTEvent, int *pno_events, double *pWeight);
	//inline void enableFixedInitialDirection(bool enable=true){use_fixed_initial_direction=enable;}
	// inline void setFixedInitialDirection(class Vector3D& k){fixed_initial_direction=k;}

	virtual bool initialize();


	bool runAnalyticalSSA(double *pIntensity);
	bool run(class GeometryList *pGeometryList=0, const int& time_frame_index=-1);
	bool run(class Result *pResult, class GeometryList *pGeometryList=0, const int& time_frame_index=-1);
	bool run(class Result *pResult, class Estimate *pEstimate, class GeometryList *pGeometryList=0, const int& time_frame_index=-1);

	void checkDirectLightContribution(class Estimate *pEstimate, double weight);
	bool calculateDirectLightRadiance(class Result *pResult);

	void writeTrajectoryASCII(FILE *fptr, struct RadiativeTransferEvent_s *RTEvent, int n);

	inline struct RadiativeTransferEvent_s *getRTEventPtr(){return RTEvent;}
	// inline struct RadiativeTransferEvent_s *getTwilightRTEventPtr(){return TwilightRTEvent;}

	inline bool *getControlFlagPtr(){return &abort_tracing;}
	inline bool isRunning(){return running;}

	// use with other programs:
#ifdef USE_GUI
	inline void setPathViewerWidget(PathViewerWidget *PathViewerWidgetPtr){pPathViewerWidget=PathViewerWidgetPtr;}
#endif

	/*// refraction
	bool findRefractiveSunDirection(const class Vector3D& p, class Vector3D *pSunDirection);
	bool traceRayWithRefraction(class Vector3D& Position, struct CellIndex_s& CI, class Vector3D& Direction, bool print_trajectory=false);*/

};

#endif //_Raytracer_h


