/**************************************************************************************************************
*
*       This file,
*
*                            /RaytracerIO.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "RaytracerIO.h"
#include "trigonometry.h"

#include "Grid.h"

RaytracerIO::RaytracerIO()
{
	IOName="raytracer";

	b_save_trajectories=false;
	b_simulate_refraction=false;
	b_consider_direct_light=true;
	// b_trace_until_sun=false;
	b_use_maximum_cross_section_method=false;
	mcs_method_trigger=1.0;
	// b_forced_scattering_mode=false;
	b_automatic_switch_selection=true;
	b_use_external_seed=false;
	external_seed=0;
	b_ASCII_trajectory_output=false;
	b_use_twilight_forcing=false;
	b_Neumann_series_truncation=false;
	// twilight_distance=0.0;
	refraction_step_size=0.2334566;
	refraction_fitness=1e-7;
	refraction_iterations_maximum=15;

	no_quadrature_trajectories=10000;
	maximum_no_quadrature_trajectories=1000000;
	no_quadrature_points_per_cell=10;
	maximum_no_trace_events=10000;

	no_quadrature_points_per_cell=5;
	strcpy(dir_output_directory, "tmp/");

	addParameter(dir_output_directory, type_directory, "dir_output_directory|dir_output", 0, "output directory");
	addParameter(output_filename_prefix, type_string, "output_filename_prefix", 0, "output filename prefix");
	addParameter(&b_use_maximum_cross_section_method, type_bool, "b_use_maximum_cross_section_method|b_MCSM", 0, "indicates whether to use the maximum crosssection method");
	addParameter(&mcs_method_trigger, type_double, "mcs_method_trigger", 0, "defines the maximal vertical optical thickness for the mcs method");
	addParameter(&b_save_trajectories, type_bool, "b_save_trajectories|b_st", 0, "if enabled the raytracer will write the trajectory data to a file");
	addParameter(&b_simulate_refraction, type_bool, "b_simulate_refraction|b_refraction|b_sr", 0, "consider refraction?");
	addParameter(&refraction_step_size, type_double, "refraction_step_size", 0, "[km] step size for the simulation of refraction");
	addParameter(&refraction_fitness, type_double, "refraction_fitness", 0, "[rad] maximal direction difference for the ray exit direction, sun direction");
	addParameter(&refraction_iterations_maximum, type_int, "refraction_iterations_maximum", 0, "maximal number of iteration steps in the refraction fitting procedure");
	addParameter(&b_consider_direct_light, type_bool, "b_consider_direct_light|b_direct_light|b_dl", 0, "consider direct (unscattered) sunlight?");
	// addParameter(&b_trace_until_sun, type_bool, "b_trace_until_sun|b_tus", 0, "trace trajectories from each scatter event to Sun?");
	addParameter(&no_quadrature_trajectories, type_int, "no_quadrature_trajectories", 0, "number of simulated trajectories");
	addParameter(&maximum_no_quadrature_trajectories, type_int, "maximum_no_quadrature_trajectories", 0, "maximum number of simulated trajectories");
	addParameter(&no_quadrature_points_per_cell, type_int, "no_quadrature_points_per_cell", 0, "number of quadrature points per cell");
	addParameter(&maximum_no_trace_events, type_int, "maximum_no_trace_events|n_max", 0, "maximum allowed number of traces events");
	// addParameter(&b_forced_scattering_mode, type_bool, "b_forced_scattering_mode|b_fsm", 0, "enables direct sampling of the first scatter event");
	 addParameter(&b_use_twilight_forcing, type_bool, "b_use_twilight_forcing|b_utf", 0, "indicates e.g. a high terminator");
	// addParameter(&twilight_distance, type_double, "[km] twilight_distance", 0, "parameter for twilight forcing");
	addParameter(&b_automatic_switch_selection, type_bool, "b_automatic_switch_selection|b_ass", 0, "automatic setting of some flags");
	addParameter(&b_Neumann_series_truncation, type_bool, "b_Neumann_series_truncation|b_Nst", 0, "limit number of scatter orders?");
	addParameter(&no_scatter_orders, type_int, "no_scatter_orders", 0, "maximum number of scatter orders");
	addParameter(&b_use_external_seed, type_bool, "b_use_external_seed|b_ues", 0, "is external seed used?");
	addParameter(&external_seed, type_int, "seed|external_seed", 0, "random number generator seed");
	addParameter(&b_ASCII_trajectory_output, type_bool, "b_ASCII_trajectory_output|b_ASCII", 0, "shall the trajectory output be human readable?");

	// streams
	log_stream=(FILE *)0;
	scattered_stream=(FILE *)0;
	unscattered_stream=(FILE *)0;
}

RaytracerIO::~RaytracerIO()
{
	closeStreams();
}

void RaytracerIO::formatScatteredTrajectoriesFilename(char *trajectories_filename)
{
	sprintf(trajectories_filename,"%s%s%s_scattered.dat", dir_base, dir_output_directory, output_filename_prefix);
}

void RaytracerIO::formatUnscatteredTrajectoriesFilename(char *trajectories_filename)
{
	sprintf(trajectories_filename,"%s%s%s_unscattered.dat", dir_base, dir_output_directory, output_filename_prefix);
}

void RaytracerIO::formatLogFilename(char *log_filename)
{
	sprintf(log_filename,"%s%s%s_log.txt", dir_base, dir_output_directory, output_filename_prefix);
}

bool RaytracerIO::openStreams()
{
	closeStreams();
	char filename[2048];
	if(b_save_trajectories)
	{
		formatScatteredTrajectoriesFilename(filename);
		scattered_stream=fopen(filename, b_ASCII_trajectory_output ? "w" : "wb");
		if(!scattered_stream)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not open scattered trajectories stream \"%s\".\n", filename);
			fprintf(stderr,"Check output directory \"%s\" and filename prefix \"%s\"!\n", dir_output_directory, output_filename_prefix);
			return false;
		}
		/* if(!b_ASCII_trajectory_output)
			fwrite(&zero, 1, sizeof(zero), scattered_stream); */
		formatUnscatteredTrajectoriesFilename(filename);
		unscattered_stream=fopen(filename, b_ASCII_trajectory_output ? "w" : "wb");
		if(!unscattered_stream)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not open unscattered trajectories stream \"%s\".\n", filename);
			fprintf(stderr, "Check output directory \"%s\" and filename prefix \"%s\"!\n", dir_output_directory, output_filename_prefix);
			return false;
		}
		/* if(!b_ASCII_trajectory_output)
			fwrite(&zero, 1, sizeof(zero), unscattered_stream); */
	}
	formatLogFilename(filename);
	log_stream=fopen(filename,"w");
	if(!log_stream)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open log file \"%s\".\n", filename);
		fprintf(stderr,"Check output directory \"%s\" and filename prefix \"%s\"!\n", dir_output_directory, output_filename_prefix);
		return false;
	}
	return true;
}

void RaytracerIO::writeSingleEventASCII(FILE *fptr, struct RadiativeTransferEvent_s *pRTEvent)
{
	const char *event_type;
	switch(pRTEvent->type)
	{
		default:
		case TE_SOURCE:
			event_type="SOURCE";
			break;
		case TE_REFRACTION_X:
			event_type="REFRACTION_X";
			break;
		case TE_REFRACTION_Y:
			event_type="REFRACTION_Y";
			break;
		case TE_REFRACTION_Z:
			event_type="REFRACTION_Z";
			break;
		case TE_ABSORPTION:
			event_type="ABSORPTION";
			break;
		case TE_GROUND_REFLECTION:
			event_type="GROUND_REFLECTION";
			break;
		case TE_GROUND_SCATTERING:
			event_type="GROUND_SCATTERING";
			break;
		case TE_RAYLEIGH_SCATTERING:
			event_type="RAYLEIGH_SCATTERING";
			break;
		case TE_ROTATIONAL_RAMAN_SCATTERING:
			event_type="ROTATIONAL_RAMAN_SCATTERING";
			break;
		case TE_PARTICLE_SCATTERING:
			event_type="PARTICLE_SCATTERING";
			break;
	}
	fprintf(fptr, "%s\t", event_type);
	double r, theta, phi;
	if(pGrid->isSphericalGrid())
	{
		pRTEvent->position.toPolar(&r, &theta, &phi);
		fprintf(fptr, "%.2lf\t%.2lf\t%.2lf\t", r - pGrid->getPlanetRadius(), Degree(theta), Degree(phi));
	}
	else
		fprintf(fptr, "%.2lf\t%.2lf\t%.2lf\t", pRTEvent->position[0], pRTEvent->position[1], pRTEvent->position[2]);
	if(pGrid->is3DGrid())
		fprintf(fptr, "%d\t%d\t%d\t", \
			pRTEvent->CellIndex.coordinates.general.iC[0], \
			pRTEvent->CellIndex.coordinates.general.iC[1], \
			pRTEvent->CellIndex.coordinates.general.iC[2]);
	else
		fprintf(fptr, "%d\t", pRTEvent->CellIndex.coordinates.general.iC[2]);
	if(pGrid->isSphericalGrid())
	{
		pRTEvent->direction.toPolar(&r, &theta, &phi);
		fprintf(fptr, "%.2lf\t%.2lf\t%.2lf\t", r, Degree(theta), Degree(phi));
	}
	else
		fprintf(fptr, "%.2le\t%.2le\t%.2le\t", pRTEvent->direction[0], pRTEvent->direction[1], pRTEvent->direction[2]);
	fprintf(fptr, "%.2lf\n", 1e9*pRTEvent->wavelength);
}

void RaytracerIO::logEvent(struct RadiativeTransferEvent_s *pRTEvent)
{
	/*// check only non refraction events
	if(pRTEvent->type>TE_ABSORPTION)
	{
		struct CellIndex_s CI;
		resetCellIndex(&CI);
		// if(!pGrid->calculateCellIndex(pRTEvent->position, pRTEvent->direction, &CI))
		if(!pGrid->calculateCellIndex(pRTEvent->position, &CI))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"event chain point outside the grid\n");
		}
		bool inconsistency=false; 
		if(pGrid->is3DGrid())
		{
			if(CI.coordinates.general.iC[0]!=pRTEvent->CellIndex.coordinates.general.iC[0] \
				|| CI.coordinates.general.iC[1]!=pRTEvent->CellIndex.coordinates.general.iC[1] \
				|| CI.coordinates.general.iC[2]!=pRTEvent->CellIndex.coordinates.general.iC[2] )
			inconsistency=true;
		}
		else
		{
			if(CI.coordinates.general.iC[2]!=pRTEvent->CellIndex.coordinates.general.iC[2])
				inconsistency=true;
		}
		if(inconsistency)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "point - cell index inconsistency\n");
		}
	}*/
	if(b_ASCII_trajectory_output)
		writeSingleEventASCII(scattered_stream, pRTEvent);
	else
		fwrite(pRTEvent, sizeof(RadiativeTransferEvent_s), 1, scattered_stream);
}

void RaytracerIO::closeStreams()
{
	if(log_stream)
	{
		fclose(log_stream);
		log_stream=(FILE *)0;
	}
	if(scattered_stream)
	{
		fclose(scattered_stream);
		scattered_stream=(FILE *)0;
	}
	if(unscattered_stream)
	{
		fclose(unscattered_stream);
		unscattered_stream=(FILE *)0;
	}
}

// END OF STREAMS   ################################################################################################################ 


