/**************************************************************************************************************
*
*       This file,
*
*                            /Result.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Result.h"
#include "Simulation.h"
#include "Grid.h"

#include <float.h>

Result::Result()
{
	GasResult=0;
	ParticleResult=0;
	ScalarHessian=0;
	AllStochasticVariables=0;

	no_stochastic_variables=0;
			
	ScalarHessianMatrixElements=0;
	ScalarJacobianElements=0;

	calculate_cell_path_length=false;
	calculate_cell_sensitivity=false;
	calculate_Jacobians=false;
	calculate_Hessian=false;
	output_polarized_functionals=false;;
		
	restrictive_precision_check=false;
	precision_threshold=0.0;

	output_enabled=true;
	normalize_with_F0=false;
	enable_event_counts=false;

	suppress_Hessian_output=false;
	suppress_Jacobian_output=false;
}

Result::~Result()
{
	if(GasResult)
		delete [] GasResult;

	if(ParticleResult)
		delete [] ParticleResult;
	
	if(ScalarHessian)
		delete [] ScalarHessian;

	if(AllStochasticVariables)
		delete [] AllStochasticVariables;
}

bool Result::allocate(class Atmosphere *pAtmosphere, bool allocate_Hessian)
{
	/*
		initialisation procedure:
			
		allocate active gases and particles
			
	*/
	// evaluate constituents_list
	class Gas *gas=pAtmosphere->getGasPtr();
	class Particle *particle=pAtmosphere->getParticlePtr();
	int i, cluster_size, no_phase_function_parameters, no_parameters;
	int no_gases=pAtmosphere->getNOGases();
	int no_particles=pAtmosphere->getNOParticles();
	
	// cell sensitivity
	if(calculate_cell_sensitivity && !CellSensitivity.allocate(pGrid->getNOCells()))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	
	// count active gases and particles and their parameters
	no_active_gases=0;
	no_active_particles=0;
	N_Parameters=0;

	for(i=0;i<no_gases;i++)
		if(gas[i].isFunctionalCalculationEnabled())
		{
			N_Parameters+=gas[i].getNOClusters();
			no_active_gases++;
		}
	for(i=0;i<no_particles;i++)
		if(particle[i].isFunctionalCalculationEnabled())
		{
			no_phase_function_parameters=particle[i].getNOPhaseFunctionParameters();
			if(particle[i].hasSinglePhaseFunction())
				no_parameters=particle[i].getNOClusters()+1+no_phase_function_parameters;
			else
				no_parameters=particle[i].getNOClusters()*(2+no_phase_function_parameters);
			N_Parameters+=no_parameters;
			no_active_particles++;
		}

	calculate_Jacobians=N_Parameters>0;

	// allocate structs
	if(GasResult)
		delete [] GasResult;
	GasResult=0;
	if(no_active_gases)
	{
		GasResult=new struct GasResult_s[no_active_gases];
		if(!GasResult)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	}

	if(ParticleResult)
		delete [] ParticleResult;
	ParticleResult=0;
	if(no_active_particles)
	{
		ParticleResult=new struct ParticleResult_s[no_active_particles];
		if(!ParticleResult)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	}
	if(AllStochasticVariables)
		delete [] AllStochasticVariables;
	AllStochasticVariables=0;
	if(calculate_Jacobians)
	{
		// allocate Jacobians and Hessians
		if(allocate_Hessian)
			no_stochastic_variables=N_Parameters*(1+N_Parameters);
		else
			no_stochastic_variables=N_Parameters;
		AllStochasticVariables=new StochasticVariable<double>[no_stochastic_variables];
		if(!AllStochasticVariables)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		ScalarJacobianElements=AllStochasticVariables;
		// adjust pointers
		int sv_index=0, index=0, no_clusters;
		if(no_active_gases)
			for(i=0;i<no_gases;i++)
				if(gas[i].isFunctionalCalculationEnabled())
				{
					no_clusters=gas[i].getNOClusters();
					GasResult[index].pGas=gas+i;
					GasResult[index].no_clusters=no_clusters;
					GasResult[index].Jacobian=AllStochasticVariables+sv_index;
					gas[i].setResultStructIndex(index);
					sv_index+=no_clusters;
					index++;
				}

		index=0;
		/*

			particle Jacobian order:
			
			single phase function
				eps1,eps2,...,epsN, ssa, p1,p2,...,pM

			several phase functions
				eps1,ssa1,p11,p12,...,p1M, eps2,ssa2,p21,pr22,...,p2M, .... , epsN,ssaN,pN1,pN2,...,pNM

		*/
		if(no_active_particles)
			for(i=0;i<no_particles;i++)
				if(particle[i].isFunctionalCalculationEnabled())
				{
					no_clusters=particle[i].getNOClusters();
					no_phase_function_parameters=particle[i].getNOPhaseFunctionParameters();
					ParticleResult[index].pParticle=particle+i;
					ParticleResult[index].no_phase_function_parameters=no_phase_function_parameters;
					ParticleResult[index].no_clusters=no_clusters;
					ParticleResult[index].Jacobian=AllStochasticVariables+sv_index;
					if(particle[i].hasSinglePhaseFunction())
					{
						ParticleResult[index].SSAAndPJacobian=ParticleResult[index].Jacobian+no_clusters;
						ParticleResult[index].cluster_size=1;
						no_parameters=no_clusters+1+no_phase_function_parameters;	
					}
					else
					{				
						ParticleResult[index].SSAAndPJacobian=0;
						cluster_size=2+no_phase_function_parameters;
						ParticleResult[index].cluster_size=cluster_size;
						no_parameters=no_clusters*cluster_size;
					}
					ParticleResult[index].no_parameters=no_parameters;
					sv_index+=no_parameters;
					particle[i].setResultStructIndex(index);
					index++;
				}
		if(ScalarHessian)
			delete [] ScalarHessian;
		ScalarHessian=0;
		if(allocate_Hessian)
		{
			ScalarHessian=new StochasticVariable<double>*[N_Parameters];
			if(!ScalarHessian)
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			ScalarHessianMatrixElements=AllStochasticVariables+sv_index;
			for(i=0;i<N_Parameters;i++)
				ScalarHessian[i]=ScalarHessianMatrixElements+i*N_Parameters;
		}
		else
			ScalarHessianMatrixElements=0;
	}
	else
		no_stochastic_variables=0;

	calculate_Hessian=allocate_Hessian;
	return true;
}

void Result::reset()
{
	direct_light_radiance.reset();

	scalar_radiance.reset();
	elastic_scalar_radiance.reset();

	vector_radiance.reset();
	elastic_vector_radiance.reset();
	
	if(enable_event_counts)
	{
		scalar_RRS_count.reset();
		scalar_multiple_RRS_count.reset();
		scalar_elastic_scatter_count.reset();
		scalar_ground_scatter_count.reset();
	}

	int i;
	// Jacobians and Hessians
	if(calculate_cell_sensitivity)
		CellSensitivity.setToZero();
	class StochasticVariable<double> *pStochasticVariable=AllStochasticVariables;
	for(i=0;i<no_stochastic_variables;i++)
		pStochasticVariable++->reset();
}

void Result::registerEstimate(class Estimate *pEstimate)
{
	// radiances
	direct_light_radiance		.registerEstimate(pEstimate->direct_light_radiance);

	scalar_radiance			.registerEstimate(pEstimate->scalar_radiance);
	elastic_scalar_radiance		.registerEstimate(pEstimate->elastic_scalar_radiance);

	vector_radiance			.registerEstimate(pEstimate->vector_radiance);
	elastic_vector_radiance		.registerEstimate(pEstimate->elastic_vector_radiance);

	if(enable_event_counts)
	{
		scalar_RRS_count		.registerEstimate(pEstimate->scalar_RRS_count);
		scalar_multiple_RRS_count	.registerEstimate(pEstimate->scalar_multiple_RRS_count);
		scalar_elastic_scatter_count	.registerEstimate(pEstimate->scalar_elastic_scatter_count);
		scalar_ground_scatter_count	.registerEstimate(pEstimate->scalar_ground_scatter_count);
	}

	if(calculate_cell_sensitivity)
		pEstimate->addCellSensitivitiesTo(CellSensitivity.getCoefficientPtr());
	// Jacobians and Hessian elements
	if(calculate_Jacobians)
	{
		int i, j, m, n;
		class StochasticVariable<double> *pStochasticVariable;
		double *pDouble;
		for(i=0;i<no_active_gases;i++)
		{
			n=GasResult[i].no_clusters;
			pStochasticVariable=GasResult[i].Jacobian;
			pDouble=pEstimate->GasProfileJacobian[i].Jacobian.getCoefficientPtr();
			for(j=0;j<n;j++)
				pStochasticVariable++ ->registerEstimate(*pDouble++ );
		}
		class Particle *pParticle;
		for(i=0;i<no_active_particles;i++)
		{
			pParticle=ParticleResult[i].pParticle;
			m=ParticleResult[i].no_parameters;
			
			pStochasticVariable=ParticleResult[i].Jacobian;
			if(pParticle->hasSinglePhaseFunction())
				pDouble=pEstimate->ParticleProfileJacobian[i].SPF_Jacobian.getCoefficientPtr();
			else
				pDouble=pEstimate->ParticleProfileJacobian[i].Jacobian.getCoefficientPtr();

			for(j=0;j<m;j++)
				pStochasticVariable++ ->registerEstimate(*pDouble++ );
		}
		if(calculate_Hessian)
		{
			pStochasticVariable=ScalarHessianMatrixElements;
			pDouble=pEstimate->Hessian.getCoefficientPtr();
			for(i=0;i<N_Parameters;i++)
				for(j=0;j<N_Parameters;j++)
					pStochasticVariable++->registerEstimate(*pDouble++);
		}
	}
}

bool Result::metAccuracyCriterium()
{
	int i, k;
	double precision;
	class StokesVector vp;
	class StochasticVariable<double> *pStochasticVariable=AllStochasticVariables;
	if(restrictive_precision_check)	// very restrictive!
	{
		double max_precision=DBL_MIN;

		precision=scalar_radiance.getPrecision();
		if(precision>max_precision)
			max_precision=precision;

		precision=direct_light_radiance.getPrecision();
		if(precision>max_precision)
			max_precision=precision;
	
		vp=vector_radiance.getPrecision();
		for(k=0;k<4;k++)
			if(vp[k]>max_precision)
				max_precision=vp[k];

		for(i=0;i<no_stochastic_variables;i++)
		{
			precision=pStochasticVariable++->getPrecision();
			if(precision>max_precision)
				max_precision=precision;
		}

		precision=max_precision;
	}
	else
	{
		double min_precision=DBL_MAX;

		precision=scalar_radiance.getPrecision();
		if(precision<min_precision)
			min_precision=precision;

		precision=direct_light_radiance.getPrecision();
		if(precision<min_precision)
			min_precision=precision;
		
		vp=vector_radiance.getPrecision();
		for(k=0;k<4;k++)
			if(vp[k]<min_precision)
				min_precision=vp[k];

		for(i=0;i<no_stochastic_variables;i++)
		{
			precision=pStochasticVariable++->getPrecision();
			if(precision<min_precision)
				min_precision=precision;
		}

		precision=min_precision;
	}
	return precision<precision_threshold;
}

void Result::finish()
{
	double f=1.0;
	if(normalize_with_F0)
		f=1.0/F0;

	direct_light_radiance.finish();
	direct_light_radiance*=f;

	scalar_radiance.finish();
	scalar_radiance*=f;

	elastic_scalar_radiance.finish();
	elastic_scalar_radiance*=f;

	vector_radiance.finish();
	vector_radiance*=f;

	elastic_vector_radiance.finish();
	elastic_vector_radiance*=f;

	if(enable_event_counts)
	{
		double d=scalar_radiance.getMean()/f;
		scalar_RRS_count.finish();
		scalar_RRS_count/=d;

		scalar_multiple_RRS_count.finish();
		scalar_multiple_RRS_count/=d;

		scalar_elastic_scatter_count.finish();
		scalar_elastic_scatter_count/=d;

		scalar_ground_scatter_count.finish();
		scalar_ground_scatter_count/=d;
	}

	int i;
	class StochasticVariable<double> *pStochasticVariable=AllStochasticVariables;
	for(i=0;i<no_stochastic_variables;i++)
	{
		pStochasticVariable->finish();
		*pStochasticVariable++*=f;
	}
}

bool Result::fillScalarJAndH(class Vector& J, class Matrix& H)
{
	if(!N_Parameters)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "no parameters specified!\n");
		return false;
	}
	if(!J.allocate(N_Parameters))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "J allocation error!\n");
		return false;
	}
	int i, n=N_Parameters;
	class StochasticVariable<double> *pScalarJacobian=AllStochasticVariables;
	double *c=J.getCoefficientPtr();
	for(i=0;i<n;i++)
		*c++=pScalarJacobian++->getMean();
	if(calculate_Hessian)
	{
		if(!H.allocate(N_Parameters, N_Parameters))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "H allocation error!\n");
			return false;
		}
		n*=N_Parameters;
		c=H.getCoefficientPtr();
		for(i=0;i<n;i++)
			*c++=pScalarJacobian++->getMean();
	}
	return true;
}

void Result::print(FILE *fptr, const char *output_format)
{
	char output_format_string[20];
	sprintf(output_format_string, "%s\t%s%%c", output_format, output_format);
	
	class StokesVector vLm, vLe;
	double I, eI, pI;
	int i, n;

	// radiances
	scalar_radiance.finish();

	I=scalar_radiance.getMean();
	eI=scalar_radiance.getStddev();
	pI=eI/I;
	fprintf(fptr, output_format_string, I, eI, '\t');
	fprintf(fptr, "%.5lf\t", pI);

	if(enable_RRS)
	{
		elastic_scalar_radiance.finish();
		fprintf(fptr, output_format_string, elastic_scalar_radiance.getMean(), elastic_scalar_radiance.getStddev(), '\t');
	}
	if(enable_event_counts)
	{
		fprintf(fptr, output_format_string, scalar_elastic_scatter_count.getMean(), scalar_elastic_scatter_count.getStddev(), '\t');
		fprintf(fptr, output_format_string, scalar_ground_scatter_count.getMean(), scalar_ground_scatter_count.getStddev(), '\t');
		if(enable_RRS)
		{
			fprintf(fptr, output_format_string, scalar_RRS_count.getMean(), scalar_RRS_count.getStddev(), '\t');
			fprintf(fptr, output_format_string, scalar_multiple_RRS_count.getMean(), scalar_multiple_RRS_count.getStddev(), '\t');
		}
	}

	direct_light_radiance.finish();
	// fprintf(fptr, output_format_string, direct_light_radiance.getMean(), direct_light_radiance.getStddev(), '\t');
	
	vector_radiance.finish();

	if(output_polarized_functionals)
	{
		vLm=vector_radiance.getMean();
		vLe=vector_radiance.getStddev();
		fprintf(fptr, output_format_string, vLm[0], vLe[0], '\t');
		fprintf(fptr, output_format_string, vLm[1], vLe[1], '\t');
		fprintf(fptr, output_format_string, vLm[2], vLe[2], '\t');
		fprintf(fptr, output_format_string, vLm[3], vLe[3], '\t');
		if(enable_RRS)
		{
			elastic_vector_radiance.finish();
			vLm=elastic_vector_radiance.getMean();
			vLe=elastic_vector_radiance.getStddev();
			fprintf(fptr, output_format_string, vLm[0], vLe[0], '\t');
			fprintf(fptr, output_format_string, vLm[1], vLe[1], '\t');
			fprintf(fptr, output_format_string, vLm[2], vLe[2], '\t');
			fprintf(fptr, output_format_string, vLm[3], vLe[3], '\t');
		}
	}

	// Jacobians and Hessian
	class StochasticVariable<double> *pStochasticVariable;
	if(calculate_Jacobians && !suppress_Jacobian_output)
	{
		pStochasticVariable=ScalarJacobianElements;
		n=N_Parameters;
		for(i=0;i<n;i++)
		{
			fprintf(fptr, output_format_string, pStochasticVariable->getMean(), pStochasticVariable->getStddev(), '\t');
			pStochasticVariable++;
		}
	}
	if(calculate_Hessian && !suppress_Hessian_output)
	{
		pStochasticVariable=ScalarHessianMatrixElements;
		n=N_Parameters*N_Parameters;
		for(i=0;i<n;i++)
		{
			fprintf(fptr, output_format_string, pStochasticVariable->getMean(), pStochasticVariable->getStddev(), '\t');
			pStochasticVariable++;
		}
	}
}

struct Result::GasResult_s *Result::findGasResultPtr(const char *name)
{
	int i;
	for(i=0;i<no_active_gases;i++)
		if(!strcmp(GasResult[i].pGas->getName(), name))
			return GasResult+i;
	return 0;
}

struct Result::ParticleResult_s *Result::findParticleResultPtr(const char *name)
{
	int i;
	for(i=0;i<no_active_particles;i++)
		if(!strcmp(ParticleResult[i].pParticle->getName(), name))
			return ParticleResult+i;
	return 0;
}

void Result::printPathIntegralResults(const char *name, class list<string>& constituents_list, FILE *fptr)
{
	int i, j, index_offset, PI_index_offset, no_derivatives, PI_index, index;
	char output_format_string[20];
	double x, x_error, y, y_error, concentration, I, II, I_error;
	class GasResult_s *pGasResult, *pPIGasResult;
	class ParticleResult_s *pParticleResult;

	// prepare output format
	sprintf(output_format_string, "%%le\t%%le%%c");

	// prepare radiance
	I=scalar_radiance.getMean();
	II=I*I;
	I_error=scalar_radiance.getStddev();

	// calculate and output path integral -------------------------
	// find gas in result list
	pPIGasResult=findGasResultPtr(name);
	if(!pPIGasResult)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "gas \"%s\" not in the result list\n", name);
		return;
	}
	PI_index_offset=pPIGasResult->pGas->getJacobianOffset();

	// calculate path integral
	x=x_error=0.0;
	for(i=0;i<pPIGasResult->no_clusters;i++)
	{
		concentration=pPIGasResult->pGas->getConcentration(i);
		x+=concentration*pPIGasResult->Jacobian[i].getMean();
		x_error+=concentration*pPIGasResult->Jacobian[i].getStddev();
	}
	x_error=1e5*(I*x_error+x*I_error)/II;
	x*=-1e5/I;

	// output path integral
	fprintf(fptr, output_format_string, x, x_error, '\t');

	// calculate and output path integral derivatives
	list<string>::iterator it;
	for(it=constituents_list.begin();it!=constituents_list.end();it++)
	{
		// determine index_offset in Hessian
		pGasResult=findGasResultPtr(it->c_str());
		if(!pGasResult)
		{
			// may be a particle?
			pParticleResult=findParticleResultPtr(it->c_str());
			if(!pParticleResult)
			{
				errorIn(__FILE__, __LINE__, __func__);
				return;
			}
			//------------------------------------
			index_offset=pParticleResult->pParticle->getJacobianOffset();
			no_derivatives=pParticleResult->no_parameters;
		}
		else
		{
			index_offset=pGasResult->pGas->getJacobianOffset();
			no_derivatives=pGasResult->no_clusters;
		}
		for(j=0;j<no_derivatives;j++)
		{
			x=0.0;
			x_error=0.0;
			// loop over path integrals gas clusters
			index=index_offset+j;
			for(i=0;i<pPIGasResult->no_clusters;i++)
			{
				PI_index=PI_index_offset+i;
				concentration=pPIGasResult->pGas->getConcentration(i);
				
				// add up x
				y = ScalarJacobianElements[PI_index].getMean();
				y *= ScalarJacobianElements[index].getMean();
				y -= I * ScalarHessian[PI_index][index].getMean();
				x += y * concentration;

				// add up x_error
				y_error = ScalarJacobianElements[PI_index].getMean()*ScalarJacobianElements[index].getStddev();
				y_error += ScalarJacobianElements[PI_index].getStddev()*ScalarJacobianElements[index].getMean();
				y_error -= I*ScalarHessian[PI_index][index].getStddev();
				y_error -= I_error*ScalarHessian[PI_index][index].getMean();
				x_error += y_error * concentration;
			}
			x_error=1e5*(x_error/II-(2.0*I_error*x)/(II*I));
			x*=1e5/II;
			fprintf(fptr, output_format_string, x, x_error, '\t');
		}
	}
}




