/**************************************************************************************************************
*
*       This file,
*
*                            /Result.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Result_h
#define _Result_h

#include "SimulationObject.h"

#include "Math/StochasticVariable.h"
#include "Math/StokesVector.h"
#include "Math/StokesMatrix.h"
#include "Math/Vector.h"
#include "Math/Matrix.h"

#include <string>
#include <list>

using namespace std;

class Result: public SimulationObject
{
	class StochasticVariable<double> *AllStochasticVariables;
	int no_stochastic_variables;
		
	class StochasticVariable<double> *ScalarHessianMatrixElements;
	class StochasticVariable<double> *ScalarJacobianElements;

	bool calculate_cell_path_length, calculate_Jacobians, calculate_Hessian, output_polarized_functionals;
	bool calculate_cell_sensitivity;
	
	// accuracy
	bool restrictive_precision_check;
	double precision_threshold;

	// output
	bool output_enabled;
	bool normalize_with_F0;
	bool enable_RRS;
	bool enable_event_counts;
	bool suppress_Hessian_output;
	bool suppress_Jacobian_output;

	// int no_air_clusters;
	public:
		
		Result();		
		~Result();

		/*
			important thing is that FunctionalLayer enables gas/particle activated flags
		*/

		bool allocate(class Atmosphere *pAtmosphere, bool allocate_Hessian);
		void inline enableOutputOfPolarizedFunctionals(bool enable=true){output_polarized_functionals=enable;}
		void inline disableRadianceHessianOutput(bool disable=true){suppress_Hessian_output=disable;}
		void inline disableRadianceJacobianOutput(bool disable=true){suppress_Jacobian_output=disable;}
		void inline enableRRS(bool enable=true){enable_RRS=enable;}
		void inline enableEventCounts(bool enable=true){enable_event_counts=enable;}

		inline void enableCalculateCellPathLength(bool enable=true){calculate_cell_path_length=enable;}
		inline void enableCalculateCellSensitivity(bool enable=true){calculate_cell_sensitivity=enable;}
		inline bool calculateCellPathLength(){return calculate_cell_path_length;}
		inline bool calculateCellSensitivity(){return calculate_cell_sensitivity;}

		void reset();
		// void resetEstimates();

		void registerEstimate(class Estimate *pEstimate);

		// void registerEstimates();

		void finish();
	
		inline void enableOutput(bool enable){output_enabled=enable;}
		inline bool outputEnabled(){return output_enabled;}

		inline bool calculateHessian(){return calculate_Hessian;}

		void print(FILE *fptr, const char *output_format);
		void printPathIntegralResults(const char *name, class list<string>& constituents_list, FILE *fptr);

		bool fillScalarJAndH(class Vector& J, class Matrix& H);

		inline void defineAccuracy(bool restrictive, double precision){restrictive_precision_check=restrictive; precision_threshold=precision;}
		bool metAccuracyCriterium();

		// radiances ----------------------------------------------
		double F0;
		inline void setF0(double I){F0=I;}
		inline void enableNormalizeWithF0(bool enable=true){normalize_with_F0=enable;}

		class StochasticVariable<double> direct_light_radiance;
		class StochasticVariable<double> scalar_radiance, elastic_scalar_radiance;
		class StochasticVariable<class StokesVector> vector_radiance, elastic_vector_radiance;

		// scatter event counts
		class StochasticVariable<double> scalar_RRS_count, scalar_multiple_RRS_count, scalar_elastic_scatter_count, scalar_ground_scatter_count;
		
		// cell path length
		class Vector CellSensitivity;		// result will be given without error

		// particle and gas Jacobians -----------------------------
		int N_Parameters;	// number of parameters: i.e. sum of no_clusters * cluster_size 

		int no_active_gases;
		struct GasResult_s
		{
			int no_clusters;
			bool logarithmic_derivative;
			class StochasticVariable<double> *Jacobian;
			class Gas *pGas;
		} *GasResult;
		struct GasResult_s *findGasResultPtr(const char *name);

		int no_active_particles;
		struct ParticleResult_s
		{
			int no_clusters, cluster_size, no_phase_function_parameters, no_parameters;
			bool logarithmic_derivative;
			class StochasticVariable<double> *Jacobian, *SSAAndPJacobian;
			class Particle *pParticle;
		} *ParticleResult;
		struct ParticleResult_s *findParticleResultPtr(const char *name);

		// Hessian data -------------------------------------------
		class StochasticVariable<double> **ScalarHessian;
};

#endif // _Result_h


