/**************************************************************************************************************
*
*       This file,
*
*                            /RingEvent.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "RingEvent.h"
#include "Air.h"
#include "error.h"
#include <math.h>

bool RingEvent::allocate()
{
	return N2.allocate(GasMixtureDCS::getMoleculePtr(Molecule::N2)) && \
		O2.allocate(GasMixtureDCS::getMoleculePtr(Molecule::O2));
}

void RingEvent::setTemperature(const double& temperature)
{
	N2.setTemperature(temperature);
	O2.setTemperature(temperature);
	this->temperature=temperature;
}

double RingEvent::calculateCrossSections(const double& wavelength)
{
	lambda=wavelength;
	// actual calculation
	N2.calculateCrossSections(wavelength);
	O2.calculateCrossSections(wavelength);

	// algebra 
	sigma_out=p_N2*N2.getSigmaOut()+p_O2*O2.getSigmaOut();
	
	sigma_in_N2=p_N2*N2.getSigmaIn();
	sigma_in_O2=p_O2*O2.getSigmaIn();
	sigma_in=sigma_in_N2+sigma_in_O2;

	// energy conservation importance sampling weight
	sigma_in_sigma_out_ratio=sigma_in/sigma_out;

	return sigma_out;
}

double RingEvent::sampleWavelength()
{
	double sigma=prand()*sigma_in;
	if(sigma<sigma_in_N2)
		return N2.sampleWavelength (sigma/sigma_in_N2);
	return O2.sampleWavelength ( (sigma-sigma_in_N2) / sigma_in_O2);
}

void RingEvent::getInScatteringWavelengths(double *pwavelength)
{
	int i, n=MolecularRRSCrossSection::getJMax();
	for(i=0;i<n;i++)
	{
		// order in the list might be different from iother orders !!!
		*pwavelength++=N2.lambda[i][2];
		*pwavelength++=N2.lambda[i][3];
		*pwavelength++=O2.lambda[i][2];
		*pwavelength++=O2.lambda[i][3];
	}
}

void RingEvent::getInScatteringCrossSections(double *psigma) 
{
	int i, n=MolecularRRSCrossSection::getJMax();
	for(i=0;i<n;i++)
	{
		// order in the list might be different from iother orders !!!
		*psigma++ = p_N2 * N2.sigma[i][2];
		*psigma++ = p_N2 * N2.sigma[i][3];
		*psigma++ = p_O2 * O2.sigma[i][2];
		*psigma++ = p_O2 * O2.sigma[i][3];
	}
}

void RingEvent::print(FILE *fptr, bool detailed)
{
	if(detailed)
	{
		/*
			double sigma_in, sigma_out;	// total cross section
			double sigma_in_N2, sigma_in_O2;
			double sigma_in_sigma_out_ratio;
			double lambda;			// center wavelength
			double temperature;
		*/
		// fprintf(fptr, "# sigma_in=%lf\n", sigma_in);
		// fprintf(fptr, "# sigma_out=%lf\n", sigma_out);
		// fprintf(fptr, "# sigma_in_N2=%lf\n", sigma_in_N2);
		// fprintf(fptr, "# sigma_in_O2=%lf\n", sigma_in_O2);
		fprintf(fptr, "# lambda=%le\n", lambda);
		fprintf(fptr, "# temperature=%le\n", temperature);
		int i, j, k,  n=MolecularRRSCrossSection::getJMax();
		MolecularRRSCrossSection *pMC[]={&N2, &O2};
		for(i=0;i<n;i++)
		{
			for(j=0;j<2;j++)
			{
				fprintf(fptr, "%le\t", pMC[j]->wT[i]);
				for(k=0;k<4;k++)
				{
					fprintf(fptr, "%.3lf\t%le", 1e9*pMC[j]->lambda[i][k], pMC[j]->sigma[i][k]);
					if(k<3)
						fputc('\t', fptr);
				}
				if(j<1)
					fputc('\t', fptr);
			}
			fputc('\n', fptr);
		}
	}
	else
	{
		fprintf(fptr, "\t%lf", temperature);
		fprintf(fptr, "\t%le", sigma_out);
		fprintf(fptr, "\t%le", sigma_in);
		/*int i, j, J_max=DiatomicMoleculeEnsemble::getJMax();
		class DiatomicMoleculeEnsemble *DAMPtr[2]={&N2, &O2};
		for(i=0;i<2;i++)
			for(j=0;j<J_max;j++)
				DAMPtr[i]->print(j, fptr);*/
		fputc('\n',fptr);
	}
}




