/**************************************************************************************************************
*
*       This file,
*
*                            /Screen.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Screen.h"

Screen::Screen()
{
	// resetMarkers();
	//raytracer.enableSaveTrajectories();
	//raytracer.setScreen(this);

	// depiction
	for(int i=0;i<1000;i++)
		enable_trace_event[i]=true;
	enable_trace_event[TE_REFRACTION_X]=false;
	enable_trace_event[TE_REFRACTION_Y]=false;
	enable_trace_event[TE_REFRACTION_Z]=false;

	PointCloudScreen.setPerspective(900.0, 1590.0);
	PointCloudScreen.setBackgroundColor(0x0);

	SurfaceGridPoints.setSurfaceGridPtr(&surface);
	SurfaceGridPoints.setGranularity(0.1);
	SurfaceGridPoints.setGridColor(0xa0a0a0);

	PCRenderList.setRenderScreenPtr(&PointCloudScreen);
	PCRenderList.addPointCollectionPtr(&TraceEvents);
	PCRenderList.addPointCollectionPtr(&SurfaceGridPoints);
	PCRenderList.resetPerspective();

	pSimulationControlFlag=getRaytracerPtr()->getControlFlagPtr();

	width=640;
	height=480;
}

bool Screen::loadScatteredTrajectories()
{
	char filename[2047];
	raytracer.formatScatteredTrajectoriesFilename(filename);
	return TraceEvents.loadFromFile(filename, enable_trace_event);
}

bool Screen::loadUnscatteredTrajectories()
{
	char filename[2047];
	raytracer.formatUnscatteredTrajectoriesFilename(filename);
	return TraceEvents.loadFromFile(filename, enable_trace_event);
}

bool Screen::initialize(int argc, char *argv[])
{
	if(argc!=9 && argc!=10)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "wrong number of arguments\n");
		return false;
	}

	// screen allocation
	width=atoi(argv[1]);
	height=atoi(argv[2]);
	
	if(!PointCloudScreen.allocate(width, height, true))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// perspective
	strcpy(fn_perspective, argv[3]);
	if(!PCRenderList.loadPerspective(fn_perspective))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	
	// decode event mask
	strcpy(event_mask, argv[4]);
	int i, n=strlen(event_mask);
	for(i=0;i<n;i++)
		enable_trace_event[i]=event_mask[i]=='1'?true:false;
	// outbase and 
	strcpy(fn_outbase, argv[5]);
	TraceEvents.setContributionCenter(atof(argv[6]));
	TraceEvents.setContributionSigma(atof(argv[7]));
	TraceEvents.recalculateColors();
	strcpy(fn_simulation, argv[8]);
	if(!readAndLoadFiles(fn_simulation) || !Simulation::initialize())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return -1;
	}
	raytracer.enableSaveTrajectories();
	return true;
}

void Screen::post_process_run(const int& run_id)
{
	char file_name[1000];
	if(run_id==0)
		sprintf(file_name, "%s.eps", fn_outbase);
	else
		sprintf(file_name, "%s_%d.eps", fn_outbase, run_id);
	
	loadScatteredTrajectories();
	PCRenderList.restorePerspective();
	PCRenderList.renderPointCollections();
	PCRenderList.exportToEPS(file_name, false);
}







