/**************************************************************************************************************
*
*       This file,
*
*                            /SeaSurface.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "SeaSurface.h"
#include <math.h>

#include "trigonometry.h"
#include "refraction.h"

SeaSurface::SeaSurface()
{
	crosswind_resolution=100;
	upwind_resolution=100;
}

bool SeaSurface::setMapResolution(const int& crosswind_resolution, const int& upwind_resolution)
{
	if(!A.allocate(upwind_resolution, crosswind_resolution) || 
		!mu_Matrix.allocate(upwind_resolution, crosswind_resolution) || 
		!zx.allocate(crosswind_resolution) ||
		!zy.allocate(upwind_resolution))
	{
		errorIn(__FILE__, __LINE__, __func__);	
		return false;
	}
	this->crosswind_resolution = crosswind_resolution;
	this->upwind_resolution = upwind_resolution;
	return true;
}

bool SeaSurface::setWindSpeed(const double& W, bool clean_surface)
{
	if(W<1.0 || W>15.0)
	{
		errorIn(__FILE__, __LINE__, __func__);	
		fprintf(stderr, "wind speed must be in the interval [1,15] m/s\n");
		return false;
	}
	if(clean_surface)
	{
		// clean surface
		sigma_crosswind = sqrt ( 0.003 + 1.92e-3 * W ) ;
		sigma_upwind = sqrt ( 3.16e-3 * W ) ;
		c_21 = 0.01 - 0.0086 * W ;
		c_03 = 0.04 - 0.033 * W ;
		c_40 = 0.4 ;
		c_22 = 0.12 ;
		c_04 = 0.23 ;
	}
	else
	{
		// slick surface
		sigma_crosswind = sqrt ( 0.003 + 0.84e-3 * W ) ;
		sigma_upwind = sqrt ( 0.005 + 0.78e-3 * W ) ;
		c_21 = 0.0 ;
		c_03 = 0.02 ;
		c_40 = 0.36 ;
		c_22 = 0.1 ;
		c_04 = 0.26 ;
	}
	int i, j;
	// 3 sigma range 	
	const double sigma_range=2.2;
	double mx, x1 = - sigma_range * sigma_crosswind;
	double my, y1 = - sigma_range * sigma_upwind;
	// ------------------------
	for(i=0;i<upwind_resolution;i++)
		zy[i]=y1+(double(2*i+1) * sigma_range * sigma_upwind ) / double ( upwind_resolution );
	for(j=0;j<crosswind_resolution;j++)
		zx[j]=x1+(double(2*j+1) * sigma_range * sigma_crosswind ) / double ( crosswind_resolution );
	// ------------------------
	for(i=0;i<upwind_resolution;i++)
	{
		my = zy[i];
		for(j=0;j<crosswind_resolution;j++)
		{
			mx = zx[j];
			A[i][j] = GramCharlierSlopePDF(mx, my);
		}
	}
	if(!initializeFromMatrix(A, zx, zy))
	{
		errorIn(__FILE__, __LINE__, __func__);	
		return false;
	}
	// calculate array of directions
	if(!face_normal.allocate(upwind_resolution, crosswind_resolution))
	{
		errorIn(__FILE__, __LINE__, __func__);	
		return false;
	}
	
	for(i=0;i<upwind_resolution;i++)
	{
		my = zy[i] ;
		for(j=0;j<crosswind_resolution;j++)
		{
			mx = zx[j] ;
			face_normal[i][j] = norm ( Vector3D( -mx, -my, 1.0 ) ) ;
		}
	}
	return true;
}

double SeaSurface::GramCharlierSlopePDF(const double& z_crosswind, const double& z_upwind)
{
	double xi = z_crosswind / sigma_crosswind;
	double xi2 = xi * xi;
	double xi4 = xi2 * xi2;

	double eta = z_upwind / sigma_upwind;
	double eta2 = eta * eta;
	double eta3 = eta2 * eta;
	double eta4 = eta2 * eta2;
	
	double f = exp ( -0.5 * ( xi2 + eta2 ) ) / ( 2.0 * M_PI * sigma_crosswind * sigma_upwind ) ;
	double a = 1.0;

	a -= 0.5 * c_21 * ( xi2 - 1.0 ) * eta ;
	a -= c_03 * ( eta3 - 3.0 * eta ) / 6.0 ;
	a += c_40 * ( xi4 - 6.0 * xi2 + 3.0 ) / 24.0 ;
	a += c_22 * ( xi2 - 1.0 ) * ( eta2 - 1.0 ) / 4.0 ;
	a += c_04 * ( eta4 - 6.0 * eta2 + 3.0 ) / 24.0 ;

	return f * a;
}

class Vector3D SeaSurface::getNormalSample()
{
	double mc, mu;
	if(!getSample(&mc, &mu))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return Vector3D(0.0, 0.0, 1.0);
	}
	return norm ( Vector3D( -mc, -mu, 1.0 ) ) ;
}

bool SeaSurface::recalculateSlopePDF(class Vector3D& k_in)
{
	int i,j;
	double weight;
	for(i=0;i<upwind_resolution;i++)
		for(j=0;j<crosswind_resolution;j++)
		{
			weight = - correctMu ( face_normal[i][j] * k_in );
			if(weight>=0.0)
			{
				s[i][j] = weight * s0[i][j];
				mu_Matrix[i][j] = weight;
			}
			else
			{
				s[i][j] = 0.0;
				mu_Matrix[i][j] = 0.0;
			}
		}
	return normalize();
}

bool SeaSurface::calculateEffectiveReflectivityMap(const int& azimuth_resolution, const int& polar_resolution, const double& refractive_index)
{
	class Vector _phi;
	class TMatrix<class Vector3D> KMatrix;
	if(!RMatrix.allocate(polar_resolution+1, azimuth_resolution+1) \
		|| !_phi.allocate(azimuth_resolution+1) \
		|| !KMatrix.allocate(polar_resolution+1, azimuth_resolution+1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	RMatrix_azimuth_resolution=azimuth_resolution;
	RMatrix_polar_resolution=polar_resolution;
	int i, j, m, n;
	class Vector3D k_in;
	double theta, phi, weight, weight_sum, R, R_sum;
	for(j=0;j<=azimuth_resolution;j++)
		_phi[i] = (2.0*M_PI*double(j)) / double(azimuth_resolution);
	for(i=0;i<=polar_resolution;i++)
	{
		theta = (0.5*M_PI*double(i)) / double(polar_resolution);
		for(j=0;j<=azimuth_resolution;j++)
		{
			phi=_phi[j];
			KMatrix[i][j] = Polar(1.0, theta, phi);
		}
	}
	// create R map
	for(i=0;i<=polar_resolution;i++)
		for(j=0;j<=azimuth_resolution;j++)
		{
			k_in = KMatrix[i][j];
			// for k_in calculate cos(theta_inc) - weighted average reflectivity
			if(!recalculateSlopePDF(k_in))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error calculating reflectivity for incident direction theta=%lf°, phi=%lf°\n"), Degree(theta), Degree(phi);
				return false;
			}
			R_sum=0.0;
			weight_sum=0.0;
			for(m=0;m<upwind_resolution;m++)
				for(n=0;n<crosswind_resolution;n++)
				{
					weight = s[i][j];
					R = calculateFresnelCoefficient(mu_Matrix[i][j], refractive_index);

					R_sum += weight * R ;
					weight_sum += weight ;
				}
			RMatrix[i][j] = R_sum / weight_sum;
		}
	return true;
}

double SeaSurface::calculateEffectiveReflectivity(class Vector3D& k_in, const class Vector3D& surface_normal)
{
	int i, j;
	double r, theta, phi;
	if(correctMu(k_in*surface_normal)>0.0)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return 0.0;
	}
	k_in.toPolar(&r, &theta, &phi);
	i=int(double(RMatrix_polar_resolution)*((2.0*theta)/M_PI));
	if(i<0 || i>=RMatrix_polar_resolution)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return 0.0;
	}
	
	j=int(double(RMatrix_azimuth_resolution)*(phi/(2.0*M_PI)));
	if(j<0 || j>=RMatrix_azimuth_resolution)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return 0.0;
	}
	return RMatrix[i][j];
}






