/**************************************************************************************************************
*
*       This file,
*
*                            /SeaSurface.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _SeaSurface_h
#define _SeaSurface_h

#include "MapSampler.h"
#include "Math/Vector3D.h"
#include "Math/TMatrix.h"

class SeaSurface: public MapSampler
{
/*
	parametrisation of the sea surface slopes after Cox and Munk 1954
*/
	int crosswind_resolution, upwind_resolution;
	double wind_speed;	// meters per second

	// parameters for the Gram-Charlier PDF:
	double sigma_crosswind, sigma_upwind;
	double c_21, c_03, c_40, c_22, c_04;

	// discrete version of the pdf:
	class Matrix A;
	class Vector zx, zy;

	// array of normal directions
	class TMatrix<Vector3D> face_normal;
	class Matrix mu_Matrix, RMatrix;

	int RMatrix_azimuth_resolution;
	int RMatrix_polar_resolution;

	bool setMapResolution(const int& crosswind_resolution, const int& upwind_resolution);

	public:

		SeaSurface();

		class Vector3D getNormalSample();
		bool recalculateSlopePDF(class Vector3D& k_in);

		bool calculateEffectiveReflectivityMap(const int& polar_resolution, const int& azimuth_resolution, const double& refractive_index);
		double calculateEffectiveReflectivity(class Vector3D& k_in, const class Vector3D& surface_normal);
		
		inline bool allocate(){return setMapResolution(crosswind_resolution, upwind_resolution);}

		inline int *getCrosswindMapResolutionPtr(){return &crosswind_resolution;}
		inline int *getUpwindMapResolutionPtr(){return &upwind_resolution;}

		bool setWindSpeed(const double& windspeed_in_meters_per_second, bool clean_surface=true);
		double GramCharlierSlopePDF(const double& z_crosswind, const double& z_upwind);
		
};

#endif // _SeaSurface_h



