/**************************************************************************************************************
*
*       This file,
*
*                            /Simulation.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/


#include "Simulation.h"
#include "SimulationObject.h"

#include "GeometryList.h"

#include "File.h"
#include "error.h"

#ifdef WIN32
	#include <dir.h>
#else
	#include <sys/stat.h>
	#include <sys/types.h>
#endif

Simulation::Simulation()
{
	IOName="simulation";

	wavelength_nm=360.0;
	b_use_HITRANDataBase=false;

	b_time_integration=false;
	b_time_in_seconds=false;
	has_valid_configuration=false;

	SimulationObject::setSimulationPtr(this);

	// IO setup
	addParameter(fn_raytracer, type_configuration, "fn_raytracer", 0, "raytracer configuration file");
	addParameter(fn_geometry, type_configuration, "fn_geometry", 0, "geometry configuration file");
	addParameter(fn_grid, type_configuration, "fn_grid", 0, "grid configuration file");
	addParameter(fn_atmosphere, type_configuration, "fn_atmosphere|fn_medium", 0, "(optical) atmosphere configuration file");
	addParameter(fn_surface, type_configuration, "fn_surface", 0, "solid surface configuration file");
	addParameter(fn_solar_irradiance, type_file, "fn_solar_irradiance|fn_Fraunhofer_spectrum|fn_F0", 0, "extra terrestrial solar irradiance spectrum file. Format: wavelength [nm] <TAB> irradiance [W/m^2]");

	addParameter(&wavelength_distribution, type_enum, "wavelength_distribution", "delta|Gauss|file|flat", "wavelength distribution of detected wavelength spectrum: delta|Gauss|flat");
	addParameter(&wavelength_nm, type_double, "wavelength_nm", 0, "wave length in nm");
	addParameter(&slit_sigma_nm, type_double, "slit_sigma_nm", 0, "slit function sigma in nm");
	addParameter(&wavelength_A_nm, type_double, "wavelength_A_nm", 0, "wave length A in nm");
	addParameter(&wavelength_B_nm, type_double, "wavelength_B_nm", 0, "wave length B in nm");
	addParameter(fn_spectral_sensitivity, type_file, "fn_spectral_sensitivity", 0, "spectral sensitivity file: 2 columns, wavelength [nm] <TAB> sensitivity (unnormalized)");

	addParameter(&b_use_HITRANDataBase, type_bool, "b_use_HITRANDataBase", 0, "setup and use HITRAN spectral database");
	addParameter(fn_HITRAN_database_configuration, type_configuration, "fn_HITRAN_database_configuration", 0, "HITRAN database reference");
	addParameter(fn_table_description, type_file, "fn_table_description|fn_data_table_format|fn_data_table_columns", 0, "file containing format of the data table to be processed");
	addParameter(fn_data_table, type_file, "fn_data_table|fn_input_data", 0, "file with data in columns as described in the file above");
	addParameter(&b_time_integration, type_bool, "b_time_integration", 0, "interpolate and integrated geometry data (explicit) in fn_data_table");
	addParameter(&b_time_in_seconds, type_bool, "b_time_in_seconds", 0, "time stamps given in seconds instead of date/time in dd.mm.yyyy hh.mm.ss?");
	addParameter(fn_time_frames, type_file, "fn_time_frames|fn_integration_times", 0, "integration time intervals definition file");
	addParameter(fn_functional_layer, type_configuration, "fn_functional_layer|fn_functionals", 0, "file describing the functionals to be calculated");

	strcpy(fn_HITRAN_database_configuration,"hitran.conf");
	strcpy(fn_raytracer, "raytracer.conf");
	strcpy(fn_geometry, "geometry.conf");
	strcpy(fn_functional_layer, "functionals.conf");
	strcpy(fn_spectral_sensitivity, "spectral_sensitivity.txt");

#ifdef WIN32
	strcpy(fn_grid, "grid\\grid.conf");
	strcpy(fn_atmosphere, "atmosphere\\atmosphere.conf");
	strcpy(fn_surface, "surface\\surface.conf");
#else
	strcpy(fn_grid, "grid/grid.conf");
	strcpy(fn_atmosphere, "atmosphere/atmosphere.conf");
	strcpy(fn_surface, "surface/surface.conf");
#endif

	strcpy(fn_solar_irradiance, "solar_irradiance.txt");
	wavelength_distribution=delta;
}

static bool determineGridType(char *file_name, enum GridIO::GridType_e& gt)
{
	class GridIO gridIO;
	if(!gridIO.read(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not read grid definition from file \"%s\"!\n", file_name);
		return false;
	}
	gt=gridIO.getGridType();
	return true;
}

double Simulation::getWavelengthSample()
{
	switch(wavelength_distribution)
	{
		case Gauss:
			return 1e-9*gaussSampleBoxMullerMethod(slit_sigma_nm, wavelength_nm);	
		case spectral_sensitivity_file:
			return 1e-9*SpectralSensitivity.takeSample();
		case flat:
			return 1e-9*(wavelength_A_nm+(wavelength_B_nm-wavelength_A_nm)*prand());
		case delta:
		default:;
	}
	return 1e-9*wavelength_nm;
}

bool Simulation::loadFiles()
{
	has_valid_configuration=false;

	// center wavelength
	if(wavelength_distribution==flat)
		wavelength_nm=0.5*(wavelength_A_nm+wavelength_B_nm);

	// ######################## load member objects
	char file_name[2048];

	// select grid:
	GridIO::GridType_e grid_type=GridIO::sp1D;
	formatGridFilename(file_name);
	if(!determineGridType(file_name, grid_type))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error determining grid type of file \"%s\"!\n", file_name);
		return false;
	}
	switch(grid_type)
	{
		case GridIO::pp1D:
			pGrid=&pp1DGrid;
			break;

		case GridIO::pp3D:
			pGrid=&pp3DGrid;
			break;

		case GridIO::sp1D:
			pGrid=&sp1DGrid;
			break;

		case GridIO::sp3D:
			pGrid=&sp3DGrid;
			break;

		default:
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"grid type not yet implemented!\n");
			return false;
	}
	pGrid->setSurfaceGridPtr(&surface);
	if(!pGrid->readAndLoadFiles(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load grid from file \"%s\"!\n", file_name);
		return false;
	}
	// SimulationObject::setGridPtr(pGrid);
	SimulationObject::updateGrid();
	surface.setGridPtr(pGrid);

	// surface
	formatSurfaceFilename(file_name);
	if(!surface.readAndLoadFiles(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load surface from file \"%s\"!\n", file_name);
		return false;
	}

	// HITRAN database
	if(b_use_HITRANDataBase)
	{
		formatHITRANDatabaseFilename(file_name);
		if(!Gas::setupHITRAN(file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"failed to open HITRAN database using file \"%s\"!\n",file_name);
			return false;
		}
	}

	// atmosphere
	formatAtmosphereFilename(file_name);
	if(!atmosphere.readAndLoadFiles(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load atmosphere from file \"%s\"!\n", file_name);
		return false;
	}
	SimulationObject::updateAtmosphere();

	// geometry
	formatGeometryFilename(file_name);
	if(!geometry.read(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load geometry from file \"%s\"!\n", file_name);
		return false;
	}
	
	// raytracer:
	formatRaytracerFilename(file_name);
	if(!raytracer.RaytracerIO::read(file_name))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load raytracer from file \"%s\"!\n", file_name);
		return false;
	}
	formatFunctionalLayerFilename(file_name);
	if(!raytracer.FunctionalLayer::read(file_name) || !raytracer.FunctionalLayer::loadFiles())
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not load functional file \"%s\"!\n", file_name);
		return false;
	}
	// load solar irradiance spectrum:
	formatSolarIrradianceFilename(file_name);
	if(!raytracer.loadSolarIrradianceSpectrum(file_name))
		return false;
	// load spectral sensitivity file
	if(wavelength_distribution==spectral_sensitivity_file)
	{
		formatFilePath(file_name, fn_spectral_sensitivity);
		if(!SpectralSensitivity.load(file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not load spectral sensitivity file \"%s\"!\n", file_name);
			return false;
		}
	}

	createBasicIOLinkList();
	return (has_valid_configuration=true);
}

bool Simulation::run(const enum RTESolvingMethod_e& method)
{
	if(!has_valid_configuration)
		return false;
	switch(method)
	{
		case MonteCarlo:
			if(!raytracer.run())
				return false;
			break;
		case AnalyticalSSA:
		{
			double intensity;
			if(!raytracer.runAnalyticalSSA(&intensity))
				return false;
			printf("%le\n", intensity);
			break;
		}
		default:
			errorIn(__FILE__, __LINE__, __func__);
			return false;
	}
	post_process_run(0);
	return true;
}

void Simulation::createBasicIOLinkList()
{
	int i, n;
	BasicIOLinkList.clear();

	// this, raytracer and geometry
	BasicIOLinkList.push_back(make_pair(this->IOName, this));
	
	class BasicIO *pRaytracerIO=(Raytracer::RaytracerIO *)&raytracer;
	BasicIOLinkList.push_back(make_pair(pRaytracerIO->getIOName(), pRaytracerIO));

	class BasicIO *pFunctionalsIO=raytracer.getFunctionalLayerIOPtr();
	BasicIOLinkList.push_back(make_pair(pFunctionalsIO->getIOName(), pFunctionalsIO));

	BasicIOLinkList.push_back(make_pair(geometry.getIOName(), &geometry));
	
	// grids ----------------------------------------------------------------
	BasicIOLinkList.push_back(make_pair("grid", pGrid));
	BasicIOLinkList.push_back(make_pair("sp1DGrid", &sp1DGrid));
	BasicIOLinkList.push_back(make_pair("sp3DGrid", &sp3DGrid));
	BasicIOLinkList.push_back(make_pair("pp1DGrid", &pp1DGrid));
	BasicIOLinkList.push_back(make_pair("pp3DGrid", &pp3DGrid));

	// atmosphere -----------------------------------------------------------
	BasicIOLinkList.push_back(make_pair(atmosphere.getIOName(), &atmosphere));
	// gases ...
	n=atmosphere.getNOGases();
	class Gas *gas=atmosphere.getGasPtr();
	for(i=0;i<n;i++)
		BasicIOLinkList.push_back(make_pair(gas[i].getName(), gas+i));
	// particles ...
	n=atmosphere.getNOParticles();
	class Particle *particle=atmosphere.getParticlePtr();
	for(i=0;i<n;i++)
		BasicIOLinkList.push_back(make_pair(particle[i].getName(), particle+i));
	// surface ----------------------------------------------------------
	BasicIOLinkList.push_back(make_pair(surface.getIOName(), &surface));
}

void Simulation::printBasicIOLinkList(FILE *fptr)
{
	list< pair<const char *, class BasicIO *> >::iterator pBasicIOLink;
	for(pBasicIOLink=BasicIOLinkList.begin();pBasicIOLink!=BasicIOLinkList.end();pBasicIOLink++)
		fprintf(fptr, "%s\n", pBasicIOLink->first);
}

bool Simulation::createIOParameterList(list<struct DataTableIOParameter_s> *pIOParameterList, char *file_name)
{
/*
	opens file file_name (usually fn_table_description) and extracts the list of parameters that shall be 
	inserted into the .conf objects
*/
	if(!has_valid_configuration)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "no valid configuration loaded!\n");
		return false;
	}
	pIOParameterList->clear();
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not open data table description file \"%s\"\n", file_name);
		return false;
	}
	// read ------------------------------------------------------------------
	int line_number=1;
	char line_buffer[1000], object_name[50], parameter_name[50];
	while(fgets(line_buffer, 1000, fptr))
	{
		remove('\n', line_buffer);
		if(!strcmp(line_buffer, "id"))
			pIOParameterList->push_back(DataTableIOParameter_s(DataTableIOParameter_s::id, string("id"), 0, 0));
		else if(!strcmp(line_buffer, "time_stamp") || !strcmp(line_buffer, "t"))
			pIOParameterList->push_back(DataTableIOParameter_s(DataTableIOParameter_s::time_stamp, string("time_stamp"), 0, 0));
		else if(!strcmp(line_buffer, "carry") || !strcmp(line_buffer, "+"))
			pIOParameterList->push_back(DataTableIOParameter_s(DataTableIOParameter_s::carry, string("carry"), 0, 0));
		else if(!strcmp(line_buffer, "ignore") || !strcmp(line_buffer, "-") || !strcmp(line_buffer, "*"))
			pIOParameterList->push_back(DataTableIOParameter_s(DataTableIOParameter_s::ignore, string("ignore"), 0, 0));
		else
		{
			if(numberOf('.', line_buffer)!=1)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in line %d of file \"%s\": missing '.' for 'object.parameter' separation!\n", line_number, file_name);
				fclose(fptr);
				return false;
			}
			extractWordFromList(line_buffer, 0, '.', object_name);
			extractWordFromList(line_buffer, 1, '.', parameter_name);
			// find object_name in BasicIOLinkList
			list< pair<const char *, class BasicIO *> >::iterator pBasicIOLink;
			for(pBasicIOLink=BasicIOLinkList.begin();pBasicIOLink!=BasicIOLinkList.end();pBasicIOLink++)
				if(!strcmp(object_name, pBasicIOLink->first))
					break;
			if(pBasicIOLink==BasicIOLinkList.end())
			{
				// did not find object_name :(
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in line %d of file \"%s\":\nobject \"%s\"not found.\nPossible parameters are listed below:\n", \
					line_number, file_name, object_name);
				printBasicIOLinkList(stderr);
				return false;
			}
			struct BasicIO::Parameter_s *pIOParameter=pBasicIOLink->second->findParameter(string(parameter_name));
			if(!pIOParameter)
			{
				fclose(fptr);
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error in line %d of file \"%s\":\nparameter \"%s\" for object \"%s\" not found.\nPossible parameters are listed below:\n", \
					line_number, file_name, parameter_name, object_name);
				printBasicIOLinkList(stderr);
				return false;
			}	
			pIOParameterList->push_back(DataTableIOParameter_s(DataTableIOParameter_s::parameter, string(parameter_name), pIOParameter, pBasicIOLink->second));
		}
		line_number++;
	}
	fclose(fptr);
	return true;
}

bool Simulation::processTable(enum RTESolvingMethod_e method)
{
	char file_name[2000];
	formatFilePath(file_name, fn_table_description);
	if(b_time_integration)
	{
		int i, n;
		class GeometryList geometryList;
		geometryList.setSimulationPtr(this);
		// reject AnalyticalSSA model for time integration
		if(method==AnalyticalSSA)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "time integration not implemented for analytical single scattering approximation!\n");
			return false;
		}
		/*// check if only geometry parameters are given in the data file
		int index=0;
		bool has_time_stamp_column=false;
		for(pDataTableIOParameter=IOParameterList.begin();pDataTableIOParameter!=IOParameterList.end();pDataTableIOParameter++)
		{
			switch(pDataTableIOParameter->column_type)
			{
				case DataTableIOParameter_s::time_stamp:
					if(index!=0)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "time_stamp found in column %d, but needs to be in the first!\n", index+1);
						return false;
					}
					has_time_stamp_column=true;
					break;
				case DataTableIOParameter_s::parameter:
					if(pDataTableIOParameter->BasicIOPtr != &geometry)
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "found parameter of class \"%s\" in %s, but ", pDataTableIOParameter->BasicIOPtr->getIOName(), file_name);
						fprintf(stderr, "only class geometry parameters allowed for time integration!\n");
						return false;
					}
					break;
			}
			index++;
		}
		if(!has_time_stamp_column)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "missing time stamp column in file %s\n", file_name);
			return false;
		}
		// let Geometry object load geometry data
		// if(!geometry.read*/
		if(!geometryList.loadGeometryDescription(file_name, b_time_in_seconds))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error loading geometry description from file %s\n", file_name);
			return false;
		}
		formatFilePath(file_name, fn_data_table);
		if(!geometryList.loadGeometryData(file_name, b_time_in_seconds))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error loading geometry description from file %s\n", file_name);
			return false;
		}
		formatFilePath(file_name, fn_time_frames);
		if(!geometryList.loadTimeFrameData(file_name, b_time_in_seconds))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		n=geometryList.getNOIntegrationFrames();
		for(i=0;i<n;i++)
		{
			if(!raytracer.run(&geometryList, i))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "file \"%s\": error processing time frame %d\n", file_name, i+1);
				return false;
			}
			post_process_run(i);
		}
	}
	else
	{
		list<struct DataTableIOParameter_s> IOParameterList;
		list<struct DataTableIOParameter_s>::iterator pDataTableIOParameter;
		if(!createIOParameterList(&IOParameterList, file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error creating IO parameter list from file \"%s\"\n", file_name);	
			return false;
		}
		// process data table
		formatFilePath(file_name, fn_data_table);
		FILE *fptr=fopen(file_name, "r");
		if(!fptr)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not open data table file \"%s\"\n", file_name);
			return false;
		}
		if(numberOfColumns(fptr)!=int(IOParameterList.size()))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "file \"%s\" has wrong number of columns\n", file_name);
			fclose(fptr);
			return false;
		}
		int i, line_number=1;
		char value_buffer[50], line_buffer[1000];
		while(fgets(line_buffer, 1000, fptr))
		{
			i=0;
			remove('\n', line_buffer);
			for(pDataTableIOParameter=IOParameterList.begin();pDataTableIOParameter!=IOParameterList.end();pDataTableIOParameter++)
			{
				extractWordFromList(line_buffer, i, '\t', value_buffer);
				switch(pDataTableIOParameter->column_type)
				{
					case DataTableIOParameter_s::parameter:
						if(!pDataTableIOParameter->BasicIOPtr->O2I(pDataTableIOParameter->parameter_name.data(), value_buffer))
						{
							errorIn(__FILE__, __LINE__, __func__);
							fprintf(stderr, "file \"%s\", line %d: \"%s\"\n", file_name, line_number, value_buffer);
							fprintf(stderr, "value \"%s\" not accepted by %s.%s\n", value_buffer, \
								pDataTableIOParameter->BasicIOPtr->getIOName(), pDataTableIOParameter->parameter_name.data());
							fclose(fptr);
							return false;
						}
						printf("%s\t", value_buffer);
						break;
					case DataTableIOParameter_s::carry:
						printf("%s\t", value_buffer);
						break;
					default:
					case DataTableIOParameter_s::ignore:
						break;
				}
				i++;
			}
			switch(method)
			{
				case MonteCarlo:
					if(!raytracer.run())
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr, "file \"%s\", line %d: \"%s\"\n", file_name, line_number, line_buffer);
						fclose(fptr);
						return false;
					}
					break;

				case AnalyticalSSA:
				{
					double intensity;
					if(!raytracer.runAnalyticalSSA(&intensity))
						return false;
					printf("%le\n", intensity);
					break;
				}
				default:
					errorIn(__FILE__, __LINE__, __func__);
					return false;
			}
			post_process_run(line_number);
			line_number++;
		}
		fclose(fptr);
	}
	return true;
}

void Simulation::createConfigurationSkeleton(char *base_directory_name)
{
	char path_name[4095], directory_name[2047], buffer[2047];
#ifdef WIN32
	char dir_character='\\';
#else
	char dir_character='/';
#endif
	strcpy(directory_name, base_directory_name);
	strcpy(dir_base, directory_name);
#ifdef WIN32
	mkdir(directory_name);
#else
	mkdir(directory_name, 0x1ed);
#endif

	sprintf(path_name, "%s%csimulation.conf", directory_name, dir_character);
	writeTemplate(path_name);

	// raytracer
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_raytracer);
	raytracer.RaytracerIO::writeTemplate(path_name);

	// geometry
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_geometry);
	geometry.writeTemplate(path_name);

	// functionals
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_functional_layer);
	raytracer.FunctionalLayer::writeTemplate(path_name);

	// grid
	class GridIO grid;
	extractPathFromFilename(fn_grid, buffer);
	sprintf(path_name, "%s%c%s", directory_name, dir_character, buffer);
#ifdef WIN32
	mkdir(path_name);
#else
	mkdir(path_name, 0x1ed);
#endif
	
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_grid);
	grid.writeTemplate(path_name);

	// surface
	extractPathFromFilename(fn_surface, buffer);
	sprintf(path_name, "%s%c%s", directory_name, dir_character, buffer);
#ifdef WIN32
	mkdir(path_name);
#else
	mkdir(path_name, 0x1ed);
#endif
	
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_surface);
	surface.writeTemplate(path_name);

	// atmosphere	
	extractPathFromFilename(fn_atmosphere, buffer);
	sprintf(path_name, "%s%c%s", directory_name, dir_character, buffer);
#ifdef WIN32
	mkdir(path_name);
#else
	mkdir(path_name, 0x1ed);
#endif
	
	sprintf(path_name, "%s%c%s", directory_name, dir_character, fn_atmosphere);
	atmosphere.write(path_name);
	
	extractPathFromFilename(path_name, directory_name);
	// a gas
	class Gas a_gas;
	sprintf(path_name, "%s%ca_gas.conf", directory_name, dir_character );
	a_gas.writeTemplate(path_name);

	// a particle
	class Particle a_particle;
	sprintf(path_name, "%s%ca_particle.conf", directory_name, dir_character);
	a_particle.writeTemplate(path_name);
}

void Simulation::prepareOpticalMedium()
{
	atmosphere.setCenterWavelength(1e-9*wavelength_nm);
	atmosphere.resetOpticalCells();
}


