/**************************************************************************************************************
*
*       This file,
*
*                            /Simulation.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Simulation_h
#define _Simulation_h

#include "Grid.h"

#include "Raytracer.h"

#include "Atmosphere.h"
#include "Surface.h"
#include "Geometry.h"

#include "Spherical1DGrid.h"
#include "Spherical3DGrid.h"

#include "PlaneParallel1DGrid.h"
#include "PlaneParallel3DGrid.h"
#include "BasicIO.h"
#include "ProbabilityDensity1D.h"

#include <list>
#include <utility>

// class GeometryList;

using namespace std;

class Simulation: public BasicIO
{

public:

	enum RTESolvingMethod_e
	{
		MonteCarlo,
		AnalyticalSSA
	};

	struct DataTableIOParameter_s
	{
		//enum ColumnType_e column_type;
		enum ColumnType_e
		{
			id,	// spectrum name
			time_stamp,
			parameter,
			carry,
			ignore
		} column_type;

		string parameter_name;
		class BasicIO::Parameter_s *pIOParameter;
		class BasicIO *BasicIOPtr;

		// constructor
		DataTableIOParameter_s(enum ColumnType_e ct, string pn, class BasicIO::Parameter_s *pIOP, class BasicIO *p)\
			 : column_type(ct), parameter_name(pn), pIOParameter(pIOP), BasicIOPtr(p) {}
	};

private:

	// IO ##########################################################

	char fn_raytracer[BasicIO_string_size];		// raytracer conf file_name
	char fn_geometry[BasicIO_string_size];		// geometry conf file_name
	char fn_functional_layer[BasicIO_string_size];	// file describing the functinals to be calculated

	char fn_grid[BasicIO_string_size];		// lattice grid
	char fn_atmosphere[BasicIO_string_size];	// (optical) atmosphere description ...
	char fn_surface[BasicIO_string_size];		// solid surface
	
	char fn_solar_irradiance[BasicIO_string_size];  // solar irradiance spectrum

	// wavelength
	enum WavelengthDistribution_e
	{
		delta,
		Gauss,
		spectral_sensitivity_file,
		flat
	} wavelength_distribution;

	double wavelength_nm;				// wavelength in nm
	double wavelength_A_nm;				// wavelength in nm
	double wavelength_B_nm;				// wavelength in nm
	double slit_sigma_nm;
	char fn_spectral_sensitivity [BasicIO_string_size];

	// HITRAN
	bool b_use_HITRANDataBase;
	char fn_HITRAN_database_configuration[BasicIO_string_size];	// HITRAN database reference

	// data table
	char fn_table_description[BasicIO_string_size];		// format of the data table
	char fn_data_table[BasicIO_string_size];		// data table in matrix form
	bool b_time_integration;				// enable time integration
	bool b_time_in_seconds;					// give time indices in seconds
	char fn_time_frames [BasicIO_string_size];		// time interval definition file

	// end of IO ##########################################################

	// configuration file_name for scripts and questions...
	bool has_valid_configuration;

	// table stuff	------------------------------------------
	list< pair<const char *, class BasicIO *> > BasicIOLinkList;
	void createBasicIOLinkList();
	void printBasicIOLinkList(FILE *fptr);

protected:

	// class GeometryList geometryList;

	// raytracer and geometry
	class Raytracer raytracer;
	class Geometry geometry;

	// grid stuff......
	class Grid *pGrid;
	class Spherical1DGrid sp1DGrid;
	class Spherical3DGrid sp3DGrid;

	class PlaneParallel1DGrid pp1DGrid;
	class PlaneParallel3DGrid pp3DGrid;

	// atmosphere and ocean...
	class Atmosphere atmosphere;

	class ProbabilityDensity1D SpectralSensitivity;
	
	// surface
	class Surface surface;

	virtual bool loadFiles();
	virtual void post_process_run(const int& run_id){}

public:

	Simulation();
	//~Simulation();

	inline void setWavelengthNM(const double& lambda_nm){wavelength_nm=lambda_nm;wavelength_distribution=delta;}
	void prepareOpticalMedium();

	void createConfigurationSkeleton(char *base_directory_name);

	bool createIOParameterList(list<struct DataTableIOParameter_s> *pIOParameterList, char *file_name);
	// configuration IO
	inline bool timeInSeconds(){return b_time_in_seconds;}
	inline bool timeIntegration(){return b_time_integration;}

	inline bool reload(){return readAndLoadFiles(conf_file_name) && initialize();}

	inline bool hasValidConfiguration(){return has_valid_configuration;}

	// member object access
	inline class Geometry *getGeometryPtr(){return &geometry;}
	inline class Raytracer *getRaytracerPtr(){return &raytracer;}
	inline class Grid *getGridPtr(){return pGrid;}
	inline class Atmosphere *getAtmospherePtr(){return &atmosphere;}
	inline class Surface *getSurfacePtr(){return &surface;}

	// running...
	bool initialize(){return has_valid_configuration?raytracer.initialize():false;}
	bool run(const enum Simulation::RTESolvingMethod_e& method=Simulation::MonteCarlo);
	bool processTable(enum RTESolvingMethod_e method=MonteCarlo);
	
	// IO ----
	inline bool useHITRANDataBase(){return b_use_HITRANDataBase;}

	inline void formatRaytracerFilename(char *filepath){formatFilePath(filepath, fn_raytracer);}
	inline void formatFunctionalLayerFilename(char *filepath){formatFilePath(filepath, fn_functional_layer);}
	inline void formatGeometryFilename(char *filepath){formatFilePath(filepath, fn_geometry);}
	inline void formatSolarIrradianceFilename(char *filepath){formatFilePath(filepath, fn_solar_irradiance);}

	inline void formatGridFilename(char *filepath){formatFilePath(filepath, fn_grid);}
	inline void formatAtmosphereFilename(char *filepath){formatFilePath(filepath, fn_atmosphere);}
	inline void formatSurfaceFilename(char *filepath){formatFilePath(filepath, fn_surface);}

	void formatHITRANDatabaseFilename(char *filepath){formatFilePath(filepath, fn_HITRAN_database_configuration);}

	inline double getCenterWavelength(){return 1e-9*wavelength_nm;}
	double getWavelengthSample();

};

#endif //_Simulation_h





