/**************************************************************************************************************
*
*       This file,
*
*                            /Spherical1DGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Spherical1DGrid.h"
#include "Geometry.h"
#include "error.h"

bool Spherical1DGrid::loadFiles()
{
	// load height grid...
	char filename[1000];
	formatAltitudeGridFilename(filename);
	
	externalX[radius].defineNameAndUnit("altitude", "km");
	if(!externalX[radius].loadFromFile(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "failed to load radial grid from file \"%s\"!\n", filename);
		return false;
	}
	Radius_Res=externalX[radius].getResolution();

	internalX[radius].defineNameAndUnit("radius", "km");
	if(!internalX[radius].setResolution(Radius_Res))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not allocate internal grid with resolution %d!\n", Radius_Res);
		return false;
	}
	// construct tracing grid
	int i;
	double *ih=internalX[radius].getGridPtr();
	double *eh=externalX[radius].getGridPtr();
	for(i=0;i<=Radius_Res;i++)
		*ih++=*eh+++getPlanetRadius();
		
	return internalX[radius].checkGrid() && calculateGroundLayerIndex();
}

bool Spherical1DGrid::contains(class Vector3D& point)
{
	double r=point.length();
	return r>=ground_radius && r<internalX[radius].getB();
}

bool Spherical1DGrid::calculateCellIntersectionPoint(class Vector3D& Position, struct CellIndex_s *pCI,\
	 class Vector3D& Direction, double *pDistance, class Vector3D *pNextPosition,\
		struct CellIndex_s *pNextCI)
{
	int iR;
	double radius, r_position=Position.length(), b, distance=1e30, th_square, *R=internalX[SphericalGrid::radius].getGridPtr();
	
	b=Position*Direction;
	bool upward_direction=b>=0.0, surface_hit=false;
	iR=pCI->coordinates.spherical.iR;
	if(upward_direction)
	{
		radius=R[iR+1];
		th_square=(radius+r_position)*(radius-r_position)+b*b;
		if(th_square<0.0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		distance=-b+sqrt(th_square);
		pNextCI->coordinates.spherical.iR=iR+1;
		if(pNextCI->coordinates.spherical.iR>=Radius_Res)
			pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		else
			pNextCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	}
	else // down
	{
		if(iR==ground_layer_index)
			radius=ground_radius;
		else
			radius=R[iR];
		th_square=(radius+r_position)*(radius-r_position)+b*b;
		if(th_square<0.0)
		{
			// no intersection with the bottom shell
			radius=R[iR+1];
			th_square=(radius+r_position)*(radius-r_position)+b*b;
			if(th_square<0.0)
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			distance=-b+sqrt(th_square);
			if(distance<0.0)
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			pNextCI->coordinates.spherical.iR=iR+1;
			if(pNextCI->coordinates.spherical.iR>=Radius_Res)
				pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
			else
				pNextCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
		}  
		else
		{
			distance=-b-sqrt(th_square);
			if(distance<0.0)
				distance=-distance;			// <---- ATTENTION HERE: trick applied!!!
			pNextCI->coordinates.spherical.iR=iR-1;
			if(pNextCI->coordinates.spherical.iR<ground_layer_index)
			{
				pNextCI->CellPosition=CELL_POSITION_SURFACE;
				surface_hit=true;
			}
			else
				pNextCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
		}
	}
	*pNextPosition=Position+distance*Direction;
	*pDistance=distance;
	if(surface_hit)
		pSurfaceGrid->adjustSurfacePointAndNormal(*pNextPosition);
	pNextCI->event=TE_REFRACTION_Z;
	return true;
}

bool Spherical1DGrid::calculateCellIndex(class Vector3D& point, struct CellIndex_s *pCI)
{
	double r=point.length();
	if(r<ground_radius)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(r>internalX[radius].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	if(!internalX[radius].findIntervallIndexFastButUnsafe(r, &(pCI->coordinates.spherical.iR)))
		return false;
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

bool Spherical1DGrid::calculateCellIndex(class Vector3D& point, class Vector3D& direction, struct CellIndex_s *pCI)
{
	double r=point.length();
	if(r<ground_radius)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(r>internalX[radius].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	int iR=pCI->coordinates.spherical.iR;
	if(!internalX[radius].findIntervallIndexFastButUnsafe(r, &iR))
		return false;
	if(internalX[radius][iR+1]==r && norm(point) * direction  > 0.0)
		iR++;
	pCI->coordinates.spherical.iR=iR;
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

void Spherical1DGrid::initializeBoxSamplingStructure(class MacroCell *pMacroCell, struct BoxDetector::SamplingStructure_s *pSamplingStructure)
{
	pSamplingStructure->x1=-M_PI;
	pSamplingStructure->x2=M_PI;
	pSamplingStructure->Nx=pSamplingStructure->x2-pSamplingStructure->x1;
	pSamplingStructure->cx=pSamplingStructure->x1;
	pSamplingStructure->dx=M_PI*1e-10;

	pSamplingStructure->y1=0.0;
	pSamplingStructure->y2=M_PI;
	pSamplingStructure->Ny=2;
	pSamplingStructure->cy=-1.0;
	pSamplingStructure->dy=M_PI*1e-10;

	double rA=pMacroCell->internal_xA(2);
	double rB=pMacroCell->internal_xB(2);

	pSamplingStructure->z1=rA;
	pSamplingStructure->z2=rB;
	pSamplingStructure->Nz=rB*rB*rB-rA*rA*rA;
	pSamplingStructure->cz=rA*rA*rA;
	pSamplingStructure->dx=1e-6;

	// sampling functions
	pSamplingStructure->takeXSample=&SamplePhiCoordinate;
	pSamplingStructure->takeYSample=&SampleThetaCoordinate;
	pSamplingStructure->takeZSample=&SampleRadiusCoordinate;

	// volumes / areas
	pSamplingStructure->size[(int)BoxDetector::fullbox]=(4.0*M_PI*pSamplingStructure->Nz)/3.0;

	pSamplingStructure->size[(int)BoxDetector::boxtopface]=4.0*M_PI*rB*rB;
	pSamplingStructure->size[(int)BoxDetector::boxbottomface]=4.0*M_PI*rA*rA;

	pSamplingStructure->size[(int)BoxDetector::boxnorthface]=0.0;
	pSamplingStructure->size[(int)BoxDetector::boxsouthface]=0.0;

	pSamplingStructure->size[(int)BoxDetector::boxeastface]=0.0;
	pSamplingStructure->size[(int)BoxDetector::boxwestface]=0.0;

	pSamplingStructure->size[(int)BoxDetector::boxface]=0.0;
	for(int i=(int)BoxDetector::boxtopface;i<=(int)BoxDetector::boxwestface;i++)
		pSamplingStructure->size[(int)BoxDetector::boxface]+=pSamplingStructure->size[i];
}

void Spherical1DGrid::printCell(FILE *fptr, struct CellIndex_s& CI)
{
	int iR=CI.coordinates.spherical.iR;
	class Grid1D& Radius=internalX[radius];
	fprintf(fptr, "r[%d]=[%lf,%lf]\n", iR, Radius[iR], Radius[iR+1]);
}

bool Spherical1DGrid::acceptGeometry(class Geometry *pGeometry)
{
	int index;
	double c=pGeometry->getAltitude();
	if(externalX[radius].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector altitude coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	return true;
}





