/**************************************************************************************************************
*
*       This file,
*
*                            /Spherical3DGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Spherical3DGrid.h"
#include "Geometry.h"

#include "trigonometry.h"
#include "Math/nan.h"
#include "error.h"


Spherical3DGrid::Spherical3DGrid()
{
	data=0;
	theta_cell_type=0;
}

Spherical3DGrid::~Spherical3DGrid()
{
	if(data)
		delete [] data;
	if(theta_cell_type)
		delete [] theta_cell_type;
}

bool Spherical3DGrid::loadFiles()
{
	// read external grids ############################################################################
	char filename[1000];

	formatAltitudeGridFilename(filename);
	externalX[radius].defineNameAndUnit("altitude", "km");
	if(!externalX[radius].loadFromFile(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "altitude grid file \"%s\"\n", filename);
		return false;
	}
	Radius_Res=externalX[radius].getResolution();

	formatColatitudeGridFilename(filename);
	externalX[theta].defineNameAndUnit("colatitude", "°");
	if(!externalX[theta].loadFromFile(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "colatitude grid file \"%s\"\n", filename);
		return false;
	}
	Theta_Res=externalX[theta].getResolution();
	if(b_cyclic_y)
	{
		if(externalX[theta][Theta_Res]!=180.0 && externalX[theta][0]!=0.0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "colatitude grid file \"%s\":\n", filename);
			fprintf(stderr, "cyclic coordinates enabled, but colatitude coordinate list not cyclic!\n");
			return false;
		}
	}

	formatLongitudeGridFilename(filename);
	externalX[phi].defineNameAndUnit("longitude", "°");
	if(!externalX[phi].loadFromFile(filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "longitude grid file \"%s\"\n", filename);
		return false;
	}
	Phi_Res=externalX[phi].getResolution();
	if(b_cyclic_x)
	{
		if(fmod(externalX[phi][Phi_Res]-externalX[phi][0], 360.0)!=0.0)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "longitude grid file \"%s\":\n", filename);
			fprintf(stderr, "cyclic coordinates enabled, but longitude coordinate list not cyclic!\n");
			return false;
		}
	}

	// grid data allocation ##########################################################################
	int Theta_Res_p1 = Theta_Res + 1 ;
	int i, n = 2 + Radius_Res + Phi_Res + 3 * Theta_Res_p1 ;
	if(data)
		delete [] data;
	data=new double[n];
	if(!data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}
	if(theta_cell_type)
		delete [] theta_cell_type;
	theta_cell_type=new int[Theta_Res];
	if(!theta_cell_type)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error!\n");
		return false;
	}

	// adjust pointers
	squared_Radius = data;
	sin2_theta = squared_Radius + ( 1 + Radius_Res ) ;
	cos2_theta = sin2_theta + Theta_Res_p1 ;
	tan2_theta = cos2_theta + Theta_Res_p1 ;
	tan_Phi = tan2_theta + Theta_Res_p1 ;

	// phi
	double x, y, *ig, *eg;
	if(!internalX[phi].setResolution(Phi_Res))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating longitude grid!\n");
		return false;
	}
	ig=internalX[phi].getGridPtr();
	eg=externalX[phi].getGridPtr();
	for(i=0;i<=Phi_Res;i++)
	{
		x=Rad(*eg++);
		*ig++=x;
		tan_Phi[i]=tan(x);
	}
	if(!internalX[phi].checkGrid())
		return false;

	// theta
	if(!internalX[theta].setResolution(Theta_Res))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating colatitude grid!\n");
		return false;
	}
	ig=internalX[theta].getGridPtr();
	eg=externalX[theta].getGridPtr();
	for(i=0;i<=Theta_Res;i++)
	{
		x=Rad(*eg++);
		*ig++=x;

		y = sin ( x ) ;
		sin2_theta [ i ] = y * y ;

		y = cos ( x ) ;
		cos2_theta [ i ] = y * y ;

		y = tan ( x ) ;
		tan2_theta [ i ] = y * y ;

		if(ISNAN(y))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"90° not allowed in colatitude grid!\n");
			return false;
		}
	}
	if(!internalX[theta].checkGrid())
		return false;

	ig=internalX[theta].getGridPtr();

	// determine theta_cell_type: 
	// -2	: northern hemisphere
	// -1	: northern hemisphere adjacent to equator
	// 0	: equatorial
	// +1	: southern hemisphere adjacent to equatorial
	// +2	: southern hemisphere
	for(i=0;i<Theta_Res;i++)
	{
		if(externalX[theta][i]<90.0)
		{
			if(externalX[theta][i+1]<90.0)
				theta_cell_type[i]=-2;
			else if(externalX[theta][i+1]==90.0)
				theta_cell_type[i]=-1;
			else
				theta_cell_type[i]=0;
		}
		else if(externalX[theta][i]==90.0)
		{
			if(externalX[theta][i+1]>90.0)
				theta_cell_type[i]=1;
			else
				errorIn(__FILE__, __LINE__, __func__);
		}
		else
		{
			if(externalX[theta][i+1]>90.0)
				theta_cell_type[i]=2;
			else
				errorIn(__FILE__, __LINE__, __func__);
		}
	}

	// radius + ground leve
	if(!internalX[radius].setResolution(Radius_Res))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating radius grid!\n");
		return false;
	}
	ig=internalX[radius].getGridPtr();
	eg=externalX[radius].getGridPtr();
	for(i=0;i<=Radius_Res;i++)
	{
		x=*eg+++getPlanetRadius();
		*ig++=x;
		squared_Radius[i]=x*x;
	}
	if(!internalX[radius].checkGrid())
		return false;
	
	return calculateGroundLayerIndex();
}

bool Spherical3DGrid::contains(class Vector3D& point)
{
	double radius, theta, phi;
	point.toPolar(&radius, &theta, &phi);
	return radius>=ground_radius && radius<=internalX[SphericalGrid::radius].getB() \
		&& theta>=internalX[SphericalGrid::theta].getA() && theta<=internalX[SphericalGrid::theta].getB() \
		&& phi>=internalX[SphericalGrid::phi].getA() && phi<=internalX[SphericalGrid::phi].getB();
}


#define THETA_INTERSECTION_NO_SOLUTION_CODE \
{\
	fprintf(stderr, "theta intersection no solution\n");\
	return false;\
}

#define THETA_INTERSECTION_NO_CODE {;}

static inline void inconsistency_error(const char *coordinate, const char *file, int line)
{
	fprintf(stderr,"file \"%s\", line %d: %s cell index - %s inconsistency!\n", file, line, coordinate, coordinate);
}

#define THETA_EPS		1e-9

bool Spherical3DGrid::calculateThetaConeIntersection(int index, double *pSolution)
{
	double ss, cc, a, p, q, radicand, sqrt_radicand, solution, sz;
	ss = sin2_theta [ index ];
	cc = cos2_theta [ index ];
	// c = c_cone [ index ] ; // c = -tan^2(theta[index])
	a = cc * kk_xy - ss * kk_z;
	if(a==0.0)
		return false;
	else
	{
		p = ( cc * pk_xy - ss * pk_z ) / a;
		q = ( cc * pp_xy - ss * pp_z ) / a;
		radicand = p * p - q ;
		if( radicand > 0.0)
		{
			bool cone_has_positive_z = internalX [ theta ] . getGridPtr() [ index ] <= 0.5 * M_PI ? true : false ;
			sqrt_radicand = sqrt ( radicand );
			solution = - p - sqrt_radicand;
			if ( solution > THETA_EPS )
			{
				sz = pz + solution * kz ;
				if( cone_has_positive_z ? ( sz > 0.0 ) : ( sz < 0.0 ) )
				{
					*pSolution = solution;
					return true;
				}
			}
			else
			{
				solution = - p + sqrt_radicand;
				if ( solution > THETA_EPS )
				{
					sz = pz + solution * kz ;
					if ( cone_has_positive_z ? ( sz > 0.0 ) : ( sz < 0.0 ) )
					{
						*pSolution = solution;
						return true;
					}
				}
			}
		}
	}
	return false;
}

bool Spherical3DGrid::calculateThetaConeIntersectionEquator(int index, double *pSolution)
{
	double cc, ss, a, p, q, radicand, sqrt_radicand, solution, sz;
	ss = sin2_theta [ index ];
	cc = cos2_theta [ index ];
	// c = c_cone [ index ] ; // c = -tan^2(theta[index])
	a = cc * kk_xy - ss * kk_z;
	if ( a == 0.0 )
		return false;
	else
	{
		p = ( cc * pk_xy - ss * pk_z ) / a;
		q = ( cc * pp_xy - ss * pp_z ) / a;
		radicand = p * p - q ;
		if ( radicand > 0.0 )
		{
			bool cone_has_positive_z = internalX [ theta ] . getGridPtr() [ index ] <= 0.5 * M_PI ? true : false ;
			sqrt_radicand = sqrt ( radicand );
			// -------------------------------------------------------------------
			solution = - p - sqrt_radicand;
			if ( solution > THETA_EPS )
			{
				sz = pz + solution * kz ;
				if( cone_has_positive_z ? sz > 0.0 : sz < 0.0 )
				{
					*pSolution = solution;
					return true;
				}
				else
				{
					solution = - p + sqrt_radicand;
					if ( solution > THETA_EPS )
					{
						sz = pz + solution * kz ;
						if( cone_has_positive_z ? sz > 0.0 : sz < 0.0 )
						{
							*pSolution = solution;
							return true;
						}
					}
				}
			}
			else
			{
				solution = - p + sqrt_radicand;
				if ( solution > THETA_EPS )
				{
					sz = pz + solution * kz ;
					if( cone_has_positive_z ? sz > 0.0 : sz < 0.0 )
					{
						*pSolution = solution;
						return true;
					}
				}
			}
		}
	}
	return false;
}

#define THETA_UPDATE_SOLUTION(DIRECTION_FLAG)\
if(solution<lambda_min)\
{\
	lambda_min=solution;\
	positive_delta=(DIRECTION_FLAG);\
	edge=EDGE_THETA;\
}



bool Spherical3DGrid::calculateCellIntersectionPoint(class Vector3D& Position, struct CellIndex_s *pCI,\
	 class Vector3D& Direction, double *pDistance, class Vector3D *pNextPosition,\
		struct CellIndex_s *pNextCI)
{
#define EDGE_RADIUS	1
#define EDGE_THETA	2
#define EDGE_PHI	3

	// data #####################################################

	px = Position[0];
	py = Position[1];
	pz = Position[2];

	kx = Direction[0];
	ky = Direction[1];
	kz = Direction[2];

	// helpers #################################################
	kk_xy = kx * kx + ky * ky ;
	kk_z = kz * kz ;

	pk_xy = px * kx + py * ky ;
	pk_z = pz * kz ;
	pk = pk_xy + pk_z ;

	pp_x = px * px ;
	pp_y = py * py ;
	pp_z = pz * pz ;

	pp_xy = pp_x + pp_y ;

	pp = pp_xy + pp_z ;

	pp_xy_pp = pp_xy * pp ;
	sqrt_pp_xy_pp = sqrt ( pp_xy_pp ) ;

	// checks ##################################################
	unsigned int edge=EDGE_RADIUS;
	bool positive_delta;

	int iPhi=pCI->coordinates.spherical.iPhi;
	int iTheta=pCI->coordinates.spherical.iTheta;
	int iRadius=pCI->coordinates.spherical.iR;

	// double radicand, root,  p, q, a, t;
	double p, q, a, t, radicand, solution, lambda_min;

	// check r=const layers ++++++++++++++++++++++++++++++++++++
	p = pk ;
	if ( pk > 0.0 )	// photon travels upwards
	{
		q = pp - squared_Radius [ iRadius + 1 ] ;	// test layer above first!
		radicand = p * p - q ;
		if( radicand < 0.0 )	// no intersection at all
		{
			inconsistency_error("radius", __FILE__, __LINE__);
			return false;
		}
		lambda_min = - p + sqrt ( radicand ) ;
		positive_delta = true ;
	}
	else // down
	{
		q = pp - squared_Radius [ iRadius ] ;	// test layer below
		radicand = p * p - q ;
		if ( radicand < 0.0 )	// no intersection at all
		{
			// no intersection with the bottom shell -> test intersection with the layer above!
			q = pp - squared_Radius[iRadius+1];	// test layer above 
			radicand = p * p - q ;
			if(radicand<0.0)	// no intersection at all
			{
				inconsistency_error("radius", __FILE__, __LINE__);
				return false;
			}
			lambda_min=-p+sqrt(radicand);
			positive_delta=true;
		}  
		else
		{
			lambda_min=-p-sqrt(radicand);
			positive_delta=false;
		}
	}

	// theta=const. intersections ++++++++++++++++++++++++++++++++++++++
	if(!b_s3d_one_colatitude_cell)
	{
		double mu_theta;
		mu_theta = ( ( kx * px + ky * py ) * pz ) / sqrt_pp_xy_pp ;
		mu_theta -= kz * sqrt ( pp_xy / pp ) ;
		switch ( theta_cell_type [ iTheta ] )
		{
			case -2:	// northern hemisphere
				if ( mu_theta > 0.0 )
				{
					if(calculateThetaConeIntersection(iTheta+1, &solution))
					{
						THETA_UPDATE_SOLUTION(true)
					}
				}
				else
				{
					if(calculateThetaConeIntersection(iTheta, &solution))
					{
						THETA_UPDATE_SOLUTION(false)
					}
					else if(calculateThetaConeIntersection(iTheta+1, &solution))
					{
						THETA_UPDATE_SOLUTION(true)
					}
				}
				break;
			case -1:	// northern hemisphere adjacent to equator
				if(mu_theta>0.0) 	// travelling to south
				{
					if(kz)
					{
						solution=-pz/kz;
						if(solution>0.0)
							THETA_UPDATE_SOLUTION(true)
					}
				}
				else	// travelling no north
				{
					if(calculateThetaConeIntersection(iTheta, &solution))
					{
						THETA_UPDATE_SOLUTION(false)
					}
				}
				break;
			
			case  0:	// equatorial
				//if(mu_theta<0.0)
				if(pz>0.0)// northern hemisphere
				{
					if(mu_theta<=0.0) // travelling northwards
					{
						if(calculateThetaConeIntersection(iTheta, &solution))
						{
							THETA_UPDATE_SOLUTION(false)
						}
					}
					else
					{
						if(calculateThetaConeIntersectionEquator(iTheta+1, &solution))
						{
							THETA_UPDATE_SOLUTION(true)
						}
					}
				}
				else if(pz==0.0)
				{
					// ...
					if(kz>0.0) // travelling northwards
					{
						if(calculateThetaConeIntersection(iTheta, &solution))
						{
							THETA_UPDATE_SOLUTION(false)
						}
					}
					else
					{
						if(calculateThetaConeIntersection(iTheta+1, &solution))
						{
							THETA_UPDATE_SOLUTION(true)
						}
					}
				}
				else // p in southern hemisphere
				{
					if(mu_theta>0.0) // travelling southwards
					{
						if(calculateThetaConeIntersection(iTheta+1, &solution))
						{
							THETA_UPDATE_SOLUTION(true)
						}
					}
					else
					{
						if(calculateThetaConeIntersectionEquator(iTheta, &solution))
						{
							THETA_UPDATE_SOLUTION(false)
						}
					}
				}
				/*if(kz<0.0)
				{
					if(calculateThetaConeIntersection(iTheta, &solution))
					{
						THETA_UPDATE_SOLUTION(false)
					}
				}
				else
				{
 					if(calculateThetaConeIntersection(iTheta+1, &solution))
					{
						THETA_UPDATE_SOLUTION(true)
					}
				}*/
				break;

			case  1:	// southern hemisphere adjacent to equatorial
				if(mu_theta>0.0)	// travelling to south
				{
					if(calculateThetaConeIntersection(iTheta+1, &solution))
					{
						THETA_UPDATE_SOLUTION(true)
					}
				}
				else
				{
					if(kz)
					{
						solution=-pz/kz;
						if(solution>0.0)
							THETA_UPDATE_SOLUTION(false)
					}
				}
				break;
			case  2:	// southern hemisphere
				if(mu_theta>0.0)
				{
					if(calculateThetaConeIntersection(iTheta+1, &solution))
					{
						THETA_UPDATE_SOLUTION(true)
					}
					else
					{
						if(calculateThetaConeIntersection(iTheta, &solution))
						{
							THETA_UPDATE_SOLUTION(false)
						}
					}
				}
				else
				{
					if(calculateThetaConeIntersection(iTheta, &solution))
					{
						THETA_UPDATE_SOLUTION(false)
					}
				}
				break;
		}
	}

	// check phi=const. +++++++++++++++++++++++++++++++++++++++++++
	if(!b_s3d_one_longitude_cell)
	{
		double mu_phi;
		mu_phi=-(kx*py)/sqrt_pp_xy_pp;
		mu_phi+=(ky*px)/sqrt_pp_xy_pp;

		// go
		if(mu_phi>0.0)
		{
			t=tan_Phi[iPhi+1];
			a=kx*t-ky;
			if(a==0.0)
			{
				inconsistency_error("phi",__FILE__,__LINE__);
				return false;
			}
			solution=(py-px*t)/a;
			if(solution>0.0 && solution<lambda_min)
			{
				lambda_min=solution;
				positive_delta=true;
				edge=EDGE_PHI;
			}
		}
		else
		{
			t=tan_Phi[iPhi];
			a=kx*t-ky;
			if(a==0.0)
			{
				inconsistency_error("phi",__FILE__,__LINE__);
				return false;
			}
			solution=(py-px*t)/a;
			if(solution>0.0 && solution<lambda_min)
			{
				lambda_min=solution;
				positive_delta=false;
				edge=EDGE_PHI;
			}
		}
	}

	// update voxel index
	*pNextCI=*pCI;
	switch(edge)
	{
		case EDGE_RADIUS:
			if(positive_delta)
			{
				pNextCI->coordinates.spherical.iR++;
				if(pNextCI->coordinates.spherical.iR>=Radius_Res)
					pNextCI->CellPosition=CELL_POSITION_ABOVE_TOA;
			}
			else
			{
				pNextCI->coordinates.spherical.iR--;
				if(pNextCI->coordinates.spherical.iR<ground_layer_index)
					pNextCI->CellPosition=CELL_POSITION_SURFACE;
			}
			pNextCI->event=TE_REFRACTION_Z;
			// lattice_grid_flag=LGZ;
			break;
		case EDGE_THETA:
			if(positive_delta)
			{
				pNextCI->coordinates.spherical.iTheta++;
				if(pNextCI->coordinates.spherical.iTheta>=Theta_Res)
				{
					if(b_cyclic_y)
						pNextCI->coordinates.spherical.iTheta=Theta_Res-1;
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_Y_B;
				}
			}
			else
			{
				pNextCI->coordinates.spherical.iTheta--;
				if(pNextCI->coordinates.spherical.iTheta<0)
				{
					if(b_cyclic_y)
						pNextCI->coordinates.spherical.iTheta=0;
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_Y_A;
				}
			}
			pNextCI->event=TE_REFRACTION_Y;
			// lattice_grid_flag=LGY;
			break;
		case EDGE_PHI:
			if(positive_delta)
			{
				pNextCI->coordinates.spherical.iPhi++;
				if(pNextCI->coordinates.spherical.iPhi>=Phi_Res)
				{
					if(b_cyclic_x)
						pNextCI->coordinates.spherical.iPhi=0;
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_X_B;
				}
			}
			else
			{
				pNextCI->coordinates.spherical.iPhi--;
				if(pNextCI->coordinates.spherical.iPhi<0)
				{
					if(b_cyclic_x)
						pNextCI->coordinates.spherical.iPhi=Phi_Res-1;
					else
						pNextCI->CellPosition=CELL_POSITION_OFF_X_A;
				}
			}
			pNextCI->event=TE_REFRACTION_X;
			// lattice_grid_flag=LGX;
			break;
		default:
			errorIn(__FILE__, __LINE__, __func__);
			return false;
	}

	if(pSurfaceGrid->useReliefData() && pSurfaceGrid->isReliefCell(pCI))
	{

		// there is maybe a surface relief intersection....
		double distance;
		class Vector3D NextPosition;
		pSurfaceGrid->calculateReliefIntersectionPoint(pCI, Position, Direction, &distance, &NextPosition);
		if(distance<lambda_min)
		{
			*pNextPosition=NextPosition;
			*pDistance=distance;
			*pNextCI=*pCI;
			pNextCI->CellPosition=CELL_POSITION_SURFACE;
			return true;
		}
	}

	*pNextPosition=Position+lambda_min*Direction;
	*pDistance=lambda_min;
	if(pNextCI->CellPosition==CELL_POSITION_SURFACE)
		pSurfaceGrid->adjustSurfacePointAndNormal(*pNextPosition);
	if(pNextCI->CellPosition==CELL_POSITION_INSIDE_ATMOSPHERE)
		if(pNextCI->coordinates.spherical.iR<0 || pNextCI->coordinates.spherical.iR>=Radius_Res
			|| pNextCI->coordinates.spherical.iTheta<0 || pNextCI->coordinates.spherical.iTheta>=Theta_Res
			|| pNextCI->coordinates.spherical.iPhi<0 || pNextCI->coordinates.spherical.iPhi>=Phi_Res)
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
	return true;
}

bool Spherical3DGrid::calculateCellIndex(class Vector3D& point, struct CellIndex_s *pCI)
{
	double radius, theta, phi;
	point.toPolar(&radius, &theta, &phi);

	if(radius<ground_radius)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(radius>internalX[SphericalGrid::radius].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	if(phi<internalX[SphericalGrid::phi].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_A;
		return false;
	}
	else if(phi>internalX[SphericalGrid::phi].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_B;
		return false;
	}
	if(theta<internalX[SphericalGrid::theta].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	else if(theta>internalX[SphericalGrid::theta].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	if(!(internalX[SphericalGrid::radius].findIntervallIndexFastButUnsafe(radius, &(pCI->coordinates.spherical.iR)) \
		&& internalX[SphericalGrid::theta].findIntervallIndexFastButUnsafe(theta, &(pCI->coordinates.spherical.iTheta)) \
		&& internalX[SphericalGrid::phi].findIntervallIndexFastButUnsafe(phi, &(pCI->coordinates.spherical.iPhi))))
		return false;
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

bool Spherical3DGrid::calculateCellIndex(class Vector3D& point, class Vector3D& direction, struct CellIndex_s *pCI)
{
	double radius, theta, phi;
	point.toPolar(&radius, &theta, &phi);
	if(radius<ground_radius)
	{
		pCI->CellPosition=CELL_POSITION_SURFACE;
		return false;
	}
	else if(radius>internalX[SphericalGrid::radius].getB())
	{
		pCI->CellPosition=CELL_POSITION_ABOVE_TOA;
		return false;
	}
	if(phi<internalX[SphericalGrid::phi].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_A;
		return false;
	}
	else if(phi>internalX[SphericalGrid::phi].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_X_B;
		return false;
	}
	if(theta<internalX[SphericalGrid::theta].getA())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	else if(theta>internalX[SphericalGrid::theta].getB())
	{
		pCI->CellPosition=CELL_POSITION_OFF_Y_A;
		return false;
	}
	int iR=pCI->coordinates.spherical.iR;
	int iTheta=pCI->coordinates.spherical.iTheta;
	int iPhi=pCI->coordinates.spherical.iPhi;
	if(!(internalX[SphericalGrid::radius].findIntervallIndexFastButUnsafe(radius, &iR) \
		&& internalX[SphericalGrid::theta].findIntervallIndexFastButUnsafe(theta, &iTheta) \
		&& internalX[SphericalGrid::phi].findIntervallIndexFastButUnsafe(phi, &iPhi)))
		return false;
	
	// adjust if necessary
	if(internalX[SphericalGrid::radius][iR+1]==radius && Vector3D::createDRVector(theta, phi) * direction  > 0.0)
		iR++;
	if(internalX[SphericalGrid::theta][iTheta+1]==theta && Vector3D::createDThetaVector(theta, phi) * direction  > 0.0)
		iTheta++;
	if(internalX[SphericalGrid::phi][iPhi+1]==phi && Vector3D::createDPhiVector(theta, phi) * direction  > 0.0)
		iPhi++;

	pCI->coordinates.spherical.iR=iR;
	pCI->coordinates.spherical.iTheta=iTheta;
	pCI->coordinates.spherical.iPhi=iPhi;
	
	pCI->CellPosition=CELL_POSITION_INSIDE_ATMOSPHERE;
	return true;
}

void Spherical3DGrid::initializeBoxSamplingStructure(class MacroCell *pMacroCell, struct BoxDetector::SamplingStructure_s *pSamplingStructure)
{
	// phi / x
	double phiA=pMacroCell->internal_xA(0);
	double phiB=pMacroCell->internal_xB(0);
	pSamplingStructure->x1=phiA;
	pSamplingStructure->x2=phiB;
	pSamplingStructure->Nx=phiB-phiA;
	pSamplingStructure->cx=phiA;
	pSamplingStructure->dx=M_PI*1e-10;

	// theta / y
	double thetaA=pMacroCell->internal_xA(1);
	double thetaB=pMacroCell->internal_xB(1);
	double sin_thetaA=sin(thetaA), cos_thetaA=cos(thetaA);
	double sin_thetaB=sin(thetaB), cos_thetaB=cos(thetaB);
	pSamplingStructure->y1=thetaA;
	pSamplingStructure->y2=thetaB;
	pSamplingStructure->Ny=cos_thetaB-cos_thetaA;
	pSamplingStructure->cy=cos_thetaA;
	pSamplingStructure->dy=M_PI*1e-10;

	// radius / z
	double rA=pMacroCell->internal_xA(2);
	double rB=pMacroCell->internal_xB(2);
	pSamplingStructure->z1=rA;
	pSamplingStructure->z2=rB;
	pSamplingStructure->Nz=rB*rB*rB-rA*rA*rA;
	pSamplingStructure->cz=rA*rA*rA;
	pSamplingStructure->dx=1e-6;

	// sampling functions
	pSamplingStructure->takeXSample=&SamplePhiCoordinate;
	pSamplingStructure->takeYSample=&SampleThetaCoordinate;
	pSamplingStructure->takeZSample=&SampleRadiusCoordinate;

	// volumes / areas
	pSamplingStructure->size[(int)BoxDetector::fullbox]=((cos_thetaB-cos_thetaA)*(rB*rB*rB-rA*rA*rA)*(phiB-phiA))/3.0;

	pSamplingStructure->size[(int)BoxDetector::boxtopface]=4.0*M_PI*rB*rB;
	pSamplingStructure->size[(int)BoxDetector::boxbottomface]=4.0*M_PI*rA*rA;

	double C_r_phi=(rB*rB-rA*rA)*(phiB-phiA)/2.0;
	pSamplingStructure->size[(int)BoxDetector::boxnorthface]=sin_thetaA*C_r_phi;
	pSamplingStructure->size[(int)BoxDetector::boxsouthface]=sin_thetaB*C_r_phi;

	double C_r_theta=(rB*rB-rA*rA)*(thetaA-thetaB)/2.0;
	pSamplingStructure->size[(int)BoxDetector::boxeastface]=C_r_theta;
	pSamplingStructure->size[(int)BoxDetector::boxwestface]=C_r_theta;

	pSamplingStructure->size[(int)BoxDetector::boxface]=0.0;
	for(int i=(int)BoxDetector::boxtopface;i<=(int)BoxDetector::boxwestface;i++)
		pSamplingStructure->size[(int)BoxDetector::boxface]+=pSamplingStructure->size[i];
}

void Spherical3DGrid::printCell(FILE *fptr, struct CellIndex_s& CI)
{
	int iR=CI.coordinates.spherical.iR;
	int iTheta=CI.coordinates.spherical.iTheta;
	int iPhi=CI.coordinates.spherical.iPhi;
	class Grid1D& Radius=internalX[radius];
	class Grid1D& Theta=internalX[theta];
	class Grid1D& Phi=internalX[phi];
	fprintf(fptr, "r[%d] = [%.3lf, %.3lf], ", iR, Radius[iR], Radius[iR+1]);
	fprintf(fptr, "theta[%d] = [%.3lf, %.3lf], ", iTheta, Degree(Theta[iTheta]), Degree(Theta[iTheta+1]));
	fprintf(fptr, "phi[%d] = [%.3lf, %.3lf]\n", iPhi, Degree(Phi[iPhi]), Degree(Phi[iPhi+1]));
}


bool Spherical3DGrid::acceptGeometry(class Geometry *pGeometry)
{
	int index;
	double c=pGeometry->getLongitude();
	if(externalX[phi].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector longitude coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	c=pGeometry->getCoLatitude();
	if(externalX[theta].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector colatitude coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	c=pGeometry->getAltitude();
	if(externalX[radius].findIndex(c, &index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "detector altitude coordinate on cell boundary, geometry rejected!\n");
		return false;
	}
	return true;
}



