/**************************************************************************************************************
*
*       This file,
*
*                            /SphericalGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "SphericalGrid.h"

#include "Math/prand.h"

#include "constants.h"
#include "trigonometry.h"

#include "Math/Matrix3D.h"
#include "error.h"

bool SphericalGrid::calculateGroundLayerIndex()
{
	// calculate ground radius
	ground_radius=planet_radius+ground_level;
	/*if(ground_radius<internalX[radius].getA() || ground_radius>internalX[radius].getB())
	{
		fprintf(stderr, "ground level %lf outside the vertical grid\n", ground_level);
		return false;
	}*/

	if(!internalX[radius].findIndex(ground_radius, &ground_layer_index))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"ground level %lf not in height list!\n", ground_level);
		return false;
	}
	return true;
}

void SphericalGrid::calculateSurfaceProjection(class Vector3D& position, double *px, double *py)
{
	double r;
	position.toPolar(&r, py, px);
}


void SphericalGrid::convertEquidistantAlbedoMapCoordinates(double *pXA, double *pYA, double *pXB, double *pYB)
{
	*pXA=Rad(*pXA);
	*pYA=Rad(*pYA);
	*pXB=Rad(*pXB);
	*pYB=Rad(*pYB);
}

void SphericalGrid::convertAlbedoMapGrids(double *x, int xres, double *y, int yres)
{
	int i;
	for(i=0;i<=xres;i++)
		x[i]=Rad(x[i]);
	for(i=0;i<=yres;i++)
		y[i]=Rad(y[i]);
}

void SphericalGrid::sampleTOAPoint(class Vector3D *pPosition, class Vector3D *pDirection, class Vector3D& SunPosition)
{
	class Vector3D SunPhotonPosition, DiscPoint, direction;
	double TOARadius=internalX[radius].getB(), r, theta, phi, p, q, radicand, lambda;

	// take location sample from within the sun
	r=SUNDIAMETER*exp(log(prand())/3.0)/2;
	theta=acos(correctMu(-1.0+2.0*prand()));
	phi=2.0*M_PI*prand();
	SunPhotonPosition.fromPolar(r, theta, phi);
	SunPhotonPosition+=SunPosition;
	do
	{
		// take a sample from the "TOARadius disc" through the origin
		r=TOARadius*sqrt(prand());
		phi=2.0*M_PI*prand();
		DiscPoint=Vector3D(cos(phi)*r, sin(phi)*r, 0.0);
		
		SunPhotonPosition.toPolar(&r, &theta, &phi);
		DiscPoint*=Matrix3D(Matrix3D::ROTY, theta);
		DiscPoint*=Matrix3D(Matrix3D::ROTZ, phi);

		// connect both points to have a direction sample
		direction=norm(DiscPoint-SunPhotonPosition);

		// calculate intersection of the ray with the "TOARadius sphere" around the origin
		p=SunPhotonPosition*direction;
		q=SunPhotonPosition*SunPhotonPosition-TOARadius*TOARadius;
		radicand=p*p-q;
	}
	while(radicand<0.0 || (lambda=-p-sqrt(radicand))<0.0);
	lambda+=1.0;	// add 1 km
	*pPosition=SunPhotonPosition+lambda*direction;
	*pDirection=direction;
}

double SphericalGrid::getVerticalExtension(const struct CellIndex_s& CI)
{
	int index=CI.coordinates.spherical.iR;
	double *g=internalX[radius].getGridPtr();
	return g[index+1]-g[index];
}

void SphericalGrid::calculateEastAxis(double& phi, double& theta, double& radius, class Vector3D& n)
{
	double sin_theta=sin(theta);
	n[0]=-sin(phi)*sin_theta;
	n[1]=cos(phi)*sin_theta;
	n[2]=cos(theta);
}

void SphericalGrid::calculateSouthAxis(double& phi, double& theta, double& radius, class Vector3D& n)
{
	double cos_theta=cos(theta);
	n[0]=cos(phi)*cos_theta;
	n[1]=sin(phi)*cos_theta;
	n[2]=-sin(theta);
}

void SphericalGrid::calculateAltitudeAxis(double& phi, double& theta, double& radius, class Vector3D& n)
{
	double sin_theta=sin(theta);
	n[0]=cos(phi)*sin_theta;
	n[1]=sin(phi)*sin_theta;
	n[2]=cos(theta);
}


