/**************************************************************************************************************
*
*       This file,
*
*                            /Surface.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Surface.h"
#include "trigonometry.h"
#include "File.h"
#include "error.h"
#include "Math/Matrix3D.h"

#include "refraction.h"

Surface::Surface()
{
        IOName="surface";
        b_use_albedo_map=false;
        b_use_reflectance_spectra=false;
        b_use_specific_albedo_map_grid=false;
        b_use_equidistant_albedo_map_grid=false;
	
	refractive_index=1.333;
	b_specular_reflection_at_surface=false;
	
	wind_direction=0.0;
	wind_speed=0.0;
	b_sea_surface=false;
	wavelength=0.0;
	
	sea_surface_R_map_polar_resolution=100;
	sea_surface_R_map_azimuth_resolution=100;

	addParameter(&b_specular_reflection_at_surface, type_bool, "b_specular_reflection_at_surface|b_Fresnel_reflection", 0, "Fresnel reflection at the surface?");
	// addParameter(&refractive_index, type_double, "refractive_index", 0, "real part of the refractive index of the medium below the surface");
	addParameter(fn_refractive_index, type_file, "fn_refractive_index", 0 , "[nm re im] refractive index table for surface");
	addParameter(&wind_speed, type_double, "wind_speed", 0, "[m/s] wind speed at the sea surface");
	addParameter(&wind_direction, type_angle, "wind_direction|wind_azimuth", 0, "[degree] wind direction relative to North");
	addParameter(&b_sea_surface, type_bool, "b_sea_surface", 0, "use Cox & Munks 1954 sea surface slope PDF for simulation of specular reflection?");
	addParameter(SeaSurfaceSlopeSampler.getCrosswindMapResolutionPtr(), type_int, "crosswind_resolution", 0, "sea surface crosswind axis slope map resolution");
	addParameter(SeaSurfaceSlopeSampler.getUpwindMapResolutionPtr(), type_int, "upwind_resolution", 0, "sea surface upwind axis slope map resolution");
	// addParameter(&sea_surface_R_map_azimuth_resolution, type_int, "sea_surface_R_map_azimuth_resolution", 0, "sea surface reflection coefficient (R) map azimuth resolution");
	// addParameter(&sea_surface_R_map_polar_resolution, type_int, "sea_surface_R_map_polar_resolution", 0, "sea surface reflection coefficient (R) map polar resolution");
	addParameter(&ground_albedo, type_double, "ground_albedo|albedo", 0, "homogeneously grey surface albedo value");
	addParameter(&b_use_albedo_map, type_bool, "b_use_albedo_map", 0, "use detailed albedo map?");
	addParameter(&b_use_reflectance_spectra, type_bool, "b_use_reflectance_spectra", 0, "use reflectance spectra?");
	addParameter(fn_ground_reflectance_spectra, type_file, "fn_ground_reflectance_spectra|fn_spectral_albedo_data", 0, "file with data table of reflectance spectra");
	addParameter(fn_ground_pixel_map, type_file, "fn_ground_pixel_map", 0, "ground pixel map instead of albedo");
	addParameter(&b_use_specific_albedo_map_grid, type_bool, "b_use_specific_albedo_map_grid", 0, "use specific albedo map grid?");
	addParameter(&b_use_equidistant_albedo_map_grid, type_bool, "b_use_equidistant_albedo_map_grid", 0, "use equidistant albedo map grid?");
	addParameter(&AlbedoMapXA, type_double, "AlbedoMapXA", 0, "albedo map XA");
	addParameter(&AlbedoMapXB, type_double, "AlbedoMapXB", 0, "albedo map XB");
	addParameter(&AlbedoMapYA, type_double, "AlbedoMapYA", 0, "albedo map YA");
	addParameter(&AlbedoMapYB, type_double, "AlbedoMapYB", 0, "albedo map YB");
	addParameter(&AlbedoMapDeltaX, type_double, "AlbedoMapDeltaX", 0, "x size of albedo map tile");
	addParameter(&AlbedoMapDeltaY, type_double, "AlbedoMapDeltaY", 0, "y size of albedo map tile");
	addParameter(fn_albedo_map_x_grid, type_file, "fn_albedo_map_x_grid", 0, "x grid file for albedo map");
	addParameter(fn_albedo_map_y_grid, type_file, "fn_albedo_map_y_grid", 0, "y grid file for albedo map");
}

Surface::~Surface()
{
	/*if(reflectance_index_map)
		delete [] reflectance_index_map;
	if(albedo_map)
		delete [] albedo_map;
	if(reflectance_data)
		delete [] reflectance_data;
	if(reflectance_table)
		delete [] reflectance_table;
	*/
}

bool Surface::loadAlbedoFiles()
{
	if(!(useAlbedoMap() || useReflectanceSpectra()))
		return true;

	// File file;
	char file_name[1000];
	// read fn_ground_reflectance_spectra
	if(useReflectanceSpectra())
	{
		int i;
		formatGroundReflectanceSpectraFilename(file_name);
		if(!ReflectanceSpectraMatrix.load(file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "could not open reflectance spectra file \"%s\"!\n", file_name);
			return false;
		}
		// inspection:
		no_reflectance_spectra=ReflectanceSpectraMatrix.getN()-1;
		no_wavelengths=ReflectanceSpectraMatrix.getM();
		wavelength_index=0;
		// adjust wavelengths
		for(i=0;i<no_wavelengths;i++)
			ReflectanceSpectraMatrix[i][0]*=1e-9;
		// read in:
		if(no_reflectance_spectra>1)
		{
			// specific albedo map grid
			formatGroundPixelMapFilename(file_name);
			// load index map
			if(!ReflectanceSpectraIndexMap.load(file_name))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "could not load ground reflectance spectra index map file \"%s\"!\n", file_name);
				return false;
			}
			albedo_map_xres=ReflectanceSpectraIndexMap.getN();
			albedo_map_yres=ReflectanceSpectraIndexMap.getM();
			// check index map
			int *index=ReflectanceSpectraIndexMap.getData(), n=albedo_map_xres*albedo_map_yres;
			for(i=0;i<n;i++)
			{
				(*index)--;	
				if(*index<0 || *index >= no_reflectance_spectra)
				{
					errorIn(__FILE__, __LINE__, __func__);
					fprintf(stderr, "index out of range [1,%d] in ground reflectance spectra index map file \"%s\"!\n", no_reflectance_spectra, file_name);
					return false;
				}
				index++;
			}
		}
		// initialize wavelength cache
		if(!ReflectanceSpectraCache.allocate(2, no_reflectance_spectra))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		ReflectanceSpectraCache.setToZero();
	}
	else
	{
		no_reflectance_spectra=0;
		no_wavelengths=0;
		wavelength_index=0;
		formatGroundPixelMapFilename(file_name);
		if(!AlbedoMap.load(file_name))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "failed to read ground albedo map from file \"%s\"\n",file_name);
			return false;
		}
		albedo_map_xres=AlbedoMap.getN();
		albedo_map_yres=AlbedoMap.getM();
		// printf("AM %d x %d\n", albedo_map_xres, albedo_map_yres);
	}
	// only reflectance spectra? then return true
	if(!b_use_albedo_map)
		return true;
	// albedo map grid
	if(b_use_specific_albedo_map_grid)
	{
		if(b_use_equidistant_albedo_map_grid)
			pGrid->convertEquidistantAlbedoMapCoordinates(&AlbedoMapXA, &AlbedoMapYA, &AlbedoMapXB, &AlbedoMapYB);
		else
		{
			// x grid ..................
			formatAlbedoMapXGridFilename(file_name);
			if(!albedo_map_x_grid.load(file_name))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error loading albedo map x grid \"%s\"\n", file_name);
				return false;
			}
			if(albedo_map_xres!=albedo_map_x_grid.getN()-1)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "x grid from file \"%s\" incompatible with albedo map in \"%s\"\n", file_name, fn_ground_pixel_map);
				return false;
			}

			// y ...........................
			formatAlbedoMapYGridFilename(file_name);
			if(!albedo_map_y_grid.load(file_name))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error loading albedo map y grid \"%s\"\n", file_name);
				return false;
			}
			if(albedo_map_yres!=albedo_map_y_grid.getN()-1)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "y grid from file \"%s\" incompatible with albedo map in \"%s\"\n", file_name, fn_ground_pixel_map);
				return false;
			}
			pGrid->convertAlbedoMapGrids(albedo_map_x_grid.getCoefficientPtr(), albedo_map_xres, albedo_map_y_grid.getCoefficientPtr(), albedo_map_yres);
			AlbedoMapXA=albedo_map_x_grid[0];
			AlbedoMapXB=albedo_map_x_grid[albedo_map_xres];
			AlbedoMapYA=albedo_map_y_grid[0];
			AlbedoMapYB=albedo_map_y_grid[albedo_map_yres];
		}
	}
	else
	{
		if(albedo_map_xres!=grid_xres || albedo_map_yres!=grid_yres)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "albedo map in \"%s\" incompatible with x/longitude or y/colatitude grid files\n", fn_ground_pixel_map);
			return false;
		}
		AlbedoMapXA=pGrid->getInternalGrid1DPtr()[0].getA();
		AlbedoMapXB=pGrid->getInternalGrid1DPtr()[0].getB();
		AlbedoMapYA=pGrid->getInternalGrid1DPtr()[1].getA();
		AlbedoMapYB=pGrid->getInternalGrid1DPtr()[1].getB();
	}
	AlbedoMapDeltaY=AlbedoMapYB-AlbedoMapYA;
	AlbedoMapDeltaX=AlbedoMapXB-AlbedoMapXA;
	am_x=0;
	am_y=0;
	return true;
}

bool Surface::setWavelength(double lambda)
{
	if(wavelength==lambda)
		return true;
	if(useReflectanceSpectra())
	{
		// find wavelength intervall
		if(lambda<ReflectanceSpectraMatrix[0][0] || lambda>ReflectanceSpectraMatrix[no_wavelengths-1][0])
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "no reflectance data for wavelength %lg!\n", lambda);
			return false;
		}
		if(!ReflectanceSpectraMatrix.findRowIndex(lambda, 0, wavelength_index))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		int i;
		double a,b,f;
		a=ReflectanceSpectraMatrix[wavelength_index][0];
		b=ReflectanceSpectraMatrix[wavelength_index+1][0];
		f=(lambda-a)/(b-a);
		for(i=0;i<no_reflectance_spectra;i++)
		{
			ReflectanceSpectraCache[0][i]=lambda;
			a=ReflectanceSpectraMatrix[wavelength_index][1+i];
			b=ReflectanceSpectraMatrix[wavelength_index+1][1+i];
			ReflectanceSpectraCache[1][i]=a+f*(b-a);
		}
		
	}
	// surface reflection
	refractive_index=RefractiveIndex.evaluate(1e9*lambda).real();
	wavelength=lambda;
	return true;
}
	
bool Surface::getAlbedoMapTileIndices(const struct CellIndex_s& CI, class Vector3D& position, int& ix, int& iy)
{
	if(b_use_specific_albedo_map_grid)
	{
		double x, y;
		pGrid->calculateSurfaceProjection(position, &x, &y);
		if(y<AlbedoMapYA || y>AlbedoMapYB || x<AlbedoMapXA || x>AlbedoMapXB)
			return false;	// outside albedo map?
		if(b_use_equidistant_albedo_map_grid)
		{
			ix=int((double(albedo_map_xres)*(x-AlbedoMapXA))/AlbedoMapDeltaX)%albedo_map_xres;
			iy=int((double(albedo_map_yres)*(y-AlbedoMapYA))/AlbedoMapDeltaY)%albedo_map_yres;
		}
		else
		{
			if(!albedo_map_y_grid.findIndex(y, am_y))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			if(!albedo_map_x_grid.findIndex(x, am_x))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return false;
			}
			ix=am_x;
			iy=am_y;
		}
	}
	else
	{
		ix=CI.coordinates.general.iC[0];
		iy=CI.coordinates.general.iC[1];
	}
	return true;
}

double Surface::getAlbedo(const struct CellIndex_s& CI, class Vector3D& position, const double& wavelength)
{
	int ix, iy;
	if(b_use_reflectance_spectra)
	{
		int rs_index;
		double a, b, f;
		if(b_use_albedo_map)
		{
			if(!getAlbedoMapTileIndices(CI, position, ix, iy))
				return ground_albedo;
			rs_index=ReflectanceSpectraIndexMap[iy][ix];
		}
		else rs_index=0;
		if(ReflectanceSpectraCache[0][rs_index]==wavelength)
			return ReflectanceSpectraCache[1][rs_index];
		if(!ReflectanceSpectraMatrix.findRowIndex(wavelength, 0, wavelength_index))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		a=ReflectanceSpectraMatrix[wavelength_index][0];
		b=ReflectanceSpectraMatrix[wavelength_index+1][0];
		f=(wavelength-a)/(b-a);
		ReflectanceSpectraCache[0][rs_index]=wavelength;
		a=ReflectanceSpectraMatrix[wavelength_index][1+rs_index];
		b=ReflectanceSpectraMatrix[wavelength_index+1][1+rs_index];
		ReflectanceSpectraCache[1][rs_index]=a+f*(b-a);
		
	}
	if(b_use_albedo_map)
	{
		if(!getAlbedoMapTileIndices(CI, position, ix, iy))
			return ground_albedo;
		return AlbedoMap[iy][ix];
	}
	return ground_albedo;
}

bool Surface::loadFiles()
{
	if(!SurfaceGrid::loadFiles() || !loadAlbedoFiles())
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	if(b_specular_reflection_at_surface)
	{
		char file_path[BasicIO_string_size];
		formatFilePath(file_path, fn_refractive_index);
		if(!RefractiveIndex.readFromFile(file_path))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error loading refractive index from file \"%s\"!\n", file_path);
			return false;
		}
		if(b_sea_surface)
		{
			if(!SeaSurfaceSlopeSampler.allocate() \
				|| !SeaSurfaceSlopeSampler.setWindSpeed(wind_speed))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error setting up sea surafce slope sampler\n");
				return false;
			}
			/*if(!SeaSurfaceSlopeSampler.calculateEffectiveReflectivityMap(sea_surface_R_map_polar_resolution, sea_surface_R_map_azimuth_resolution, refractive_index))
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr, "error setting up sea surafce slope sampler\n");
				return false;
			}*/
		}
	}
	return true;
}

double Surface::calculateReflectionCoefficient(class Vector3D& p, class Vector3D& k, double *pWeight)
{
	if(!b_specular_reflection_at_surface)
		return 0.0;
	double mu_inc, R;
	class Vector3D reflection_surface_normal; //, k_relative_to_upwind;
	if(b_sea_surface)
	{
		double wind_angle, r, theta, phi, weight; 
		wind_angle = - 0.5 * M_PI - wind_direction ;
		p.toPolar(&r, &theta, &phi);
		if(wind_speed)
		{
			/*
			k_relative_to_upwind = k;

			// 1. transform k into wind system
			k_relative_to_upwind *= Matrix3D(Matrix3D::ROTZ, -phi);
			if(pGrid->isSphericalGrid())
				k_relative_to_upwind *= Matrix3D(Matrix3D::ROTY, -theta);
			k_relative_to_upwind *= Matrix3D(Matrix3D::ROTZ, -wind_angle);

			// 2. recalculate PDF considering the incident direction of the photon
			if(!SeaSurfaceSlopeSampler.recalculateSlopePDF(k_relative_to_upwind))
			{
				errorIn(__FILE__, __LINE__, __func__);
				return 0.0;
			}
			
			// 3. sample from modified PDF
			*/
			do
			{
				reflection_surface_normal = SeaSurfaceSlopeSampler.getNormalSample();
				// rotate by - 90° - wind direction
				reflection_surface_normal *= Matrix3D(Matrix3D::ROTZ, wind_angle);
				if(pGrid->isSphericalGrid())
					reflection_surface_normal *= Matrix3D(Matrix3D::ROTY, theta);
				reflection_surface_normal *= Matrix3D(Matrix3D::ROTZ, phi);
				weight = -correctMu(reflection_surface_normal*k);
			} while(weight<=0.0);
			*pWeight *= weight;
		}
		else
		{
			reflection_surface_normal = Vector3D ( 0.0, 0.0, 1.0 ) ;
			// rotate by - 90° - wind direction
			reflection_surface_normal *= Matrix3D(Matrix3D::ROTZ, wind_angle);
			if(pGrid->isSphericalGrid())
				reflection_surface_normal *= Matrix3D(Matrix3D::ROTY, theta);
			reflection_surface_normal *= Matrix3D(Matrix3D::ROTZ, phi);
		}
		// setSurfaceNormal ( reflection_surface_normal );
	}
	else
		reflection_surface_normal=getSurfaceNormal();
	mu_inc=correctMu(-reflection_surface_normal*k);
	R=calculateFresnelCoefficient(mu_inc, refractive_index);
	return R;
}











