/**************************************************************************************************************
*
*       This file,
*
*                            /Surface.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _Surface_h
#define _Surface_h

#include "SeaSurface.h"
#include "SurfaceGrid.h"
#include "Math/Matrix.h"
#include "Math/Vector.h"
#include "Math/TMatrix.h"

#include "Math/LinearelyInterpolatedFunctionTemplate.h"

class Surface: public SurfaceGrid
{
	// IO #########################################################################
	double ground_albedo;					// homogeneously grey surface albedo value

	bool b_use_reflectance_spectra;				// do we use reflectance spectra?
	bool b_use_albedo_map;					// use detailed albedo map?
	
	char fn_ground_pixel_map[BasicIO_string_size];			// instead of albedo
	char fn_ground_reflectance_spectra[BasicIO_string_size];	// file with reflectance data table

	// specific albedo map grid...
	bool b_use_specific_albedo_map_grid;	// use specific albedo map grid?
	bool b_use_equidistant_albedo_map_grid;	// use equidistant albedo map grid?

	// yes:
	double AlbedoMapXA;	// albedo map XA
	double AlbedoMapXB;	// albedo map XB
	double AlbedoMapYA;	// albedo map YA
	double AlbedoMapYB;	// albedo map YB

	double AlbedoMapDeltaX;	// x size of albedo map tile
	double AlbedoMapDeltaY; // y size of albedo map tile
	// no:
	char fn_albedo_map_x_grid[BasicIO_string_size];			// instead of albedo
	char fn_albedo_map_y_grid[BasicIO_string_size];			// instead of albedo

	// specular reflection
	bool b_specular_reflection_at_surface;
	char fn_refractive_index[BasicIO_string_size];			// refractive index file

	// sea surface
	bool b_sea_surface;
	double wind_speed;
	double wind_direction;

	int sea_surface_R_map_polar_resolution;
	int sea_surface_R_map_azimuth_resolution;

	// #############################################################################

	double wavelength;

	/*
		mainly describes the albedo properties of the surface.
		The parent class SurfaceGrid contains the description of the surface relief
	*/
	
	int no_reflectance_spectra, no_wavelengths;

	int wavelength_index;

	class TMatrix<int> ReflectanceSpectraIndexMap;
	class Matrix ReflectanceSpectraMatrix;
	class Matrix ReflectanceSpectraCache;

	// albedo map stuff
	// int default_albedo_map_index;
	double albedo1D;
	class Matrix AlbedoMap;		// used for both: grey and colored albedo
	int albedo_map_xres, albedo_map_yres;
	class Vector albedo_map_x_grid, albedo_map_y_grid;
	
	int am_x, am_y;	// position of the albedo map tile
	
	bool loadAlbedoFiles();

	bool setWavelength(double wavelength);

	// specular reflection 
	double refractive_index;
	class LinearelyInterpolatedFunctionTemplate< complex<double> > RefractiveIndex;
	class SeaSurface SeaSurfaceSlopeSampler;

protected:

	virtual bool loadFiles();

public:

	Surface();
	~Surface();
	
	inline void formatGroundPixelMapFilename(char *filepath){formatFilePath(filepath,fn_ground_pixel_map);}
	inline void formatGroundReflectanceSpectraFilename(char *filepath){formatFilePath(filepath,fn_ground_reflectance_spectra);}
	inline void formatAlbedoMapXGridFilename(char *filepath){formatFilePath(filepath,fn_albedo_map_x_grid);}
	inline void formatAlbedoMapYGridFilename(char *filepath){formatFilePath(filepath,fn_albedo_map_y_grid);}

	// flag access
	inline bool useReflectanceSpectra(){return b_use_reflectance_spectra;}
	inline bool useSpecificAlbedoMapGrid(){return b_use_specific_albedo_map_grid;}
	inline bool useAlbedoMap(){return b_use_albedo_map;}
	inline bool useEquidistantAlbedoMapGrid(){return b_use_equidistant_albedo_map_grid;}

	bool getAlbedoMapTileIndices(const struct CellIndex_s& CI, class Vector3D& position, int& ix, int& iy);

	// inline functions
	inline double getAlbedo(){return albedo1D;}
	// inline double *getAlbedoMap(){return albedo_map;}
	inline void getAlbedoMapResolution(int *pxres, int *pyres){*pxres=albedo_map_xres; *pyres=albedo_map_yres;}

	double calculateReflectionCoefficient(class Vector3D& p, class Vector3D& k, double *pWeight);
	double getAlbedo(const struct CellIndex_s& VI, class Vector3D& position, const double& wavelength);

	// inline int getDefaultAlbedoMapIndex(){return default_albedo_map_index;}

	inline int getNOReflectanceSpectra(){return no_reflectance_spectra;}

};

#endif //_Surface_h

