/**************************************************************************************************************
*
*       This file,
*
*                            /SurfaceGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "SurfaceGrid.h"
#include "File.h"
#include "Math/Matrix3D.h"

#include "Math/prand.h"
#include "trigonometry.h"
#include "error.h"

#include <float.h>

#include "Math/minmax.h"
#include "constants.h"

SurfaceGrid::SurfaceGrid()
{
	// IOName defined later
	pGrid=(class Grid *)0;

	b_use_relief_data=false;

	addParameter(&b_use_relief_data, type_bool, "b_use_relief_data|b_relief", 0, "use topograhic surface altitude data?");
	addParameter(fn_relief_data_file , type_file, "fn_relief_data_file|fn_relief", 0, "file which holds the ground relief data...");
}

bool SurfaceGrid::loadFiles()
{
	if(!b_use_relief_data)
		return true;
	if(pGrid->getGridDimension()==1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"reliefs and 1D grids are incompatible.\n");
		return false;
	}

	// try to open the file
	char relief_filename[1000];
	formatReliefDataFilename(relief_filename);
	/*File file;
	if(!file.open(relief_filename, (char *)"r"))
	{
		fprintf(stderr,"Could not open relief definition file \"%s\"!\n", relief_filename);
		return false;
	}*/

	// is the format correct?
	pGrid->getGroundMapResolution(&grid_xres, &grid_yres);
	if(!height_data.load(relief_filename))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"could not open relief definition file \"%s\"!\n", relief_filename);
		return false;
	}
	int x, xres=height_data.getN()-1;
	int y, yres=height_data.getM()-1;
	if(xres!=grid_xres || yres!=grid_yres)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"relief dimension %dx%d of file \"%s\" does not fit to grid resolution %dx%d...\n", \
				xres, yres, relief_filename, grid_xres, grid_yres);
		return false;
	}

	/*// read-in heights:
	if(!height_data.allocate(1+xres, 1+yres))
	{
		fprintf(stderr, "(%dx%d) relief height map: allocation error!\n", xres, yres);
		return false;
	}
	if(!file.readTable(yres+1, xres+1, sizeof(double), (char *)"lf", height))
	{
		fprintf(stderr, "failed to read (%dx%d) relief map from file \"%s\"\n", xres, yres, relief_filename);
		return false;
	}
	if(
	int xres=file.numberOfColumns()-1, yres=file.numberOfRows()-1;*/

	// convert height to grid heights....
	double *h=height_data.getDataPtr();
	for(y=0;y<=grid_yres;y++)
		for(x=0;x<=grid_xres;x++)
			pGrid->convertHeightToGridHeight(*h++);

	// allocate min/max cell index arrays
	if(!cell_index_data.allocate(2*xres*yres))
	/*if(cell_index_data)
		delete [] cell_index_data;
	cell_index_data=new int[2*xres*yres];
	if(!cell_index_data)*/
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr,"(2x%dx%d) relief cell indices: allocation error!\n",xres,yres);
		return false;
	}
	min_cell_index=cell_index_data.getCoefficientPtr();
	max_cell_index=min_cell_index+xres*yres;

	return calculateReliefPoints();
}

bool SurfaceGrid::calculateReliefPoints()
{
	/*
	if(pointData)
		delete [] pointData;
	pointData=new class Vector3D[(grid_xres+1)*(grid_yres+1)+grid_xres*grid_yres];
	if(!pointData)*/
	if(!point_data.allocate((grid_xres+1)*(grid_yres+1)+grid_xres*grid_yres))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "allocation error in calculateReliefPoints()!\n");
		return false;
	}
	middlePoint=point_data.getCoefficientPtr();
	gridPoint=middlePoint+grid_xres*grid_yres;
	maximal_cell_index=0;

	// get x and y grid from pGrid:
	class Grid1D *grid1D=pGrid->getInternalGrid1DPtr();
	double *grid_x=grid1D[0].getGridPtr();
	double *grid_y=grid1D[1].getGridPtr();
	double *grid_z=grid1D[2].getGridPtr();
	// data for both GridType_e s
	class Vector3D P[5];
	double h[5], ax, ay, hmax, hmin, *height=height_data.getDataPtr();
	int x, y, xres=grid_xres+1, grid_zres=grid1D[2].getResolution(), ih, index;
	for(y=0;y<grid_yres;y++)
		for(x=0;x<grid_xres;x++)
		{
			index=y*xres+x;
			h[4]=h[0]=height[index];
			pGrid->convertToVector3D(grid_x[x], grid_y[y], h[0], P[0]);
			gridPoint[index]=P[0];
			
			index=y*xres+x+1;
			h[4]+=h[1]=height[index];
			pGrid->convertToVector3D(grid_x[x+1], grid_y[y], h[1], P[1]);
			gridPoint[index]=P[1];
			
			index=(y+1)*xres+x+1;
			h[4]+=h[2]=height[index];
			pGrid->convertToVector3D(grid_x[x+1], grid_y[y+1], h[2], P[2]);
			gridPoint[index]=P[2];
			
			index=(y+1)*xres+x;
			h[4]+=h[3]=height[index];
			pGrid->convertToVector3D(grid_x[x], grid_y[y+1], h[3], P[3]);
			gridPoint[index]=P[3];
			
			h[4]*=0.25;

			ax=0.5*(grid_x[x]+grid_x[x+1]);
			ay=0.5*(grid_y[y]+grid_y[y+1]);

			pGrid->convertToVector3D(ax, ay, h[4], P[4]);
			middlePoint[index=y*grid_xres+x]=P[4];

			// determine minimal and maximal height of the surface grid cell
			hmax=max(max(h[0],h[1]),max(h[2],h[3]));
			hmin=min(min(h[0],h[1]),min(h[2],h[3]));

			if(pGrid->isSphericalGrid())
			{
				// make consistence/validity check for spherical grids: calculate 
				// the four extremal point, being closest to the earth center and
				// calculate the height of the middle point. In the following section
				// it is checked, whether this point is under the surface.
				double mu_x, mu_y, re, a, b, c, d, e, det;
				class Vector3D E0=P[4], Ex, Ey, Pe[4];				
				for(int i=0;i<4;i++)
				{
					Ex=P[i]-E0;
					Ey=P[(i+1)%4]-E0;
					a=Ex*Ex;
					b=Ey*Ey;
					c=Ex*Ey;
					d=E0*Ex;
					e=E0*Ey;
					det=a*b-c*c;
					if(det==0.0) // will not happen
					{
						errorIn(__FILE__, __LINE__, __func__);
						fprintf(stderr,"relief, minimal point calculation: determinant is zero!\n");
						return false;
					}
					mu_x=(c*e-b*d)/det;
					mu_y=(c*d-a*e)/det;
					Pe[i]=E0+mu_x*Ex+mu_y*Ey;
					re=sqrt(Pe[i]*Pe[i]);
					if(re<hmin)
						hmin=re;
				}
			}
			if(hmin<pGrid->getGroundLevel())
			{
				min_cell_index[index]=0;
				ih=0;
			}
			else
				for(ih=1;ih<=grid_zres;ih++)
					if(grid_z[ih]>hmin)
					{
						min_cell_index[index]=ih-1;
						break;
					}
			if(ih==grid_zres)
			{
				errorIn(__FILE__, __LINE__, __func__);
				fprintf(stderr,"assertion failed!\n");
				fprintf(stderr,"h_min=%fkm\th_max=%fkm\n",hmin-pGrid->getGroundLevel(),hmax-pGrid->getGroundLevel());
				fprintf(stderr,"check row %d and column %d of the relief file\n",x+1,y+1);
				return false;
			}

			for(;ih<grid_zres;ih++)
				if(grid_z[ih]>hmax)
				{
					max_cell_index[index]=ih;
					if(ih>maximal_cell_index)
						maximal_cell_index=ih;
					break;
				}
		}
	return true;
}

bool SurfaceGrid::isReliefCell(struct CellIndex_s *pCI)
{
	int iZ=pCI->coordinates.general.iC[CELL_INDEX_HEIGHT];
	int local_maximal_cell_index=pCI->coordinates.general.iC[CELL_INDEX_Y]*grid_xres+pCI->coordinates.general.iC[CELL_INDEX_X];
	if(iZ>maximal_cell_index || iZ>local_maximal_cell_index)
		return false;	// above the maximum cell index
	return true;
}

#define RELIEF_MIN_ZERO		0.0
#define RELIEF_MAX_ONE		1.0
#define RELIEF_FACE_EPS		1e-8

void SurfaceGrid::calculateReliefIntersectionPoint(struct CellIndex_s *pCI, class Vector3D& position, class Vector3D& direction, double *pDistance, class Vector3D *pNextPosition)
{
	int ix=pCI->coordinates.general.iC[CELL_INDEX_X],
		iy=pCI->coordinates.general.iC[CELL_INDEX_Y],
		//iz=pCI->coordinates.general.iC[CELL_INDEX_HEIGHT],
		xres=grid_xres+1,i,imin=-1;
	class Vector3D *pCornerPoint[4], E0, D, v, r[8], *pr1=r, *pr2=r+4;
	double sd, s1, s2, distance_minimum=DBL_MAX;
	class Matrix3D m;

	pCornerPoint[0]=gridPoint+iy*xres+ix;
	pCornerPoint[1]=pCornerPoint[0]+xres;
	pCornerPoint[2]=pCornerPoint[1]+1;
	pCornerPoint[3]=pCornerPoint[2]-xres;
	E0=middlePoint[iy*grid_xres+ix];
	D=E0-position;
	for(i=0;i<4;i++)
	{
		*pr1=E0-*(pCornerPoint[i]);
		*pr2=E0-*(pCornerPoint[(i+1)%4]);
		m=Matrix3D(direction, *pr1++, *pr2++).inverse();
		v=m*D;
		v.toCartesic(&sd,&s1,&s2);
		if(sd>=0.0)
		{
			//found_at_least_one_face=true;
			if(sd<distance_minimum && s1>=RELIEF_MIN_ZERO && s2>=RELIEF_MIN_ZERO && s1+s2<=RELIEF_MAX_ONE)
			{
				// check if exitpoint is in the correct quarter
				distance_minimum=sd;
				imin=i;
			}
		}
	}
	if(imin<0)
	{
		*pDistance=DBL_MAX;
		return;
	}
	// update raytracing variables
	surface_normal=norm(r[4+imin].cross(r[imin]));
	*pDistance=distance_minimum;
	*pNextPosition=position+direction*distance_minimum+surface_normal*RELIEF_FACE_EPS;
	surface_point=*pNextPosition;
}
/*
void SurfaceGrid::sampleLambertianScatterDirection(double *pTheta, double *pPhi, class Vector3D& direction)
{
	// Lambert Reflector: sample phi_s and mu_s
	double phi_s=prand()*(2.0*M_PI);
	double mu_s=correctMu(prand());
	double sin_theta_s=sqrt(1.0-mu_s*mu_s);

	// construct scattered vector in z-direction
	double x=cos(phi_s)*sin_theta_s;
	double y=sin(phi_s)*sin_theta_s;
	double z=mu_s;

	double theta, phi, rx, ry, rz, sina, cosa;
	surface_normal.toPolar(&theta,&phi);

	*pTheta=theta;
	*pPhi=phi;

	// rotate it back: theta

	//	a[0][0]=cos(val);	a[0][1]=0.0;		a[0][2]=sin(val); 
	//	a[1][0]=0.0;		a[1][1]=1.0;		a[1][2]=0.0; 
	//	a[2][0]=-sin(val);	a[2][1]=0.0;		a[2][2]=cos(val); 		

	sina=sin(theta);
	cosa=cos(theta);

	rx=x*cosa+z*sina;
	rz=-x*sina+z*cosa;
	// update x and z
	x=rx;
	z=rz;

	// rotate it back phi

	//	a[0][0]=cos(val);	a[0][1]=-sin(val);	a[0][2]=0.0; 
	//	a[1][0]=sin(val);	a[1][1]=cos(val);	a[1][2]=0.0; 
	//	a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=1.0; 					

	sina=sin(phi);
	cosa=cos(phi);

	rx=x*cosa-y*sina;
	ry=x*sina+y*cosa;
	
	direction[0]=rx;
	direction[1]=ry;
	direction[2]=rz;
}*/

void SurfaceGrid::getReliefData(int *pxres, int *pyres, class Vector3D **pGridPoint, class Vector3D **pMiddlePoint)
{
	*pxres=grid_xres;
	*pyres=grid_yres;
	*pGridPoint=gridPoint;
	*pMiddlePoint=middlePoint;
}

void SurfaceGrid::calculateSpecularlyReflectedDirection(class Vector3D& face_normal, class Vector3D& direction)
{
	direction-=2.0*(face_normal*direction)*face_normal;
}

