/**************************************************************************************************************
*
*       This file,
*
*                            /SurfaceGrid.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _SurfaceGrid_h
#define _SurfaceGrid_h

#include "Grid.h"

#include "Math/Vector3D.h"
#include "Math/TVector.h"
#include "Math/Matrix.h"

#include "cell.h"

#include "BasicIO.h"

#include "coordinate_sampling.h"

#include "SimulationObject.h"

class SurfaceGrid: public BasicIO, public SimulationObject
{

	// IO section ##########################################################################################
	// surface grid
	char fn_relief_data_file[BasicIO_string_size];		// file which holds the ground profile data...
	bool b_use_relief_data;					// indicates if the lattice file should contain relief data

	// ##########################################################################################

	// information concerning cells
	class TVector<int> cell_index_data;
	int *max_cell_index;		// maximal cell index
	int *min_cell_index;		// minimal cell index	
	int maximal_cell_index;

	// relief points
	class TVector<Vector3D> point_data;	// point data
	class Vector3D *gridPoint;	// the points on the grid
	class Vector3D *middlePoint;	// points in the mesh middles

	// internal variable for the relief intersection point 
	// and its surface normal:
	class Vector3D surface_normal, surface_point;

protected:

	int grid_xres, grid_yres;	// resolution of the relief map
	class Matrix height_data;

	virtual bool loadFiles();
	bool calculateReliefPoints();

	// class Grid *pGrid;

public:

	SurfaceGrid();
	
	/*inline void setGridPtr(class Grid *GridPtr){pGrid=GridPtr;}
	inline class Grid *getGridPtr(){return pGrid;}*/

	bool isReliefCell(struct CellIndex_s *pCI);
	void calculateReliefIntersectionPoint(struct CellIndex_s *pCI, class Vector3D& position, class Vector3D& direction, double *pDistance, class Vector3D *pNextPosition);
	
	inline void defineSurfacePointAndNormal(class Vector3D& point, class Vector3D& normal){surface_normal=normal; surface_point=point;}
	inline void adjustSurfacePointAndNormal(class Vector3D& point){surface_normal=point.getUnityVector(); surface_point=point;}	// use with spherical grids only!

	inline void setSurfaceNormal(class Vector3D& normal){surface_normal=normal;}
	inline class Vector3D& getSurfaceNormal(){return surface_normal;}
	inline class Vector3D& getSurfacePoint(){return surface_point;}

	inline bool useReliefData(){return b_use_relief_data;}

	// void sampleLambertianScatterDirection(double *pTheta, double *pPhi, class Vector3D& direction);
	inline void sampleLambertianScatterDirection(class Vector3D& direction)
		{return sampleLambertianDirection(surface_normal, direction);}

	static inline void sampleLambertianScatterDirection(class Vector3D& face_normal, class Vector3D& direction)
		{return sampleLambertianDirection(face_normal, direction);}
	static void calculateSpecularlyReflectedDirection(class Vector3D& face_normal, class Vector3D& direction);

	void getReliefData(int *pxres, int *pyres, class Vector3D **pGridPoint, class Vector3D **pMiddlePoint);

	inline void formatReliefDataFilename(char *filepath){formatFilePath(filepath, fn_relief_data_file);}

};

#endif //_SurfaceGrid_h

