/**************************************************************************************************************
*
*       This file,
*
*                            /Table.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "Table.h"
#include "error.h"

#include <memory.h>
#include <math.h>

Table::Table()
{
	data=(double *)0;
	matrix_element=(double **)0;
}

Table::~Table()
{
	if(data)
		delete [] data;
	if(matrix_element)
		delete [] matrix_element;
}

bool Table::allocate( int m, int n)
{
	if(data)
		delete [] data;
	data=new double[m*n];
	if(matrix_element)
		delete [] matrix_element;
	matrix_element=new double*[m];
	if(!data || !matrix_element)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not allocate %d x %d table!\n", m, n);
		return false;
	}
	int i;
	double *ptr=data;
	for(i=0;i<m;i++)
	{
		matrix_element[i]=ptr;
		ptr+=n;
	}
	no_rows=m;
	no_columns=n;
	return true;
}

bool Table::addColumns(int n)
{
	if(!data)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "table object not initialized!\n");
		return false;
	}
	int i, j, new_no_columns=no_columns+n, size=sizeof(double)*no_columns;
	double *new_data=new double[no_rows*new_no_columns], *ptr;
	if(!new_data)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not allocate %d x %d table!\n", no_rows, new_no_columns);
		return false;
	}
	// copy old_data
	ptr=new_data;
	for(i=0;i<no_rows;i++)
	{
		memcpy(ptr, matrix_element[i], size);
		matrix_element[i]=ptr;
		ptr+=new_no_columns;
	}
	delete [] data;
	data=new_data;
	no_columns=new_no_columns;	
	return true;
}

bool Table::loadFrom(char *filename)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not load table from file \"%s\"!\n", filename);
		return false;
	}
	no_rows=numberOfRows(fptr);
	no_columns=numberOfColumns(fptr);
	if(!allocate(no_rows, no_columns))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not allocate table according to file \"%s\"!\n", filename);
		return false;
	}
	int i, j;
	double *ptr=data;
	for(i=0;i<no_rows;i++)
		for(j=0;j<no_columns;j++)
			if(fscanf(fptr, j<no_columns-1?"%lg\t":"%lg\n", ptr++)!=1)
			{
				errorIn(__FILE__, __LINE__);
				fprintf(stderr, "file \"%s\": error in line %d column %d\n", filename, i+1, j+1);
				fclose(fptr);
				return false;
			}
	fclose(fptr);
	return true;
}

bool Table::saveTo(char *filename)
{
	FILE *fptr=fopen(filename, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "could not save table to file \"%s\"!\n", filename);
		return false;
	}
	writeTo(fptr);
	fclose(fptr);
	return true;
}

void Table::writeTo(FILE *fptr)
{
	int i,j;
	double *me;
	for(i=0;i<no_rows;i++)
	{
		me=matrix_element[i];
		for(j=0;j<no_columns;j++)
			fprintf(fptr,j==0?"%lg":"\t%lg",me[j]); 
		fputc('\n', fptr);
	}
}

// operations ############################################################

bool Table::fillColumnWithRowNumbers(int iC)
{
	for(int i=0;i<no_rows;i++)
		matrix_element[i][iC]=double(i);
}

bool Table::prepareResultColumn(int *piCResult)
{
	bool append_result_column= *piCResult==-1;
	if(!(append_result_column||validColumnIndex(*piCResult)))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"invalid result colum index (iCResult=%d)\n", *piCResult);
		return false;
	}
	if(append_result_column)
	{
		*piCResult=no_columns;
		if(!addColumns(1))
			return false;
	}
	return true;
}

bool Table::applyElementaryOperation(int iCa, char opcode, int iCb, int iCResult)
{
	if(!(validColumnIndex(iCa) && validColumnIndex(iCb) && prepareResultColumn(&iCResult)))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"invalid column indices (iCa=%d, iCb=%d, iCResult=%d)\n", iCa, iCb, iCResult);
		return false;
	}
	register int i;
	register double *c;
	switch(opcode)
	{
		case '+':
			for(i=0;i<no_rows;i++)
			{
				c=matrix_element[i];
				c[iCResult]=c[iCa]+c[iCb];
			}
			break;
		case '-':
			for(i=0;i<no_rows;i++)
			{
				c=matrix_element[i];
				c[iCResult]=c[iCa]-c[iCb];
			}
			break;
		case '*':
			for(i=0;i<no_rows;i++)
			{
				c=matrix_element[i];
				c[iCResult]=c[iCa]*c[iCb];
			}
			break;
		case '/':
			for(i=0;i<no_rows;i++)
			{
				c=matrix_element[i];
				c[iCResult]=c[iCa]/c[iCb];
			}
			break;
		case '^':
			for(i=0;i<no_rows;i++)
			{
				c=matrix_element[i];
				c[iCResult]=pow(c[iCa], c[iCb]);
			}
			break;
		default:
			errorIn(__FILE__, __LINE__);
			fprintf(stderr, "opcode \'%c\' not implemented!\n", opcode);
			return false;
	}
	return true;
}

bool Table::apply1DGaussFilter(int iCx, int iCy, double sigma, int iCResult)
{
	if(!(validColumnIndex(iCx) && validColumnIndex(iCy) && prepareResultColumn(&iCResult)))
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr, "invalid column indices (iCx=%d, iCy=%d, iCResult=%d)\n", iCx, iCy, iCResult);
		return false;
	}
	// allocations, quick access and initialization
	int i, j;
	double *tmp_data=new double[3*no_rows];
	if(!tmp_data)
		return false;
	double *x, *y, *dx;
	x=tmp_data;
	y=x+no_rows;
	dx=y+no_rows;
	for(i=0;i<no_rows;i++)
	{
		x[i]=matrix_element[i][iCx];
		y[i]=matrix_element[i][iCy];
		dx[i]=0.5*(i==0?(x[i+1]-x[i]):(i<no_rows-1?(x[i+1]-x[i-1]):(x[i]-x[i-1])));
	}
	// apply filter
	double dxj, xi, xj, d, w, w_sum, r_sum=0.0;
	for(i=0;i<no_rows;i++)
	{
		xi=x[i];
		w_sum=0.0;
		r_sum=0.0;
		for(j=0;j<no_rows;j++)
		{
			xj=x[j];
			d=(xj-xi)/sigma;
			if(fabs(d)<=4.0)
			{
				w=exp(-0.5*d*d)*dx[j];
				r_sum+=w*y[j];
				w_sum+=w;
			}
		}
		matrix_element[i][iCResult]=r_sum/w_sum;
	}
	delete [] tmp_data;
	return true;
}

bool Table::findExtrema(int iC, list<int> *pExtremaIndexList)
{
	if(!validColumnIndex(iC))
		return false;
	int i;
	bool sign, new_sign;
	double y, y_new;
	pExtremaIndexList->clear();
	y=matrix_element[0][iC];
	sign=matrix_element[1][iC]>y;
	for(i=1;i<no_rows;i++)
	{
		y_new=matrix_element[i][iC];
		new_sign=y_new>y;
		if(sign!=new_sign)
		{
			pExtremaIndexList->push_back(i-1);
			sign=new_sign;
		}
		y=y_new;
	}
	return true;
}

bool Table::findRoots(int iCx, int iCy, list<double> *pRootList)
{
	if(!(validColumnIndex(iCx) && validColumnIndex(iCy)))
		return false;
	int i;
	bool sign, new_sign;
	double yi, yim1, xi, xim1;
	pRootList->clear();
	sign=(yim1=matrix_element[0][iCy])>0.0;
	for(i=1;i<no_rows;i++)
	{
		new_sign=(yi=matrix_element[i][iCy]>0.0);
		if(sign!=new_sign)
		{
			xi=matrix_element[i][iCx];
			xim1=matrix_element[i-1][iCx];
			pRootList->push_back((yim1*(xim1-xi))/(yi-yim1));
			sign=new_sign;
		}
		yim1=yi;
	}
	return false;
}





