/**************************************************************************************************************
*
*       This file,
*
*                            /VarianceReductionGrid.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "VarianceReductionGrid.h"
#include "BinaryGridFile.h"

double VarianceReductionGrid::WW_p=1.0;
double VarianceReductionGrid::WW_C=10.0;
double VarianceReductionGrid::ISThreshold=0.1;

VarianceReductionGrid::VarianceReductionGrid()
{
	cell=0;
	QuadraturePoint=0;
	no_allocated_quadrature_points=0;
	no_cells=0;	
	no_samples=0;
	total_accumulated_contribution=0.0;
	importance_norm=0.0;
}

VarianceReductionGrid::~VarianceReductionGrid()
{
	if(cell)
		delete [] cell;
	if(QuadraturePoint)
		delete [] QuadraturePoint;
}

bool VarianceReductionGrid::allocate(int no_cells, int no_mu_bins, int no_phi_bins, int no_quadrature_points)
{
	if(cell)
		delete [] cell;
	cell=new struct Cell_s[no_cells];
	if(QuadraturePoint)
		delete [] QuadraturePoint;
	QuadraturePoint=new struct QuadraturePoint_s[no_quadrature_points];
	if(!cell || !QuadraturePoint)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	memset(cell, 0, sizeof(struct Cell_s)*no_cells);
	memset(QuadraturePoint, 0, sizeof(struct QuadraturePoint_s)*no_quadrature_points);
	no_allocated_quadrature_points=no_quadrature_points;
	this->no_cells=no_cells;
	this->no_quadrature_points=0;
	importance_norm=0.0;
	for(int i=0;i<no_cells;i++)
	{
		if(!cell[i].allocate(no_mu_bins, no_phi_bins)) // , she_order))
		{
			errorIn(__FILE__, __LINE__, __func__);
			return false;
		}
		cell[i].pSunDirectionRotationMatrix=&SunDirectionRotationMatrix;
		cell[i].p_importance_norm=&importance_norm;
	}
	this->no_mu_bins=no_mu_bins;
	this->no_phi_bins=no_phi_bins;
	// this->she_order=she_order;
	return true;
}

void VarianceReductionGrid::resetQuadraturePoints()
{
	memset(QuadraturePoint, 0, sizeof(struct QuadraturePoint_s)*no_quadrature_points);
	no_quadrature_points=0;
}

void VarianceReductionGrid::registerQuadraturePoints()
{
	if(no_quadrature_points<=0)
		return;
	
	// start with end of quadrature point list
	int i;
	struct Cell_s *pCell;
	struct QuadraturePoint_s *pQP=QuadraturePoint+no_quadrature_points-1, *pQPTarget=0;
	double c;
	if(no_quadrature_points<=1)
	{
		// only one scatter event... (seldom)
		c=pQP->accumulated_contribution+pQP->contribution;
		pCell=cell+pQP->cell_index;
		pCell->registerContribution(c, pQP->weight, pQP->k);
	}
	else
	{
		for(i=1;i<no_quadrature_points;i++)
		{
			c=pQP->accumulated_contribution+pQP->contribution;
			pQPTarget=QuadraturePoint+pQP->predecessor;
			pCell=cell+pQPTarget->cell_index;
			pCell->registerContribution(c, pQP->weight, pQP->k);
			pQP--;
			// put accumulated contributions to predecessor
			pQPTarget->accumulated_contribution+=c;
		}
		// reached at the first scatter event
		c=pQP->accumulated_contribution+pQP->contribution;
		pCell=cell+source_cell_index;
		pCell->registerContribution(c, pQP->weight, pQP->k);
		
		// register total contribution
		total_accumulated_contribution+=c;
		no_samples++;
		importance_norm=total_accumulated_contribution/(double(no_samples));
	}
	
	resetQuadraturePoints();
}

void VarianceReductionGrid::registerQuadraturePoint(const int& cell_index, const Vector3D& p, const Vector3D& k, const double& contribution, const double & weight)
{
	struct QuadraturePoint_s *pQP=QuadraturePoint+no_quadrature_points;
	pQP->cell_index=cell_index;
	pQP->p=p;
	pQP->k=k;
	pQP->contribution=contribution;
	pQP->weight=weight;
	pQP->predecessor=current_predecessor;
	no_quadrature_points++;
}

/*void VarianceReductionGrid::updateDIVRSettings()
{
	int i;
	double area=0.0, threshold, initial_importance=0.0, initial_importance_weight_sum=0.0, weight, cell_importance;
	struct Cell_s *pCell=cell;
	// calculate initial importance	
	for(i=0;i<no_cells;i++)
	{
		area+=pCell->absolute_importance_measure=pCell->no_samples*pCell->contribution_sum;
		pCell++;
	}
	threshold=0.01*area/double(no_cells);
	pCell=cell;
	for(i=0;i<no_cells;i++)
	{
		if(pCell->absolute_importance_measure>threshold)
		{
			pCell->calculateTotalWeightWindow();
			weight=pCell->total_accumulated_contribution;
			pCell->importance=cell_importance=pCell->contribution_sum/pCell->weight_sum;
			initial_importance+=cell_importance*weight;
			initial_importance_weight_sum+=weight;
			pCell->enableDIS=true;
		}
		else
			pCell->enableDIS=false;
		pCell++;	
	}
	initial_importance/=initial_importance_weight_sum;
	double importance_norm=1.0/initial_importance;
	pCell=cell;
	for(i=0;i<no_cells;i++)
	{
		if(pCell->enableDIS)
		{
			pCell->relative_importance=pCell->importance/initial_importance;
			pCell->importance_norm=importance_norm;
		}
		pCell++;
	}
}*/

/*bool VarianceReductionGrid::exportDirectionalImportancesToBinaryGridFile(char *file_name)
{
	class BinaryGridFile BGF;
	if(!BGF.allocate(no_phi_bins, no_mu_bins, no_cells, 1))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;		
	}
	// write data to BinaryGridFile object
	int x,y,z, i;
	double *data=BGF.getCellPtr();
	for(z=0;z<no_cells;z++)
	{
		//cell[z].DirectionalImportance[0][0]=0.2;
		for(y=0;y<no_mu_bins;y++)
			for(x=0;x<no_phi_bins;x++)
			{
				i=BGF.calculateIndex(x, y, z, 0);
				if(cell[z].enableDIS[y][x])
					data[i]=cell[z].DirectionalImportance[y][x];
				else
					data[i]=-cell[z].DirectionalImportance[y][x];
			}
	}
	// save to file:
	return BGF.writeToFile(file_name);
}*/

bool VarianceReductionGrid::exportImportancesToBinaryGridFile(char *file_name)
{
	class BinaryGridFile BGF;
	int xres,yres,zres, x,y,z, i,j,k=0, no_layers=2+no_mu_bins*no_phi_bins;

	xres=pGrid->getGrid1DResolution(0);
	yres=pGrid->getGrid1DResolution(1);
	zres=pGrid->getGrid1DResolution(2);
	
	if(!BGF.allocate(xres, yres, zres, no_layers))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;		
	}
	double *data=BGF.getCellPtr(), M1, M2, w, re;
	for(z=0;z<zres;z++)
		for(y=0;y<yres;y++)
			for(x=0;x<xres;x++)
			{
				w=cell[k].weight_sum;
				if(w==0.0)
				{
					*data++=0.0;
					*data++=0.0;
				}
				else
				{
					M1=cell[k].absolute_importance_sum/w;
					M2=cell[k].absolute_squared_importance_sum/w;
					// write to BG file
					re=sqrt((M2-M1*M1)/w)/M1;
					if(re<VarianceReductionGrid::ISThreshold)
						*data++=M1/importance_norm;
					else
						*data++=0.0;
					*data++=re;

				}
				// directional importances
				for(i=0;i<no_mu_bins;i++)
					for(j=0;j<no_phi_bins;j++)
					{
						if(cell[k].enableDIS[i][j])
							*data=cell[k].DirectionalImportance[i][j];
						else
							*data=-cell[k].DirectionalImportance[i][j];
						data++;
					}
				k++;
			}
	// save to file:
	return BGF.writeToFile(file_name);
}

/*bool VarianceReductionGrid::writeCellData(char *file_name)
{
	FILE *fptr=fopen(file_name, "w");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not write to file \"%s\"!\n", file_name);
		return false;
	}
	for(int i=0;i<no_cells;i++)
		cell[i].writeTo(fptr);
	fclose(fptr);
	return true;
}*/

bool VarianceReductionGrid::save(char *file_name)
{
	FILE *fptr=fopen(file_name, "wb");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not write to file \"%s\"!\n", file_name);
		return false;
	}
	fwrite(&no_cells, sizeof(int), 1, fptr); 
	fwrite(&no_mu_bins, sizeof(int), 1, fptr); 
	fwrite(&no_phi_bins, sizeof(int), 1, fptr); 
	fwrite(&importance_norm, sizeof(double), 1, fptr); 
	
	int i, n=no_mu_bins*no_phi_bins;
	for(i=0;i<no_cells;i++)
	{
		// fwrite(&(cell[i].relative_importance), sizeof(double), 1, fptr);
		// fwrite(&(cell[i].importance_norm), sizeof(double), 1, fptr);
		fwrite(cell[i].DirectionalImportance.getCoefficientPtr(), sizeof(double), n, fptr);
		fwrite(cell[i].DirectionalImportanceVariance.getCoefficientPtr(), sizeof(double), n, fptr);
	}
	fclose(fptr);
	return true;
}

bool VarianceReductionGrid::load(char *file_name)
{
	FILE *fptr=fopen(file_name, "rb");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not read file \"%s\"!\n", file_name);
		return false;
	}
	int no_cells, no_mu_bins, no_phi_bins;
	fread(&no_cells, sizeof(int), 1, fptr); 
	fread(&no_mu_bins, sizeof(int), 1, fptr); 
	fread(&no_phi_bins, sizeof(int), 1, fptr); 
	fread(&importance_norm, sizeof(double), 1, fptr); 
	if(no_cells!=this->no_cells || no_mu_bins!=this->no_mu_bins || no_phi_bins!=this->no_phi_bins)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "file \"%s\" incompatible with current settings\n", file_name);
		fclose(fptr);
		return false;
	}
	int i, n=no_mu_bins*no_phi_bins;
	for(i=0;i<no_cells;i++)
	{
		// fread(&(cell[i].relative_importance), sizeof(double), 1, fptr);
		// fread(&(cell[i].importance_norm), sizeof(double), 1, fptr);
		fread(cell[i].DirectionalImportance.getCoefficientPtr(), sizeof(double), n, fptr);
		fread(cell[i].DirectionalImportanceVariance.getCoefficientPtr(), sizeof(double), n, fptr);
		cell[i].calculateEnableDISMap();
	}
	fclose(fptr);
	return true;
}

void VarianceReductionGrid::linkToAtmosphere(class Atmosphere *pAtmosphere)
{
	int i, n=pGrid->getNOCells();
	// struct Atmosphere::OpticalCell_s *oc=pAtmosphere->getOpticalCellPtr();
	for(i=0;i<n;i++)
		cell[i].pOC=pAtmosphere->getOpticalCellPtr(i);
}

void VarianceReductionGrid::setSunDirection(const class Vector3D& k)
{
	double r, theta, phi;
	SunDirection=k;
	k.toPolar(&r, &theta, &phi);
	// SunDirectionRotationMatrix = Matrix3D ( Matrix3D::ROTZ, -phi ) ;
	// SunDirectionRotationMatrix* = Matrix3D ( Matrix3D::ROTY, -theta ) ;
	SunDirectionRotationMatrix = Matrix3D ( Matrix3D::ROTY, -theta ) * Matrix3D ( Matrix3D::ROTZ, -phi );
}

void VarianceReductionGrid::determineActiveCells()
{

}


