/**************************************************************************************************************
*
*       This file,
*
*                            /VarianceReductionGrid.h,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#ifndef _VarianceReductionGrid_h
#define _VarianceReductionGrid_h

#include "SimulationObject.h"
#include "Atmosphere.h"

#include "Math/Vector3D.h"
#include "Math/Matrix3D.h"

#include "Math/Matrix.h"
#include "Math/TMatrix.h"

// static inline double& getWeightWindowRange(){return WW_C;}
// static inline double& getWeightWindowPartitionRatio(){return WW_p;}

class VarianceReductionGrid: public SimulationObject
{
	public:
		static double WW_C;
		static double WW_p;
		static double ISThreshold;

		struct Cell_s
		{
			// Cell_s() : importance_norm(0.0) {}
			Cell_s();
			~Cell_s();

			bool *DIS_flag_data, **enableDIS; // , enableDISInCell;
			void reset();

			bool allocate(int no_mu_bins, int no_phi_bins);// , int she_order);

			/*double weight_splitting_threshold;
			double weight_optimum;
			double weight_russian_roulette_threshold;*/

			class Atmosphere::OpticalCell_s *pOC;
			void calculateDISMapIndices(const class Vector3D& k, int& i_mu, int& i_phi);

			// class Vector Mu;
			class Matrix DirectionalImportance;
			class Matrix DirectionalImportanceVariance;

			// omnidirectional importance
			double absolute_importance, absolute_importance_sum, absolute_squared_importance_sum, weight_sum;
			
			// class SphericalHarmonicsExpansion she;
			// void calculateDirectionalImportances();
			
			double *p_importance_norm;
			class Matrix3D *pSunDirectionRotationMatrix;
			
			// quantities that are accumulated:	
			void registerContribution(const double& contribution, const double& weight, const class Vector3D& k);

			/*int no_samples;
			double weight_sum;
			double contribution_sum;
			double absolute_importance_measure;*/
			
			// contributions related to the first scatter event in the trajectory
			// the hole contribution of a trajectory is accumulated there
			// the weight is 1

			/*int total_no_samples;
			double total_accumulated_contribution;
			inline void registerTotalContribution(const double& c){total_accumulated_contribution+=c; total_no_samples++;}

			// variance and expectation of contribution-weighted weight
			double wc_sum;
			double wwc_sum;
			
			double mean_weight, weight_variance;

			class Vector3D k_sum;*/
			class Matrix AbsoluteDirectionalImportance;
			class Matrix AbsoluteDirectionalImportanceSquared;
			class Matrix DirectionalWeightSum;
			
			// tools for variance reduction
			// void calculateTotalWeightWindow();
			void calculateWeightWindow(class Vector3D& k, bool& on, double& wl, double &wo, double& wu);
			void calculateEnableDISMap();

			/*bool enableDIS;
			int DIS_samples_threshold;
			double importance, relative_importance, importance_norm;*/
			// void sampleImportantDirection(class Vector3D& k, double& p);
			// void writeTo(FILE *fptr);
		};

		struct QuadraturePoint_s
		{
			int cell_index;
			class Vector3D p, k;
			double contribution, weight;

			// tree navigation
			int predecessor;

			// accumulation
			double accumulated_contribution;

			QuadraturePoint_s(){cell_index=predecessor=0;}
		};

	private:

		// cells
		struct Cell_s *cell;
		int no_cells, no_mu_bins, no_phi_bins, source_cell_index;
		// bool enableDIS;
		
		// quadrature points list
		struct QuadraturePoint_s *QuadraturePoint;
		int no_quadrature_points, no_allocated_quadrature_points;
		int current_predecessor;
		
		class Vector3D SunDirection;
		class Matrix3D SunDirectionRotationMatrix;

		// total contribution
		double total_accumulated_contribution;
		double importance_norm;
		int no_samples;
		
	public:

		VarianceReductionGrid();
		~VarianceReductionGrid();

		bool save(char *file_name);
		bool load(char *file_name);
	
		void linkToAtmosphere(class Atmosphere *pAtmosphere);
		void determineActiveCells();
		void setSunDirection(const class Vector3D& k);
		
		inline void setSourceCellIndex(const int &i){source_cell_index=i;}

		bool allocate(int no_cells, int no_mu_bins, int no_azimuth_bins, int no_quadrature_points);
		void registerQuadraturePoints();
		void registerQuadraturePoint(const int& cell_index, const Vector3D& p, const Vector3D& k, const double& contribution, const double& weight);
		
		inline void calculateWeightWindow(const int& cell_index, class Vector3D& k, bool& on, double& wl, double &wo, double& wu)
			{cell[cell_index].calculateWeightWindow(k, on, wl, wo, wu);}

		inline int getCurrentQuadraturePointIndex(){return no_quadrature_points-1;}	// call AFTER registering an event!
		inline void setPredecessorIndex(const int& i){current_predecessor=i;}
		inline void setCurrentPredecessorIndex(){current_predecessor=no_quadrature_points-1;}
		
		bool exportImportancesToBinaryGridFile(char *file_name);
		// bool exportDirectionalImportancesToBinaryGridFile(char *file_name);
		// bool writeCellData(char *file_name);
		inline void resetCells(){memset(cell, 0, sizeof(struct Cell_s)*no_cells);}
		void resetQuadraturePoints();
};

#endif // _VarianceReductionGrid_h






