/**************************************************************************************************************
*
*       This file,
*
*                            /VarianceReductionGridCell.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "VarianceReductionGrid.h"
#include "trigonometry.h"

VarianceReductionGrid::Cell_s::Cell_s()
{
	DIS_flag_data=0;
	enableDIS=0;
	absolute_importance_sum=0.0;
	absolute_squared_importance_sum=0.0;
	weight_sum=0.0;
}

VarianceReductionGrid::Cell_s::~Cell_s()
{
	if(DIS_flag_data)
		delete [] DIS_flag_data;
	if(enableDIS)
		delete [] enableDIS;
}

void VarianceReductionGrid::Cell_s::reset()
{
	DirectionalImportance.setToZero();
	DirectionalImportanceVariance.setToZero();
	AbsoluteDirectionalImportance.setToZero();
	AbsoluteDirectionalImportanceSquared.setToZero();
	DirectionalWeightSum.setToZero();
	memset(DIS_flag_data, 0, sizeof(bool)*DirectionalImportance.getNumberOfElements());
	pOC=0;
	absolute_importance_sum=0.0;
	absolute_squared_importance_sum=0.0;
	weight_sum=0.0;
}

bool VarianceReductionGrid::Cell_s::allocate(int no_mu_bins, int no_phi_bins) //, int she_order)
{
	int i, n=no_mu_bins*no_phi_bins;
	if(DIS_flag_data)
		delete [] DIS_flag_data;
	DIS_flag_data=new bool[n];
	if(enableDIS)
		delete [] enableDIS;
	enableDIS=new bool *[no_mu_bins];
	if(!enableDIS || !DIS_flag_data)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}

	// assign pointers
	for(i=0;i<no_mu_bins;i++)
		enableDIS[i]=DIS_flag_data+i*no_phi_bins;

	// return matrix initialisation 
	return AbsoluteDirectionalImportance.allocate(no_mu_bins, no_phi_bins) \
		&& AbsoluteDirectionalImportanceSquared.allocate(no_mu_bins, no_phi_bins) \
		&& DirectionalImportance.allocate(no_mu_bins, no_phi_bins) \
		&& DirectionalImportanceVariance.allocate(no_mu_bins, no_phi_bins) \
		&& DirectionalWeightSum.allocate(no_mu_bins, no_phi_bins);
}

void VarianceReductionGrid::Cell_s::calculateDISMapIndices(const class Vector3D& k, int& i_mu, int& i_phi)
{
	class Vector3D k_sdr=(*pSunDirectionRotationMatrix)*k;

	double phi=atan2(k_sdr[1], k_sdr[0]), mu=correctMu(k_sdr[2]);
	
	int N_mu=AbsoluteDirectionalImportance.getM();
	int N_phi=AbsoluteDirectionalImportance.getN();

	// i_mu=int((double(N_mu)*(1.0+mu))/2.0)%N_mu;
	i_mu=int(double(N_mu)*pOC->HGDIS.evaluateCDF(mu))%N_mu;
	i_phi=int((double(N_phi)*(M_PI+phi))/(2.0*M_PI))%N_phi;
	
	/*
	if(i_mu<0 || i_mu>=N_mu || i_phi<0 || i_phi>=N_phi)
		errorIn(__FILE__, __LINE__, __func__);
	*/
}

void VarianceReductionGrid::Cell_s::registerContribution(const double& contribution, const double& weight, const class Vector3D& k)
{
	double Sw, Scw, Scwcw, f_k, var_f_k, re, squared_contribution=contribution*contribution;

	// register contribution 
	absolute_importance_sum+=contribution;
	absolute_squared_importance_sum+=squared_contribution;
	weight_sum+=weight;
	
	// DIS ########################################################################
	int i_mu, i_phi;
	calculateDISMapIndices(k, i_mu, i_phi);

	// register contribution in direction map
	AbsoluteDirectionalImportance[i_mu][i_phi]+=contribution;
	AbsoluteDirectionalImportanceSquared[i_mu][i_phi]+=squared_contribution;
	DirectionalWeightSum[i_mu][i_phi]+=weight;

	// update directional importance
	Sw=DirectionalWeightSum[i_mu][i_phi];
	Scw=AbsoluteDirectionalImportance[i_mu][i_phi];
	Scwcw=AbsoluteDirectionalImportanceSquared[i_mu][i_phi];
	f_k=Scw/Sw;
	var_f_k=(Scwcw-Scw*Scw/Sw)/Sw;
	re=sqrt(var_f_k)/f_k;
	if(re!=0.0 && re<VarianceReductionGrid::ISThreshold)
	{
		DirectionalImportance[i_mu][i_phi]=f_k;
		DirectionalImportanceVariance[i_mu][i_phi]=var_f_k;
		enableDIS[i_mu][i_phi]=true;
	}
}

void VarianceReductionGrid::Cell_s::calculateEnableDISMap()
{
	int i_mu, i_phi, N_mu, N_phi;
	N_mu=AbsoluteDirectionalImportance.getM();
	N_phi=AbsoluteDirectionalImportance.getN();
	double re;
	for(i_mu=0;i_mu<N_mu;i_mu++)
		for(i_phi=0;i_phi<N_phi;i_phi++)
		{
			re=sqrt(DirectionalImportanceVariance[i_mu][i_phi])/DirectionalImportance[i_mu][i_phi];
			if(re!=0.0 && re<VarianceReductionGrid::ISThreshold)
				enableDIS[i_mu][i_phi]=true;
		}
}


/*
void VarianceReductionGrid::Cell_s::calculateDirectionalImportances()
{
	if(contribution_sum==0.0 || weight_sum==0.0)
		return;
	double directional_weight_sum, *ca, *cr, *cw, di; 
	int i, n, M, N;

	double mean_directional_importance=0.0;

	ca=AbsoluteDirectionalImportance.getCoefficientPtr();
	cr=DirectionalImportance.getCoefficientPtr();
	cw=DirectionalWeightSum.getCoefficientPtr();

	M=DirectionalImportance.getM();
	N=DirectionalImportance.getN();
	n=M*N;

	int no_valid_cells=0;
	
	for(i=0;i<n;i++)
	{
		directional_weight_sum=*cw++;
		if(directional_weight_sum)
		{
			di=*ca++/directional_weight_sum;
			// mean_directional_importance+=di;
			// no_valid_cells++;
			*cr++=di;
		}
		else
		{
			*cr++=0.0;
			ca++;
		}
	}

	importance=contribution_sum/weight_sum;
	mean_weight=wc_sum/contribution_sum;

	double mean_ww=wwc_sum/contribution_sum;
	weight_variance=mean_ww-mean_weight*mean_weight;
}
*/

// #define WW_C		2
// #define WW_p		1.1

static inline void WeightWindowFunction(const double& importance, double& wl, double& wo, double& wu)
{
	wo=1.0/importance;
	wl=wo/(1.0+VarianceReductionGrid::WW_p*(VarianceReductionGrid::WW_C-1.0));
	wu=VarianceReductionGrid::WW_C*wl;
}

void VarianceReductionGrid::Cell_s::calculateWeightWindow(class Vector3D& k, bool& on, double& wl, double &wo, double& wu)
{
	int i_mu, i_phi;
	calculateDISMapIndices(k, i_mu, i_phi);
	on=enableDIS[i_mu][i_phi];
	if(on)
	{
		double Phi=DirectionalImportance[i_mu][i_phi] / *p_importance_norm;
		WeightWindowFunction(Phi, wl, wo, wu);
	}
}








