/**************************************************************************************************************
*
*       This file,
*
*                            /VarianceReductionLayer.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "VarianceReductionLayer.h"
#include "Geometry.h"

VarianceReductionLayer::VarianceReductionLayer()
{
	// b_enable_directional_importance_sampling=false;
	b_generate_variance_reduction_parameter_file=false;
	b_load_variance_reduction_parameter_file=false;
	strcpy(fn_variance_reduction_parameter_file, "default.vrp");

	psi_sampling_probability=0.0;
	delta_mu=1e-5;

	g_sun=0.0;
	maximum_recursion_depth=0;

	mu_split=0.0;
	w_split=1.0;
	w_russian_roulette=0.1;
	p_split=0.1;
	b_importance_sampling_on_root_trajectory=false;
	no_phi_bins=12;
	no_mu_bins=20;

	// estimated_scatter_order=0;
	// variance reduction
	// addParameter(&b_enable_directional_importance_sampling, type_bool, "b_enable_directional_importance_sampling|b_enable_DIS", 0, "use directional importance sampling (memory consumption!)?");
	addParameter(&b_generate_variance_reduction_parameter_file, type_bool, "b_generate_variance_reduction_parameter_file|b_generate_VRP_file", 0, "generate variance reduction parameter file?");
	addParameter(&b_load_variance_reduction_parameter_file, type_bool, "b_load_variance_reduction_parameter_file|b_load_VRP_file", 0, "load variance reduction parameter file?");
	addParameter(fn_variance_reduction_parameter_file, type_file, "fn_variance_reduction_parameter_file|fn_VRP", 0, "variance reduction parameter file");
	addParameter(&psi_sampling_probability, type_double, "psi_sampling_probability|p_psi|psp", 0, "psi (importance) sampling probability");
	addParameter(&p_split, type_double, "p_split|ps", 0, "splitting probability");
	addParameter(&delta_mu, type_double, "delta_mu|dmu", 0, "delta mu for (delta distribution) forward scattering");
	addParameter(&g_sun, type_double, "g_sun", 0, "asymmetry parameter for variance reduction");
	addParameter(&maximum_recursion_depth, type_int, "maximum_recursion_depth|max_rcd", 0, "maximum recursion depth");
	addParameter(&mu_split, type_double, "mu_split|mu_s", 0, "mu split");
	addParameter(&w_split, type_double, "w_split|w_s", 0, "lower weight limit for splitting");
	addParameter(&w_russian_roulette, type_double, "w_russian_roulette|w_rr", 0, "upper weight limit for russian roulette");
	addParameter(&b_importance_sampling_on_root_trajectory, type_bool, "b_importance_sampling_on_root_trajectory|b_isort", 0, "use importance sampling on root trajectory?");
	addParameter(&no_mu_bins, type_int, "no_mu_bins", 0, "number of mu=kz=cos(theta) bins");
	addParameter(&no_phi_bins, type_int, "no_phi_bins", 0, "number of phi (azimuth) bins");

	//addParameter(&VarianceReductionGrid::WW_C, type_double, "WW_C|weight_window_size", 0, "size of the weight window");
	//addParameter(&VarianceReductionGrid::WW_p, type_double, "WW_p|weight_window_partition", 0, "partition ratio of the weight window");
	//addParameter(&VarianceReductionGrid::ISThreshold, type_double, "ISThreshold|importance_sampling_threshold", 0, "relative precision threshold for importance sampling");
	
}

VarianceReductionLayer::~VarianceReductionLayer()
{
}

bool VarianceReductionLayer::setup(int maximum_no_trace_events)
{
	if(!VRGrid.allocate(no_cells, no_mu_bins, no_phi_bins, maximum_no_trace_events)) // , b_enable_directional_importance_sampling))//, she_order))
	{
		errorIn(__FILE__, __LINE__, __func__);
		return false;
	}
	VRGrid.linkToAtmosphere(pAtmosphere);
	VRGrid.setSunDirection(pGeometry->getSunDirection());

	if(b_load_variance_reduction_parameter_file && !VRGrid.load(fn_variance_reduction_parameter_file))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading variance reduction parameter file!\n");
		return false;
	}
	/*if(!VarianceReductionParameterMatrix.allocate(no_cells, 9))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "variance reduction: error allocating parameter matrix\n");
		return false;
	}
	VarianceReductionParameterMatrix.setToZero();
	if(!ScatterOrderContributionMatrix.allocate(maximum_no_trace_events, 7))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "variance reduction: error allocating scatter order contribution matrix\n");
		return false;
	}
	ScatterOrderContributionMatrix.setToZero();*/
// en/dis-able VRP update
	update_variance_reduction_parameters = !b_load_variance_reduction_parameter_file;
	return true;
}

