/**************************************************************************************************************
*
*       This file,
*
*                            /coordinate_sampling.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "coordinate_sampling.h"
#include "Math/prand.h"
#include "trigonometry.h"

#include <math.h>

double SampleCartesicCoordinate(const double& N, const double& c)
{
	return c+prand()*N;
}

double SamplePhiCoordinate(const double& N, const double& c)
{
	return c+prand()*N;
}

double SampleThetaCoordinate(const double& N, const double& c)
{
	return acos(c+prand()*N);
}

double SampleRadiusCoordinate(const double& N, const double& c)
{
	return exp(log(prand()*N+c)/3.0);
}

void sampleRotationSymmetricDirection(class Vector3D& reference_direction, class Vector3D& direction, const double& mu_s)
{
	double phi_s=(2.0*M_PI)*prand();
	double sin_theta_s=sqrt(1.0-mu_s*mu_s);

	// construct scattered vector in z-direction
	double x=cos(phi_s)*sin_theta_s;
	double y=sin(phi_s)*sin_theta_s;
	double z=mu_s;

	double theta, phi, rx, ry, rz, sina, cosa;
	reference_direction.toPolar(&theta, &phi);

	// rotate it back: theta

	//	a[0][0]=cos(val);	a[0][1]=0.0;		a[0][2]=sin(val); 
	//	a[1][0]=0.0;		a[1][1]=1.0;		a[1][2]=0.0; 
	//	a[2][0]=-sin(val);	a[2][1]=0.0;		a[2][2]=cos(val); 		

	sina=sin(theta);
	cosa=cos(theta);

	rx=x*cosa+z*sina;
	rz=-x*sina+z*cosa;
	// update x and z
	x=rx;
	z=rz;

	// rotate it back phi

	//	a[0][0]=cos(val);	a[0][1]=-sin(val);	a[0][2]=0.0; 
	//	a[1][0]=sin(val);	a[1][1]=cos(val);	a[1][2]=0.0; 
	//	a[2][0]=0.0;		a[2][1]=0.0;		a[2][2]=1.0; 					

	sina=sin(phi);
	cosa=cos(phi);

	rx=x*cosa-y*sina;
	ry=x*sina+y*cosa;

	direction[0]=rx;
	direction[1]=ry;
	direction[2]=rz;
}

