/**************************************************************************************************************
*
*       This file,
*
*                            /fileio.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "fileio.h"

int numberOfRows(char *filename, bool robust_behaviour)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr) return -1;
	int rows=numberOfRows(fptr, robust_behaviour);
	fclose(fptr);
	return rows;	
}

int numberOfRows(FILE *fptr, bool robust_behaviour)
{
	int rows=0, c;
	char line_buffer[100];
	long filepos=ftell(fptr);
	fseek(fptr,SEEK_SET,0);
	bool is_new_line=true;
	while((c=fgetc(fptr))!=EOF)
	{
		switch(c)
		{
			case '\n':
				rows++;
				is_new_line=true;
				break;
			case '#':
				if(is_new_line && robust_behaviour)
					fgets(line_buffer, 100, fptr);	// eat comment line
			default:
				is_new_line=false;
		}	
	}
	fseek(fptr, SEEK_SET, filepos);
	return rows;	
}

int numberOfColumns(char *filename, bool robust_behaviour)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr) return -1;
	int ntabs=numberOfColumns(fptr, robust_behaviour);
	fclose(fptr);
	return ntabs;
}

int numberOfColumns(FILE *fptr, bool robust_behaviour)
{
	long filepos=ftell(fptr);
	fseek(fptr,SEEK_SET,0);
	int character, ntabs=0;
	if(robust_behaviour)
	{
		char line_buffer[100];
		while(1)
		{
			character=fgetc(fptr);
			fseek(fptr, -1, SEEK_CUR);	// rewind file ptr
			if(character=='#')
				fgets(line_buffer, 100, fptr);	// eat comment line
			else break;
		}
		bool last_character_is_a_spacer=true;
		do
		{
			character=fgetc(fptr);
			switch(character)
			{
				case ' ':
				case '\t':
					last_character_is_a_spacer=true;
					break;

				case EOF:
				case '\n':
					break;

				default:
					if(last_character_is_a_spacer)
					{
						ntabs++;
						last_character_is_a_spacer=false;
					}
			}
		}
		while(character!=EOF && character!='\n');
	}
	else
	{
		int c;
		while((c=fgetc(fptr))!='\n' && c!=EOF) if(c=='\t') ntabs++;
		ntabs++;
	}
	fseek(fptr,SEEK_SET,filepos);
	return ntabs;
}

bool readTable(FILE *fptr, int rows, int columns, int element_size, char *format_string, void *data)
{
	int row, column, fields_read;
	if(rows>1)
	{
		char irf[20], ref[20];	// in-row-format and row-end-format
		sprintf(irf,"%%%s\t",format_string);
		sprintf(ref,"%%%s\n",format_string);
		for(row=0;row<rows;row++)
			for(column=0;column<columns;column++)
			{
				if(column<columns-1)
					fields_read=fscanf(fptr,irf,data);
				else
					fields_read=fscanf(fptr,ref,data);
				if(fields_read!=1)
				{
					fprintf(stderr,"error in column %d, row %d!\n",column+1,row+1);
					fprintf(stderr,"read %d field(s)!\n",fields_read);
					return false;
				}
				data=(void *)((char *)data+element_size);
			}
	}
	else
	{
		char format[20];	// in-row-format and row-end-format
		sprintf(format,"%%%s\\n",format_string);
		for(row=0;row<rows;row++)
		{
			fields_read=fscanf(fptr,format_string,data);
			if(fields_read!=1)
			{
				fprintf(stderr,"error in row %d!\n",row+1);
				fprintf(stderr,"read %d field(s)!\n",fields_read);
				return false;
			}
			data=(void *)((char *)data+element_size);
		}
	}
	return true;
}

bool doesFileExist(char *filename)
{
	FILE *fptr=fopen(filename,"r");
	if(!fptr)
		return false;
	fclose(fptr);
	return true;
}

