/**************************************************************************************************************
*
*       This file,
*
*                            /main_BGF.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "BinaryGridFile.h"
#include "Math/Matrix.h"
#include "error.h"

#include <stdlib.h>

void print_usage(FILE *fptr)
{
	fprintf(fptr,"bgf_tool file command\n\n");
	fprintf(fptr,"command is:\n");
	fprintf(fptr,"\t-i\t: information on the content of file\n");
	fprintf(fptr,"\t-a l p\t: add and print voxels of layer l and plane p of file\n");
	fprintf(fptr,"\t-x l p i\t: extract layer l plane p with index i...\n");
	fprintf(fptr,"\t-xeps l p i epsfile\t: extract eps picture from layer l plane p with index i into file epsfile\n");
	fprintf(fptr,"\nbugs to: Tim.Deutschmann@iup.uni-heidelberg.de\n\n");
	/*fprintf(fptr,"");
	fprintf(fptr,"");
	fprintf(fptr,"");*/
}

int main(int argc, char *argv[])
{
	if(argc<3)
	{
		print_usage(stdout);
		return 0;
	}
	char *filename=argv[1];
	BinaryGridFile brf;
	if(!brf.loadFromFile(filename))
		return -1;
	char *command=argv[2];
	if(!strcmp(command,"-i"))
	{
		brf.writeFileHeaderInfo(stdout);
		return 0;
	}
	else if(!strcmp(command, "-a"))
	{
		class Matrix result, plane;
		int layer_index=atoi(argv[3])-1;
		char plane_character=argv[4][0];
		enum BinaryGridFile::GridPlane_e GridPlane;
		if(!brf.Character2GridPlane(plane_character,&GridPlane))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"plane character has to be 'x','y' or 'z'!\n");
			return -1;
		}
		brf.setLayerIndex(layer_index);
		brf.setPlaneAndIndex(GridPlane, 0);
		int xres,yres,zres, x,y,z;
		xres=brf.getXRes();
		yres=brf.getYRes();
		zres=brf.getNumberOfPlanes();
		if(!result.allocate(yres, xres) || !plane.allocate(yres, xres))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "allocation error!\n");
			return -1;
		}
		result.setToZero();
		for(z=0;z<zres;z++)
		{
			brf.setPlaneAndIndex(GridPlane, z);
			brf.getPlaneData(plane.getCoefficientPtr());
			for(y=0;y<yres;y++)
				for(x=0;x<xres;x++)
					result[y][x]+=plane[y][x];
		}
		result.dumpTo(stdout);
	}
	else if(!strcmp(command,"-x") || !strcmp(command,"-xeps"))
	{
		if(!(argc==6 || argc==7))
		{
			print_usage(stderr);
			return -1;
		}
		int layer_index=atoi(argv[3])-1;
		char plane_character=argv[4][0];
		int plane_index=atoi(argv[5])-1;
		enum BinaryGridFile::GridPlane_e GridPlane;
		if(!brf.Character2GridPlane(plane_character,&GridPlane))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr,"plane character has to be 'x','y' or 'z'!\n");
			return -1;
		}
		brf.setLayerIndex(layer_index);
		brf.setPlaneAndIndex(GridPlane, plane_index);
		if(!strcmp(command, "-x"))
			brf.exportToASCII(stdout);
		else
		{
			if(argc!=7)
			{
				print_usage(stderr);
				return -1;
			}
			char *epsfilename=argv[6];
			if(!brf.exportSliceToEPS(epsfilename, true))
			{
				fprintf(stderr,"error during export to eps file \"%s\"!\n",epsfilename);
				return -1;
			}
		}
		return 0;
	}
	return 0;
}

