/**************************************************************************************************************
*
*       This file,
*
*                            /main_MieTool.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "MieModule.h"
#include "error.h"

#include <stdlib.h>

void print_usage(FILE *fptr)
{
	fprintf(fptr, "usage: MieTool lambda re im [ [ s radius ] | [ sp | sr | st ] radius N_phf ]\n");
	//                      0       1     2  3     4   5          4    4   4       5     6          
}

int main(int argc, char *argv[])
{
	class MieModule MM;
	
	if(argc<5)
	{
		print_usage(stdout);
		return 0;
	}
	char *command=argv[4];
	double lambda=atof(argv[1]);
	double re=atof(argv[2]);
	double im=atof(argv[3]);
	if(!strcmp(command, "sp") || !strcmp(command, "sr") || !strcmp(command, "st"))
	{
		if(argc!=7)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "wrong number of arguments!\n");
			print_usage(stderr);
			return -1;
		}
		double radius=atof(argv[5]);
		int i, N=atoi(argv[6]);
		if(!MM.setParameters(lambda, re, im, radius))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error setting Mie module parameters!\n");
			return -1;
		}
		MM.calculateMieCoefficients();
		if(!strcmp(command, "sp"))
		{
			double mu;
			class StokesMatrix P;
			double Qext, Qsca, ssa, g, epsilon;
			class LegendrePhaseFunction F11, F12, F33, F34;
			MM.developLegendrePhaseFunctions(F11, F12, F33, F34, MM.getNOCoefficients(), N);
			Qext=MM.getExtinctionEfficiency();
			Qsca=MM.getScatteringEfficiency();
			ssa=MM.getSingleScatteringAlbedo();
			g=MM.getAsymmetryParameter();
			epsilon=Qext*M_PI*radius*radius;
			printf("# %.3lf\t%le\t%.5lf\t%.3lf\n", Qext, epsilon, ssa, g);
			for(i=0;i<=N;i++)
			{
				mu=-1.0+double(2*i)/double(N);
				MM.calculatePhaseMatrix(mu, P);
				printf("%.4lf\t%le\t%le\t%le\t%le", mu, P[0][0], P[0][1], P[2][2], P[3][2]);
				printf("\t%le\t%le\t%le\t%le\n", F11.LegendrePolynomial::evaluate(mu), F12.LegendrePolynomial::evaluate(mu), F33.LegendrePolynomial::evaluate(mu), F34.LegendrePolynomial::evaluate(mu));
			}
		}
		if(!strcmp(command, "sr"))
		{
			double mu;
			class StokesMatrix P;
			double Qext, Qsca, ssa, g, epsilon;
			Qext=MM.getExtinctionEfficiency();
			Qsca=MM.getScatteringEfficiency();
			ssa=MM.getSingleScatteringAlbedo();
			g=MM.getAsymmetryParameter();
			epsilon=Qext*M_PI*radius*radius;
			printf("# %.3lf\t%le\t%.5lf\t%.3lf\n", Qext, epsilon, ssa, g);
			for(i=0;i<=N;i++)
			{
				mu=-1.0+double(2*i)/double(N);
				MM.calculatePhaseMatrix(mu, P);
				printf("%.4lf\t%le\t%le\t%le\t%le\n", mu, P[0][0], P[0][1], P[2][2], P[3][2]);
			}
		}
		else if(!strcmp(command, "st"))
			MM.printMieCoefficients(stdout);
	}
	else if(!strcmp(command, "s"))
	{
		if(argc!=6)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "wrong number of arguments!\n");
			print_usage(stderr);
			return -1;
		}
		double radius=atof(argv[5]);
		if(!MM.setParameters(lambda, re, im, radius))
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error setting Mie module parameters!\n");
			return -1;
		}
		MM.calculateMieCoefficients();
		double Qext, Qsca, Qabs, ssa, g, epsilon;
		Qext=MM.getExtinctionEfficiency();
		Qsca=MM.getScatteringEfficiency();
		Qabs=Qext-Qsca;
		ssa=MM.getSingleScatteringAlbedo();
		g=MM.getAsymmetryParameter();
		epsilon=Qext*M_PI*radius*radius;
		printf("%le\t%le\t%le\t", Qext, Qsca, Qabs);
		printf("%le\t%le\t%le\n", epsilon, ssa, g);
	}
	/*else if(!strcmp(command, "bulk"))
	{
		if(argc!=10 && argc!=9 && argc!=8)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "wrong number of arguments!\n");
			print_usage(stderr);
			return -1;
		}
		if(!strcmp(argv[5], "LN") || !strcmp(argv[5], "lognormal"))
		{
			int N_dsd=atoi(argv[6]);
			double radius =atof(argv[7]);
			double sigma_log=atof(argv[8]);
			int N_phf=atoi(argv[9]);
			// bool doLSE;
			if(!MM.processLogNormalDSD(lambda, radius, sigma_log, re, im, N_dsd, N_phf))
			{
				errorIn(__FILE__, __LINE__,__func__);
				return -1;
			}
		}
		else if(!strcmp(argv[5], "bins"))
		{
			// detailed n(r) given
			if(argc==8)// cross section, ssa, g
				MM.calculateCrossSectionSSAAndG(argv[6], argv[7]);
			else if(argc==9)
				MM.calculateBulkDetailedPhaseFunction(argv[6], argv[7], atoi(argv[8]));
		}
	}*/
	return 0;
}


