/**************************************************************************************************************
*
*       This file,
*
*                            /refraction.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "refraction.h"
#include "Math/Matrix3D.h"
#include "error.h"

#ifdef WIN32
#include <float.h>
#endif

class Vector3D doRefraction(const Vector3D& normal, const Vector3D& dir_IN, const double &n_IN, const double &n_OUT)
{
/*
	normal:		normal vector of surface
	dir_IN:		incoming photon direction
	n_IN:		refraction index of medium which is left
	n_OUT:		refraction index of medium which is entered
	a_IN: 		incoming angle to normal
	a_OUT: 		outgoing angle to normal
	
	axis=normal x dir_IN
	
	Snell's law:
	
	n1*sin(a1)=n2*sin(a2)
	a2=arcsin(n1/n2*sin(a1))

	v. Friedeburgs Snells law mod:

	a2=arcsin(n1/(n1+dn*cos(a2))*sin(a1))
*/	
	class Vector3D axis=normal.cross(dir_IN);
	double a_IN=asin(axis.length()/(dir_IN.length()*normal.length()));
	double sin_alpha=(n_IN*sin(a_IN))/n_OUT;
	if(sin_alpha>1.0)
		sin_alpha=1.0;
	else if(sin_alpha<-1.0)
		sin_alpha=-1.0;
	double a_OUT=asin(sin_alpha);
#ifdef _DEBUG
	if(_isnan(a_IN) || _isnan(a_OUT) || !axis.isValid())
	{
		errorIn(__FILE__, __LINE__);
		fprintf(stderr,"nan number\n");
	}
#endif
	return axisRotation(axis,fabs(a_OUT-a_IN))*dir_IN;	
}

double calculateFresnelCoefficient(const double& mu, const double& refractive_index)
{
	double cos_theta_in=mu;
	double sin_theta_in=sqrt(1.0-cos_theta_in*cos_theta_in);

	double sin_psi=sin_theta_in/refractive_index;
	double cos_psi=sqrt(1.0-sin_psi*sin_psi);
	
	double rft=refractive_index*cos_psi;
	rft=(cos_theta_in-rft)/(cos_theta_in+rft);
	rft*=rft;

	double rfp=refractive_index*cos_theta_in;
	rfp=(rfp-cos_psi)/(rfp+cos_psi);
	rfp*=rfp;
	return 0.5*(rft+rfp);
}


