/**************************************************************************************************************
*
*       This file,
*
*                            /slit_function/main.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "LinearelyInterpolatedFunctionTemplate.h"
#include "Matrix.h"
#include "error.h"

int main(int argc, char *argv[])
{
	if(argc!=6)
	{
		printf("cwsf spectrum slit_function dx normalize_sf? dx_grid?\n");
		return 0;
	}
	LinearelyInterpolatedFunctionTemplate<double> slit_function;
	LinearelyInterpolatedFunctionTemplate<double> spectrum;
	class Matrix sf;
	double dx, x;
	if(!spectrum.readFromFile(argv[1]))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loading spectrum from file \"%s\"\n", argv[1]);
		return -1;
	}
	if(!slit_function.readFromFile(argv[2]))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error loadig slit function from file \"%s\"\n", argv[2]);
		return -1;
	}
	dx=atof(argv[3]);
	// create sf matrix
	double sf_xmin = slit_function.xmin();
	double sf_xmax = slit_function.xmax();
	int i, j, M = int((sf_xmax-sf_xmin)/dx)+1;
	if(!sf.allocate(M, 2))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error allocating slit function matrix\n");
		return -1;
	}
	double sf_norm=0.0;
	bool normalize=atoi(argv[4])==1?true:false;
	bool use_dx_grid=atoi(argv[5])==1?true:false;
	for(i=0;i<M;i++)
	{
		x=sf_xmin+double(i)*dx;
		sf[i][0]=x;
		sf[i][1]=slit_function.evaluate(x);
		sf_norm+=fabs(sf[i][1]);
	}
	int no_points=spectrum.getNOPoints();
	double *lambda=spectrum.getXPtr(), l, I, s;
	if(use_dx_grid)
	{
		double l_start=spectrum.getXPtr()[0];
		double l_end=spectrum.getXPtr()[no_points-1];
		l=l_start;
		do
		{
			s=0.0;
			for(j=0;j<M;j++)
			{
				x = l + sf[j][0];
				I = spectrum.evaluate(x);
				s +=  I * sf[j][1];
			}
			if(normalize)
				s /= sf_norm;
			else
				s *= dx;
			printf("%lg\t%lg\t%lg\n", l, spectrum.evaluate(l), s);
			l+=dx;
		}while(l<l_end);
	}
	else
	{
		for(i=0;i<no_points;i++)
		{
			s=0.0;
			l=lambda[i];
			for(j=0;j<M;j++)
			{
				x = l + sf[j][0];
				I = spectrum.evaluate(x);
				s +=  I * sf[j][1];
			}
			if(normalize)
				s /= sf_norm;
			else
				s *= dx;
			printf("%lg\t%lg\t%lg\n", l, spectrum.evaluate(l), s);
		}
	}
	return 0;
} 





