/**************************************************************************************************************
*
*       This file,
*
*                            /stringio.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "stringio.h"
#include "error.h"
#include <time.h>

#ifdef WIN32
	#include <string.h>
	#include <stdio.h>
#else
	#include <string>
#endif

int numberOf(char character, const char *text)
{
	int n=0;
	while(*text) if(*text++==character) n++;
	return n;
}

void remove(char character, char *text)
{
	if(!*text) return;
	size_t n=strlen(text);
	char *buffer=new char[n+1],*source_charptr=text,*destination_charptr=buffer,c;
	do
	{
		c=*source_charptr++;
		if(c!=character)
		{
			*destination_charptr=c;
			destination_charptr++;
		}
	}
	while(c!='\0');
	//*destination_charptr='\0';// terminate string
	strcpy(text,buffer);
	*(--destination_charptr)=character;
	delete [] buffer;
}

bool extractWordFromList(const char *text, int index, char separation_character, char *buffer)
{
	char character;
	while(index>0)
	{
		character=*text++;
		if(character==separation_character)
			index--;
		if(character=='\0' && index>0)
			return false;
	}
	while(1)
	{
		*buffer=*text;
		if(*text=='\0')
			return true;
		else if(*text==separation_character)
		{
			*buffer='\0';
			return true;
		}
		text++; buffer++;
	}
	return false;
}

#ifdef WIN32
	#define	DIR_CHAR	'\\'
#else
	#define DIR_CHAR	'/'
#endif	

bool extractPathFromFilename(const char *filename, char *path_buffer, char *name)
{
	const char *charptr=filename+strlen(filename)-1;
	while(*charptr--!=DIR_CHAR)
		if(charptr<filename)
			return false; // attention: this is part of the loop
	charptr++;
	int index=int(charptr-filename)+1;
	strncpy(path_buffer, filename, index);
	path_buffer[index]='\0';
	if(name)
		strcpy(name,++charptr);
	return true;
}

bool extractBracketExpression(const char *text, char b1, char b2, char *buffer)
{
	if(!numberOf(b1,text) || !numberOf(b2,text))
		return false;
	char character;
	const char *charptr=text;
	do
	{
		character=*charptr++;
		if(character=='\0')
			return false;
	}while(character!=b1);
	do
	{
		character=*charptr++;
		*buffer=character;
		buffer++;
		if(character=='\0')
			return false;
	}while(character!=b2);
	*(--buffer)='\0';
	return true;
}

bool extractBool(const char *value_string, bool *aboolptr)
{
	if(!strcmp(value_string,"yes") || !strcmp(value_string,"1") || !strcmp(value_string,"true"))
		*aboolptr=true;
	else if(!strcmp(value_string,"no") || !strcmp(value_string,"0") || !strcmp(value_string,"false"))
		*aboolptr=false;
	else
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error extracting boolian value from %s\n", value_string);
		return false;
	}
	return true;
}

char *strncpyz(char *destination, const char *source, size_t n)
{
	char *ret=strncpy(destination,source,n);
	destination[n]='\0';
	return ret;
}

void condenseLine(char *line_buffer) // removes space/tab/extra character polutions between the columns and replace them by one TAB character
{
	char buffer[10000],*char_ptr=buffer,character,*source_ptr=line_buffer;
	bool last_character_empty=true;
	do
	{
		character=*source_ptr++;
		switch(character)
		{
			case ' ':
			case '\t':
				if(last_character_empty)
					continue;
				last_character_empty=true;
				*char_ptr++='\t';
				break;
			case 10:
				break;
			default:
				*char_ptr++=character;
				if(character!=0)
					last_character_empty=false;
		}
	}while(character);	
	if(last_character_empty)
		buffer[strlen(buffer)-1]=0;
	strcpy(line_buffer,buffer);
}

bool calculateSecondsInEpoche(int *pSiE, char *date_string, char *time_string)
{
/*
  struct tm {
               int tm_sec;	// seconds
               int tm_min;	// minutes
               int tm_hour;	// hours
               int tm_mday;	// day of the month
               int tm_mon;	// month
               int tm_year;	// year
               int tm_wday;	// day of the week
               int tm_yday;	// day in the year
               int tm_isdst;	// daylight saving time
           };
*/	
	struct tm time_struct;
	memset(&time_struct, 0, sizeof(struct tm));
#ifdef WIN32
	if(sscanf(date_string, "%d.%d.%d", &time_struct.tm_mday, &time_struct.tm_mon, &time_struct.tm_year)!=3)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not extract date from \"%s\"\n", date_string);
		return false;
	}
	time_struct.tm_mday--;
	time_struct.tm_mon--;
	time_struct.tm_year+=100;
	if(sscanf(date_string, "%d.%d.%d", &time_struct.tm_hour, &time_struct.tm_min, &time_struct.tm_sec)!=3)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not extract time from \"%s\"\n", time_string);
		return false;
	}
#else
	if(!strptime(date_string, "%d.%m.%y", &time_struct))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not extract date from \"%s\"\n", date_string);
		return false;
	}
	if(!strptime(time_string, "%H:%M:%S", &time_struct))
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "could not extract time from \"%s\"\n", time_string);
		return false;
	}
#endif
	int seconds_in_epoche=mktime(&time_struct);
	if(seconds_in_epoche==-1)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fprintf(stderr, "error calculating seconds in epoche for date \"%s\" and time \"%s\"\n", date_string, time_string);
		return false;
	}
	*pSiE=seconds_in_epoche;
	return true;	
}

void formatTimeFromSecondsInEpoche(time_t SiE, char *time_string)
{
	struct tm *ptm=gmtime(&SiE);
	// strftime(time_stamp, 20, "%d.%m.%y %H:%M:%S", ptm);
	sprintf(time_string, "%02d.%02d.%02d %02d:%02d:%02d", ptm->tm_mday, ptm->tm_mon+1, ptm->tm_year-100, (ptm->tm_hour+1)%24, ptm->tm_min, ptm->tm_sec); 
}

void createListOfStrings(char *x_separated_list, char x, class list<string>& list_of_strings, const char *empty_list_string)
{
	list_of_strings.clear();
	if(!x_separated_list[0] || !strcmp(x_separated_list, empty_list_string))
		return;
	char name_buffer[100];
	int i, no_separators=numberOf(x, x_separated_list);
	condenseLine(x_separated_list);
	for(i=0;i<=no_separators;i++)
	{
		extractWordFromList(x_separated_list, i, ',', name_buffer);
		list_of_strings.push_back(string(name_buffer));
	}
}






