/**************************************************************************************************************
*
*       This file,
*
*                            /tools/makeMatrix.cpp,
*
*        is part of the publically available source code of McArtim 3.
*
*       You may not use the source code in ANY way for commercial purposes,
*       i.e. to generate profit from any derivations (modifications and/or compilations, etc.)
*       of the source code without my explicit permission. ANY commercial usage has to be allowed by me,
*       Tim Deutschmann, and will be charged with a profit share of at least 5%.
*
*       You may freely study, modify, compile and redistribute the following source code
*
*	                     FOR SCIENTIFIC PURPOSES ONLY.
*
*       The modified source code has to be made publically available at the latest on the day of
*       publishing the effects of the modification. All derivations of this source code shall remain
*       to be 'open source'.
*	
*       If you use the source code or the programs compiled from the original or
*       from the modified files you must properly cite the author of the original code by refering
*       to the following publications:
*	
*
*       Tim Deutschmann: Atmospheric Radiative Transfer Modelling with Monte Carlo Methods,
*       master thesis, University of Heidelberg, 2009.
*
*       Deutschmann, T., Beirle, S., Frieß, U., Grzegorski, M., Kern, C., Kritten, L.,
*       Platt, U., Prados-Román, C., Pukı̄te, J., Wagner, T., Werner, B., and Pfeilsticker, K.:
*       The Monte Carlo Atmospheric Radiative Transfer Model McArtim:
*       Introduction and Validation of Jacobians and 3D Features,
*       Journal of Quantitative Spectroscopy and Radiative Transfer, 112(6):1119-1137, 2011.
*
*       Tim Deutschmann: On Modeling Elastic and Inelastic Polarized Radiation Transport in the
*       Earth Atmosphere with Monte Carlo Methods, PhD thesis, University of Leipzig, 2014.
*
*       The official webpage of the McArtim source code 3.0 is
*
*                        http://www.tim-deutschmann.de/McArtim/index.html
*       
*       You may contact me through Tim.Deutschmann@posteo.de.
*
*       Have fun with the code!
*
*
*
*	                                                 Tim Deutschmann, Dossenheim, May 11th 2018.
*
**********************************************************************************************************/

#include "../stringio.h"
#include "../fileio.h"
#include "../error.h"

#include <stdlib.h>

#include <list>
#include <string>

using namespace std;

bool listHasEntry(list<string>& clist, const string& entry, int& index)
{
	int i=0;
	list<string>::iterator it;
	for(it=clist.begin();it!=clist.end();it++)
	{
		if(!it->compare(entry))
		{
			index=i;
			return true;
		}
		i++;
	}
	return false;
}

int main(int argc, char *argv[])
{
	if(argc!=5)
	{
		printf("usage: makeMatrix file x_colum y_column r_column\n");
		return 0;
	}
	char *file_name=argv[1];
	int xc=atoi(argv[2])-1;
	int yc=atoi(argv[3])-1;
	int rc=atoi(argv[4])-1;
	// open file
	FILE *fptr=fopen(file_name, "r");
	if(!fptr)
	{
		errorIn(__FILE__, __LINE__, __func__);
		return -1;
	}
	// go
	int i, j, n, no_rows=0; // =numberOfRows(fptr);
	// coordinate lists
	list<string> xlist, ylist;
	int xindex, yindex;
	
	// determine buffer size
	int buffer_size=0, c;
	do
	{
		n=0;
		//
		while((c=fgetc(fptr))!='\n' && c!=EOF && c!='\r') n++;
		if(n>buffer_size)
			buffer_size=n;
		no_rows++;
	}while(c!=EOF);
	no_rows--;

	buffer_size+=2;	// to be sure :)
	char *line_buffer=new char [2*buffer_size];
	char *name_buffer=line_buffer+buffer_size;
	if(!line_buffer)
	{
		errorIn(__FILE__, __LINE__, __func__);
		fclose(fptr);
		return -1;
	}

	// ---------------------------------------------------
	rewind(fptr);
	// ---------------------------------------------------

	//  read file line by line
	// determine dimension of the matrix
	string name;
	for(i=0;i<no_rows;i++)
	{
		if(fgets(line_buffer, buffer_size, fptr)!=line_buffer)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d of file \"%s\"!\n", i+1, file_name);
			delete [] line_buffer;
			fclose(fptr);
			return -1;
		}
		// read x column
		extractWordFromList(line_buffer, xc, '\t', name_buffer);
		name=string(name_buffer);
		if(!listHasEntry(xlist, name, xindex))
			xlist.push_back(name);
		// read y column
		extractWordFromList(line_buffer, yc, '\t', name_buffer);
		name=string(name_buffer);
		if(!listHasEntry(ylist, name, yindex))
			ylist.push_back(name);
	}
	// allocate result array
	int Nx=xlist.size(), Ny=ylist.size();
	double *result_data=new double [Nx*Ny+Nx+Ny];
	double **result=new double *[Ny];
	if(!result_data || !result)
	{
		errorIn(__FILE__, __LINE__, __func__);
		delete [] line_buffer;
		fclose(fptr);
		return -1;
	}
	double *x=result_data+Nx*Ny, *y=x+Nx;


	// adjust pointers
	for(i=0;i<Ny;i++)
		result[i]=result_data+i*Nx;


	// convert coordinate lists
	list<string>::iterator it;

	// convert xlist to x
	i=0;
	for(it=xlist.begin();it!=xlist.end();it++)
		x[i++]=atof(it->c_str());	

	// convert ylist to y
	i=0;
	for(it=ylist.begin();it!=ylist.end();it++)
		y[i++]=atof(it->c_str());	
	
	// ---------------------------------------------------
	rewind(fptr);
	// ---------------------------------------------------

	// read-in result
	for(i=0;i<no_rows;i++)
	{
		if(fgets(line_buffer, buffer_size, fptr)!=line_buffer)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d of file \"%s\"!\n", i+1, file_name);
			delete [] line_buffer;
			delete [] result_data;
			fclose(fptr);
			return -1;
		}

		// read x column
		extractWordFromList(line_buffer, xc, '\t', name_buffer);
		name=string(name_buffer);
		if(!listHasEntry(xlist, name, xindex) || xindex<0 || xindex>=Nx)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d of file \"%s\"!\n", i+1, file_name);
			fprintf(stderr, "xvalue \"%s\" not in coordinate list!\n", name_buffer);
			delete [] line_buffer;
			delete [] result_data;
			delete [] result;
			fclose(fptr);
			return -1;
		}

		// read y column
		extractWordFromList(line_buffer, yc, '\t', name_buffer);
		name=string(name_buffer);
		if(!listHasEntry(ylist, name, yindex) || yindex<0 || yindex>=Ny)
		{
			errorIn(__FILE__, __LINE__, __func__);
			fprintf(stderr, "error in line %d of file \"%s\"!\n", i+1, file_name);
			fprintf(stderr, "yvalue \"%s\" not in coordinate list!\n", name_buffer);
			delete [] line_buffer;
			delete [] result_data;
			delete [] result;
			fclose(fptr);
			return -1;
		}

		// read result
		extractWordFromList(line_buffer, rc, '\t', name_buffer);
		result[yindex][xindex]=atof(name_buffer);
	}
	fclose(fptr);

	// output
	printf("0");
	for(j=0;j<Nx;j++)
		printf("\t%le", x[j]);
	fputc('\n', stdout);
	
	for(i=0;i<Ny;i++)
	{
		printf("%le", y[i]);
		for(j=0;j<Nx;j++)
			printf("\t%le", result[i][j]);
		fputc('\n', stdout);
	}
	// the end :) 
	delete [] line_buffer;
	delete [] result_data;
	delete [] result;
	return 0;
}










